/* Riot v4.0.0-alpha.0, @license MIT */
(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory() :
  typeof define === 'function' && define.amd ? define(factory) :
  (global = global || self, global.riot = factory());
}(this, function () { 'use strict';

  const
    COMPONENTS_IMPLEMENTATION_MAP = new Map(),
    COMPONENTS_CREATION_MAP = new WeakMap(),
    MIXINS_MAP = new Map(),
    IS_DIRECTIVE = 'is';

  var globals = /*#__PURE__*/Object.freeze({
    COMPONENTS_IMPLEMENTATION_MAP: COMPONENTS_IMPLEMENTATION_MAP,
    COMPONENTS_CREATION_MAP: COMPONENTS_CREATION_MAP,
    MIXINS_MAP: MIXINS_MAP,
    IS_DIRECTIVE: IS_DIRECTIVE
  });

  /**
   * Quick type checking
   * @param   {*} element - anything
   * @param   {string} type - type definition
   * @returns {boolean} true if the type corresponds
   */
  function checkType(element, type) {
    return typeof element === type
  }

  /**
   * Check that will be passed if its argument is a function
   * @param   {*} value - value to check
   * @returns {boolean} - true if the value is a function
   */
  function isFunction(value) {
    return checkType(value, 'function')
  }

  /**
   * Check that will be passed if its argument is a string
   * @param   {*} value - value to check
   * @returns {boolean} - true if the value is a string
   */
  function isString(value) {
    return checkType(value, 'string')
  }

  /* eslint-disable fp/no-mutating-methods */
  /**
   * Throw an error
   * @param {string} error - error message
   * @returns {undefined} it's a IO void function
   */
  function panic(error) {
    throw new Error(error)
  }

  /**
   * Call the first argument received only if it's a function otherwise return it as it is
   * @param   {*} source - anything
   * @returns {*} anything
   */
  function callOrAssign(source) {
    return isFunction(source) ? source() : source
  }

  // doese simply nothing
  function noop() {
    return this
  }

  /**
   * Autobind the methods of a source object to itself
   * @param   {Object} source - probably a riot tag instance
   * @param   {Array<string>} methods - list of the methods to autobind
   * @returns {Object} the original object received
   */
  function autobindMethods(source, methods) {
    methods.forEach(method => {
      source[method] = source[method].bind(source);
    });

    return source
  }

  /**
   * Helper function to set an immutable property
   * @param   {Object} source - object where the new property will be set
   * @param   {string} key - object key where the new property will be stored
   * @param   {*} value - value of the new property
   * @param   {Object} options - set the propery overriding the default options
   * @returns {Object} - the original object modified
   */
  function defineProperty(source, key, value, options = {}) {
    Object.defineProperty(source, key, {
      value,
      enumerable: false,
      writable: false,
      configurable: true,
      ...options
    });

    return source
  }

  /**
   * Define multiple properties on a target object
   * @param   {Object} source - object where the new properties will be set
   * @param   {Object} properties - object containing as key pair the key + value properties
   * @param   {Object} options - set the propery overriding the default options
   * @returns {Object} the original object modified
   */
  function defineProperties(source, properties, options) {
    Object.entries(properties).forEach(([key, value]) => {
      defineProperty(source, key, value, options);
    });

    return source
  }

  /**
   * Evaluate a list of attribute expressions
   * @param   {Array} attributes - attribute expressions generated by the riot compiler
   * @param   {Object} scope - current scope
   * @returns {Object} key value pairs with the result of the computation
   */
  function evaluateAttributeExpressions(attributes, scope) {
    return attributes.reduce((acc, attribute) => {
      const value = attribute.evaluate(scope);

      if (attribute.name) {
        acc[attribute.name] = value;
      } else {
        Object.assign(acc, value);
      }

      return acc
    }, {})
  }

  /**
   * Shorter and fast way to select multiple nodes in the DOM
   * @param   {string} selector - DOM selector
   * @param   {Object} context - DOM node where the targets of our search will is located
   * @returns {Array} dom nodes found
   */
  function $$(selector, context) {
    if (isString(selector)) return Array.from((context || document).querySelectorAll(selector))
    return domToArray(selector)
  }

  /**
   * Select a single DOM element
   * @param   {string} selector - DOM selector
   * @param   {Object} context - DOM node where the targets of our search will is located
   * @returns {HTMLElement} DOM node found
   */
  function $(selector, context) {
    if (isString(selector)) return (context || document).querySelector(selector)
    return selector
  }

  /**
   * Get the document window
   * @returns {Object} window object
   */
  function getWindow() {
    return typeof window === 'undefined' ? /* istanbul ignore next */ undefined : window
  }

  /**
   * Converts any DOM node/s to a loopable array
   * @param   { HTMLElement|NodeList } els - single html element or a node list
   * @returns { Array } always a loopable object
   */
  function domToArray(els) {
    // can this object be already looped?
    if (!Array.isArray(els)) {
      // is it a node list?
      if (
        /^\[object (HTMLCollection|NodeList|Object)\]$/
          .test(Object.prototype.toString.call(els))
          && typeof els.length === 'number'
      )
        return Array.from(els)
      else
        // if it's a single node
        // it will be returned as "array" with one single entry
        return [els]
    }
    // this object could be looped out of the box
    return els
  }

  /**
   * Get the value of any DOM attribute on a node
   * @param   {HTMLElement} element - DOM node we want to inspect
   * @param   {string} name - name of the attribute we want to get
   * @returns {string|undefined} the node attribute if it exists
   */
  function getAttribute(element, name) {
    return element.getAttribute(name)
  }

  /**
   * Set the value of any DOM attribute
   * @param   {HTMLElement} element - DOM node we to update
   * @param   {string} name - name of the attribute we want to set
   * @param   {string} value - the value of the atribute to set
   * @returns {undefined} void function
   */
  function setAttribute(element, name, value) {
    if (isString(value)) {
      element.setAttribute(name, value);
    }
  }

  /**
   * Get all the element attributes as object
   * @param   {HTMLElement} element - DOM node we want to parse
   * @returns {Object} all the attributes found as a key value pairs
   */
  function getAttributes(element) {
    return Array.from(element.attributes).reduce((acc, attribute) => {
      acc[attribute.name] = attribute.value;
      return acc
    }, {})
  }

  /**
   * Set multiple DOM attributes
   * @param   {HTMLElement} element target element
   * @param   {Object} attributes - object containing the attributes key values
   * @returns {HTMLElement} - the original element received
   */
  function setAttributes(element, attributes) {
    Object.entries(attributes).forEach(([key, value]) => {
      setAttribute(element, key, value);
    });
    return element
  }

  /**
   * Get the tag name of any DOM node
   * @param   {HTMLElement} element - DOM node we want to inspect
   * @returns {string} name to identify this dom node in riot
   */
  function getName(element) {
    return getAttribute(element, IS_DIRECTIVE) || element.tagName.toLowerCase()
  }

  /**
   * Remove the child nodes from any DOM node
   * @param   {HTMLElement} node - target node
   * @returns {undefined}
   */
  function cleanNode(node) {
    const children = node.childNodes;
    children.forEach(n => node.removeChild(n));
  }

  const EACH = 0;
  const IF = 1;
  const SIMPLE = 2;
  const TAG = 3;

  var bindingTypes = {
    EACH,
    IF,
    SIMPLE,
    TAG
  };

  /*! (c) Andrea Giammarchi - ISC */
  var self$1 = null || /* istanbul ignore next */ {};
  try { self$1.Map = Map; }
  catch (Map) {
    self$1.Map = function Map() {
      var i = 0;
      var k = [];
      var v = [];
      return {
        delete: function (key) {
          var had = contains(key);
          if (had) {
            k.splice(i, 1);
            v.splice(i, 1);
          }
          return had;
        },
        get: function get(key) {
          return contains(key) ? v[i] : void 0;
        },
        has: function has(key) {
          return contains(key);
        },
        set: function set(key, value) {
          v[contains(key) ? i : (k.push(key) - 1)] = value;
          return this;
        }
      };
      function contains(v) {
        i = k.indexOf(v);
        return -1 < i;
      }
    };
  }
  var Map$1 = self$1.Map;

  const append = (get, parent, children, start, end, before) => {
    if ((end - start) < 2)
      parent.insertBefore(get(children[start], 1), before);
    else {
      const fragment = parent.ownerDocument.createDocumentFragment();
      while (start < end)
        fragment.appendChild(get(children[start++], 1));
      parent.insertBefore(fragment, before);
    }
  };

  const eqeq = (a, b) => a == b;

  const identity = O => O;

  const indexOf = (
    moreNodes,
    moreStart,
    moreEnd,
    lessNodes,
    lessStart,
    lessEnd,
    compare
  ) => {
    const length = lessEnd - lessStart;
    /* istanbul ignore if */
    if (length < 1)
      return -1;
    while ((moreEnd - moreStart) >= length) {
      let m = moreStart;
      let l = lessStart;
      while (
        m < moreEnd &&
        l < lessEnd &&
        compare(moreNodes[m], lessNodes[l])
      ) {
        m++;
        l++;
      }
      if (l === lessEnd)
        return moreStart;
      moreStart = m + 1;
    }
    return -1;
  };

  const isReversed = (
    futureNodes,
    futureEnd,
    currentNodes,
    currentStart,
    currentEnd,
    compare
  ) => {
    while (
      currentStart < currentEnd &&
      compare(
        currentNodes[currentStart],
        futureNodes[futureEnd - 1]
      )) {
        currentStart++;
        futureEnd--;
      }  return futureEnd === 0;
  };

  const next = (get, list, i, length, before) => i < length ?
                get(list[i], 0) :
                (0 < i ?
                  get(list[i - 1], -0).nextSibling :
                  before);

  const remove = (get, parent, children, start, end) => {
    if ((end - start) < 2)
      parent.removeChild(get(children[start], -1));
    else {
      const range = parent.ownerDocument.createRange();
      range.setStartBefore(get(children[start], -1));
      range.setEndAfter(get(children[end - 1], -1));
      range.deleteContents();
    }
  };

  // - - - - - - - - - - - - - - - - - - -
  // diff related constants and utilities
  // - - - - - - - - - - - - - - - - - - -

  const DELETION = -1;
  const INSERTION = 1;
  const SKIP = 0;
  const SKIP_OND = 50;

  const HS = (
    futureNodes,
    futureStart,
    futureEnd,
    futureChanges,
    currentNodes,
    currentStart,
    currentEnd,
    currentChanges
  ) => {

    let k = 0;
    /* istanbul ignore next */
    let minLen = futureChanges < currentChanges ? futureChanges : currentChanges;
    const link = Array(minLen++);
    const tresh = Array(minLen);
    tresh[0] = -1;

    for (let i = 1; i < minLen; i++)
      tresh[i] = currentEnd;

    const keymap = new Map$1;
    for (let i = currentStart; i < currentEnd; i++)
      keymap.set(currentNodes[i], i);

    for (let i = futureStart; i < futureEnd; i++) {
      const idxInOld = keymap.get(futureNodes[i]);
      if (idxInOld != null) {
        k = findK(tresh, minLen, idxInOld);
        /* istanbul ignore else */
        if (-1 < k) {
          tresh[k] = idxInOld;
          link[k] = {
            newi: i,
            oldi: idxInOld,
            prev: link[k - 1]
          };
        }
      }
    }

    k = --minLen;
    --currentEnd;
    while (tresh[k] > currentEnd) --k;

    minLen = currentChanges + futureChanges - k;
    const diff = Array(minLen);
    let ptr = link[k];
    --futureEnd;
    while (ptr) {
      const {newi, oldi} = ptr;
      while (futureEnd > newi) {
        diff[--minLen] = INSERTION;
        --futureEnd;
      }
      while (currentEnd > oldi) {
        diff[--minLen] = DELETION;
        --currentEnd;
      }
      diff[--minLen] = SKIP;
      --futureEnd;
      --currentEnd;
      ptr = ptr.prev;
    }
    while (futureEnd >= futureStart) {
      diff[--minLen] = INSERTION;
      --futureEnd;
    }
    while (currentEnd >= currentStart) {
      diff[--minLen] = DELETION;
      --currentEnd;
    }
    return diff;
  };

  // this is pretty much the same petit-dom code without the delete map part
  // https://github.com/yelouafi/petit-dom/blob/bd6f5c919b5ae5297be01612c524c40be45f14a7/src/vdom.js#L556-L561
  const OND = (
    futureNodes,
    futureStart,
    rows,
    currentNodes,
    currentStart,
    cols,
    compare
  ) => {
    const length = rows + cols;
    const v = [];
    let d, k, r, c, pv, cv, pd;
    outer: for (d = 0; d <= length; d++) {
      /* istanbul ignore if */
      if (d > SKIP_OND)
        return null;
      pd = d - 1;
      /* istanbul ignore next */
      pv = d ? v[d - 1] : [0, 0];
      cv = v[d] = [];
      for (k = -d; k <= d; k += 2) {
        if (k === -d || (k !== d && pv[pd + k - 1] < pv[pd + k + 1])) {
          c = pv[pd + k + 1];
        } else {
          c = pv[pd + k - 1] + 1;
        }
        r = c - k;
        while (
          c < cols &&
          r < rows &&
          compare(
            currentNodes[currentStart + c],
            futureNodes[futureStart + r]
          )
        ) {
          c++;
          r++;
        }
        if (c === cols && r === rows) {
          break outer;
        }
        cv[d + k] = c;
      }
    }

    const diff = Array(d / 2 + length / 2);
    let diffIdx = diff.length - 1;
    for (d = v.length - 1; d >= 0; d--) {
      while (
        c > 0 &&
        r > 0 &&
        compare(
          currentNodes[currentStart + c - 1],
          futureNodes[futureStart + r - 1]
        )
      ) {
        // diagonal edge = equality
        diff[diffIdx--] = SKIP;
        c--;
        r--;
      }
      if (!d)
        break;
      pd = d - 1;
      /* istanbul ignore next */
      pv = d ? v[d - 1] : [0, 0];
      k = c - r;
      if (k === -d || (k !== d && pv[pd + k - 1] < pv[pd + k + 1])) {
        // vertical edge = insertion
        r--;
        diff[diffIdx--] = INSERTION;
      } else {
        // horizontal edge = deletion
        c--;
        diff[diffIdx--] = DELETION;
      }
    }
    return diff;
  };

  const applyDiff = (
    diff,
    get,
    parentNode,
    futureNodes,
    futureStart,
    currentNodes,
    currentStart,
    currentLength,
    before
  ) => {
    const live = new Map$1;
    const length = diff.length;
    let currentIndex = currentStart;
    let i = 0;
    while (i < length) {
      switch (diff[i++]) {
        case SKIP:
          futureStart++;
          currentIndex++;
          break;
        case INSERTION:
          // TODO: bulk appends for sequential nodes
          live.set(futureNodes[futureStart], 1);
          append(
            get,
            parentNode,
            futureNodes,
            futureStart++,
            futureStart,
            currentIndex < currentLength ?
              get(currentNodes[currentIndex], 1) :
              before
          );
          break;
        case DELETION:
          currentIndex++;
          break;
      }
    }
    i = 0;
    while (i < length) {
      switch (diff[i++]) {
        case SKIP:
          currentStart++;
          break;
        case DELETION:
          // TODO: bulk removes for sequential nodes
          if (live.has(currentNodes[currentStart]))
            currentStart++;
          else
            remove(
              get,
              parentNode,
              currentNodes,
              currentStart++,
              currentStart
            );
          break;
      }
    }
  };

  const findK = (ktr, length, j) => {
    let lo = 1;
    let hi = length;
    while (lo < hi) {
      const mid = ((lo + hi) / 2) >>> 0;
      if (j < ktr[mid])
        hi = mid;
      else
        lo = mid + 1;
    }
    return lo;
  };

  const smartDiff = (
    get,
    parentNode,
    futureNodes,
    futureStart,
    futureEnd,
    futureChanges,
    currentNodes,
    currentStart,
    currentEnd,
    currentChanges,
    currentLength,
    compare,
    before
  ) => {
    applyDiff(
      OND(
        futureNodes,
        futureStart,
        futureChanges,
        currentNodes,
        currentStart,
        currentChanges,
        compare
      ) ||
      HS(
        futureNodes,
        futureStart,
        futureEnd,
        futureChanges,
        currentNodes,
        currentStart,
        currentEnd,
        currentChanges
      ),
      get,
      parentNode,
      futureNodes,
      futureStart,
      currentNodes,
      currentStart,
      currentLength,
      before
    );
  };

  /*! (c) 2018 Andrea Giammarchi (ISC) */

  const domdiff = (
    parentNode,     // where changes happen
    currentNodes,   // Array of current items/nodes
    futureNodes,    // Array of future items/nodes
    options         // optional object with one of the following properties
                    //  before: domNode
                    //  compare(generic, generic) => true if same generic
                    //  node(generic) => Node
  ) => {
    if (!options)
      options = {};

    const compare = options.compare || eqeq;
    const get = options.node || identity;
    const before = options.before == null ? null : get(options.before, 0);

    const currentLength = currentNodes.length;
    let currentEnd = currentLength;
    let currentStart = 0;

    let futureEnd = futureNodes.length;
    let futureStart = 0;

    // common prefix
    while (
      currentStart < currentEnd &&
      futureStart < futureEnd &&
      compare(currentNodes[currentStart], futureNodes[futureStart])
    ) {
      currentStart++;
      futureStart++;
    }

    // common suffix
    while (
      currentStart < currentEnd &&
      futureStart < futureEnd &&
      compare(currentNodes[currentEnd - 1], futureNodes[futureEnd - 1])
    ) {
      currentEnd--;
      futureEnd--;
    }

    const currentSame = currentStart === currentEnd;
    const futureSame = futureStart === futureEnd;

    // same list
    if (currentSame && futureSame)
      return futureNodes;

    // only stuff to add
    if (currentSame && futureStart < futureEnd) {
      append(
        get,
        parentNode,
        futureNodes,
        futureStart,
        futureEnd,
        next(get, currentNodes, currentStart, currentLength, before)
      );
      return futureNodes;
    }

    // only stuff to remove
    if (futureSame && currentStart < currentEnd) {
      remove(
        get,
        parentNode,
        currentNodes,
        currentStart,
        currentEnd
      );
      return futureNodes;
    }

    const currentChanges = currentEnd - currentStart;
    const futureChanges = futureEnd - futureStart;
    let i = -1;

    // 2 simple indels: the shortest sequence is a subsequence of the longest
    if (currentChanges < futureChanges) {
      i = indexOf(
        futureNodes,
        futureStart,
        futureEnd,
        currentNodes,
        currentStart,
        currentEnd,
        compare
      );
      // inner diff
      if (-1 < i) {
        append(
          get,
          parentNode,
          futureNodes,
          futureStart,
          i,
          get(currentNodes[currentStart], 0)
        );
        append(
          get,
          parentNode,
          futureNodes,
          i + currentChanges,
          futureEnd,
          next(get, currentNodes, currentEnd, currentLength, before)
        );
        return futureNodes;
      }
    }
    /* istanbul ignore else */
    else if (futureChanges < currentChanges) {
      i = indexOf(
        currentNodes,
        currentStart,
        currentEnd,
        futureNodes,
        futureStart,
        futureEnd,
        compare
      );
      // outer diff
      if (-1 < i) {
        remove(
          get,
          parentNode,
          currentNodes,
          currentStart,
          i
        );
        remove(
          get,
          parentNode,
          currentNodes,
          i + futureChanges,
          currentEnd
        );
        return futureNodes;
      }
    }

    // common case with one replacement for many nodes
    // or many nodes replaced for a single one
    /* istanbul ignore else */
    if ((currentChanges < 2 || futureChanges < 2)) {
      append(
        get,
        parentNode,
        futureNodes,
        futureStart,
        futureEnd,
        get(currentNodes[currentStart], 0)
      );
      remove(
        get,
        parentNode,
        currentNodes,
        currentStart,
        currentEnd
      );
      return futureNodes;
    }

    // the half match diff part has been skipped in petit-dom
    // https://github.com/yelouafi/petit-dom/blob/bd6f5c919b5ae5297be01612c524c40be45f14a7/src/vdom.js#L391-L397
    // accordingly, I think it's safe to skip in here too
    // if one day it'll come out like the speediest thing ever to do
    // then I might add it in here too

    // Extra: before going too fancy, what about reversed lists ?
    //        This should bail out pretty quickly if that's not the case.
    if (
      currentChanges === futureChanges &&
      isReversed(
        futureNodes,
        futureEnd,
        currentNodes,
        currentStart,
        currentEnd,
        compare
      )
    ) {
      append(
        get,
        parentNode,
        futureNodes,
        futureStart,
        futureEnd,
        next(get, currentNodes, currentEnd, currentLength, before)
      );
      return futureNodes;
    }

    // last resort through a smart diff
    smartDiff(
      get,
      parentNode,
      futureNodes,
      futureStart,
      futureEnd,
      futureChanges,
      currentNodes,
      currentStart,
      currentEnd,
      currentChanges,
      currentLength,
      compare,
      before
    );

    return futureNodes;
  };

  /**
   * Safe expression/bindings value evaluation, in case of errors we return a fallback value
   * @param   {Function} fn  - function to evaluate
   * @param   {*}        fallback - a fallback return value
   * @param   {boolean}  debug - if true the error will be logged
   * @returns {*} result of the computation or a fallback value
   */

  function evalOrFallback(fn, fallback, debug) {
    try {
      return fn()
    } catch (error) {
      if (debug) {
        console.error(debug); // eslint-disable-line
      }

      return fallback
    }
  }

  const EachBinding = Object.seal({
    // dynamic binding properties
    childrenMap: null,
    node: null,
    root: null,
    condition: null,
    evaluate: null,
    template: null,
    tags: [],
    getKey: null,
    indexName: null,
    itemName: null,
    afterPlaceholder: null,
    placeholder: null,

    // API methods
    mount(scope) {
      return this.update(scope)
    },
    update(scope) {
      const { placeholder } = this;
      const collection = evalOrFallback(() => this.evaluate(scope), []);
      const items = collection ? Array.from(collection) : [];
      const parent = placeholder.parentNode;

      // prepare the diffing
      const { newChildrenMap, batches, futureNodes } = loopItems(items, scope, this);

      /**
       * DOM Updates
       */
      const before = this.tags[this.tags.length - 1];
      domdiff(parent, this.tags, futureNodes, {
        before: before ? before.nextSibling : placeholder.nextSibling
      });

      // trigger the mounts and the updates
      batches.forEach(fn => fn());

      // update the children map
      this.childrenMap = newChildrenMap;
      this.tags = futureNodes;

      return this
    },
    unmount() {
      Array
        .from(this.childrenMap.values())
        .forEach(({tag, context}) => {
          tag.unmount(context, true);
        });

      this.childrenMap = new Map();
      this.tags = [];

      return this
    }
  });

  /**
   * Check whether a tag must be filtered from a loop
   * @param   {Function} condition - filter function
   * @param   {Object} context - argument passed to the filter function
   * @returns {boolean} true if this item should be skipped
   */
  function mustFilterItem(condition, context) {
    return condition ? (condition(context) === false) : false
  }

  /**
   * Get the context of the looped tag
   * @param   {string} options.itemName - key to identify the looped item in the new context
   * @param   {string} options.indexName - key to identify the index of the looped item
   * @param   {number} options.index - current index
   * @param   {*} options.item - collection item looped
   * @param   {*} options.scope - current parent scope
   * @returns {Object} enhanced scope object
   */
  function getContext({itemName, indexName, index, item, scope}) {
    const context = {
      [itemName]: item,
      ...scope
    };

    if (indexName) {
      return {
        [indexName]: index,
        ...context
      }
    }

    return context
  }


  /**
   * Loop the current tag items
   * @param   { Array } items - tag collection
   * @param   { * } scope - tag scope
   * @param   { EeachBinding } binding - each binding object instance
   * @returns { Object } data
   * @returns { Map } data.newChildrenMap - a Map containing the new children tags structure
   * @returns { Array } data.batches - array containing functions the tags lifecycle functions to trigger
   * @returns { Array } data.futureNodes - array containing the nodes we need to diff
   */
  function loopItems(items, scope, binding) {
    const { condition, template, childrenMap, itemName, getKey, indexName, root } = binding;
    const filteredItems = new Set();
    const newChildrenMap = new Map();
    const batches = [];
    const futureNodes = [];

    items.forEach((item, i) => {
      // the real item index should be subtracted to the items that were filtered
      const index = i - filteredItems.size;
      const context = getContext({itemName, indexName, index, item, scope});
      const key = getKey ? getKey(context) : index;
      const oldItem = childrenMap.get(key);

      if (mustFilterItem(condition, context)) {
        filteredItems.add(oldItem);
        return
      }

      const tag = oldItem ? oldItem.tag : template.clone(root);
      const el = oldItem ? tag.el : root.cloneNode();

      if (!oldItem) {
        batches.push(() => tag.mount(el, context));
      } else {
        batches.push(() => tag.update(context));
      }

      futureNodes.push(el);

      // update the children map
      newChildrenMap.set(key, {
        tag,
        context,
        index
      });
    });

    return {
      newChildrenMap,
      batches,
      futureNodes
    }
  }

  function create(node, { evaluate, condition, itemName, indexName, getKey, template }) {
    const placeholder = document.createTextNode('');
    const parent = node.parentNode;
    const root = node.cloneNode();
    const offset = Array.from(parent.childNodes).indexOf(node);

    parent.insertBefore(placeholder, node);
    parent.removeChild(node);

    return {
      ...EachBinding,
      childrenMap: new Map(),
      node,
      root,
      offset,
      condition,
      evaluate,
      template,
      getKey,
      indexName,
      itemName,
      placeholder
    }
  }

  /**
   * Binding responsible for the `if` directive
   */
  const IfBinding = Object.seal({
    // dynamic binding properties
    node: null,
    evaluate: null,
    placeholder: null,
    template: '',

    // API methods
    mount(scope) {
      swap(this.placeholder, this.node);
      return this.update(scope)
    },
    update(scope) {
      const value = !!evalOrFallback(() => this.evaluate(scope), false);
      const mustMount = !this.value && value;
      const mustUnmount = this.value && !value;

      switch (true) {
      case mustMount:
        swap(this.node, this.placeholder);
        if (this.template) {
          this.template = this.template.clone(this.node);
          this.template.mount(this.node, scope);
        }
        break
      case mustUnmount:
        swap(this.placeholder, this.node);
        this.unmount(scope);
        break
      default:
        if (value) this.template.update(scope);
      }

      this.value = value;

      return this
    },
    unmount(scope) {
      const { template } = this;

      if (template) {
        template.unmount(scope);
      }

      return this
    }
  });

  function swap(inNode, outNode) {
    const parent = outNode.parentNode;
    parent.insertBefore(inNode, outNode);
    parent.removeChild(outNode);
  }

  function create$1(node, { evaluate, template }) {
    return {
      ...IfBinding,
      node,
      evaluate,
      placeholder: document.createTextNode(''),
      template
    }
  }

  const ATTRIBUTE = 0;
  const EVENT = 1;
  const TEXT = 2;
  const VALUE = 3;

  var expressionTypes = {
    ATTRIBUTE,
    EVENT,
    TEXT,
    VALUE
  };

  const REMOVE_ATTRIBUTE = 'removeAttribute';
  const SET_ATTIBUTE = 'setAttribute';

  /**
   * Add all the attributes provided
   * @param   {HTMLElement} node - target node
   * @param   {Object} attributes - object containing the attributes names and values
   * @returns {undefined} sorry it's a void function :(
   */
  function setAllAttributes(node, attributes) {
    Object
      .entries(attributes)
      .forEach(([name, value]) => attributeExpression(node, { name }, value));
  }

  /**
   * Remove all the attributes provided
   * @param   {HTMLElement} node - target node
   * @param   {Object} attributes - object containing all the attribute names
   * @returns {undefined} sorry it's a void function :(
   */
  function removeAllAttributes(node, attributes) {
    Object
      .keys(attributes)
      .forEach(attribute => node.removeAttribute(attribute));
  }

  /**
   * This methods handles the DOM attributes updates
   * @param   {HTMLElement} node - target node
   * @param   {Object} expression - expression object
   * @param   {string} expression.name - attribute name
   * @param   {*} value - new expression value
   * @param   {*} oldValue - the old expression cached value
   * @returns {undefined}
   */
  function attributeExpression(node, { name }, value, oldValue) {
    // is it a spread operator? {...attributes}
    if (!name) {
      // is the value still truthy?
      if (value) {
        setAllAttributes(node, value);
      } else if (oldValue) {
        // otherwise remove all the old attributes
        removeAllAttributes(node, oldValue);
      }

      return
    }

    // handle boolean attributes
    if (typeof value === 'boolean') {
      node[name] = value;
    }

    node[getMethod(value)](name, normalizeValue(name, value));
  }

  /**
   * Get the attribute modifier method
   * @param   {*} value - if truthy we return `setAttribute` othewise `removeAttribute`
   * @returns {string} the node attribute modifier method name
   */
  function getMethod(value) {
    return value ? SET_ATTIBUTE : REMOVE_ATTRIBUTE
  }

  /**
   * Get the value as string
   * @param   {string} name - attribute name
   * @param   {*} value - user input value
   * @returns {string} input value as string
   */
  function normalizeValue(name, value) {
    // be sure that expressions like selected={ true } will be always rendered as selected='selected'
    if (value === true) return name

    // array values will be joined with spaces
    return Array.isArray(value) ? value.join(' ') : value
  }

  /**
   * Set a new event listener
   * @param   {HTMLElement} node - target node
   * @param   {Object} expression - expression object
   * @param   {string} expression.name - event name
   * @param   {*} value - new expression value
   * @returns {undefined}
   */
  function eventExpression(node, { name }, value) {
    node[name] = value;
  }

  /**
   * This methods handles a simple text expression update
   * @param   {HTMLElement} node - target node
   * @param   {Object} expression - expression object
   * @param   {number} expression.childNodeIndex - index to find the text node to update
   * @param   {*} value - new expression value
   * @returns {undefined}
   */
  function textExpression(node, { childNodeIndex }, value) {
    const target = node.childNodes[childNodeIndex];
    const val = normalizeValue$1(value);

    // replace the target if it's a placeholder comment
    if (target.nodeType === Node.COMMENT_NODE) {
      const textNode = document.createTextNode(val);
      node.replaceChild(textNode, target);
    } else {
      target.data = normalizeValue$1(val);
    }
  }

  /**
   * Normalize the user value in order to render a empty string in case of falsy values
   * @param   {*} value - user input value
   * @returns {string} hopefully a string
   */
  function normalizeValue$1(value) {
    return value != null ? value : ''
  }

  /**
   * This methods handles the input fileds value updates
   * @param   {HTMLElement} node - target node
   * @param   {Object} expression - expression object
   * @param   {*} value - new expression value
   * @returns {undefined}
   */
  function valueExpression(node, expression, value) {
    node.value = value;
  }

  var expressions = {
    [ATTRIBUTE]: attributeExpression,
    [EVENT]: eventExpression,
    [TEXT]: textExpression,
    [VALUE]: valueExpression
  };

  const Expression = Object.seal({
    // Static props
    node: null,
    value: null,

    // API methods
    /**
     * Mount the expression evaluating its initial value
     * @param   {*} scope - argument passed to the expression to evaluate its current values
     * @returns {Expression} self
     */
    mount(scope) {
      // hopefully a pure function
      this.value = this.evaluate(scope);

      // IO() DOM updates
      apply(this, this.value);

      return this
    },
    /**
     * Update the expression if its value changed
     * @param   {*} scope - argument passed to the expression to evaluate its current values
     * @returns {Expression} self
     */
    update(scope) {
      // pure function
      const value = this.evaluate(scope);

      if (this.value !== value) {
        // IO() DOM updates
        apply(this, value);
        this.value = value;
      }

      return this
    },
    /**
     * Expression teardown method
     * @returns {Expression} self
     */
    unmount() {
      return this
    }
  });

  /**
   * IO() function to handle the DOM updates
   * @param {Expression} expression - expression object
   * @param {*} value - current expression value
   * @returns {undefined}
   */
  function apply(expression, value) {
    return expressions[expression.type](expression.node, expression, value, expression.value)
  }

  function create$2(node, data) {
    return {
      ...Expression,
      ...data,
      node
    }
  }

  /**
   * Create a flat object having as keys a list of methods that if dispatched will propagate
   * on the whole collection
   * @param   {Array} collection - collection to iterate
   * @param   {Array<string>} methods - methods to execute on each item of the collection
   * @param   {*} context - context returned by the new methods created
   * @returns {Object} a new object to simplify the the nested methods dispatching
   */
  function flattenCollectionMethods(collection, methods, context) {
    return methods.reduce((acc, method) => {
      return {
        ...acc,
        [method]: (scope) => {
          return collection.map(item => item[method](scope)) && context
        }
      }
    }, {})
  }

  function create$3(node, { expressions }) {
    return {
      ...flattenCollectionMethods(
        expressions.map(expression => create$2(node, expression)),
        ['mount', 'update', 'unmount']
      )
    }
  }

  /**
   * Create a new tag object if it was registered before, otherwise fallback to the simple
   * template chunk
   * @param   {Function} component - component factory function
   * @param   {Array<Object>} slots - array containing the slots markup
   * @param   {Array} attributes - dynamic attributes that will be received by the tag element
   * @returns {TagImplementation|TemplateChunk} a tag implementation or a template chunk as fallback
   */
  function getTag(component, slots = [], attributes = []) {
    // if this tag was registered before we will return its implementation
    if (component) {
      return component({ slots, attributes })
    }

    // otherwise we return a template chunk
    return create$6(slotsToMarkup(slots), [
      ...slotBindings(slots), {
      // the attributes should be registered as binding
      // if we fallback to a normal template chunk
        expressions: attributes.map(attr => {
          return {
            type: ATTRIBUTE,
            ...attr
          }
        })
      }
    ])
  }


  /**
   * Merge all the slots bindings into a single array
   * @param   {Array<Object>} slots - slots collection
   * @returns {Array<Bindings>} flatten bindings array
   */
  function slotBindings(slots) {
    return slots.reduce((acc, { bindings }) => acc.concat(bindings), [])
  }

  /**
   * Merge all the slots together in a single markup string
   * @param   {Array<Object>} slots - slots collection
   * @returns {string} markup of all the slots in a single string
   */
  function slotsToMarkup(slots) {
    return slots.reduce((acc, slot) => {
      return acc + slot.html
    }, '')
  }


  const TagBinding = Object.seal({
    // dynamic binding properties
    node: null,
    evaluate: null,
    name: null,
    slots: null,
    tag: null,
    attributes: null,
    getComponent: null,

    mount(scope) {
      return this.update(scope)
    },
    update(scope) {
      const name = this.evaluate(scope);

      // simple update
      if (name === this.name) {
        this.tag.update(scope);
      } else {
        // unmount the old tag if it exists
        if (this.tag) {
          this.tag.unmount(scope);
        }

        // mount the new tag
        this.name = name;
        this.tag = getTag(this.getComponent(name), this.slots, this.attributes);
        this.tag.mount(this.node, scope);
      }

      return this
    },
    unmount(scope) {
      if (this.tag) {
        this.tag.unmount(scope);
      }

      return this
    }
  });

  function create$4(node, { evaluate, getComponent, slots, attributes }) {
    return {
      ...TagBinding,
      node,
      evaluate,
      slots,
      attributes,
      getComponent
    }
  }

  var bindings = {
    [IF]: create$1,
    [SIMPLE]: create$3,
    [EACH]: create,
    [TAG]: create$4
  };

  /**
   * Bind a new expression object to a DOM node
   * @param   {HTMLElement} root - DOM node where to bind the expression
   * @param   {Object} binding - binding data
   * @returns {Expression} Expression object
   */
  function create$5(root, binding) {
    const { selector, type, redundantAttribute, expressions } = binding;
    // find the node to apply the bindings
    const node = selector ? root.querySelector(selector) : root;
    // remove eventually additional attributes created only to select this node
    if (redundantAttribute) node.removeAttribute(redundantAttribute);

    // init the binding
    return (bindings[type] || bindings[SIMPLE])(
      node,
      {
        ...binding,
        expressions: expressions || []
      }
    )
  }

  /**
   * Check if an element is part of an svg
   * @param   {HTMLElement}  el - element to check
   * @returns {boolean} true if we are in an svg context
   */
  function isSvg(el) {
    const owner = el.ownerSVGElement;

    return !!owner || owner === null
  }

  // in this case a simple innerHTML is enough
  function createHTMLTree(html) {
    const template = document.createElement('template');
    template.innerHTML = html;
    return template.content
  }

  // for svg nodes we need a bit more work
  function creteSVGTree(html, container) {
    // create the SVGNode
    const svgNode = container.ownerDocument.importNode(
      new window.DOMParser()
        .parseFromString(
          `<svg xmlns="http://www.w3.org/2000/svg">${html}</svg>`,
          'application/xml'
        )
        .documentElement,
      true
    );

    return svgNode
  }

  /**
   * Create the DOM that will be injected
   * @param {Object} root - DOM node to find out the context where the fragment will be created
   * @param   {string} html - DOM to create as string
   * @returns {HTMLDocumentFragment|HTMLElement} a new html fragment
   */
  function createDOMTree(root, html) {
    if (isSvg(root)) return creteSVGTree(html, root)

    return createHTMLTree(html)
  }

  /**
   * Move all the child nodes from a source tag to another
   * @param   {HTMLElement} source - source node
   * @param   {HTMLElement} target - target node
   * @returns {undefined} it's a void method ¯\_(ツ)_/¯
   */

  // Ignore this helper because it's needed only for svg tags
  /* istanbul ignore next */
  function moveChildren(source, target) {
    if (source.firstChild) {
      target.appendChild(source.firstChild);
      moveChildren(source, target);
    }
  }

  const SVG_RE = /svg/i;

  /**
   * Inject the DOM tree into a target node
   * @param   {HTMLElement} el - target element
   * @param   {HTMLFragment|SVGElement} dom - dom tree to inject
   * @returns {undefined}
   */
  function injectDOM(el, dom) {
    if (SVG_RE.test(el.tagName)) {
      moveChildren(dom, el);
    } else {
      el.appendChild(dom);
    }
  }

  /**
   * Create the Template DOM skeleton
   * @param   {HTMLElement} el - root node where the DOM will be injected
   * @param   {string} html - markup that will be injected into the root node
   * @returns {HTMLFragment} fragment that will be injected into the root node
   */
  function createTemplateDOM(el, html) {
    return html && (typeof html === 'string' ?
      createDOMTree(el, html) :
      html)
  }

  /**
   * Template Chunk model
   * @type {Object}
   */
  const TemplateChunk = Object.freeze({
    // Static props
    bindings: null,
    bindingsData: null,
    html: null,
    dom: null,
    el: null,

    // API methods
    /**
     * Attach the template to a DOM node
     * @param   {HTMLElement} el - target DOM node
     * @param   {*} scope - template data
     * @returns {TemplateChunk} self
     */
    mount(el, scope) {
      if (!el) throw new Error('Please provide DOM node to mount properly your template')

      if (this.el) this.unmount(scope);

      this.el = el;

      // create lazily the template fragment only once if it hasn't been created before
      this.dom = this.dom || createTemplateDOM(el, this.html);

      if (this.dom) injectDOM(el, this.dom.cloneNode(true));

      // create the bindings
      this.bindings = this.bindingsData.map(binding => create$5(this.el, binding));
      this.bindings.forEach(b => b.mount(scope));

      return this
    },
    /**
     * Update the template with fresh data
     * @param   {*} scope - template data
     * @returns {TemplateChunk} self
     */
    update(scope) {
      this.bindings.forEach(b => b.update(scope));

      return this
    },
    /**
     * Remove the template from the node where it was initially mounted
     * @param   {*} scope - template data
     * @param   {boolean} mustRemoveRoot - if true remove the root element
     * @returns {TemplateChunk} self
     */
    unmount(scope, mustRemoveRoot) {
      if (this.el) {
        this.bindings.forEach(b => b.unmount(scope));
        cleanNode(this.el);

        if (mustRemoveRoot) {
          this.el.parentNode.removeChild(this.el);
        }

        this.el = null;
      }

      return this
    },
    /**
     * Clone the template chunk
     * @param   {HTMLElement} el - template target DOM node
     * @returns {TemplateChunk} a clone of this object resetting the this.el property
     */
    clone(el) {
      // make sure that the DOM gets created before cloning the template
      this.dom = this.dom || createTemplateDOM(el, this.html);

      return {
        ...this,
        el: null
      }
    }
  });

  /**
   * Create a template chunk wiring also the bindings
   * @param   {string|HTMLElement} html - template string
   * @param   {Array} bindings - bindings collection
   * @returns {TemplateChunk} a new TemplateChunk copy
   */
  function create$6(html, bindings = []) {
    return {
      ...TemplateChunk,
      html,
      bindingsData: bindings
    }
  }

  /**
   * Method used to bind expressions to a DOM node
   * @param   {string|HTMLElement} html - your static template html structure
   * @param   {Array} bindings - list of the expressions to bind to update the markup
   * @returns {TemplateChunk} a new TemplateChunk object having the `update`,`mount`, `unmount` and `clone` methods
   *
   * @example
   *
   * riotDOMBindings
   *  .template(
   *   `<div expr0><!----></div><div><p expr1><!----><section expr2></section></p>`,
   *   [
   *     {
   *       selector: '[expr0]',
   *       redundantAttribute: 'expr0',
   *       expressions: [
   *         {
   *           type: expressionTypes.TEXT,
   *           childNodeIndex: 0,
   *           evaluate(scope) {
   *             return scope.time;
   *           },
   *         },
   *       ],
   *     },
   *     {
   *       selector: '[expr1]',
   *       redundantAttribute: 'expr1',
   *       expressions: [
   *         {
   *           type: expressionTypes.TEXT,
   *           childNodeIndex: 0,
   *           evaluate(scope) {
   *             return scope.name;
   *           },
   *         },
   *         {
   *           type: 'attribute',
   *           name: 'style',
   *           evaluate(scope) {
   *             return scope.style;
   *           },
   *         },
   *       ],
   *     },
   *     {
   *       selector: '[expr2]',
   *       redundantAttribute: 'expr2',
   *       type: bindingTypes.IF,
   *       evaluate(scope) {
   *         return scope.isVisible;
   *       },
   *       template: riotDOMBindings.template('hello there'),
   *     },
   *   ]
   * )
   */

  const WIN = getWindow();
  const CSS_BY_NAME = new Map();

  // skip the following code on the server
  const styleNode = WIN && ((() => {
    // create a new style element with the correct type
    const newNode = document.createElement('style');
    setAttribute(newNode, 'type', 'text/css');
    document.head.appendChild(newNode);

    return newNode
  }))();

  /**
   * Object that will be used to inject and manage the css of every tag instance
   */
  var cssManager = {
    /**
     * Save a tag style to be later injected into DOM
     * @param { string } name - if it's passed we will map the css to a tagname
     * @param { string } css - css string
     * @returns {Object} self
     */
    add(name, css) {
      CSS_BY_NAME.set(name, css);
      this.inject();
      return this
    },
    /**
     * Inject all previously saved tag styles into DOM
     * innerHTML seems slow: http://jsperf.com/riot-insert-style
     * @returns {Object} self
     */
    inject() {
      if (!styleNode) return this
      styleNode.innerHTML = [...CSS_BY_NAME.values()].join('\n');
      return this
    },

    /**
     * Remove a tag style from the DOM
     * @param {string} name a registered tagname
     * @returns {Object} self
     */
    remove(name) {
      if (CSS_BY_NAME.has(name)) {
        CSS_BY_NAME.delete(name);
        this.inject();
      }

      return this
    }
  };

  /**
   * Function to curry any javascript method
   * @param   {Function}  fn - the target function we want to curry
   * @param   {...[args]} acc - initial arguments
   * @returns {Function|*} it will return a function until the target function
   *                       will receive all of its arguments
   */
  function curry(fn, ...acc) {
    return (...args) => {
      args = [...acc, ...args];

      return args.length < fn.length ?
        curry(fn, ...args) :
        fn(...args)
    }
  }

  const COMPONENT_CORE = Object.freeze({
    // component helpers
    $(selector){ return $(selector, this.root) },
    $$(selector){ return $$(selector, this.root) },
    mixin(name) {
      // extend this component with this mixin
      Object.assing(this, MIXINS_MAP.get(name));
    },
    // defined during the component creation
    css: null,
    template: null
  });

  const COMPONENT_LIFECYCLE_METHODS = Object.freeze({
    onBeforeMount: noop,
    onMounted: noop,
    onBeforeUpdate: noop,
    onUpdated: noop,
    onBeforeUnmount: noop,
    onUnmounted: noop
  });

  const EMPTY_TEMPLATE_INTERFACE = {
    update: noop,
    mount: noop,
    unmount: noop,
    clone: noop
  };

  /**
   * Component definition function
   * @param   {Object} implementation - the componen implementation will be generated via compiler
   * @param   {Object} component - the component initial properties
   * @returns {Object} a new component implementation object
   */
  function defineComponent({css, template: template$$1, tag, name}) {
    const componentAPI = callOrAssign(tag) || {};

    // add the component css into the DOM
    if (css && name) cssManager.add(name, css);

    return curry(createComponent)(defineProperties({
      ...COMPONENT_LIFECYCLE_METHODS,
      ...componentAPI,
      // defined during the component creation
      state: {},
      props: {},
      slots: null,
      root: null
    }, {
      // these properties should not be overriden
      ...COMPONENT_CORE,
      css,
      template: template$$1 ? template$$1(
        create$6,
        expressionTypes,
        bindingTypes,
        function(name) {
          return (componentAPI.components || {})[name] || COMPONENTS_IMPLEMENTATION_MAP.get(name)
        }
      ) : EMPTY_TEMPLATE_INTERFACE
    }))
  }

  /**
   * Evaluate the component properties either from its real attributes or from its attribute expressions
   * @param   {HTMLElement} element - component root
   * @param   {Array}  attributeExpressions - attribute expressions generated by the riot compiler
   * @param   {Object} scope - current scope
   * @returns {Object} attributes key value pairs
   */
  function evaluateProps(element, attributeExpressions = [], scope) {
    return attributeExpressions.length ?
      evaluateAttributeExpressions(attributeExpressions, scope) :
      getAttributes(element)
  }

  /**
   * Component creation factory function
   * @param   {Object} component - a component implementation previously defined
   * @param   {Array} options.slots - component slots generated via riot compiler
   * @param   {Array} options.attributes - attribute expressions generated via riot compiler
   * @returns {Riot.Component} a riot component instance
   */
  function createComponent(component, {slots, attributes}) {
    // if this component was manually mounted its DOM attributes are likely not attribute expressions
    // generated via riot compiler
    const shouldSetAttributes = attributes && attributes.length;

    return autobindMethods(
      defineProperties(Object.create(component), {
        slots,
        mount(element, state = {}, props = {}) {
          this.props = evaluateProps(element, attributes, props);
          this.state = callOrAssign(state);

          defineProperties(this, {
            root: element,
            template: this.template.clone(element)
          });

          this.onBeforeMount();
          shouldSetAttributes && setAttributes(this.root, this.props);
          this.template.mount(element, this);
          this.onMounted();

          return this
        },
        update(state = {}, props = {}) {
          const newProps = evaluateProps(this.root, attributes, props);

          if (this.onBeforeUpdate(newProps, state) === false) return

          this.props = {
            ...this.props,
            ...newProps
          };

          this.state = {
            ...this.state,
            ...state
          };

          shouldSetAttributes && setAttributes(this.root, this.props);
          this.template.update(this);
          this.onUpdated();

          return this
        },
        unmount(removeRoot) {
          this.onBeforeUnmount();
          this.template.unmount(this, removeRoot === true);
          this.onUnmounted();

          return this
        }
      }),
      Object.keys(component).filter(prop => isFunction(component[prop]))
    )
  }

  /**
   * Component initialization function starting from a DOM node
   * @param   {HTMLElement} element - element to upgrade
   * @param   {Object} initialState - initial component state
   * @param   {string} componentName - component id
   * @returns {Object} a new component instance bound to a DOM node
   */
  function mountComponent(element, initialState, componentName) {
    const name = componentName || getName(element);
    if (!COMPONENTS_IMPLEMENTATION_MAP.has(name)) panic(`The component named "${name}" was never registered`);
    const component = COMPONENTS_IMPLEMENTATION_MAP.get(name)({});
    COMPONENTS_CREATION_MAP.set(element, component);

    return component.mount(element, {}, initialState)
  }

  const { COMPONENTS_CREATION_MAP: COMPONENTS_CREATION_MAP$1, COMPONENTS_IMPLEMENTATION_MAP: COMPONENTS_IMPLEMENTATION_MAP$1, MIXINS_MAP: MIXINS_MAP$1 } = globals;

  /**
   * Riot public api
   */

  /**
   * Register a custom tag by name
   * @param   {string} name - component name
   * @param   {Object} implementation - tag implementation
   * @returns {Object} object representing our tag implementation
   */
  function register(name, {css, template, tag}) {
    if (COMPONENTS_IMPLEMENTATION_MAP$1.has(name)) panic(`The component "${name}" was already registered`);

    return COMPONENTS_IMPLEMENTATION_MAP$1.set(name, (...args) => {
      const component = defineComponent({
        css,
        template,
        tag,
        name
      })(...args);

      // this object will be provided to the tag bindings generated via compiler
      // the bindings will not be able to update the components state, they will only pass down
      // the parentScope updates
      return {
        mount(element, parentScope, state) {
          return component.mount(element, state, parentScope)
        },
        update(parentScope, state) {
          return component.update(state, parentScope)
        },
        unmount() {
          return component.unmount()
        }
      }
    })
  }

  /**
   * Unregister a riot web component
   * @param   {string} name - component name
   * @returns {boolean} true if deleted
   */
  function unregister(name) {
    if (COMPONENTS_IMPLEMENTATION_MAP$1.has(name)) return COMPONENTS_IMPLEMENTATION_MAP$1.delete(name)
    cssManager.remove(name);
    return false
  }

  /**
   * Mounting function that will work only for the components that were globally registered
   * @param   {string|HTMLElement} selector - query for the selection or a DOM element
   * @param   {Object} initialState - the initial component state
   * @param   {string} name - optional component name
   * @returns {Array} list of nodes upgraded
   */
  function mount(selector, initialState, name) {
    return $$(selector).map((element) => mountComponent(element, initialState, name))
  }

  /**
   * Sweet unmounting helper function for the DOM node mounted manually by the user
   * @param   {string|HTMLElement} selector - query for the selection or a DOM element
   * @returns {Array} list of nodes unmounted
   */
  function unmount(selector) {
    return $$(selector).map((element) => {
      if (COMPONENTS_CREATION_MAP$1.has(element)) {
        COMPONENTS_CREATION_MAP$1.get(element).unmount();
      }
      return element
    })
  }

  /**
   * Define a mixin
   * @param   {string} name - mixin id
   * @param   {Object|Function} mixin - mixin logic
   * @returns {Map} the map containing all the mixins
   */
  function mixin(name, mixin) {
    if (MIXINS_MAP$1.has(name)) panic(`The mixin "${name}" was already defined`);

    MIXINS_MAP$1.set(name, callOrAssign(mixin));

    return MIXINS_MAP$1
  }

  /**
   * Function to define an anonymous component
   * @param   {Object} component - this object should contain the component implementation,
   * like css and/or template/render function
   * @param   {Object} slotsAndAttributes - object containing the slots or attribute expressions
   * you shouldn't normally need it but it might be handy for testing
   * @returns {Riot.Component} a riot component instance
   */
  const component = ({css, template, ...rest}, slotsAndAttributes = {}) => defineComponent({
    css,
    template,
    tag: rest
  })(slotsAndAttributes);

  /** @type {string} current riot version */
  const version = 'v4.0.0-alpha.0';

  // expose some internal stuff that might be used from external tools
  const __ = {
    cssManager,
    globals
  };

  var riot = /*#__PURE__*/Object.freeze({
    register: register,
    unregister: unregister,
    mount: mount,
    unmount: unmount,
    mixin: mixin,
    component: component,
    version: version,
    __: __
  });

  var commonjsGlobal = typeof window !== 'undefined' ? window : typeof global !== 'undefined' ? global : typeof self !== 'undefined' ? self : {};

  function unwrapExports (x) {
  	return x && x.__esModule && Object.prototype.hasOwnProperty.call(x, 'default') ? x.default : x;
  }

  function createCommonjsModule(fn, module) {
  	return module = { exports: {} }, fn(module, module.exports), module.exports;
  }

  function getCjsExportFromNamespace (n) {
  	return n && n.default || n;
  }

  var _empty_module = {};

  var _empty_module$1 = /*#__PURE__*/Object.freeze({
    default: _empty_module
  });

  var Ap = Array.prototype;
  var slice = Ap.slice;
  var Op = Object.prototype;
  var objToStr = Op.toString;
  var funObjStr = objToStr.call(function(){});
  var strObjStr = objToStr.call("");
  var hasOwn = Op.hasOwnProperty;

  var types = function () {

      var exports = {};

      // A type is an object with a .check method that takes a value and returns
      // true or false according to whether the value matches the type.

      function Type(check, name) {
          var self = this;
          if (!(self instanceof Type)) {
              throw new Error("Type constructor cannot be invoked without 'new'");
          }

          // Unfortunately we can't elegantly reuse isFunction and isString,
          // here, because this code is executed while defining those types.
          if (objToStr.call(check) !== funObjStr) {
              throw new Error(check + " is not a function");
          }

          // The `name` parameter can be either a function or a string.
          var nameObjStr = objToStr.call(name);
          if (!(nameObjStr === funObjStr ||
            nameObjStr === strObjStr)) {
              throw new Error(name + " is neither a function nor a string");
          }

          Object.defineProperties(self, {
              name: {value: name},
              check: {
                  value: function (value, deep) {
                      var result = check.call(self, value, deep);
                      if (!result && deep && objToStr.call(deep) === funObjStr)
                          deep(self, value);
                      return result;
                  }
              }
          });
      }

      var Tp = Type.prototype;

      // Throughout this file we use Object.defineProperty to prevent
      // redefinition of exported properties.
      exports.Type = Type;

      // Like .check, except that failure triggers an AssertionError.
      Tp.assert = function (value, deep) {
          if (!this.check(value, deep)) {
              var str = shallowStringify(value);
              throw new Error(str + " does not match type " + this);
          }
          return true;
      };

      function shallowStringify(value) {
          if (isObject.check(value))
              return "{" + Object.keys(value).map(function (key) {
                    return key + ": " + value[key];
                }).join(", ") + "}";

          if (isArray.check(value))
              return "[" + value.map(shallowStringify).join(", ") + "]";

          return JSON.stringify(value);
      }

      Tp.toString = function () {
          var name = this.name;

          if (isString.check(name))
              return name;

          if (isFunction.check(name))
              return name.call(this) + "";

          return name + " type";
      };

      var builtInCtorFns = [];
      var builtInCtorTypes = [];
      var builtInTypes = {};
      exports.builtInTypes = builtInTypes;

      function defBuiltInType(example, name) {
          var objStr = objToStr.call(example);

          var type = new Type(function (value) {
              return objToStr.call(value) === objStr;
          }, name);

          builtInTypes[name] = type;

          if (example && typeof example.constructor === "function") {
              builtInCtorFns.push(example.constructor);
              builtInCtorTypes.push(type);
          }

          return type;
      }

      // These types check the underlying [[Class]] attribute of the given
      // value, rather than using the problematic typeof operator. Note however
      // that no subtyping is considered; so, for instance, isObject.check
      // returns false for [], /./, new Date, and null.
      var isString = defBuiltInType("truthy", "string");
      var isFunction = defBuiltInType(function () {}, "function");
      var isArray = defBuiltInType([], "array");
      var isObject = defBuiltInType({}, "object");
      var isRegExp = defBuiltInType(/./, "RegExp");
      var isDate = defBuiltInType(new Date, "Date");
      var isNumber = defBuiltInType(3, "number");
      var isBoolean = defBuiltInType(true, "boolean");
      var isNull = defBuiltInType(null, "null");
      var isUndefined = defBuiltInType(void 0, "undefined");

      // There are a number of idiomatic ways of expressing types, so this
      // function serves to coerce them all to actual Type objects. Note that
      // providing the name argument is not necessary in most cases.
      function toType(from, name) {
          // The toType function should of course be idempotent.
          if (from instanceof Type)
              return from;

          // The Def type is used as a helper for constructing compound
          // interface types for AST nodes.
          if (from instanceof Def)
              return from.type;

          // Support [ElemType] syntax.
          if (isArray.check(from))
              return Type.fromArray(from);

          // Support { someField: FieldType, ... } syntax.
          if (isObject.check(from))
              return Type.fromObject(from);

          if (isFunction.check(from)) {
              var bicfIndex = builtInCtorFns.indexOf(from);
              if (bicfIndex >= 0) {
                  return builtInCtorTypes[bicfIndex];
              }

              // If isFunction.check(from), and from is not a built-in
              // constructor, assume from is a binary predicate function we can
              // use to define the type.
              return new Type(from, name);
          }

          // As a last resort, toType returns a type that matches any value that
          // is === from. This is primarily useful for literal values like
          // toType(null), but it has the additional advantage of allowing
          // toType to be a total function.
          return new Type(function (value) {
              return value === from;
          }, isUndefined.check(name) ? function () {
              return from + "";
          } : name);
      }

      // Returns a type that matches the given value iff any of type1, type2,
      // etc. match the value.
      Type.or = function (/* type1, type2, ... */) {
          var types = [];
          var len = arguments.length;
          for (var i = 0; i < len; ++i)
              types.push(toType(arguments[i]));

          return new Type(function (value, deep) {
              for (var i = 0; i < len; ++i)
                  if (types[i].check(value, deep))
                      return true;
              return false;
          }, function () {
              return types.join(" | ");
          });
      };

      Type.fromArray = function (arr) {
          if (!isArray.check(arr)) {
              throw new Error("");
          }
          if (arr.length !== 1) {
              throw new Error("only one element type is permitted for typed arrays");
          }
          return toType(arr[0]).arrayOf();
      };

      Tp.arrayOf = function () {
          var elemType = this;
          return new Type(function (value, deep) {
              return isArray.check(value) && value.every(function (elem) {
                    return elemType.check(elem, deep);
                });
          }, function () {
              return "[" + elemType + "]";
          });
      };

      Type.fromObject = function (obj) {
          var fields = Object.keys(obj).map(function (name) {
              return new Field(name, obj[name]);
          });

          return new Type(function (value, deep) {
              return isObject.check(value) && fields.every(function (field) {
                    return field.type.check(value[field.name], deep);
                });
          }, function () {
              return "{ " + fields.join(", ") + " }";
          });
      };

      function Field(name, type, defaultFn, hidden) {
          var self = this;

          if (!(self instanceof Field)) {
              throw new Error("Field constructor cannot be invoked without 'new'");
          }
          isString.assert(name);

          type = toType(type);

          var properties = {
              name: {value: name},
              type: {value: type},
              hidden: {value: !!hidden}
          };

          if (isFunction.check(defaultFn)) {
              properties.defaultFn = {value: defaultFn};
          }

          Object.defineProperties(self, properties);
      }

      var Fp = Field.prototype;

      Fp.toString = function () {
          return JSON.stringify(this.name) + ": " + this.type;
      };

      Fp.getValue = function (obj) {
          var value = obj[this.name];

          if (!isUndefined.check(value))
              return value;

          if (this.defaultFn)
              value = this.defaultFn.call(obj);

          return value;
      };

      // Define a type whose name is registered in a namespace (the defCache) so
      // that future definitions will return the same type given the same name.
      // In particular, this system allows for circular and forward definitions.
      // The Def object d returned from Type.def may be used to configure the
      // type d.type by calling methods such as d.bases, d.build, and d.field.
      Type.def = function (typeName) {
          isString.assert(typeName);
          return hasOwn.call(defCache, typeName)
            ? defCache[typeName]
            : defCache[typeName] = new Def(typeName);
      };

      // In order to return the same Def instance every time Type.def is called
      // with a particular name, those instances need to be stored in a cache.
      var defCache = Object.create(null);

      function Def(typeName) {
          var self = this;
          if (!(self instanceof Def)) {
              throw new Error("Def constructor cannot be invoked without 'new'");
          }

          Object.defineProperties(self, {
              typeName: {value: typeName},
              baseNames: {value: []},
              ownFields: {value: Object.create(null)},

              // These two are populated during finalization.
              allSupertypes: {value: Object.create(null)}, // Includes own typeName.
              supertypeList: {value: []}, // Linear inheritance hierarchy.
              allFields: {value: Object.create(null)}, // Includes inherited fields.
              fieldNames: {value: []}, // Non-hidden keys of allFields.

              type: {
                  value: new Type(function (value, deep) {
                      return self.check(value, deep);
                  }, typeName)
              }
          });
      }

      Def.fromValue = function (value) {
          if (value && typeof value === "object") {
              var type = value.type;
              if (typeof type === "string" &&
                hasOwn.call(defCache, type)) {
                  var d = defCache[type];
                  if (d.finalized) {
                      return d;
                  }
              }
          }

          return null;
      };

      var Dp = Def.prototype;

      Dp.isSupertypeOf = function (that) {
          if (that instanceof Def) {
              if (this.finalized !== true ||
                that.finalized !== true) {
                  throw new Error("");
              }
              return hasOwn.call(that.allSupertypes, this.typeName);
          } else {
              throw new Error(that + " is not a Def");
          }
      };

      // Note that the list returned by this function is a copy of the internal
      // supertypeList, *without* the typeName itself as the first element.
      exports.getSupertypeNames = function (typeName) {
          if (!hasOwn.call(defCache, typeName)) {
              throw new Error("");
          }
          var d = defCache[typeName];
          if (d.finalized !== true) {
              throw new Error("");
          }
          return d.supertypeList.slice(1);
      };

      // Returns an object mapping from every known type in the defCache to the
      // most specific supertype whose name is an own property of the candidates
      // object.
      exports.computeSupertypeLookupTable = function (candidates) {
          var table = {};
          var typeNames = Object.keys(defCache);
          var typeNameCount = typeNames.length;

          for (var i = 0; i < typeNameCount; ++i) {
              var typeName = typeNames[i];
              var d = defCache[typeName];
              if (d.finalized !== true) {
                  throw new Error("" + typeName);
              }
              for (var j = 0; j < d.supertypeList.length; ++j) {
                  var superTypeName = d.supertypeList[j];
                  if (hasOwn.call(candidates, superTypeName)) {
                      table[typeName] = superTypeName;
                      break;
                  }
              }
          }

          return table;
      };

      Dp.checkAllFields = function (value, deep) {
          var allFields = this.allFields;
          if (this.finalized !== true) {
              throw new Error("" + this.typeName);
          }

          function checkFieldByName(name) {
              var field = allFields[name];
              var type = field.type;
              var child = field.getValue(value);
              return type.check(child, deep);
          }

          return isObject.check(value)
            && Object.keys(allFields).every(checkFieldByName);
      };

      Dp.check = function (value, deep) {
          if (this.finalized !== true) {
              throw new Error(
                "prematurely checking unfinalized type " + this.typeName
              );
          }

          // A Def type can only match an object value.
          if (!isObject.check(value))
              return false;

          var vDef = Def.fromValue(value);
          if (!vDef) {
              // If we couldn't infer the Def associated with the given value,
              // and we expected it to be a SourceLocation or a Position, it was
              // probably just missing a "type" field (because Esprima does not
              // assign a type property to such nodes). Be optimistic and let
              // this.checkAllFields make the final decision.
              if (this.typeName === "SourceLocation" ||
                this.typeName === "Position") {
                  return this.checkAllFields(value, deep);
              }

              // Calling this.checkAllFields for any other type of node is both
              // bad for performance and way too forgiving.
              return false;
          }

          // If checking deeply and vDef === this, then we only need to call
          // checkAllFields once. Calling checkAllFields is too strict when deep
          // is false, because then we only care about this.isSupertypeOf(vDef).
          if (deep && vDef === this)
              return this.checkAllFields(value, deep);

          // In most cases we rely exclusively on isSupertypeOf to make O(1)
          // subtyping determinations. This suffices in most situations outside
          // of unit tests, since interface conformance is checked whenever new
          // instances are created using builder functions.
          if (!this.isSupertypeOf(vDef))
              return false;

          // The exception is when deep is true; then, we recursively check all
          // fields.
          if (!deep)
              return true;

          // Use the more specific Def (vDef) to perform the deep check, but
          // shallow-check fields defined by the less specific Def (this).
          return vDef.checkAllFields(value, deep)
            && this.checkAllFields(value, false);
      };

      Dp.bases = function () {
          var args = slice.call(arguments);
          var bases = this.baseNames;

          if (this.finalized) {
              if (args.length !== bases.length) {
                  throw new Error("");
              }
              for (var i = 0; i < args.length; i++) {
                  if (args[i] !== bases[i]) {
                      throw new Error("");
                  }
              }
              return this;
          }

          args.forEach(function (baseName) {
              isString.assert(baseName);

              // This indexOf lookup may be O(n), but the typical number of base
              // names is very small, and indexOf is a native Array method.
              if (bases.indexOf(baseName) < 0)
                  bases.push(baseName);
          });

          return this; // For chaining.
      };

      // False by default until .build(...) is called on an instance.
      Object.defineProperty(Dp, "buildable", {value: false});

      var builders = {};
      exports.builders = builders;

      // This object is used as prototype for any node created by a builder.
      var nodePrototype = {};

      // Call this function to define a new method to be shared by all AST
       // nodes. The replaced method (if any) is returned for easy wrapping.
      exports.defineMethod = function (name, func) {
          var old = nodePrototype[name];

          // Pass undefined as func to delete nodePrototype[name].
          if (isUndefined.check(func)) {
              delete nodePrototype[name];

          } else {
              isFunction.assert(func);

              Object.defineProperty(nodePrototype, name, {
                  enumerable: true, // For discoverability.
                  configurable: true, // For delete proto[name].
                  value: func
              });
          }

          return old;
      };

      var isArrayOfString = isString.arrayOf();

      // Calling the .build method of a Def simultaneously marks the type as
      // buildable (by defining builders[getBuilderName(typeName)]) and
      // specifies the order of arguments that should be passed to the builder
      // function to create an instance of the type.
      Dp.build = function (/* param1, param2, ... */) {
          var self = this;

          var newBuildParams = slice.call(arguments);
          isArrayOfString.assert(newBuildParams);

          // Calling Def.prototype.build multiple times has the effect of merely
          // redefining this property.
          Object.defineProperty(self, "buildParams", {
              value: newBuildParams,
              writable: false,
              enumerable: false,
              configurable: true
          });

          if (self.buildable) {
              // If this Def is already buildable, update self.buildParams and
              // continue using the old builder function.
              return self;
          }

          // Every buildable type will have its "type" field filled in
          // automatically. This includes types that are not subtypes of Node,
          // like SourceLocation, but that seems harmless (TODO?).
          self.field("type", String, function () { return self.typeName });

          // Override Dp.buildable for this Def instance.
          Object.defineProperty(self, "buildable", {value: true});

          function addParam(built, param, arg, isArgAvailable) {
              if (hasOwn.call(built, param))
                  return;

              var all = self.allFields;
              if (!hasOwn.call(all, param)) {
                  throw new Error("" + param);
              }

              var field = all[param];
              var type = field.type;
              var value;

              if (isArgAvailable) {
                  value = arg;
              } else if (field.defaultFn) {
                  // Expose the partially-built object to the default
                  // function as its `this` object.
                  value = field.defaultFn.call(built);
              } else {
                  var message = "no value or default function given for field " +
                    JSON.stringify(param) + " of " + self.typeName + "(" +
                    self.buildParams.map(function (name) {
                        return all[name];
                    }).join(", ") + ")";
                  throw new Error(message);
              }

              if (!type.check(value)) {
                  throw new Error(
                    shallowStringify(value) +
                    " does not match field " + field +
                    " of type " + self.typeName
                  );
              }

              built[param] = value;
          }

          // Calling the builder function will construct an instance of the Def,
          // with positional arguments mapped to the fields original passed to .build.
          // If not enough arguments are provided, the default value for the remaining fields
          // will be used.
          function builder() {
              var args = arguments;
              var argc = args.length;
              
              if (!self.finalized) {
                  throw new Error(
                      "attempting to instantiate unfinalized type " +
                      self.typeName
                  );
              }

              var built = Object.create(nodePrototype);

              self.buildParams.forEach(function (param, i) {
                  if (i < argc) {
                      addParam(built, param, args[i], true);
                  } else {
                      addParam(built, param, null, false);
                  }
              });

              Object.keys(self.allFields).forEach(function (param) {
                  // Use the default value.
                  addParam(built, param, null, false);
              });

              // Make sure that the "type" field was filled automatically.
              if (built.type !== self.typeName) {
                  throw new Error("");
              }

              return built;
          }

          // Calling .from on the builder function will construct an instance of the Def,
          // using field values from the passed object. For fields missing from the passed object,
          // their default value will be used.
          builder.from = function (obj) {
              if (!self.finalized) {
                  throw new Error(
                      "attempting to instantiate unfinalized type " +
                      self.typeName
                  );
              }
              
              var built = Object.create(nodePrototype);

              Object.keys(self.allFields).forEach(function (param) {
                  if (hasOwn.call(obj, param)) {
                      addParam(built, param, obj[param], true);
                  } else {
                      addParam(built, param, null, false);
                  }
              });

              // Make sure that the "type" field was filled automatically.
              if (built.type !== self.typeName) {
                  throw new Error("");
              }

              return built;
          };

          Object.defineProperty(builders, getBuilderName(self.typeName), {
              enumerable: true,
              value: builder
          });

          return self; // For chaining.
      };

      function getBuilderName(typeName) {
          return typeName.replace(/^[A-Z]+/, function (upperCasePrefix) {
              var len = upperCasePrefix.length;
              switch (len) {
                  case 0: return "";
                  // If there's only one initial capital letter, just lower-case it.
                  case 1: return upperCasePrefix.toLowerCase();
                  default:
                      // If there's more than one initial capital letter, lower-case
                      // all but the last one, so that XMLDefaultDeclaration (for
                      // example) becomes xmlDefaultDeclaration.
                      return upperCasePrefix.slice(
                          0, len - 1).toLowerCase() +
                        upperCasePrefix.charAt(len - 1);
              }
          });
      }
      exports.getBuilderName = getBuilderName;

      function getStatementBuilderName(typeName) {
          typeName = getBuilderName(typeName);
          return typeName.replace(/(Expression)?$/, "Statement");
      }
      exports.getStatementBuilderName = getStatementBuilderName;

      // The reason fields are specified using .field(...) instead of an object
      // literal syntax is somewhat subtle: the object literal syntax would
      // support only one key and one value, but with .field(...) we can pass
      // any number of arguments to specify the field.
      Dp.field = function (name, type, defaultFn, hidden) {
          if (this.finalized) {
              console.error("Ignoring attempt to redefine field " +
                JSON.stringify(name) + " of finalized type " +
                JSON.stringify(this.typeName));
              return this;
          }
          this.ownFields[name] = new Field(name, type, defaultFn, hidden);
          return this; // For chaining.
      };

      var namedTypes = {};
      exports.namedTypes = namedTypes;

      // Like Object.keys, but aware of what fields each AST type should have.
      function getFieldNames(object) {
          var d = Def.fromValue(object);
          if (d) {
              return d.fieldNames.slice(0);
          }

          if ("type" in object) {
              throw new Error(
                "did not recognize object of type " +
                JSON.stringify(object.type)
              );
          }

          return Object.keys(object);
      }
      exports.getFieldNames = getFieldNames;

      // Get the value of an object property, taking object.type and default
      // functions into account.
      function getFieldValue(object, fieldName) {
          var d = Def.fromValue(object);
          if (d) {
              var field = d.allFields[fieldName];
              if (field) {
                  return field.getValue(object);
              }
          }

          return object && object[fieldName];
      }
      exports.getFieldValue = getFieldValue;

      // Iterate over all defined fields of an object, including those missing
      // or undefined, passing each field name and effective value (as returned
      // by getFieldValue) to the callback. If the object has no corresponding
      // Def, the callback will never be called.
      exports.eachField = function (object, callback, context) {
          getFieldNames(object).forEach(function (name) {
              callback.call(this, name, getFieldValue(object, name));
          }, context);
      };

      // Similar to eachField, except that iteration stops as soon as the
      // callback returns a truthy value. Like Array.prototype.some, the final
      // result is either true or false to indicates whether the callback
      // returned true for any element or not.
      exports.someField = function (object, callback, context) {
          return getFieldNames(object).some(function (name) {
              return callback.call(this, name, getFieldValue(object, name));
          }, context);
      };

      // This property will be overridden as true by individual Def instances
      // when they are finalized.
      Object.defineProperty(Dp, "finalized", {value: false});

      Dp.finalize = function () {
          var self = this;

          // It's not an error to finalize a type more than once, but only the
          // first call to .finalize does anything.
          if (!self.finalized) {
              var allFields = self.allFields;
              var allSupertypes = self.allSupertypes;

              self.baseNames.forEach(function (name) {
                  var def = defCache[name];
                  if (def instanceof Def) {
                      def.finalize();
                      extend(allFields, def.allFields);
                      extend(allSupertypes, def.allSupertypes);
                  } else {
                      var message = "unknown supertype name " +
                        JSON.stringify(name) +
                        " for subtype " +
                        JSON.stringify(self.typeName);
                      throw new Error(message);
                  }
              });

              // TODO Warn if fields are overridden with incompatible types.
              extend(allFields, self.ownFields);
              allSupertypes[self.typeName] = self;

              self.fieldNames.length = 0;
              for (var fieldName in allFields) {
                  if (hasOwn.call(allFields, fieldName) &&
                      !allFields[fieldName].hidden) {
                          self.fieldNames.push(fieldName);
                  }
              }

              // Types are exported only once they have been finalized.
              Object.defineProperty(namedTypes, self.typeName, {
                  enumerable: true,
                  value: self.type
              });

              Object.defineProperty(self, "finalized", {value: true});

              // A linearization of the inheritance hierarchy.
              populateSupertypeList(self.typeName, self.supertypeList);

              if (self.buildable && self.supertypeList.lastIndexOf("Expression") >= 0) {
                  wrapExpressionBuilderWithStatement(self.typeName);
              }
          }
      };

      // Adds an additional builder for Expression subtypes
      // that wraps the built Expression in an ExpressionStatements.
      function wrapExpressionBuilderWithStatement(typeName) {
          var wrapperName = getStatementBuilderName(typeName);

          // skip if the builder already exists
          if (builders[wrapperName]) return;

          // the builder function to wrap with builders.ExpressionStatement
          var wrapped = builders[getBuilderName(typeName)];

          // skip if there is nothing to wrap
          if (!wrapped) return;

          builders[wrapperName] = function () {
              return builders.expressionStatement(wrapped.apply(builders, arguments));
          };
      }

      function populateSupertypeList(typeName, list) {
          list.length = 0;
          list.push(typeName);

          var lastSeen = Object.create(null);

          for (var pos = 0; pos < list.length; ++pos) {
              typeName = list[pos];
              var d = defCache[typeName];
              if (d.finalized !== true) {
                  throw new Error("");
              }

              // If we saw typeName earlier in the breadth-first traversal,
              // delete the last-seen occurrence.
              if (hasOwn.call(lastSeen, typeName)) {
                  delete list[lastSeen[typeName]];
              }

              // Record the new index of the last-seen occurrence of typeName.
              lastSeen[typeName] = pos;

              // Enqueue the base names of this type.
              list.push.apply(list, d.baseNames);
          }

          // Compaction loop to remove array holes.
          for (var to = 0, from = to, len = list.length; from < len; ++from) {
              if (hasOwn.call(list, from)) {
                  list[to++] = list[from];
              }
          }

          list.length = to;
      }

      function extend(into, from) {
          Object.keys(from).forEach(function (name) {
              into[name] = from[name];
          });

          return into;
      }
      exports.finalize = function () {
          Object.keys(defCache).forEach(function (name) {
              defCache[name].finalize();
          });
      };

      return exports;
  };

  var equiv = function (fork) {
      var types$$1 = fork.use(types);
      var getFieldNames = types$$1.getFieldNames;
      var getFieldValue = types$$1.getFieldValue;
      var isArray = types$$1.builtInTypes.array;
      var isObject = types$$1.builtInTypes.object;
      var isDate = types$$1.builtInTypes.Date;
      var isRegExp = types$$1.builtInTypes.RegExp;
      var hasOwn = Object.prototype.hasOwnProperty;

      function astNodesAreEquivalent(a, b, problemPath) {
          if (isArray.check(problemPath)) {
              problemPath.length = 0;
          } else {
              problemPath = null;
          }

          return areEquivalent(a, b, problemPath);
      }

      astNodesAreEquivalent.assert = function (a, b) {
          var problemPath = [];
          if (!astNodesAreEquivalent(a, b, problemPath)) {
              if (problemPath.length === 0) {
                  if (a !== b) {
                      throw new Error("Nodes must be equal");
                  }
              } else {
                  throw new Error(
                    "Nodes differ in the following path: " +
                    problemPath.map(subscriptForProperty).join("")
                  );
              }
          }
      };

      function subscriptForProperty(property) {
          if (/[_$a-z][_$a-z0-9]*/i.test(property)) {
              return "." + property;
          }
          return "[" + JSON.stringify(property) + "]";
      }

      function areEquivalent(a, b, problemPath) {
          if (a === b) {
              return true;
          }

          if (isArray.check(a)) {
              return arraysAreEquivalent(a, b, problemPath);
          }

          if (isObject.check(a)) {
              return objectsAreEquivalent(a, b, problemPath);
          }

          if (isDate.check(a)) {
              return isDate.check(b) && (+a === +b);
          }

          if (isRegExp.check(a)) {
              return isRegExp.check(b) && (
                  a.source === b.source &&
                  a.global === b.global &&
                  a.multiline === b.multiline &&
                  a.ignoreCase === b.ignoreCase
                );
          }

          return a == b;
      }

      function arraysAreEquivalent(a, b, problemPath) {
          isArray.assert(a);
          var aLength = a.length;

          if (!isArray.check(b) || b.length !== aLength) {
              if (problemPath) {
                  problemPath.push("length");
              }
              return false;
          }

          for (var i = 0; i < aLength; ++i) {
              if (problemPath) {
                  problemPath.push(i);
              }

              if (i in a !== i in b) {
                  return false;
              }

              if (!areEquivalent(a[i], b[i], problemPath)) {
                  return false;
              }

              if (problemPath) {
                  var problemPathTail = problemPath.pop();
                  if (problemPathTail !== i) {
                      throw new Error("" + problemPathTail);
                  }
              }
          }

          return true;
      }

      function objectsAreEquivalent(a, b, problemPath) {
          isObject.assert(a);
          if (!isObject.check(b)) {
              return false;
          }

          // Fast path for a common property of AST nodes.
          if (a.type !== b.type) {
              if (problemPath) {
                  problemPath.push("type");
              }
              return false;
          }

          var aNames = getFieldNames(a);
          var aNameCount = aNames.length;

          var bNames = getFieldNames(b);
          var bNameCount = bNames.length;

          if (aNameCount === bNameCount) {
              for (var i = 0; i < aNameCount; ++i) {
                  var name = aNames[i];
                  var aChild = getFieldValue(a, name);
                  var bChild = getFieldValue(b, name);

                  if (problemPath) {
                      problemPath.push(name);
                  }

                  if (!areEquivalent(aChild, bChild, problemPath)) {
                      return false;
                  }

                  if (problemPath) {
                      var problemPathTail = problemPath.pop();
                      if (problemPathTail !== name) {
                          throw new Error("" + problemPathTail);
                      }
                  }
              }

              return true;
          }

          if (!problemPath) {
              return false;
          }

          // Since aNameCount !== bNameCount, we need to find some name that's
          // missing in aNames but present in bNames, or vice-versa.

          var seenNames = Object.create(null);

          for (i = 0; i < aNameCount; ++i) {
              seenNames[aNames[i]] = true;
          }

          for (i = 0; i < bNameCount; ++i) {
              name = bNames[i];

              if (!hasOwn.call(seenNames, name)) {
                  problemPath.push(name);
                  return false;
              }

              delete seenNames[name];
          }

          for (name in seenNames) {
              problemPath.push(name);
              break;
          }

          return false;
      }
      
      return astNodesAreEquivalent;
  };

  var Op$1 = Object.prototype;
  var hasOwn$1 = Op$1.hasOwnProperty;

  var path = function (fork) {
      var types$$1 = fork.use(types);
      var isArray = types$$1.builtInTypes.array;
      var isNumber = types$$1.builtInTypes.number;

      function Path(value, parentPath, name) {
          if (!(this instanceof Path)) {
              throw new Error("Path constructor cannot be invoked without 'new'");
          }

          if (parentPath) {
              if (!(parentPath instanceof Path)) {
                  throw new Error("");
              }
          } else {
              parentPath = null;
              name = null;
          }

          // The value encapsulated by this Path, generally equal to
          // parentPath.value[name] if we have a parentPath.
          this.value = value;

          // The immediate parent Path of this Path.
          this.parentPath = parentPath;

          // The name of the property of parentPath.value through which this
          // Path's value was reached.
          this.name = name;

          // Calling path.get("child") multiple times always returns the same
          // child Path object, for both performance and consistency reasons.
          this.__childCache = null;
      }

      var Pp = Path.prototype;

      function getChildCache(path) {
          // Lazily create the child cache. This also cheapens cache
          // invalidation, since you can just reset path.__childCache to null.
          return path.__childCache || (path.__childCache = Object.create(null));
      }

      function getChildPath(path, name) {
          var cache = getChildCache(path);
          var actualChildValue = path.getValueProperty(name);
          var childPath = cache[name];
          if (!hasOwn$1.call(cache, name) ||
            // Ensure consistency between cache and reality.
            childPath.value !== actualChildValue) {
              childPath = cache[name] = new path.constructor(
                actualChildValue, path, name
              );
          }
          return childPath;
      }

  // This method is designed to be overridden by subclasses that need to
  // handle missing properties, etc.
      Pp.getValueProperty = function getValueProperty(name) {
          return this.value[name];
      };

      Pp.get = function get(name) {
          var path = this;
          var names = arguments;
          var count = names.length;

          for (var i = 0; i < count; ++i) {
              path = getChildPath(path, names[i]);
          }

          return path;
      };

      Pp.each = function each(callback, context) {
          var childPaths = [];
          var len = this.value.length;
          var i = 0;

          // Collect all the original child paths before invoking the callback.
          for (var i = 0; i < len; ++i) {
              if (hasOwn$1.call(this.value, i)) {
                  childPaths[i] = this.get(i);
              }
          }

          // Invoke the callback on just the original child paths, regardless of
          // any modifications made to the array by the callback. I chose these
          // semantics over cleverly invoking the callback on new elements because
          // this way is much easier to reason about.
          context = context || this;
          for (i = 0; i < len; ++i) {
              if (hasOwn$1.call(childPaths, i)) {
                  callback.call(context, childPaths[i]);
              }
          }
      };

      Pp.map = function map(callback, context) {
          var result = [];

          this.each(function (childPath) {
              result.push(callback.call(this, childPath));
          }, context);

          return result;
      };

      Pp.filter = function filter(callback, context) {
          var result = [];

          this.each(function (childPath) {
              if (callback.call(this, childPath)) {
                  result.push(childPath);
              }
          }, context);

          return result;
      };

      function emptyMoves() {}
      function getMoves(path, offset, start, end) {
          isArray.assert(path.value);

          if (offset === 0) {
              return emptyMoves;
          }

          var length = path.value.length;
          if (length < 1) {
              return emptyMoves;
          }

          var argc = arguments.length;
          if (argc === 2) {
              start = 0;
              end = length;
          } else if (argc === 3) {
              start = Math.max(start, 0);
              end = length;
          } else {
              start = Math.max(start, 0);
              end = Math.min(end, length);
          }

          isNumber.assert(start);
          isNumber.assert(end);

          var moves = Object.create(null);
          var cache = getChildCache(path);

          for (var i = start; i < end; ++i) {
              if (hasOwn$1.call(path.value, i)) {
                  var childPath = path.get(i);
                  if (childPath.name !== i) {
                      throw new Error("");
                  }
                  var newIndex = i + offset;
                  childPath.name = newIndex;
                  moves[newIndex] = childPath;
                  delete cache[i];
              }
          }

          delete cache.length;

          return function () {
              for (var newIndex in moves) {
                  var childPath = moves[newIndex];
                  if (childPath.name !== +newIndex) {
                      throw new Error("");
                  }
                  cache[newIndex] = childPath;
                  path.value[newIndex] = childPath.value;
              }
          };
      }

      Pp.shift = function shift() {
          var move = getMoves(this, -1);
          var result = this.value.shift();
          move();
          return result;
      };

      Pp.unshift = function unshift(node) {
          var move = getMoves(this, arguments.length);
          var result = this.value.unshift.apply(this.value, arguments);
          move();
          return result;
      };

      Pp.push = function push(node) {
          isArray.assert(this.value);
          delete getChildCache(this).length;
          return this.value.push.apply(this.value, arguments);
      };

      Pp.pop = function pop() {
          isArray.assert(this.value);
          var cache = getChildCache(this);
          delete cache[this.value.length - 1];
          delete cache.length;
          return this.value.pop();
      };

      Pp.insertAt = function insertAt(index, node) {
          var argc = arguments.length;
          var move = getMoves(this, argc - 1, index);
          if (move === emptyMoves) {
              return this;
          }

          index = Math.max(index, 0);

          for (var i = 1; i < argc; ++i) {
              this.value[index + i - 1] = arguments[i];
          }

          move();

          return this;
      };

      Pp.insertBefore = function insertBefore(node) {
          var pp = this.parentPath;
          var argc = arguments.length;
          var insertAtArgs = [this.name];
          for (var i = 0; i < argc; ++i) {
              insertAtArgs.push(arguments[i]);
          }
          return pp.insertAt.apply(pp, insertAtArgs);
      };

      Pp.insertAfter = function insertAfter(node) {
          var pp = this.parentPath;
          var argc = arguments.length;
          var insertAtArgs = [this.name + 1];
          for (var i = 0; i < argc; ++i) {
              insertAtArgs.push(arguments[i]);
          }
          return pp.insertAt.apply(pp, insertAtArgs);
      };

      function repairRelationshipWithParent(path) {
          if (!(path instanceof Path)) {
              throw new Error("");
          }

          var pp = path.parentPath;
          if (!pp) {
              // Orphan paths have no relationship to repair.
              return path;
          }

          var parentValue = pp.value;
          var parentCache = getChildCache(pp);

          // Make sure parentCache[path.name] is populated.
          if (parentValue[path.name] === path.value) {
              parentCache[path.name] = path;
          } else if (isArray.check(parentValue)) {
              // Something caused path.name to become out of date, so attempt to
              // recover by searching for path.value in parentValue.
              var i = parentValue.indexOf(path.value);
              if (i >= 0) {
                  parentCache[path.name = i] = path;
              }
          } else {
              // If path.value disagrees with parentValue[path.name], and
              // path.name is not an array index, let path.value become the new
              // parentValue[path.name] and update parentCache accordingly.
              parentValue[path.name] = path.value;
              parentCache[path.name] = path;
          }

          if (parentValue[path.name] !== path.value) {
              throw new Error("");
          }
          if (path.parentPath.get(path.name) !== path) {
              throw new Error("");
          }

          return path;
      }

      Pp.replace = function replace(replacement) {
          var results = [];
          var parentValue = this.parentPath.value;
          var parentCache = getChildCache(this.parentPath);
          var count = arguments.length;

          repairRelationshipWithParent(this);

          if (isArray.check(parentValue)) {
              var originalLength = parentValue.length;
              var move = getMoves(this.parentPath, count - 1, this.name + 1);

              var spliceArgs = [this.name, 1];
              for (var i = 0; i < count; ++i) {
                  spliceArgs.push(arguments[i]);
              }

              var splicedOut = parentValue.splice.apply(parentValue, spliceArgs);

              if (splicedOut[0] !== this.value) {
                  throw new Error("");
              }
              if (parentValue.length !== (originalLength - 1 + count)) {
                  throw new Error("");
              }

              move();

              if (count === 0) {
                  delete this.value;
                  delete parentCache[this.name];
                  this.__childCache = null;

              } else {
                  if (parentValue[this.name] !== replacement) {
                      throw new Error("");
                  }

                  if (this.value !== replacement) {
                      this.value = replacement;
                      this.__childCache = null;
                  }

                  for (i = 0; i < count; ++i) {
                      results.push(this.parentPath.get(this.name + i));
                  }

                  if (results[0] !== this) {
                      throw new Error("");
                  }
              }

          } else if (count === 1) {
              if (this.value !== replacement) {
                  this.__childCache = null;
              }
              this.value = parentValue[this.name] = replacement;
              results.push(this);

          } else if (count === 0) {
              delete parentValue[this.name];
              delete this.value;
              this.__childCache = null;

              // Leave this path cached as parentCache[this.name], even though
              // it no longer has a value defined.

          } else {
              throw new Error("Could not replace path");
          }

          return results;
      };

      return Path;
  };

  var hasOwn$2 = Object.prototype.hasOwnProperty;

  var scope = function (fork) {
      var types$$1 = fork.use(types);
      var Type = types$$1.Type;
      var namedTypes = types$$1.namedTypes;
      var Node = namedTypes.Node;
      var Expression = namedTypes.Expression;
      var isArray = types$$1.builtInTypes.array;
      var b = types$$1.builders;

      function Scope(path, parentScope) {
          if (!(this instanceof Scope)) {
              throw new Error("Scope constructor cannot be invoked without 'new'");
          }
          if (!(path instanceof fork.use(nodePath))) {
              throw new Error("");
          }
          ScopeType.assert(path.value);

          var depth;

          if (parentScope) {
              if (!(parentScope instanceof Scope)) {
                  throw new Error("");
              }
              depth = parentScope.depth + 1;
          } else {
              parentScope = null;
              depth = 0;
          }

          Object.defineProperties(this, {
              path: { value: path },
              node: { value: path.value },
              isGlobal: { value: !parentScope, enumerable: true },
              depth: { value: depth },
              parent: { value: parentScope },
              bindings: { value: {} },
              types: { value: {} },
          });
      }

      var scopeTypes = [
          // Program nodes introduce global scopes.
          namedTypes.Program,

          // Function is the supertype of FunctionExpression,
          // FunctionDeclaration, ArrowExpression, etc.
          namedTypes.Function,

          // In case you didn't know, the caught parameter shadows any variable
          // of the same name in an outer scope.
          namedTypes.CatchClause
      ];

      var ScopeType = Type.or.apply(Type, scopeTypes);

      Scope.isEstablishedBy = function(node) {
          return ScopeType.check(node);
      };

      var Sp = Scope.prototype;

  // Will be overridden after an instance lazily calls scanScope.
      Sp.didScan = false;

      Sp.declares = function(name) {
          this.scan();
          return hasOwn$2.call(this.bindings, name);
      };

      Sp.declaresType = function(name) {
          this.scan();
          return hasOwn$2.call(this.types, name);
      };

      Sp.declareTemporary = function(prefix) {
          if (prefix) {
              if (!/^[a-z$_]/i.test(prefix)) {
                  throw new Error("");
              }
          } else {
              prefix = "t$";
          }

          // Include this.depth in the name to make sure the name does not
          // collide with any variables in nested/enclosing scopes.
          prefix += this.depth.toString(36) + "$";

          this.scan();

          var index = 0;
          while (this.declares(prefix + index)) {
              ++index;
          }

          var name = prefix + index;
          return this.bindings[name] = types$$1.builders.identifier(name);
      };

      Sp.injectTemporary = function(identifier, init) {
          identifier || (identifier = this.declareTemporary());

          var bodyPath = this.path.get("body");
          if (namedTypes.BlockStatement.check(bodyPath.value)) {
              bodyPath = bodyPath.get("body");
          }

          bodyPath.unshift(
            b.variableDeclaration(
              "var",
              [b.variableDeclarator(identifier, init || null)]
            )
          );

          return identifier;
      };

      Sp.scan = function(force) {
          if (force || !this.didScan) {
              for (var name in this.bindings) {
                  // Empty out this.bindings, just in cases.
                  delete this.bindings[name];
              }
              scanScope(this.path, this.bindings, this.types);
              this.didScan = true;
          }
      };

      Sp.getBindings = function () {
          this.scan();
          return this.bindings;
      };

      Sp.getTypes = function () {
          this.scan();
          return this.types;
      };

      function scanScope(path, bindings, scopeTypes) {
          var node = path.value;
          ScopeType.assert(node);

          if (namedTypes.CatchClause.check(node)) {
              // A catch clause establishes a new scope but the only variable
              // bound in that scope is the catch parameter. Any other
              // declarations create bindings in the outer scope.
              addPattern(path.get("param"), bindings);

          } else {
              recursiveScanScope(path, bindings, scopeTypes);
          }
      }

      function recursiveScanScope(path, bindings, scopeTypes) {
          var node = path.value;

          if (path.parent &&
            namedTypes.FunctionExpression.check(path.parent.node) &&
            path.parent.node.id) {
              addPattern(path.parent.get("id"), bindings);
          }

          if (!node) ; else if (isArray.check(node)) {
              path.each(function(childPath) {
                  recursiveScanChild(childPath, bindings, scopeTypes);
              });

          } else if (namedTypes.Function.check(node)) {
              path.get("params").each(function(paramPath) {
                  addPattern(paramPath, bindings);
              });

              recursiveScanChild(path.get("body"), bindings, scopeTypes);

          } else if (namedTypes.TypeAlias && namedTypes.TypeAlias.check(node)) {
              addTypePattern(path.get("id"), scopeTypes);

          } else if (namedTypes.VariableDeclarator.check(node)) {
              addPattern(path.get("id"), bindings);
              recursiveScanChild(path.get("init"), bindings, scopeTypes);

          } else if (node.type === "ImportSpecifier" ||
            node.type === "ImportNamespaceSpecifier" ||
            node.type === "ImportDefaultSpecifier") {
              addPattern(
                // Esprima used to use the .name field to refer to the local
                // binding identifier for ImportSpecifier nodes, but .id for
                // ImportNamespaceSpecifier and ImportDefaultSpecifier nodes.
                // ESTree/Acorn/ESpree use .local for all three node types.
                path.get(node.local ? "local" :
                  node.name ? "name" : "id"),
                bindings
              );

          } else if (Node.check(node) && !Expression.check(node)) {
              types$$1.eachField(node, function(name, child) {
                  var childPath = path.get(name);
                  if (!pathHasValue(childPath, child)) {
                      throw new Error("");
                  }
                  recursiveScanChild(childPath, bindings, scopeTypes);
              });
          }
      }

      function pathHasValue(path, value) {
          if (path.value === value) {
              return true;
          }

          // Empty arrays are probably produced by defaults.emptyArray, in which
          // case is makes sense to regard them as equivalent, if not ===.
          if (Array.isArray(path.value) &&
            path.value.length === 0 &&
            Array.isArray(value) &&
            value.length === 0) {
              return true;
          }

          return false;
      }

      function recursiveScanChild(path, bindings, scopeTypes) {
          var node = path.value;

          if (!node || Expression.check(node)) ; else if (namedTypes.FunctionDeclaration.check(node) &&
                     node.id !== null) {
              addPattern(path.get("id"), bindings);

          } else if (namedTypes.ClassDeclaration &&
            namedTypes.ClassDeclaration.check(node)) {
              addPattern(path.get("id"), bindings);

          } else if (ScopeType.check(node)) {
              if (namedTypes.CatchClause.check(node)) {
                  var catchParamName = node.param.name;
                  var hadBinding = hasOwn$2.call(bindings, catchParamName);

                  // Any declarations that occur inside the catch body that do
                  // not have the same name as the catch parameter should count
                  // as bindings in the outer scope.
                  recursiveScanScope(path.get("body"), bindings, scopeTypes);

                  // If a new binding matching the catch parameter name was
                  // created while scanning the catch body, ignore it because it
                  // actually refers to the catch parameter and not the outer
                  // scope that we're currently scanning.
                  if (!hadBinding) {
                      delete bindings[catchParamName];
                  }
              }

          } else {
              recursiveScanScope(path, bindings, scopeTypes);
          }
      }

      function addPattern(patternPath, bindings) {
          var pattern = patternPath.value;
          namedTypes.Pattern.assert(pattern);

          if (namedTypes.Identifier.check(pattern)) {
              if (hasOwn$2.call(bindings, pattern.name)) {
                  bindings[pattern.name].push(patternPath);
              } else {
                  bindings[pattern.name] = [patternPath];
              }

          } else if (namedTypes.AssignmentPattern &&
            namedTypes.AssignmentPattern.check(pattern)) {
            addPattern(patternPath.get('left'), bindings);

          } else if (namedTypes.ObjectPattern &&
            namedTypes.ObjectPattern.check(pattern)) {
              patternPath.get('properties').each(function(propertyPath) {
                  var property = propertyPath.value;
                  if (namedTypes.Pattern.check(property)) {
                      addPattern(propertyPath, bindings);
                  } else  if (namedTypes.Property.check(property)) {
                      addPattern(propertyPath.get('value'), bindings);
                  } else if (namedTypes.SpreadProperty &&
                    namedTypes.SpreadProperty.check(property)) {
                      addPattern(propertyPath.get('argument'), bindings);
                  }
              });

          } else if (namedTypes.ArrayPattern &&
            namedTypes.ArrayPattern.check(pattern)) {
              patternPath.get('elements').each(function(elementPath) {
                  var element = elementPath.value;
                  if (namedTypes.Pattern.check(element)) {
                      addPattern(elementPath, bindings);
                  } else if (namedTypes.SpreadElement &&
                    namedTypes.SpreadElement.check(element)) {
                      addPattern(elementPath.get("argument"), bindings);
                  }
              });

          } else if (namedTypes.PropertyPattern &&
            namedTypes.PropertyPattern.check(pattern)) {
              addPattern(patternPath.get('pattern'), bindings);

          } else if ((namedTypes.SpreadElementPattern &&
            namedTypes.SpreadElementPattern.check(pattern)) ||
            (namedTypes.SpreadPropertyPattern &&
            namedTypes.SpreadPropertyPattern.check(pattern))) {
              addPattern(patternPath.get('argument'), bindings);
          }
      }

      function addTypePattern(patternPath, types$$1) {
          var pattern = patternPath.value;
          namedTypes.Pattern.assert(pattern);

          if (namedTypes.Identifier.check(pattern)) {
              if (hasOwn$2.call(types$$1, pattern.name)) {
                  types$$1[pattern.name].push(patternPath);
              } else {
                  types$$1[pattern.name] = [patternPath];
              }

          }
      }

      Sp.lookup = function(name) {
          for (var scope = this; scope; scope = scope.parent)
              if (scope.declares(name))
                  break;
          return scope;
      };

      Sp.lookupType = function(name) {
          for (var scope = this; scope; scope = scope.parent)
              if (scope.declaresType(name))
                  break;
          return scope;
      };

      Sp.getGlobalScope = function() {
          var scope = this;
          while (!scope.isGlobal)
              scope = scope.parent;
          return scope;
      };

      return Scope;
  };

  var nodePath = function (fork) {
      var types$$1 = fork.use(types);
      var n = types$$1.namedTypes;
      var b = types$$1.builders;
      var isNumber = types$$1.builtInTypes.number;
      var isArray = types$$1.builtInTypes.array;
      var Path = fork.use(path);
      var Scope = fork.use(scope);

      function NodePath(value, parentPath, name) {
          if (!(this instanceof NodePath)) {
              throw new Error("NodePath constructor cannot be invoked without 'new'");
          }
          Path.call(this, value, parentPath, name);
      }

      var NPp = NodePath.prototype = Object.create(Path.prototype, {
          constructor: {
              value: NodePath,
              enumerable: false,
              writable: true,
              configurable: true
          }
      });

      Object.defineProperties(NPp, {
          node: {
              get: function () {
                  Object.defineProperty(this, "node", {
                      configurable: true, // Enable deletion.
                      value: this._computeNode()
                  });

                  return this.node;
              }
          },

          parent: {
              get: function () {
                  Object.defineProperty(this, "parent", {
                      configurable: true, // Enable deletion.
                      value: this._computeParent()
                  });

                  return this.parent;
              }
          },

          scope: {
              get: function () {
                  Object.defineProperty(this, "scope", {
                      configurable: true, // Enable deletion.
                      value: this._computeScope()
                  });

                  return this.scope;
              }
          }
      });

      NPp.replace = function () {
          delete this.node;
          delete this.parent;
          delete this.scope;
          return Path.prototype.replace.apply(this, arguments);
      };

      NPp.prune = function () {
          var remainingNodePath = this.parent;

          this.replace();

          return cleanUpNodesAfterPrune(remainingNodePath);
      };

      // The value of the first ancestor Path whose value is a Node.
      NPp._computeNode = function () {
          var value = this.value;
          if (n.Node.check(value)) {
              return value;
          }

          var pp = this.parentPath;
          return pp && pp.node || null;
      };

      // The first ancestor Path whose value is a Node distinct from this.node.
      NPp._computeParent = function () {
          var value = this.value;
          var pp = this.parentPath;

          if (!n.Node.check(value)) {
              while (pp && !n.Node.check(pp.value)) {
                  pp = pp.parentPath;
              }

              if (pp) {
                  pp = pp.parentPath;
              }
          }

          while (pp && !n.Node.check(pp.value)) {
              pp = pp.parentPath;
          }

          return pp || null;
      };

      // The closest enclosing scope that governs this node.
      NPp._computeScope = function () {
          var value = this.value;
          var pp = this.parentPath;
          var scope$$1 = pp && pp.scope;

          if (n.Node.check(value) &&
            Scope.isEstablishedBy(value)) {
              scope$$1 = new Scope(this, scope$$1);
          }

          return scope$$1 || null;
      };

      NPp.getValueProperty = function (name) {
          return types$$1.getFieldValue(this.value, name);
      };

      /**
       * Determine whether this.node needs to be wrapped in parentheses in order
       * for a parser to reproduce the same local AST structure.
       *
       * For instance, in the expression `(1 + 2) * 3`, the BinaryExpression
       * whose operator is "+" needs parentheses, because `1 + 2 * 3` would
       * parse differently.
       *
       * If assumeExpressionContext === true, we don't worry about edge cases
       * like an anonymous FunctionExpression appearing lexically first in its
       * enclosing statement and thus needing parentheses to avoid being parsed
       * as a FunctionDeclaration with a missing name.
       */
      NPp.needsParens = function (assumeExpressionContext) {
          var pp = this.parentPath;
          if (!pp) {
              return false;
          }

          var node = this.value;

          // Only expressions need parentheses.
          if (!n.Expression.check(node)) {
              return false;
          }

          // Identifiers never need parentheses.
          if (node.type === "Identifier") {
              return false;
          }

          while (!n.Node.check(pp.value)) {
              pp = pp.parentPath;
              if (!pp) {
                  return false;
              }
          }

          var parent = pp.value;

          switch (node.type) {
              case "UnaryExpression":
              case "SpreadElement":
              case "SpreadProperty":
                  return parent.type === "MemberExpression"
                    && this.name === "object"
                    && parent.object === node;

              case "BinaryExpression":
              case "LogicalExpression":
                  switch (parent.type) {
                      case "CallExpression":
                          return this.name === "callee"
                            && parent.callee === node;

                      case "UnaryExpression":
                      case "SpreadElement":
                      case "SpreadProperty":
                          return true;

                      case "MemberExpression":
                          return this.name === "object"
                            && parent.object === node;

                      case "BinaryExpression":
                      case "LogicalExpression":
                          var po = parent.operator;
                          var pp = PRECEDENCE[po];
                          var no = node.operator;
                          var np = PRECEDENCE[no];

                          if (pp > np) {
                              return true;
                          }

                          if (pp === np && this.name === "right") {
                              if (parent.right !== node) {
                                  throw new Error("Nodes must be equal");
                              }
                              return true;
                          }

                      default:
                          return false;
                  }

              case "SequenceExpression":
                  switch (parent.type) {
                      case "ForStatement":
                          // Although parentheses wouldn't hurt around sequence
                          // expressions in the head of for loops, traditional style
                          // dictates that e.g. i++, j++ should not be wrapped with
                          // parentheses.
                          return false;

                      case "ExpressionStatement":
                          return this.name !== "expression";

                      default:
                          // Otherwise err on the side of overparenthesization, adding
                          // explicit exceptions above if this proves overzealous.
                          return true;
                  }

              case "YieldExpression":
                  switch (parent.type) {
                      case "BinaryExpression":
                      case "LogicalExpression":
                      case "UnaryExpression":
                      case "SpreadElement":
                      case "SpreadProperty":
                      case "CallExpression":
                      case "MemberExpression":
                      case "NewExpression":
                      case "ConditionalExpression":
                      case "YieldExpression":
                          return true;

                      default:
                          return false;
                  }

              case "Literal":
                  return parent.type === "MemberExpression"
                    && isNumber.check(node.value)
                    && this.name === "object"
                    && parent.object === node;

              case "AssignmentExpression":
              case "ConditionalExpression":
                  switch (parent.type) {
                      case "UnaryExpression":
                      case "SpreadElement":
                      case "SpreadProperty":
                      case "BinaryExpression":
                      case "LogicalExpression":
                          return true;

                      case "CallExpression":
                          return this.name === "callee"
                            && parent.callee === node;

                      case "ConditionalExpression":
                          return this.name === "test"
                            && parent.test === node;

                      case "MemberExpression":
                          return this.name === "object"
                            && parent.object === node;

                      default:
                          return false;
                  }

              default:
                  if (parent.type === "NewExpression" &&
                    this.name === "callee" &&
                    parent.callee === node) {
                      return containsCallExpression(node);
                  }
          }

          if (assumeExpressionContext !== true &&
            !this.canBeFirstInStatement() &&
            this.firstInStatement())
              return true;

          return false;
      };

      function isBinary(node) {
          return n.BinaryExpression.check(node)
            || n.LogicalExpression.check(node);
      }

      var PRECEDENCE = {};
      [["||"],
          ["&&"],
          ["|"],
          ["^"],
          ["&"],
          ["==", "===", "!=", "!=="],
          ["<", ">", "<=", ">=", "in", "instanceof"],
          [">>", "<<", ">>>"],
          ["+", "-"],
          ["*", "/", "%"]
      ].forEach(function (tier, i) {
          tier.forEach(function (op) {
              PRECEDENCE[op] = i;
          });
      });

      function containsCallExpression(node) {
          if (n.CallExpression.check(node)) {
              return true;
          }

          if (isArray.check(node)) {
              return node.some(containsCallExpression);
          }

          if (n.Node.check(node)) {
              return types$$1.someField(node, function (name, child) {
                  return containsCallExpression(child);
              });
          }

          return false;
      }

      NPp.canBeFirstInStatement = function () {
          var node = this.node;
          return !n.FunctionExpression.check(node)
            && !n.ObjectExpression.check(node);
      };

      NPp.firstInStatement = function () {
          return firstInStatement(this);
      };

      function firstInStatement(path$$1) {
          for (var node, parent; path$$1.parent; path$$1 = path$$1.parent) {
              node = path$$1.node;
              parent = path$$1.parent.node;

              if (n.BlockStatement.check(parent) &&
                path$$1.parent.name === "body" &&
                path$$1.name === 0) {
                  if (parent.body[0] !== node) {
                      throw new Error("Nodes must be equal");
                  }
                  return true;
              }

              if (n.ExpressionStatement.check(parent) &&
                path$$1.name === "expression") {
                  if (parent.expression !== node) {
                      throw new Error("Nodes must be equal");
                  }
                  return true;
              }

              if (n.SequenceExpression.check(parent) &&
                path$$1.parent.name === "expressions" &&
                path$$1.name === 0) {
                  if (parent.expressions[0] !== node) {
                      throw new Error("Nodes must be equal");
                  }
                  continue;
              }

              if (n.CallExpression.check(parent) &&
                path$$1.name === "callee") {
                  if (parent.callee !== node) {
                      throw new Error("Nodes must be equal");
                  }
                  continue;
              }

              if (n.MemberExpression.check(parent) &&
                path$$1.name === "object") {
                  if (parent.object !== node) {
                      throw new Error("Nodes must be equal");
                  }
                  continue;
              }

              if (n.ConditionalExpression.check(parent) &&
                path$$1.name === "test") {
                  if (parent.test !== node) {
                      throw new Error("Nodes must be equal");
                  }
                  continue;
              }

              if (isBinary(parent) &&
                path$$1.name === "left") {
                  if (parent.left !== node) {
                      throw new Error("Nodes must be equal");
                  }
                  continue;
              }

              if (n.UnaryExpression.check(parent) &&
                !parent.prefix &&
                path$$1.name === "argument") {
                  if (parent.argument !== node) {
                      throw new Error("Nodes must be equal");
                  }
                  continue;
              }

              return false;
          }

          return true;
      }

      /**
       * Pruning certain nodes will result in empty or incomplete nodes, here we clean those nodes up.
       */
      function cleanUpNodesAfterPrune(remainingNodePath) {
          if (n.VariableDeclaration.check(remainingNodePath.node)) {
              var declarations = remainingNodePath.get('declarations').value;
              if (!declarations || declarations.length === 0) {
                  return remainingNodePath.prune();
              }
          } else if (n.ExpressionStatement.check(remainingNodePath.node)) {
              if (!remainingNodePath.get('expression').value) {
                  return remainingNodePath.prune();
              }
          } else if (n.IfStatement.check(remainingNodePath.node)) {
              cleanUpIfStatementAfterPrune(remainingNodePath);
          }

          return remainingNodePath;
      }

      function cleanUpIfStatementAfterPrune(ifStatement) {
          var testExpression = ifStatement.get('test').value;
          var alternate = ifStatement.get('alternate').value;
          var consequent = ifStatement.get('consequent').value;

          if (!consequent && !alternate) {
              var testExpressionStatement = b.expressionStatement(testExpression);

              ifStatement.replace(testExpressionStatement);
          } else if (!consequent && alternate) {
              var negatedTestExpression = b.unaryExpression('!', testExpression, true);

              if (n.UnaryExpression.check(testExpression) && testExpression.operator === '!') {
                  negatedTestExpression = testExpression.argument;
              }

              ifStatement.get("test").replace(negatedTestExpression);
              ifStatement.get("consequent").replace(alternate);
              ifStatement.get("alternate").replace();
          }
      }

      return NodePath;
  };

  var hasOwn$3 = Object.prototype.hasOwnProperty;

  var pathVisitor = function (fork) {
      var types$$1 = fork.use(types);
      var NodePath = fork.use(nodePath);
      var Printable = types$$1.namedTypes.Printable;
      var isArray = types$$1.builtInTypes.array;
      var isObject = types$$1.builtInTypes.object;
      var isFunction = types$$1.builtInTypes.function;
      var undefined;

      function PathVisitor() {
          if (!(this instanceof PathVisitor)) {
              throw new Error(
                "PathVisitor constructor cannot be invoked without 'new'"
              );
          }

          // Permanent state.
          this._reusableContextStack = [];

          this._methodNameTable = computeMethodNameTable(this);
          this._shouldVisitComments =
            hasOwn$3.call(this._methodNameTable, "Block") ||
            hasOwn$3.call(this._methodNameTable, "Line");

          this.Context = makeContextConstructor(this);

          // State reset every time PathVisitor.prototype.visit is called.
          this._visiting = false;
          this._changeReported = false;
      }

      function computeMethodNameTable(visitor) {
          var typeNames = Object.create(null);

          for (var methodName in visitor) {
              if (/^visit[A-Z]/.test(methodName)) {
                  typeNames[methodName.slice("visit".length)] = true;
              }
          }

          var supertypeTable = types$$1.computeSupertypeLookupTable(typeNames);
          var methodNameTable = Object.create(null);

          var typeNames = Object.keys(supertypeTable);
          var typeNameCount = typeNames.length;
          for (var i = 0; i < typeNameCount; ++i) {
              var typeName = typeNames[i];
              methodName = "visit" + supertypeTable[typeName];
              if (isFunction.check(visitor[methodName])) {
                  methodNameTable[typeName] = methodName;
              }
          }

          return methodNameTable;
      }

      PathVisitor.fromMethodsObject = function fromMethodsObject(methods) {
          if (methods instanceof PathVisitor) {
              return methods;
          }

          if (!isObject.check(methods)) {
              // An empty visitor?
              return new PathVisitor;
          }

          function Visitor() {
              if (!(this instanceof Visitor)) {
                  throw new Error(
                    "Visitor constructor cannot be invoked without 'new'"
                  );
              }
              PathVisitor.call(this);
          }

          var Vp = Visitor.prototype = Object.create(PVp);
          Vp.constructor = Visitor;

          extend(Vp, methods);
          extend(Visitor, PathVisitor);

          isFunction.assert(Visitor.fromMethodsObject);
          isFunction.assert(Visitor.visit);

          return new Visitor;
      };

      function extend(target, source) {
          for (var property in source) {
              if (hasOwn$3.call(source, property)) {
                  target[property] = source[property];
              }
          }

          return target;
      }

      PathVisitor.visit = function visit(node, methods) {
          return PathVisitor.fromMethodsObject(methods).visit(node);
      };

      var PVp = PathVisitor.prototype;

      PVp.visit = function () {
          if (this._visiting) {
              throw new Error(
                "Recursively calling visitor.visit(path) resets visitor state. " +
                "Try this.visit(path) or this.traverse(path) instead."
              );
          }

          // Private state that needs to be reset before every traversal.
          this._visiting = true;
          this._changeReported = false;
          this._abortRequested = false;

          var argc = arguments.length;
          var args = new Array(argc);
          for (var i = 0; i < argc; ++i) {
              args[i] = arguments[i];
          }

          if (!(args[0] instanceof NodePath)) {
              args[0] = new NodePath({root: args[0]}).get("root");
          }

          // Called with the same arguments as .visit.
          this.reset.apply(this, args);

          try {
              var root = this.visitWithoutReset(args[0]);
              var didNotThrow = true;
          } finally {
              this._visiting = false;

              if (!didNotThrow && this._abortRequested) {
                  // If this.visitWithoutReset threw an exception and
                  // this._abortRequested was set to true, return the root of
                  // the AST instead of letting the exception propagate, so that
                  // client code does not have to provide a try-catch block to
                  // intercept the AbortRequest exception.  Other kinds of
                  // exceptions will propagate without being intercepted and
                  // rethrown by a catch block, so their stacks will accurately
                  // reflect the original throwing context.
                  return args[0].value;
              }
          }

          return root;
      };

      PVp.AbortRequest = function AbortRequest() {};
      PVp.abort = function () {
          var visitor = this;
          visitor._abortRequested = true;
          var request = new visitor.AbortRequest();

          // If you decide to catch this exception and stop it from propagating,
          // make sure to call its cancel method to avoid silencing other
          // exceptions that might be thrown later in the traversal.
          request.cancel = function () {
              visitor._abortRequested = false;
          };

          throw request;
      };

      PVp.reset = function (path/*, additional arguments */) {
          // Empty stub; may be reassigned or overridden by subclasses.
      };

      PVp.visitWithoutReset = function (path) {
          if (this instanceof this.Context) {
              // Since this.Context.prototype === this, there's a chance we
              // might accidentally call context.visitWithoutReset. If that
              // happens, re-invoke the method against context.visitor.
              return this.visitor.visitWithoutReset(path);
          }

          if (!(path instanceof NodePath)) {
              throw new Error("");
          }

          var value = path.value;

          var methodName = value &&
            typeof value === "object" &&
            typeof value.type === "string" &&
            this._methodNameTable[value.type];

          if (methodName) {
              var context = this.acquireContext(path);
              try {
                  return context.invokeVisitorMethod(methodName);
              } finally {
                  this.releaseContext(context);
              }

          } else {
              // If there was no visitor method to call, visit the children of
              // this node generically.
              return visitChildren(path, this);
          }
      };

      function visitChildren(path, visitor) {
          if (!(path instanceof NodePath)) {
              throw new Error("");
          }
          if (!(visitor instanceof PathVisitor)) {
              throw new Error("");
          }

          var value = path.value;

          if (isArray.check(value)) {
              path.each(visitor.visitWithoutReset, visitor);
          } else if (!isObject.check(value)) ; else {
              var childNames = types$$1.getFieldNames(value);

              // The .comments field of the Node type is hidden, so we only
              // visit it if the visitor defines visitBlock or visitLine, and
              // value.comments is defined.
              if (visitor._shouldVisitComments &&
                value.comments &&
                childNames.indexOf("comments") < 0) {
                  childNames.push("comments");
              }

              var childCount = childNames.length;
              var childPaths = [];

              for (var i = 0; i < childCount; ++i) {
                  var childName = childNames[i];
                  if (!hasOwn$3.call(value, childName)) {
                      value[childName] = types$$1.getFieldValue(value, childName);
                  }
                  childPaths.push(path.get(childName));
              }

              for (var i = 0; i < childCount; ++i) {
                  visitor.visitWithoutReset(childPaths[i]);
              }
          }

          return path.value;
      }

      PVp.acquireContext = function (path) {
          if (this._reusableContextStack.length === 0) {
              return new this.Context(path);
          }
          return this._reusableContextStack.pop().reset(path);
      };

      PVp.releaseContext = function (context) {
          if (!(context instanceof this.Context)) {
              throw new Error("");
          }
          this._reusableContextStack.push(context);
          context.currentPath = null;
      };

      PVp.reportChanged = function () {
          this._changeReported = true;
      };

      PVp.wasChangeReported = function () {
          return this._changeReported;
      };

      function makeContextConstructor(visitor) {
          function Context(path) {
              if (!(this instanceof Context)) {
                  throw new Error("");
              }
              if (!(this instanceof PathVisitor)) {
                  throw new Error("");
              }
              if (!(path instanceof NodePath)) {
                  throw new Error("");
              }

              Object.defineProperty(this, "visitor", {
                  value: visitor,
                  writable: false,
                  enumerable: true,
                  configurable: false
              });

              this.currentPath = path;
              this.needToCallTraverse = true;

              Object.seal(this);
          }

          if (!(visitor instanceof PathVisitor)) {
              throw new Error("");
          }

          // Note that the visitor object is the prototype of Context.prototype,
          // so all visitor methods are inherited by context objects.
          var Cp = Context.prototype = Object.create(visitor);

          Cp.constructor = Context;
          extend(Cp, sharedContextProtoMethods);

          return Context;
      }

  // Every PathVisitor has a different this.Context constructor and
  // this.Context.prototype object, but those prototypes can all use the
  // same reset, invokeVisitorMethod, and traverse function objects.
      var sharedContextProtoMethods = Object.create(null);

      sharedContextProtoMethods.reset =
        function reset(path) {
            if (!(this instanceof this.Context)) {
                throw new Error("");
            }
            if (!(path instanceof NodePath)) {
                throw new Error("");
            }

            this.currentPath = path;
            this.needToCallTraverse = true;

            return this;
        };

      sharedContextProtoMethods.invokeVisitorMethod =
        function invokeVisitorMethod(methodName) {
            if (!(this instanceof this.Context)) {
                throw new Error("");
            }
            if (!(this.currentPath instanceof NodePath)) {
                throw new Error("");
            }

            var result = this.visitor[methodName].call(this, this.currentPath);

            if (result === false) {
                // Visitor methods return false to indicate that they have handled
                // their own traversal needs, and we should not complain if
                // this.needToCallTraverse is still true.
                this.needToCallTraverse = false;

            } else if (result !== undefined) {
                // Any other non-undefined value returned from the visitor method
                // is interpreted as a replacement value.
                this.currentPath = this.currentPath.replace(result)[0];

                if (this.needToCallTraverse) {
                    // If this.traverse still hasn't been called, visit the
                    // children of the replacement node.
                    this.traverse(this.currentPath);
                }
            }

            if (this.needToCallTraverse !== false) {
                throw new Error(
                  "Must either call this.traverse or return false in " + methodName
                );
            }

            var path = this.currentPath;
            return path && path.value;
        };

      sharedContextProtoMethods.traverse =
        function traverse(path, newVisitor) {
            if (!(this instanceof this.Context)) {
                throw new Error("");
            }
            if (!(path instanceof NodePath)) {
                throw new Error("");
            }
            if (!(this.currentPath instanceof NodePath)) {
                throw new Error("");
            }

            this.needToCallTraverse = false;

            return visitChildren(path, PathVisitor.fromMethodsObject(
              newVisitor || this.visitor
            ));
        };

      sharedContextProtoMethods.visit =
        function visit(path, newVisitor) {
            if (!(this instanceof this.Context)) {
                throw new Error("");
            }
            if (!(path instanceof NodePath)) {
                throw new Error("");
            }
            if (!(this.currentPath instanceof NodePath)) {
                throw new Error("");
            }

            this.needToCallTraverse = false;

            return PathVisitor.fromMethodsObject(
              newVisitor || this.visitor
            ).visitWithoutReset(path);
        };

      sharedContextProtoMethods.reportChanged = function reportChanged() {
          this.visitor.reportChanged();
      };

      sharedContextProtoMethods.abort = function abort() {
          this.needToCallTraverse = false;
          this.visitor.abort();
      };

      return PathVisitor;
  };

  var fork = function (defs) {
      var used = [];
      var usedResult = [];
      var fork = {};

      function use(plugin) {
          var idx = used.indexOf(plugin);
          if (idx === -1) {
              idx = used.length;
              used.push(plugin);
              usedResult[idx] = plugin(fork);
          }
          return usedResult[idx];
      }

      fork.use = use;

      var types$$1 = use(types);

      defs.forEach(use);

      types$$1.finalize();

      var exports = {
          Type: types$$1.Type,
          builtInTypes: types$$1.builtInTypes,
          namedTypes: types$$1.namedTypes,
          builders: types$$1.builders,
          defineMethod: types$$1.defineMethod,
          getFieldNames: types$$1.getFieldNames,
          getFieldValue: types$$1.getFieldValue,
          eachField: types$$1.eachField,
          someField: types$$1.someField,
          getSupertypeNames: types$$1.getSupertypeNames,
          astNodesAreEquivalent: use(equiv),
          finalize: types$$1.finalize,
          Path: use(path),
          NodePath: use(nodePath),
          PathVisitor: use(pathVisitor),
          use: use
      };

      exports.visit = exports.PathVisitor.visit;

      return exports;
  };

  var shared = function (fork) {
      var exports = {};
      var types$$1 = fork.use(types);
      var Type = types$$1.Type;
      var builtin = types$$1.builtInTypes;
      var isNumber = builtin.number;

      // An example of constructing a new type with arbitrary constraints from
      // an existing type.
      exports.geq = function (than) {
          return new Type(function (value) {
              return isNumber.check(value) && value >= than;
          }, isNumber + " >= " + than);
      };

      // Default value-returning functions that may optionally be passed as a
      // third argument to Def.prototype.field.
      exports.defaults = {
          // Functions were used because (among other reasons) that's the most
          // elegant way to allow for the emptyArray one always to give a new
          // array instance.
          "null": function () { return null },
          "emptyArray": function () { return [] },
          "false": function () { return false },
          "true": function () { return true },
          "undefined": function () {}
      };

      var naiveIsPrimitive = Type.or(
        builtin.string,
        builtin.number,
        builtin.boolean,
        builtin.null,
        builtin.undefined
      );

      exports.isPrimitive = new Type(function (value) {
          if (value === null)
              return true;
          var type = typeof value;
          return !(type === "object" ||
          type === "function");
      }, naiveIsPrimitive.toString());

      return exports;
  };

  var core = function (fork) {
      var types$$1 = fork.use(types);
      var Type = types$$1.Type;
      var def = Type.def;
      var or = Type.or;
      var shared$$1 = fork.use(shared);
      var defaults = shared$$1.defaults;
      var geq = shared$$1.geq;

      // Abstract supertype of all syntactic entities that are allowed to have a
      // .loc field.
      def("Printable")
          .field("loc", or(
              def("SourceLocation"),
              null
          ), defaults["null"], true);

      def("Node")
          .bases("Printable")
          .field("type", String)
          .field("comments", or(
              [def("Comment")],
              null
          ), defaults["null"], true);

      def("SourceLocation")
          .build("start", "end", "source")
          .field("start", def("Position"))
          .field("end", def("Position"))
          .field("source", or(String, null), defaults["null"]);

      def("Position")
          .build("line", "column")
          .field("line", geq(1))
          .field("column", geq(0));

      def("File")
          .bases("Node")
          .build("program", "name")
          .field("program", def("Program"))
          .field("name", or(String, null), defaults["null"]);

      def("Program")
          .bases("Node")
          .build("body")
          .field("body", [def("Statement")]);

      def("Function")
          .bases("Node")
          .field("id", or(def("Identifier"), null), defaults["null"])
          .field("params", [def("Pattern")])
          .field("body", def("BlockStatement"));

      def("Statement").bases("Node");

  // The empty .build() here means that an EmptyStatement can be constructed
  // (i.e. it's not abstract) but that it needs no arguments.
      def("EmptyStatement").bases("Statement").build();

      def("BlockStatement")
          .bases("Statement")
          .build("body")
          .field("body", [def("Statement")]);

      // TODO Figure out how to silently coerce Expressions to
      // ExpressionStatements where a Statement was expected.
      def("ExpressionStatement")
          .bases("Statement")
          .build("expression")
          .field("expression", def("Expression"));

      def("IfStatement")
          .bases("Statement")
          .build("test", "consequent", "alternate")
          .field("test", def("Expression"))
          .field("consequent", def("Statement"))
          .field("alternate", or(def("Statement"), null), defaults["null"]);

      def("LabeledStatement")
          .bases("Statement")
          .build("label", "body")
          .field("label", def("Identifier"))
          .field("body", def("Statement"));

      def("BreakStatement")
          .bases("Statement")
          .build("label")
          .field("label", or(def("Identifier"), null), defaults["null"]);

      def("ContinueStatement")
          .bases("Statement")
          .build("label")
          .field("label", or(def("Identifier"), null), defaults["null"]);

      def("WithStatement")
          .bases("Statement")
          .build("object", "body")
          .field("object", def("Expression"))
          .field("body", def("Statement"));

      def("SwitchStatement")
          .bases("Statement")
          .build("discriminant", "cases", "lexical")
          .field("discriminant", def("Expression"))
          .field("cases", [def("SwitchCase")])
          .field("lexical", Boolean, defaults["false"]);

      def("ReturnStatement")
          .bases("Statement")
          .build("argument")
          .field("argument", or(def("Expression"), null));

      def("ThrowStatement")
          .bases("Statement")
          .build("argument")
          .field("argument", def("Expression"));

      def("TryStatement")
          .bases("Statement")
          .build("block", "handler", "finalizer")
          .field("block", def("BlockStatement"))
          .field("handler", or(def("CatchClause"), null), function () {
              return this.handlers && this.handlers[0] || null;
          })
          .field("handlers", [def("CatchClause")], function () {
              return this.handler ? [this.handler] : [];
          }, true) // Indicates this field is hidden from eachField iteration.
          .field("guardedHandlers", [def("CatchClause")], defaults.emptyArray)
          .field("finalizer", or(def("BlockStatement"), null), defaults["null"]);

      def("CatchClause")
          .bases("Node")
          .build("param", "guard", "body")
          // https://github.com/tc39/proposal-optional-catch-binding
          .field("param", or(def("Pattern"), null), defaults["null"])
          .field("guard", or(def("Expression"), null), defaults["null"])
          .field("body", def("BlockStatement"));

      def("WhileStatement")
          .bases("Statement")
          .build("test", "body")
          .field("test", def("Expression"))
          .field("body", def("Statement"));

      def("DoWhileStatement")
          .bases("Statement")
          .build("body", "test")
          .field("body", def("Statement"))
          .field("test", def("Expression"));

      def("ForStatement")
          .bases("Statement")
          .build("init", "test", "update", "body")
          .field("init", or(
              def("VariableDeclaration"),
              def("Expression"),
              null))
          .field("test", or(def("Expression"), null))
          .field("update", or(def("Expression"), null))
          .field("body", def("Statement"));

      def("ForInStatement")
          .bases("Statement")
          .build("left", "right", "body")
          .field("left", or(
              def("VariableDeclaration"),
              def("Expression")))
          .field("right", def("Expression"))
          .field("body", def("Statement"));

      def("DebuggerStatement").bases("Statement").build();

      def("Declaration").bases("Statement");

      def("FunctionDeclaration")
          .bases("Function", "Declaration")
          .build("id", "params", "body")
          .field("id", def("Identifier"));

      def("FunctionExpression")
          .bases("Function", "Expression")
          .build("id", "params", "body");

      def("VariableDeclaration")
          .bases("Declaration")
          .build("kind", "declarations")
          .field("kind", or("var", "let", "const"))
          .field("declarations", [def("VariableDeclarator")]);

      def("VariableDeclarator")
          .bases("Node")
          .build("id", "init")
          .field("id", def("Pattern"))
          .field("init", or(def("Expression"), null));

      // TODO Are all Expressions really Patterns?
      def("Expression").bases("Node", "Pattern");

      def("ThisExpression").bases("Expression").build();

      def("ArrayExpression")
          .bases("Expression")
          .build("elements")
          .field("elements", [or(def("Expression"), null)]);

      def("ObjectExpression")
          .bases("Expression")
          .build("properties")
          .field("properties", [def("Property")]);

      // TODO Not in the Mozilla Parser API, but used by Esprima.
      def("Property")
          .bases("Node") // Want to be able to visit Property Nodes.
          .build("kind", "key", "value")
          .field("kind", or("init", "get", "set"))
          .field("key", or(def("Literal"), def("Identifier")))
          .field("value", def("Expression"));

      def("SequenceExpression")
          .bases("Expression")
          .build("expressions")
          .field("expressions", [def("Expression")]);

      var UnaryOperator = or(
          "-", "+", "!", "~",
          "typeof", "void", "delete");

      def("UnaryExpression")
          .bases("Expression")
          .build("operator", "argument", "prefix")
          .field("operator", UnaryOperator)
          .field("argument", def("Expression"))
          // Esprima doesn't bother with this field, presumably because it's
          // always true for unary operators.
          .field("prefix", Boolean, defaults["true"]);

      var BinaryOperator = or(
          "==", "!=", "===", "!==",
          "<", "<=", ">", ">=",
          "<<", ">>", ">>>",
          "+", "-", "*", "/", "%", "**",
          "&", // TODO Missing from the Parser API.
          "|", "^", "in",
          "instanceof", "..");

      def("BinaryExpression")
          .bases("Expression")
          .build("operator", "left", "right")
          .field("operator", BinaryOperator)
          .field("left", def("Expression"))
          .field("right", def("Expression"));

      var AssignmentOperator = or(
          "=", "+=", "-=", "*=", "/=", "%=",
          "<<=", ">>=", ">>>=",
          "|=", "^=", "&=");

      def("AssignmentExpression")
          .bases("Expression")
          .build("operator", "left", "right")
          .field("operator", AssignmentOperator)
          .field("left", def("Pattern"))
          .field("right", def("Expression"));

      var UpdateOperator = or("++", "--");

      def("UpdateExpression")
          .bases("Expression")
          .build("operator", "argument", "prefix")
          .field("operator", UpdateOperator)
          .field("argument", def("Expression"))
          .field("prefix", Boolean);

      var LogicalOperator = or("||", "&&");

      def("LogicalExpression")
          .bases("Expression")
          .build("operator", "left", "right")
          .field("operator", LogicalOperator)
          .field("left", def("Expression"))
          .field("right", def("Expression"));

      def("ConditionalExpression")
          .bases("Expression")
          .build("test", "consequent", "alternate")
          .field("test", def("Expression"))
          .field("consequent", def("Expression"))
          .field("alternate", def("Expression"));

      def("NewExpression")
          .bases("Expression")
          .build("callee", "arguments")
          .field("callee", def("Expression"))
          // The Mozilla Parser API gives this type as [or(def("Expression"),
          // null)], but null values don't really make sense at the call site.
          // TODO Report this nonsense.
          .field("arguments", [def("Expression")]);

      def("CallExpression")
          .bases("Expression")
          .build("callee", "arguments")
          .field("callee", def("Expression"))
          // See comment for NewExpression above.
          .field("arguments", [def("Expression")]);

      def("MemberExpression")
          .bases("Expression")
          .build("object", "property", "computed")
          .field("object", def("Expression"))
          .field("property", or(def("Identifier"), def("Expression")))
          .field("computed", Boolean, function () {
              var type = this.property.type;
              if (type === 'Literal' ||
                  type === 'MemberExpression' ||
                  type === 'BinaryExpression') {
                  return true;
              }
              return false;
          });

      def("Pattern").bases("Node");

      def("SwitchCase")
          .bases("Node")
          .build("test", "consequent")
          .field("test", or(def("Expression"), null))
          .field("consequent", [def("Statement")]);

      def("Identifier")
          // But aren't Expressions and Patterns already Nodes? TODO Report this.
          .bases("Node", "Expression", "Pattern")
          .build("name")
          .field("name", String)
          .field("optional", Boolean, defaults["false"]);

      def("Literal")
          // But aren't Expressions already Nodes? TODO Report this.
          .bases("Node", "Expression")
          .build("value")
          .field("value", or(String, Boolean, null, Number, RegExp))
          .field("regex", or({
              pattern: String,
              flags: String
          }, null), function () {
              if (this.value instanceof RegExp) {
                  var flags = "";

                  if (this.value.ignoreCase) flags += "i";
                  if (this.value.multiline) flags += "m";
                  if (this.value.global) flags += "g";

                  return {
                      pattern: this.value.source,
                      flags: flags
                  };
              }

              return null;
          });

      // Abstract (non-buildable) comment supertype. Not a Node.
      def("Comment")
          .bases("Printable")
          .field("value", String)
          // A .leading comment comes before the node, whereas a .trailing
          // comment comes after it. These two fields should not both be true,
          // but they might both be false when the comment falls inside a node
          // and the node has no children for the comment to lead or trail,
          // e.g. { /*dangling*/ }.
          .field("leading", Boolean, defaults["true"])
          .field("trailing", Boolean, defaults["false"]);
  };

  var es6 = function (fork) {
    fork.use(core);
    var types$$1 = fork.use(types);
    var def = types$$1.Type.def;
    var or = types$$1.Type.or;
    var defaults = fork.use(shared).defaults;

    def("Function")
      .field("generator", Boolean, defaults["false"])
      .field("expression", Boolean, defaults["false"])
      .field("defaults", [or(def("Expression"), null)], defaults.emptyArray)
      // TODO This could be represented as a RestElement in .params.
      .field("rest", or(def("Identifier"), null), defaults["null"]);

    // The ESTree way of representing a ...rest parameter.
    def("RestElement")
      .bases("Pattern")
      .build("argument")
      .field("argument", def("Pattern"))
      .field("typeAnnotation", // for Babylon. Flow parser puts it on the identifier
        or(def("TypeAnnotation"), def("TSTypeAnnotation"), null), defaults["null"]);

    def("SpreadElementPattern")
      .bases("Pattern")
      .build("argument")
      .field("argument", def("Pattern"));

    def("FunctionDeclaration")
      .build("id", "params", "body", "generator", "expression");

    def("FunctionExpression")
      .build("id", "params", "body", "generator", "expression");

    // The Parser API calls this ArrowExpression, but Esprima and all other
    // actual parsers use ArrowFunctionExpression.
    def("ArrowFunctionExpression")
      .bases("Function", "Expression")
      .build("params", "body", "expression")
      // The forced null value here is compatible with the overridden
      // definition of the "id" field in the Function interface.
      .field("id", null, defaults["null"])
      // Arrow function bodies are allowed to be expressions.
      .field("body", or(def("BlockStatement"), def("Expression")))
      // The current spec forbids arrow generators, so I have taken the
      // liberty of enforcing that. TODO Report this.
      .field("generator", false, defaults["false"]);

    def("ForOfStatement")
      .bases("Statement")
      .build("left", "right", "body")
      .field("left", or(
        def("VariableDeclaration"),
        def("Pattern")))
      .field("right", def("Expression"))
      .field("body", def("Statement"));

    def("YieldExpression")
      .bases("Expression")
      .build("argument", "delegate")
      .field("argument", or(def("Expression"), null))
      .field("delegate", Boolean, defaults["false"]);

    def("GeneratorExpression")
      .bases("Expression")
      .build("body", "blocks", "filter")
      .field("body", def("Expression"))
      .field("blocks", [def("ComprehensionBlock")])
      .field("filter", or(def("Expression"), null));

    def("ComprehensionExpression")
      .bases("Expression")
      .build("body", "blocks", "filter")
      .field("body", def("Expression"))
      .field("blocks", [def("ComprehensionBlock")])
      .field("filter", or(def("Expression"), null));

    def("ComprehensionBlock")
      .bases("Node")
      .build("left", "right", "each")
      .field("left", def("Pattern"))
      .field("right", def("Expression"))
      .field("each", Boolean);

    def("Property")
      .field("key", or(def("Literal"), def("Identifier"), def("Expression")))
      .field("value", or(def("Expression"), def("Pattern")))
      .field("method", Boolean, defaults["false"])
      .field("shorthand", Boolean, defaults["false"])
      .field("computed", Boolean, defaults["false"]);

    def("PropertyPattern")
      .bases("Pattern")
      .build("key", "pattern")
      .field("key", or(def("Literal"), def("Identifier"), def("Expression")))
      .field("pattern", def("Pattern"))
      .field("computed", Boolean, defaults["false"]);

    def("ObjectPattern")
      .bases("Pattern")
      .build("properties")
      .field("properties", [or(def("PropertyPattern"), def("Property"))]);

    def("ArrayPattern")
      .bases("Pattern")
      .build("elements")
      .field("elements", [or(def("Pattern"), null)]);

    def("MethodDefinition")
      .bases("Declaration")
      .build("kind", "key", "value", "static")
      .field("kind", or("constructor", "method", "get", "set"))
      .field("key", def("Expression"))
      .field("value", def("Function"))
      .field("computed", Boolean, defaults["false"])
      .field("static", Boolean, defaults["false"]);

    def("SpreadElement")
      .bases("Node")
      .build("argument")
      .field("argument", def("Expression"));

    def("ArrayExpression")
      .field("elements", [or(
        def("Expression"),
        def("SpreadElement"),
        def("RestElement"),
        null
      )]);

    def("NewExpression")
      .field("arguments", [or(def("Expression"), def("SpreadElement"))]);

    def("CallExpression")
      .field("arguments", [or(def("Expression"), def("SpreadElement"))]);

    // Note: this node type is *not* an AssignmentExpression with a Pattern on
    // the left-hand side! The existing AssignmentExpression type already
    // supports destructuring assignments. AssignmentPattern nodes may appear
    // wherever a Pattern is allowed, and the right-hand side represents a
    // default value to be destructured against the left-hand side, if no
    // value is otherwise provided. For example: default parameter values.
    def("AssignmentPattern")
      .bases("Pattern")
      .build("left", "right")
      .field("left", def("Pattern"))
      .field("right", def("Expression"));

    var ClassBodyElement = or(
      def("MethodDefinition"),
      def("VariableDeclarator"),
      def("ClassPropertyDefinition"),
      def("ClassProperty")
    );

    def("ClassProperty")
      .bases("Declaration")
      .build("key")
      .field("key", or(def("Literal"), def("Identifier"), def("Expression")))
      .field("computed", Boolean, defaults["false"]);

    def("ClassPropertyDefinition") // static property
      .bases("Declaration")
      .build("definition")
      // Yes, Virginia, circular definitions are permitted.
      .field("definition", ClassBodyElement);

    def("ClassBody")
      .bases("Declaration")
      .build("body")
      .field("body", [ClassBodyElement]);

    def("ClassDeclaration")
      .bases("Declaration")
      .build("id", "body", "superClass")
      .field("id", or(def("Identifier"), null))
      .field("body", def("ClassBody"))
      .field("superClass", or(def("Expression"), null), defaults["null"]);

    def("ClassExpression")
      .bases("Expression")
      .build("id", "body", "superClass")
      .field("id", or(def("Identifier"), null), defaults["null"])
      .field("body", def("ClassBody"))
      .field("superClass", or(def("Expression"), null), defaults["null"]);

    // Specifier and ModuleSpecifier are abstract non-standard types
    // introduced for definitional convenience.
    def("Specifier").bases("Node");

    // This supertype is shared/abused by both def/babel.js and
    // def/esprima.js. In the future, it will be possible to load only one set
    // of definitions appropriate for a given parser, but until then we must
    // rely on default functions to reconcile the conflicting AST formats.
    def("ModuleSpecifier")
      .bases("Specifier")
      // This local field is used by Babel/Acorn. It should not technically
      // be optional in the Babel/Acorn AST format, but it must be optional
      // in the Esprima AST format.
      .field("local", or(def("Identifier"), null), defaults["null"])
      // The id and name fields are used by Esprima. The id field should not
      // technically be optional in the Esprima AST format, but it must be
      // optional in the Babel/Acorn AST format.
      .field("id", or(def("Identifier"), null), defaults["null"])
      .field("name", or(def("Identifier"), null), defaults["null"]);

    // Like ModuleSpecifier, except type:"ImportSpecifier" and buildable.
    // import {<id [as name]>} from ...;
    def("ImportSpecifier")
      .bases("ModuleSpecifier")
      .build("id", "name");

    // import <* as id> from ...;
    def("ImportNamespaceSpecifier")
      .bases("ModuleSpecifier")
      .build("id");

    // import <id> from ...;
    def("ImportDefaultSpecifier")
      .bases("ModuleSpecifier")
      .build("id");

    def("ImportDeclaration")
      .bases("Declaration")
      .build("specifiers", "source", "importKind")
      .field("specifiers", [or(
        def("ImportSpecifier"),
        def("ImportNamespaceSpecifier"),
        def("ImportDefaultSpecifier")
      )], defaults.emptyArray)
      .field("source", def("Literal"))
      .field("importKind", or(
        "value",
        "type"
      ), function() {
        return "value";
      });

    def("TaggedTemplateExpression")
      .bases("Expression")
      .build("tag", "quasi")
      .field("tag", def("Expression"))
      .field("quasi", def("TemplateLiteral"));

    def("TemplateLiteral")
      .bases("Expression")
      .build("quasis", "expressions")
      .field("quasis", [def("TemplateElement")])
      .field("expressions", [def("Expression")]);

    def("TemplateElement")
      .bases("Node")
      .build("value", "tail")
      .field("value", {"cooked": String, "raw": String})
      .field("tail", Boolean);
  };

  var es7 = function (fork) {
    fork.use(es6);

    var types$$1 = fork.use(types);
    var def = types$$1.Type.def;
    var or = types$$1.Type.or;
    var builtin = types$$1.builtInTypes;
    var defaults = fork.use(shared).defaults;

    def("Function")
      .field("async", Boolean, defaults["false"]);

    def("SpreadProperty")
      .bases("Node")
      .build("argument")
      .field("argument", def("Expression"));

    def("ObjectExpression")
      .field("properties", [or(
        def("Property"),
        def("SpreadProperty"),
        def("SpreadElement")
      )]);

    def("SpreadPropertyPattern")
      .bases("Pattern")
      .build("argument")
      .field("argument", def("Pattern"));

    def("ObjectPattern")
      .field("properties", [or(
        def("Property"),
        def("PropertyPattern"),
        def("SpreadPropertyPattern")
      )]);

    def("AwaitExpression")
      .bases("Expression")
      .build("argument", "all")
      .field("argument", or(def("Expression"), null))
      .field("all", Boolean, defaults["false"]);
  };

  var mozilla = function (fork) {
      fork.use(core);
      var types$$1 = fork.use(types);
      var def = types$$1.Type.def;
      var or = types$$1.Type.or;
      var shared$$1 = fork.use(shared);
      var geq = shared$$1.geq;
      var defaults = shared$$1.defaults;

      def("Function")
          // SpiderMonkey allows expression closures: function(x) x+1
          .field("body", or(def("BlockStatement"), def("Expression")));

      def("ForInStatement")
          .build("left", "right", "body", "each")
          .field("each", Boolean, defaults["false"]);

      def("LetStatement")
          .bases("Statement")
          .build("head", "body")
          // TODO Deviating from the spec by reusing VariableDeclarator here.
          .field("head", [def("VariableDeclarator")])
          .field("body", def("Statement"));

      def("LetExpression")
          .bases("Expression")
          .build("head", "body")
          // TODO Deviating from the spec by reusing VariableDeclarator here.
          .field("head", [def("VariableDeclarator")])
          .field("body", def("Expression"));

      def("GraphExpression")
          .bases("Expression")
          .build("index", "expression")
          .field("index", geq(0))
          .field("expression", def("Literal"));

      def("GraphIndexExpression")
          .bases("Expression")
          .build("index")
          .field("index", geq(0));
  };

  var e4x = function (fork) {
      fork.use(core);
      var types$$1 = fork.use(types);
      var def = types$$1.Type.def;
      var or = types$$1.Type.or;

      // Note that none of these types are buildable because the Mozilla Parser
      // API doesn't specify any builder functions, and nobody uses E4X anymore.

      def("XMLDefaultDeclaration")
          .bases("Declaration")
          .field("namespace", def("Expression"));

      def("XMLAnyName").bases("Expression");

      def("XMLQualifiedIdentifier")
          .bases("Expression")
          .field("left", or(def("Identifier"), def("XMLAnyName")))
          .field("right", or(def("Identifier"), def("Expression")))
          .field("computed", Boolean);

      def("XMLFunctionQualifiedIdentifier")
          .bases("Expression")
          .field("right", or(def("Identifier"), def("Expression")))
          .field("computed", Boolean);

      def("XMLAttributeSelector")
          .bases("Expression")
          .field("attribute", def("Expression"));

      def("XMLFilterExpression")
          .bases("Expression")
          .field("left", def("Expression"))
          .field("right", def("Expression"));

      def("XMLElement")
          .bases("XML", "Expression")
          .field("contents", [def("XML")]);

      def("XMLList")
          .bases("XML", "Expression")
          .field("contents", [def("XML")]);

      def("XML").bases("Node");

      def("XMLEscape")
          .bases("XML")
          .field("expression", def("Expression"));

      def("XMLText")
          .bases("XML")
          .field("text", String);

      def("XMLStartTag")
          .bases("XML")
          .field("contents", [def("XML")]);

      def("XMLEndTag")
          .bases("XML")
          .field("contents", [def("XML")]);

      def("XMLPointTag")
          .bases("XML")
          .field("contents", [def("XML")]);

      def("XMLName")
          .bases("XML")
          .field("contents", or(String, [def("XML")]));

      def("XMLAttribute")
          .bases("XML")
          .field("value", String);

      def("XMLCdata")
          .bases("XML")
          .field("contents", String);

      def("XMLComment")
          .bases("XML")
          .field("contents", String);

      def("XMLProcessingInstruction")
          .bases("XML")
          .field("target", String)
          .field("contents", or(String, null));
  };

  var jsx = function (fork) {
    fork.use(es7);

    var types$$1 = fork.use(types);
    var def = types$$1.Type.def;
    var or = types$$1.Type.or;
    var defaults = fork.use(shared).defaults;

    def("JSXAttribute")
      .bases("Node")
      .build("name", "value")
      .field("name", or(def("JSXIdentifier"), def("JSXNamespacedName")))
      .field("value", or(
        def("Literal"), // attr="value"
        def("JSXExpressionContainer"), // attr={value}
        null // attr= or just attr
      ), defaults["null"]);

    def("JSXIdentifier")
      .bases("Identifier")
      .build("name")
      .field("name", String);

    def("JSXNamespacedName")
      .bases("Node")
      .build("namespace", "name")
      .field("namespace", def("JSXIdentifier"))
      .field("name", def("JSXIdentifier"));

    def("JSXMemberExpression")
      .bases("MemberExpression")
      .build("object", "property")
      .field("object", or(def("JSXIdentifier"), def("JSXMemberExpression")))
      .field("property", def("JSXIdentifier"))
      .field("computed", Boolean, defaults.false);

    var JSXElementName = or(
      def("JSXIdentifier"),
      def("JSXNamespacedName"),
      def("JSXMemberExpression")
    );

    def("JSXSpreadAttribute")
      .bases("Node")
      .build("argument")
      .field("argument", def("Expression"));

    var JSXAttributes = [or(
      def("JSXAttribute"),
      def("JSXSpreadAttribute")
    )];

    def("JSXExpressionContainer")
      .bases("Expression")
      .build("expression")
      .field("expression", def("Expression"));

    def("JSXElement")
      .bases("Expression")
      .build("openingElement", "closingElement", "children")
      .field("openingElement", def("JSXOpeningElement"))
      .field("closingElement", or(def("JSXClosingElement"), null), defaults["null"])
      .field("children", [or(
        def("JSXElement"),
        def("JSXExpressionContainer"),
        def("JSXFragment"),
        def("JSXText"),
        def("Literal") // TODO Esprima should return JSXText instead.
      )], defaults.emptyArray)
      .field("name", JSXElementName, function () {
        // Little-known fact: the `this` object inside a default function
        // is none other than the partially-built object itself, and any
        // fields initialized directly from builder function arguments
        // (like openingElement, closingElement, and children) are
        // guaranteed to be available.
        return this.openingElement.name;
      }, true) // hidden from traversal
      .field("selfClosing", Boolean, function () {
        return this.openingElement.selfClosing;
      }, true) // hidden from traversal
      .field("attributes", JSXAttributes, function () {
        return this.openingElement.attributes;
      }, true); // hidden from traversal

    def("JSXOpeningElement")
      .bases("Node") // TODO Does this make sense? Can't really be an JSXElement.
      .build("name", "attributes", "selfClosing")
      .field("name", JSXElementName)
      .field("attributes", JSXAttributes, defaults.emptyArray)
      .field("selfClosing", Boolean, defaults["false"]);

    def("JSXClosingElement")
      .bases("Node") // TODO Same concern.
      .build("name")
      .field("name", JSXElementName);

    def("JSXFragment")
      .bases("Expression")
      .build("openingElement", "closingElement", "children")
      .field("openingElement", def("JSXOpeningFragment"))
      .field("closingElement", def("JSXClosingFragment"))
      .field("children", [or(
        def("JSXElement"),
        def("JSXExpressionContainer"),
        def("JSXFragment"),
        def("JSXText"),
        def("Literal") // TODO Esprima should return JSXText instead.
      )], defaults.emptyArray);

    def("JSXOpeningFragment")
      .bases("Node") // TODO Same concern.
      .build();

    def("JSXClosingFragment")
      .bases("Node") // TODO Same concern.
      .build();

    def("JSXText")
      .bases("Literal")
      .build("value")
      .field("value", String);

    def("JSXEmptyExpression").bases("Expression").build();

    // This PR has caused many people issues, but supporting it seems like a
    // good idea anyway: https://github.com/babel/babel/pull/4988
    def("JSXSpreadChild")
      .bases("Expression")
      .build("expression")
      .field("expression", def("Expression"));
  };

  var flow = function (fork) {
    fork.use(es7);

    var types$$1 = fork.use(types);
    var def = types$$1.Type.def;
    var or = types$$1.Type.or;
    var defaults = fork.use(shared).defaults;

    // Base types

    def("Flow").bases("Node");
    def("FlowType").bases("Flow");

    // Type annotations

    def("AnyTypeAnnotation")
      .bases("FlowType")
      .build();

    def("EmptyTypeAnnotation")
      .bases("FlowType")
      .build();

    def("MixedTypeAnnotation")
      .bases("FlowType")
      .build();

    def("VoidTypeAnnotation")
      .bases("FlowType")
      .build();

    def("NumberTypeAnnotation")
      .bases("FlowType")
      .build();

    def("NumberLiteralTypeAnnotation")
      .bases("FlowType")
      .build("value", "raw")
      .field("value", Number)
      .field("raw", String);

    // Babylon 6 differs in AST from Flow
    // same as NumberLiteralTypeAnnotation
    def("NumericLiteralTypeAnnotation")
      .bases("FlowType")
      .build("value", "raw")
      .field("value", Number)
      .field("raw", String);

    def("StringTypeAnnotation")
      .bases("FlowType")
      .build();

    def("StringLiteralTypeAnnotation")
      .bases("FlowType")
      .build("value", "raw")
      .field("value", String)
      .field("raw", String);

    def("BooleanTypeAnnotation")
      .bases("FlowType")
      .build();

    def("BooleanLiteralTypeAnnotation")
      .bases("FlowType")
      .build("value", "raw")
      .field("value", Boolean)
      .field("raw", String);

    def("TypeAnnotation")
      .bases("Node")
      .build("typeAnnotation")
      .field("typeAnnotation", def("FlowType"));

    def("NullableTypeAnnotation")
      .bases("FlowType")
      .build("typeAnnotation")
      .field("typeAnnotation", def("FlowType"));

    def("NullLiteralTypeAnnotation")
      .bases("FlowType")
      .build();

    def("NullTypeAnnotation")
      .bases("FlowType")
      .build();

    def("ThisTypeAnnotation")
      .bases("FlowType")
      .build();

    def("ExistsTypeAnnotation")
      .bases("FlowType")
      .build();

    def("ExistentialTypeParam")
      .bases("FlowType")
      .build();

    def("FunctionTypeAnnotation")
      .bases("FlowType")
      .build("params", "returnType", "rest", "typeParameters")
      .field("params", [def("FunctionTypeParam")])
      .field("returnType", def("FlowType"))
      .field("rest", or(def("FunctionTypeParam"), null))
      .field("typeParameters", or(def("TypeParameterDeclaration"), null));

    def("FunctionTypeParam")
      .bases("Node")
      .build("name", "typeAnnotation", "optional")
      .field("name", def("Identifier"))
      .field("typeAnnotation", def("FlowType"))
      .field("optional", Boolean);

    def("ArrayTypeAnnotation")
      .bases("FlowType")
      .build("elementType")
      .field("elementType", def("FlowType"));

    def("ObjectTypeAnnotation")
      .bases("FlowType")
      .build("properties", "indexers", "callProperties")
      .field("properties", [
        or(def("ObjectTypeProperty"),
           def("ObjectTypeSpreadProperty"))
      ])
      .field("indexers", [def("ObjectTypeIndexer")], defaults.emptyArray)
      .field("callProperties",
             [def("ObjectTypeCallProperty")],
             defaults.emptyArray)
      .field("exact", Boolean, defaults["false"]);

    def("Variance")
      .bases("Node")
      .build("kind")
      .field("kind", or("plus", "minus"));

    var LegacyVariance = or(
      def("Variance"),
      "plus",
      "minus",
      null
    );

    def("ObjectTypeProperty")
      .bases("Node")
      .build("key", "value", "optional")
      .field("key", or(def("Literal"), def("Identifier")))
      .field("value", def("FlowType"))
      .field("optional", Boolean)
      .field("variance", LegacyVariance, defaults["null"]);

    def("ObjectTypeIndexer")
      .bases("Node")
      .build("id", "key", "value")
      .field("id", def("Identifier"))
      .field("key", def("FlowType"))
      .field("value", def("FlowType"))
      .field("variance", LegacyVariance, defaults["null"]);

    def("ObjectTypeCallProperty")
      .bases("Node")
      .build("value")
      .field("value", def("FunctionTypeAnnotation"))
      .field("static", Boolean, defaults["false"]);

    def("QualifiedTypeIdentifier")
      .bases("Node")
      .build("qualification", "id")
      .field("qualification",
             or(def("Identifier"),
                def("QualifiedTypeIdentifier")))
      .field("id", def("Identifier"));

    def("GenericTypeAnnotation")
      .bases("FlowType")
      .build("id", "typeParameters")
      .field("id", or(def("Identifier"), def("QualifiedTypeIdentifier")))
      .field("typeParameters", or(def("TypeParameterInstantiation"), null));

    def("MemberTypeAnnotation")
      .bases("FlowType")
      .build("object", "property")
      .field("object", def("Identifier"))
      .field("property",
             or(def("MemberTypeAnnotation"),
                def("GenericTypeAnnotation")));

    def("UnionTypeAnnotation")
      .bases("FlowType")
      .build("types")
      .field("types", [def("FlowType")]);

    def("IntersectionTypeAnnotation")
      .bases("FlowType")
      .build("types")
      .field("types", [def("FlowType")]);

    def("TypeofTypeAnnotation")
      .bases("FlowType")
      .build("argument")
      .field("argument", def("FlowType"));

    def("ObjectTypeSpreadProperty")
      .bases("Node")
      .build("argument")
      .field("argument", def("FlowType"));

    def("Identifier")
      .field("typeAnnotation", or(def("TypeAnnotation"), null), defaults["null"]);

    def("ObjectPattern")
      .field("typeAnnotation", or(def("TypeAnnotation"), null), defaults["null"]);

    def("TypeParameterDeclaration")
      .bases("Node")
      .build("params")
      .field("params", [def("TypeParameter")]);

    def("TypeParameterInstantiation")
      .bases("Node")
      .build("params")
      .field("params", [def("FlowType")]);

    def("TypeParameter")
      .bases("FlowType")
      .build("name", "variance", "bound")
      .field("name", String)
      .field("variance", LegacyVariance, defaults["null"])
      .field("bound",
             or(def("TypeAnnotation"), null),
             defaults["null"]);

    def("Function")
      .field("returnType",
             or(def("TypeAnnotation"), null),
             defaults["null"])
      .field("typeParameters",
             or(def("TypeParameterDeclaration"), null),
             defaults["null"]);

    def("ClassProperty")
      .build("key", "value", "typeAnnotation", "static")
      .field("value", or(def("Expression"), null))
      .field("typeAnnotation", or(def("TypeAnnotation"), null))
      .field("static", Boolean, defaults["false"])
      .field("variance", LegacyVariance, defaults["null"]);

    ["ClassDeclaration",
     "ClassExpression",
    ].forEach(typeName => {
      def(typeName)
        .field("typeParameters",
               or(def("TypeParameterDeclaration"), null),
               defaults["null"])
        .field("superTypeParameters",
               or([def("GenericTypeAnnotation")], null),
               defaults["null"]);
    });

    def("ClassImplements")
      .bases("Node")
      .build("id")
      .field("id", def("Identifier"))
      .field("superClass", or(def("Expression"), null), defaults["null"])
      .field("typeParameters",
             or(def("TypeParameterInstantiation"), null),
             defaults["null"]);

    ["ClassDeclaration",
     "ClassExpression",
    ].forEach(typeName => {
      def(typeName)
        .field("implements", [def("ClassImplements")], defaults.emptyArray);
    });

    def("InterfaceDeclaration")
      .bases("Declaration")
      .build("id", "body", "extends")
      .field("id", def("Identifier"))
      .field("typeParameters",
             or(def("TypeParameterDeclaration"), null),
             defaults["null"])
      .field("body", def("ObjectTypeAnnotation"))
      .field("extends", [def("InterfaceExtends")]);

    def("DeclareInterface")
      .bases("InterfaceDeclaration")
      .build("id", "body", "extends");

    def("InterfaceExtends")
      .bases("Node")
      .build("id")
      .field("id", def("Identifier"))
      .field("typeParameters", or(def("TypeParameterInstantiation"), null));

    def("TypeAlias")
      .bases("Declaration")
      .build("id", "typeParameters", "right")
      .field("id", def("Identifier"))
      .field("typeParameters", or(def("TypeParameterDeclaration"), null))
      .field("right", def("FlowType"));

    def("OpaqueType")
      .bases("Declaration")
      .build("id", "typeParameters", "impltype", "supertype")
      .field("id", def("Identifier"))
      .field("typeParameters", or(def("TypeParameterDeclaration"), null))
      .field("implType", def("FlowType"))
      .field("superType", def("FlowType"));

    def("DeclareTypeAlias")
      .bases("TypeAlias")
      .build("id", "typeParameters", "right");

    def("DeclareOpaqueType")
      .bases("TypeAlias")
      .build("id", "typeParameters", "supertype");

    def("TypeCastExpression")
      .bases("Expression")
      .build("expression", "typeAnnotation")
      .field("expression", def("Expression"))
      .field("typeAnnotation", def("TypeAnnotation"));

    def("TupleTypeAnnotation")
      .bases("FlowType")
      .build("types")
      .field("types", [def("FlowType")]);

    def("DeclareVariable")
      .bases("Statement")
      .build("id")
      .field("id", def("Identifier"));

    def("DeclareFunction")
      .bases("Statement")
      .build("id")
      .field("id", def("Identifier"));

    def("DeclareClass")
      .bases("InterfaceDeclaration")
      .build("id");

    def("DeclareModule")
      .bases("Statement")
      .build("id", "body")
      .field("id", or(def("Identifier"), def("Literal")))
      .field("body", def("BlockStatement"));

    def("DeclareModuleExports")
      .bases("Statement")
      .build("typeAnnotation")
      .field("typeAnnotation", def("TypeAnnotation"));

    def("DeclareExportDeclaration")
      .bases("Declaration")
      .build("default", "declaration", "specifiers", "source")
      .field("default", Boolean)
      .field("declaration", or(
        def("DeclareVariable"),
        def("DeclareFunction"),
        def("DeclareClass"),
        def("FlowType"), // Implies default.
        null
      ))
      .field("specifiers", [or(
        def("ExportSpecifier"),
        def("ExportBatchSpecifier")
      )], defaults.emptyArray)
      .field("source", or(
        def("Literal"),
        null
      ), defaults["null"]);

    def("DeclareExportAllDeclaration")
      .bases("Declaration")
      .build("source")
      .field("source", or(
        def("Literal"),
        null
      ), defaults["null"]);

    def("FlowPredicate").bases("Flow");

    def("InferredPredicate")
      .bases("FlowPredicate")
      .build();

    def("DeclaredPredicate")
      .bases("FlowPredicate")
      .build("value")
      .field("value", def("Expression"));
  };

  var esprima = function (fork) {
    fork.use(es7);

    var types$$1 = fork.use(types);
    var defaults = fork.use(shared).defaults;
    var def = types$$1.Type.def;
    var or = types$$1.Type.or;

    def("VariableDeclaration")
      .field("declarations", [or(
        def("VariableDeclarator"),
        def("Identifier") // Esprima deviation.
      )]);

    def("Property")
      .field("value", or(
        def("Expression"),
        def("Pattern") // Esprima deviation.
      ));

    def("ArrayPattern")
      .field("elements", [or(
        def("Pattern"),
        def("SpreadElement"),
        null
      )]);

    def("ObjectPattern")
      .field("properties", [or(
        def("Property"),
        def("PropertyPattern"),
        def("SpreadPropertyPattern"),
        def("SpreadProperty") // Used by Esprima.
      )]);

    // Like ModuleSpecifier, except type:"ExportSpecifier" and buildable.
    // export {<id [as name]>} [from ...];
    def("ExportSpecifier")
      .bases("ModuleSpecifier")
      .build("id", "name");

    // export <*> from ...;
    def("ExportBatchSpecifier")
      .bases("Specifier")
      .build();

    def("ExportDeclaration")
      .bases("Declaration")
      .build("default", "declaration", "specifiers", "source")
      .field("default", Boolean)
      .field("declaration", or(
        def("Declaration"),
        def("Expression"), // Implies default.
        null
      ))
      .field("specifiers", [or(
        def("ExportSpecifier"),
        def("ExportBatchSpecifier")
      )], defaults.emptyArray)
      .field("source", or(
        def("Literal"),
        null
      ), defaults["null"]);

    def("Block")
      .bases("Comment")
      .build("value", /*optional:*/ "leading", "trailing");

    def("Line")
      .bases("Comment")
      .build("value", /*optional:*/ "leading", "trailing");
  };

  var babelCore = function (fork) {
    fork.use(es7);

    var types$$1 = fork.use(types);
    var defaults = fork.use(shared).defaults;
    var def = types$$1.Type.def;
    var or = types$$1.Type.or;

    def("Noop")
      .bases("Statement")
      .build();

    def("DoExpression")
      .bases("Expression")
      .build("body")
      .field("body", [def("Statement")]);

    def("Super")
      .bases("Expression")
      .build();

    def("BindExpression")
      .bases("Expression")
      .build("object", "callee")
      .field("object", or(def("Expression"), null))
      .field("callee", def("Expression"));

    def("Decorator")
      .bases("Node")
      .build("expression")
      .field("expression", def("Expression"));

    def("Property")
      .field("decorators",
             or([def("Decorator")], null),
             defaults["null"]);

    def("MethodDefinition")
      .field("decorators",
             or([def("Decorator")], null),
             defaults["null"]);

    def("MetaProperty")
      .bases("Expression")
      .build("meta", "property")
      .field("meta", def("Identifier"))
      .field("property", def("Identifier"));

    def("ParenthesizedExpression")
      .bases("Expression")
      .build("expression")
      .field("expression", def("Expression"));

    def("ImportSpecifier")
      .bases("ModuleSpecifier")
      .build("imported", "local")
      .field("imported", def("Identifier"));

    def("ImportDefaultSpecifier")
      .bases("ModuleSpecifier")
      .build("local");

    def("ImportNamespaceSpecifier")
      .bases("ModuleSpecifier")
      .build("local");

    def("ExportDefaultDeclaration")
      .bases("Declaration")
      .build("declaration")
      .field("declaration", or(def("Declaration"), def("Expression")));

    def("ExportNamedDeclaration")
      .bases("Declaration")
      .build("declaration", "specifiers", "source")
      .field("declaration", or(def("Declaration"), null))
      .field("specifiers", [def("ExportSpecifier")], defaults.emptyArray)
      .field("source", or(def("Literal"), null), defaults["null"]);

    def("ExportSpecifier")
      .bases("ModuleSpecifier")
      .build("local", "exported")
      .field("exported", def("Identifier"));

    def("ExportNamespaceSpecifier")
      .bases("Specifier")
      .build("exported")
      .field("exported", def("Identifier"));

    def("ExportDefaultSpecifier")
      .bases("Specifier")
      .build("exported")
      .field("exported", def("Identifier"));

    def("ExportAllDeclaration")
      .bases("Declaration")
      .build("exported", "source")
      .field("exported", or(def("Identifier"), null))
      .field("source", def("Literal"));

    def("CommentBlock")
      .bases("Comment")
      .build("value", /*optional:*/ "leading", "trailing");

    def("CommentLine")
      .bases("Comment")
      .build("value", /*optional:*/ "leading", "trailing");

    def("Directive")
      .bases("Node")
      .build("value")
      .field("value", def("DirectiveLiteral"));

    def("DirectiveLiteral")
      .bases("Node", "Expression")
      .build("value")
      .field("value", String, defaults["use strict"]);

    def("BlockStatement")
      .bases("Statement")
      .build("body")
      .field("body", [def("Statement")])
      .field("directives", [def("Directive")], defaults.emptyArray);

    def("Program")
      .bases("Node")
      .build("body")
      .field("body", [def("Statement")])
      .field("directives", [def("Directive")], defaults.emptyArray);

    // Split Literal
    def("StringLiteral")
      .bases("Literal")
      .build("value")
      .field("value", String);

    def("NumericLiteral")
      .bases("Literal")
      .build("value")
      .field("value", Number)
      .field("raw", or(String, null), defaults["null"])
      .field("extra", {
        rawValue: Number,
        raw: String
      }, function getDefault() {
        return {
          rawValue: this.value,
          raw: this.value + ""
        }
      });

    def("BigIntLiteral")
      .bases("Literal")
      .build("value")
      // Only String really seems appropriate here, since BigInt values
      // often exceed the limits of JS numbers.
      .field("value", or(String, Number))
      .field("extra", {
        rawValue: String,
        raw: String
      }, function getDefault() {
        return {
          rawValue: String(this.value),
          raw: this.value + "n"
        };
      });

    def("NullLiteral")
      .bases("Literal")
      .build()
      .field("value", null, defaults["null"]);

    def("BooleanLiteral")
      .bases("Literal")
      .build("value")
      .field("value", Boolean);

    def("RegExpLiteral")
      .bases("Literal")
      .build("pattern", "flags")
      .field("pattern", String)
      .field("flags", String)
      .field("value", RegExp, function () {
        return new RegExp(this.pattern, this.flags);
      });

    var ObjectExpressionProperty = or(
      def("Property"),
      def("ObjectMethod"),
      def("ObjectProperty"),
      def("SpreadProperty"),
      def("SpreadElement")
    );

    // Split Property -> ObjectProperty and ObjectMethod
    def("ObjectExpression")
      .bases("Expression")
      .build("properties")
      .field("properties", [ObjectExpressionProperty]);

    // ObjectMethod hoist .value properties to own properties
    def("ObjectMethod")
      .bases("Node", "Function")
      .build("kind", "key", "params", "body", "computed")
      .field("kind", or("method", "get", "set"))
      .field("key", or(def("Literal"), def("Identifier"), def("Expression")))
      .field("params", [def("Pattern")])
      .field("body", def("BlockStatement"))
      .field("computed", Boolean, defaults["false"])
      .field("generator", Boolean, defaults["false"])
      .field("async", Boolean, defaults["false"])
      .field("accessibility", // TypeScript
             or(def("Literal"), null),
             defaults["null"])
      .field("decorators",
             or([def("Decorator")], null),
             defaults["null"]);

    def("ObjectProperty")
      .bases("Node")
      .build("key", "value")
      .field("key", or(def("Literal"), def("Identifier"), def("Expression")))
      .field("value", or(def("Expression"), def("Pattern")))
      .field("accessibility", // TypeScript
             or(def("Literal"), null),
             defaults["null"])
      .field("computed", Boolean, defaults["false"]);

    var ClassBodyElement = or(
      def("MethodDefinition"),
      def("VariableDeclarator"),
      def("ClassPropertyDefinition"),
      def("ClassProperty"),
      def("ClassMethod")
    );

    // MethodDefinition -> ClassMethod
    def("ClassBody")
      .bases("Declaration")
      .build("body")
      .field("body", [ClassBodyElement]);

    def("ClassMethod")
      .bases("Declaration", "Function")
      .build("kind", "key", "params", "body", "computed", "static")
      .field("kind", or("get", "set", "method", "constructor"))
      .field("key", or(def("Literal"), def("Identifier"), def("Expression")))
      .field("params", [def("Pattern")])
      .field("body", def("BlockStatement"))
      .field("computed", Boolean, defaults["false"])
      .field("static", Boolean, defaults["false"])
      .field("generator", Boolean, defaults["false"])
      .field("async", Boolean, defaults["false"])
      .field("decorators",
             or([def("Decorator")], null),
             defaults["null"]);

    var ObjectPatternProperty = or(
      def("Property"),
      def("PropertyPattern"),
      def("SpreadPropertyPattern"),
      def("SpreadProperty"), // Used by Esprima
      def("ObjectProperty"), // Babel 6
      def("RestProperty") // Babel 6
    );

    // Split into RestProperty and SpreadProperty
    def("ObjectPattern")
      .bases("Pattern")
      .build("properties")
      .field("properties", [ObjectPatternProperty])
      .field("decorators",
             or([def("Decorator")], null),
             defaults["null"]);

    def("SpreadProperty")
      .bases("Node")
      .build("argument")
      .field("argument", def("Expression"));

    def("RestProperty")
      .bases("Node")
      .build("argument")
      .field("argument", def("Expression"));

    def("ForAwaitStatement")
      .bases("Statement")
      .build("left", "right", "body")
      .field("left", or(
        def("VariableDeclaration"),
        def("Expression")))
      .field("right", def("Expression"))
      .field("body", def("Statement"));

    // The callee node of a dynamic import(...) expression.
    def("Import")
      .bases("Expression")
      .build();
  };

  var babel = function (fork) {
    fork.use(babelCore);
    fork.use(flow);
  };

  var typescript = function (fork) {
    // Since TypeScript is parsed by Babylon, include the core Babylon types
    // but omit the Flow-related types.
    fork.use(babelCore);

    var types$$1 = fork.use(types);
    var n = types$$1.namedTypes;
    var def = types$$1.Type.def;
    var or = types$$1.Type.or;
    var defaults = fork.use(shared).defaults;
    var StringLiteral = new types$$1.Type(function (value, deep) {
      if (n.StringLiteral &&
          n.StringLiteral.check(value, deep)) {
        return true
      }
      if (n.Literal &&
          n.Literal.check(value, deep) &&
          typeof value.value === "string") {
        return true;
      }
      return false;
    }, "StringLiteral");

    def("TSType")
      .bases("Node");

    var IdOrQualifiedName = or(
      def("Identifier"),
      def("TSQualifiedName")
    );

    def("TSTypeReference")
      .bases("TSType")
      .build("typeName", "typeParameters")
      .field("typeName", IdOrQualifiedName)
      .field("typeParameters",
             or(def("TSTypeParameterInstantiation"), null),
             defaults["null"]);

    // An abstract (non-buildable) base type that provide a commonly-needed
    // optional .typeParameters field.
    def("TSHasOptionalTypeParameters")
      .field("typeParameters",
             or(def("TSTypeParameterDeclaration"), null),
             defaults["null"]);

    // An abstract (non-buildable) base type that provide a commonly-needed
    // optional .typeAnnotation field.
    def("TSHasOptionalTypeAnnotation")
      .field("typeAnnotation",
             or(def("TSTypeAnnotation"), null),
             defaults["null"]);

    def("TSQualifiedName")
      .bases("Node")
      .build("left", "right")
      .field("left", IdOrQualifiedName)
      .field("right", IdOrQualifiedName);

    def("TSAsExpression")
      .bases("Expression")
      .build("expression")
      .field("expression", def("Expression"))
      .field("typeAnnotation", def("TSType"))
      .field("extra",
             or({ parenthesized: Boolean }, null),
             defaults["null"]);

    def("TSNonNullExpression")
      .bases("Expression")
      .build("expression")
      .field("expression", def("Expression"));

    [ // Define all the simple keyword types.
      "TSAnyKeyword",
      "TSBooleanKeyword",
      "TSNeverKeyword",
      "TSNullKeyword",
      "TSNumberKeyword",
      "TSObjectKeyword",
      "TSStringKeyword",
      "TSSymbolKeyword",
      "TSUndefinedKeyword",
      "TSUnknownKeyword",
      "TSVoidKeyword",
      "TSThisType",
    ].forEach(keywordType => {
      def(keywordType)
        .bases("TSType")
        .build();
    });

    def("TSArrayType")
      .bases("TSType")
      .build("elementType")
      .field("elementType", def("TSType"));

    def("TSLiteralType")
      .bases("TSType")
      .build("literal")
      .field("literal",
             or(def("NumericLiteral"),
                def("StringLiteral"),
                def("BooleanLiteral")));

    ["TSUnionType",
     "TSIntersectionType",
    ].forEach(typeName => {
      def(typeName)
        .bases("TSType")
        .build("types")
        .field("types", [def("TSType")]);
    });

    def("TSConditionalType")
      .bases("TSType")
      .build("checkType", "extendsType", "trueType", "falseType")
      .field("checkType", def("TSType"))
      .field("extendsType", def("TSType"))
      .field("trueType", def("TSType"))
      .field("falseType", def("TSType"));

    def("TSInferType")
      .bases("TSType")
      .build("typeParameter")
      .field("typeParameter", def("TSTypeParameter"));

    def("TSParenthesizedType")
      .bases("TSType")
      .build("typeAnnotation")
      .field("typeAnnotation", def("TSType"));

    var ParametersType = [or(
      def("Identifier"),
      def("RestElement")
    )];

    ["TSFunctionType",
     "TSConstructorType",
    ].forEach(typeName => {
      def(typeName)
        .bases("TSType",
               "TSHasOptionalTypeParameters",
               "TSHasOptionalTypeAnnotation")
        .build("parameters")
        .field("parameters", ParametersType);
    });

    def("TSDeclareFunction")
      .bases("Declaration", "TSHasOptionalTypeParameters")
      .build("id", "params", "returnType")
      .field("declare", Boolean, defaults["false"])
      .field("async", Boolean, defaults["false"])
      .field("generator", Boolean, defaults["false"])
      .field("id", or(def("Identifier"), null), defaults["null"])
      .field("params", [def("Pattern")])
      // tSFunctionTypeAnnotationCommon
      .field("returnType",
             or(def("TSTypeAnnotation"),
                def("Noop"), // Still used?
                null),
             defaults["null"]);

    def("TSDeclareMethod")
      .bases("Declaration", "TSHasOptionalTypeParameters")
      .build("key", "params", "returnType")
      .field("async", Boolean, defaults["false"])
      .field("generator", Boolean, defaults["false"])
      .field("params", [def("Pattern")])
      // classMethodOrPropertyCommon
      .field("abstract", Boolean, defaults["false"])
      .field("accessibility",
             or("public", "private", "protected", void 0),
             defaults["undefined"])
      .field("static", Boolean, defaults["false"])
      .field("computed", Boolean, defaults["false"])
      .field("optional", Boolean, defaults["false"])
      .field("key", or(
        def("Identifier"),
        def("StringLiteral"),
        def("NumericLiteral"),
        // Only allowed if .computed is true.
        def("Expression")
      ))
      // classMethodOrDeclareMethodCommon
      .field("kind",
             or("get", "set", "method", "constructor"),
             function getDefault() { return "method"; })
      .field("access", // Not "accessibility"?
             or("public", "private", "protected", void 0),
             defaults["undefined"])
      .field("decorators",
             or([def("Decorator")], null),
             defaults["null"])
      // tSFunctionTypeAnnotationCommon
      .field("returnType",
             or(def("TSTypeAnnotation"),
                def("Noop"), // Still used?
                null),
             defaults["null"]);

    def("TSMappedType")
      .bases("TSType")
      .build("typeParameter", "typeAnnotation")
      .field("readonly", Boolean, defaults["false"])
      .field("typeParameter", def("TSTypeParameter"))
      .field("optional", Boolean, defaults["false"])
      .field("typeAnnotation",
             or(def("TSType"), null),
             defaults["null"]);

    def("TSTupleType")
      .bases("TSType")
      .build("elementTypes")
      .field("elementTypes", [def("TSType")]);

    def("TSRestType")
      .bases("TSType")
      .build("typeAnnotation")
      .field("typeAnnotation", def("TSType"));

    def("TSOptionalType")
      .bases("TSType")
      .build("typeAnnotation")
      .field("typeAnnotation", def("TSType"));

    def("TSIndexedAccessType")
      .bases("TSType")
      .build("objectType", "indexType")
      .field("objectType", def("TSType"))
      .field("indexType", def("TSType"));

    def("TSTypeOperator")
      .bases("TSType")
      .build("operator")
      .field("operator", String)
      .field("typeAnnotation", def("TSType"));

    def("TSTypeAnnotation")
      .bases("Node")
      .build("typeAnnotation")
      .field("typeAnnotation",
             or(def("TSType"),
                def("TSTypeAnnotation")));

    def("TSIndexSignature")
      .bases("Declaration", "TSHasOptionalTypeAnnotation")
      .build("parameters")
      .field("parameters", [def("Identifier")]) // Length === 1
      .field("readonly", Boolean, defaults["false"]);

    def("TSPropertySignature")
      .bases("Declaration", "TSHasOptionalTypeAnnotation")
      .build("key")
      .field("key", def("Expression"))
      .field("computed", Boolean, defaults["false"])
      .field("readonly", Boolean, defaults["false"])
      .field("optional", Boolean, defaults["false"])
      .field("initializer",
             or(def("Expression"), null),
             defaults["null"]);

    def("TSMethodSignature")
      .bases("Declaration",
             "TSHasOptionalTypeParameters",
             "TSHasOptionalTypeAnnotation")
      .build("key")
      .field("key", def("Expression"))
      .field("computed", Boolean, defaults["false"])
      .field("optional", Boolean, defaults["false"])
      .field("parameters", ParametersType);

    def("TSTypePredicate")
      .bases("TSTypeAnnotation")
      .build("parameterName", "typeAnnotation")
      .field("parameterName",
             or(def("Identifier"),
                def("TSThisType")))
      .field("typeAnnotation", def("TSTypeAnnotation"));

    ["TSCallSignatureDeclaration",
     "TSConstructSignatureDeclaration",
    ].forEach(typeName => {
      def(typeName)
        .bases("Declaration",
               "TSHasOptionalTypeParameters",
               "TSHasOptionalTypeAnnotation")
        .build("parameters")
        .field("parameters", ParametersType);
    });

    def("TSEnumMember")
      .bases("Node")
      .build("id", "initializer")
      .field("id", or(def("Identifier"), StringLiteral))
      .field("initializer",
             or(def("Expression"), null),
             defaults["null"]);

    def("TSTypeQuery")
      .bases("TSType")
      .build("exprName")
      .field("exprName", def("Identifier"));

    // Inferred from Babylon's tsParseTypeMember method.
    var TSTypeMember = or(
      def("TSCallSignatureDeclaration"),
      def("TSConstructSignatureDeclaration"),
      def("TSIndexSignature"),
      def("TSMethodSignature"),
      def("TSPropertySignature")
    );

    def("TSTypeLiteral")
      .bases("TSType")
      .build("members")
      .field("members", [TSTypeMember]);

    def("TSTypeParameter")
      .bases("Identifier")
      .field("name", String)
      .field("constraint", or(def("TSType"), null), defaults["null"])
      .field("default", or(def("TSType"), null), defaults["null"]);

    def("TSTypeAssertion")
      .bases("Expression")
      .build("typeAnnotation", "expression")
      .field("typeAnnotation", def("TSType"))
      .field("expression", def("Expression"))
      .field("extra",
             or({ parenthesized: Boolean }, null),
             defaults["null"]);

    def("TSTypeParameterDeclaration")
      .bases("Declaration")
      .build("params")
      .field("params", [def("TSTypeParameter")]);

    def("TSTypeParameterInstantiation")
      .bases("Node")
      .build("params")
      .field("params", [def("TSType")]);

    def("TSEnumDeclaration")
      .bases("Declaration")
      .build("id", "members")
      .field("id", def("Identifier"))
      .field("const", Boolean, defaults["false"])
      .field("declare", Boolean, defaults["false"])
      .field("members", [def("TSEnumMember")])
      .field("initializer",
             or(def("Expression"), null),
             defaults["null"]);

    def("TSTypeAliasDeclaration")
      .bases("Declaration", "TSHasOptionalTypeParameters")
      .build("id")
      .field("id", def("Identifier"))
      .field("declare", Boolean, defaults["false"])
      .field("typeAnnotation", def("TSType"));

    def("TSModuleBlock")
      .bases("Node")
      .build("body")
      .field("body", [def("Statement")]);

    def("TSModuleDeclaration")
      .bases("Declaration")
      .build("id", "body")
      .field("id", or(StringLiteral, IdOrQualifiedName))
      .field("declare", Boolean, defaults["false"])
      .field("global", Boolean, defaults["false"])
      .field("body",
             or(def("TSModuleBlock"),
                def("TSModuleDeclaration"),
                null),
             defaults["null"]);

    def("TSImportEqualsDeclaration")
      .bases("Declaration")
      .build("id", "moduleReference")
      .field("id", def("Identifier"))
      .field("isExport", Boolean, defaults["false"])
      .field("moduleReference",
             or(IdOrQualifiedName,
                def("TSExternalModuleReference")));

    def("TSExternalModuleReference")
      .bases("Declaration")
      .build("expression")
      .field("expression", StringLiteral);

    def("TSExportAssignment")
      .bases("Statement")
      .build("expression")
      .field("expression", def("Expression"));

    def("TSNamespaceExportDeclaration")
      .bases("Declaration")
      .build("id")
      .field("id", def("Identifier"));

    def("TSInterfaceBody")
      .bases("Node")
      .build("body")
      .field("body", [TSTypeMember]);

    def("TSExpressionWithTypeArguments")
      .bases("TSType")
      .build("expression", "typeParameters")
      .field("expression", IdOrQualifiedName)
      .field("typeParameters",
             or(def("TSTypeParameterInstantiation"), null),
             defaults["null"]);

    def("TSInterfaceDeclaration")
      .bases("Declaration", "TSHasOptionalTypeParameters")
      .build("id", "body")
      .field("id", IdOrQualifiedName)
      .field("declare", Boolean, defaults["false"])
      .field("extends",
             or([def("TSExpressionWithTypeArguments")], null),
             defaults["null"])
      .field("body", def("TSInterfaceBody"));

    ["ClassDeclaration",
     "ClassExpression",
    ].forEach(typeName => {
      def(typeName)
        .field("implements",
               [def("TSExpressionWithTypeArguments")],
               defaults.emptyArray);
    });

    def("TSParameterProperty")
      .bases("Pattern")
      .build("parameter")
      .field("accessibility",
             or("public", "private", "protected", void 0),
             defaults["undefined"])
      .field("readonly", Boolean, defaults["false"])
      .field("parameter", or(def("Identifier"),
                             def("AssignmentPattern")));

    // Defined already in es6 and babel-core.
    def("ClassBody")
      .field("body", [or(
        def("MethodDefinition"),
        def("VariableDeclarator"),
        def("ClassPropertyDefinition"),
        def("ClassProperty"),
        def("ClassMethod"),
        // Just need to add these types:
        def("TSDeclareMethod"),
        TSTypeMember
      )]);
  };

  var esProposals = function (fork) {
    fork.use(core);

    var types$$1 = fork.use(types);
    var Type = types$$1.Type;
    var def = types$$1.Type.def;
    var or = Type.or;

    var shared$$1 = fork.use(shared);
    var defaults = shared$$1.defaults;


    // https://github.com/tc39/proposal-optional-chaining
    // `a?.b` as per https://github.com/estree/estree/issues/146
    def("OptionalMemberExpression")
      .bases("MemberExpression")
      .build("object", "property", "computed", "optional")
      .field("optional", Boolean, defaults["true"]);

    // a?.b()
    def("OptionalCallExpression")
      .bases("CallExpression")
      .build("callee", "arguments", "optional")
      .field("optional", Boolean, defaults["true"]);


    // https://github.com/tc39/proposal-nullish-coalescing
    // `a ?? b` as per https://github.com/babel/babylon/pull/761/files
    var LogicalOperator = or("||", "&&", "??");

    def("LogicalExpression")
      .field("operator", LogicalOperator);
  };

  var main = fork([
    // This core module of AST types captures ES5 as it is parsed today by
    // git://github.com/ariya/esprima.git#master.
    core,

    // Feel free to add to or remove from this list of extension modules to
    // configure the precise type hierarchy that you need.
    es6,
    es7,
    mozilla,
    e4x,
    jsx,
    flow,
    esprima,
    babel,
    typescript,
    esProposals
  ]);

  // This module was originally created so that Recast could add its own
  // custom types to the AST type system (in particular, the File type), but
  // those types are now incorporated into ast-types, so this module doesn't
  // have much to do anymore. Still, it might prove useful in the future.
  var types$1 = main;

  /* -*- Mode: js; js-indent-level: 2; -*- */
  /*
   * Copyright 2011 Mozilla Foundation and contributors
   * Licensed under the New BSD license. See LICENSE or:
   * http://opensource.org/licenses/BSD-3-Clause
   */

  var intToCharMap = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/'.split('');

  /**
   * Encode an integer in the range of 0 to 63 to a single base 64 digit.
   */
  var encode = function (number) {
    if (0 <= number && number < intToCharMap.length) {
      return intToCharMap[number];
    }
    throw new TypeError("Must be between 0 and 63: " + number);
  };

  /**
   * Decode a single base 64 character code digit to an integer. Returns -1 on
   * failure.
   */
  var decode = function (charCode) {
    var bigA = 65;     // 'A'
    var bigZ = 90;     // 'Z'

    var littleA = 97;  // 'a'
    var littleZ = 122; // 'z'

    var zero = 48;     // '0'
    var nine = 57;     // '9'

    var plus = 43;     // '+'
    var slash = 47;    // '/'

    var littleOffset = 26;
    var numberOffset = 52;

    // 0 - 25: ABCDEFGHIJKLMNOPQRSTUVWXYZ
    if (bigA <= charCode && charCode <= bigZ) {
      return (charCode - bigA);
    }

    // 26 - 51: abcdefghijklmnopqrstuvwxyz
    if (littleA <= charCode && charCode <= littleZ) {
      return (charCode - littleA + littleOffset);
    }

    // 52 - 61: 0123456789
    if (zero <= charCode && charCode <= nine) {
      return (charCode - zero + numberOffset);
    }

    // 62: +
    if (charCode == plus) {
      return 62;
    }

    // 63: /
    if (charCode == slash) {
      return 63;
    }

    // Invalid base64 digit.
    return -1;
  };

  var base64 = {
  	encode: encode,
  	decode: decode
  };

  /* -*- Mode: js; js-indent-level: 2; -*- */
  /*
   * Copyright 2011 Mozilla Foundation and contributors
   * Licensed under the New BSD license. See LICENSE or:
   * http://opensource.org/licenses/BSD-3-Clause
   *
   * Based on the Base 64 VLQ implementation in Closure Compiler:
   * https://code.google.com/p/closure-compiler/source/browse/trunk/src/com/google/debugging/sourcemap/Base64VLQ.java
   *
   * Copyright 2011 The Closure Compiler Authors. All rights reserved.
   * Redistribution and use in source and binary forms, with or without
   * modification, are permitted provided that the following conditions are
   * met:
   *
   *  * Redistributions of source code must retain the above copyright
   *    notice, this list of conditions and the following disclaimer.
   *  * Redistributions in binary form must reproduce the above
   *    copyright notice, this list of conditions and the following
   *    disclaimer in the documentation and/or other materials provided
   *    with the distribution.
   *  * Neither the name of Google Inc. nor the names of its
   *    contributors may be used to endorse or promote products derived
   *    from this software without specific prior written permission.
   *
   * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
   */



  // A single base 64 digit can contain 6 bits of data. For the base 64 variable
  // length quantities we use in the source map spec, the first bit is the sign,
  // the next four bits are the actual value, and the 6th bit is the
  // continuation bit. The continuation bit tells us whether there are more
  // digits in this value following this digit.
  //
  //   Continuation
  //   |    Sign
  //   |    |
  //   V    V
  //   101011

  var VLQ_BASE_SHIFT = 5;

  // binary: 100000
  var VLQ_BASE = 1 << VLQ_BASE_SHIFT;

  // binary: 011111
  var VLQ_BASE_MASK = VLQ_BASE - 1;

  // binary: 100000
  var VLQ_CONTINUATION_BIT = VLQ_BASE;

  /**
   * Converts from a two-complement value to a value where the sign bit is
   * placed in the least significant bit.  For example, as decimals:
   *   1 becomes 2 (10 binary), -1 becomes 3 (11 binary)
   *   2 becomes 4 (100 binary), -2 becomes 5 (101 binary)
   */
  function toVLQSigned(aValue) {
    return aValue < 0
      ? ((-aValue) << 1) + 1
      : (aValue << 1) + 0;
  }

  /**
   * Converts to a two-complement value from a value where the sign bit is
   * placed in the least significant bit.  For example, as decimals:
   *   2 (10 binary) becomes 1, 3 (11 binary) becomes -1
   *   4 (100 binary) becomes 2, 5 (101 binary) becomes -2
   */
  function fromVLQSigned(aValue) {
    var isNegative = (aValue & 1) === 1;
    var shifted = aValue >> 1;
    return isNegative
      ? -shifted
      : shifted;
  }

  /**
   * Returns the base 64 VLQ encoded value.
   */
  var encode$1 = function base64VLQ_encode(aValue) {
    var encoded = "";
    var digit;

    var vlq = toVLQSigned(aValue);

    do {
      digit = vlq & VLQ_BASE_MASK;
      vlq >>>= VLQ_BASE_SHIFT;
      if (vlq > 0) {
        // There are still more digits in this value, so we must make sure the
        // continuation bit is marked.
        digit |= VLQ_CONTINUATION_BIT;
      }
      encoded += base64.encode(digit);
    } while (vlq > 0);

    return encoded;
  };

  /**
   * Decodes the next base 64 VLQ value from the given string and returns the
   * value and the rest of the string via the out parameter.
   */
  var decode$1 = function base64VLQ_decode(aStr, aIndex, aOutParam) {
    var strLen = aStr.length;
    var result = 0;
    var shift = 0;
    var continuation, digit;

    do {
      if (aIndex >= strLen) {
        throw new Error("Expected more digits in base 64 VLQ value.");
      }

      digit = base64.decode(aStr.charCodeAt(aIndex++));
      if (digit === -1) {
        throw new Error("Invalid base64 digit: " + aStr.charAt(aIndex - 1));
      }

      continuation = !!(digit & VLQ_CONTINUATION_BIT);
      digit &= VLQ_BASE_MASK;
      result = result + (digit << shift);
      shift += VLQ_BASE_SHIFT;
    } while (continuation);

    aOutParam.value = fromVLQSigned(result);
    aOutParam.rest = aIndex;
  };

  var base64Vlq = {
  	encode: encode$1,
  	decode: decode$1
  };

  var util = createCommonjsModule(function (module, exports) {
  /* -*- Mode: js; js-indent-level: 2; -*- */
  /*
   * Copyright 2011 Mozilla Foundation and contributors
   * Licensed under the New BSD license. See LICENSE or:
   * http://opensource.org/licenses/BSD-3-Clause
   */

  /**
   * This is a helper function for getting values from parameter/options
   * objects.
   *
   * @param args The object we are extracting values from
   * @param name The name of the property we are getting.
   * @param defaultValue An optional value to return if the property is missing
   * from the object. If this is not specified and the property is missing, an
   * error will be thrown.
   */
  function getArg(aArgs, aName, aDefaultValue) {
    if (aName in aArgs) {
      return aArgs[aName];
    } else if (arguments.length === 3) {
      return aDefaultValue;
    } else {
      throw new Error('"' + aName + '" is a required argument.');
    }
  }
  exports.getArg = getArg;

  var urlRegexp = /^(?:([\w+\-.]+):)?\/\/(?:(\w+:\w+)@)?([\w.-]*)(?::(\d+))?(.*)$/;
  var dataUrlRegexp = /^data:.+\,.+$/;

  function urlParse(aUrl) {
    var match = aUrl.match(urlRegexp);
    if (!match) {
      return null;
    }
    return {
      scheme: match[1],
      auth: match[2],
      host: match[3],
      port: match[4],
      path: match[5]
    };
  }
  exports.urlParse = urlParse;

  function urlGenerate(aParsedUrl) {
    var url = '';
    if (aParsedUrl.scheme) {
      url += aParsedUrl.scheme + ':';
    }
    url += '//';
    if (aParsedUrl.auth) {
      url += aParsedUrl.auth + '@';
    }
    if (aParsedUrl.host) {
      url += aParsedUrl.host;
    }
    if (aParsedUrl.port) {
      url += ":" + aParsedUrl.port;
    }
    if (aParsedUrl.path) {
      url += aParsedUrl.path;
    }
    return url;
  }
  exports.urlGenerate = urlGenerate;

  /**
   * Normalizes a path, or the path portion of a URL:
   *
   * - Replaces consecutive slashes with one slash.
   * - Removes unnecessary '.' parts.
   * - Removes unnecessary '<dir>/..' parts.
   *
   * Based on code in the Node.js 'path' core module.
   *
   * @param aPath The path or url to normalize.
   */
  function normalize(aPath) {
    var path = aPath;
    var url = urlParse(aPath);
    if (url) {
      if (!url.path) {
        return aPath;
      }
      path = url.path;
    }
    var isAbsolute = exports.isAbsolute(path);

    var parts = path.split(/\/+/);
    for (var part, up = 0, i = parts.length - 1; i >= 0; i--) {
      part = parts[i];
      if (part === '.') {
        parts.splice(i, 1);
      } else if (part === '..') {
        up++;
      } else if (up > 0) {
        if (part === '') {
          // The first part is blank if the path is absolute. Trying to go
          // above the root is a no-op. Therefore we can remove all '..' parts
          // directly after the root.
          parts.splice(i + 1, up);
          up = 0;
        } else {
          parts.splice(i, 2);
          up--;
        }
      }
    }
    path = parts.join('/');

    if (path === '') {
      path = isAbsolute ? '/' : '.';
    }

    if (url) {
      url.path = path;
      return urlGenerate(url);
    }
    return path;
  }
  exports.normalize = normalize;

  /**
   * Joins two paths/URLs.
   *
   * @param aRoot The root path or URL.
   * @param aPath The path or URL to be joined with the root.
   *
   * - If aPath is a URL or a data URI, aPath is returned, unless aPath is a
   *   scheme-relative URL: Then the scheme of aRoot, if any, is prepended
   *   first.
   * - Otherwise aPath is a path. If aRoot is a URL, then its path portion
   *   is updated with the result and aRoot is returned. Otherwise the result
   *   is returned.
   *   - If aPath is absolute, the result is aPath.
   *   - Otherwise the two paths are joined with a slash.
   * - Joining for example 'http://' and 'www.example.com' is also supported.
   */
  function join(aRoot, aPath) {
    if (aRoot === "") {
      aRoot = ".";
    }
    if (aPath === "") {
      aPath = ".";
    }
    var aPathUrl = urlParse(aPath);
    var aRootUrl = urlParse(aRoot);
    if (aRootUrl) {
      aRoot = aRootUrl.path || '/';
    }

    // `join(foo, '//www.example.org')`
    if (aPathUrl && !aPathUrl.scheme) {
      if (aRootUrl) {
        aPathUrl.scheme = aRootUrl.scheme;
      }
      return urlGenerate(aPathUrl);
    }

    if (aPathUrl || aPath.match(dataUrlRegexp)) {
      return aPath;
    }

    // `join('http://', 'www.example.com')`
    if (aRootUrl && !aRootUrl.host && !aRootUrl.path) {
      aRootUrl.host = aPath;
      return urlGenerate(aRootUrl);
    }

    var joined = aPath.charAt(0) === '/'
      ? aPath
      : normalize(aRoot.replace(/\/+$/, '') + '/' + aPath);

    if (aRootUrl) {
      aRootUrl.path = joined;
      return urlGenerate(aRootUrl);
    }
    return joined;
  }
  exports.join = join;

  exports.isAbsolute = function (aPath) {
    return aPath.charAt(0) === '/' || urlRegexp.test(aPath);
  };

  /**
   * Make a path relative to a URL or another path.
   *
   * @param aRoot The root path or URL.
   * @param aPath The path or URL to be made relative to aRoot.
   */
  function relative(aRoot, aPath) {
    if (aRoot === "") {
      aRoot = ".";
    }

    aRoot = aRoot.replace(/\/$/, '');

    // It is possible for the path to be above the root. In this case, simply
    // checking whether the root is a prefix of the path won't work. Instead, we
    // need to remove components from the root one by one, until either we find
    // a prefix that fits, or we run out of components to remove.
    var level = 0;
    while (aPath.indexOf(aRoot + '/') !== 0) {
      var index = aRoot.lastIndexOf("/");
      if (index < 0) {
        return aPath;
      }

      // If the only part of the root that is left is the scheme (i.e. http://,
      // file:///, etc.), one or more slashes (/), or simply nothing at all, we
      // have exhausted all components, so the path is not relative to the root.
      aRoot = aRoot.slice(0, index);
      if (aRoot.match(/^([^\/]+:\/)?\/*$/)) {
        return aPath;
      }

      ++level;
    }

    // Make sure we add a "../" for each component we removed from the root.
    return Array(level + 1).join("../") + aPath.substr(aRoot.length + 1);
  }
  exports.relative = relative;

  var supportsNullProto = (function () {
    var obj = Object.create(null);
    return !('__proto__' in obj);
  }());

  function identity (s) {
    return s;
  }

  /**
   * Because behavior goes wacky when you set `__proto__` on objects, we
   * have to prefix all the strings in our set with an arbitrary character.
   *
   * See https://github.com/mozilla/source-map/pull/31 and
   * https://github.com/mozilla/source-map/issues/30
   *
   * @param String aStr
   */
  function toSetString(aStr) {
    if (isProtoString(aStr)) {
      return '$' + aStr;
    }

    return aStr;
  }
  exports.toSetString = supportsNullProto ? identity : toSetString;

  function fromSetString(aStr) {
    if (isProtoString(aStr)) {
      return aStr.slice(1);
    }

    return aStr;
  }
  exports.fromSetString = supportsNullProto ? identity : fromSetString;

  function isProtoString(s) {
    if (!s) {
      return false;
    }

    var length = s.length;

    if (length < 9 /* "__proto__".length */) {
      return false;
    }

    if (s.charCodeAt(length - 1) !== 95  /* '_' */ ||
        s.charCodeAt(length - 2) !== 95  /* '_' */ ||
        s.charCodeAt(length - 3) !== 111 /* 'o' */ ||
        s.charCodeAt(length - 4) !== 116 /* 't' */ ||
        s.charCodeAt(length - 5) !== 111 /* 'o' */ ||
        s.charCodeAt(length - 6) !== 114 /* 'r' */ ||
        s.charCodeAt(length - 7) !== 112 /* 'p' */ ||
        s.charCodeAt(length - 8) !== 95  /* '_' */ ||
        s.charCodeAt(length - 9) !== 95  /* '_' */) {
      return false;
    }

    for (var i = length - 10; i >= 0; i--) {
      if (s.charCodeAt(i) !== 36 /* '$' */) {
        return false;
      }
    }

    return true;
  }

  /**
   * Comparator between two mappings where the original positions are compared.
   *
   * Optionally pass in `true` as `onlyCompareGenerated` to consider two
   * mappings with the same original source/line/column, but different generated
   * line and column the same. Useful when searching for a mapping with a
   * stubbed out mapping.
   */
  function compareByOriginalPositions(mappingA, mappingB, onlyCompareOriginal) {
    var cmp = strcmp(mappingA.source, mappingB.source);
    if (cmp !== 0) {
      return cmp;
    }

    cmp = mappingA.originalLine - mappingB.originalLine;
    if (cmp !== 0) {
      return cmp;
    }

    cmp = mappingA.originalColumn - mappingB.originalColumn;
    if (cmp !== 0 || onlyCompareOriginal) {
      return cmp;
    }

    cmp = mappingA.generatedColumn - mappingB.generatedColumn;
    if (cmp !== 0) {
      return cmp;
    }

    cmp = mappingA.generatedLine - mappingB.generatedLine;
    if (cmp !== 0) {
      return cmp;
    }

    return strcmp(mappingA.name, mappingB.name);
  }
  exports.compareByOriginalPositions = compareByOriginalPositions;

  /**
   * Comparator between two mappings with deflated source and name indices where
   * the generated positions are compared.
   *
   * Optionally pass in `true` as `onlyCompareGenerated` to consider two
   * mappings with the same generated line and column, but different
   * source/name/original line and column the same. Useful when searching for a
   * mapping with a stubbed out mapping.
   */
  function compareByGeneratedPositionsDeflated(mappingA, mappingB, onlyCompareGenerated) {
    var cmp = mappingA.generatedLine - mappingB.generatedLine;
    if (cmp !== 0) {
      return cmp;
    }

    cmp = mappingA.generatedColumn - mappingB.generatedColumn;
    if (cmp !== 0 || onlyCompareGenerated) {
      return cmp;
    }

    cmp = strcmp(mappingA.source, mappingB.source);
    if (cmp !== 0) {
      return cmp;
    }

    cmp = mappingA.originalLine - mappingB.originalLine;
    if (cmp !== 0) {
      return cmp;
    }

    cmp = mappingA.originalColumn - mappingB.originalColumn;
    if (cmp !== 0) {
      return cmp;
    }

    return strcmp(mappingA.name, mappingB.name);
  }
  exports.compareByGeneratedPositionsDeflated = compareByGeneratedPositionsDeflated;

  function strcmp(aStr1, aStr2) {
    if (aStr1 === aStr2) {
      return 0;
    }

    if (aStr1 === null) {
      return 1; // aStr2 !== null
    }

    if (aStr2 === null) {
      return -1; // aStr1 !== null
    }

    if (aStr1 > aStr2) {
      return 1;
    }

    return -1;
  }

  /**
   * Comparator between two mappings with inflated source and name strings where
   * the generated positions are compared.
   */
  function compareByGeneratedPositionsInflated(mappingA, mappingB) {
    var cmp = mappingA.generatedLine - mappingB.generatedLine;
    if (cmp !== 0) {
      return cmp;
    }

    cmp = mappingA.generatedColumn - mappingB.generatedColumn;
    if (cmp !== 0) {
      return cmp;
    }

    cmp = strcmp(mappingA.source, mappingB.source);
    if (cmp !== 0) {
      return cmp;
    }

    cmp = mappingA.originalLine - mappingB.originalLine;
    if (cmp !== 0) {
      return cmp;
    }

    cmp = mappingA.originalColumn - mappingB.originalColumn;
    if (cmp !== 0) {
      return cmp;
    }

    return strcmp(mappingA.name, mappingB.name);
  }
  exports.compareByGeneratedPositionsInflated = compareByGeneratedPositionsInflated;

  /**
   * Strip any JSON XSSI avoidance prefix from the string (as documented
   * in the source maps specification), and then parse the string as
   * JSON.
   */
  function parseSourceMapInput(str) {
    return JSON.parse(str.replace(/^\)]}'[^\n]*\n/, ''));
  }
  exports.parseSourceMapInput = parseSourceMapInput;

  /**
   * Compute the URL of a source given the the source root, the source's
   * URL, and the source map's URL.
   */
  function computeSourceURL(sourceRoot, sourceURL, sourceMapURL) {
    sourceURL = sourceURL || '';

    if (sourceRoot) {
      // This follows what Chrome does.
      if (sourceRoot[sourceRoot.length - 1] !== '/' && sourceURL[0] !== '/') {
        sourceRoot += '/';
      }
      // The spec says:
      //   Line 4: An optional source root, useful for relocating source
      //   files on a server or removing repeated values in the
      //   “sources” entry.  This value is prepended to the individual
      //   entries in the “source” field.
      sourceURL = sourceRoot + sourceURL;
    }

    // Historically, SourceMapConsumer did not take the sourceMapURL as
    // a parameter.  This mode is still somewhat supported, which is why
    // this code block is conditional.  However, it's preferable to pass
    // the source map URL to SourceMapConsumer, so that this function
    // can implement the source URL resolution algorithm as outlined in
    // the spec.  This block is basically the equivalent of:
    //    new URL(sourceURL, sourceMapURL).toString()
    // ... except it avoids using URL, which wasn't available in the
    // older releases of node still supported by this library.
    //
    // The spec says:
    //   If the sources are not absolute URLs after prepending of the
    //   “sourceRoot”, the sources are resolved relative to the
    //   SourceMap (like resolving script src in a html document).
    if (sourceMapURL) {
      var parsed = urlParse(sourceMapURL);
      if (!parsed) {
        throw new Error("sourceMapURL could not be parsed");
      }
      if (parsed.path) {
        // Strip the last path component, but keep the "/".
        var index = parsed.path.lastIndexOf('/');
        if (index >= 0) {
          parsed.path = parsed.path.substring(0, index + 1);
        }
      }
      sourceURL = join(urlGenerate(parsed), sourceURL);
    }

    return normalize(sourceURL);
  }
  exports.computeSourceURL = computeSourceURL;
  });
  var util_1 = util.getArg;
  var util_2 = util.urlParse;
  var util_3 = util.urlGenerate;
  var util_4 = util.normalize;
  var util_5 = util.join;
  var util_6 = util.isAbsolute;
  var util_7 = util.relative;
  var util_8 = util.toSetString;
  var util_9 = util.fromSetString;
  var util_10 = util.compareByOriginalPositions;
  var util_11 = util.compareByGeneratedPositionsDeflated;
  var util_12 = util.compareByGeneratedPositionsInflated;
  var util_13 = util.parseSourceMapInput;
  var util_14 = util.computeSourceURL;

  /* -*- Mode: js; js-indent-level: 2; -*- */
  /*
   * Copyright 2011 Mozilla Foundation and contributors
   * Licensed under the New BSD license. See LICENSE or:
   * http://opensource.org/licenses/BSD-3-Clause
   */


  var has = Object.prototype.hasOwnProperty;
  var hasNativeMap = typeof Map !== "undefined";

  /**
   * A data structure which is a combination of an array and a set. Adding a new
   * member is O(1), testing for membership is O(1), and finding the index of an
   * element is O(1). Removing elements from the set is not supported. Only
   * strings are supported for membership.
   */
  function ArraySet() {
    this._array = [];
    this._set = hasNativeMap ? new Map() : Object.create(null);
  }

  /**
   * Static method for creating ArraySet instances from an existing array.
   */
  ArraySet.fromArray = function ArraySet_fromArray(aArray, aAllowDuplicates) {
    var set = new ArraySet();
    for (var i = 0, len = aArray.length; i < len; i++) {
      set.add(aArray[i], aAllowDuplicates);
    }
    return set;
  };

  /**
   * Return how many unique items are in this ArraySet. If duplicates have been
   * added, than those do not count towards the size.
   *
   * @returns Number
   */
  ArraySet.prototype.size = function ArraySet_size() {
    return hasNativeMap ? this._set.size : Object.getOwnPropertyNames(this._set).length;
  };

  /**
   * Add the given string to this set.
   *
   * @param String aStr
   */
  ArraySet.prototype.add = function ArraySet_add(aStr, aAllowDuplicates) {
    var sStr = hasNativeMap ? aStr : util.toSetString(aStr);
    var isDuplicate = hasNativeMap ? this.has(aStr) : has.call(this._set, sStr);
    var idx = this._array.length;
    if (!isDuplicate || aAllowDuplicates) {
      this._array.push(aStr);
    }
    if (!isDuplicate) {
      if (hasNativeMap) {
        this._set.set(aStr, idx);
      } else {
        this._set[sStr] = idx;
      }
    }
  };

  /**
   * Is the given string a member of this set?
   *
   * @param String aStr
   */
  ArraySet.prototype.has = function ArraySet_has(aStr) {
    if (hasNativeMap) {
      return this._set.has(aStr);
    } else {
      var sStr = util.toSetString(aStr);
      return has.call(this._set, sStr);
    }
  };

  /**
   * What is the index of the given string in the array?
   *
   * @param String aStr
   */
  ArraySet.prototype.indexOf = function ArraySet_indexOf(aStr) {
    if (hasNativeMap) {
      var idx = this._set.get(aStr);
      if (idx >= 0) {
          return idx;
      }
    } else {
      var sStr = util.toSetString(aStr);
      if (has.call(this._set, sStr)) {
        return this._set[sStr];
      }
    }

    throw new Error('"' + aStr + '" is not in the set.');
  };

  /**
   * What is the element at the given index?
   *
   * @param Number aIdx
   */
  ArraySet.prototype.at = function ArraySet_at(aIdx) {
    if (aIdx >= 0 && aIdx < this._array.length) {
      return this._array[aIdx];
    }
    throw new Error('No element indexed by ' + aIdx);
  };

  /**
   * Returns the array representation of this set (which has the proper indices
   * indicated by indexOf). Note that this is a copy of the internal array used
   * for storing the members so that no one can mess with internal state.
   */
  ArraySet.prototype.toArray = function ArraySet_toArray() {
    return this._array.slice();
  };

  var ArraySet_1 = ArraySet;

  var arraySet = {
  	ArraySet: ArraySet_1
  };

  /* -*- Mode: js; js-indent-level: 2; -*- */
  /*
   * Copyright 2014 Mozilla Foundation and contributors
   * Licensed under the New BSD license. See LICENSE or:
   * http://opensource.org/licenses/BSD-3-Clause
   */



  /**
   * Determine whether mappingB is after mappingA with respect to generated
   * position.
   */
  function generatedPositionAfter(mappingA, mappingB) {
    // Optimized for most common case
    var lineA = mappingA.generatedLine;
    var lineB = mappingB.generatedLine;
    var columnA = mappingA.generatedColumn;
    var columnB = mappingB.generatedColumn;
    return lineB > lineA || lineB == lineA && columnB >= columnA ||
           util.compareByGeneratedPositionsInflated(mappingA, mappingB) <= 0;
  }

  /**
   * A data structure to provide a sorted view of accumulated mappings in a
   * performance conscious manner. It trades a neglibable overhead in general
   * case for a large speedup in case of mappings being added in order.
   */
  function MappingList() {
    this._array = [];
    this._sorted = true;
    // Serves as infimum
    this._last = {generatedLine: -1, generatedColumn: 0};
  }

  /**
   * Iterate through internal items. This method takes the same arguments that
   * `Array.prototype.forEach` takes.
   *
   * NOTE: The order of the mappings is NOT guaranteed.
   */
  MappingList.prototype.unsortedForEach =
    function MappingList_forEach(aCallback, aThisArg) {
      this._array.forEach(aCallback, aThisArg);
    };

  /**
   * Add the given source mapping.
   *
   * @param Object aMapping
   */
  MappingList.prototype.add = function MappingList_add(aMapping) {
    if (generatedPositionAfter(this._last, aMapping)) {
      this._last = aMapping;
      this._array.push(aMapping);
    } else {
      this._sorted = false;
      this._array.push(aMapping);
    }
  };

  /**
   * Returns the flat, sorted array of mappings. The mappings are sorted by
   * generated position.
   *
   * WARNING: This method returns internal data without copying, for
   * performance. The return value must NOT be mutated, and should be treated as
   * an immutable borrow. If you want to take ownership, you must make your own
   * copy.
   */
  MappingList.prototype.toArray = function MappingList_toArray() {
    if (!this._sorted) {
      this._array.sort(util.compareByGeneratedPositionsInflated);
      this._sorted = true;
    }
    return this._array;
  };

  var MappingList_1 = MappingList;

  var mappingList = {
  	MappingList: MappingList_1
  };

  /* -*- Mode: js; js-indent-level: 2; -*- */
  /*
   * Copyright 2011 Mozilla Foundation and contributors
   * Licensed under the New BSD license. See LICENSE or:
   * http://opensource.org/licenses/BSD-3-Clause
   */



  var ArraySet$1 = arraySet.ArraySet;
  var MappingList$1 = mappingList.MappingList;

  /**
   * An instance of the SourceMapGenerator represents a source map which is
   * being built incrementally. You may pass an object with the following
   * properties:
   *
   *   - file: The filename of the generated source.
   *   - sourceRoot: A root for all relative URLs in this source map.
   */
  function SourceMapGenerator(aArgs) {
    if (!aArgs) {
      aArgs = {};
    }
    this._file = util.getArg(aArgs, 'file', null);
    this._sourceRoot = util.getArg(aArgs, 'sourceRoot', null);
    this._skipValidation = util.getArg(aArgs, 'skipValidation', false);
    this._sources = new ArraySet$1();
    this._names = new ArraySet$1();
    this._mappings = new MappingList$1();
    this._sourcesContents = null;
  }

  SourceMapGenerator.prototype._version = 3;

  /**
   * Creates a new SourceMapGenerator based on a SourceMapConsumer
   *
   * @param aSourceMapConsumer The SourceMap.
   */
  SourceMapGenerator.fromSourceMap =
    function SourceMapGenerator_fromSourceMap(aSourceMapConsumer) {
      var sourceRoot = aSourceMapConsumer.sourceRoot;
      var generator = new SourceMapGenerator({
        file: aSourceMapConsumer.file,
        sourceRoot: sourceRoot
      });
      aSourceMapConsumer.eachMapping(function (mapping) {
        var newMapping = {
          generated: {
            line: mapping.generatedLine,
            column: mapping.generatedColumn
          }
        };

        if (mapping.source != null) {
          newMapping.source = mapping.source;
          if (sourceRoot != null) {
            newMapping.source = util.relative(sourceRoot, newMapping.source);
          }

          newMapping.original = {
            line: mapping.originalLine,
            column: mapping.originalColumn
          };

          if (mapping.name != null) {
            newMapping.name = mapping.name;
          }
        }

        generator.addMapping(newMapping);
      });
      aSourceMapConsumer.sources.forEach(function (sourceFile) {
        var sourceRelative = sourceFile;
        if (sourceRoot !== null) {
          sourceRelative = util.relative(sourceRoot, sourceFile);
        }

        if (!generator._sources.has(sourceRelative)) {
          generator._sources.add(sourceRelative);
        }

        var content = aSourceMapConsumer.sourceContentFor(sourceFile);
        if (content != null) {
          generator.setSourceContent(sourceFile, content);
        }
      });
      return generator;
    };

  /**
   * Add a single mapping from original source line and column to the generated
   * source's line and column for this source map being created. The mapping
   * object should have the following properties:
   *
   *   - generated: An object with the generated line and column positions.
   *   - original: An object with the original line and column positions.
   *   - source: The original source file (relative to the sourceRoot).
   *   - name: An optional original token name for this mapping.
   */
  SourceMapGenerator.prototype.addMapping =
    function SourceMapGenerator_addMapping(aArgs) {
      var generated = util.getArg(aArgs, 'generated');
      var original = util.getArg(aArgs, 'original', null);
      var source = util.getArg(aArgs, 'source', null);
      var name = util.getArg(aArgs, 'name', null);

      if (!this._skipValidation) {
        this._validateMapping(generated, original, source, name);
      }

      if (source != null) {
        source = String(source);
        if (!this._sources.has(source)) {
          this._sources.add(source);
        }
      }

      if (name != null) {
        name = String(name);
        if (!this._names.has(name)) {
          this._names.add(name);
        }
      }

      this._mappings.add({
        generatedLine: generated.line,
        generatedColumn: generated.column,
        originalLine: original != null && original.line,
        originalColumn: original != null && original.column,
        source: source,
        name: name
      });
    };

  /**
   * Set the source content for a source file.
   */
  SourceMapGenerator.prototype.setSourceContent =
    function SourceMapGenerator_setSourceContent(aSourceFile, aSourceContent) {
      var source = aSourceFile;
      if (this._sourceRoot != null) {
        source = util.relative(this._sourceRoot, source);
      }

      if (aSourceContent != null) {
        // Add the source content to the _sourcesContents map.
        // Create a new _sourcesContents map if the property is null.
        if (!this._sourcesContents) {
          this._sourcesContents = Object.create(null);
        }
        this._sourcesContents[util.toSetString(source)] = aSourceContent;
      } else if (this._sourcesContents) {
        // Remove the source file from the _sourcesContents map.
        // If the _sourcesContents map is empty, set the property to null.
        delete this._sourcesContents[util.toSetString(source)];
        if (Object.keys(this._sourcesContents).length === 0) {
          this._sourcesContents = null;
        }
      }
    };

  /**
   * Applies the mappings of a sub-source-map for a specific source file to the
   * source map being generated. Each mapping to the supplied source file is
   * rewritten using the supplied source map. Note: The resolution for the
   * resulting mappings is the minimium of this map and the supplied map.
   *
   * @param aSourceMapConsumer The source map to be applied.
   * @param aSourceFile Optional. The filename of the source file.
   *        If omitted, SourceMapConsumer's file property will be used.
   * @param aSourceMapPath Optional. The dirname of the path to the source map
   *        to be applied. If relative, it is relative to the SourceMapConsumer.
   *        This parameter is needed when the two source maps aren't in the same
   *        directory, and the source map to be applied contains relative source
   *        paths. If so, those relative source paths need to be rewritten
   *        relative to the SourceMapGenerator.
   */
  SourceMapGenerator.prototype.applySourceMap =
    function SourceMapGenerator_applySourceMap(aSourceMapConsumer, aSourceFile, aSourceMapPath) {
      var sourceFile = aSourceFile;
      // If aSourceFile is omitted, we will use the file property of the SourceMap
      if (aSourceFile == null) {
        if (aSourceMapConsumer.file == null) {
          throw new Error(
            'SourceMapGenerator.prototype.applySourceMap requires either an explicit source file, ' +
            'or the source map\'s "file" property. Both were omitted.'
          );
        }
        sourceFile = aSourceMapConsumer.file;
      }
      var sourceRoot = this._sourceRoot;
      // Make "sourceFile" relative if an absolute Url is passed.
      if (sourceRoot != null) {
        sourceFile = util.relative(sourceRoot, sourceFile);
      }
      // Applying the SourceMap can add and remove items from the sources and
      // the names array.
      var newSources = new ArraySet$1();
      var newNames = new ArraySet$1();

      // Find mappings for the "sourceFile"
      this._mappings.unsortedForEach(function (mapping) {
        if (mapping.source === sourceFile && mapping.originalLine != null) {
          // Check if it can be mapped by the source map, then update the mapping.
          var original = aSourceMapConsumer.originalPositionFor({
            line: mapping.originalLine,
            column: mapping.originalColumn
          });
          if (original.source != null) {
            // Copy mapping
            mapping.source = original.source;
            if (aSourceMapPath != null) {
              mapping.source = util.join(aSourceMapPath, mapping.source);
            }
            if (sourceRoot != null) {
              mapping.source = util.relative(sourceRoot, mapping.source);
            }
            mapping.originalLine = original.line;
            mapping.originalColumn = original.column;
            if (original.name != null) {
              mapping.name = original.name;
            }
          }
        }

        var source = mapping.source;
        if (source != null && !newSources.has(source)) {
          newSources.add(source);
        }

        var name = mapping.name;
        if (name != null && !newNames.has(name)) {
          newNames.add(name);
        }

      }, this);
      this._sources = newSources;
      this._names = newNames;

      // Copy sourcesContents of applied map.
      aSourceMapConsumer.sources.forEach(function (sourceFile) {
        var content = aSourceMapConsumer.sourceContentFor(sourceFile);
        if (content != null) {
          if (aSourceMapPath != null) {
            sourceFile = util.join(aSourceMapPath, sourceFile);
          }
          if (sourceRoot != null) {
            sourceFile = util.relative(sourceRoot, sourceFile);
          }
          this.setSourceContent(sourceFile, content);
        }
      }, this);
    };

  /**
   * A mapping can have one of the three levels of data:
   *
   *   1. Just the generated position.
   *   2. The Generated position, original position, and original source.
   *   3. Generated and original position, original source, as well as a name
   *      token.
   *
   * To maintain consistency, we validate that any new mapping being added falls
   * in to one of these categories.
   */
  SourceMapGenerator.prototype._validateMapping =
    function SourceMapGenerator_validateMapping(aGenerated, aOriginal, aSource,
                                                aName) {
      // When aOriginal is truthy but has empty values for .line and .column,
      // it is most likely a programmer error. In this case we throw a very
      // specific error message to try to guide them the right way.
      // For example: https://github.com/Polymer/polymer-bundler/pull/519
      if (aOriginal && typeof aOriginal.line !== 'number' && typeof aOriginal.column !== 'number') {
          throw new Error(
              'original.line and original.column are not numbers -- you probably meant to omit ' +
              'the original mapping entirely and only map the generated position. If so, pass ' +
              'null for the original mapping instead of an object with empty or null values.'
          );
      }

      if (aGenerated && 'line' in aGenerated && 'column' in aGenerated
          && aGenerated.line > 0 && aGenerated.column >= 0
          && !aOriginal && !aSource && !aName) {
        // Case 1.
        return;
      }
      else if (aGenerated && 'line' in aGenerated && 'column' in aGenerated
               && aOriginal && 'line' in aOriginal && 'column' in aOriginal
               && aGenerated.line > 0 && aGenerated.column >= 0
               && aOriginal.line > 0 && aOriginal.column >= 0
               && aSource) {
        // Cases 2 and 3.
        return;
      }
      else {
        throw new Error('Invalid mapping: ' + JSON.stringify({
          generated: aGenerated,
          source: aSource,
          original: aOriginal,
          name: aName
        }));
      }
    };

  /**
   * Serialize the accumulated mappings in to the stream of base 64 VLQs
   * specified by the source map format.
   */
  SourceMapGenerator.prototype._serializeMappings =
    function SourceMapGenerator_serializeMappings() {
      var previousGeneratedColumn = 0;
      var previousGeneratedLine = 1;
      var previousOriginalColumn = 0;
      var previousOriginalLine = 0;
      var previousName = 0;
      var previousSource = 0;
      var result = '';
      var next;
      var mapping;
      var nameIdx;
      var sourceIdx;

      var mappings = this._mappings.toArray();
      for (var i = 0, len = mappings.length; i < len; i++) {
        mapping = mappings[i];
        next = '';

        if (mapping.generatedLine !== previousGeneratedLine) {
          previousGeneratedColumn = 0;
          while (mapping.generatedLine !== previousGeneratedLine) {
            next += ';';
            previousGeneratedLine++;
          }
        }
        else {
          if (i > 0) {
            if (!util.compareByGeneratedPositionsInflated(mapping, mappings[i - 1])) {
              continue;
            }
            next += ',';
          }
        }

        next += base64Vlq.encode(mapping.generatedColumn
                                   - previousGeneratedColumn);
        previousGeneratedColumn = mapping.generatedColumn;

        if (mapping.source != null) {
          sourceIdx = this._sources.indexOf(mapping.source);
          next += base64Vlq.encode(sourceIdx - previousSource);
          previousSource = sourceIdx;

          // lines are stored 0-based in SourceMap spec version 3
          next += base64Vlq.encode(mapping.originalLine - 1
                                     - previousOriginalLine);
          previousOriginalLine = mapping.originalLine - 1;

          next += base64Vlq.encode(mapping.originalColumn
                                     - previousOriginalColumn);
          previousOriginalColumn = mapping.originalColumn;

          if (mapping.name != null) {
            nameIdx = this._names.indexOf(mapping.name);
            next += base64Vlq.encode(nameIdx - previousName);
            previousName = nameIdx;
          }
        }

        result += next;
      }

      return result;
    };

  SourceMapGenerator.prototype._generateSourcesContent =
    function SourceMapGenerator_generateSourcesContent(aSources, aSourceRoot) {
      return aSources.map(function (source) {
        if (!this._sourcesContents) {
          return null;
        }
        if (aSourceRoot != null) {
          source = util.relative(aSourceRoot, source);
        }
        var key = util.toSetString(source);
        return Object.prototype.hasOwnProperty.call(this._sourcesContents, key)
          ? this._sourcesContents[key]
          : null;
      }, this);
    };

  /**
   * Externalize the source map.
   */
  SourceMapGenerator.prototype.toJSON =
    function SourceMapGenerator_toJSON() {
      var map = {
        version: this._version,
        sources: this._sources.toArray(),
        names: this._names.toArray(),
        mappings: this._serializeMappings()
      };
      if (this._file != null) {
        map.file = this._file;
      }
      if (this._sourceRoot != null) {
        map.sourceRoot = this._sourceRoot;
      }
      if (this._sourcesContents) {
        map.sourcesContent = this._generateSourcesContent(map.sources, map.sourceRoot);
      }

      return map;
    };

  /**
   * Render the source map being generated to a string.
   */
  SourceMapGenerator.prototype.toString =
    function SourceMapGenerator_toString() {
      return JSON.stringify(this.toJSON());
    };

  var SourceMapGenerator_1 = SourceMapGenerator;

  var sourceMapGenerator = {
  	SourceMapGenerator: SourceMapGenerator_1
  };

  var binarySearch = createCommonjsModule(function (module, exports) {
  /* -*- Mode: js; js-indent-level: 2; -*- */
  /*
   * Copyright 2011 Mozilla Foundation and contributors
   * Licensed under the New BSD license. See LICENSE or:
   * http://opensource.org/licenses/BSD-3-Clause
   */

  exports.GREATEST_LOWER_BOUND = 1;
  exports.LEAST_UPPER_BOUND = 2;

  /**
   * Recursive implementation of binary search.
   *
   * @param aLow Indices here and lower do not contain the needle.
   * @param aHigh Indices here and higher do not contain the needle.
   * @param aNeedle The element being searched for.
   * @param aHaystack The non-empty array being searched.
   * @param aCompare Function which takes two elements and returns -1, 0, or 1.
   * @param aBias Either 'binarySearch.GREATEST_LOWER_BOUND' or
   *     'binarySearch.LEAST_UPPER_BOUND'. Specifies whether to return the
   *     closest element that is smaller than or greater than the one we are
   *     searching for, respectively, if the exact element cannot be found.
   */
  function recursiveSearch(aLow, aHigh, aNeedle, aHaystack, aCompare, aBias) {
    // This function terminates when one of the following is true:
    //
    //   1. We find the exact element we are looking for.
    //
    //   2. We did not find the exact element, but we can return the index of
    //      the next-closest element.
    //
    //   3. We did not find the exact element, and there is no next-closest
    //      element than the one we are searching for, so we return -1.
    var mid = Math.floor((aHigh - aLow) / 2) + aLow;
    var cmp = aCompare(aNeedle, aHaystack[mid], true);
    if (cmp === 0) {
      // Found the element we are looking for.
      return mid;
    }
    else if (cmp > 0) {
      // Our needle is greater than aHaystack[mid].
      if (aHigh - mid > 1) {
        // The element is in the upper half.
        return recursiveSearch(mid, aHigh, aNeedle, aHaystack, aCompare, aBias);
      }

      // The exact needle element was not found in this haystack. Determine if
      // we are in termination case (3) or (2) and return the appropriate thing.
      if (aBias == exports.LEAST_UPPER_BOUND) {
        return aHigh < aHaystack.length ? aHigh : -1;
      } else {
        return mid;
      }
    }
    else {
      // Our needle is less than aHaystack[mid].
      if (mid - aLow > 1) {
        // The element is in the lower half.
        return recursiveSearch(aLow, mid, aNeedle, aHaystack, aCompare, aBias);
      }

      // we are in termination case (3) or (2) and return the appropriate thing.
      if (aBias == exports.LEAST_UPPER_BOUND) {
        return mid;
      } else {
        return aLow < 0 ? -1 : aLow;
      }
    }
  }

  /**
   * This is an implementation of binary search which will always try and return
   * the index of the closest element if there is no exact hit. This is because
   * mappings between original and generated line/col pairs are single points,
   * and there is an implicit region between each of them, so a miss just means
   * that you aren't on the very start of a region.
   *
   * @param aNeedle The element you are looking for.
   * @param aHaystack The array that is being searched.
   * @param aCompare A function which takes the needle and an element in the
   *     array and returns -1, 0, or 1 depending on whether the needle is less
   *     than, equal to, or greater than the element, respectively.
   * @param aBias Either 'binarySearch.GREATEST_LOWER_BOUND' or
   *     'binarySearch.LEAST_UPPER_BOUND'. Specifies whether to return the
   *     closest element that is smaller than or greater than the one we are
   *     searching for, respectively, if the exact element cannot be found.
   *     Defaults to 'binarySearch.GREATEST_LOWER_BOUND'.
   */
  exports.search = function search(aNeedle, aHaystack, aCompare, aBias) {
    if (aHaystack.length === 0) {
      return -1;
    }

    var index = recursiveSearch(-1, aHaystack.length, aNeedle, aHaystack,
                                aCompare, aBias || exports.GREATEST_LOWER_BOUND);
    if (index < 0) {
      return -1;
    }

    // We have found either the exact element, or the next-closest element than
    // the one we are searching for. However, there may be more than one such
    // element. Make sure we always return the smallest of these.
    while (index - 1 >= 0) {
      if (aCompare(aHaystack[index], aHaystack[index - 1], true) !== 0) {
        break;
      }
      --index;
    }

    return index;
  };
  });
  var binarySearch_1 = binarySearch.GREATEST_LOWER_BOUND;
  var binarySearch_2 = binarySearch.LEAST_UPPER_BOUND;
  var binarySearch_3 = binarySearch.search;

  /* -*- Mode: js; js-indent-level: 2; -*- */
  /*
   * Copyright 2011 Mozilla Foundation and contributors
   * Licensed under the New BSD license. See LICENSE or:
   * http://opensource.org/licenses/BSD-3-Clause
   */

  // It turns out that some (most?) JavaScript engines don't self-host
  // `Array.prototype.sort`. This makes sense because C++ will likely remain
  // faster than JS when doing raw CPU-intensive sorting. However, when using a
  // custom comparator function, calling back and forth between the VM's C++ and
  // JIT'd JS is rather slow *and* loses JIT type information, resulting in
  // worse generated code for the comparator function than would be optimal. In
  // fact, when sorting with a comparator, these costs outweigh the benefits of
  // sorting in C++. By using our own JS-implemented Quick Sort (below), we get
  // a ~3500ms mean speed-up in `bench/bench.html`.

  /**
   * Swap the elements indexed by `x` and `y` in the array `ary`.
   *
   * @param {Array} ary
   *        The array.
   * @param {Number} x
   *        The index of the first item.
   * @param {Number} y
   *        The index of the second item.
   */
  function swap$1(ary, x, y) {
    var temp = ary[x];
    ary[x] = ary[y];
    ary[y] = temp;
  }

  /**
   * Returns a random integer within the range `low .. high` inclusive.
   *
   * @param {Number} low
   *        The lower bound on the range.
   * @param {Number} high
   *        The upper bound on the range.
   */
  function randomIntInRange(low, high) {
    return Math.round(low + (Math.random() * (high - low)));
  }

  /**
   * The Quick Sort algorithm.
   *
   * @param {Array} ary
   *        An array to sort.
   * @param {function} comparator
   *        Function to use to compare two items.
   * @param {Number} p
   *        Start index of the array
   * @param {Number} r
   *        End index of the array
   */
  function doQuickSort(ary, comparator, p, r) {
    // If our lower bound is less than our upper bound, we (1) partition the
    // array into two pieces and (2) recurse on each half. If it is not, this is
    // the empty array and our base case.

    if (p < r) {
      // (1) Partitioning.
      //
      // The partitioning chooses a pivot between `p` and `r` and moves all
      // elements that are less than or equal to the pivot to the before it, and
      // all the elements that are greater than it after it. The effect is that
      // once partition is done, the pivot is in the exact place it will be when
      // the array is put in sorted order, and it will not need to be moved
      // again. This runs in O(n) time.

      // Always choose a random pivot so that an input array which is reverse
      // sorted does not cause O(n^2) running time.
      var pivotIndex = randomIntInRange(p, r);
      var i = p - 1;

      swap$1(ary, pivotIndex, r);
      var pivot = ary[r];

      // Immediately after `j` is incremented in this loop, the following hold
      // true:
      //
      //   * Every element in `ary[p .. i]` is less than or equal to the pivot.
      //
      //   * Every element in `ary[i+1 .. j-1]` is greater than the pivot.
      for (var j = p; j < r; j++) {
        if (comparator(ary[j], pivot) <= 0) {
          i += 1;
          swap$1(ary, i, j);
        }
      }

      swap$1(ary, i + 1, j);
      var q = i + 1;

      // (2) Recurse on each half.

      doQuickSort(ary, comparator, p, q - 1);
      doQuickSort(ary, comparator, q + 1, r);
    }
  }

  /**
   * Sort the given array in-place with the given comparator function.
   *
   * @param {Array} ary
   *        An array to sort.
   * @param {function} comparator
   *        Function to use to compare two items.
   */
  var quickSort_1 = function (ary, comparator) {
    doQuickSort(ary, comparator, 0, ary.length - 1);
  };

  var quickSort = {
  	quickSort: quickSort_1
  };

  /* -*- Mode: js; js-indent-level: 2; -*- */
  /*
   * Copyright 2011 Mozilla Foundation and contributors
   * Licensed under the New BSD license. See LICENSE or:
   * http://opensource.org/licenses/BSD-3-Clause
   */



  var ArraySet$2 = arraySet.ArraySet;

  var quickSort$1 = quickSort.quickSort;

  function SourceMapConsumer(aSourceMap, aSourceMapURL) {
    var sourceMap = aSourceMap;
    if (typeof aSourceMap === 'string') {
      sourceMap = util.parseSourceMapInput(aSourceMap);
    }

    return sourceMap.sections != null
      ? new IndexedSourceMapConsumer(sourceMap, aSourceMapURL)
      : new BasicSourceMapConsumer(sourceMap, aSourceMapURL);
  }

  SourceMapConsumer.fromSourceMap = function(aSourceMap, aSourceMapURL) {
    return BasicSourceMapConsumer.fromSourceMap(aSourceMap, aSourceMapURL);
  };

  /**
   * The version of the source mapping spec that we are consuming.
   */
  SourceMapConsumer.prototype._version = 3;

  // `__generatedMappings` and `__originalMappings` are arrays that hold the
  // parsed mapping coordinates from the source map's "mappings" attribute. They
  // are lazily instantiated, accessed via the `_generatedMappings` and
  // `_originalMappings` getters respectively, and we only parse the mappings
  // and create these arrays once queried for a source location. We jump through
  // these hoops because there can be many thousands of mappings, and parsing
  // them is expensive, so we only want to do it if we must.
  //
  // Each object in the arrays is of the form:
  //
  //     {
  //       generatedLine: The line number in the generated code,
  //       generatedColumn: The column number in the generated code,
  //       source: The path to the original source file that generated this
  //               chunk of code,
  //       originalLine: The line number in the original source that
  //                     corresponds to this chunk of generated code,
  //       originalColumn: The column number in the original source that
  //                       corresponds to this chunk of generated code,
  //       name: The name of the original symbol which generated this chunk of
  //             code.
  //     }
  //
  // All properties except for `generatedLine` and `generatedColumn` can be
  // `null`.
  //
  // `_generatedMappings` is ordered by the generated positions.
  //
  // `_originalMappings` is ordered by the original positions.

  SourceMapConsumer.prototype.__generatedMappings = null;
  Object.defineProperty(SourceMapConsumer.prototype, '_generatedMappings', {
    configurable: true,
    enumerable: true,
    get: function () {
      if (!this.__generatedMappings) {
        this._parseMappings(this._mappings, this.sourceRoot);
      }

      return this.__generatedMappings;
    }
  });

  SourceMapConsumer.prototype.__originalMappings = null;
  Object.defineProperty(SourceMapConsumer.prototype, '_originalMappings', {
    configurable: true,
    enumerable: true,
    get: function () {
      if (!this.__originalMappings) {
        this._parseMappings(this._mappings, this.sourceRoot);
      }

      return this.__originalMappings;
    }
  });

  SourceMapConsumer.prototype._charIsMappingSeparator =
    function SourceMapConsumer_charIsMappingSeparator(aStr, index) {
      var c = aStr.charAt(index);
      return c === ";" || c === ",";
    };

  /**
   * Parse the mappings in a string in to a data structure which we can easily
   * query (the ordered arrays in the `this.__generatedMappings` and
   * `this.__originalMappings` properties).
   */
  SourceMapConsumer.prototype._parseMappings =
    function SourceMapConsumer_parseMappings(aStr, aSourceRoot) {
      throw new Error("Subclasses must implement _parseMappings");
    };

  SourceMapConsumer.GENERATED_ORDER = 1;
  SourceMapConsumer.ORIGINAL_ORDER = 2;

  SourceMapConsumer.GREATEST_LOWER_BOUND = 1;
  SourceMapConsumer.LEAST_UPPER_BOUND = 2;

  /**
   * Iterate over each mapping between an original source/line/column and a
   * generated line/column in this source map.
   *
   * @param Function aCallback
   *        The function that is called with each mapping.
   * @param Object aContext
   *        Optional. If specified, this object will be the value of `this` every
   *        time that `aCallback` is called.
   * @param aOrder
   *        Either `SourceMapConsumer.GENERATED_ORDER` or
   *        `SourceMapConsumer.ORIGINAL_ORDER`. Specifies whether you want to
   *        iterate over the mappings sorted by the generated file's line/column
   *        order or the original's source/line/column order, respectively. Defaults to
   *        `SourceMapConsumer.GENERATED_ORDER`.
   */
  SourceMapConsumer.prototype.eachMapping =
    function SourceMapConsumer_eachMapping(aCallback, aContext, aOrder) {
      var context = aContext || null;
      var order = aOrder || SourceMapConsumer.GENERATED_ORDER;

      var mappings;
      switch (order) {
      case SourceMapConsumer.GENERATED_ORDER:
        mappings = this._generatedMappings;
        break;
      case SourceMapConsumer.ORIGINAL_ORDER:
        mappings = this._originalMappings;
        break;
      default:
        throw new Error("Unknown order of iteration.");
      }

      var sourceRoot = this.sourceRoot;
      mappings.map(function (mapping) {
        var source = mapping.source === null ? null : this._sources.at(mapping.source);
        source = util.computeSourceURL(sourceRoot, source, this._sourceMapURL);
        return {
          source: source,
          generatedLine: mapping.generatedLine,
          generatedColumn: mapping.generatedColumn,
          originalLine: mapping.originalLine,
          originalColumn: mapping.originalColumn,
          name: mapping.name === null ? null : this._names.at(mapping.name)
        };
      }, this).forEach(aCallback, context);
    };

  /**
   * Returns all generated line and column information for the original source,
   * line, and column provided. If no column is provided, returns all mappings
   * corresponding to a either the line we are searching for or the next
   * closest line that has any mappings. Otherwise, returns all mappings
   * corresponding to the given line and either the column we are searching for
   * or the next closest column that has any offsets.
   *
   * The only argument is an object with the following properties:
   *
   *   - source: The filename of the original source.
   *   - line: The line number in the original source.  The line number is 1-based.
   *   - column: Optional. the column number in the original source.
   *    The column number is 0-based.
   *
   * and an array of objects is returned, each with the following properties:
   *
   *   - line: The line number in the generated source, or null.  The
   *    line number is 1-based.
   *   - column: The column number in the generated source, or null.
   *    The column number is 0-based.
   */
  SourceMapConsumer.prototype.allGeneratedPositionsFor =
    function SourceMapConsumer_allGeneratedPositionsFor(aArgs) {
      var line = util.getArg(aArgs, 'line');

      // When there is no exact match, BasicSourceMapConsumer.prototype._findMapping
      // returns the index of the closest mapping less than the needle. By
      // setting needle.originalColumn to 0, we thus find the last mapping for
      // the given line, provided such a mapping exists.
      var needle = {
        source: util.getArg(aArgs, 'source'),
        originalLine: line,
        originalColumn: util.getArg(aArgs, 'column', 0)
      };

      needle.source = this._findSourceIndex(needle.source);
      if (needle.source < 0) {
        return [];
      }

      var mappings = [];

      var index = this._findMapping(needle,
                                    this._originalMappings,
                                    "originalLine",
                                    "originalColumn",
                                    util.compareByOriginalPositions,
                                    binarySearch.LEAST_UPPER_BOUND);
      if (index >= 0) {
        var mapping = this._originalMappings[index];

        if (aArgs.column === undefined) {
          var originalLine = mapping.originalLine;

          // Iterate until either we run out of mappings, or we run into
          // a mapping for a different line than the one we found. Since
          // mappings are sorted, this is guaranteed to find all mappings for
          // the line we found.
          while (mapping && mapping.originalLine === originalLine) {
            mappings.push({
              line: util.getArg(mapping, 'generatedLine', null),
              column: util.getArg(mapping, 'generatedColumn', null),
              lastColumn: util.getArg(mapping, 'lastGeneratedColumn', null)
            });

            mapping = this._originalMappings[++index];
          }
        } else {
          var originalColumn = mapping.originalColumn;

          // Iterate until either we run out of mappings, or we run into
          // a mapping for a different line than the one we were searching for.
          // Since mappings are sorted, this is guaranteed to find all mappings for
          // the line we are searching for.
          while (mapping &&
                 mapping.originalLine === line &&
                 mapping.originalColumn == originalColumn) {
            mappings.push({
              line: util.getArg(mapping, 'generatedLine', null),
              column: util.getArg(mapping, 'generatedColumn', null),
              lastColumn: util.getArg(mapping, 'lastGeneratedColumn', null)
            });

            mapping = this._originalMappings[++index];
          }
        }
      }

      return mappings;
    };

  var SourceMapConsumer_1 = SourceMapConsumer;

  /**
   * A BasicSourceMapConsumer instance represents a parsed source map which we can
   * query for information about the original file positions by giving it a file
   * position in the generated source.
   *
   * The first parameter is the raw source map (either as a JSON string, or
   * already parsed to an object). According to the spec, source maps have the
   * following attributes:
   *
   *   - version: Which version of the source map spec this map is following.
   *   - sources: An array of URLs to the original source files.
   *   - names: An array of identifiers which can be referrenced by individual mappings.
   *   - sourceRoot: Optional. The URL root from which all sources are relative.
   *   - sourcesContent: Optional. An array of contents of the original source files.
   *   - mappings: A string of base64 VLQs which contain the actual mappings.
   *   - file: Optional. The generated file this source map is associated with.
   *
   * Here is an example source map, taken from the source map spec[0]:
   *
   *     {
   *       version : 3,
   *       file: "out.js",
   *       sourceRoot : "",
   *       sources: ["foo.js", "bar.js"],
   *       names: ["src", "maps", "are", "fun"],
   *       mappings: "AA,AB;;ABCDE;"
   *     }
   *
   * The second parameter, if given, is a string whose value is the URL
   * at which the source map was found.  This URL is used to compute the
   * sources array.
   *
   * [0]: https://docs.google.com/document/d/1U1RGAehQwRypUTovF1KRlpiOFze0b-_2gc6fAH0KY0k/edit?pli=1#
   */
  function BasicSourceMapConsumer(aSourceMap, aSourceMapURL) {
    var sourceMap = aSourceMap;
    if (typeof aSourceMap === 'string') {
      sourceMap = util.parseSourceMapInput(aSourceMap);
    }

    var version = util.getArg(sourceMap, 'version');
    var sources = util.getArg(sourceMap, 'sources');
    // Sass 3.3 leaves out the 'names' array, so we deviate from the spec (which
    // requires the array) to play nice here.
    var names = util.getArg(sourceMap, 'names', []);
    var sourceRoot = util.getArg(sourceMap, 'sourceRoot', null);
    var sourcesContent = util.getArg(sourceMap, 'sourcesContent', null);
    var mappings = util.getArg(sourceMap, 'mappings');
    var file = util.getArg(sourceMap, 'file', null);

    // Once again, Sass deviates from the spec and supplies the version as a
    // string rather than a number, so we use loose equality checking here.
    if (version != this._version) {
      throw new Error('Unsupported version: ' + version);
    }

    if (sourceRoot) {
      sourceRoot = util.normalize(sourceRoot);
    }

    sources = sources
      .map(String)
      // Some source maps produce relative source paths like "./foo.js" instead of
      // "foo.js".  Normalize these first so that future comparisons will succeed.
      // See bugzil.la/1090768.
      .map(util.normalize)
      // Always ensure that absolute sources are internally stored relative to
      // the source root, if the source root is absolute. Not doing this would
      // be particularly problematic when the source root is a prefix of the
      // source (valid, but why??). See github issue #199 and bugzil.la/1188982.
      .map(function (source) {
        return sourceRoot && util.isAbsolute(sourceRoot) && util.isAbsolute(source)
          ? util.relative(sourceRoot, source)
          : source;
      });

    // Pass `true` below to allow duplicate names and sources. While source maps
    // are intended to be compressed and deduplicated, the TypeScript compiler
    // sometimes generates source maps with duplicates in them. See Github issue
    // #72 and bugzil.la/889492.
    this._names = ArraySet$2.fromArray(names.map(String), true);
    this._sources = ArraySet$2.fromArray(sources, true);

    this._absoluteSources = this._sources.toArray().map(function (s) {
      return util.computeSourceURL(sourceRoot, s, aSourceMapURL);
    });

    this.sourceRoot = sourceRoot;
    this.sourcesContent = sourcesContent;
    this._mappings = mappings;
    this._sourceMapURL = aSourceMapURL;
    this.file = file;
  }

  BasicSourceMapConsumer.prototype = Object.create(SourceMapConsumer.prototype);
  BasicSourceMapConsumer.prototype.consumer = SourceMapConsumer;

  /**
   * Utility function to find the index of a source.  Returns -1 if not
   * found.
   */
  BasicSourceMapConsumer.prototype._findSourceIndex = function(aSource) {
    var relativeSource = aSource;
    if (this.sourceRoot != null) {
      relativeSource = util.relative(this.sourceRoot, relativeSource);
    }

    if (this._sources.has(relativeSource)) {
      return this._sources.indexOf(relativeSource);
    }

    // Maybe aSource is an absolute URL as returned by |sources|.  In
    // this case we can't simply undo the transform.
    var i;
    for (i = 0; i < this._absoluteSources.length; ++i) {
      if (this._absoluteSources[i] == aSource) {
        return i;
      }
    }

    return -1;
  };

  /**
   * Create a BasicSourceMapConsumer from a SourceMapGenerator.
   *
   * @param SourceMapGenerator aSourceMap
   *        The source map that will be consumed.
   * @param String aSourceMapURL
   *        The URL at which the source map can be found (optional)
   * @returns BasicSourceMapConsumer
   */
  BasicSourceMapConsumer.fromSourceMap =
    function SourceMapConsumer_fromSourceMap(aSourceMap, aSourceMapURL) {
      var smc = Object.create(BasicSourceMapConsumer.prototype);

      var names = smc._names = ArraySet$2.fromArray(aSourceMap._names.toArray(), true);
      var sources = smc._sources = ArraySet$2.fromArray(aSourceMap._sources.toArray(), true);
      smc.sourceRoot = aSourceMap._sourceRoot;
      smc.sourcesContent = aSourceMap._generateSourcesContent(smc._sources.toArray(),
                                                              smc.sourceRoot);
      smc.file = aSourceMap._file;
      smc._sourceMapURL = aSourceMapURL;
      smc._absoluteSources = smc._sources.toArray().map(function (s) {
        return util.computeSourceURL(smc.sourceRoot, s, aSourceMapURL);
      });

      // Because we are modifying the entries (by converting string sources and
      // names to indices into the sources and names ArraySets), we have to make
      // a copy of the entry or else bad things happen. Shared mutable state
      // strikes again! See github issue #191.

      var generatedMappings = aSourceMap._mappings.toArray().slice();
      var destGeneratedMappings = smc.__generatedMappings = [];
      var destOriginalMappings = smc.__originalMappings = [];

      for (var i = 0, length = generatedMappings.length; i < length; i++) {
        var srcMapping = generatedMappings[i];
        var destMapping = new Mapping;
        destMapping.generatedLine = srcMapping.generatedLine;
        destMapping.generatedColumn = srcMapping.generatedColumn;

        if (srcMapping.source) {
          destMapping.source = sources.indexOf(srcMapping.source);
          destMapping.originalLine = srcMapping.originalLine;
          destMapping.originalColumn = srcMapping.originalColumn;

          if (srcMapping.name) {
            destMapping.name = names.indexOf(srcMapping.name);
          }

          destOriginalMappings.push(destMapping);
        }

        destGeneratedMappings.push(destMapping);
      }

      quickSort$1(smc.__originalMappings, util.compareByOriginalPositions);

      return smc;
    };

  /**
   * The version of the source mapping spec that we are consuming.
   */
  BasicSourceMapConsumer.prototype._version = 3;

  /**
   * The list of original sources.
   */
  Object.defineProperty(BasicSourceMapConsumer.prototype, 'sources', {
    get: function () {
      return this._absoluteSources.slice();
    }
  });

  /**
   * Provide the JIT with a nice shape / hidden class.
   */
  function Mapping() {
    this.generatedLine = 0;
    this.generatedColumn = 0;
    this.source = null;
    this.originalLine = null;
    this.originalColumn = null;
    this.name = null;
  }

  /**
   * Parse the mappings in a string in to a data structure which we can easily
   * query (the ordered arrays in the `this.__generatedMappings` and
   * `this.__originalMappings` properties).
   */
  BasicSourceMapConsumer.prototype._parseMappings =
    function SourceMapConsumer_parseMappings(aStr, aSourceRoot) {
      var generatedLine = 1;
      var previousGeneratedColumn = 0;
      var previousOriginalLine = 0;
      var previousOriginalColumn = 0;
      var previousSource = 0;
      var previousName = 0;
      var length = aStr.length;
      var index = 0;
      var cachedSegments = {};
      var temp = {};
      var originalMappings = [];
      var generatedMappings = [];
      var mapping, str, segment, end, value;

      while (index < length) {
        if (aStr.charAt(index) === ';') {
          generatedLine++;
          index++;
          previousGeneratedColumn = 0;
        }
        else if (aStr.charAt(index) === ',') {
          index++;
        }
        else {
          mapping = new Mapping();
          mapping.generatedLine = generatedLine;

          // Because each offset is encoded relative to the previous one,
          // many segments often have the same encoding. We can exploit this
          // fact by caching the parsed variable length fields of each segment,
          // allowing us to avoid a second parse if we encounter the same
          // segment again.
          for (end = index; end < length; end++) {
            if (this._charIsMappingSeparator(aStr, end)) {
              break;
            }
          }
          str = aStr.slice(index, end);

          segment = cachedSegments[str];
          if (segment) {
            index += str.length;
          } else {
            segment = [];
            while (index < end) {
              base64Vlq.decode(aStr, index, temp);
              value = temp.value;
              index = temp.rest;
              segment.push(value);
            }

            if (segment.length === 2) {
              throw new Error('Found a source, but no line and column');
            }

            if (segment.length === 3) {
              throw new Error('Found a source and line, but no column');
            }

            cachedSegments[str] = segment;
          }

          // Generated column.
          mapping.generatedColumn = previousGeneratedColumn + segment[0];
          previousGeneratedColumn = mapping.generatedColumn;

          if (segment.length > 1) {
            // Original source.
            mapping.source = previousSource + segment[1];
            previousSource += segment[1];

            // Original line.
            mapping.originalLine = previousOriginalLine + segment[2];
            previousOriginalLine = mapping.originalLine;
            // Lines are stored 0-based
            mapping.originalLine += 1;

            // Original column.
            mapping.originalColumn = previousOriginalColumn + segment[3];
            previousOriginalColumn = mapping.originalColumn;

            if (segment.length > 4) {
              // Original name.
              mapping.name = previousName + segment[4];
              previousName += segment[4];
            }
          }

          generatedMappings.push(mapping);
          if (typeof mapping.originalLine === 'number') {
            originalMappings.push(mapping);
          }
        }
      }

      quickSort$1(generatedMappings, util.compareByGeneratedPositionsDeflated);
      this.__generatedMappings = generatedMappings;

      quickSort$1(originalMappings, util.compareByOriginalPositions);
      this.__originalMappings = originalMappings;
    };

  /**
   * Find the mapping that best matches the hypothetical "needle" mapping that
   * we are searching for in the given "haystack" of mappings.
   */
  BasicSourceMapConsumer.prototype._findMapping =
    function SourceMapConsumer_findMapping(aNeedle, aMappings, aLineName,
                                           aColumnName, aComparator, aBias) {
      // To return the position we are searching for, we must first find the
      // mapping for the given position and then return the opposite position it
      // points to. Because the mappings are sorted, we can use binary search to
      // find the best mapping.

      if (aNeedle[aLineName] <= 0) {
        throw new TypeError('Line must be greater than or equal to 1, got '
                            + aNeedle[aLineName]);
      }
      if (aNeedle[aColumnName] < 0) {
        throw new TypeError('Column must be greater than or equal to 0, got '
                            + aNeedle[aColumnName]);
      }

      return binarySearch.search(aNeedle, aMappings, aComparator, aBias);
    };

  /**
   * Compute the last column for each generated mapping. The last column is
   * inclusive.
   */
  BasicSourceMapConsumer.prototype.computeColumnSpans =
    function SourceMapConsumer_computeColumnSpans() {
      for (var index = 0; index < this._generatedMappings.length; ++index) {
        var mapping = this._generatedMappings[index];

        // Mappings do not contain a field for the last generated columnt. We
        // can come up with an optimistic estimate, however, by assuming that
        // mappings are contiguous (i.e. given two consecutive mappings, the
        // first mapping ends where the second one starts).
        if (index + 1 < this._generatedMappings.length) {
          var nextMapping = this._generatedMappings[index + 1];

          if (mapping.generatedLine === nextMapping.generatedLine) {
            mapping.lastGeneratedColumn = nextMapping.generatedColumn - 1;
            continue;
          }
        }

        // The last mapping for each line spans the entire line.
        mapping.lastGeneratedColumn = Infinity;
      }
    };

  /**
   * Returns the original source, line, and column information for the generated
   * source's line and column positions provided. The only argument is an object
   * with the following properties:
   *
   *   - line: The line number in the generated source.  The line number
   *     is 1-based.
   *   - column: The column number in the generated source.  The column
   *     number is 0-based.
   *   - bias: Either 'SourceMapConsumer.GREATEST_LOWER_BOUND' or
   *     'SourceMapConsumer.LEAST_UPPER_BOUND'. Specifies whether to return the
   *     closest element that is smaller than or greater than the one we are
   *     searching for, respectively, if the exact element cannot be found.
   *     Defaults to 'SourceMapConsumer.GREATEST_LOWER_BOUND'.
   *
   * and an object is returned with the following properties:
   *
   *   - source: The original source file, or null.
   *   - line: The line number in the original source, or null.  The
   *     line number is 1-based.
   *   - column: The column number in the original source, or null.  The
   *     column number is 0-based.
   *   - name: The original identifier, or null.
   */
  BasicSourceMapConsumer.prototype.originalPositionFor =
    function SourceMapConsumer_originalPositionFor(aArgs) {
      var needle = {
        generatedLine: util.getArg(aArgs, 'line'),
        generatedColumn: util.getArg(aArgs, 'column')
      };

      var index = this._findMapping(
        needle,
        this._generatedMappings,
        "generatedLine",
        "generatedColumn",
        util.compareByGeneratedPositionsDeflated,
        util.getArg(aArgs, 'bias', SourceMapConsumer.GREATEST_LOWER_BOUND)
      );

      if (index >= 0) {
        var mapping = this._generatedMappings[index];

        if (mapping.generatedLine === needle.generatedLine) {
          var source = util.getArg(mapping, 'source', null);
          if (source !== null) {
            source = this._sources.at(source);
            source = util.computeSourceURL(this.sourceRoot, source, this._sourceMapURL);
          }
          var name = util.getArg(mapping, 'name', null);
          if (name !== null) {
            name = this._names.at(name);
          }
          return {
            source: source,
            line: util.getArg(mapping, 'originalLine', null),
            column: util.getArg(mapping, 'originalColumn', null),
            name: name
          };
        }
      }

      return {
        source: null,
        line: null,
        column: null,
        name: null
      };
    };

  /**
   * Return true if we have the source content for every source in the source
   * map, false otherwise.
   */
  BasicSourceMapConsumer.prototype.hasContentsOfAllSources =
    function BasicSourceMapConsumer_hasContentsOfAllSources() {
      if (!this.sourcesContent) {
        return false;
      }
      return this.sourcesContent.length >= this._sources.size() &&
        !this.sourcesContent.some(function (sc) { return sc == null; });
    };

  /**
   * Returns the original source content. The only argument is the url of the
   * original source file. Returns null if no original source content is
   * available.
   */
  BasicSourceMapConsumer.prototype.sourceContentFor =
    function SourceMapConsumer_sourceContentFor(aSource, nullOnMissing) {
      if (!this.sourcesContent) {
        return null;
      }

      var index = this._findSourceIndex(aSource);
      if (index >= 0) {
        return this.sourcesContent[index];
      }

      var relativeSource = aSource;
      if (this.sourceRoot != null) {
        relativeSource = util.relative(this.sourceRoot, relativeSource);
      }

      var url;
      if (this.sourceRoot != null
          && (url = util.urlParse(this.sourceRoot))) {
        // XXX: file:// URIs and absolute paths lead to unexpected behavior for
        // many users. We can help them out when they expect file:// URIs to
        // behave like it would if they were running a local HTTP server. See
        // https://bugzilla.mozilla.org/show_bug.cgi?id=885597.
        var fileUriAbsPath = relativeSource.replace(/^file:\/\//, "");
        if (url.scheme == "file"
            && this._sources.has(fileUriAbsPath)) {
          return this.sourcesContent[this._sources.indexOf(fileUriAbsPath)]
        }

        if ((!url.path || url.path == "/")
            && this._sources.has("/" + relativeSource)) {
          return this.sourcesContent[this._sources.indexOf("/" + relativeSource)];
        }
      }

      // This function is used recursively from
      // IndexedSourceMapConsumer.prototype.sourceContentFor. In that case, we
      // don't want to throw if we can't find the source - we just want to
      // return null, so we provide a flag to exit gracefully.
      if (nullOnMissing) {
        return null;
      }
      else {
        throw new Error('"' + relativeSource + '" is not in the SourceMap.');
      }
    };

  /**
   * Returns the generated line and column information for the original source,
   * line, and column positions provided. The only argument is an object with
   * the following properties:
   *
   *   - source: The filename of the original source.
   *   - line: The line number in the original source.  The line number
   *     is 1-based.
   *   - column: The column number in the original source.  The column
   *     number is 0-based.
   *   - bias: Either 'SourceMapConsumer.GREATEST_LOWER_BOUND' or
   *     'SourceMapConsumer.LEAST_UPPER_BOUND'. Specifies whether to return the
   *     closest element that is smaller than or greater than the one we are
   *     searching for, respectively, if the exact element cannot be found.
   *     Defaults to 'SourceMapConsumer.GREATEST_LOWER_BOUND'.
   *
   * and an object is returned with the following properties:
   *
   *   - line: The line number in the generated source, or null.  The
   *     line number is 1-based.
   *   - column: The column number in the generated source, or null.
   *     The column number is 0-based.
   */
  BasicSourceMapConsumer.prototype.generatedPositionFor =
    function SourceMapConsumer_generatedPositionFor(aArgs) {
      var source = util.getArg(aArgs, 'source');
      source = this._findSourceIndex(source);
      if (source < 0) {
        return {
          line: null,
          column: null,
          lastColumn: null
        };
      }

      var needle = {
        source: source,
        originalLine: util.getArg(aArgs, 'line'),
        originalColumn: util.getArg(aArgs, 'column')
      };

      var index = this._findMapping(
        needle,
        this._originalMappings,
        "originalLine",
        "originalColumn",
        util.compareByOriginalPositions,
        util.getArg(aArgs, 'bias', SourceMapConsumer.GREATEST_LOWER_BOUND)
      );

      if (index >= 0) {
        var mapping = this._originalMappings[index];

        if (mapping.source === needle.source) {
          return {
            line: util.getArg(mapping, 'generatedLine', null),
            column: util.getArg(mapping, 'generatedColumn', null),
            lastColumn: util.getArg(mapping, 'lastGeneratedColumn', null)
          };
        }
      }

      return {
        line: null,
        column: null,
        lastColumn: null
      };
    };

  var BasicSourceMapConsumer_1 = BasicSourceMapConsumer;

  /**
   * An IndexedSourceMapConsumer instance represents a parsed source map which
   * we can query for information. It differs from BasicSourceMapConsumer in
   * that it takes "indexed" source maps (i.e. ones with a "sections" field) as
   * input.
   *
   * The first parameter is a raw source map (either as a JSON string, or already
   * parsed to an object). According to the spec for indexed source maps, they
   * have the following attributes:
   *
   *   - version: Which version of the source map spec this map is following.
   *   - file: Optional. The generated file this source map is associated with.
   *   - sections: A list of section definitions.
   *
   * Each value under the "sections" field has two fields:
   *   - offset: The offset into the original specified at which this section
   *       begins to apply, defined as an object with a "line" and "column"
   *       field.
   *   - map: A source map definition. This source map could also be indexed,
   *       but doesn't have to be.
   *
   * Instead of the "map" field, it's also possible to have a "url" field
   * specifying a URL to retrieve a source map from, but that's currently
   * unsupported.
   *
   * Here's an example source map, taken from the source map spec[0], but
   * modified to omit a section which uses the "url" field.
   *
   *  {
   *    version : 3,
   *    file: "app.js",
   *    sections: [{
   *      offset: {line:100, column:10},
   *      map: {
   *        version : 3,
   *        file: "section.js",
   *        sources: ["foo.js", "bar.js"],
   *        names: ["src", "maps", "are", "fun"],
   *        mappings: "AAAA,E;;ABCDE;"
   *      }
   *    }],
   *  }
   *
   * The second parameter, if given, is a string whose value is the URL
   * at which the source map was found.  This URL is used to compute the
   * sources array.
   *
   * [0]: https://docs.google.com/document/d/1U1RGAehQwRypUTovF1KRlpiOFze0b-_2gc6fAH0KY0k/edit#heading=h.535es3xeprgt
   */
  function IndexedSourceMapConsumer(aSourceMap, aSourceMapURL) {
    var sourceMap = aSourceMap;
    if (typeof aSourceMap === 'string') {
      sourceMap = util.parseSourceMapInput(aSourceMap);
    }

    var version = util.getArg(sourceMap, 'version');
    var sections = util.getArg(sourceMap, 'sections');

    if (version != this._version) {
      throw new Error('Unsupported version: ' + version);
    }

    this._sources = new ArraySet$2();
    this._names = new ArraySet$2();

    var lastOffset = {
      line: -1,
      column: 0
    };
    this._sections = sections.map(function (s) {
      if (s.url) {
        // The url field will require support for asynchronicity.
        // See https://github.com/mozilla/source-map/issues/16
        throw new Error('Support for url field in sections not implemented.');
      }
      var offset = util.getArg(s, 'offset');
      var offsetLine = util.getArg(offset, 'line');
      var offsetColumn = util.getArg(offset, 'column');

      if (offsetLine < lastOffset.line ||
          (offsetLine === lastOffset.line && offsetColumn < lastOffset.column)) {
        throw new Error('Section offsets must be ordered and non-overlapping.');
      }
      lastOffset = offset;

      return {
        generatedOffset: {
          // The offset fields are 0-based, but we use 1-based indices when
          // encoding/decoding from VLQ.
          generatedLine: offsetLine + 1,
          generatedColumn: offsetColumn + 1
        },
        consumer: new SourceMapConsumer(util.getArg(s, 'map'), aSourceMapURL)
      }
    });
  }

  IndexedSourceMapConsumer.prototype = Object.create(SourceMapConsumer.prototype);
  IndexedSourceMapConsumer.prototype.constructor = SourceMapConsumer;

  /**
   * The version of the source mapping spec that we are consuming.
   */
  IndexedSourceMapConsumer.prototype._version = 3;

  /**
   * The list of original sources.
   */
  Object.defineProperty(IndexedSourceMapConsumer.prototype, 'sources', {
    get: function () {
      var sources = [];
      for (var i = 0; i < this._sections.length; i++) {
        for (var j = 0; j < this._sections[i].consumer.sources.length; j++) {
          sources.push(this._sections[i].consumer.sources[j]);
        }
      }
      return sources;
    }
  });

  /**
   * Returns the original source, line, and column information for the generated
   * source's line and column positions provided. The only argument is an object
   * with the following properties:
   *
   *   - line: The line number in the generated source.  The line number
   *     is 1-based.
   *   - column: The column number in the generated source.  The column
   *     number is 0-based.
   *
   * and an object is returned with the following properties:
   *
   *   - source: The original source file, or null.
   *   - line: The line number in the original source, or null.  The
   *     line number is 1-based.
   *   - column: The column number in the original source, or null.  The
   *     column number is 0-based.
   *   - name: The original identifier, or null.
   */
  IndexedSourceMapConsumer.prototype.originalPositionFor =
    function IndexedSourceMapConsumer_originalPositionFor(aArgs) {
      var needle = {
        generatedLine: util.getArg(aArgs, 'line'),
        generatedColumn: util.getArg(aArgs, 'column')
      };

      // Find the section containing the generated position we're trying to map
      // to an original position.
      var sectionIndex = binarySearch.search(needle, this._sections,
        function(needle, section) {
          var cmp = needle.generatedLine - section.generatedOffset.generatedLine;
          if (cmp) {
            return cmp;
          }

          return (needle.generatedColumn -
                  section.generatedOffset.generatedColumn);
        });
      var section = this._sections[sectionIndex];

      if (!section) {
        return {
          source: null,
          line: null,
          column: null,
          name: null
        };
      }

      return section.consumer.originalPositionFor({
        line: needle.generatedLine -
          (section.generatedOffset.generatedLine - 1),
        column: needle.generatedColumn -
          (section.generatedOffset.generatedLine === needle.generatedLine
           ? section.generatedOffset.generatedColumn - 1
           : 0),
        bias: aArgs.bias
      });
    };

  /**
   * Return true if we have the source content for every source in the source
   * map, false otherwise.
   */
  IndexedSourceMapConsumer.prototype.hasContentsOfAllSources =
    function IndexedSourceMapConsumer_hasContentsOfAllSources() {
      return this._sections.every(function (s) {
        return s.consumer.hasContentsOfAllSources();
      });
    };

  /**
   * Returns the original source content. The only argument is the url of the
   * original source file. Returns null if no original source content is
   * available.
   */
  IndexedSourceMapConsumer.prototype.sourceContentFor =
    function IndexedSourceMapConsumer_sourceContentFor(aSource, nullOnMissing) {
      for (var i = 0; i < this._sections.length; i++) {
        var section = this._sections[i];

        var content = section.consumer.sourceContentFor(aSource, true);
        if (content) {
          return content;
        }
      }
      if (nullOnMissing) {
        return null;
      }
      else {
        throw new Error('"' + aSource + '" is not in the SourceMap.');
      }
    };

  /**
   * Returns the generated line and column information for the original source,
   * line, and column positions provided. The only argument is an object with
   * the following properties:
   *
   *   - source: The filename of the original source.
   *   - line: The line number in the original source.  The line number
   *     is 1-based.
   *   - column: The column number in the original source.  The column
   *     number is 0-based.
   *
   * and an object is returned with the following properties:
   *
   *   - line: The line number in the generated source, or null.  The
   *     line number is 1-based. 
   *   - column: The column number in the generated source, or null.
   *     The column number is 0-based.
   */
  IndexedSourceMapConsumer.prototype.generatedPositionFor =
    function IndexedSourceMapConsumer_generatedPositionFor(aArgs) {
      for (var i = 0; i < this._sections.length; i++) {
        var section = this._sections[i];

        // Only consider this section if the requested source is in the list of
        // sources of the consumer.
        if (section.consumer._findSourceIndex(util.getArg(aArgs, 'source')) === -1) {
          continue;
        }
        var generatedPosition = section.consumer.generatedPositionFor(aArgs);
        if (generatedPosition) {
          var ret = {
            line: generatedPosition.line +
              (section.generatedOffset.generatedLine - 1),
            column: generatedPosition.column +
              (section.generatedOffset.generatedLine === generatedPosition.line
               ? section.generatedOffset.generatedColumn - 1
               : 0)
          };
          return ret;
        }
      }

      return {
        line: null,
        column: null
      };
    };

  /**
   * Parse the mappings in a string in to a data structure which we can easily
   * query (the ordered arrays in the `this.__generatedMappings` and
   * `this.__originalMappings` properties).
   */
  IndexedSourceMapConsumer.prototype._parseMappings =
    function IndexedSourceMapConsumer_parseMappings(aStr, aSourceRoot) {
      this.__generatedMappings = [];
      this.__originalMappings = [];
      for (var i = 0; i < this._sections.length; i++) {
        var section = this._sections[i];
        var sectionMappings = section.consumer._generatedMappings;
        for (var j = 0; j < sectionMappings.length; j++) {
          var mapping = sectionMappings[j];

          var source = section.consumer._sources.at(mapping.source);
          source = util.computeSourceURL(section.consumer.sourceRoot, source, this._sourceMapURL);
          this._sources.add(source);
          source = this._sources.indexOf(source);

          var name = null;
          if (mapping.name) {
            name = section.consumer._names.at(mapping.name);
            this._names.add(name);
            name = this._names.indexOf(name);
          }

          // The mappings coming from the consumer for the section have
          // generated positions relative to the start of the section, so we
          // need to offset them to be relative to the start of the concatenated
          // generated file.
          var adjustedMapping = {
            source: source,
            generatedLine: mapping.generatedLine +
              (section.generatedOffset.generatedLine - 1),
            generatedColumn: mapping.generatedColumn +
              (section.generatedOffset.generatedLine === mapping.generatedLine
              ? section.generatedOffset.generatedColumn - 1
              : 0),
            originalLine: mapping.originalLine,
            originalColumn: mapping.originalColumn,
            name: name
          };

          this.__generatedMappings.push(adjustedMapping);
          if (typeof adjustedMapping.originalLine === 'number') {
            this.__originalMappings.push(adjustedMapping);
          }
        }
      }

      quickSort$1(this.__generatedMappings, util.compareByGeneratedPositionsDeflated);
      quickSort$1(this.__originalMappings, util.compareByOriginalPositions);
    };

  var IndexedSourceMapConsumer_1 = IndexedSourceMapConsumer;

  var sourceMapConsumer = {
  	SourceMapConsumer: SourceMapConsumer_1,
  	BasicSourceMapConsumer: BasicSourceMapConsumer_1,
  	IndexedSourceMapConsumer: IndexedSourceMapConsumer_1
  };

  /* -*- Mode: js; js-indent-level: 2; -*- */
  /*
   * Copyright 2011 Mozilla Foundation and contributors
   * Licensed under the New BSD license. See LICENSE or:
   * http://opensource.org/licenses/BSD-3-Clause
   */

  var SourceMapGenerator$1 = sourceMapGenerator.SourceMapGenerator;


  // Matches a Windows-style `\r\n` newline or a `\n` newline used by all other
  // operating systems these days (capturing the result).
  var REGEX_NEWLINE = /(\r?\n)/;

  // Newline character code for charCodeAt() comparisons
  var NEWLINE_CODE = 10;

  // Private symbol for identifying `SourceNode`s when multiple versions of
  // the source-map library are loaded. This MUST NOT CHANGE across
  // versions!
  var isSourceNode = "$$$isSourceNode$$$";

  /**
   * SourceNodes provide a way to abstract over interpolating/concatenating
   * snippets of generated JavaScript source code while maintaining the line and
   * column information associated with the original source code.
   *
   * @param aLine The original line number.
   * @param aColumn The original column number.
   * @param aSource The original source's filename.
   * @param aChunks Optional. An array of strings which are snippets of
   *        generated JS, or other SourceNodes.
   * @param aName The original identifier.
   */
  function SourceNode(aLine, aColumn, aSource, aChunks, aName) {
    this.children = [];
    this.sourceContents = {};
    this.line = aLine == null ? null : aLine;
    this.column = aColumn == null ? null : aColumn;
    this.source = aSource == null ? null : aSource;
    this.name = aName == null ? null : aName;
    this[isSourceNode] = true;
    if (aChunks != null) this.add(aChunks);
  }

  /**
   * Creates a SourceNode from generated code and a SourceMapConsumer.
   *
   * @param aGeneratedCode The generated code
   * @param aSourceMapConsumer The SourceMap for the generated code
   * @param aRelativePath Optional. The path that relative sources in the
   *        SourceMapConsumer should be relative to.
   */
  SourceNode.fromStringWithSourceMap =
    function SourceNode_fromStringWithSourceMap(aGeneratedCode, aSourceMapConsumer, aRelativePath) {
      // The SourceNode we want to fill with the generated code
      // and the SourceMap
      var node = new SourceNode();

      // All even indices of this array are one line of the generated code,
      // while all odd indices are the newlines between two adjacent lines
      // (since `REGEX_NEWLINE` captures its match).
      // Processed fragments are accessed by calling `shiftNextLine`.
      var remainingLines = aGeneratedCode.split(REGEX_NEWLINE);
      var remainingLinesIndex = 0;
      var shiftNextLine = function() {
        var lineContents = getNextLine();
        // The last line of a file might not have a newline.
        var newLine = getNextLine() || "";
        return lineContents + newLine;

        function getNextLine() {
          return remainingLinesIndex < remainingLines.length ?
              remainingLines[remainingLinesIndex++] : undefined;
        }
      };

      // We need to remember the position of "remainingLines"
      var lastGeneratedLine = 1, lastGeneratedColumn = 0;

      // The generate SourceNodes we need a code range.
      // To extract it current and last mapping is used.
      // Here we store the last mapping.
      var lastMapping = null;

      aSourceMapConsumer.eachMapping(function (mapping) {
        if (lastMapping !== null) {
          // We add the code from "lastMapping" to "mapping":
          // First check if there is a new line in between.
          if (lastGeneratedLine < mapping.generatedLine) {
            // Associate first line with "lastMapping"
            addMappingWithCode(lastMapping, shiftNextLine());
            lastGeneratedLine++;
            lastGeneratedColumn = 0;
            // The remaining code is added without mapping
          } else {
            // There is no new line in between.
            // Associate the code between "lastGeneratedColumn" and
            // "mapping.generatedColumn" with "lastMapping"
            var nextLine = remainingLines[remainingLinesIndex] || '';
            var code = nextLine.substr(0, mapping.generatedColumn -
                                          lastGeneratedColumn);
            remainingLines[remainingLinesIndex] = nextLine.substr(mapping.generatedColumn -
                                                lastGeneratedColumn);
            lastGeneratedColumn = mapping.generatedColumn;
            addMappingWithCode(lastMapping, code);
            // No more remaining code, continue
            lastMapping = mapping;
            return;
          }
        }
        // We add the generated code until the first mapping
        // to the SourceNode without any mapping.
        // Each line is added as separate string.
        while (lastGeneratedLine < mapping.generatedLine) {
          node.add(shiftNextLine());
          lastGeneratedLine++;
        }
        if (lastGeneratedColumn < mapping.generatedColumn) {
          var nextLine = remainingLines[remainingLinesIndex] || '';
          node.add(nextLine.substr(0, mapping.generatedColumn));
          remainingLines[remainingLinesIndex] = nextLine.substr(mapping.generatedColumn);
          lastGeneratedColumn = mapping.generatedColumn;
        }
        lastMapping = mapping;
      }, this);
      // We have processed all mappings.
      if (remainingLinesIndex < remainingLines.length) {
        if (lastMapping) {
          // Associate the remaining code in the current line with "lastMapping"
          addMappingWithCode(lastMapping, shiftNextLine());
        }
        // and add the remaining lines without any mapping
        node.add(remainingLines.splice(remainingLinesIndex).join(""));
      }

      // Copy sourcesContent into SourceNode
      aSourceMapConsumer.sources.forEach(function (sourceFile) {
        var content = aSourceMapConsumer.sourceContentFor(sourceFile);
        if (content != null) {
          if (aRelativePath != null) {
            sourceFile = util.join(aRelativePath, sourceFile);
          }
          node.setSourceContent(sourceFile, content);
        }
      });

      return node;

      function addMappingWithCode(mapping, code) {
        if (mapping === null || mapping.source === undefined) {
          node.add(code);
        } else {
          var source = aRelativePath
            ? util.join(aRelativePath, mapping.source)
            : mapping.source;
          node.add(new SourceNode(mapping.originalLine,
                                  mapping.originalColumn,
                                  source,
                                  code,
                                  mapping.name));
        }
      }
    };

  /**
   * Add a chunk of generated JS to this source node.
   *
   * @param aChunk A string snippet of generated JS code, another instance of
   *        SourceNode, or an array where each member is one of those things.
   */
  SourceNode.prototype.add = function SourceNode_add(aChunk) {
    if (Array.isArray(aChunk)) {
      aChunk.forEach(function (chunk) {
        this.add(chunk);
      }, this);
    }
    else if (aChunk[isSourceNode] || typeof aChunk === "string") {
      if (aChunk) {
        this.children.push(aChunk);
      }
    }
    else {
      throw new TypeError(
        "Expected a SourceNode, string, or an array of SourceNodes and strings. Got " + aChunk
      );
    }
    return this;
  };

  /**
   * Add a chunk of generated JS to the beginning of this source node.
   *
   * @param aChunk A string snippet of generated JS code, another instance of
   *        SourceNode, or an array where each member is one of those things.
   */
  SourceNode.prototype.prepend = function SourceNode_prepend(aChunk) {
    if (Array.isArray(aChunk)) {
      for (var i = aChunk.length-1; i >= 0; i--) {
        this.prepend(aChunk[i]);
      }
    }
    else if (aChunk[isSourceNode] || typeof aChunk === "string") {
      this.children.unshift(aChunk);
    }
    else {
      throw new TypeError(
        "Expected a SourceNode, string, or an array of SourceNodes and strings. Got " + aChunk
      );
    }
    return this;
  };

  /**
   * Walk over the tree of JS snippets in this node and its children. The
   * walking function is called once for each snippet of JS and is passed that
   * snippet and the its original associated source's line/column location.
   *
   * @param aFn The traversal function.
   */
  SourceNode.prototype.walk = function SourceNode_walk(aFn) {
    var chunk;
    for (var i = 0, len = this.children.length; i < len; i++) {
      chunk = this.children[i];
      if (chunk[isSourceNode]) {
        chunk.walk(aFn);
      }
      else {
        if (chunk !== '') {
          aFn(chunk, { source: this.source,
                       line: this.line,
                       column: this.column,
                       name: this.name });
        }
      }
    }
  };

  /**
   * Like `String.prototype.join` except for SourceNodes. Inserts `aStr` between
   * each of `this.children`.
   *
   * @param aSep The separator.
   */
  SourceNode.prototype.join = function SourceNode_join(aSep) {
    var newChildren;
    var i;
    var len = this.children.length;
    if (len > 0) {
      newChildren = [];
      for (i = 0; i < len-1; i++) {
        newChildren.push(this.children[i]);
        newChildren.push(aSep);
      }
      newChildren.push(this.children[i]);
      this.children = newChildren;
    }
    return this;
  };

  /**
   * Call String.prototype.replace on the very right-most source snippet. Useful
   * for trimming whitespace from the end of a source node, etc.
   *
   * @param aPattern The pattern to replace.
   * @param aReplacement The thing to replace the pattern with.
   */
  SourceNode.prototype.replaceRight = function SourceNode_replaceRight(aPattern, aReplacement) {
    var lastChild = this.children[this.children.length - 1];
    if (lastChild[isSourceNode]) {
      lastChild.replaceRight(aPattern, aReplacement);
    }
    else if (typeof lastChild === 'string') {
      this.children[this.children.length - 1] = lastChild.replace(aPattern, aReplacement);
    }
    else {
      this.children.push(''.replace(aPattern, aReplacement));
    }
    return this;
  };

  /**
   * Set the source content for a source file. This will be added to the SourceMapGenerator
   * in the sourcesContent field.
   *
   * @param aSourceFile The filename of the source file
   * @param aSourceContent The content of the source file
   */
  SourceNode.prototype.setSourceContent =
    function SourceNode_setSourceContent(aSourceFile, aSourceContent) {
      this.sourceContents[util.toSetString(aSourceFile)] = aSourceContent;
    };

  /**
   * Walk over the tree of SourceNodes. The walking function is called for each
   * source file content and is passed the filename and source content.
   *
   * @param aFn The traversal function.
   */
  SourceNode.prototype.walkSourceContents =
    function SourceNode_walkSourceContents(aFn) {
      for (var i = 0, len = this.children.length; i < len; i++) {
        if (this.children[i][isSourceNode]) {
          this.children[i].walkSourceContents(aFn);
        }
      }

      var sources = Object.keys(this.sourceContents);
      for (var i = 0, len = sources.length; i < len; i++) {
        aFn(util.fromSetString(sources[i]), this.sourceContents[sources[i]]);
      }
    };

  /**
   * Return the string representation of this source node. Walks over the tree
   * and concatenates all the various snippets together to one string.
   */
  SourceNode.prototype.toString = function SourceNode_toString() {
    var str = "";
    this.walk(function (chunk) {
      str += chunk;
    });
    return str;
  };

  /**
   * Returns the string representation of this source node along with a source
   * map.
   */
  SourceNode.prototype.toStringWithSourceMap = function SourceNode_toStringWithSourceMap(aArgs) {
    var generated = {
      code: "",
      line: 1,
      column: 0
    };
    var map = new SourceMapGenerator$1(aArgs);
    var sourceMappingActive = false;
    var lastOriginalSource = null;
    var lastOriginalLine = null;
    var lastOriginalColumn = null;
    var lastOriginalName = null;
    this.walk(function (chunk, original) {
      generated.code += chunk;
      if (original.source !== null
          && original.line !== null
          && original.column !== null) {
        if(lastOriginalSource !== original.source
           || lastOriginalLine !== original.line
           || lastOriginalColumn !== original.column
           || lastOriginalName !== original.name) {
          map.addMapping({
            source: original.source,
            original: {
              line: original.line,
              column: original.column
            },
            generated: {
              line: generated.line,
              column: generated.column
            },
            name: original.name
          });
        }
        lastOriginalSource = original.source;
        lastOriginalLine = original.line;
        lastOriginalColumn = original.column;
        lastOriginalName = original.name;
        sourceMappingActive = true;
      } else if (sourceMappingActive) {
        map.addMapping({
          generated: {
            line: generated.line,
            column: generated.column
          }
        });
        lastOriginalSource = null;
        sourceMappingActive = false;
      }
      for (var idx = 0, length = chunk.length; idx < length; idx++) {
        if (chunk.charCodeAt(idx) === NEWLINE_CODE) {
          generated.line++;
          generated.column = 0;
          // Mappings end at eol
          if (idx + 1 === length) {
            lastOriginalSource = null;
            sourceMappingActive = false;
          } else if (sourceMappingActive) {
            map.addMapping({
              source: original.source,
              original: {
                line: original.line,
                column: original.column
              },
              generated: {
                line: generated.line,
                column: generated.column
              },
              name: original.name
            });
          }
        } else {
          generated.column++;
        }
      }
    });
    this.walkSourceContents(function (sourceFile, sourceContent) {
      map.setSourceContent(sourceFile, sourceContent);
    });

    return { code: generated.code, map: map };
  };

  var SourceNode_1 = SourceNode;

  var sourceNode = {
  	SourceNode: SourceNode_1
  };

  /*
   * Copyright 2009-2011 Mozilla Foundation and contributors
   * Licensed under the New BSD license. See LICENSE.txt or:
   * http://opensource.org/licenses/BSD-3-Clause
   */
  var SourceMapGenerator$2 = sourceMapGenerator.SourceMapGenerator;
  var SourceMapConsumer$1 = sourceMapConsumer.SourceMapConsumer;
  var SourceNode$1 = sourceNode.SourceNode;

  var sourceMap = {
  	SourceMapGenerator: SourceMapGenerator$2,
  	SourceMapConsumer: SourceMapConsumer$1,
  	SourceNode: SourceNode$1
  };

  var require$$1 = getCjsExportFromNamespace(_empty_module$1);

  var util_1$1 = createCommonjsModule(function (module, exports) {
  var getFieldValue = types$1.getFieldValue;
  var n = types$1.namedTypes;

  var SourceMapConsumer = sourceMap.SourceMapConsumer;
  var SourceMapGenerator = sourceMap.SourceMapGenerator;
  var hasOwn = Object.prototype.hasOwnProperty;
  var util = exports;

  function getOption(options, key, defaultValue) {
    if (options && hasOwn.call(options, key)) {
      return options[key];
    }
    return defaultValue;
  }
  util.getOption = getOption;

  function getUnionOfKeys() {
    var result = {};
    var argc = arguments.length;
    for (var i = 0; i < argc; ++i) {
      var keys = Object.keys(arguments[i]);
      var keyCount = keys.length;
      for (var j = 0; j < keyCount; ++j) {
        result[keys[j]] = true;
      }
    }
    return result;
  }
  util.getUnionOfKeys = getUnionOfKeys;

  function comparePos(pos1, pos2) {
    return (pos1.line - pos2.line) || (pos1.column - pos2.column);
  }
  util.comparePos = comparePos;

  function copyPos(pos) {
    return {
      line: pos.line,
      column: pos.column
    };
  }
  util.copyPos = copyPos;

  util.composeSourceMaps = function(formerMap, latterMap) {
    if (formerMap) {
      if (!latterMap) {
        return formerMap;
      }
    } else {
      return latterMap || null;
    }

    var smcFormer = new SourceMapConsumer(formerMap);
    var smcLatter = new SourceMapConsumer(latterMap);
    var smg = new SourceMapGenerator({
      file: latterMap.file,
      sourceRoot: latterMap.sourceRoot
    });

    var sourcesToContents = {};

    smcLatter.eachMapping(function(mapping) {
      var origPos = smcFormer.originalPositionFor({
        line: mapping.originalLine,
        column: mapping.originalColumn
      });

      var sourceName = origPos.source;
      if (sourceName === null) {
        return;
      }

      smg.addMapping({
        source: sourceName,
        original: copyPos(origPos),
        generated: {
          line: mapping.generatedLine,
          column: mapping.generatedColumn
        },
        name: mapping.name
      });

      var sourceContent = smcFormer.sourceContentFor(sourceName);
      if (sourceContent && !hasOwn.call(sourcesToContents, sourceName)) {
        sourcesToContents[sourceName] = sourceContent;
        smg.setSourceContent(sourceName, sourceContent);
      }
    });

    return smg.toJSON();
  };

  util.getTrueLoc = function(node, lines) {
    // It's possible that node is newly-created (not parsed by Esprima),
    // in which case it probably won't have a .loc property (or an
    // .original property for that matter). That's fine; we'll just
    // pretty-print it as usual.
    if (!node.loc) {
      return null;
    }

    var result = {
      start: node.loc.start,
      end: node.loc.end
    };

    function include(node) {
      expandLoc(result, node.loc);
    }

    // If the node is an export declaration and its .declaration has any
    // decorators, their locations might contribute to the true start/end
    // positions of the export declaration node.
    if (node.declaration &&
        node.declaration.decorators &&
        util.isExportDeclaration(node)) {
      node.declaration.decorators.forEach(include);
    }

    if (comparePos(result.start, result.end) < 0) {
      // Trim leading whitespace.
      result.start = copyPos(result.start);
      lines.skipSpaces(result.start, false, true);

      if (comparePos(result.start, result.end) < 0) {
        // Trim trailing whitespace, if the end location is not already the
        // same as the start location.
        result.end = copyPos(result.end);
        lines.skipSpaces(result.end, true, true);
      }
    }

    // If the node has any comments, their locations might contribute to
    // the true start/end positions of the node.
    if (node.comments) {
      node.comments.forEach(include);
    }

    return result;
  };

  function expandLoc(parentLoc, childLoc) {
    if (parentLoc && childLoc) {
      if (comparePos(childLoc.start, parentLoc.start) < 0) {
        parentLoc.start = childLoc.start;
      }

      if (comparePos(parentLoc.end, childLoc.end) < 0) {
        parentLoc.end = childLoc.end;
      }
    }
  }

  util.fixFaultyLocations = function(node, lines) {
    var loc = node.loc;
    if (loc) {
      if (loc.start.line < 1) {
        loc.start.line = 1;
      }

      if (loc.end.line < 1) {
        loc.end.line = 1;
      }
    }

    if (node.type === "File") {
      // Babylon returns File nodes whose .loc.{start,end} do not include
      // leading or trailing whitespace.
      loc.start = lines.firstPos();
      loc.end = lines.lastPos();
    }

    fixForLoopHead(node, lines);
    fixTemplateLiteral(node, lines);

    if (loc && node.decorators) {
      // Expand the .loc of the node responsible for printing the decorators
      // (here, the decorated node) so that it includes node.decorators.
      node.decorators.forEach(function (decorator) {
        expandLoc(loc, decorator.loc);
      });

    } else if (node.declaration && util.isExportDeclaration(node)) {
      // Nullify .loc information for the child declaration so that we never
      // try to reprint it without also reprinting the export declaration.
      node.declaration.loc = null;

      // Expand the .loc of the node responsible for printing the decorators
      // (here, the export declaration) so that it includes node.decorators.
      var decorators = node.declaration.decorators;
      if (decorators) {
        decorators.forEach(function (decorator) {
          expandLoc(loc, decorator.loc);
        });
      }

    } else if ((n.MethodDefinition && n.MethodDefinition.check(node)) ||
               (n.Property.check(node) && (node.method || node.shorthand))) {
      // If the node is a MethodDefinition or a .method or .shorthand
      // Property, then the location information stored in
      // node.value.loc is very likely untrustworthy (just the {body}
      // part of a method, or nothing in the case of shorthand
      // properties), so we null out that information to prevent
      // accidental reuse of bogus source code during reprinting.
      node.value.loc = null;

      if (n.FunctionExpression.check(node.value)) {
        // FunctionExpression method values should be anonymous,
        // because their .id fields are ignored anyway.
        node.value.id = null;
      }

    } else if (node.type === "ObjectTypeProperty") {
      var loc = node.loc;
      var end = loc && loc.end;
      if (end) {
        end = copyPos(end);
        if (lines.prevPos(end) &&
            lines.charAt(end) === ",") {
          // Some parsers accidentally include trailing commas in the
          // .loc.end information for ObjectTypeProperty nodes.
          if ((end = lines.skipSpaces(end, true, true))) {
            loc.end = end;
          }
        }
      }
    }
  };

  function fixForLoopHead(node, lines) {
    if (node.type !== "ForStatement") {
      return;
    }

    function fix(child) {
      var loc = child && child.loc;
      var start = loc && loc.start;
      var end = loc && copyPos(loc.end);

      while (start && end && comparePos(start, end) < 0) {
        lines.prevPos(end);
        if (lines.charAt(end) === ";") {
          // Update child.loc.end to *exclude* the ';' character.
          loc.end.line = end.line;
          loc.end.column = end.column;
        } else {
          break;
        }
      }
    }

    fix(node.init);
    fix(node.test);
    fix(node.update);
  }

  function fixTemplateLiteral(node, lines) {
    if (node.type !== "TemplateLiteral") {
      return;
    }

    if (node.quasis.length === 0) {
      // If there are no quasi elements, then there is nothing to fix.
      return;
    }

    // First we need to exclude the opening ` from the .loc of the first
    // quasi element, in case the parser accidentally decided to include it.
    var afterLeftBackTickPos = copyPos(node.loc.start);
    var firstQuasi = node.quasis[0];
    if (comparePos(firstQuasi.loc.start, afterLeftBackTickPos) < 0) {
      firstQuasi.loc.start = afterLeftBackTickPos;
    }

    // Next we need to exclude the closing ` from the .loc of the last quasi
    // element, in case the parser accidentally decided to include it.
    var rightBackTickPos = copyPos(node.loc.end);
    var lastQuasi = node.quasis[node.quasis.length - 1];
    if (comparePos(rightBackTickPos, lastQuasi.loc.end) < 0) {
      lastQuasi.loc.end = rightBackTickPos;
    }

    // Now we need to exclude ${ and } characters from the .loc's of all
    // quasi elements, since some parsers accidentally include them.
    node.expressions.forEach(function (expr, i) {
      // Rewind from expr.loc.start over any whitespace and the ${ that
      // precedes the expression. The position of the $ should be the same
      // as the .loc.end of the preceding quasi element, but some parsers
      // accidentally include the ${ in the .loc of the quasi element.
      var dollarCurlyPos = lines.skipSpaces(expr.loc.start, true, false);
      if (lines.prevPos(dollarCurlyPos) &&
          lines.charAt(dollarCurlyPos) === "{" &&
          lines.prevPos(dollarCurlyPos) &&
          lines.charAt(dollarCurlyPos) === "$") {
        var quasiBefore = node.quasis[i];
        if (comparePos(dollarCurlyPos, quasiBefore.loc.end) < 0) {
          quasiBefore.loc.end = dollarCurlyPos;
        }
      }

      // Likewise, some parsers accidentally include the } that follows
      // the expression in the .loc of the following quasi element.
      var rightCurlyPos = lines.skipSpaces(expr.loc.end, false, false);
      if (lines.charAt(rightCurlyPos) === "}") {
        // Now rightCurlyPos is technically the position just after the }.
        var quasiAfter = node.quasis[i + 1];
        if (comparePos(quasiAfter.loc.start, rightCurlyPos) < 0) {
          quasiAfter.loc.start = rightCurlyPos;
        }
      }
    });
  }

  util.isExportDeclaration = function (node) {
    if (node) switch (node.type) {
    case "ExportDeclaration":
    case "ExportDefaultDeclaration":
    case "ExportDefaultSpecifier":
    case "DeclareExportDeclaration":
    case "ExportNamedDeclaration":
    case "ExportAllDeclaration":
      return true;
    }

    return false;
  };

  util.getParentExportDeclaration = function (path) {
    var parentNode = path.getParentNode();
    if (path.getName() === "declaration" &&
        util.isExportDeclaration(parentNode)) {
      return parentNode;
    }

    return null;
  };

  util.isTrailingCommaEnabled = function(options, context) {
    var trailingComma = options.trailingComma;
    if (typeof trailingComma === "object") {
      return !!trailingComma[context];
    }
    return !!trailingComma;
  };
  });

  /* -*- Mode: js; js-indent-level: 2; -*- */
  /*
   * Copyright 2011 Mozilla Foundation and contributors
   * Licensed under the New BSD license. See LICENSE or:
   * http://opensource.org/licenses/BSD-3-Clause
   */

  const intToCharMap$1 = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/".split("");

  /**
   * Encode an integer in the range of 0 to 63 to a single base 64 digit.
   */
  var encode$2 = function(number) {
    if (0 <= number && number < intToCharMap$1.length) {
      return intToCharMap$1[number];
    }
    throw new TypeError("Must be between 0 and 63: " + number);
  };

  var base64$1 = {
  	encode: encode$2
  };

  /* -*- Mode: js; js-indent-level: 2; -*- */
  /*
   * Copyright 2011 Mozilla Foundation and contributors
   * Licensed under the New BSD license. See LICENSE or:
   * http://opensource.org/licenses/BSD-3-Clause
   *
   * Based on the Base 64 VLQ implementation in Closure Compiler:
   * https://code.google.com/p/closure-compiler/source/browse/trunk/src/com/google/debugging/sourcemap/Base64VLQ.java
   *
   * Copyright 2011 The Closure Compiler Authors. All rights reserved.
   * Redistribution and use in source and binary forms, with or without
   * modification, are permitted provided that the following conditions are
   * met:
   *
   *  * Redistributions of source code must retain the above copyright
   *    notice, this list of conditions and the following disclaimer.
   *  * Redistributions in binary form must reproduce the above
   *    copyright notice, this list of conditions and the following
   *    disclaimer in the documentation and/or other materials provided
   *    with the distribution.
   *  * Neither the name of Google Inc. nor the names of its
   *    contributors may be used to endorse or promote products derived
   *    from this software without specific prior written permission.
   *
   * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
   */



  // A single base 64 digit can contain 6 bits of data. For the base 64 variable
  // length quantities we use in the source map spec, the first bit is the sign,
  // the next four bits are the actual value, and the 6th bit is the
  // continuation bit. The continuation bit tells us whether there are more
  // digits in this value following this digit.
  //
  //   Continuation
  //   |    Sign
  //   |    |
  //   V    V
  //   101011

  const VLQ_BASE_SHIFT$1 = 5;

  // binary: 100000
  const VLQ_BASE$1 = 1 << VLQ_BASE_SHIFT$1;

  // binary: 011111
  const VLQ_BASE_MASK$1 = VLQ_BASE$1 - 1;

  // binary: 100000
  const VLQ_CONTINUATION_BIT$1 = VLQ_BASE$1;

  /**
   * Converts from a two-complement value to a value where the sign bit is
   * placed in the least significant bit.  For example, as decimals:
   *   1 becomes 2 (10 binary), -1 becomes 3 (11 binary)
   *   2 becomes 4 (100 binary), -2 becomes 5 (101 binary)
   */
  function toVLQSigned$1(aValue) {
    return aValue < 0
      ? ((-aValue) << 1) + 1
      : (aValue << 1) + 0;
  }

  /**
   * Returns the base 64 VLQ encoded value.
   */
  var encode$3 = function base64VLQ_encode(aValue) {
    let encoded = "";
    let digit;

    let vlq = toVLQSigned$1(aValue);

    do {
      digit = vlq & VLQ_BASE_MASK$1;
      vlq >>>= VLQ_BASE_SHIFT$1;
      if (vlq > 0) {
        // There are still more digits in this value, so we must make sure the
        // continuation bit is marked.
        digit |= VLQ_CONTINUATION_BIT$1;
      }
      encoded += base64$1.encode(digit);
    } while (vlq > 0);

    return encoded;
  };

  var base64Vlq$1 = {
  	encode: encode$3
  };

  var util$1 = createCommonjsModule(function (module, exports) {
  /* -*- Mode: js; js-indent-level: 2; -*- */
  /*
   * Copyright 2011 Mozilla Foundation and contributors
   * Licensed under the New BSD license. See LICENSE or:
   * http://opensource.org/licenses/BSD-3-Clause
   */

  /**
   * This is a helper function for getting values from parameter/options
   * objects.
   *
   * @param args The object we are extracting values from
   * @param name The name of the property we are getting.
   * @param defaultValue An optional value to return if the property is missing
   * from the object. If this is not specified and the property is missing, an
   * error will be thrown.
   */
  function getArg(aArgs, aName, aDefaultValue) {
    if (aName in aArgs) {
      return aArgs[aName];
    } else if (arguments.length === 3) {
      return aDefaultValue;
    }
      throw new Error('"' + aName + '" is a required argument.');

  }
  exports.getArg = getArg;

  const urlRegexp = /^(?:([\w+\-.]+):)?\/\/(?:(\w+:\w+)@)?([\w.-]*)(?::(\d+))?(.*)$/;
  const dataUrlRegexp = /^data:.+\,.+$/;

  function urlParse(aUrl) {
    const match = aUrl.match(urlRegexp);
    if (!match) {
      return null;
    }
    return {
      scheme: match[1],
      auth: match[2],
      host: match[3],
      port: match[4],
      path: match[5]
    };
  }
  exports.urlParse = urlParse;

  function urlGenerate(aParsedUrl) {
    let url = "";
    if (aParsedUrl.scheme) {
      url += aParsedUrl.scheme + ":";
    }
    url += "//";
    if (aParsedUrl.auth) {
      url += aParsedUrl.auth + "@";
    }
    if (aParsedUrl.host) {
      url += aParsedUrl.host;
    }
    if (aParsedUrl.port) {
      url += ":" + aParsedUrl.port;
    }
    if (aParsedUrl.path) {
      url += aParsedUrl.path;
    }
    return url;
  }
  exports.urlGenerate = urlGenerate;

  const MAX_CACHED_INPUTS = 32;

  /**
   * Takes some function `f(input) -> result` and returns a memoized version of
   * `f`.
   *
   * We keep at most `MAX_CACHED_INPUTS` memoized results of `f` alive. The
   * memoization is a dumb-simple, linear least-recently-used cache.
   */
  function lruMemoize(f) {
    const cache = [];

    return function(input) {
      for (let i = 0; i < cache.length; i++) {
        if (cache[i].input === input) {
          const temp = cache[0];
          cache[0] = cache[i];
          cache[i] = temp;
          return cache[0].result;
        }
      }

      const result = f(input);

      cache.unshift({
        input,
        result,
      });

      if (cache.length > MAX_CACHED_INPUTS) {
        cache.pop();
      }

      return result;
    };
  }

  /**
   * Normalizes a path, or the path portion of a URL:
   *
   * - Replaces consecutive slashes with one slash.
   * - Removes unnecessary '.' parts.
   * - Removes unnecessary '<dir>/..' parts.
   *
   * Based on code in the Node.js 'path' core module.
   *
   * @param aPath The path or url to normalize.
   */
  const normalize = lruMemoize(function normalize(aPath) {
    let path = aPath;
    const url = urlParse(aPath);
    if (url) {
      if (!url.path) {
        return aPath;
      }
      path = url.path;
    }
    const isAbsolute = exports.isAbsolute(path);

    // Split the path into parts between `/` characters. This is much faster than
    // using `.split(/\/+/g)`.
    const parts = [];
    let start = 0;
    let i = 0;
    while (true) {
      start = i;
      i = path.indexOf("/", start);
      if (i === -1) {
        parts.push(path.slice(start));
        break;
      } else {
        parts.push(path.slice(start, i));
        while (i < path.length && path[i] === "/") {
          i++;
        }
      }
    }

    let up = 0;
    for (i = parts.length - 1; i >= 0; i--) {
      const part = parts[i];
      if (part === ".") {
        parts.splice(i, 1);
      } else if (part === "..") {
        up++;
      } else if (up > 0) {
        if (part === "") {
          // The first part is blank if the path is absolute. Trying to go
          // above the root is a no-op. Therefore we can remove all '..' parts
          // directly after the root.
          parts.splice(i + 1, up);
          up = 0;
        } else {
          parts.splice(i, 2);
          up--;
        }
      }
    }
    path = parts.join("/");

    if (path === "") {
      path = isAbsolute ? "/" : ".";
    }

    if (url) {
      url.path = path;
      return urlGenerate(url);
    }
    return path;
  });
  exports.normalize = normalize;

  /**
   * Joins two paths/URLs.
   *
   * @param aRoot The root path or URL.
   * @param aPath The path or URL to be joined with the root.
   *
   * - If aPath is a URL or a data URI, aPath is returned, unless aPath is a
   *   scheme-relative URL: Then the scheme of aRoot, if any, is prepended
   *   first.
   * - Otherwise aPath is a path. If aRoot is a URL, then its path portion
   *   is updated with the result and aRoot is returned. Otherwise the result
   *   is returned.
   *   - If aPath is absolute, the result is aPath.
   *   - Otherwise the two paths are joined with a slash.
   * - Joining for example 'http://' and 'www.example.com' is also supported.
   */
  function join(aRoot, aPath) {
    if (aRoot === "") {
      aRoot = ".";
    }
    if (aPath === "") {
      aPath = ".";
    }
    const aPathUrl = urlParse(aPath);
    const aRootUrl = urlParse(aRoot);
    if (aRootUrl) {
      aRoot = aRootUrl.path || "/";
    }

    // `join(foo, '//www.example.org')`
    if (aPathUrl && !aPathUrl.scheme) {
      if (aRootUrl) {
        aPathUrl.scheme = aRootUrl.scheme;
      }
      return urlGenerate(aPathUrl);
    }

    if (aPathUrl || aPath.match(dataUrlRegexp)) {
      return aPath;
    }

    // `join('http://', 'www.example.com')`
    if (aRootUrl && !aRootUrl.host && !aRootUrl.path) {
      aRootUrl.host = aPath;
      return urlGenerate(aRootUrl);
    }

    const joined = aPath.charAt(0) === "/"
      ? aPath
      : normalize(aRoot.replace(/\/+$/, "") + "/" + aPath);

    if (aRootUrl) {
      aRootUrl.path = joined;
      return urlGenerate(aRootUrl);
    }
    return joined;
  }
  exports.join = join;

  exports.isAbsolute = function(aPath) {
    return aPath.charAt(0) === "/" || urlRegexp.test(aPath);
  };

  /**
   * Make a path relative to a URL or another path.
   *
   * @param aRoot The root path or URL.
   * @param aPath The path or URL to be made relative to aRoot.
   */
  function relative(aRoot, aPath) {
    if (aRoot === "") {
      aRoot = ".";
    }

    aRoot = aRoot.replace(/\/$/, "");

    // It is possible for the path to be above the root. In this case, simply
    // checking whether the root is a prefix of the path won't work. Instead, we
    // need to remove components from the root one by one, until either we find
    // a prefix that fits, or we run out of components to remove.
    let level = 0;
    while (aPath.indexOf(aRoot + "/") !== 0) {
      const index = aRoot.lastIndexOf("/");
      if (index < 0) {
        return aPath;
      }

      // If the only part of the root that is left is the scheme (i.e. http://,
      // file:///, etc.), one or more slashes (/), or simply nothing at all, we
      // have exhausted all components, so the path is not relative to the root.
      aRoot = aRoot.slice(0, index);
      if (aRoot.match(/^([^\/]+:\/)?\/*$/)) {
        return aPath;
      }

      ++level;
    }

    // Make sure we add a "../" for each component we removed from the root.
    return Array(level + 1).join("../") + aPath.substr(aRoot.length + 1);
  }
  exports.relative = relative;

  const supportsNullProto = (function() {
    const obj = Object.create(null);
    return !("__proto__" in obj);
  }());

  function identity(s) {
    return s;
  }

  /**
   * Because behavior goes wacky when you set `__proto__` on objects, we
   * have to prefix all the strings in our set with an arbitrary character.
   *
   * See https://github.com/mozilla/source-map/pull/31 and
   * https://github.com/mozilla/source-map/issues/30
   *
   * @param String aStr
   */
  function toSetString(aStr) {
    if (isProtoString(aStr)) {
      return "$" + aStr;
    }

    return aStr;
  }
  exports.toSetString = supportsNullProto ? identity : toSetString;

  function fromSetString(aStr) {
    if (isProtoString(aStr)) {
      return aStr.slice(1);
    }

    return aStr;
  }
  exports.fromSetString = supportsNullProto ? identity : fromSetString;

  function isProtoString(s) {
    if (!s) {
      return false;
    }

    const length = s.length;

    if (length < 9 /* "__proto__".length */) {
      return false;
    }

    /* eslint-disable no-multi-spaces */
    if (s.charCodeAt(length - 1) !== 95  /* '_' */ ||
        s.charCodeAt(length - 2) !== 95  /* '_' */ ||
        s.charCodeAt(length - 3) !== 111 /* 'o' */ ||
        s.charCodeAt(length - 4) !== 116 /* 't' */ ||
        s.charCodeAt(length - 5) !== 111 /* 'o' */ ||
        s.charCodeAt(length - 6) !== 114 /* 'r' */ ||
        s.charCodeAt(length - 7) !== 112 /* 'p' */ ||
        s.charCodeAt(length - 8) !== 95  /* '_' */ ||
        s.charCodeAt(length - 9) !== 95  /* '_' */) {
      return false;
    }
    /* eslint-enable no-multi-spaces */

    for (let i = length - 10; i >= 0; i--) {
      if (s.charCodeAt(i) !== 36 /* '$' */) {
        return false;
      }
    }

    return true;
  }

  /**
   * Comparator between two mappings where the original positions are compared.
   *
   * Optionally pass in `true` as `onlyCompareGenerated` to consider two
   * mappings with the same original source/line/column, but different generated
   * line and column the same. Useful when searching for a mapping with a
   * stubbed out mapping.
   */
  function compareByOriginalPositions(mappingA, mappingB, onlyCompareOriginal) {
    let cmp = strcmp(mappingA.source, mappingB.source);
    if (cmp !== 0) {
      return cmp;
    }

    cmp = mappingA.originalLine - mappingB.originalLine;
    if (cmp !== 0) {
      return cmp;
    }

    cmp = mappingA.originalColumn - mappingB.originalColumn;
    if (cmp !== 0 || onlyCompareOriginal) {
      return cmp;
    }

    cmp = mappingA.generatedColumn - mappingB.generatedColumn;
    if (cmp !== 0) {
      return cmp;
    }

    cmp = mappingA.generatedLine - mappingB.generatedLine;
    if (cmp !== 0) {
      return cmp;
    }

    return strcmp(mappingA.name, mappingB.name);
  }
  exports.compareByOriginalPositions = compareByOriginalPositions;

  /**
   * Comparator between two mappings with deflated source and name indices where
   * the generated positions are compared.
   *
   * Optionally pass in `true` as `onlyCompareGenerated` to consider two
   * mappings with the same generated line and column, but different
   * source/name/original line and column the same. Useful when searching for a
   * mapping with a stubbed out mapping.
   */
  function compareByGeneratedPositionsDeflated(mappingA, mappingB, onlyCompareGenerated) {
    let cmp = mappingA.generatedLine - mappingB.generatedLine;
    if (cmp !== 0) {
      return cmp;
    }

    cmp = mappingA.generatedColumn - mappingB.generatedColumn;
    if (cmp !== 0 || onlyCompareGenerated) {
      return cmp;
    }

    cmp = strcmp(mappingA.source, mappingB.source);
    if (cmp !== 0) {
      return cmp;
    }

    cmp = mappingA.originalLine - mappingB.originalLine;
    if (cmp !== 0) {
      return cmp;
    }

    cmp = mappingA.originalColumn - mappingB.originalColumn;
    if (cmp !== 0) {
      return cmp;
    }

    return strcmp(mappingA.name, mappingB.name);
  }
  exports.compareByGeneratedPositionsDeflated = compareByGeneratedPositionsDeflated;

  function strcmp(aStr1, aStr2) {
    if (aStr1 === aStr2) {
      return 0;
    }

    if (aStr1 === null) {
      return 1; // aStr2 !== null
    }

    if (aStr2 === null) {
      return -1; // aStr1 !== null
    }

    if (aStr1 > aStr2) {
      return 1;
    }

    return -1;
  }

  /**
   * Comparator between two mappings with inflated source and name strings where
   * the generated positions are compared.
   */
  function compareByGeneratedPositionsInflated(mappingA, mappingB) {
    let cmp = mappingA.generatedLine - mappingB.generatedLine;
    if (cmp !== 0) {
      return cmp;
    }

    cmp = mappingA.generatedColumn - mappingB.generatedColumn;
    if (cmp !== 0) {
      return cmp;
    }

    cmp = strcmp(mappingA.source, mappingB.source);
    if (cmp !== 0) {
      return cmp;
    }

    cmp = mappingA.originalLine - mappingB.originalLine;
    if (cmp !== 0) {
      return cmp;
    }

    cmp = mappingA.originalColumn - mappingB.originalColumn;
    if (cmp !== 0) {
      return cmp;
    }

    return strcmp(mappingA.name, mappingB.name);
  }
  exports.compareByGeneratedPositionsInflated = compareByGeneratedPositionsInflated;

  /**
   * Strip any JSON XSSI avoidance prefix from the string (as documented
   * in the source maps specification), and then parse the string as
   * JSON.
   */
  function parseSourceMapInput(str) {
    return JSON.parse(str.replace(/^\)]}'[^\n]*\n/, ""));
  }
  exports.parseSourceMapInput = parseSourceMapInput;

  /**
   * Compute the URL of a source given the the source root, the source's
   * URL, and the source map's URL.
   */
  function computeSourceURL(sourceRoot, sourceURL, sourceMapURL) {
    sourceURL = sourceURL || "";

    if (sourceRoot) {
      // This follows what Chrome does.
      if (sourceRoot[sourceRoot.length - 1] !== "/" && sourceURL[0] !== "/") {
        sourceRoot += "/";
      }
      // The spec says:
      //   Line 4: An optional source root, useful for relocating source
      //   files on a server or removing repeated values in the
      //   “sources” entry.  This value is prepended to the individual
      //   entries in the “source” field.
      sourceURL = sourceRoot + sourceURL;
    }

    // Historically, SourceMapConsumer did not take the sourceMapURL as
    // a parameter.  This mode is still somewhat supported, which is why
    // this code block is conditional.  However, it's preferable to pass
    // the source map URL to SourceMapConsumer, so that this function
    // can implement the source URL resolution algorithm as outlined in
    // the spec.  This block is basically the equivalent of:
    //    new URL(sourceURL, sourceMapURL).toString()
    // ... except it avoids using URL, which wasn't available in the
    // older releases of node still supported by this library.
    //
    // The spec says:
    //   If the sources are not absolute URLs after prepending of the
    //   “sourceRoot”, the sources are resolved relative to the
    //   SourceMap (like resolving script src in a html document).
    if (sourceMapURL) {
      const parsed = urlParse(sourceMapURL);
      if (!parsed) {
        throw new Error("sourceMapURL could not be parsed");
      }
      if (parsed.path) {
        // Strip the last path component, but keep the "/".
        const index = parsed.path.lastIndexOf("/");
        if (index >= 0) {
          parsed.path = parsed.path.substring(0, index + 1);
        }
      }
      sourceURL = join(urlGenerate(parsed), sourceURL);
    }

    return normalize(sourceURL);
  }
  exports.computeSourceURL = computeSourceURL;
  });
  var util_1$2 = util$1.getArg;
  var util_2$1 = util$1.urlParse;
  var util_3$1 = util$1.urlGenerate;
  var util_4$1 = util$1.normalize;
  var util_5$1 = util$1.join;
  var util_6$1 = util$1.isAbsolute;
  var util_7$1 = util$1.relative;
  var util_8$1 = util$1.toSetString;
  var util_9$1 = util$1.fromSetString;
  var util_10$1 = util$1.compareByOriginalPositions;
  var util_11$1 = util$1.compareByGeneratedPositionsDeflated;
  var util_12$1 = util$1.compareByGeneratedPositionsInflated;
  var util_13$1 = util$1.parseSourceMapInput;
  var util_14$1 = util$1.computeSourceURL;

  /* -*- Mode: js; js-indent-level: 2; -*- */
  /*
   * Copyright 2011 Mozilla Foundation and contributors
   * Licensed under the New BSD license. See LICENSE or:
   * http://opensource.org/licenses/BSD-3-Clause
   */

  /**
   * A data structure which is a combination of an array and a set. Adding a new
   * member is O(1), testing for membership is O(1), and finding the index of an
   * element is O(1). Removing elements from the set is not supported. Only
   * strings are supported for membership.
   */
  class ArraySet$3 {
    constructor() {
      this._array = [];
      this._set = new Map();
    }

    /**
     * Static method for creating ArraySet instances from an existing array.
     */
    static fromArray(aArray, aAllowDuplicates) {
      const set = new ArraySet$3();
      for (let i = 0, len = aArray.length; i < len; i++) {
        set.add(aArray[i], aAllowDuplicates);
      }
      return set;
    }

    /**
     * Return how many unique items are in this ArraySet. If duplicates have been
     * added, than those do not count towards the size.
     *
     * @returns Number
     */
    size() {
      return this._set.size;
    }

    /**
     * Add the given string to this set.
     *
     * @param String aStr
     */
    add(aStr, aAllowDuplicates) {
      const isDuplicate = this.has(aStr);
      const idx = this._array.length;
      if (!isDuplicate || aAllowDuplicates) {
        this._array.push(aStr);
      }
      if (!isDuplicate) {
        this._set.set(aStr, idx);
      }
    }

    /**
     * Is the given string a member of this set?
     *
     * @param String aStr
     */
    has(aStr) {
        return this._set.has(aStr);
    }

    /**
     * What is the index of the given string in the array?
     *
     * @param String aStr
     */
    indexOf(aStr) {
      const idx = this._set.get(aStr);
      if (idx >= 0) {
          return idx;
      }
      throw new Error('"' + aStr + '" is not in the set.');
    }

    /**
     * What is the element at the given index?
     *
     * @param Number aIdx
     */
    at(aIdx) {
      if (aIdx >= 0 && aIdx < this._array.length) {
        return this._array[aIdx];
      }
      throw new Error("No element indexed by " + aIdx);
    }

    /**
     * Returns the array representation of this set (which has the proper indices
     * indicated by indexOf). Note that this is a copy of the internal array used
     * for storing the members so that no one can mess with internal state.
     */
    toArray() {
      return this._array.slice();
    }
  }
  var ArraySet_1$1 = ArraySet$3;

  var arraySet$1 = {
  	ArraySet: ArraySet_1$1
  };

  /* -*- Mode: js; js-indent-level: 2; -*- */
  /*
   * Copyright 2014 Mozilla Foundation and contributors
   * Licensed under the New BSD license. See LICENSE or:
   * http://opensource.org/licenses/BSD-3-Clause
   */



  /**
   * Determine whether mappingB is after mappingA with respect to generated
   * position.
   */
  function generatedPositionAfter$1(mappingA, mappingB) {
    // Optimized for most common case
    const lineA = mappingA.generatedLine;
    const lineB = mappingB.generatedLine;
    const columnA = mappingA.generatedColumn;
    const columnB = mappingB.generatedColumn;
    return lineB > lineA || lineB == lineA && columnB >= columnA ||
           util$1.compareByGeneratedPositionsInflated(mappingA, mappingB) <= 0;
  }

  /**
   * A data structure to provide a sorted view of accumulated mappings in a
   * performance conscious manner. It trades a negligible overhead in general
   * case for a large speedup in case of mappings being added in order.
   */
  class MappingList$2 {
    constructor() {
      this._array = [];
      this._sorted = true;
      // Serves as infimum
      this._last = {generatedLine: -1, generatedColumn: 0};
    }

    /**
     * Iterate through internal items. This method takes the same arguments that
     * `Array.prototype.forEach` takes.
     *
     * NOTE: The order of the mappings is NOT guaranteed.
     */
    unsortedForEach(aCallback, aThisArg) {
      this._array.forEach(aCallback, aThisArg);
    }

    /**
     * Add the given source mapping.
     *
     * @param Object aMapping
     */
    add(aMapping) {
      if (generatedPositionAfter$1(this._last, aMapping)) {
        this._last = aMapping;
        this._array.push(aMapping);
      } else {
        this._sorted = false;
        this._array.push(aMapping);
      }
    }

    /**
     * Returns the flat, sorted array of mappings. The mappings are sorted by
     * generated position.
     *
     * WARNING: This method returns internal data without copying, for
     * performance. The return value must NOT be mutated, and should be treated as
     * an immutable borrow. If you want to take ownership, you must make your own
     * copy.
     */
    toArray() {
      if (!this._sorted) {
        this._array.sort(util$1.compareByGeneratedPositionsInflated);
        this._sorted = true;
      }
      return this._array;
    }
  }

  var MappingList_1$1 = MappingList$2;

  var mappingList$1 = {
  	MappingList: MappingList_1$1
  };

  /* -*- Mode: js; js-indent-level: 2; -*- */
  /*
   * Copyright 2011 Mozilla Foundation and contributors
   * Licensed under the New BSD license. See LICENSE or:
   * http://opensource.org/licenses/BSD-3-Clause
   */



  const ArraySet$4 = arraySet$1.ArraySet;
  const MappingList$3 = mappingList$1.MappingList;

  /**
   * An instance of the SourceMapGenerator represents a source map which is
   * being built incrementally. You may pass an object with the following
   * properties:
   *
   *   - file: The filename of the generated source.
   *   - sourceRoot: A root for all relative URLs in this source map.
   */
  class SourceMapGenerator$3 {
    constructor(aArgs) {
      if (!aArgs) {
        aArgs = {};
      }
      this._file = util$1.getArg(aArgs, "file", null);
      this._sourceRoot = util$1.getArg(aArgs, "sourceRoot", null);
      this._skipValidation = util$1.getArg(aArgs, "skipValidation", false);
      this._sources = new ArraySet$4();
      this._names = new ArraySet$4();
      this._mappings = new MappingList$3();
      this._sourcesContents = null;
    }

    /**
     * Creates a new SourceMapGenerator based on a SourceMapConsumer
     *
     * @param aSourceMapConsumer The SourceMap.
     */
    static fromSourceMap(aSourceMapConsumer) {
      const sourceRoot = aSourceMapConsumer.sourceRoot;
      const generator = new SourceMapGenerator$3({
        file: aSourceMapConsumer.file,
        sourceRoot
      });
      aSourceMapConsumer.eachMapping(function(mapping) {
        const newMapping = {
          generated: {
            line: mapping.generatedLine,
            column: mapping.generatedColumn
          }
        };

        if (mapping.source != null) {
          newMapping.source = mapping.source;
          if (sourceRoot != null) {
            newMapping.source = util$1.relative(sourceRoot, newMapping.source);
          }

          newMapping.original = {
            line: mapping.originalLine,
            column: mapping.originalColumn
          };

          if (mapping.name != null) {
            newMapping.name = mapping.name;
          }
        }

        generator.addMapping(newMapping);
      });
      aSourceMapConsumer.sources.forEach(function(sourceFile) {
        let sourceRelative = sourceFile;
        if (sourceRoot !== null) {
          sourceRelative = util$1.relative(sourceRoot, sourceFile);
        }

        if (!generator._sources.has(sourceRelative)) {
          generator._sources.add(sourceRelative);
        }

        const content = aSourceMapConsumer.sourceContentFor(sourceFile);
        if (content != null) {
          generator.setSourceContent(sourceFile, content);
        }
      });
      return generator;
    }

    /**
     * Add a single mapping from original source line and column to the generated
     * source's line and column for this source map being created. The mapping
     * object should have the following properties:
     *
     *   - generated: An object with the generated line and column positions.
     *   - original: An object with the original line and column positions.
     *   - source: The original source file (relative to the sourceRoot).
     *   - name: An optional original token name for this mapping.
     */
    addMapping(aArgs) {
      const generated = util$1.getArg(aArgs, "generated");
      const original = util$1.getArg(aArgs, "original", null);
      let source = util$1.getArg(aArgs, "source", null);
      let name = util$1.getArg(aArgs, "name", null);

      if (!this._skipValidation) {
        this._validateMapping(generated, original, source, name);
      }

      if (source != null) {
        source = String(source);
        if (!this._sources.has(source)) {
          this._sources.add(source);
        }
      }

      if (name != null) {
        name = String(name);
        if (!this._names.has(name)) {
          this._names.add(name);
        }
      }

      this._mappings.add({
        generatedLine: generated.line,
        generatedColumn: generated.column,
        originalLine: original != null && original.line,
        originalColumn: original != null && original.column,
        source,
        name
      });
    }

    /**
     * Set the source content for a source file.
     */
    setSourceContent(aSourceFile, aSourceContent) {
      let source = aSourceFile;
      if (this._sourceRoot != null) {
        source = util$1.relative(this._sourceRoot, source);
      }

      if (aSourceContent != null) {
        // Add the source content to the _sourcesContents map.
        // Create a new _sourcesContents map if the property is null.
        if (!this._sourcesContents) {
          this._sourcesContents = Object.create(null);
        }
        this._sourcesContents[util$1.toSetString(source)] = aSourceContent;
      } else if (this._sourcesContents) {
        // Remove the source file from the _sourcesContents map.
        // If the _sourcesContents map is empty, set the property to null.
        delete this._sourcesContents[util$1.toSetString(source)];
        if (Object.keys(this._sourcesContents).length === 0) {
          this._sourcesContents = null;
        }
      }
    }

    /**
     * Applies the mappings of a sub-source-map for a specific source file to the
     * source map being generated. Each mapping to the supplied source file is
     * rewritten using the supplied source map. Note: The resolution for the
     * resulting mappings is the minimium of this map and the supplied map.
     *
     * @param aSourceMapConsumer The source map to be applied.
     * @param aSourceFile Optional. The filename of the source file.
     *        If omitted, SourceMapConsumer's file property will be used.
     * @param aSourceMapPath Optional. The dirname of the path to the source map
     *        to be applied. If relative, it is relative to the SourceMapConsumer.
     *        This parameter is needed when the two source maps aren't in the same
     *        directory, and the source map to be applied contains relative source
     *        paths. If so, those relative source paths need to be rewritten
     *        relative to the SourceMapGenerator.
     */
    applySourceMap(aSourceMapConsumer, aSourceFile, aSourceMapPath) {
      let sourceFile = aSourceFile;
      // If aSourceFile is omitted, we will use the file property of the SourceMap
      if (aSourceFile == null) {
        if (aSourceMapConsumer.file == null) {
          throw new Error(
            "SourceMapGenerator.prototype.applySourceMap requires either an explicit source file, " +
            'or the source map\'s "file" property. Both were omitted.'
          );
        }
        sourceFile = aSourceMapConsumer.file;
      }
      const sourceRoot = this._sourceRoot;
      // Make "sourceFile" relative if an absolute Url is passed.
      if (sourceRoot != null) {
        sourceFile = util$1.relative(sourceRoot, sourceFile);
      }
      // Applying the SourceMap can add and remove items from the sources and
      // the names array.
      const newSources = this._mappings.toArray().length > 0
        ? new ArraySet$4()
        : this._sources;
      const newNames = new ArraySet$4();

      // Find mappings for the "sourceFile"
      this._mappings.unsortedForEach(function(mapping) {
        if (mapping.source === sourceFile && mapping.originalLine != null) {
          // Check if it can be mapped by the source map, then update the mapping.
          const original = aSourceMapConsumer.originalPositionFor({
            line: mapping.originalLine,
            column: mapping.originalColumn
          });
          if (original.source != null) {
            // Copy mapping
            mapping.source = original.source;
            if (aSourceMapPath != null) {
              mapping.source = util$1.join(aSourceMapPath, mapping.source);
            }
            if (sourceRoot != null) {
              mapping.source = util$1.relative(sourceRoot, mapping.source);
            }
            mapping.originalLine = original.line;
            mapping.originalColumn = original.column;
            if (original.name != null) {
              mapping.name = original.name;
            }
          }
        }

        const source = mapping.source;
        if (source != null && !newSources.has(source)) {
          newSources.add(source);
        }

        const name = mapping.name;
        if (name != null && !newNames.has(name)) {
          newNames.add(name);
        }

      }, this);
      this._sources = newSources;
      this._names = newNames;

      // Copy sourcesContents of applied map.
      aSourceMapConsumer.sources.forEach(function(srcFile) {
        const content = aSourceMapConsumer.sourceContentFor(srcFile);
        if (content != null) {
          if (aSourceMapPath != null) {
            srcFile = util$1.join(aSourceMapPath, srcFile);
          }
          if (sourceRoot != null) {
            srcFile = util$1.relative(sourceRoot, srcFile);
          }
          this.setSourceContent(srcFile, content);
        }
      }, this);
    }

    /**
     * A mapping can have one of the three levels of data:
     *
     *   1. Just the generated position.
     *   2. The Generated position, original position, and original source.
     *   3. Generated and original position, original source, as well as a name
     *      token.
     *
     * To maintain consistency, we validate that any new mapping being added falls
     * in to one of these categories.
     */
    _validateMapping(aGenerated, aOriginal, aSource, aName) {
      // When aOriginal is truthy but has empty values for .line and .column,
      // it is most likely a programmer error. In this case we throw a very
      // specific error message to try to guide them the right way.
      // For example: https://github.com/Polymer/polymer-bundler/pull/519
      if (aOriginal && typeof aOriginal.line !== "number" && typeof aOriginal.column !== "number") {
          throw new Error(
              "original.line and original.column are not numbers -- you probably meant to omit " +
              "the original mapping entirely and only map the generated position. If so, pass " +
              "null for the original mapping instead of an object with empty or null values."
          );
      }

      if (aGenerated && "line" in aGenerated && "column" in aGenerated
          && aGenerated.line > 0 && aGenerated.column >= 0
          && !aOriginal && !aSource && !aName) ; else if (aGenerated && "line" in aGenerated && "column" in aGenerated
               && aOriginal && "line" in aOriginal && "column" in aOriginal
               && aGenerated.line > 0 && aGenerated.column >= 0
               && aOriginal.line > 0 && aOriginal.column >= 0
               && aSource) ; else {
        throw new Error("Invalid mapping: " + JSON.stringify({
          generated: aGenerated,
          source: aSource,
          original: aOriginal,
          name: aName
        }));
      }
    }

    /**
     * Serialize the accumulated mappings in to the stream of base 64 VLQs
     * specified by the source map format.
     */
    _serializeMappings() {
      let previousGeneratedColumn = 0;
      let previousGeneratedLine = 1;
      let previousOriginalColumn = 0;
      let previousOriginalLine = 0;
      let previousName = 0;
      let previousSource = 0;
      let result = "";
      let next;
      let mapping;
      let nameIdx;
      let sourceIdx;

      const mappings = this._mappings.toArray();
      for (let i = 0, len = mappings.length; i < len; i++) {
        mapping = mappings[i];
        next = "";

        if (mapping.generatedLine !== previousGeneratedLine) {
          previousGeneratedColumn = 0;
          while (mapping.generatedLine !== previousGeneratedLine) {
            next += ";";
            previousGeneratedLine++;
          }
        } else if (i > 0) {
          if (!util$1.compareByGeneratedPositionsInflated(mapping, mappings[i - 1])) {
            continue;
          }
          next += ",";
        }

        next += base64Vlq$1.encode(mapping.generatedColumn
                                   - previousGeneratedColumn);
        previousGeneratedColumn = mapping.generatedColumn;

        if (mapping.source != null) {
          sourceIdx = this._sources.indexOf(mapping.source);
          next += base64Vlq$1.encode(sourceIdx - previousSource);
          previousSource = sourceIdx;

          // lines are stored 0-based in SourceMap spec version 3
          next += base64Vlq$1.encode(mapping.originalLine - 1
                                     - previousOriginalLine);
          previousOriginalLine = mapping.originalLine - 1;

          next += base64Vlq$1.encode(mapping.originalColumn
                                     - previousOriginalColumn);
          previousOriginalColumn = mapping.originalColumn;

          if (mapping.name != null) {
            nameIdx = this._names.indexOf(mapping.name);
            next += base64Vlq$1.encode(nameIdx - previousName);
            previousName = nameIdx;
          }
        }

        result += next;
      }

      return result;
    }

    _generateSourcesContent(aSources, aSourceRoot) {
      return aSources.map(function(source) {
        if (!this._sourcesContents) {
          return null;
        }
        if (aSourceRoot != null) {
          source = util$1.relative(aSourceRoot, source);
        }
        const key = util$1.toSetString(source);
        return Object.prototype.hasOwnProperty.call(this._sourcesContents, key)
          ? this._sourcesContents[key]
          : null;
      }, this);
    }

    /**
     * Externalize the source map.
     */
    toJSON() {
      const map = {
        version: this._version,
        sources: this._sources.toArray(),
        names: this._names.toArray(),
        mappings: this._serializeMappings()
      };
      if (this._file != null) {
        map.file = this._file;
      }
      if (this._sourceRoot != null) {
        map.sourceRoot = this._sourceRoot;
      }
      if (this._sourcesContents) {
        map.sourcesContent = this._generateSourcesContent(map.sources, map.sourceRoot);
      }

      return map;
    }

    /**
     * Render the source map being generated to a string.
     */
    toString() {
      return JSON.stringify(this.toJSON());
    }
  }

  SourceMapGenerator$3.prototype._version = 3;
  var SourceMapGenerator_1$1 = SourceMapGenerator$3;

  var sourceMapGenerator$1 = {
  	SourceMapGenerator: SourceMapGenerator_1$1
  };

  var binarySearch$1 = createCommonjsModule(function (module, exports) {
  /* -*- Mode: js; js-indent-level: 2; -*- */
  /*
   * Copyright 2011 Mozilla Foundation and contributors
   * Licensed under the New BSD license. See LICENSE or:
   * http://opensource.org/licenses/BSD-3-Clause
   */

  exports.GREATEST_LOWER_BOUND = 1;
  exports.LEAST_UPPER_BOUND = 2;

  /**
   * Recursive implementation of binary search.
   *
   * @param aLow Indices here and lower do not contain the needle.
   * @param aHigh Indices here and higher do not contain the needle.
   * @param aNeedle The element being searched for.
   * @param aHaystack The non-empty array being searched.
   * @param aCompare Function which takes two elements and returns -1, 0, or 1.
   * @param aBias Either 'binarySearch.GREATEST_LOWER_BOUND' or
   *     'binarySearch.LEAST_UPPER_BOUND'. Specifies whether to return the
   *     closest element that is smaller than or greater than the one we are
   *     searching for, respectively, if the exact element cannot be found.
   */
  function recursiveSearch(aLow, aHigh, aNeedle, aHaystack, aCompare, aBias) {
    // This function terminates when one of the following is true:
    //
    //   1. We find the exact element we are looking for.
    //
    //   2. We did not find the exact element, but we can return the index of
    //      the next-closest element.
    //
    //   3. We did not find the exact element, and there is no next-closest
    //      element than the one we are searching for, so we return -1.
    const mid = Math.floor((aHigh - aLow) / 2) + aLow;
    const cmp = aCompare(aNeedle, aHaystack[mid], true);
    if (cmp === 0) {
      // Found the element we are looking for.
      return mid;
    } else if (cmp > 0) {
      // Our needle is greater than aHaystack[mid].
      if (aHigh - mid > 1) {
        // The element is in the upper half.
        return recursiveSearch(mid, aHigh, aNeedle, aHaystack, aCompare, aBias);
      }

      // The exact needle element was not found in this haystack. Determine if
      // we are in termination case (3) or (2) and return the appropriate thing.
      if (aBias == exports.LEAST_UPPER_BOUND) {
        return aHigh < aHaystack.length ? aHigh : -1;
      }
      return mid;
    }

    // Our needle is less than aHaystack[mid].
    if (mid - aLow > 1) {
      // The element is in the lower half.
      return recursiveSearch(aLow, mid, aNeedle, aHaystack, aCompare, aBias);
    }

    // we are in termination case (3) or (2) and return the appropriate thing.
    if (aBias == exports.LEAST_UPPER_BOUND) {
      return mid;
    }
    return aLow < 0 ? -1 : aLow;
  }

  /**
   * This is an implementation of binary search which will always try and return
   * the index of the closest element if there is no exact hit. This is because
   * mappings between original and generated line/col pairs are single points,
   * and there is an implicit region between each of them, so a miss just means
   * that you aren't on the very start of a region.
   *
   * @param aNeedle The element you are looking for.
   * @param aHaystack The array that is being searched.
   * @param aCompare A function which takes the needle and an element in the
   *     array and returns -1, 0, or 1 depending on whether the needle is less
   *     than, equal to, or greater than the element, respectively.
   * @param aBias Either 'binarySearch.GREATEST_LOWER_BOUND' or
   *     'binarySearch.LEAST_UPPER_BOUND'. Specifies whether to return the
   *     closest element that is smaller than or greater than the one we are
   *     searching for, respectively, if the exact element cannot be found.
   *     Defaults to 'binarySearch.GREATEST_LOWER_BOUND'.
   */
  exports.search = function search(aNeedle, aHaystack, aCompare, aBias) {
    if (aHaystack.length === 0) {
      return -1;
    }

    let index = recursiveSearch(-1, aHaystack.length, aNeedle, aHaystack,
                                aCompare, aBias || exports.GREATEST_LOWER_BOUND);
    if (index < 0) {
      return -1;
    }

    // We have found either the exact element, or the next-closest element than
    // the one we are searching for. However, there may be more than one such
    // element. Make sure we always return the smallest of these.
    while (index - 1 >= 0) {
      if (aCompare(aHaystack[index], aHaystack[index - 1], true) !== 0) {
        break;
      }
      --index;
    }

    return index;
  };
  });
  var binarySearch_1$1 = binarySearch$1.GREATEST_LOWER_BOUND;
  var binarySearch_2$1 = binarySearch$1.LEAST_UPPER_BOUND;
  var binarySearch_3$1 = binarySearch$1.search;

  var readWasm = createCommonjsModule(function (module) {
  if (typeof fetch === "function") {
    // Web version of reading a wasm file into an array buffer.

    let mappingsWasmUrl = null;

    module.exports = function readWasm() {
      if (typeof mappingsWasmUrl !== "string") {
        throw new Error("You must provide the URL of lib/mappings.wasm by calling " +
                        "SourceMapConsumer.initialize({ 'lib/mappings.wasm': ... }) " +
                        "before using SourceMapConsumer");
      }

      return fetch(mappingsWasmUrl)
        .then(response => response.arrayBuffer());
    };

    module.exports.initialize = url => mappingsWasmUrl = url;
  } else {
    // Node version of reading a wasm file into an array buffer.
    const fs = require$$1;
    const path = require$$1;

    module.exports = function readWasm() {
      return new Promise((resolve, reject) => {
        const wasmPath = path.join(__dirname, "mappings.wasm");
        fs.readFile(wasmPath, null, (error, data) => {
          if (error) {
            reject(error);
            return;
          }

          resolve(data.buffer);
        });
      });
    };

    module.exports.initialize = _ => {
      console.debug("SourceMapConsumer.initialize is a no-op when running in node.js");
    };
  }
  });
  var readWasm_1 = readWasm.initialize;

  /**
   * Provide the JIT with a nice shape / hidden class.
   */
  function Mapping$1() {
    this.generatedLine = 0;
    this.generatedColumn = 0;
    this.lastGeneratedColumn = null;
    this.source = null;
    this.originalLine = null;
    this.originalColumn = null;
    this.name = null;
  }

  let cachedWasm = null;

  var wasm = function wasm() {
    if (cachedWasm) {
      return cachedWasm;
    }

    const callbackStack = [];

    cachedWasm = readWasm().then(buffer => {
        return WebAssembly.instantiate(buffer, {
          env: {
            mapping_callback(
              generatedLine,
              generatedColumn,

              hasLastGeneratedColumn,
              lastGeneratedColumn,

              hasOriginal,
              source,
              originalLine,
              originalColumn,

              hasName,
              name
            ) {
              const mapping = new Mapping$1();
              // JS uses 1-based line numbers, wasm uses 0-based.
              mapping.generatedLine = generatedLine + 1;
              mapping.generatedColumn = generatedColumn;

              if (hasLastGeneratedColumn) {
                // JS uses inclusive last generated column, wasm uses exclusive.
                mapping.lastGeneratedColumn = lastGeneratedColumn - 1;
              }

              if (hasOriginal) {
                mapping.source = source;
                // JS uses 1-based line numbers, wasm uses 0-based.
                mapping.originalLine = originalLine + 1;
                mapping.originalColumn = originalColumn;

                if (hasName) {
                  mapping.name = name;
                }
              }

              callbackStack[callbackStack.length - 1](mapping);
            },

            start_all_generated_locations_for() { console.time("all_generated_locations_for"); },
            end_all_generated_locations_for() { console.timeEnd("all_generated_locations_for"); },

            start_compute_column_spans() { console.time("compute_column_spans"); },
            end_compute_column_spans() { console.timeEnd("compute_column_spans"); },

            start_generated_location_for() { console.time("generated_location_for"); },
            end_generated_location_for() { console.timeEnd("generated_location_for"); },

            start_original_location_for() { console.time("original_location_for"); },
            end_original_location_for() { console.timeEnd("original_location_for"); },

            start_parse_mappings() { console.time("parse_mappings"); },
            end_parse_mappings() { console.timeEnd("parse_mappings"); },

            start_sort_by_generated_location() { console.time("sort_by_generated_location"); },
            end_sort_by_generated_location() { console.timeEnd("sort_by_generated_location"); },

            start_sort_by_original_location() { console.time("sort_by_original_location"); },
            end_sort_by_original_location() { console.timeEnd("sort_by_original_location"); },
          }
        });
    }).then(Wasm => {
      return {
        exports: Wasm.instance.exports,
        withMappingCallback: (mappingCallback, f) => {
          callbackStack.push(mappingCallback);
          try {
            f();
          } finally {
            callbackStack.pop();
          }
        }
      };
    }).then(null, e => {
      cachedWasm = null;
      throw e;
    });

    return cachedWasm;
  };

  /* -*- Mode: js; js-indent-level: 2; -*- */
  /*
   * Copyright 2011 Mozilla Foundation and contributors
   * Licensed under the New BSD license. See LICENSE or:
   * http://opensource.org/licenses/BSD-3-Clause
   */



  const ArraySet$5 = arraySet$1.ArraySet;
   // eslint-disable-line no-unused-vars



  const INTERNAL = Symbol("smcInternal");

  class SourceMapConsumer$2 {
    constructor(aSourceMap, aSourceMapURL) {
      // If the constructor was called by super(), just return Promise<this>.
      // Yes, this is a hack to retain the pre-existing API of the base-class
      // constructor also being an async factory function.
      if (aSourceMap == INTERNAL) {
        return Promise.resolve(this);
      }

      return _factory(aSourceMap, aSourceMapURL);
    }

    static initialize(opts) {
      readWasm.initialize(opts["lib/mappings.wasm"]);
    }

    static fromSourceMap(aSourceMap, aSourceMapURL) {
      return _factoryBSM(aSourceMap, aSourceMapURL);
    }

    /**
     * Construct a new `SourceMapConsumer` from `rawSourceMap` and `sourceMapUrl`
     * (see the `SourceMapConsumer` constructor for details. Then, invoke the `async
     * function f(SourceMapConsumer) -> T` with the newly constructed consumer, wait
     * for `f` to complete, call `destroy` on the consumer, and return `f`'s return
     * value.
     *
     * You must not use the consumer after `f` completes!
     *
     * By using `with`, you do not have to remember to manually call `destroy` on
     * the consumer, since it will be called automatically once `f` completes.
     *
     * ```js
     * const xSquared = await SourceMapConsumer.with(
     *   myRawSourceMap,
     *   null,
     *   async function (consumer) {
     *     // Use `consumer` inside here and don't worry about remembering
     *     // to call `destroy`.
     *
     *     const x = await whatever(consumer);
     *     return x * x;
     *   }
     * );
     *
     * // You may not use that `consumer` anymore out here; it has
     * // been destroyed. But you can use `xSquared`.
     * console.log(xSquared);
     * ```
     */
    static with(rawSourceMap, sourceMapUrl, f) {
      // Note: The `acorn` version that `webpack` currently depends on doesn't
      // support `async` functions, and the nodes that we support don't all have
      // `.finally`. Therefore, this is written a bit more convolutedly than it
      // should really be.

      let consumer = null;
      const promise = new SourceMapConsumer$2(rawSourceMap, sourceMapUrl);
      return promise
        .then(c => {
          consumer = c;
          return f(c);
        })
        .then(x => {
          if (consumer) {
            consumer.destroy();
          }
          return x;
        }, e => {
          if (consumer) {
            consumer.destroy();
          }
          throw e;
        });
    }

    /**
     * Parse the mappings in a string in to a data structure which we can easily
     * query (the ordered arrays in the `this.__generatedMappings` and
     * `this.__originalMappings` properties).
     */
    _parseMappings(aStr, aSourceRoot) {
      throw new Error("Subclasses must implement _parseMappings");
    }

    /**
     * Iterate over each mapping between an original source/line/column and a
     * generated line/column in this source map.
     *
     * @param Function aCallback
     *        The function that is called with each mapping.
     * @param Object aContext
     *        Optional. If specified, this object will be the value of `this` every
     *        time that `aCallback` is called.
     * @param aOrder
     *        Either `SourceMapConsumer.GENERATED_ORDER` or
     *        `SourceMapConsumer.ORIGINAL_ORDER`. Specifies whether you want to
     *        iterate over the mappings sorted by the generated file's line/column
     *        order or the original's source/line/column order, respectively. Defaults to
     *        `SourceMapConsumer.GENERATED_ORDER`.
     */
    eachMapping(aCallback, aContext, aOrder) {
      throw new Error("Subclasses must implement eachMapping");
    }

    /**
     * Returns all generated line and column information for the original source,
     * line, and column provided. If no column is provided, returns all mappings
     * corresponding to a either the line we are searching for or the next
     * closest line that has any mappings. Otherwise, returns all mappings
     * corresponding to the given line and either the column we are searching for
     * or the next closest column that has any offsets.
     *
     * The only argument is an object with the following properties:
     *
     *   - source: The filename of the original source.
     *   - line: The line number in the original source.  The line number is 1-based.
     *   - column: Optional. the column number in the original source.
     *    The column number is 0-based.
     *
     * and an array of objects is returned, each with the following properties:
     *
     *   - line: The line number in the generated source, or null.  The
     *    line number is 1-based.
     *   - column: The column number in the generated source, or null.
     *    The column number is 0-based.
     */
    allGeneratedPositionsFor(aArgs) {
      throw new Error("Subclasses must implement allGeneratedPositionsFor");
    }

    destroy() {
      throw new Error("Subclasses must implement destroy");
    }
  }

  /**
   * The version of the source mapping spec that we are consuming.
   */
  SourceMapConsumer$2.prototype._version = 3;
  SourceMapConsumer$2.GENERATED_ORDER = 1;
  SourceMapConsumer$2.ORIGINAL_ORDER = 2;

  SourceMapConsumer$2.GREATEST_LOWER_BOUND = 1;
  SourceMapConsumer$2.LEAST_UPPER_BOUND = 2;

  /**
   * A BasicSourceMapConsumer instance represents a parsed source map which we can
   * query for information about the original file positions by giving it a file
   * position in the generated source.
   *
   * The first parameter is the raw source map (either as a JSON string, or
   * already parsed to an object). According to the spec, source maps have the
   * following attributes:
   *
   *   - version: Which version of the source map spec this map is following.
   *   - sources: An array of URLs to the original source files.
   *   - names: An array of identifiers which can be referenced by individual mappings.
   *   - sourceRoot: Optional. The URL root from which all sources are relative.
   *   - sourcesContent: Optional. An array of contents of the original source files.
   *   - mappings: A string of base64 VLQs which contain the actual mappings.
   *   - file: Optional. The generated file this source map is associated with.
   *
   * Here is an example source map, taken from the source map spec[0]:
   *
   *     {
   *       version : 3,
   *       file: "out.js",
   *       sourceRoot : "",
   *       sources: ["foo.js", "bar.js"],
   *       names: ["src", "maps", "are", "fun"],
   *       mappings: "AA,AB;;ABCDE;"
   *     }
   *
   * The second parameter, if given, is a string whose value is the URL
   * at which the source map was found.  This URL is used to compute the
   * sources array.
   *
   * [0]: https://docs.google.com/document/d/1U1RGAehQwRypUTovF1KRlpiOFze0b-_2gc6fAH0KY0k/edit?pli=1#
   */
  class BasicSourceMapConsumer$1 extends SourceMapConsumer$2 {
    constructor(aSourceMap, aSourceMapURL) {
      return super(INTERNAL).then(that => {
        let sourceMap = aSourceMap;
        if (typeof aSourceMap === "string") {
          sourceMap = util$1.parseSourceMapInput(aSourceMap);
        }

        const version = util$1.getArg(sourceMap, "version");
        let sources = util$1.getArg(sourceMap, "sources");
        // Sass 3.3 leaves out the 'names' array, so we deviate from the spec (which
        // requires the array) to play nice here.
        const names = util$1.getArg(sourceMap, "names", []);
        let sourceRoot = util$1.getArg(sourceMap, "sourceRoot", null);
        const sourcesContent = util$1.getArg(sourceMap, "sourcesContent", null);
        const mappings = util$1.getArg(sourceMap, "mappings");
        const file = util$1.getArg(sourceMap, "file", null);

        // Once again, Sass deviates from the spec and supplies the version as a
        // string rather than a number, so we use loose equality checking here.
        if (version != that._version) {
          throw new Error("Unsupported version: " + version);
        }

        if (sourceRoot) {
          sourceRoot = util$1.normalize(sourceRoot);
        }

        sources = sources
          .map(String)
          // Some source maps produce relative source paths like "./foo.js" instead of
          // "foo.js".  Normalize these first so that future comparisons will succeed.
          // See bugzil.la/1090768.
          .map(util$1.normalize)
          // Always ensure that absolute sources are internally stored relative to
          // the source root, if the source root is absolute. Not doing this would
          // be particularly problematic when the source root is a prefix of the
          // source (valid, but why??). See github issue #199 and bugzil.la/1188982.
          .map(function(source) {
            return sourceRoot && util$1.isAbsolute(sourceRoot) && util$1.isAbsolute(source)
              ? util$1.relative(sourceRoot, source)
              : source;
          });

        // Pass `true` below to allow duplicate names and sources. While source maps
        // are intended to be compressed and deduplicated, the TypeScript compiler
        // sometimes generates source maps with duplicates in them. See Github issue
        // #72 and bugzil.la/889492.
        that._names = ArraySet$5.fromArray(names.map(String), true);
        that._sources = ArraySet$5.fromArray(sources, true);

        that._absoluteSources = that._sources.toArray().map(function(s) {
          return util$1.computeSourceURL(sourceRoot, s, aSourceMapURL);
        });

        that.sourceRoot = sourceRoot;
        that.sourcesContent = sourcesContent;
        that._mappings = mappings;
        that._sourceMapURL = aSourceMapURL;
        that.file = file;

        that._computedColumnSpans = false;
        that._mappingsPtr = 0;
        that._wasm = null;

        return wasm().then(w => {
          that._wasm = w;
          return that;
        });
      });
    }

    /**
     * Utility function to find the index of a source.  Returns -1 if not
     * found.
     */
    _findSourceIndex(aSource) {
      let relativeSource = aSource;
      if (this.sourceRoot != null) {
        relativeSource = util$1.relative(this.sourceRoot, relativeSource);
      }

      if (this._sources.has(relativeSource)) {
        return this._sources.indexOf(relativeSource);
      }

      // Maybe aSource is an absolute URL as returned by |sources|.  In
      // this case we can't simply undo the transform.
      for (let i = 0; i < this._absoluteSources.length; ++i) {
        if (this._absoluteSources[i] == aSource) {
          return i;
        }
      }

      return -1;
    }

    /**
     * Create a BasicSourceMapConsumer from a SourceMapGenerator.
     *
     * @param SourceMapGenerator aSourceMap
     *        The source map that will be consumed.
     * @param String aSourceMapURL
     *        The URL at which the source map can be found (optional)
     * @returns BasicSourceMapConsumer
     */
    static fromSourceMap(aSourceMap, aSourceMapURL) {
      return new BasicSourceMapConsumer$1(aSourceMap.toString());
    }

    get sources() {
      return this._absoluteSources.slice();
    }

    _getMappingsPtr() {
      if (this._mappingsPtr === 0) {
        this._parseMappings(this._mappings, this.sourceRoot);
      }

      return this._mappingsPtr;
    }

    /**
     * Parse the mappings in a string in to a data structure which we can easily
     * query (the ordered arrays in the `this.__generatedMappings` and
     * `this.__originalMappings` properties).
     */
    _parseMappings(aStr, aSourceRoot) {
      const size = aStr.length;

      const mappingsBufPtr = this._wasm.exports.allocate_mappings(size);
      const mappingsBuf = new Uint8Array(this._wasm.exports.memory.buffer, mappingsBufPtr, size);
      for (let i = 0; i < size; i++) {
        mappingsBuf[i] = aStr.charCodeAt(i);
      }

      const mappingsPtr = this._wasm.exports.parse_mappings(mappingsBufPtr);

      if (!mappingsPtr) {
        const error = this._wasm.exports.get_last_error();
        let msg = `Error parsing mappings (code ${error}): `;

        // XXX: keep these error codes in sync with `fitzgen/source-map-mappings`.
        switch (error) {
          case 1:
            msg += "the mappings contained a negative line, column, source index, or name index";
            break;
          case 2:
            msg += "the mappings contained a number larger than 2**32";
            break;
          case 3:
            msg += "reached EOF while in the middle of parsing a VLQ";
            break;
          case 4:
            msg += "invalid base 64 character while parsing a VLQ";
            break;
          default:
            msg += "unknown error code";
            break;
        }

        throw new Error(msg);
      }

      this._mappingsPtr = mappingsPtr;
    }

    eachMapping(aCallback, aContext, aOrder) {
      const context = aContext || null;
      const order = aOrder || SourceMapConsumer$2.GENERATED_ORDER;
      const sourceRoot = this.sourceRoot;

      this._wasm.withMappingCallback(
        mapping => {
          if (mapping.source !== null) {
            mapping.source = this._sources.at(mapping.source);
            mapping.source = util$1.computeSourceURL(sourceRoot, mapping.source, this._sourceMapURL);

            if (mapping.name !== null) {
              mapping.name = this._names.at(mapping.name);
            }
          }

          aCallback.call(context, mapping);
        },
        () => {
          switch (order) {
          case SourceMapConsumer$2.GENERATED_ORDER:
            this._wasm.exports.by_generated_location(this._getMappingsPtr());
            break;
          case SourceMapConsumer$2.ORIGINAL_ORDER:
            this._wasm.exports.by_original_location(this._getMappingsPtr());
            break;
          default:
            throw new Error("Unknown order of iteration.");
          }
        }
      );
    }

    allGeneratedPositionsFor(aArgs) {
      let source = util$1.getArg(aArgs, "source");
      const originalLine = util$1.getArg(aArgs, "line");
      const originalColumn = aArgs.column || 0;

      source = this._findSourceIndex(source);
      if (source < 0) {
        return [];
      }

      if (originalLine < 1) {
        throw new Error("Line numbers must be >= 1");
      }

      if (originalColumn < 0) {
        throw new Error("Column numbers must be >= 0");
      }

      const mappings = [];

      this._wasm.withMappingCallback(
        m => {
          let lastColumn = m.lastGeneratedColumn;
          if (this._computedColumnSpans && lastColumn === null) {
            lastColumn = Infinity;
          }
          mappings.push({
            line: m.generatedLine,
            column: m.generatedColumn,
            lastColumn,
          });
        }, () => {
          this._wasm.exports.all_generated_locations_for(
            this._getMappingsPtr(),
            source,
            originalLine - 1,
            "column" in aArgs,
            originalColumn
          );
        }
      );

      return mappings;
    }

    destroy() {
      if (this._mappingsPtr !== 0) {
        this._wasm.exports.free_mappings(this._mappingsPtr);
        this._mappingsPtr = 0;
      }
    }

    /**
     * Compute the last column for each generated mapping. The last column is
     * inclusive.
     */
    computeColumnSpans() {
      if (this._computedColumnSpans) {
        return;
      }

      this._wasm.exports.compute_column_spans(this._getMappingsPtr());
      this._computedColumnSpans = true;
    }

    /**
     * Returns the original source, line, and column information for the generated
     * source's line and column positions provided. The only argument is an object
     * with the following properties:
     *
     *   - line: The line number in the generated source.  The line number
     *     is 1-based.
     *   - column: The column number in the generated source.  The column
     *     number is 0-based.
     *   - bias: Either 'SourceMapConsumer.GREATEST_LOWER_BOUND' or
     *     'SourceMapConsumer.LEAST_UPPER_BOUND'. Specifies whether to return the
     *     closest element that is smaller than or greater than the one we are
     *     searching for, respectively, if the exact element cannot be found.
     *     Defaults to 'SourceMapConsumer.GREATEST_LOWER_BOUND'.
     *
     * and an object is returned with the following properties:
     *
     *   - source: The original source file, or null.
     *   - line: The line number in the original source, or null.  The
     *     line number is 1-based.
     *   - column: The column number in the original source, or null.  The
     *     column number is 0-based.
     *   - name: The original identifier, or null.
     */
    originalPositionFor(aArgs) {
      const needle = {
        generatedLine: util$1.getArg(aArgs, "line"),
        generatedColumn: util$1.getArg(aArgs, "column")
      };

      if (needle.generatedLine < 1) {
        throw new Error("Line numbers must be >= 1");
      }

      if (needle.generatedColumn < 0) {
        throw new Error("Column numbers must be >= 0");
      }

      let bias = util$1.getArg(aArgs, "bias", SourceMapConsumer$2.GREATEST_LOWER_BOUND);
      if (bias == null) {
        bias = SourceMapConsumer$2.GREATEST_LOWER_BOUND;
      }

      let mapping;
      this._wasm.withMappingCallback(m => mapping = m, () => {
        this._wasm.exports.original_location_for(
          this._getMappingsPtr(),
          needle.generatedLine - 1,
          needle.generatedColumn,
          bias
        );
      });

      if (mapping) {
        if (mapping.generatedLine === needle.generatedLine) {
          let source = util$1.getArg(mapping, "source", null);
          if (source !== null) {
            source = this._sources.at(source);
            source = util$1.computeSourceURL(this.sourceRoot, source, this._sourceMapURL);
          }

          let name = util$1.getArg(mapping, "name", null);
          if (name !== null) {
            name = this._names.at(name);
          }

          return {
            source,
            line: util$1.getArg(mapping, "originalLine", null),
            column: util$1.getArg(mapping, "originalColumn", null),
            name
          };
        }
      }

      return {
        source: null,
        line: null,
        column: null,
        name: null
      };
    }

    /**
     * Return true if we have the source content for every source in the source
     * map, false otherwise.
     */
    hasContentsOfAllSources() {
      if (!this.sourcesContent) {
        return false;
      }
      return this.sourcesContent.length >= this._sources.size() &&
        !this.sourcesContent.some(function(sc) { return sc == null; });
    }

    /**
     * Returns the original source content. The only argument is the url of the
     * original source file. Returns null if no original source content is
     * available.
     */
    sourceContentFor(aSource, nullOnMissing) {
      if (!this.sourcesContent) {
        return null;
      }

      const index = this._findSourceIndex(aSource);
      if (index >= 0) {
        return this.sourcesContent[index];
      }

      let relativeSource = aSource;
      if (this.sourceRoot != null) {
        relativeSource = util$1.relative(this.sourceRoot, relativeSource);
      }

      let url;
      if (this.sourceRoot != null
          && (url = util$1.urlParse(this.sourceRoot))) {
        // XXX: file:// URIs and absolute paths lead to unexpected behavior for
        // many users. We can help them out when they expect file:// URIs to
        // behave like it would if they were running a local HTTP server. See
        // https://bugzilla.mozilla.org/show_bug.cgi?id=885597.
        const fileUriAbsPath = relativeSource.replace(/^file:\/\//, "");
        if (url.scheme == "file"
            && this._sources.has(fileUriAbsPath)) {
          return this.sourcesContent[this._sources.indexOf(fileUriAbsPath)];
        }

        if ((!url.path || url.path == "/")
            && this._sources.has("/" + relativeSource)) {
          return this.sourcesContent[this._sources.indexOf("/" + relativeSource)];
        }
      }

      // This function is used recursively from
      // IndexedSourceMapConsumer.prototype.sourceContentFor. In that case, we
      // don't want to throw if we can't find the source - we just want to
      // return null, so we provide a flag to exit gracefully.
      if (nullOnMissing) {
        return null;
      }

      throw new Error('"' + relativeSource + '" is not in the SourceMap.');
    }

    /**
     * Returns the generated line and column information for the original source,
     * line, and column positions provided. The only argument is an object with
     * the following properties:
     *
     *   - source: The filename of the original source.
     *   - line: The line number in the original source.  The line number
     *     is 1-based.
     *   - column: The column number in the original source.  The column
     *     number is 0-based.
     *   - bias: Either 'SourceMapConsumer.GREATEST_LOWER_BOUND' or
     *     'SourceMapConsumer.LEAST_UPPER_BOUND'. Specifies whether to return the
     *     closest element that is smaller than or greater than the one we are
     *     searching for, respectively, if the exact element cannot be found.
     *     Defaults to 'SourceMapConsumer.GREATEST_LOWER_BOUND'.
     *
     * and an object is returned with the following properties:
     *
     *   - line: The line number in the generated source, or null.  The
     *     line number is 1-based.
     *   - column: The column number in the generated source, or null.
     *     The column number is 0-based.
     */
    generatedPositionFor(aArgs) {
      let source = util$1.getArg(aArgs, "source");
      source = this._findSourceIndex(source);
      if (source < 0) {
        return {
          line: null,
          column: null,
          lastColumn: null
        };
      }

      const needle = {
        source,
        originalLine: util$1.getArg(aArgs, "line"),
        originalColumn: util$1.getArg(aArgs, "column")
      };

      if (needle.originalLine < 1) {
        throw new Error("Line numbers must be >= 1");
      }

      if (needle.originalColumn < 0) {
        throw new Error("Column numbers must be >= 0");
      }

      let bias = util$1.getArg(aArgs, "bias", SourceMapConsumer$2.GREATEST_LOWER_BOUND);
      if (bias == null) {
        bias = SourceMapConsumer$2.GREATEST_LOWER_BOUND;
      }

      let mapping;
      this._wasm.withMappingCallback(m => mapping = m, () => {
        this._wasm.exports.generated_location_for(
          this._getMappingsPtr(),
          needle.source,
          needle.originalLine - 1,
          needle.originalColumn,
          bias
        );
      });

      if (mapping) {
        if (mapping.source === needle.source) {
          let lastColumn = mapping.lastGeneratedColumn;
          if (this._computedColumnSpans && lastColumn === null) {
            lastColumn = Infinity;
          }
          return {
            line: util$1.getArg(mapping, "generatedLine", null),
            column: util$1.getArg(mapping, "generatedColumn", null),
            lastColumn,
          };
        }
      }

      return {
        line: null,
        column: null,
        lastColumn: null
      };
    }
  }

  BasicSourceMapConsumer$1.prototype.consumer = SourceMapConsumer$2;

  /**
   * An IndexedSourceMapConsumer instance represents a parsed source map which
   * we can query for information. It differs from BasicSourceMapConsumer in
   * that it takes "indexed" source maps (i.e. ones with a "sections" field) as
   * input.
   *
   * The first parameter is a raw source map (either as a JSON string, or already
   * parsed to an object). According to the spec for indexed source maps, they
   * have the following attributes:
   *
   *   - version: Which version of the source map spec this map is following.
   *   - file: Optional. The generated file this source map is associated with.
   *   - sections: A list of section definitions.
   *
   * Each value under the "sections" field has two fields:
   *   - offset: The offset into the original specified at which this section
   *       begins to apply, defined as an object with a "line" and "column"
   *       field.
   *   - map: A source map definition. This source map could also be indexed,
   *       but doesn't have to be.
   *
   * Instead of the "map" field, it's also possible to have a "url" field
   * specifying a URL to retrieve a source map from, but that's currently
   * unsupported.
   *
   * Here's an example source map, taken from the source map spec[0], but
   * modified to omit a section which uses the "url" field.
   *
   *  {
   *    version : 3,
   *    file: "app.js",
   *    sections: [{
   *      offset: {line:100, column:10},
   *      map: {
   *        version : 3,
   *        file: "section.js",
   *        sources: ["foo.js", "bar.js"],
   *        names: ["src", "maps", "are", "fun"],
   *        mappings: "AAAA,E;;ABCDE;"
   *      }
   *    }],
   *  }
   *
   * The second parameter, if given, is a string whose value is the URL
   * at which the source map was found.  This URL is used to compute the
   * sources array.
   *
   * [0]: https://docs.google.com/document/d/1U1RGAehQwRypUTovF1KRlpiOFze0b-_2gc6fAH0KY0k/edit#heading=h.535es3xeprgt
   */
  class IndexedSourceMapConsumer$1 extends SourceMapConsumer$2 {
    constructor(aSourceMap, aSourceMapURL) {
      return super(INTERNAL).then(that => {
        let sourceMap = aSourceMap;
        if (typeof aSourceMap === "string") {
          sourceMap = util$1.parseSourceMapInput(aSourceMap);
        }

        const version = util$1.getArg(sourceMap, "version");
        const sections = util$1.getArg(sourceMap, "sections");

        if (version != that._version) {
          throw new Error("Unsupported version: " + version);
        }

        that._sources = new ArraySet$5();
        that._names = new ArraySet$5();
        that.__generatedMappings = null;
        that.__originalMappings = null;
        that.__generatedMappingsUnsorted = null;
        that.__originalMappingsUnsorted = null;

        let lastOffset = {
          line: -1,
          column: 0
        };
        return Promise.all(sections.map(s => {
          if (s.url) {
            // The url field will require support for asynchronicity.
            // See https://github.com/mozilla/source-map/issues/16
            throw new Error("Support for url field in sections not implemented.");
          }
          const offset = util$1.getArg(s, "offset");
          const offsetLine = util$1.getArg(offset, "line");
          const offsetColumn = util$1.getArg(offset, "column");

          if (offsetLine < lastOffset.line ||
              (offsetLine === lastOffset.line && offsetColumn < lastOffset.column)) {
            throw new Error("Section offsets must be ordered and non-overlapping.");
          }
          lastOffset = offset;

          const cons = new SourceMapConsumer$2(util$1.getArg(s, "map"), aSourceMapURL);
          return cons.then(consumer => {
            return {
              generatedOffset: {
                // The offset fields are 0-based, but we use 1-based indices when
                // encoding/decoding from VLQ.
                generatedLine: offsetLine + 1,
                generatedColumn: offsetColumn + 1
              },
              consumer
            };
          });
        })).then(s => {
          that._sections = s;
          return that;
        });
      });
    }

    // `__generatedMappings` and `__originalMappings` are arrays that hold the
    // parsed mapping coordinates from the source map's "mappings" attribute. They
    // are lazily instantiated, accessed via the `_generatedMappings` and
    // `_originalMappings` getters respectively, and we only parse the mappings
    // and create these arrays once queried for a source location. We jump through
    // these hoops because there can be many thousands of mappings, and parsing
    // them is expensive, so we only want to do it if we must.
    //
    // Each object in the arrays is of the form:
    //
    //     {
    //       generatedLine: The line number in the generated code,
    //       generatedColumn: The column number in the generated code,
    //       source: The path to the original source file that generated this
    //               chunk of code,
    //       originalLine: The line number in the original source that
    //                     corresponds to this chunk of generated code,
    //       originalColumn: The column number in the original source that
    //                       corresponds to this chunk of generated code,
    //       name: The name of the original symbol which generated this chunk of
    //             code.
    //     }
    //
    // All properties except for `generatedLine` and `generatedColumn` can be
    // `null`.
    //
    // `_generatedMappings` is ordered by the generated positions.
    //
    // `_originalMappings` is ordered by the original positions.
    get _generatedMappings() {
      if (!this.__generatedMappings) {
        this._sortGeneratedMappings();
      }

      return this.__generatedMappings;
    }

    get _originalMappings() {
      if (!this.__originalMappings) {
        this._sortOriginalMappings();
      }

      return this.__originalMappings;
    }

    get _generatedMappingsUnsorted() {
      if (!this.__generatedMappingsUnsorted) {
        this._parseMappings(this._mappings, this.sourceRoot);
      }

      return this.__generatedMappingsUnsorted;
    }

    get _originalMappingsUnsorted() {
      if (!this.__originalMappingsUnsorted) {
        this._parseMappings(this._mappings, this.sourceRoot);
      }

      return this.__originalMappingsUnsorted;
    }

    _sortGeneratedMappings() {
      const mappings = this._generatedMappingsUnsorted;
      mappings.sort(util$1.compareByGeneratedPositionsDeflated);
      this.__generatedMappings = mappings;
    }

    _sortOriginalMappings() {
      const mappings = this._originalMappingsUnsorted;
      mappings.sort(util$1.compareByOriginalPositions);
      this.__originalMappings = mappings;
    }

    /**
     * The list of original sources.
     */
    get sources() {
      const sources = [];
      for (let i = 0; i < this._sections.length; i++) {
        for (let j = 0; j < this._sections[i].consumer.sources.length; j++) {
          sources.push(this._sections[i].consumer.sources[j]);
        }
      }
      return sources;
    }

    /**
     * Returns the original source, line, and column information for the generated
     * source's line and column positions provided. The only argument is an object
     * with the following properties:
     *
     *   - line: The line number in the generated source.  The line number
     *     is 1-based.
     *   - column: The column number in the generated source.  The column
     *     number is 0-based.
     *
     * and an object is returned with the following properties:
     *
     *   - source: The original source file, or null.
     *   - line: The line number in the original source, or null.  The
     *     line number is 1-based.
     *   - column: The column number in the original source, or null.  The
     *     column number is 0-based.
     *   - name: The original identifier, or null.
     */
    originalPositionFor(aArgs) {
      const needle = {
        generatedLine: util$1.getArg(aArgs, "line"),
        generatedColumn: util$1.getArg(aArgs, "column")
      };

      // Find the section containing the generated position we're trying to map
      // to an original position.
      const sectionIndex = binarySearch$1.search(needle, this._sections,
        function(aNeedle, section) {
          const cmp = aNeedle.generatedLine - section.generatedOffset.generatedLine;
          if (cmp) {
            return cmp;
          }

          return (aNeedle.generatedColumn -
                  section.generatedOffset.generatedColumn);
        });
      const section = this._sections[sectionIndex];

      if (!section) {
        return {
          source: null,
          line: null,
          column: null,
          name: null
        };
      }

      return section.consumer.originalPositionFor({
        line: needle.generatedLine -
          (section.generatedOffset.generatedLine - 1),
        column: needle.generatedColumn -
          (section.generatedOffset.generatedLine === needle.generatedLine
           ? section.generatedOffset.generatedColumn - 1
           : 0),
        bias: aArgs.bias
      });
    }

    /**
     * Return true if we have the source content for every source in the source
     * map, false otherwise.
     */
    hasContentsOfAllSources() {
      return this._sections.every(function(s) {
        return s.consumer.hasContentsOfAllSources();
      });
    }

    /**
     * Returns the original source content. The only argument is the url of the
     * original source file. Returns null if no original source content is
     * available.
     */
    sourceContentFor(aSource, nullOnMissing) {
      for (let i = 0; i < this._sections.length; i++) {
        const section = this._sections[i];

        const content = section.consumer.sourceContentFor(aSource, true);
        if (content) {
          return content;
        }
      }
      if (nullOnMissing) {
        return null;
      }
      throw new Error('"' + aSource + '" is not in the SourceMap.');
    }

    /**
     * Returns the generated line and column information for the original source,
     * line, and column positions provided. The only argument is an object with
     * the following properties:
     *
     *   - source: The filename of the original source.
     *   - line: The line number in the original source.  The line number
     *     is 1-based.
     *   - column: The column number in the original source.  The column
     *     number is 0-based.
     *
     * and an object is returned with the following properties:
     *
     *   - line: The line number in the generated source, or null.  The
     *     line number is 1-based.
     *   - column: The column number in the generated source, or null.
     *     The column number is 0-based.
     */
    generatedPositionFor(aArgs) {
      for (let i = 0; i < this._sections.length; i++) {
        const section = this._sections[i];

        // Only consider this section if the requested source is in the list of
        // sources of the consumer.
        if (section.consumer._findSourceIndex(util$1.getArg(aArgs, "source")) === -1) {
          continue;
        }
        const generatedPosition = section.consumer.generatedPositionFor(aArgs);
        if (generatedPosition) {
          const ret = {
            line: generatedPosition.line +
              (section.generatedOffset.generatedLine - 1),
            column: generatedPosition.column +
              (section.generatedOffset.generatedLine === generatedPosition.line
               ? section.generatedOffset.generatedColumn - 1
               : 0)
          };
          return ret;
        }
      }

      return {
        line: null,
        column: null
      };
    }

    /**
     * Parse the mappings in a string in to a data structure which we can easily
     * query (the ordered arrays in the `this.__generatedMappings` and
     * `this.__originalMappings` properties).
     */
    _parseMappings(aStr, aSourceRoot) {
      const generatedMappings = this.__generatedMappingsUnsorted = [];
      const originalMappings = this.__originalMappingsUnsorted = [];
      for (let i = 0; i < this._sections.length; i++) {
        const section = this._sections[i];

        const sectionMappings = [];
        section.consumer.eachMapping(m => sectionMappings.push(m));

        for (let j = 0; j < sectionMappings.length; j++) {
          const mapping = sectionMappings[j];

          // TODO: test if null is correct here.  The original code used
          // `source`, which would actually have gotten used as null because
          // var's get hoisted.
          // See: https://github.com/mozilla/source-map/issues/333
          let source = util$1.computeSourceURL(section.consumer.sourceRoot, null, this._sourceMapURL);
          this._sources.add(source);
          source = this._sources.indexOf(source);

          let name = null;
          if (mapping.name) {
            this._names.add(mapping.name);
            name = this._names.indexOf(mapping.name);
          }

          // The mappings coming from the consumer for the section have
          // generated positions relative to the start of the section, so we
          // need to offset them to be relative to the start of the concatenated
          // generated file.
          const adjustedMapping = {
            source,
            generatedLine: mapping.generatedLine +
              (section.generatedOffset.generatedLine - 1),
            generatedColumn: mapping.generatedColumn +
              (section.generatedOffset.generatedLine === mapping.generatedLine
              ? section.generatedOffset.generatedColumn - 1
              : 0),
            originalLine: mapping.originalLine,
            originalColumn: mapping.originalColumn,
            name
          };

          generatedMappings.push(adjustedMapping);
          if (typeof adjustedMapping.originalLine === "number") {
            originalMappings.push(adjustedMapping);
          }
        }
      }
    }

    eachMapping(aCallback, aContext, aOrder) {
      const context = aContext || null;
      const order = aOrder || SourceMapConsumer$2.GENERATED_ORDER;

      let mappings;
      switch (order) {
      case SourceMapConsumer$2.GENERATED_ORDER:
        mappings = this._generatedMappings;
        break;
      case SourceMapConsumer$2.ORIGINAL_ORDER:
        mappings = this._originalMappings;
        break;
      default:
        throw new Error("Unknown order of iteration.");
      }

      const sourceRoot = this.sourceRoot;
      mappings.map(function(mapping) {
        let source = null;
        if (mapping.source !== null) {
          source = this._sources.at(mapping.source);
          source = util$1.computeSourceURL(sourceRoot, source, this._sourceMapURL);
        }
        return {
          source,
          generatedLine: mapping.generatedLine,
          generatedColumn: mapping.generatedColumn,
          originalLine: mapping.originalLine,
          originalColumn: mapping.originalColumn,
          name: mapping.name === null ? null : this._names.at(mapping.name)
        };
      }, this).forEach(aCallback, context);
    }

    /**
     * Find the mapping that best matches the hypothetical "needle" mapping that
     * we are searching for in the given "haystack" of mappings.
     */
    _findMapping(aNeedle, aMappings, aLineName,
                aColumnName, aComparator, aBias) {
      // To return the position we are searching for, we must first find the
      // mapping for the given position and then return the opposite position it
      // points to. Because the mappings are sorted, we can use binary search to
      // find the best mapping.

      if (aNeedle[aLineName] <= 0) {
        throw new TypeError("Line must be greater than or equal to 1, got "
                            + aNeedle[aLineName]);
      }
      if (aNeedle[aColumnName] < 0) {
        throw new TypeError("Column must be greater than or equal to 0, got "
                            + aNeedle[aColumnName]);
      }

      return binarySearch$1.search(aNeedle, aMappings, aComparator, aBias);
    }

    allGeneratedPositionsFor(aArgs) {
      const line = util$1.getArg(aArgs, "line");

      // When there is no exact match, BasicSourceMapConsumer.prototype._findMapping
      // returns the index of the closest mapping less than the needle. By
      // setting needle.originalColumn to 0, we thus find the last mapping for
      // the given line, provided such a mapping exists.
      const needle = {
        source: util$1.getArg(aArgs, "source"),
        originalLine: line,
        originalColumn: util$1.getArg(aArgs, "column", 0)
      };

      needle.source = this._findSourceIndex(needle.source);
      if (needle.source < 0) {
        return [];
      }

      if (needle.originalLine < 1) {
        throw new Error("Line numbers must be >= 1");
      }

      if (needle.originalColumn < 0) {
        throw new Error("Column numbers must be >= 0");
      }

      const mappings = [];

      let index = this._findMapping(needle,
                                    this._originalMappings,
                                    "originalLine",
                                    "originalColumn",
                                    util$1.compareByOriginalPositions,
                                    binarySearch$1.LEAST_UPPER_BOUND);
      if (index >= 0) {
        let mapping = this._originalMappings[index];

        if (aArgs.column === undefined) {
          const originalLine = mapping.originalLine;

          // Iterate until either we run out of mappings, or we run into
          // a mapping for a different line than the one we found. Since
          // mappings are sorted, this is guaranteed to find all mappings for
          // the line we found.
          while (mapping && mapping.originalLine === originalLine) {
            let lastColumn = mapping.lastGeneratedColumn;
            if (this._computedColumnSpans && lastColumn === null) {
              lastColumn = Infinity;
            }
            mappings.push({
              line: util$1.getArg(mapping, "generatedLine", null),
              column: util$1.getArg(mapping, "generatedColumn", null),
              lastColumn,
            });

            mapping = this._originalMappings[++index];
          }
        } else {
          const originalColumn = mapping.originalColumn;

          // Iterate until either we run out of mappings, or we run into
          // a mapping for a different line than the one we were searching for.
          // Since mappings are sorted, this is guaranteed to find all mappings for
          // the line we are searching for.
          while (mapping &&
                 mapping.originalLine === line &&
                 mapping.originalColumn == originalColumn) {
            let lastColumn = mapping.lastGeneratedColumn;
            if (this._computedColumnSpans && lastColumn === null) {
              lastColumn = Infinity;
            }
            mappings.push({
              line: util$1.getArg(mapping, "generatedLine", null),
              column: util$1.getArg(mapping, "generatedColumn", null),
              lastColumn,
            });

            mapping = this._originalMappings[++index];
          }
        }
      }

      return mappings;
    }

    destroy() {
      for (let i = 0; i < this._sections.length; i++) {
        this._sections[i].consumer.destroy();
      }
    }
  }

  /*
   * Cheat to get around inter-twingled classes.  `factory()` can be at the end
   * where it has access to non-hoisted classes, but it gets hoisted itself.
   */
  function _factory(aSourceMap, aSourceMapURL) {
    let sourceMap = aSourceMap;
    if (typeof aSourceMap === "string") {
      sourceMap = util$1.parseSourceMapInput(aSourceMap);
    }

    const consumer = sourceMap.sections != null
        ? new IndexedSourceMapConsumer$1(sourceMap, aSourceMapURL)
        : new BasicSourceMapConsumer$1(sourceMap, aSourceMapURL);
    return Promise.resolve(consumer);
  }

  function _factoryBSM(aSourceMap, aSourceMapURL) {
    return BasicSourceMapConsumer$1.fromSourceMap(aSourceMap, aSourceMapURL);
  }

  /*
   * Copyright 2009-2011 Mozilla Foundation and contributors
   * Licensed under the New BSD license. See LICENSE.txt or:
   * http://opensource.org/licenses/BSD-3-Clause
   */
  var SourceMapGenerator$5 = sourceMapGenerator$1.SourceMapGenerator;

  var esprima$1 = createCommonjsModule(function (module, exports) {
  (function webpackUniversalModuleDefinition(root, factory) {
  /* istanbul ignore next */
  	module.exports = factory();
  })(commonjsGlobal, function() {
  return /******/ (function(modules) { // webpackBootstrap
  /******/ 	// The module cache
  /******/ 	var installedModules = {};

  /******/ 	// The require function
  /******/ 	function __webpack_require__(moduleId) {

  /******/ 		// Check if module is in cache
  /* istanbul ignore if */
  /******/ 		if(installedModules[moduleId])
  /******/ 			return installedModules[moduleId].exports;

  /******/ 		// Create a new module (and put it into the cache)
  /******/ 		var module = installedModules[moduleId] = {
  /******/ 			exports: {},
  /******/ 			id: moduleId,
  /******/ 			loaded: false
  /******/ 		};

  /******/ 		// Execute the module function
  /******/ 		modules[moduleId].call(module.exports, module, module.exports, __webpack_require__);

  /******/ 		// Flag the module as loaded
  /******/ 		module.loaded = true;

  /******/ 		// Return the exports of the module
  /******/ 		return module.exports;
  /******/ 	}


  /******/ 	// expose the modules object (__webpack_modules__)
  /******/ 	__webpack_require__.m = modules;

  /******/ 	// expose the module cache
  /******/ 	__webpack_require__.c = installedModules;

  /******/ 	// __webpack_public_path__
  /******/ 	__webpack_require__.p = "";

  /******/ 	// Load entry module and return exports
  /******/ 	return __webpack_require__(0);
  /******/ })
  /************************************************************************/
  /******/ ([
  /* 0 */
  /***/ function(module, exports, __webpack_require__) {
  	/*
  	  Copyright JS Foundation and other contributors, https://js.foundation/

  	  Redistribution and use in source and binary forms, with or without
  	  modification, are permitted provided that the following conditions are met:

  	    * Redistributions of source code must retain the above copyright
  	      notice, this list of conditions and the following disclaimer.
  	    * Redistributions in binary form must reproduce the above copyright
  	      notice, this list of conditions and the following disclaimer in the
  	      documentation and/or other materials provided with the distribution.

  	  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  	  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  	  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  	  ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
  	  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
  	  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  	  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  	  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  	  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
  	  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  	*/
  	Object.defineProperty(exports, "__esModule", { value: true });
  	var comment_handler_1 = __webpack_require__(1);
  	var jsx_parser_1 = __webpack_require__(3);
  	var parser_1 = __webpack_require__(8);
  	var tokenizer_1 = __webpack_require__(15);
  	function parse(code, options, delegate) {
  	    var commentHandler = null;
  	    var proxyDelegate = function (node, metadata) {
  	        if (delegate) {
  	            delegate(node, metadata);
  	        }
  	        if (commentHandler) {
  	            commentHandler.visit(node, metadata);
  	        }
  	    };
  	    var parserDelegate = (typeof delegate === 'function') ? proxyDelegate : null;
  	    var collectComment = false;
  	    if (options) {
  	        collectComment = (typeof options.comment === 'boolean' && options.comment);
  	        var attachComment = (typeof options.attachComment === 'boolean' && options.attachComment);
  	        if (collectComment || attachComment) {
  	            commentHandler = new comment_handler_1.CommentHandler();
  	            commentHandler.attach = attachComment;
  	            options.comment = true;
  	            parserDelegate = proxyDelegate;
  	        }
  	    }
  	    var isModule = false;
  	    if (options && typeof options.sourceType === 'string') {
  	        isModule = (options.sourceType === 'module');
  	    }
  	    var parser;
  	    if (options && typeof options.jsx === 'boolean' && options.jsx) {
  	        parser = new jsx_parser_1.JSXParser(code, options, parserDelegate);
  	    }
  	    else {
  	        parser = new parser_1.Parser(code, options, parserDelegate);
  	    }
  	    var program = isModule ? parser.parseModule() : parser.parseScript();
  	    var ast = program;
  	    if (collectComment && commentHandler) {
  	        ast.comments = commentHandler.comments;
  	    }
  	    if (parser.config.tokens) {
  	        ast.tokens = parser.tokens;
  	    }
  	    if (parser.config.tolerant) {
  	        ast.errors = parser.errorHandler.errors;
  	    }
  	    return ast;
  	}
  	exports.parse = parse;
  	function parseModule(code, options, delegate) {
  	    var parsingOptions = options || {};
  	    parsingOptions.sourceType = 'module';
  	    return parse(code, parsingOptions, delegate);
  	}
  	exports.parseModule = parseModule;
  	function parseScript(code, options, delegate) {
  	    var parsingOptions = options || {};
  	    parsingOptions.sourceType = 'script';
  	    return parse(code, parsingOptions, delegate);
  	}
  	exports.parseScript = parseScript;
  	function tokenize(code, options, delegate) {
  	    var tokenizer = new tokenizer_1.Tokenizer(code, options);
  	    var tokens;
  	    tokens = [];
  	    try {
  	        while (true) {
  	            var token = tokenizer.getNextToken();
  	            if (!token) {
  	                break;
  	            }
  	            if (delegate) {
  	                token = delegate(token);
  	            }
  	            tokens.push(token);
  	        }
  	    }
  	    catch (e) {
  	        tokenizer.errorHandler.tolerate(e);
  	    }
  	    if (tokenizer.errorHandler.tolerant) {
  	        tokens.errors = tokenizer.errors();
  	    }
  	    return tokens;
  	}
  	exports.tokenize = tokenize;
  	var syntax_1 = __webpack_require__(2);
  	exports.Syntax = syntax_1.Syntax;
  	// Sync with *.json manifests.
  	exports.version = '4.0.0';


  /***/ },
  /* 1 */
  /***/ function(module, exports, __webpack_require__) {
  	Object.defineProperty(exports, "__esModule", { value: true });
  	var syntax_1 = __webpack_require__(2);
  	var CommentHandler = (function () {
  	    function CommentHandler() {
  	        this.attach = false;
  	        this.comments = [];
  	        this.stack = [];
  	        this.leading = [];
  	        this.trailing = [];
  	    }
  	    CommentHandler.prototype.insertInnerComments = function (node, metadata) {
  	        //  innnerComments for properties empty block
  	        //  `function a() {/** comments **\/}`
  	        if (node.type === syntax_1.Syntax.BlockStatement && node.body.length === 0) {
  	            var innerComments = [];
  	            for (var i = this.leading.length - 1; i >= 0; --i) {
  	                var entry = this.leading[i];
  	                if (metadata.end.offset >= entry.start) {
  	                    innerComments.unshift(entry.comment);
  	                    this.leading.splice(i, 1);
  	                    this.trailing.splice(i, 1);
  	                }
  	            }
  	            if (innerComments.length) {
  	                node.innerComments = innerComments;
  	            }
  	        }
  	    };
  	    CommentHandler.prototype.findTrailingComments = function (metadata) {
  	        var trailingComments = [];
  	        if (this.trailing.length > 0) {
  	            for (var i = this.trailing.length - 1; i >= 0; --i) {
  	                var entry_1 = this.trailing[i];
  	                if (entry_1.start >= metadata.end.offset) {
  	                    trailingComments.unshift(entry_1.comment);
  	                }
  	            }
  	            this.trailing.length = 0;
  	            return trailingComments;
  	        }
  	        var entry = this.stack[this.stack.length - 1];
  	        if (entry && entry.node.trailingComments) {
  	            var firstComment = entry.node.trailingComments[0];
  	            if (firstComment && firstComment.range[0] >= metadata.end.offset) {
  	                trailingComments = entry.node.trailingComments;
  	                delete entry.node.trailingComments;
  	            }
  	        }
  	        return trailingComments;
  	    };
  	    CommentHandler.prototype.findLeadingComments = function (metadata) {
  	        var leadingComments = [];
  	        var target;
  	        while (this.stack.length > 0) {
  	            var entry = this.stack[this.stack.length - 1];
  	            if (entry && entry.start >= metadata.start.offset) {
  	                target = entry.node;
  	                this.stack.pop();
  	            }
  	            else {
  	                break;
  	            }
  	        }
  	        if (target) {
  	            var count = target.leadingComments ? target.leadingComments.length : 0;
  	            for (var i = count - 1; i >= 0; --i) {
  	                var comment = target.leadingComments[i];
  	                if (comment.range[1] <= metadata.start.offset) {
  	                    leadingComments.unshift(comment);
  	                    target.leadingComments.splice(i, 1);
  	                }
  	            }
  	            if (target.leadingComments && target.leadingComments.length === 0) {
  	                delete target.leadingComments;
  	            }
  	            return leadingComments;
  	        }
  	        for (var i = this.leading.length - 1; i >= 0; --i) {
  	            var entry = this.leading[i];
  	            if (entry.start <= metadata.start.offset) {
  	                leadingComments.unshift(entry.comment);
  	                this.leading.splice(i, 1);
  	            }
  	        }
  	        return leadingComments;
  	    };
  	    CommentHandler.prototype.visitNode = function (node, metadata) {
  	        if (node.type === syntax_1.Syntax.Program && node.body.length > 0) {
  	            return;
  	        }
  	        this.insertInnerComments(node, metadata);
  	        var trailingComments = this.findTrailingComments(metadata);
  	        var leadingComments = this.findLeadingComments(metadata);
  	        if (leadingComments.length > 0) {
  	            node.leadingComments = leadingComments;
  	        }
  	        if (trailingComments.length > 0) {
  	            node.trailingComments = trailingComments;
  	        }
  	        this.stack.push({
  	            node: node,
  	            start: metadata.start.offset
  	        });
  	    };
  	    CommentHandler.prototype.visitComment = function (node, metadata) {
  	        var type = (node.type[0] === 'L') ? 'Line' : 'Block';
  	        var comment = {
  	            type: type,
  	            value: node.value
  	        };
  	        if (node.range) {
  	            comment.range = node.range;
  	        }
  	        if (node.loc) {
  	            comment.loc = node.loc;
  	        }
  	        this.comments.push(comment);
  	        if (this.attach) {
  	            var entry = {
  	                comment: {
  	                    type: type,
  	                    value: node.value,
  	                    range: [metadata.start.offset, metadata.end.offset]
  	                },
  	                start: metadata.start.offset
  	            };
  	            if (node.loc) {
  	                entry.comment.loc = node.loc;
  	            }
  	            node.type = type;
  	            this.leading.push(entry);
  	            this.trailing.push(entry);
  	        }
  	    };
  	    CommentHandler.prototype.visit = function (node, metadata) {
  	        if (node.type === 'LineComment') {
  	            this.visitComment(node, metadata);
  	        }
  	        else if (node.type === 'BlockComment') {
  	            this.visitComment(node, metadata);
  	        }
  	        else if (this.attach) {
  	            this.visitNode(node, metadata);
  	        }
  	    };
  	    return CommentHandler;
  	}());
  	exports.CommentHandler = CommentHandler;


  /***/ },
  /* 2 */
  /***/ function(module, exports) {
  	Object.defineProperty(exports, "__esModule", { value: true });
  	exports.Syntax = {
  	    AssignmentExpression: 'AssignmentExpression',
  	    AssignmentPattern: 'AssignmentPattern',
  	    ArrayExpression: 'ArrayExpression',
  	    ArrayPattern: 'ArrayPattern',
  	    ArrowFunctionExpression: 'ArrowFunctionExpression',
  	    AwaitExpression: 'AwaitExpression',
  	    BlockStatement: 'BlockStatement',
  	    BinaryExpression: 'BinaryExpression',
  	    BreakStatement: 'BreakStatement',
  	    CallExpression: 'CallExpression',
  	    CatchClause: 'CatchClause',
  	    ClassBody: 'ClassBody',
  	    ClassDeclaration: 'ClassDeclaration',
  	    ClassExpression: 'ClassExpression',
  	    ConditionalExpression: 'ConditionalExpression',
  	    ContinueStatement: 'ContinueStatement',
  	    DoWhileStatement: 'DoWhileStatement',
  	    DebuggerStatement: 'DebuggerStatement',
  	    EmptyStatement: 'EmptyStatement',
  	    ExportAllDeclaration: 'ExportAllDeclaration',
  	    ExportDefaultDeclaration: 'ExportDefaultDeclaration',
  	    ExportNamedDeclaration: 'ExportNamedDeclaration',
  	    ExportSpecifier: 'ExportSpecifier',
  	    ExpressionStatement: 'ExpressionStatement',
  	    ForStatement: 'ForStatement',
  	    ForOfStatement: 'ForOfStatement',
  	    ForInStatement: 'ForInStatement',
  	    FunctionDeclaration: 'FunctionDeclaration',
  	    FunctionExpression: 'FunctionExpression',
  	    Identifier: 'Identifier',
  	    IfStatement: 'IfStatement',
  	    ImportDeclaration: 'ImportDeclaration',
  	    ImportDefaultSpecifier: 'ImportDefaultSpecifier',
  	    ImportNamespaceSpecifier: 'ImportNamespaceSpecifier',
  	    ImportSpecifier: 'ImportSpecifier',
  	    Literal: 'Literal',
  	    LabeledStatement: 'LabeledStatement',
  	    LogicalExpression: 'LogicalExpression',
  	    MemberExpression: 'MemberExpression',
  	    MetaProperty: 'MetaProperty',
  	    MethodDefinition: 'MethodDefinition',
  	    NewExpression: 'NewExpression',
  	    ObjectExpression: 'ObjectExpression',
  	    ObjectPattern: 'ObjectPattern',
  	    Program: 'Program',
  	    Property: 'Property',
  	    RestElement: 'RestElement',
  	    ReturnStatement: 'ReturnStatement',
  	    SequenceExpression: 'SequenceExpression',
  	    SpreadElement: 'SpreadElement',
  	    Super: 'Super',
  	    SwitchCase: 'SwitchCase',
  	    SwitchStatement: 'SwitchStatement',
  	    TaggedTemplateExpression: 'TaggedTemplateExpression',
  	    TemplateElement: 'TemplateElement',
  	    TemplateLiteral: 'TemplateLiteral',
  	    ThisExpression: 'ThisExpression',
  	    ThrowStatement: 'ThrowStatement',
  	    TryStatement: 'TryStatement',
  	    UnaryExpression: 'UnaryExpression',
  	    UpdateExpression: 'UpdateExpression',
  	    VariableDeclaration: 'VariableDeclaration',
  	    VariableDeclarator: 'VariableDeclarator',
  	    WhileStatement: 'WhileStatement',
  	    WithStatement: 'WithStatement',
  	    YieldExpression: 'YieldExpression'
  	};


  /***/ },
  /* 3 */
  /***/ function(module, exports, __webpack_require__) {
  /* istanbul ignore next */
  	var __extends = (this && this.__extends) || (function () {
  	    var extendStatics = Object.setPrototypeOf ||
  	        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
  	        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
  	    return function (d, b) {
  	        extendStatics(d, b);
  	        function __() { this.constructor = d; }
  	        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
  	    };
  	})();
  	Object.defineProperty(exports, "__esModule", { value: true });
  	var character_1 = __webpack_require__(4);
  	var JSXNode = __webpack_require__(5);
  	var jsx_syntax_1 = __webpack_require__(6);
  	var Node = __webpack_require__(7);
  	var parser_1 = __webpack_require__(8);
  	var token_1 = __webpack_require__(13);
  	var xhtml_entities_1 = __webpack_require__(14);
  	token_1.TokenName[100 /* Identifier */] = 'JSXIdentifier';
  	token_1.TokenName[101 /* Text */] = 'JSXText';
  	// Fully qualified element name, e.g. <svg:path> returns "svg:path"
  	function getQualifiedElementName(elementName) {
  	    var qualifiedName;
  	    switch (elementName.type) {
  	        case jsx_syntax_1.JSXSyntax.JSXIdentifier:
  	            var id = elementName;
  	            qualifiedName = id.name;
  	            break;
  	        case jsx_syntax_1.JSXSyntax.JSXNamespacedName:
  	            var ns = elementName;
  	            qualifiedName = getQualifiedElementName(ns.namespace) + ':' +
  	                getQualifiedElementName(ns.name);
  	            break;
  	        case jsx_syntax_1.JSXSyntax.JSXMemberExpression:
  	            var expr = elementName;
  	            qualifiedName = getQualifiedElementName(expr.object) + '.' +
  	                getQualifiedElementName(expr.property);
  	            break;
  	        /* istanbul ignore next */
  	        default:
  	            break;
  	    }
  	    return qualifiedName;
  	}
  	var JSXParser = (function (_super) {
  	    __extends(JSXParser, _super);
  	    function JSXParser(code, options, delegate) {
  	        return _super.call(this, code, options, delegate) || this;
  	    }
  	    JSXParser.prototype.parsePrimaryExpression = function () {
  	        return this.match('<') ? this.parseJSXRoot() : _super.prototype.parsePrimaryExpression.call(this);
  	    };
  	    JSXParser.prototype.startJSX = function () {
  	        // Unwind the scanner before the lookahead token.
  	        this.scanner.index = this.startMarker.index;
  	        this.scanner.lineNumber = this.startMarker.line;
  	        this.scanner.lineStart = this.startMarker.index - this.startMarker.column;
  	    };
  	    JSXParser.prototype.finishJSX = function () {
  	        // Prime the next lookahead.
  	        this.nextToken();
  	    };
  	    JSXParser.prototype.reenterJSX = function () {
  	        this.startJSX();
  	        this.expectJSX('}');
  	        // Pop the closing '}' added from the lookahead.
  	        if (this.config.tokens) {
  	            this.tokens.pop();
  	        }
  	    };
  	    JSXParser.prototype.createJSXNode = function () {
  	        this.collectComments();
  	        return {
  	            index: this.scanner.index,
  	            line: this.scanner.lineNumber,
  	            column: this.scanner.index - this.scanner.lineStart
  	        };
  	    };
  	    JSXParser.prototype.createJSXChildNode = function () {
  	        return {
  	            index: this.scanner.index,
  	            line: this.scanner.lineNumber,
  	            column: this.scanner.index - this.scanner.lineStart
  	        };
  	    };
  	    JSXParser.prototype.scanXHTMLEntity = function (quote) {
  	        var result = '&';
  	        var valid = true;
  	        var terminated = false;
  	        var numeric = false;
  	        var hex = false;
  	        while (!this.scanner.eof() && valid && !terminated) {
  	            var ch = this.scanner.source[this.scanner.index];
  	            if (ch === quote) {
  	                break;
  	            }
  	            terminated = (ch === ';');
  	            result += ch;
  	            ++this.scanner.index;
  	            if (!terminated) {
  	                switch (result.length) {
  	                    case 2:
  	                        // e.g. '&#123;'
  	                        numeric = (ch === '#');
  	                        break;
  	                    case 3:
  	                        if (numeric) {
  	                            // e.g. '&#x41;'
  	                            hex = (ch === 'x');
  	                            valid = hex || character_1.Character.isDecimalDigit(ch.charCodeAt(0));
  	                            numeric = numeric && !hex;
  	                        }
  	                        break;
  	                    default:
  	                        valid = valid && !(numeric && !character_1.Character.isDecimalDigit(ch.charCodeAt(0)));
  	                        valid = valid && !(hex && !character_1.Character.isHexDigit(ch.charCodeAt(0)));
  	                        break;
  	                }
  	            }
  	        }
  	        if (valid && terminated && result.length > 2) {
  	            // e.g. '&#x41;' becomes just '#x41'
  	            var str = result.substr(1, result.length - 2);
  	            if (numeric && str.length > 1) {
  	                result = String.fromCharCode(parseInt(str.substr(1), 10));
  	            }
  	            else if (hex && str.length > 2) {
  	                result = String.fromCharCode(parseInt('0' + str.substr(1), 16));
  	            }
  	            else if (!numeric && !hex && xhtml_entities_1.XHTMLEntities[str]) {
  	                result = xhtml_entities_1.XHTMLEntities[str];
  	            }
  	        }
  	        return result;
  	    };
  	    // Scan the next JSX token. This replaces Scanner#lex when in JSX mode.
  	    JSXParser.prototype.lexJSX = function () {
  	        var cp = this.scanner.source.charCodeAt(this.scanner.index);
  	        // < > / : = { }
  	        if (cp === 60 || cp === 62 || cp === 47 || cp === 58 || cp === 61 || cp === 123 || cp === 125) {
  	            var value = this.scanner.source[this.scanner.index++];
  	            return {
  	                type: 7 /* Punctuator */,
  	                value: value,
  	                lineNumber: this.scanner.lineNumber,
  	                lineStart: this.scanner.lineStart,
  	                start: this.scanner.index - 1,
  	                end: this.scanner.index
  	            };
  	        }
  	        // " '
  	        if (cp === 34 || cp === 39) {
  	            var start = this.scanner.index;
  	            var quote = this.scanner.source[this.scanner.index++];
  	            var str = '';
  	            while (!this.scanner.eof()) {
  	                var ch = this.scanner.source[this.scanner.index++];
  	                if (ch === quote) {
  	                    break;
  	                }
  	                else if (ch === '&') {
  	                    str += this.scanXHTMLEntity(quote);
  	                }
  	                else {
  	                    str += ch;
  	                }
  	            }
  	            return {
  	                type: 8 /* StringLiteral */,
  	                value: str,
  	                lineNumber: this.scanner.lineNumber,
  	                lineStart: this.scanner.lineStart,
  	                start: start,
  	                end: this.scanner.index
  	            };
  	        }
  	        // ... or .
  	        if (cp === 46) {
  	            var n1 = this.scanner.source.charCodeAt(this.scanner.index + 1);
  	            var n2 = this.scanner.source.charCodeAt(this.scanner.index + 2);
  	            var value = (n1 === 46 && n2 === 46) ? '...' : '.';
  	            var start = this.scanner.index;
  	            this.scanner.index += value.length;
  	            return {
  	                type: 7 /* Punctuator */,
  	                value: value,
  	                lineNumber: this.scanner.lineNumber,
  	                lineStart: this.scanner.lineStart,
  	                start: start,
  	                end: this.scanner.index
  	            };
  	        }
  	        // `
  	        if (cp === 96) {
  	            // Only placeholder, since it will be rescanned as a real assignment expression.
  	            return {
  	                type: 10 /* Template */,
  	                value: '',
  	                lineNumber: this.scanner.lineNumber,
  	                lineStart: this.scanner.lineStart,
  	                start: this.scanner.index,
  	                end: this.scanner.index
  	            };
  	        }
  	        // Identifer can not contain backslash (char code 92).
  	        if (character_1.Character.isIdentifierStart(cp) && (cp !== 92)) {
  	            var start = this.scanner.index;
  	            ++this.scanner.index;
  	            while (!this.scanner.eof()) {
  	                var ch = this.scanner.source.charCodeAt(this.scanner.index);
  	                if (character_1.Character.isIdentifierPart(ch) && (ch !== 92)) {
  	                    ++this.scanner.index;
  	                }
  	                else if (ch === 45) {
  	                    // Hyphen (char code 45) can be part of an identifier.
  	                    ++this.scanner.index;
  	                }
  	                else {
  	                    break;
  	                }
  	            }
  	            var id = this.scanner.source.slice(start, this.scanner.index);
  	            return {
  	                type: 100 /* Identifier */,
  	                value: id,
  	                lineNumber: this.scanner.lineNumber,
  	                lineStart: this.scanner.lineStart,
  	                start: start,
  	                end: this.scanner.index
  	            };
  	        }
  	        return this.scanner.lex();
  	    };
  	    JSXParser.prototype.nextJSXToken = function () {
  	        this.collectComments();
  	        this.startMarker.index = this.scanner.index;
  	        this.startMarker.line = this.scanner.lineNumber;
  	        this.startMarker.column = this.scanner.index - this.scanner.lineStart;
  	        var token = this.lexJSX();
  	        this.lastMarker.index = this.scanner.index;
  	        this.lastMarker.line = this.scanner.lineNumber;
  	        this.lastMarker.column = this.scanner.index - this.scanner.lineStart;
  	        if (this.config.tokens) {
  	            this.tokens.push(this.convertToken(token));
  	        }
  	        return token;
  	    };
  	    JSXParser.prototype.nextJSXText = function () {
  	        this.startMarker.index = this.scanner.index;
  	        this.startMarker.line = this.scanner.lineNumber;
  	        this.startMarker.column = this.scanner.index - this.scanner.lineStart;
  	        var start = this.scanner.index;
  	        var text = '';
  	        while (!this.scanner.eof()) {
  	            var ch = this.scanner.source[this.scanner.index];
  	            if (ch === '{' || ch === '<') {
  	                break;
  	            }
  	            ++this.scanner.index;
  	            text += ch;
  	            if (character_1.Character.isLineTerminator(ch.charCodeAt(0))) {
  	                ++this.scanner.lineNumber;
  	                if (ch === '\r' && this.scanner.source[this.scanner.index] === '\n') {
  	                    ++this.scanner.index;
  	                }
  	                this.scanner.lineStart = this.scanner.index;
  	            }
  	        }
  	        this.lastMarker.index = this.scanner.index;
  	        this.lastMarker.line = this.scanner.lineNumber;
  	        this.lastMarker.column = this.scanner.index - this.scanner.lineStart;
  	        var token = {
  	            type: 101 /* Text */,
  	            value: text,
  	            lineNumber: this.scanner.lineNumber,
  	            lineStart: this.scanner.lineStart,
  	            start: start,
  	            end: this.scanner.index
  	        };
  	        if ((text.length > 0) && this.config.tokens) {
  	            this.tokens.push(this.convertToken(token));
  	        }
  	        return token;
  	    };
  	    JSXParser.prototype.peekJSXToken = function () {
  	        var state = this.scanner.saveState();
  	        this.scanner.scanComments();
  	        var next = this.lexJSX();
  	        this.scanner.restoreState(state);
  	        return next;
  	    };
  	    // Expect the next JSX token to match the specified punctuator.
  	    // If not, an exception will be thrown.
  	    JSXParser.prototype.expectJSX = function (value) {
  	        var token = this.nextJSXToken();
  	        if (token.type !== 7 /* Punctuator */ || token.value !== value) {
  	            this.throwUnexpectedToken(token);
  	        }
  	    };
  	    // Return true if the next JSX token matches the specified punctuator.
  	    JSXParser.prototype.matchJSX = function (value) {
  	        var next = this.peekJSXToken();
  	        return next.type === 7 /* Punctuator */ && next.value === value;
  	    };
  	    JSXParser.prototype.parseJSXIdentifier = function () {
  	        var node = this.createJSXNode();
  	        var token = this.nextJSXToken();
  	        if (token.type !== 100 /* Identifier */) {
  	            this.throwUnexpectedToken(token);
  	        }
  	        return this.finalize(node, new JSXNode.JSXIdentifier(token.value));
  	    };
  	    JSXParser.prototype.parseJSXElementName = function () {
  	        var node = this.createJSXNode();
  	        var elementName = this.parseJSXIdentifier();
  	        if (this.matchJSX(':')) {
  	            var namespace = elementName;
  	            this.expectJSX(':');
  	            var name_1 = this.parseJSXIdentifier();
  	            elementName = this.finalize(node, new JSXNode.JSXNamespacedName(namespace, name_1));
  	        }
  	        else if (this.matchJSX('.')) {
  	            while (this.matchJSX('.')) {
  	                var object = elementName;
  	                this.expectJSX('.');
  	                var property = this.parseJSXIdentifier();
  	                elementName = this.finalize(node, new JSXNode.JSXMemberExpression(object, property));
  	            }
  	        }
  	        return elementName;
  	    };
  	    JSXParser.prototype.parseJSXAttributeName = function () {
  	        var node = this.createJSXNode();
  	        var attributeName;
  	        var identifier = this.parseJSXIdentifier();
  	        if (this.matchJSX(':')) {
  	            var namespace = identifier;
  	            this.expectJSX(':');
  	            var name_2 = this.parseJSXIdentifier();
  	            attributeName = this.finalize(node, new JSXNode.JSXNamespacedName(namespace, name_2));
  	        }
  	        else {
  	            attributeName = identifier;
  	        }
  	        return attributeName;
  	    };
  	    JSXParser.prototype.parseJSXStringLiteralAttribute = function () {
  	        var node = this.createJSXNode();
  	        var token = this.nextJSXToken();
  	        if (token.type !== 8 /* StringLiteral */) {
  	            this.throwUnexpectedToken(token);
  	        }
  	        var raw = this.getTokenRaw(token);
  	        return this.finalize(node, new Node.Literal(token.value, raw));
  	    };
  	    JSXParser.prototype.parseJSXExpressionAttribute = function () {
  	        var node = this.createJSXNode();
  	        this.expectJSX('{');
  	        this.finishJSX();
  	        if (this.match('}')) {
  	            this.tolerateError('JSX attributes must only be assigned a non-empty expression');
  	        }
  	        var expression = this.parseAssignmentExpression();
  	        this.reenterJSX();
  	        return this.finalize(node, new JSXNode.JSXExpressionContainer(expression));
  	    };
  	    JSXParser.prototype.parseJSXAttributeValue = function () {
  	        return this.matchJSX('{') ? this.parseJSXExpressionAttribute() :
  	            this.matchJSX('<') ? this.parseJSXElement() : this.parseJSXStringLiteralAttribute();
  	    };
  	    JSXParser.prototype.parseJSXNameValueAttribute = function () {
  	        var node = this.createJSXNode();
  	        var name = this.parseJSXAttributeName();
  	        var value = null;
  	        if (this.matchJSX('=')) {
  	            this.expectJSX('=');
  	            value = this.parseJSXAttributeValue();
  	        }
  	        return this.finalize(node, new JSXNode.JSXAttribute(name, value));
  	    };
  	    JSXParser.prototype.parseJSXSpreadAttribute = function () {
  	        var node = this.createJSXNode();
  	        this.expectJSX('{');
  	        this.expectJSX('...');
  	        this.finishJSX();
  	        var argument = this.parseAssignmentExpression();
  	        this.reenterJSX();
  	        return this.finalize(node, new JSXNode.JSXSpreadAttribute(argument));
  	    };
  	    JSXParser.prototype.parseJSXAttributes = function () {
  	        var attributes = [];
  	        while (!this.matchJSX('/') && !this.matchJSX('>')) {
  	            var attribute = this.matchJSX('{') ? this.parseJSXSpreadAttribute() :
  	                this.parseJSXNameValueAttribute();
  	            attributes.push(attribute);
  	        }
  	        return attributes;
  	    };
  	    JSXParser.prototype.parseJSXOpeningElement = function () {
  	        var node = this.createJSXNode();
  	        this.expectJSX('<');
  	        var name = this.parseJSXElementName();
  	        var attributes = this.parseJSXAttributes();
  	        var selfClosing = this.matchJSX('/');
  	        if (selfClosing) {
  	            this.expectJSX('/');
  	        }
  	        this.expectJSX('>');
  	        return this.finalize(node, new JSXNode.JSXOpeningElement(name, selfClosing, attributes));
  	    };
  	    JSXParser.prototype.parseJSXBoundaryElement = function () {
  	        var node = this.createJSXNode();
  	        this.expectJSX('<');
  	        if (this.matchJSX('/')) {
  	            this.expectJSX('/');
  	            var name_3 = this.parseJSXElementName();
  	            this.expectJSX('>');
  	            return this.finalize(node, new JSXNode.JSXClosingElement(name_3));
  	        }
  	        var name = this.parseJSXElementName();
  	        var attributes = this.parseJSXAttributes();
  	        var selfClosing = this.matchJSX('/');
  	        if (selfClosing) {
  	            this.expectJSX('/');
  	        }
  	        this.expectJSX('>');
  	        return this.finalize(node, new JSXNode.JSXOpeningElement(name, selfClosing, attributes));
  	    };
  	    JSXParser.prototype.parseJSXEmptyExpression = function () {
  	        var node = this.createJSXChildNode();
  	        this.collectComments();
  	        this.lastMarker.index = this.scanner.index;
  	        this.lastMarker.line = this.scanner.lineNumber;
  	        this.lastMarker.column = this.scanner.index - this.scanner.lineStart;
  	        return this.finalize(node, new JSXNode.JSXEmptyExpression());
  	    };
  	    JSXParser.prototype.parseJSXExpressionContainer = function () {
  	        var node = this.createJSXNode();
  	        this.expectJSX('{');
  	        var expression;
  	        if (this.matchJSX('}')) {
  	            expression = this.parseJSXEmptyExpression();
  	            this.expectJSX('}');
  	        }
  	        else {
  	            this.finishJSX();
  	            expression = this.parseAssignmentExpression();
  	            this.reenterJSX();
  	        }
  	        return this.finalize(node, new JSXNode.JSXExpressionContainer(expression));
  	    };
  	    JSXParser.prototype.parseJSXChildren = function () {
  	        var children = [];
  	        while (!this.scanner.eof()) {
  	            var node = this.createJSXChildNode();
  	            var token = this.nextJSXText();
  	            if (token.start < token.end) {
  	                var raw = this.getTokenRaw(token);
  	                var child = this.finalize(node, new JSXNode.JSXText(token.value, raw));
  	                children.push(child);
  	            }
  	            if (this.scanner.source[this.scanner.index] === '{') {
  	                var container = this.parseJSXExpressionContainer();
  	                children.push(container);
  	            }
  	            else {
  	                break;
  	            }
  	        }
  	        return children;
  	    };
  	    JSXParser.prototype.parseComplexJSXElement = function (el) {
  	        var stack = [];
  	        while (!this.scanner.eof()) {
  	            el.children = el.children.concat(this.parseJSXChildren());
  	            var node = this.createJSXChildNode();
  	            var element = this.parseJSXBoundaryElement();
  	            if (element.type === jsx_syntax_1.JSXSyntax.JSXOpeningElement) {
  	                var opening = element;
  	                if (opening.selfClosing) {
  	                    var child = this.finalize(node, new JSXNode.JSXElement(opening, [], null));
  	                    el.children.push(child);
  	                }
  	                else {
  	                    stack.push(el);
  	                    el = { node: node, opening: opening, closing: null, children: [] };
  	                }
  	            }
  	            if (element.type === jsx_syntax_1.JSXSyntax.JSXClosingElement) {
  	                el.closing = element;
  	                var open_1 = getQualifiedElementName(el.opening.name);
  	                var close_1 = getQualifiedElementName(el.closing.name);
  	                if (open_1 !== close_1) {
  	                    this.tolerateError('Expected corresponding JSX closing tag for %0', open_1);
  	                }
  	                if (stack.length > 0) {
  	                    var child = this.finalize(el.node, new JSXNode.JSXElement(el.opening, el.children, el.closing));
  	                    el = stack[stack.length - 1];
  	                    el.children.push(child);
  	                    stack.pop();
  	                }
  	                else {
  	                    break;
  	                }
  	            }
  	        }
  	        return el;
  	    };
  	    JSXParser.prototype.parseJSXElement = function () {
  	        var node = this.createJSXNode();
  	        var opening = this.parseJSXOpeningElement();
  	        var children = [];
  	        var closing = null;
  	        if (!opening.selfClosing) {
  	            var el = this.parseComplexJSXElement({ node: node, opening: opening, closing: closing, children: children });
  	            children = el.children;
  	            closing = el.closing;
  	        }
  	        return this.finalize(node, new JSXNode.JSXElement(opening, children, closing));
  	    };
  	    JSXParser.prototype.parseJSXRoot = function () {
  	        // Pop the opening '<' added from the lookahead.
  	        if (this.config.tokens) {
  	            this.tokens.pop();
  	        }
  	        this.startJSX();
  	        var element = this.parseJSXElement();
  	        this.finishJSX();
  	        return element;
  	    };
  	    JSXParser.prototype.isStartOfExpression = function () {
  	        return _super.prototype.isStartOfExpression.call(this) || this.match('<');
  	    };
  	    return JSXParser;
  	}(parser_1.Parser));
  	exports.JSXParser = JSXParser;


  /***/ },
  /* 4 */
  /***/ function(module, exports) {
  	Object.defineProperty(exports, "__esModule", { value: true });
  	// See also tools/generate-unicode-regex.js.
  	var Regex = {
  	    // Unicode v8.0.0 NonAsciiIdentifierStart:
  	    NonAsciiIdentifierStart: /[\xAA\xB5\xBA\xC0-\xD6\xD8-\xF6\xF8-\u02C1\u02C6-\u02D1\u02E0-\u02E4\u02EC\u02EE\u0370-\u0374\u0376\u0377\u037A-\u037D\u037F\u0386\u0388-\u038A\u038C\u038E-\u03A1\u03A3-\u03F5\u03F7-\u0481\u048A-\u052F\u0531-\u0556\u0559\u0561-\u0587\u05D0-\u05EA\u05F0-\u05F2\u0620-\u064A\u066E\u066F\u0671-\u06D3\u06D5\u06E5\u06E6\u06EE\u06EF\u06FA-\u06FC\u06FF\u0710\u0712-\u072F\u074D-\u07A5\u07B1\u07CA-\u07EA\u07F4\u07F5\u07FA\u0800-\u0815\u081A\u0824\u0828\u0840-\u0858\u08A0-\u08B4\u0904-\u0939\u093D\u0950\u0958-\u0961\u0971-\u0980\u0985-\u098C\u098F\u0990\u0993-\u09A8\u09AA-\u09B0\u09B2\u09B6-\u09B9\u09BD\u09CE\u09DC\u09DD\u09DF-\u09E1\u09F0\u09F1\u0A05-\u0A0A\u0A0F\u0A10\u0A13-\u0A28\u0A2A-\u0A30\u0A32\u0A33\u0A35\u0A36\u0A38\u0A39\u0A59-\u0A5C\u0A5E\u0A72-\u0A74\u0A85-\u0A8D\u0A8F-\u0A91\u0A93-\u0AA8\u0AAA-\u0AB0\u0AB2\u0AB3\u0AB5-\u0AB9\u0ABD\u0AD0\u0AE0\u0AE1\u0AF9\u0B05-\u0B0C\u0B0F\u0B10\u0B13-\u0B28\u0B2A-\u0B30\u0B32\u0B33\u0B35-\u0B39\u0B3D\u0B5C\u0B5D\u0B5F-\u0B61\u0B71\u0B83\u0B85-\u0B8A\u0B8E-\u0B90\u0B92-\u0B95\u0B99\u0B9A\u0B9C\u0B9E\u0B9F\u0BA3\u0BA4\u0BA8-\u0BAA\u0BAE-\u0BB9\u0BD0\u0C05-\u0C0C\u0C0E-\u0C10\u0C12-\u0C28\u0C2A-\u0C39\u0C3D\u0C58-\u0C5A\u0C60\u0C61\u0C85-\u0C8C\u0C8E-\u0C90\u0C92-\u0CA8\u0CAA-\u0CB3\u0CB5-\u0CB9\u0CBD\u0CDE\u0CE0\u0CE1\u0CF1\u0CF2\u0D05-\u0D0C\u0D0E-\u0D10\u0D12-\u0D3A\u0D3D\u0D4E\u0D5F-\u0D61\u0D7A-\u0D7F\u0D85-\u0D96\u0D9A-\u0DB1\u0DB3-\u0DBB\u0DBD\u0DC0-\u0DC6\u0E01-\u0E30\u0E32\u0E33\u0E40-\u0E46\u0E81\u0E82\u0E84\u0E87\u0E88\u0E8A\u0E8D\u0E94-\u0E97\u0E99-\u0E9F\u0EA1-\u0EA3\u0EA5\u0EA7\u0EAA\u0EAB\u0EAD-\u0EB0\u0EB2\u0EB3\u0EBD\u0EC0-\u0EC4\u0EC6\u0EDC-\u0EDF\u0F00\u0F40-\u0F47\u0F49-\u0F6C\u0F88-\u0F8C\u1000-\u102A\u103F\u1050-\u1055\u105A-\u105D\u1061\u1065\u1066\u106E-\u1070\u1075-\u1081\u108E\u10A0-\u10C5\u10C7\u10CD\u10D0-\u10FA\u10FC-\u1248\u124A-\u124D\u1250-\u1256\u1258\u125A-\u125D\u1260-\u1288\u128A-\u128D\u1290-\u12B0\u12B2-\u12B5\u12B8-\u12BE\u12C0\u12C2-\u12C5\u12C8-\u12D6\u12D8-\u1310\u1312-\u1315\u1318-\u135A\u1380-\u138F\u13A0-\u13F5\u13F8-\u13FD\u1401-\u166C\u166F-\u167F\u1681-\u169A\u16A0-\u16EA\u16EE-\u16F8\u1700-\u170C\u170E-\u1711\u1720-\u1731\u1740-\u1751\u1760-\u176C\u176E-\u1770\u1780-\u17B3\u17D7\u17DC\u1820-\u1877\u1880-\u18A8\u18AA\u18B0-\u18F5\u1900-\u191E\u1950-\u196D\u1970-\u1974\u1980-\u19AB\u19B0-\u19C9\u1A00-\u1A16\u1A20-\u1A54\u1AA7\u1B05-\u1B33\u1B45-\u1B4B\u1B83-\u1BA0\u1BAE\u1BAF\u1BBA-\u1BE5\u1C00-\u1C23\u1C4D-\u1C4F\u1C5A-\u1C7D\u1CE9-\u1CEC\u1CEE-\u1CF1\u1CF5\u1CF6\u1D00-\u1DBF\u1E00-\u1F15\u1F18-\u1F1D\u1F20-\u1F45\u1F48-\u1F4D\u1F50-\u1F57\u1F59\u1F5B\u1F5D\u1F5F-\u1F7D\u1F80-\u1FB4\u1FB6-\u1FBC\u1FBE\u1FC2-\u1FC4\u1FC6-\u1FCC\u1FD0-\u1FD3\u1FD6-\u1FDB\u1FE0-\u1FEC\u1FF2-\u1FF4\u1FF6-\u1FFC\u2071\u207F\u2090-\u209C\u2102\u2107\u210A-\u2113\u2115\u2118-\u211D\u2124\u2126\u2128\u212A-\u2139\u213C-\u213F\u2145-\u2149\u214E\u2160-\u2188\u2C00-\u2C2E\u2C30-\u2C5E\u2C60-\u2CE4\u2CEB-\u2CEE\u2CF2\u2CF3\u2D00-\u2D25\u2D27\u2D2D\u2D30-\u2D67\u2D6F\u2D80-\u2D96\u2DA0-\u2DA6\u2DA8-\u2DAE\u2DB0-\u2DB6\u2DB8-\u2DBE\u2DC0-\u2DC6\u2DC8-\u2DCE\u2DD0-\u2DD6\u2DD8-\u2DDE\u3005-\u3007\u3021-\u3029\u3031-\u3035\u3038-\u303C\u3041-\u3096\u309B-\u309F\u30A1-\u30FA\u30FC-\u30FF\u3105-\u312D\u3131-\u318E\u31A0-\u31BA\u31F0-\u31FF\u3400-\u4DB5\u4E00-\u9FD5\uA000-\uA48C\uA4D0-\uA4FD\uA500-\uA60C\uA610-\uA61F\uA62A\uA62B\uA640-\uA66E\uA67F-\uA69D\uA6A0-\uA6EF\uA717-\uA71F\uA722-\uA788\uA78B-\uA7AD\uA7B0-\uA7B7\uA7F7-\uA801\uA803-\uA805\uA807-\uA80A\uA80C-\uA822\uA840-\uA873\uA882-\uA8B3\uA8F2-\uA8F7\uA8FB\uA8FD\uA90A-\uA925\uA930-\uA946\uA960-\uA97C\uA984-\uA9B2\uA9CF\uA9E0-\uA9E4\uA9E6-\uA9EF\uA9FA-\uA9FE\uAA00-\uAA28\uAA40-\uAA42\uAA44-\uAA4B\uAA60-\uAA76\uAA7A\uAA7E-\uAAAF\uAAB1\uAAB5\uAAB6\uAAB9-\uAABD\uAAC0\uAAC2\uAADB-\uAADD\uAAE0-\uAAEA\uAAF2-\uAAF4\uAB01-\uAB06\uAB09-\uAB0E\uAB11-\uAB16\uAB20-\uAB26\uAB28-\uAB2E\uAB30-\uAB5A\uAB5C-\uAB65\uAB70-\uABE2\uAC00-\uD7A3\uD7B0-\uD7C6\uD7CB-\uD7FB\uF900-\uFA6D\uFA70-\uFAD9\uFB00-\uFB06\uFB13-\uFB17\uFB1D\uFB1F-\uFB28\uFB2A-\uFB36\uFB38-\uFB3C\uFB3E\uFB40\uFB41\uFB43\uFB44\uFB46-\uFBB1\uFBD3-\uFD3D\uFD50-\uFD8F\uFD92-\uFDC7\uFDF0-\uFDFB\uFE70-\uFE74\uFE76-\uFEFC\uFF21-\uFF3A\uFF41-\uFF5A\uFF66-\uFFBE\uFFC2-\uFFC7\uFFCA-\uFFCF\uFFD2-\uFFD7\uFFDA-\uFFDC]|\uD800[\uDC00-\uDC0B\uDC0D-\uDC26\uDC28-\uDC3A\uDC3C\uDC3D\uDC3F-\uDC4D\uDC50-\uDC5D\uDC80-\uDCFA\uDD40-\uDD74\uDE80-\uDE9C\uDEA0-\uDED0\uDF00-\uDF1F\uDF30-\uDF4A\uDF50-\uDF75\uDF80-\uDF9D\uDFA0-\uDFC3\uDFC8-\uDFCF\uDFD1-\uDFD5]|\uD801[\uDC00-\uDC9D\uDD00-\uDD27\uDD30-\uDD63\uDE00-\uDF36\uDF40-\uDF55\uDF60-\uDF67]|\uD802[\uDC00-\uDC05\uDC08\uDC0A-\uDC35\uDC37\uDC38\uDC3C\uDC3F-\uDC55\uDC60-\uDC76\uDC80-\uDC9E\uDCE0-\uDCF2\uDCF4\uDCF5\uDD00-\uDD15\uDD20-\uDD39\uDD80-\uDDB7\uDDBE\uDDBF\uDE00\uDE10-\uDE13\uDE15-\uDE17\uDE19-\uDE33\uDE60-\uDE7C\uDE80-\uDE9C\uDEC0-\uDEC7\uDEC9-\uDEE4\uDF00-\uDF35\uDF40-\uDF55\uDF60-\uDF72\uDF80-\uDF91]|\uD803[\uDC00-\uDC48\uDC80-\uDCB2\uDCC0-\uDCF2]|\uD804[\uDC03-\uDC37\uDC83-\uDCAF\uDCD0-\uDCE8\uDD03-\uDD26\uDD50-\uDD72\uDD76\uDD83-\uDDB2\uDDC1-\uDDC4\uDDDA\uDDDC\uDE00-\uDE11\uDE13-\uDE2B\uDE80-\uDE86\uDE88\uDE8A-\uDE8D\uDE8F-\uDE9D\uDE9F-\uDEA8\uDEB0-\uDEDE\uDF05-\uDF0C\uDF0F\uDF10\uDF13-\uDF28\uDF2A-\uDF30\uDF32\uDF33\uDF35-\uDF39\uDF3D\uDF50\uDF5D-\uDF61]|\uD805[\uDC80-\uDCAF\uDCC4\uDCC5\uDCC7\uDD80-\uDDAE\uDDD8-\uDDDB\uDE00-\uDE2F\uDE44\uDE80-\uDEAA\uDF00-\uDF19]|\uD806[\uDCA0-\uDCDF\uDCFF\uDEC0-\uDEF8]|\uD808[\uDC00-\uDF99]|\uD809[\uDC00-\uDC6E\uDC80-\uDD43]|[\uD80C\uD840-\uD868\uD86A-\uD86C\uD86F-\uD872][\uDC00-\uDFFF]|\uD80D[\uDC00-\uDC2E]|\uD811[\uDC00-\uDE46]|\uD81A[\uDC00-\uDE38\uDE40-\uDE5E\uDED0-\uDEED\uDF00-\uDF2F\uDF40-\uDF43\uDF63-\uDF77\uDF7D-\uDF8F]|\uD81B[\uDF00-\uDF44\uDF50\uDF93-\uDF9F]|\uD82C[\uDC00\uDC01]|\uD82F[\uDC00-\uDC6A\uDC70-\uDC7C\uDC80-\uDC88\uDC90-\uDC99]|\uD835[\uDC00-\uDC54\uDC56-\uDC9C\uDC9E\uDC9F\uDCA2\uDCA5\uDCA6\uDCA9-\uDCAC\uDCAE-\uDCB9\uDCBB\uDCBD-\uDCC3\uDCC5-\uDD05\uDD07-\uDD0A\uDD0D-\uDD14\uDD16-\uDD1C\uDD1E-\uDD39\uDD3B-\uDD3E\uDD40-\uDD44\uDD46\uDD4A-\uDD50\uDD52-\uDEA5\uDEA8-\uDEC0\uDEC2-\uDEDA\uDEDC-\uDEFA\uDEFC-\uDF14\uDF16-\uDF34\uDF36-\uDF4E\uDF50-\uDF6E\uDF70-\uDF88\uDF8A-\uDFA8\uDFAA-\uDFC2\uDFC4-\uDFCB]|\uD83A[\uDC00-\uDCC4]|\uD83B[\uDE00-\uDE03\uDE05-\uDE1F\uDE21\uDE22\uDE24\uDE27\uDE29-\uDE32\uDE34-\uDE37\uDE39\uDE3B\uDE42\uDE47\uDE49\uDE4B\uDE4D-\uDE4F\uDE51\uDE52\uDE54\uDE57\uDE59\uDE5B\uDE5D\uDE5F\uDE61\uDE62\uDE64\uDE67-\uDE6A\uDE6C-\uDE72\uDE74-\uDE77\uDE79-\uDE7C\uDE7E\uDE80-\uDE89\uDE8B-\uDE9B\uDEA1-\uDEA3\uDEA5-\uDEA9\uDEAB-\uDEBB]|\uD869[\uDC00-\uDED6\uDF00-\uDFFF]|\uD86D[\uDC00-\uDF34\uDF40-\uDFFF]|\uD86E[\uDC00-\uDC1D\uDC20-\uDFFF]|\uD873[\uDC00-\uDEA1]|\uD87E[\uDC00-\uDE1D]/,
  	    // Unicode v8.0.0 NonAsciiIdentifierPart:
  	    NonAsciiIdentifierPart: /[\xAA\xB5\xB7\xBA\xC0-\xD6\xD8-\xF6\xF8-\u02C1\u02C6-\u02D1\u02E0-\u02E4\u02EC\u02EE\u0300-\u0374\u0376\u0377\u037A-\u037D\u037F\u0386-\u038A\u038C\u038E-\u03A1\u03A3-\u03F5\u03F7-\u0481\u0483-\u0487\u048A-\u052F\u0531-\u0556\u0559\u0561-\u0587\u0591-\u05BD\u05BF\u05C1\u05C2\u05C4\u05C5\u05C7\u05D0-\u05EA\u05F0-\u05F2\u0610-\u061A\u0620-\u0669\u066E-\u06D3\u06D5-\u06DC\u06DF-\u06E8\u06EA-\u06FC\u06FF\u0710-\u074A\u074D-\u07B1\u07C0-\u07F5\u07FA\u0800-\u082D\u0840-\u085B\u08A0-\u08B4\u08E3-\u0963\u0966-\u096F\u0971-\u0983\u0985-\u098C\u098F\u0990\u0993-\u09A8\u09AA-\u09B0\u09B2\u09B6-\u09B9\u09BC-\u09C4\u09C7\u09C8\u09CB-\u09CE\u09D7\u09DC\u09DD\u09DF-\u09E3\u09E6-\u09F1\u0A01-\u0A03\u0A05-\u0A0A\u0A0F\u0A10\u0A13-\u0A28\u0A2A-\u0A30\u0A32\u0A33\u0A35\u0A36\u0A38\u0A39\u0A3C\u0A3E-\u0A42\u0A47\u0A48\u0A4B-\u0A4D\u0A51\u0A59-\u0A5C\u0A5E\u0A66-\u0A75\u0A81-\u0A83\u0A85-\u0A8D\u0A8F-\u0A91\u0A93-\u0AA8\u0AAA-\u0AB0\u0AB2\u0AB3\u0AB5-\u0AB9\u0ABC-\u0AC5\u0AC7-\u0AC9\u0ACB-\u0ACD\u0AD0\u0AE0-\u0AE3\u0AE6-\u0AEF\u0AF9\u0B01-\u0B03\u0B05-\u0B0C\u0B0F\u0B10\u0B13-\u0B28\u0B2A-\u0B30\u0B32\u0B33\u0B35-\u0B39\u0B3C-\u0B44\u0B47\u0B48\u0B4B-\u0B4D\u0B56\u0B57\u0B5C\u0B5D\u0B5F-\u0B63\u0B66-\u0B6F\u0B71\u0B82\u0B83\u0B85-\u0B8A\u0B8E-\u0B90\u0B92-\u0B95\u0B99\u0B9A\u0B9C\u0B9E\u0B9F\u0BA3\u0BA4\u0BA8-\u0BAA\u0BAE-\u0BB9\u0BBE-\u0BC2\u0BC6-\u0BC8\u0BCA-\u0BCD\u0BD0\u0BD7\u0BE6-\u0BEF\u0C00-\u0C03\u0C05-\u0C0C\u0C0E-\u0C10\u0C12-\u0C28\u0C2A-\u0C39\u0C3D-\u0C44\u0C46-\u0C48\u0C4A-\u0C4D\u0C55\u0C56\u0C58-\u0C5A\u0C60-\u0C63\u0C66-\u0C6F\u0C81-\u0C83\u0C85-\u0C8C\u0C8E-\u0C90\u0C92-\u0CA8\u0CAA-\u0CB3\u0CB5-\u0CB9\u0CBC-\u0CC4\u0CC6-\u0CC8\u0CCA-\u0CCD\u0CD5\u0CD6\u0CDE\u0CE0-\u0CE3\u0CE6-\u0CEF\u0CF1\u0CF2\u0D01-\u0D03\u0D05-\u0D0C\u0D0E-\u0D10\u0D12-\u0D3A\u0D3D-\u0D44\u0D46-\u0D48\u0D4A-\u0D4E\u0D57\u0D5F-\u0D63\u0D66-\u0D6F\u0D7A-\u0D7F\u0D82\u0D83\u0D85-\u0D96\u0D9A-\u0DB1\u0DB3-\u0DBB\u0DBD\u0DC0-\u0DC6\u0DCA\u0DCF-\u0DD4\u0DD6\u0DD8-\u0DDF\u0DE6-\u0DEF\u0DF2\u0DF3\u0E01-\u0E3A\u0E40-\u0E4E\u0E50-\u0E59\u0E81\u0E82\u0E84\u0E87\u0E88\u0E8A\u0E8D\u0E94-\u0E97\u0E99-\u0E9F\u0EA1-\u0EA3\u0EA5\u0EA7\u0EAA\u0EAB\u0EAD-\u0EB9\u0EBB-\u0EBD\u0EC0-\u0EC4\u0EC6\u0EC8-\u0ECD\u0ED0-\u0ED9\u0EDC-\u0EDF\u0F00\u0F18\u0F19\u0F20-\u0F29\u0F35\u0F37\u0F39\u0F3E-\u0F47\u0F49-\u0F6C\u0F71-\u0F84\u0F86-\u0F97\u0F99-\u0FBC\u0FC6\u1000-\u1049\u1050-\u109D\u10A0-\u10C5\u10C7\u10CD\u10D0-\u10FA\u10FC-\u1248\u124A-\u124D\u1250-\u1256\u1258\u125A-\u125D\u1260-\u1288\u128A-\u128D\u1290-\u12B0\u12B2-\u12B5\u12B8-\u12BE\u12C0\u12C2-\u12C5\u12C8-\u12D6\u12D8-\u1310\u1312-\u1315\u1318-\u135A\u135D-\u135F\u1369-\u1371\u1380-\u138F\u13A0-\u13F5\u13F8-\u13FD\u1401-\u166C\u166F-\u167F\u1681-\u169A\u16A0-\u16EA\u16EE-\u16F8\u1700-\u170C\u170E-\u1714\u1720-\u1734\u1740-\u1753\u1760-\u176C\u176E-\u1770\u1772\u1773\u1780-\u17D3\u17D7\u17DC\u17DD\u17E0-\u17E9\u180B-\u180D\u1810-\u1819\u1820-\u1877\u1880-\u18AA\u18B0-\u18F5\u1900-\u191E\u1920-\u192B\u1930-\u193B\u1946-\u196D\u1970-\u1974\u1980-\u19AB\u19B0-\u19C9\u19D0-\u19DA\u1A00-\u1A1B\u1A20-\u1A5E\u1A60-\u1A7C\u1A7F-\u1A89\u1A90-\u1A99\u1AA7\u1AB0-\u1ABD\u1B00-\u1B4B\u1B50-\u1B59\u1B6B-\u1B73\u1B80-\u1BF3\u1C00-\u1C37\u1C40-\u1C49\u1C4D-\u1C7D\u1CD0-\u1CD2\u1CD4-\u1CF6\u1CF8\u1CF9\u1D00-\u1DF5\u1DFC-\u1F15\u1F18-\u1F1D\u1F20-\u1F45\u1F48-\u1F4D\u1F50-\u1F57\u1F59\u1F5B\u1F5D\u1F5F-\u1F7D\u1F80-\u1FB4\u1FB6-\u1FBC\u1FBE\u1FC2-\u1FC4\u1FC6-\u1FCC\u1FD0-\u1FD3\u1FD6-\u1FDB\u1FE0-\u1FEC\u1FF2-\u1FF4\u1FF6-\u1FFC\u200C\u200D\u203F\u2040\u2054\u2071\u207F\u2090-\u209C\u20D0-\u20DC\u20E1\u20E5-\u20F0\u2102\u2107\u210A-\u2113\u2115\u2118-\u211D\u2124\u2126\u2128\u212A-\u2139\u213C-\u213F\u2145-\u2149\u214E\u2160-\u2188\u2C00-\u2C2E\u2C30-\u2C5E\u2C60-\u2CE4\u2CEB-\u2CF3\u2D00-\u2D25\u2D27\u2D2D\u2D30-\u2D67\u2D6F\u2D7F-\u2D96\u2DA0-\u2DA6\u2DA8-\u2DAE\u2DB0-\u2DB6\u2DB8-\u2DBE\u2DC0-\u2DC6\u2DC8-\u2DCE\u2DD0-\u2DD6\u2DD8-\u2DDE\u2DE0-\u2DFF\u3005-\u3007\u3021-\u302F\u3031-\u3035\u3038-\u303C\u3041-\u3096\u3099-\u309F\u30A1-\u30FA\u30FC-\u30FF\u3105-\u312D\u3131-\u318E\u31A0-\u31BA\u31F0-\u31FF\u3400-\u4DB5\u4E00-\u9FD5\uA000-\uA48C\uA4D0-\uA4FD\uA500-\uA60C\uA610-\uA62B\uA640-\uA66F\uA674-\uA67D\uA67F-\uA6F1\uA717-\uA71F\uA722-\uA788\uA78B-\uA7AD\uA7B0-\uA7B7\uA7F7-\uA827\uA840-\uA873\uA880-\uA8C4\uA8D0-\uA8D9\uA8E0-\uA8F7\uA8FB\uA8FD\uA900-\uA92D\uA930-\uA953\uA960-\uA97C\uA980-\uA9C0\uA9CF-\uA9D9\uA9E0-\uA9FE\uAA00-\uAA36\uAA40-\uAA4D\uAA50-\uAA59\uAA60-\uAA76\uAA7A-\uAAC2\uAADB-\uAADD\uAAE0-\uAAEF\uAAF2-\uAAF6\uAB01-\uAB06\uAB09-\uAB0E\uAB11-\uAB16\uAB20-\uAB26\uAB28-\uAB2E\uAB30-\uAB5A\uAB5C-\uAB65\uAB70-\uABEA\uABEC\uABED\uABF0-\uABF9\uAC00-\uD7A3\uD7B0-\uD7C6\uD7CB-\uD7FB\uF900-\uFA6D\uFA70-\uFAD9\uFB00-\uFB06\uFB13-\uFB17\uFB1D-\uFB28\uFB2A-\uFB36\uFB38-\uFB3C\uFB3E\uFB40\uFB41\uFB43\uFB44\uFB46-\uFBB1\uFBD3-\uFD3D\uFD50-\uFD8F\uFD92-\uFDC7\uFDF0-\uFDFB\uFE00-\uFE0F\uFE20-\uFE2F\uFE33\uFE34\uFE4D-\uFE4F\uFE70-\uFE74\uFE76-\uFEFC\uFF10-\uFF19\uFF21-\uFF3A\uFF3F\uFF41-\uFF5A\uFF66-\uFFBE\uFFC2-\uFFC7\uFFCA-\uFFCF\uFFD2-\uFFD7\uFFDA-\uFFDC]|\uD800[\uDC00-\uDC0B\uDC0D-\uDC26\uDC28-\uDC3A\uDC3C\uDC3D\uDC3F-\uDC4D\uDC50-\uDC5D\uDC80-\uDCFA\uDD40-\uDD74\uDDFD\uDE80-\uDE9C\uDEA0-\uDED0\uDEE0\uDF00-\uDF1F\uDF30-\uDF4A\uDF50-\uDF7A\uDF80-\uDF9D\uDFA0-\uDFC3\uDFC8-\uDFCF\uDFD1-\uDFD5]|\uD801[\uDC00-\uDC9D\uDCA0-\uDCA9\uDD00-\uDD27\uDD30-\uDD63\uDE00-\uDF36\uDF40-\uDF55\uDF60-\uDF67]|\uD802[\uDC00-\uDC05\uDC08\uDC0A-\uDC35\uDC37\uDC38\uDC3C\uDC3F-\uDC55\uDC60-\uDC76\uDC80-\uDC9E\uDCE0-\uDCF2\uDCF4\uDCF5\uDD00-\uDD15\uDD20-\uDD39\uDD80-\uDDB7\uDDBE\uDDBF\uDE00-\uDE03\uDE05\uDE06\uDE0C-\uDE13\uDE15-\uDE17\uDE19-\uDE33\uDE38-\uDE3A\uDE3F\uDE60-\uDE7C\uDE80-\uDE9C\uDEC0-\uDEC7\uDEC9-\uDEE6\uDF00-\uDF35\uDF40-\uDF55\uDF60-\uDF72\uDF80-\uDF91]|\uD803[\uDC00-\uDC48\uDC80-\uDCB2\uDCC0-\uDCF2]|\uD804[\uDC00-\uDC46\uDC66-\uDC6F\uDC7F-\uDCBA\uDCD0-\uDCE8\uDCF0-\uDCF9\uDD00-\uDD34\uDD36-\uDD3F\uDD50-\uDD73\uDD76\uDD80-\uDDC4\uDDCA-\uDDCC\uDDD0-\uDDDA\uDDDC\uDE00-\uDE11\uDE13-\uDE37\uDE80-\uDE86\uDE88\uDE8A-\uDE8D\uDE8F-\uDE9D\uDE9F-\uDEA8\uDEB0-\uDEEA\uDEF0-\uDEF9\uDF00-\uDF03\uDF05-\uDF0C\uDF0F\uDF10\uDF13-\uDF28\uDF2A-\uDF30\uDF32\uDF33\uDF35-\uDF39\uDF3C-\uDF44\uDF47\uDF48\uDF4B-\uDF4D\uDF50\uDF57\uDF5D-\uDF63\uDF66-\uDF6C\uDF70-\uDF74]|\uD805[\uDC80-\uDCC5\uDCC7\uDCD0-\uDCD9\uDD80-\uDDB5\uDDB8-\uDDC0\uDDD8-\uDDDD\uDE00-\uDE40\uDE44\uDE50-\uDE59\uDE80-\uDEB7\uDEC0-\uDEC9\uDF00-\uDF19\uDF1D-\uDF2B\uDF30-\uDF39]|\uD806[\uDCA0-\uDCE9\uDCFF\uDEC0-\uDEF8]|\uD808[\uDC00-\uDF99]|\uD809[\uDC00-\uDC6E\uDC80-\uDD43]|[\uD80C\uD840-\uD868\uD86A-\uD86C\uD86F-\uD872][\uDC00-\uDFFF]|\uD80D[\uDC00-\uDC2E]|\uD811[\uDC00-\uDE46]|\uD81A[\uDC00-\uDE38\uDE40-\uDE5E\uDE60-\uDE69\uDED0-\uDEED\uDEF0-\uDEF4\uDF00-\uDF36\uDF40-\uDF43\uDF50-\uDF59\uDF63-\uDF77\uDF7D-\uDF8F]|\uD81B[\uDF00-\uDF44\uDF50-\uDF7E\uDF8F-\uDF9F]|\uD82C[\uDC00\uDC01]|\uD82F[\uDC00-\uDC6A\uDC70-\uDC7C\uDC80-\uDC88\uDC90-\uDC99\uDC9D\uDC9E]|\uD834[\uDD65-\uDD69\uDD6D-\uDD72\uDD7B-\uDD82\uDD85-\uDD8B\uDDAA-\uDDAD\uDE42-\uDE44]|\uD835[\uDC00-\uDC54\uDC56-\uDC9C\uDC9E\uDC9F\uDCA2\uDCA5\uDCA6\uDCA9-\uDCAC\uDCAE-\uDCB9\uDCBB\uDCBD-\uDCC3\uDCC5-\uDD05\uDD07-\uDD0A\uDD0D-\uDD14\uDD16-\uDD1C\uDD1E-\uDD39\uDD3B-\uDD3E\uDD40-\uDD44\uDD46\uDD4A-\uDD50\uDD52-\uDEA5\uDEA8-\uDEC0\uDEC2-\uDEDA\uDEDC-\uDEFA\uDEFC-\uDF14\uDF16-\uDF34\uDF36-\uDF4E\uDF50-\uDF6E\uDF70-\uDF88\uDF8A-\uDFA8\uDFAA-\uDFC2\uDFC4-\uDFCB\uDFCE-\uDFFF]|\uD836[\uDE00-\uDE36\uDE3B-\uDE6C\uDE75\uDE84\uDE9B-\uDE9F\uDEA1-\uDEAF]|\uD83A[\uDC00-\uDCC4\uDCD0-\uDCD6]|\uD83B[\uDE00-\uDE03\uDE05-\uDE1F\uDE21\uDE22\uDE24\uDE27\uDE29-\uDE32\uDE34-\uDE37\uDE39\uDE3B\uDE42\uDE47\uDE49\uDE4B\uDE4D-\uDE4F\uDE51\uDE52\uDE54\uDE57\uDE59\uDE5B\uDE5D\uDE5F\uDE61\uDE62\uDE64\uDE67-\uDE6A\uDE6C-\uDE72\uDE74-\uDE77\uDE79-\uDE7C\uDE7E\uDE80-\uDE89\uDE8B-\uDE9B\uDEA1-\uDEA3\uDEA5-\uDEA9\uDEAB-\uDEBB]|\uD869[\uDC00-\uDED6\uDF00-\uDFFF]|\uD86D[\uDC00-\uDF34\uDF40-\uDFFF]|\uD86E[\uDC00-\uDC1D\uDC20-\uDFFF]|\uD873[\uDC00-\uDEA1]|\uD87E[\uDC00-\uDE1D]|\uDB40[\uDD00-\uDDEF]/
  	};
  	exports.Character = {
  	    /* tslint:disable:no-bitwise */
  	    fromCodePoint: function (cp) {
  	        return (cp < 0x10000) ? String.fromCharCode(cp) :
  	            String.fromCharCode(0xD800 + ((cp - 0x10000) >> 10)) +
  	                String.fromCharCode(0xDC00 + ((cp - 0x10000) & 1023));
  	    },
  	    // https://tc39.github.io/ecma262/#sec-white-space
  	    isWhiteSpace: function (cp) {
  	        return (cp === 0x20) || (cp === 0x09) || (cp === 0x0B) || (cp === 0x0C) || (cp === 0xA0) ||
  	            (cp >= 0x1680 && [0x1680, 0x2000, 0x2001, 0x2002, 0x2003, 0x2004, 0x2005, 0x2006, 0x2007, 0x2008, 0x2009, 0x200A, 0x202F, 0x205F, 0x3000, 0xFEFF].indexOf(cp) >= 0);
  	    },
  	    // https://tc39.github.io/ecma262/#sec-line-terminators
  	    isLineTerminator: function (cp) {
  	        return (cp === 0x0A) || (cp === 0x0D) || (cp === 0x2028) || (cp === 0x2029);
  	    },
  	    // https://tc39.github.io/ecma262/#sec-names-and-keywords
  	    isIdentifierStart: function (cp) {
  	        return (cp === 0x24) || (cp === 0x5F) ||
  	            (cp >= 0x41 && cp <= 0x5A) ||
  	            (cp >= 0x61 && cp <= 0x7A) ||
  	            (cp === 0x5C) ||
  	            ((cp >= 0x80) && Regex.NonAsciiIdentifierStart.test(exports.Character.fromCodePoint(cp)));
  	    },
  	    isIdentifierPart: function (cp) {
  	        return (cp === 0x24) || (cp === 0x5F) ||
  	            (cp >= 0x41 && cp <= 0x5A) ||
  	            (cp >= 0x61 && cp <= 0x7A) ||
  	            (cp >= 0x30 && cp <= 0x39) ||
  	            (cp === 0x5C) ||
  	            ((cp >= 0x80) && Regex.NonAsciiIdentifierPart.test(exports.Character.fromCodePoint(cp)));
  	    },
  	    // https://tc39.github.io/ecma262/#sec-literals-numeric-literals
  	    isDecimalDigit: function (cp) {
  	        return (cp >= 0x30 && cp <= 0x39); // 0..9
  	    },
  	    isHexDigit: function (cp) {
  	        return (cp >= 0x30 && cp <= 0x39) ||
  	            (cp >= 0x41 && cp <= 0x46) ||
  	            (cp >= 0x61 && cp <= 0x66); // a..f
  	    },
  	    isOctalDigit: function (cp) {
  	        return (cp >= 0x30 && cp <= 0x37); // 0..7
  	    }
  	};


  /***/ },
  /* 5 */
  /***/ function(module, exports, __webpack_require__) {
  	Object.defineProperty(exports, "__esModule", { value: true });
  	var jsx_syntax_1 = __webpack_require__(6);
  	/* tslint:disable:max-classes-per-file */
  	var JSXClosingElement = (function () {
  	    function JSXClosingElement(name) {
  	        this.type = jsx_syntax_1.JSXSyntax.JSXClosingElement;
  	        this.name = name;
  	    }
  	    return JSXClosingElement;
  	}());
  	exports.JSXClosingElement = JSXClosingElement;
  	var JSXElement = (function () {
  	    function JSXElement(openingElement, children, closingElement) {
  	        this.type = jsx_syntax_1.JSXSyntax.JSXElement;
  	        this.openingElement = openingElement;
  	        this.children = children;
  	        this.closingElement = closingElement;
  	    }
  	    return JSXElement;
  	}());
  	exports.JSXElement = JSXElement;
  	var JSXEmptyExpression = (function () {
  	    function JSXEmptyExpression() {
  	        this.type = jsx_syntax_1.JSXSyntax.JSXEmptyExpression;
  	    }
  	    return JSXEmptyExpression;
  	}());
  	exports.JSXEmptyExpression = JSXEmptyExpression;
  	var JSXExpressionContainer = (function () {
  	    function JSXExpressionContainer(expression) {
  	        this.type = jsx_syntax_1.JSXSyntax.JSXExpressionContainer;
  	        this.expression = expression;
  	    }
  	    return JSXExpressionContainer;
  	}());
  	exports.JSXExpressionContainer = JSXExpressionContainer;
  	var JSXIdentifier = (function () {
  	    function JSXIdentifier(name) {
  	        this.type = jsx_syntax_1.JSXSyntax.JSXIdentifier;
  	        this.name = name;
  	    }
  	    return JSXIdentifier;
  	}());
  	exports.JSXIdentifier = JSXIdentifier;
  	var JSXMemberExpression = (function () {
  	    function JSXMemberExpression(object, property) {
  	        this.type = jsx_syntax_1.JSXSyntax.JSXMemberExpression;
  	        this.object = object;
  	        this.property = property;
  	    }
  	    return JSXMemberExpression;
  	}());
  	exports.JSXMemberExpression = JSXMemberExpression;
  	var JSXAttribute = (function () {
  	    function JSXAttribute(name, value) {
  	        this.type = jsx_syntax_1.JSXSyntax.JSXAttribute;
  	        this.name = name;
  	        this.value = value;
  	    }
  	    return JSXAttribute;
  	}());
  	exports.JSXAttribute = JSXAttribute;
  	var JSXNamespacedName = (function () {
  	    function JSXNamespacedName(namespace, name) {
  	        this.type = jsx_syntax_1.JSXSyntax.JSXNamespacedName;
  	        this.namespace = namespace;
  	        this.name = name;
  	    }
  	    return JSXNamespacedName;
  	}());
  	exports.JSXNamespacedName = JSXNamespacedName;
  	var JSXOpeningElement = (function () {
  	    function JSXOpeningElement(name, selfClosing, attributes) {
  	        this.type = jsx_syntax_1.JSXSyntax.JSXOpeningElement;
  	        this.name = name;
  	        this.selfClosing = selfClosing;
  	        this.attributes = attributes;
  	    }
  	    return JSXOpeningElement;
  	}());
  	exports.JSXOpeningElement = JSXOpeningElement;
  	var JSXSpreadAttribute = (function () {
  	    function JSXSpreadAttribute(argument) {
  	        this.type = jsx_syntax_1.JSXSyntax.JSXSpreadAttribute;
  	        this.argument = argument;
  	    }
  	    return JSXSpreadAttribute;
  	}());
  	exports.JSXSpreadAttribute = JSXSpreadAttribute;
  	var JSXText = (function () {
  	    function JSXText(value, raw) {
  	        this.type = jsx_syntax_1.JSXSyntax.JSXText;
  	        this.value = value;
  	        this.raw = raw;
  	    }
  	    return JSXText;
  	}());
  	exports.JSXText = JSXText;


  /***/ },
  /* 6 */
  /***/ function(module, exports) {
  	Object.defineProperty(exports, "__esModule", { value: true });
  	exports.JSXSyntax = {
  	    JSXAttribute: 'JSXAttribute',
  	    JSXClosingElement: 'JSXClosingElement',
  	    JSXElement: 'JSXElement',
  	    JSXEmptyExpression: 'JSXEmptyExpression',
  	    JSXExpressionContainer: 'JSXExpressionContainer',
  	    JSXIdentifier: 'JSXIdentifier',
  	    JSXMemberExpression: 'JSXMemberExpression',
  	    JSXNamespacedName: 'JSXNamespacedName',
  	    JSXOpeningElement: 'JSXOpeningElement',
  	    JSXSpreadAttribute: 'JSXSpreadAttribute',
  	    JSXText: 'JSXText'
  	};


  /***/ },
  /* 7 */
  /***/ function(module, exports, __webpack_require__) {
  	Object.defineProperty(exports, "__esModule", { value: true });
  	var syntax_1 = __webpack_require__(2);
  	/* tslint:disable:max-classes-per-file */
  	var ArrayExpression = (function () {
  	    function ArrayExpression(elements) {
  	        this.type = syntax_1.Syntax.ArrayExpression;
  	        this.elements = elements;
  	    }
  	    return ArrayExpression;
  	}());
  	exports.ArrayExpression = ArrayExpression;
  	var ArrayPattern = (function () {
  	    function ArrayPattern(elements) {
  	        this.type = syntax_1.Syntax.ArrayPattern;
  	        this.elements = elements;
  	    }
  	    return ArrayPattern;
  	}());
  	exports.ArrayPattern = ArrayPattern;
  	var ArrowFunctionExpression = (function () {
  	    function ArrowFunctionExpression(params, body, expression) {
  	        this.type = syntax_1.Syntax.ArrowFunctionExpression;
  	        this.id = null;
  	        this.params = params;
  	        this.body = body;
  	        this.generator = false;
  	        this.expression = expression;
  	        this.async = false;
  	    }
  	    return ArrowFunctionExpression;
  	}());
  	exports.ArrowFunctionExpression = ArrowFunctionExpression;
  	var AssignmentExpression = (function () {
  	    function AssignmentExpression(operator, left, right) {
  	        this.type = syntax_1.Syntax.AssignmentExpression;
  	        this.operator = operator;
  	        this.left = left;
  	        this.right = right;
  	    }
  	    return AssignmentExpression;
  	}());
  	exports.AssignmentExpression = AssignmentExpression;
  	var AssignmentPattern = (function () {
  	    function AssignmentPattern(left, right) {
  	        this.type = syntax_1.Syntax.AssignmentPattern;
  	        this.left = left;
  	        this.right = right;
  	    }
  	    return AssignmentPattern;
  	}());
  	exports.AssignmentPattern = AssignmentPattern;
  	var AsyncArrowFunctionExpression = (function () {
  	    function AsyncArrowFunctionExpression(params, body, expression) {
  	        this.type = syntax_1.Syntax.ArrowFunctionExpression;
  	        this.id = null;
  	        this.params = params;
  	        this.body = body;
  	        this.generator = false;
  	        this.expression = expression;
  	        this.async = true;
  	    }
  	    return AsyncArrowFunctionExpression;
  	}());
  	exports.AsyncArrowFunctionExpression = AsyncArrowFunctionExpression;
  	var AsyncFunctionDeclaration = (function () {
  	    function AsyncFunctionDeclaration(id, params, body) {
  	        this.type = syntax_1.Syntax.FunctionDeclaration;
  	        this.id = id;
  	        this.params = params;
  	        this.body = body;
  	        this.generator = false;
  	        this.expression = false;
  	        this.async = true;
  	    }
  	    return AsyncFunctionDeclaration;
  	}());
  	exports.AsyncFunctionDeclaration = AsyncFunctionDeclaration;
  	var AsyncFunctionExpression = (function () {
  	    function AsyncFunctionExpression(id, params, body) {
  	        this.type = syntax_1.Syntax.FunctionExpression;
  	        this.id = id;
  	        this.params = params;
  	        this.body = body;
  	        this.generator = false;
  	        this.expression = false;
  	        this.async = true;
  	    }
  	    return AsyncFunctionExpression;
  	}());
  	exports.AsyncFunctionExpression = AsyncFunctionExpression;
  	var AwaitExpression = (function () {
  	    function AwaitExpression(argument) {
  	        this.type = syntax_1.Syntax.AwaitExpression;
  	        this.argument = argument;
  	    }
  	    return AwaitExpression;
  	}());
  	exports.AwaitExpression = AwaitExpression;
  	var BinaryExpression = (function () {
  	    function BinaryExpression(operator, left, right) {
  	        var logical = (operator === '||' || operator === '&&');
  	        this.type = logical ? syntax_1.Syntax.LogicalExpression : syntax_1.Syntax.BinaryExpression;
  	        this.operator = operator;
  	        this.left = left;
  	        this.right = right;
  	    }
  	    return BinaryExpression;
  	}());
  	exports.BinaryExpression = BinaryExpression;
  	var BlockStatement = (function () {
  	    function BlockStatement(body) {
  	        this.type = syntax_1.Syntax.BlockStatement;
  	        this.body = body;
  	    }
  	    return BlockStatement;
  	}());
  	exports.BlockStatement = BlockStatement;
  	var BreakStatement = (function () {
  	    function BreakStatement(label) {
  	        this.type = syntax_1.Syntax.BreakStatement;
  	        this.label = label;
  	    }
  	    return BreakStatement;
  	}());
  	exports.BreakStatement = BreakStatement;
  	var CallExpression = (function () {
  	    function CallExpression(callee, args) {
  	        this.type = syntax_1.Syntax.CallExpression;
  	        this.callee = callee;
  	        this.arguments = args;
  	    }
  	    return CallExpression;
  	}());
  	exports.CallExpression = CallExpression;
  	var CatchClause = (function () {
  	    function CatchClause(param, body) {
  	        this.type = syntax_1.Syntax.CatchClause;
  	        this.param = param;
  	        this.body = body;
  	    }
  	    return CatchClause;
  	}());
  	exports.CatchClause = CatchClause;
  	var ClassBody = (function () {
  	    function ClassBody(body) {
  	        this.type = syntax_1.Syntax.ClassBody;
  	        this.body = body;
  	    }
  	    return ClassBody;
  	}());
  	exports.ClassBody = ClassBody;
  	var ClassDeclaration = (function () {
  	    function ClassDeclaration(id, superClass, body) {
  	        this.type = syntax_1.Syntax.ClassDeclaration;
  	        this.id = id;
  	        this.superClass = superClass;
  	        this.body = body;
  	    }
  	    return ClassDeclaration;
  	}());
  	exports.ClassDeclaration = ClassDeclaration;
  	var ClassExpression = (function () {
  	    function ClassExpression(id, superClass, body) {
  	        this.type = syntax_1.Syntax.ClassExpression;
  	        this.id = id;
  	        this.superClass = superClass;
  	        this.body = body;
  	    }
  	    return ClassExpression;
  	}());
  	exports.ClassExpression = ClassExpression;
  	var ComputedMemberExpression = (function () {
  	    function ComputedMemberExpression(object, property) {
  	        this.type = syntax_1.Syntax.MemberExpression;
  	        this.computed = true;
  	        this.object = object;
  	        this.property = property;
  	    }
  	    return ComputedMemberExpression;
  	}());
  	exports.ComputedMemberExpression = ComputedMemberExpression;
  	var ConditionalExpression = (function () {
  	    function ConditionalExpression(test, consequent, alternate) {
  	        this.type = syntax_1.Syntax.ConditionalExpression;
  	        this.test = test;
  	        this.consequent = consequent;
  	        this.alternate = alternate;
  	    }
  	    return ConditionalExpression;
  	}());
  	exports.ConditionalExpression = ConditionalExpression;
  	var ContinueStatement = (function () {
  	    function ContinueStatement(label) {
  	        this.type = syntax_1.Syntax.ContinueStatement;
  	        this.label = label;
  	    }
  	    return ContinueStatement;
  	}());
  	exports.ContinueStatement = ContinueStatement;
  	var DebuggerStatement = (function () {
  	    function DebuggerStatement() {
  	        this.type = syntax_1.Syntax.DebuggerStatement;
  	    }
  	    return DebuggerStatement;
  	}());
  	exports.DebuggerStatement = DebuggerStatement;
  	var Directive = (function () {
  	    function Directive(expression, directive) {
  	        this.type = syntax_1.Syntax.ExpressionStatement;
  	        this.expression = expression;
  	        this.directive = directive;
  	    }
  	    return Directive;
  	}());
  	exports.Directive = Directive;
  	var DoWhileStatement = (function () {
  	    function DoWhileStatement(body, test) {
  	        this.type = syntax_1.Syntax.DoWhileStatement;
  	        this.body = body;
  	        this.test = test;
  	    }
  	    return DoWhileStatement;
  	}());
  	exports.DoWhileStatement = DoWhileStatement;
  	var EmptyStatement = (function () {
  	    function EmptyStatement() {
  	        this.type = syntax_1.Syntax.EmptyStatement;
  	    }
  	    return EmptyStatement;
  	}());
  	exports.EmptyStatement = EmptyStatement;
  	var ExportAllDeclaration = (function () {
  	    function ExportAllDeclaration(source) {
  	        this.type = syntax_1.Syntax.ExportAllDeclaration;
  	        this.source = source;
  	    }
  	    return ExportAllDeclaration;
  	}());
  	exports.ExportAllDeclaration = ExportAllDeclaration;
  	var ExportDefaultDeclaration = (function () {
  	    function ExportDefaultDeclaration(declaration) {
  	        this.type = syntax_1.Syntax.ExportDefaultDeclaration;
  	        this.declaration = declaration;
  	    }
  	    return ExportDefaultDeclaration;
  	}());
  	exports.ExportDefaultDeclaration = ExportDefaultDeclaration;
  	var ExportNamedDeclaration = (function () {
  	    function ExportNamedDeclaration(declaration, specifiers, source) {
  	        this.type = syntax_1.Syntax.ExportNamedDeclaration;
  	        this.declaration = declaration;
  	        this.specifiers = specifiers;
  	        this.source = source;
  	    }
  	    return ExportNamedDeclaration;
  	}());
  	exports.ExportNamedDeclaration = ExportNamedDeclaration;
  	var ExportSpecifier = (function () {
  	    function ExportSpecifier(local, exported) {
  	        this.type = syntax_1.Syntax.ExportSpecifier;
  	        this.exported = exported;
  	        this.local = local;
  	    }
  	    return ExportSpecifier;
  	}());
  	exports.ExportSpecifier = ExportSpecifier;
  	var ExpressionStatement = (function () {
  	    function ExpressionStatement(expression) {
  	        this.type = syntax_1.Syntax.ExpressionStatement;
  	        this.expression = expression;
  	    }
  	    return ExpressionStatement;
  	}());
  	exports.ExpressionStatement = ExpressionStatement;
  	var ForInStatement = (function () {
  	    function ForInStatement(left, right, body) {
  	        this.type = syntax_1.Syntax.ForInStatement;
  	        this.left = left;
  	        this.right = right;
  	        this.body = body;
  	        this.each = false;
  	    }
  	    return ForInStatement;
  	}());
  	exports.ForInStatement = ForInStatement;
  	var ForOfStatement = (function () {
  	    function ForOfStatement(left, right, body) {
  	        this.type = syntax_1.Syntax.ForOfStatement;
  	        this.left = left;
  	        this.right = right;
  	        this.body = body;
  	    }
  	    return ForOfStatement;
  	}());
  	exports.ForOfStatement = ForOfStatement;
  	var ForStatement = (function () {
  	    function ForStatement(init, test, update, body) {
  	        this.type = syntax_1.Syntax.ForStatement;
  	        this.init = init;
  	        this.test = test;
  	        this.update = update;
  	        this.body = body;
  	    }
  	    return ForStatement;
  	}());
  	exports.ForStatement = ForStatement;
  	var FunctionDeclaration = (function () {
  	    function FunctionDeclaration(id, params, body, generator) {
  	        this.type = syntax_1.Syntax.FunctionDeclaration;
  	        this.id = id;
  	        this.params = params;
  	        this.body = body;
  	        this.generator = generator;
  	        this.expression = false;
  	        this.async = false;
  	    }
  	    return FunctionDeclaration;
  	}());
  	exports.FunctionDeclaration = FunctionDeclaration;
  	var FunctionExpression = (function () {
  	    function FunctionExpression(id, params, body, generator) {
  	        this.type = syntax_1.Syntax.FunctionExpression;
  	        this.id = id;
  	        this.params = params;
  	        this.body = body;
  	        this.generator = generator;
  	        this.expression = false;
  	        this.async = false;
  	    }
  	    return FunctionExpression;
  	}());
  	exports.FunctionExpression = FunctionExpression;
  	var Identifier = (function () {
  	    function Identifier(name) {
  	        this.type = syntax_1.Syntax.Identifier;
  	        this.name = name;
  	    }
  	    return Identifier;
  	}());
  	exports.Identifier = Identifier;
  	var IfStatement = (function () {
  	    function IfStatement(test, consequent, alternate) {
  	        this.type = syntax_1.Syntax.IfStatement;
  	        this.test = test;
  	        this.consequent = consequent;
  	        this.alternate = alternate;
  	    }
  	    return IfStatement;
  	}());
  	exports.IfStatement = IfStatement;
  	var ImportDeclaration = (function () {
  	    function ImportDeclaration(specifiers, source) {
  	        this.type = syntax_1.Syntax.ImportDeclaration;
  	        this.specifiers = specifiers;
  	        this.source = source;
  	    }
  	    return ImportDeclaration;
  	}());
  	exports.ImportDeclaration = ImportDeclaration;
  	var ImportDefaultSpecifier = (function () {
  	    function ImportDefaultSpecifier(local) {
  	        this.type = syntax_1.Syntax.ImportDefaultSpecifier;
  	        this.local = local;
  	    }
  	    return ImportDefaultSpecifier;
  	}());
  	exports.ImportDefaultSpecifier = ImportDefaultSpecifier;
  	var ImportNamespaceSpecifier = (function () {
  	    function ImportNamespaceSpecifier(local) {
  	        this.type = syntax_1.Syntax.ImportNamespaceSpecifier;
  	        this.local = local;
  	    }
  	    return ImportNamespaceSpecifier;
  	}());
  	exports.ImportNamespaceSpecifier = ImportNamespaceSpecifier;
  	var ImportSpecifier = (function () {
  	    function ImportSpecifier(local, imported) {
  	        this.type = syntax_1.Syntax.ImportSpecifier;
  	        this.local = local;
  	        this.imported = imported;
  	    }
  	    return ImportSpecifier;
  	}());
  	exports.ImportSpecifier = ImportSpecifier;
  	var LabeledStatement = (function () {
  	    function LabeledStatement(label, body) {
  	        this.type = syntax_1.Syntax.LabeledStatement;
  	        this.label = label;
  	        this.body = body;
  	    }
  	    return LabeledStatement;
  	}());
  	exports.LabeledStatement = LabeledStatement;
  	var Literal = (function () {
  	    function Literal(value, raw) {
  	        this.type = syntax_1.Syntax.Literal;
  	        this.value = value;
  	        this.raw = raw;
  	    }
  	    return Literal;
  	}());
  	exports.Literal = Literal;
  	var MetaProperty = (function () {
  	    function MetaProperty(meta, property) {
  	        this.type = syntax_1.Syntax.MetaProperty;
  	        this.meta = meta;
  	        this.property = property;
  	    }
  	    return MetaProperty;
  	}());
  	exports.MetaProperty = MetaProperty;
  	var MethodDefinition = (function () {
  	    function MethodDefinition(key, computed, value, kind, isStatic) {
  	        this.type = syntax_1.Syntax.MethodDefinition;
  	        this.key = key;
  	        this.computed = computed;
  	        this.value = value;
  	        this.kind = kind;
  	        this.static = isStatic;
  	    }
  	    return MethodDefinition;
  	}());
  	exports.MethodDefinition = MethodDefinition;
  	var Module = (function () {
  	    function Module(body) {
  	        this.type = syntax_1.Syntax.Program;
  	        this.body = body;
  	        this.sourceType = 'module';
  	    }
  	    return Module;
  	}());
  	exports.Module = Module;
  	var NewExpression = (function () {
  	    function NewExpression(callee, args) {
  	        this.type = syntax_1.Syntax.NewExpression;
  	        this.callee = callee;
  	        this.arguments = args;
  	    }
  	    return NewExpression;
  	}());
  	exports.NewExpression = NewExpression;
  	var ObjectExpression = (function () {
  	    function ObjectExpression(properties) {
  	        this.type = syntax_1.Syntax.ObjectExpression;
  	        this.properties = properties;
  	    }
  	    return ObjectExpression;
  	}());
  	exports.ObjectExpression = ObjectExpression;
  	var ObjectPattern = (function () {
  	    function ObjectPattern(properties) {
  	        this.type = syntax_1.Syntax.ObjectPattern;
  	        this.properties = properties;
  	    }
  	    return ObjectPattern;
  	}());
  	exports.ObjectPattern = ObjectPattern;
  	var Property = (function () {
  	    function Property(kind, key, computed, value, method, shorthand) {
  	        this.type = syntax_1.Syntax.Property;
  	        this.key = key;
  	        this.computed = computed;
  	        this.value = value;
  	        this.kind = kind;
  	        this.method = method;
  	        this.shorthand = shorthand;
  	    }
  	    return Property;
  	}());
  	exports.Property = Property;
  	var RegexLiteral = (function () {
  	    function RegexLiteral(value, raw, pattern, flags) {
  	        this.type = syntax_1.Syntax.Literal;
  	        this.value = value;
  	        this.raw = raw;
  	        this.regex = { pattern: pattern, flags: flags };
  	    }
  	    return RegexLiteral;
  	}());
  	exports.RegexLiteral = RegexLiteral;
  	var RestElement = (function () {
  	    function RestElement(argument) {
  	        this.type = syntax_1.Syntax.RestElement;
  	        this.argument = argument;
  	    }
  	    return RestElement;
  	}());
  	exports.RestElement = RestElement;
  	var ReturnStatement = (function () {
  	    function ReturnStatement(argument) {
  	        this.type = syntax_1.Syntax.ReturnStatement;
  	        this.argument = argument;
  	    }
  	    return ReturnStatement;
  	}());
  	exports.ReturnStatement = ReturnStatement;
  	var Script = (function () {
  	    function Script(body) {
  	        this.type = syntax_1.Syntax.Program;
  	        this.body = body;
  	        this.sourceType = 'script';
  	    }
  	    return Script;
  	}());
  	exports.Script = Script;
  	var SequenceExpression = (function () {
  	    function SequenceExpression(expressions) {
  	        this.type = syntax_1.Syntax.SequenceExpression;
  	        this.expressions = expressions;
  	    }
  	    return SequenceExpression;
  	}());
  	exports.SequenceExpression = SequenceExpression;
  	var SpreadElement = (function () {
  	    function SpreadElement(argument) {
  	        this.type = syntax_1.Syntax.SpreadElement;
  	        this.argument = argument;
  	    }
  	    return SpreadElement;
  	}());
  	exports.SpreadElement = SpreadElement;
  	var StaticMemberExpression = (function () {
  	    function StaticMemberExpression(object, property) {
  	        this.type = syntax_1.Syntax.MemberExpression;
  	        this.computed = false;
  	        this.object = object;
  	        this.property = property;
  	    }
  	    return StaticMemberExpression;
  	}());
  	exports.StaticMemberExpression = StaticMemberExpression;
  	var Super = (function () {
  	    function Super() {
  	        this.type = syntax_1.Syntax.Super;
  	    }
  	    return Super;
  	}());
  	exports.Super = Super;
  	var SwitchCase = (function () {
  	    function SwitchCase(test, consequent) {
  	        this.type = syntax_1.Syntax.SwitchCase;
  	        this.test = test;
  	        this.consequent = consequent;
  	    }
  	    return SwitchCase;
  	}());
  	exports.SwitchCase = SwitchCase;
  	var SwitchStatement = (function () {
  	    function SwitchStatement(discriminant, cases) {
  	        this.type = syntax_1.Syntax.SwitchStatement;
  	        this.discriminant = discriminant;
  	        this.cases = cases;
  	    }
  	    return SwitchStatement;
  	}());
  	exports.SwitchStatement = SwitchStatement;
  	var TaggedTemplateExpression = (function () {
  	    function TaggedTemplateExpression(tag, quasi) {
  	        this.type = syntax_1.Syntax.TaggedTemplateExpression;
  	        this.tag = tag;
  	        this.quasi = quasi;
  	    }
  	    return TaggedTemplateExpression;
  	}());
  	exports.TaggedTemplateExpression = TaggedTemplateExpression;
  	var TemplateElement = (function () {
  	    function TemplateElement(value, tail) {
  	        this.type = syntax_1.Syntax.TemplateElement;
  	        this.value = value;
  	        this.tail = tail;
  	    }
  	    return TemplateElement;
  	}());
  	exports.TemplateElement = TemplateElement;
  	var TemplateLiteral = (function () {
  	    function TemplateLiteral(quasis, expressions) {
  	        this.type = syntax_1.Syntax.TemplateLiteral;
  	        this.quasis = quasis;
  	        this.expressions = expressions;
  	    }
  	    return TemplateLiteral;
  	}());
  	exports.TemplateLiteral = TemplateLiteral;
  	var ThisExpression = (function () {
  	    function ThisExpression() {
  	        this.type = syntax_1.Syntax.ThisExpression;
  	    }
  	    return ThisExpression;
  	}());
  	exports.ThisExpression = ThisExpression;
  	var ThrowStatement = (function () {
  	    function ThrowStatement(argument) {
  	        this.type = syntax_1.Syntax.ThrowStatement;
  	        this.argument = argument;
  	    }
  	    return ThrowStatement;
  	}());
  	exports.ThrowStatement = ThrowStatement;
  	var TryStatement = (function () {
  	    function TryStatement(block, handler, finalizer) {
  	        this.type = syntax_1.Syntax.TryStatement;
  	        this.block = block;
  	        this.handler = handler;
  	        this.finalizer = finalizer;
  	    }
  	    return TryStatement;
  	}());
  	exports.TryStatement = TryStatement;
  	var UnaryExpression = (function () {
  	    function UnaryExpression(operator, argument) {
  	        this.type = syntax_1.Syntax.UnaryExpression;
  	        this.operator = operator;
  	        this.argument = argument;
  	        this.prefix = true;
  	    }
  	    return UnaryExpression;
  	}());
  	exports.UnaryExpression = UnaryExpression;
  	var UpdateExpression = (function () {
  	    function UpdateExpression(operator, argument, prefix) {
  	        this.type = syntax_1.Syntax.UpdateExpression;
  	        this.operator = operator;
  	        this.argument = argument;
  	        this.prefix = prefix;
  	    }
  	    return UpdateExpression;
  	}());
  	exports.UpdateExpression = UpdateExpression;
  	var VariableDeclaration = (function () {
  	    function VariableDeclaration(declarations, kind) {
  	        this.type = syntax_1.Syntax.VariableDeclaration;
  	        this.declarations = declarations;
  	        this.kind = kind;
  	    }
  	    return VariableDeclaration;
  	}());
  	exports.VariableDeclaration = VariableDeclaration;
  	var VariableDeclarator = (function () {
  	    function VariableDeclarator(id, init) {
  	        this.type = syntax_1.Syntax.VariableDeclarator;
  	        this.id = id;
  	        this.init = init;
  	    }
  	    return VariableDeclarator;
  	}());
  	exports.VariableDeclarator = VariableDeclarator;
  	var WhileStatement = (function () {
  	    function WhileStatement(test, body) {
  	        this.type = syntax_1.Syntax.WhileStatement;
  	        this.test = test;
  	        this.body = body;
  	    }
  	    return WhileStatement;
  	}());
  	exports.WhileStatement = WhileStatement;
  	var WithStatement = (function () {
  	    function WithStatement(object, body) {
  	        this.type = syntax_1.Syntax.WithStatement;
  	        this.object = object;
  	        this.body = body;
  	    }
  	    return WithStatement;
  	}());
  	exports.WithStatement = WithStatement;
  	var YieldExpression = (function () {
  	    function YieldExpression(argument, delegate) {
  	        this.type = syntax_1.Syntax.YieldExpression;
  	        this.argument = argument;
  	        this.delegate = delegate;
  	    }
  	    return YieldExpression;
  	}());
  	exports.YieldExpression = YieldExpression;


  /***/ },
  /* 8 */
  /***/ function(module, exports, __webpack_require__) {
  	Object.defineProperty(exports, "__esModule", { value: true });
  	var assert_1 = __webpack_require__(9);
  	var error_handler_1 = __webpack_require__(10);
  	var messages_1 = __webpack_require__(11);
  	var Node = __webpack_require__(7);
  	var scanner_1 = __webpack_require__(12);
  	var syntax_1 = __webpack_require__(2);
  	var token_1 = __webpack_require__(13);
  	var ArrowParameterPlaceHolder = 'ArrowParameterPlaceHolder';
  	var Parser = (function () {
  	    function Parser(code, options, delegate) {
  	        if (options === void 0) { options = {}; }
  	        this.config = {
  	            range: (typeof options.range === 'boolean') && options.range,
  	            loc: (typeof options.loc === 'boolean') && options.loc,
  	            source: null,
  	            tokens: (typeof options.tokens === 'boolean') && options.tokens,
  	            comment: (typeof options.comment === 'boolean') && options.comment,
  	            tolerant: (typeof options.tolerant === 'boolean') && options.tolerant
  	        };
  	        if (this.config.loc && options.source && options.source !== null) {
  	            this.config.source = String(options.source);
  	        }
  	        this.delegate = delegate;
  	        this.errorHandler = new error_handler_1.ErrorHandler();
  	        this.errorHandler.tolerant = this.config.tolerant;
  	        this.scanner = new scanner_1.Scanner(code, this.errorHandler);
  	        this.scanner.trackComment = this.config.comment;
  	        this.operatorPrecedence = {
  	            ')': 0,
  	            ';': 0,
  	            ',': 0,
  	            '=': 0,
  	            ']': 0,
  	            '||': 1,
  	            '&&': 2,
  	            '|': 3,
  	            '^': 4,
  	            '&': 5,
  	            '==': 6,
  	            '!=': 6,
  	            '===': 6,
  	            '!==': 6,
  	            '<': 7,
  	            '>': 7,
  	            '<=': 7,
  	            '>=': 7,
  	            '<<': 8,
  	            '>>': 8,
  	            '>>>': 8,
  	            '+': 9,
  	            '-': 9,
  	            '*': 11,
  	            '/': 11,
  	            '%': 11
  	        };
  	        this.lookahead = {
  	            type: 2 /* EOF */,
  	            value: '',
  	            lineNumber: this.scanner.lineNumber,
  	            lineStart: 0,
  	            start: 0,
  	            end: 0
  	        };
  	        this.hasLineTerminator = false;
  	        this.context = {
  	            isModule: false,
  	            await: false,
  	            allowIn: true,
  	            allowStrictDirective: true,
  	            allowYield: true,
  	            firstCoverInitializedNameError: null,
  	            isAssignmentTarget: false,
  	            isBindingElement: false,
  	            inFunctionBody: false,
  	            inIteration: false,
  	            inSwitch: false,
  	            labelSet: {},
  	            strict: false
  	        };
  	        this.tokens = [];
  	        this.startMarker = {
  	            index: 0,
  	            line: this.scanner.lineNumber,
  	            column: 0
  	        };
  	        this.lastMarker = {
  	            index: 0,
  	            line: this.scanner.lineNumber,
  	            column: 0
  	        };
  	        this.nextToken();
  	        this.lastMarker = {
  	            index: this.scanner.index,
  	            line: this.scanner.lineNumber,
  	            column: this.scanner.index - this.scanner.lineStart
  	        };
  	    }
  	    Parser.prototype.throwError = function (messageFormat) {
  	        var values = [];
  	        for (var _i = 1; _i < arguments.length; _i++) {
  	            values[_i - 1] = arguments[_i];
  	        }
  	        var args = Array.prototype.slice.call(arguments, 1);
  	        var msg = messageFormat.replace(/%(\d)/g, function (whole, idx) {
  	            assert_1.assert(idx < args.length, 'Message reference must be in range');
  	            return args[idx];
  	        });
  	        var index = this.lastMarker.index;
  	        var line = this.lastMarker.line;
  	        var column = this.lastMarker.column + 1;
  	        throw this.errorHandler.createError(index, line, column, msg);
  	    };
  	    Parser.prototype.tolerateError = function (messageFormat) {
  	        var values = [];
  	        for (var _i = 1; _i < arguments.length; _i++) {
  	            values[_i - 1] = arguments[_i];
  	        }
  	        var args = Array.prototype.slice.call(arguments, 1);
  	        var msg = messageFormat.replace(/%(\d)/g, function (whole, idx) {
  	            assert_1.assert(idx < args.length, 'Message reference must be in range');
  	            return args[idx];
  	        });
  	        var index = this.lastMarker.index;
  	        var line = this.scanner.lineNumber;
  	        var column = this.lastMarker.column + 1;
  	        this.errorHandler.tolerateError(index, line, column, msg);
  	    };
  	    // Throw an exception because of the token.
  	    Parser.prototype.unexpectedTokenError = function (token, message) {
  	        var msg = message || messages_1.Messages.UnexpectedToken;
  	        var value;
  	        if (token) {
  	            if (!message) {
  	                msg = (token.type === 2 /* EOF */) ? messages_1.Messages.UnexpectedEOS :
  	                    (token.type === 3 /* Identifier */) ? messages_1.Messages.UnexpectedIdentifier :
  	                        (token.type === 6 /* NumericLiteral */) ? messages_1.Messages.UnexpectedNumber :
  	                            (token.type === 8 /* StringLiteral */) ? messages_1.Messages.UnexpectedString :
  	                                (token.type === 10 /* Template */) ? messages_1.Messages.UnexpectedTemplate :
  	                                    messages_1.Messages.UnexpectedToken;
  	                if (token.type === 4 /* Keyword */) {
  	                    if (this.scanner.isFutureReservedWord(token.value)) {
  	                        msg = messages_1.Messages.UnexpectedReserved;
  	                    }
  	                    else if (this.context.strict && this.scanner.isStrictModeReservedWord(token.value)) {
  	                        msg = messages_1.Messages.StrictReservedWord;
  	                    }
  	                }
  	            }
  	            value = token.value;
  	        }
  	        else {
  	            value = 'ILLEGAL';
  	        }
  	        msg = msg.replace('%0', value);
  	        if (token && typeof token.lineNumber === 'number') {
  	            var index = token.start;
  	            var line = token.lineNumber;
  	            var lastMarkerLineStart = this.lastMarker.index - this.lastMarker.column;
  	            var column = token.start - lastMarkerLineStart + 1;
  	            return this.errorHandler.createError(index, line, column, msg);
  	        }
  	        else {
  	            var index = this.lastMarker.index;
  	            var line = this.lastMarker.line;
  	            var column = this.lastMarker.column + 1;
  	            return this.errorHandler.createError(index, line, column, msg);
  	        }
  	    };
  	    Parser.prototype.throwUnexpectedToken = function (token, message) {
  	        throw this.unexpectedTokenError(token, message);
  	    };
  	    Parser.prototype.tolerateUnexpectedToken = function (token, message) {
  	        this.errorHandler.tolerate(this.unexpectedTokenError(token, message));
  	    };
  	    Parser.prototype.collectComments = function () {
  	        if (!this.config.comment) {
  	            this.scanner.scanComments();
  	        }
  	        else {
  	            var comments = this.scanner.scanComments();
  	            if (comments.length > 0 && this.delegate) {
  	                for (var i = 0; i < comments.length; ++i) {
  	                    var e = comments[i];
  	                    var node = void 0;
  	                    node = {
  	                        type: e.multiLine ? 'BlockComment' : 'LineComment',
  	                        value: this.scanner.source.slice(e.slice[0], e.slice[1])
  	                    };
  	                    if (this.config.range) {
  	                        node.range = e.range;
  	                    }
  	                    if (this.config.loc) {
  	                        node.loc = e.loc;
  	                    }
  	                    var metadata = {
  	                        start: {
  	                            line: e.loc.start.line,
  	                            column: e.loc.start.column,
  	                            offset: e.range[0]
  	                        },
  	                        end: {
  	                            line: e.loc.end.line,
  	                            column: e.loc.end.column,
  	                            offset: e.range[1]
  	                        }
  	                    };
  	                    this.delegate(node, metadata);
  	                }
  	            }
  	        }
  	    };
  	    // From internal representation to an external structure
  	    Parser.prototype.getTokenRaw = function (token) {
  	        return this.scanner.source.slice(token.start, token.end);
  	    };
  	    Parser.prototype.convertToken = function (token) {
  	        var t = {
  	            type: token_1.TokenName[token.type],
  	            value: this.getTokenRaw(token)
  	        };
  	        if (this.config.range) {
  	            t.range = [token.start, token.end];
  	        }
  	        if (this.config.loc) {
  	            t.loc = {
  	                start: {
  	                    line: this.startMarker.line,
  	                    column: this.startMarker.column
  	                },
  	                end: {
  	                    line: this.scanner.lineNumber,
  	                    column: this.scanner.index - this.scanner.lineStart
  	                }
  	            };
  	        }
  	        if (token.type === 9 /* RegularExpression */) {
  	            var pattern = token.pattern;
  	            var flags = token.flags;
  	            t.regex = { pattern: pattern, flags: flags };
  	        }
  	        return t;
  	    };
  	    Parser.prototype.nextToken = function () {
  	        var token = this.lookahead;
  	        this.lastMarker.index = this.scanner.index;
  	        this.lastMarker.line = this.scanner.lineNumber;
  	        this.lastMarker.column = this.scanner.index - this.scanner.lineStart;
  	        this.collectComments();
  	        if (this.scanner.index !== this.startMarker.index) {
  	            this.startMarker.index = this.scanner.index;
  	            this.startMarker.line = this.scanner.lineNumber;
  	            this.startMarker.column = this.scanner.index - this.scanner.lineStart;
  	        }
  	        var next = this.scanner.lex();
  	        this.hasLineTerminator = (token.lineNumber !== next.lineNumber);
  	        if (next && this.context.strict && next.type === 3 /* Identifier */) {
  	            if (this.scanner.isStrictModeReservedWord(next.value)) {
  	                next.type = 4 /* Keyword */;
  	            }
  	        }
  	        this.lookahead = next;
  	        if (this.config.tokens && next.type !== 2 /* EOF */) {
  	            this.tokens.push(this.convertToken(next));
  	        }
  	        return token;
  	    };
  	    Parser.prototype.nextRegexToken = function () {
  	        this.collectComments();
  	        var token = this.scanner.scanRegExp();
  	        if (this.config.tokens) {
  	            // Pop the previous token, '/' or '/='
  	            // This is added from the lookahead token.
  	            this.tokens.pop();
  	            this.tokens.push(this.convertToken(token));
  	        }
  	        // Prime the next lookahead.
  	        this.lookahead = token;
  	        this.nextToken();
  	        return token;
  	    };
  	    Parser.prototype.createNode = function () {
  	        return {
  	            index: this.startMarker.index,
  	            line: this.startMarker.line,
  	            column: this.startMarker.column
  	        };
  	    };
  	    Parser.prototype.startNode = function (token) {
  	        return {
  	            index: token.start,
  	            line: token.lineNumber,
  	            column: token.start - token.lineStart
  	        };
  	    };
  	    Parser.prototype.finalize = function (marker, node) {
  	        if (this.config.range) {
  	            node.range = [marker.index, this.lastMarker.index];
  	        }
  	        if (this.config.loc) {
  	            node.loc = {
  	                start: {
  	                    line: marker.line,
  	                    column: marker.column,
  	                },
  	                end: {
  	                    line: this.lastMarker.line,
  	                    column: this.lastMarker.column
  	                }
  	            };
  	            if (this.config.source) {
  	                node.loc.source = this.config.source;
  	            }
  	        }
  	        if (this.delegate) {
  	            var metadata = {
  	                start: {
  	                    line: marker.line,
  	                    column: marker.column,
  	                    offset: marker.index
  	                },
  	                end: {
  	                    line: this.lastMarker.line,
  	                    column: this.lastMarker.column,
  	                    offset: this.lastMarker.index
  	                }
  	            };
  	            this.delegate(node, metadata);
  	        }
  	        return node;
  	    };
  	    // Expect the next token to match the specified punctuator.
  	    // If not, an exception will be thrown.
  	    Parser.prototype.expect = function (value) {
  	        var token = this.nextToken();
  	        if (token.type !== 7 /* Punctuator */ || token.value !== value) {
  	            this.throwUnexpectedToken(token);
  	        }
  	    };
  	    // Quietly expect a comma when in tolerant mode, otherwise delegates to expect().
  	    Parser.prototype.expectCommaSeparator = function () {
  	        if (this.config.tolerant) {
  	            var token = this.lookahead;
  	            if (token.type === 7 /* Punctuator */ && token.value === ',') {
  	                this.nextToken();
  	            }
  	            else if (token.type === 7 /* Punctuator */ && token.value === ';') {
  	                this.nextToken();
  	                this.tolerateUnexpectedToken(token);
  	            }
  	            else {
  	                this.tolerateUnexpectedToken(token, messages_1.Messages.UnexpectedToken);
  	            }
  	        }
  	        else {
  	            this.expect(',');
  	        }
  	    };
  	    // Expect the next token to match the specified keyword.
  	    // If not, an exception will be thrown.
  	    Parser.prototype.expectKeyword = function (keyword) {
  	        var token = this.nextToken();
  	        if (token.type !== 4 /* Keyword */ || token.value !== keyword) {
  	            this.throwUnexpectedToken(token);
  	        }
  	    };
  	    // Return true if the next token matches the specified punctuator.
  	    Parser.prototype.match = function (value) {
  	        return this.lookahead.type === 7 /* Punctuator */ && this.lookahead.value === value;
  	    };
  	    // Return true if the next token matches the specified keyword
  	    Parser.prototype.matchKeyword = function (keyword) {
  	        return this.lookahead.type === 4 /* Keyword */ && this.lookahead.value === keyword;
  	    };
  	    // Return true if the next token matches the specified contextual keyword
  	    // (where an identifier is sometimes a keyword depending on the context)
  	    Parser.prototype.matchContextualKeyword = function (keyword) {
  	        return this.lookahead.type === 3 /* Identifier */ && this.lookahead.value === keyword;
  	    };
  	    // Return true if the next token is an assignment operator
  	    Parser.prototype.matchAssign = function () {
  	        if (this.lookahead.type !== 7 /* Punctuator */) {
  	            return false;
  	        }
  	        var op = this.lookahead.value;
  	        return op === '=' ||
  	            op === '*=' ||
  	            op === '**=' ||
  	            op === '/=' ||
  	            op === '%=' ||
  	            op === '+=' ||
  	            op === '-=' ||
  	            op === '<<=' ||
  	            op === '>>=' ||
  	            op === '>>>=' ||
  	            op === '&=' ||
  	            op === '^=' ||
  	            op === '|=';
  	    };
  	    // Cover grammar support.
  	    //
  	    // When an assignment expression position starts with an left parenthesis, the determination of the type
  	    // of the syntax is to be deferred arbitrarily long until the end of the parentheses pair (plus a lookahead)
  	    // or the first comma. This situation also defers the determination of all the expressions nested in the pair.
  	    //
  	    // There are three productions that can be parsed in a parentheses pair that needs to be determined
  	    // after the outermost pair is closed. They are:
  	    //
  	    //   1. AssignmentExpression
  	    //   2. BindingElements
  	    //   3. AssignmentTargets
  	    //
  	    // In order to avoid exponential backtracking, we use two flags to denote if the production can be
  	    // binding element or assignment target.
  	    //
  	    // The three productions have the relationship:
  	    //
  	    //   BindingElements ⊆ AssignmentTargets ⊆ AssignmentExpression
  	    //
  	    // with a single exception that CoverInitializedName when used directly in an Expression, generates
  	    // an early error. Therefore, we need the third state, firstCoverInitializedNameError, to track the
  	    // first usage of CoverInitializedName and report it when we reached the end of the parentheses pair.
  	    //
  	    // isolateCoverGrammar function runs the given parser function with a new cover grammar context, and it does not
  	    // effect the current flags. This means the production the parser parses is only used as an expression. Therefore
  	    // the CoverInitializedName check is conducted.
  	    //
  	    // inheritCoverGrammar function runs the given parse function with a new cover grammar context, and it propagates
  	    // the flags outside of the parser. This means the production the parser parses is used as a part of a potential
  	    // pattern. The CoverInitializedName check is deferred.
  	    Parser.prototype.isolateCoverGrammar = function (parseFunction) {
  	        var previousIsBindingElement = this.context.isBindingElement;
  	        var previousIsAssignmentTarget = this.context.isAssignmentTarget;
  	        var previousFirstCoverInitializedNameError = this.context.firstCoverInitializedNameError;
  	        this.context.isBindingElement = true;
  	        this.context.isAssignmentTarget = true;
  	        this.context.firstCoverInitializedNameError = null;
  	        var result = parseFunction.call(this);
  	        if (this.context.firstCoverInitializedNameError !== null) {
  	            this.throwUnexpectedToken(this.context.firstCoverInitializedNameError);
  	        }
  	        this.context.isBindingElement = previousIsBindingElement;
  	        this.context.isAssignmentTarget = previousIsAssignmentTarget;
  	        this.context.firstCoverInitializedNameError = previousFirstCoverInitializedNameError;
  	        return result;
  	    };
  	    Parser.prototype.inheritCoverGrammar = function (parseFunction) {
  	        var previousIsBindingElement = this.context.isBindingElement;
  	        var previousIsAssignmentTarget = this.context.isAssignmentTarget;
  	        var previousFirstCoverInitializedNameError = this.context.firstCoverInitializedNameError;
  	        this.context.isBindingElement = true;
  	        this.context.isAssignmentTarget = true;
  	        this.context.firstCoverInitializedNameError = null;
  	        var result = parseFunction.call(this);
  	        this.context.isBindingElement = this.context.isBindingElement && previousIsBindingElement;
  	        this.context.isAssignmentTarget = this.context.isAssignmentTarget && previousIsAssignmentTarget;
  	        this.context.firstCoverInitializedNameError = previousFirstCoverInitializedNameError || this.context.firstCoverInitializedNameError;
  	        return result;
  	    };
  	    Parser.prototype.consumeSemicolon = function () {
  	        if (this.match(';')) {
  	            this.nextToken();
  	        }
  	        else if (!this.hasLineTerminator) {
  	            if (this.lookahead.type !== 2 /* EOF */ && !this.match('}')) {
  	                this.throwUnexpectedToken(this.lookahead);
  	            }
  	            this.lastMarker.index = this.startMarker.index;
  	            this.lastMarker.line = this.startMarker.line;
  	            this.lastMarker.column = this.startMarker.column;
  	        }
  	    };
  	    // https://tc39.github.io/ecma262/#sec-primary-expression
  	    Parser.prototype.parsePrimaryExpression = function () {
  	        var node = this.createNode();
  	        var expr;
  	        var token, raw;
  	        switch (this.lookahead.type) {
  	            case 3 /* Identifier */:
  	                if ((this.context.isModule || this.context.await) && this.lookahead.value === 'await') {
  	                    this.tolerateUnexpectedToken(this.lookahead);
  	                }
  	                expr = this.matchAsyncFunction() ? this.parseFunctionExpression() : this.finalize(node, new Node.Identifier(this.nextToken().value));
  	                break;
  	            case 6 /* NumericLiteral */:
  	            case 8 /* StringLiteral */:
  	                if (this.context.strict && this.lookahead.octal) {
  	                    this.tolerateUnexpectedToken(this.lookahead, messages_1.Messages.StrictOctalLiteral);
  	                }
  	                this.context.isAssignmentTarget = false;
  	                this.context.isBindingElement = false;
  	                token = this.nextToken();
  	                raw = this.getTokenRaw(token);
  	                expr = this.finalize(node, new Node.Literal(token.value, raw));
  	                break;
  	            case 1 /* BooleanLiteral */:
  	                this.context.isAssignmentTarget = false;
  	                this.context.isBindingElement = false;
  	                token = this.nextToken();
  	                raw = this.getTokenRaw(token);
  	                expr = this.finalize(node, new Node.Literal(token.value === 'true', raw));
  	                break;
  	            case 5 /* NullLiteral */:
  	                this.context.isAssignmentTarget = false;
  	                this.context.isBindingElement = false;
  	                token = this.nextToken();
  	                raw = this.getTokenRaw(token);
  	                expr = this.finalize(node, new Node.Literal(null, raw));
  	                break;
  	            case 10 /* Template */:
  	                expr = this.parseTemplateLiteral();
  	                break;
  	            case 7 /* Punctuator */:
  	                switch (this.lookahead.value) {
  	                    case '(':
  	                        this.context.isBindingElement = false;
  	                        expr = this.inheritCoverGrammar(this.parseGroupExpression);
  	                        break;
  	                    case '[':
  	                        expr = this.inheritCoverGrammar(this.parseArrayInitializer);
  	                        break;
  	                    case '{':
  	                        expr = this.inheritCoverGrammar(this.parseObjectInitializer);
  	                        break;
  	                    case '/':
  	                    case '/=':
  	                        this.context.isAssignmentTarget = false;
  	                        this.context.isBindingElement = false;
  	                        this.scanner.index = this.startMarker.index;
  	                        token = this.nextRegexToken();
  	                        raw = this.getTokenRaw(token);
  	                        expr = this.finalize(node, new Node.RegexLiteral(token.regex, raw, token.pattern, token.flags));
  	                        break;
  	                    default:
  	                        expr = this.throwUnexpectedToken(this.nextToken());
  	                }
  	                break;
  	            case 4 /* Keyword */:
  	                if (!this.context.strict && this.context.allowYield && this.matchKeyword('yield')) {
  	                    expr = this.parseIdentifierName();
  	                }
  	                else if (!this.context.strict && this.matchKeyword('let')) {
  	                    expr = this.finalize(node, new Node.Identifier(this.nextToken().value));
  	                }
  	                else {
  	                    this.context.isAssignmentTarget = false;
  	                    this.context.isBindingElement = false;
  	                    if (this.matchKeyword('function')) {
  	                        expr = this.parseFunctionExpression();
  	                    }
  	                    else if (this.matchKeyword('this')) {
  	                        this.nextToken();
  	                        expr = this.finalize(node, new Node.ThisExpression());
  	                    }
  	                    else if (this.matchKeyword('class')) {
  	                        expr = this.parseClassExpression();
  	                    }
  	                    else {
  	                        expr = this.throwUnexpectedToken(this.nextToken());
  	                    }
  	                }
  	                break;
  	            default:
  	                expr = this.throwUnexpectedToken(this.nextToken());
  	        }
  	        return expr;
  	    };
  	    // https://tc39.github.io/ecma262/#sec-array-initializer
  	    Parser.prototype.parseSpreadElement = function () {
  	        var node = this.createNode();
  	        this.expect('...');
  	        var arg = this.inheritCoverGrammar(this.parseAssignmentExpression);
  	        return this.finalize(node, new Node.SpreadElement(arg));
  	    };
  	    Parser.prototype.parseArrayInitializer = function () {
  	        var node = this.createNode();
  	        var elements = [];
  	        this.expect('[');
  	        while (!this.match(']')) {
  	            if (this.match(',')) {
  	                this.nextToken();
  	                elements.push(null);
  	            }
  	            else if (this.match('...')) {
  	                var element = this.parseSpreadElement();
  	                if (!this.match(']')) {
  	                    this.context.isAssignmentTarget = false;
  	                    this.context.isBindingElement = false;
  	                    this.expect(',');
  	                }
  	                elements.push(element);
  	            }
  	            else {
  	                elements.push(this.inheritCoverGrammar(this.parseAssignmentExpression));
  	                if (!this.match(']')) {
  	                    this.expect(',');
  	                }
  	            }
  	        }
  	        this.expect(']');
  	        return this.finalize(node, new Node.ArrayExpression(elements));
  	    };
  	    // https://tc39.github.io/ecma262/#sec-object-initializer
  	    Parser.prototype.parsePropertyMethod = function (params) {
  	        this.context.isAssignmentTarget = false;
  	        this.context.isBindingElement = false;
  	        var previousStrict = this.context.strict;
  	        var previousAllowStrictDirective = this.context.allowStrictDirective;
  	        this.context.allowStrictDirective = params.simple;
  	        var body = this.isolateCoverGrammar(this.parseFunctionSourceElements);
  	        if (this.context.strict && params.firstRestricted) {
  	            this.tolerateUnexpectedToken(params.firstRestricted, params.message);
  	        }
  	        if (this.context.strict && params.stricted) {
  	            this.tolerateUnexpectedToken(params.stricted, params.message);
  	        }
  	        this.context.strict = previousStrict;
  	        this.context.allowStrictDirective = previousAllowStrictDirective;
  	        return body;
  	    };
  	    Parser.prototype.parsePropertyMethodFunction = function () {
  	        var isGenerator = false;
  	        var node = this.createNode();
  	        var previousAllowYield = this.context.allowYield;
  	        this.context.allowYield = false;
  	        var params = this.parseFormalParameters();
  	        var method = this.parsePropertyMethod(params);
  	        this.context.allowYield = previousAllowYield;
  	        return this.finalize(node, new Node.FunctionExpression(null, params.params, method, isGenerator));
  	    };
  	    Parser.prototype.parsePropertyMethodAsyncFunction = function () {
  	        var node = this.createNode();
  	        var previousAllowYield = this.context.allowYield;
  	        var previousAwait = this.context.await;
  	        this.context.allowYield = false;
  	        this.context.await = true;
  	        var params = this.parseFormalParameters();
  	        var method = this.parsePropertyMethod(params);
  	        this.context.allowYield = previousAllowYield;
  	        this.context.await = previousAwait;
  	        return this.finalize(node, new Node.AsyncFunctionExpression(null, params.params, method));
  	    };
  	    Parser.prototype.parseObjectPropertyKey = function () {
  	        var node = this.createNode();
  	        var token = this.nextToken();
  	        var key;
  	        switch (token.type) {
  	            case 8 /* StringLiteral */:
  	            case 6 /* NumericLiteral */:
  	                if (this.context.strict && token.octal) {
  	                    this.tolerateUnexpectedToken(token, messages_1.Messages.StrictOctalLiteral);
  	                }
  	                var raw = this.getTokenRaw(token);
  	                key = this.finalize(node, new Node.Literal(token.value, raw));
  	                break;
  	            case 3 /* Identifier */:
  	            case 1 /* BooleanLiteral */:
  	            case 5 /* NullLiteral */:
  	            case 4 /* Keyword */:
  	                key = this.finalize(node, new Node.Identifier(token.value));
  	                break;
  	            case 7 /* Punctuator */:
  	                if (token.value === '[') {
  	                    key = this.isolateCoverGrammar(this.parseAssignmentExpression);
  	                    this.expect(']');
  	                }
  	                else {
  	                    key = this.throwUnexpectedToken(token);
  	                }
  	                break;
  	            default:
  	                key = this.throwUnexpectedToken(token);
  	        }
  	        return key;
  	    };
  	    Parser.prototype.isPropertyKey = function (key, value) {
  	        return (key.type === syntax_1.Syntax.Identifier && key.name === value) ||
  	            (key.type === syntax_1.Syntax.Literal && key.value === value);
  	    };
  	    Parser.prototype.parseObjectProperty = function (hasProto) {
  	        var node = this.createNode();
  	        var token = this.lookahead;
  	        var kind;
  	        var key = null;
  	        var value = null;
  	        var computed = false;
  	        var method = false;
  	        var shorthand = false;
  	        var isAsync = false;
  	        if (token.type === 3 /* Identifier */) {
  	            var id = token.value;
  	            this.nextToken();
  	            computed = this.match('[');
  	            isAsync = !this.hasLineTerminator && (id === 'async') &&
  	                !this.match(':') && !this.match('(') && !this.match('*');
  	            key = isAsync ? this.parseObjectPropertyKey() : this.finalize(node, new Node.Identifier(id));
  	        }
  	        else if (this.match('*')) {
  	            this.nextToken();
  	        }
  	        else {
  	            computed = this.match('[');
  	            key = this.parseObjectPropertyKey();
  	        }
  	        var lookaheadPropertyKey = this.qualifiedPropertyName(this.lookahead);
  	        if (token.type === 3 /* Identifier */ && !isAsync && token.value === 'get' && lookaheadPropertyKey) {
  	            kind = 'get';
  	            computed = this.match('[');
  	            key = this.parseObjectPropertyKey();
  	            this.context.allowYield = false;
  	            value = this.parseGetterMethod();
  	        }
  	        else if (token.type === 3 /* Identifier */ && !isAsync && token.value === 'set' && lookaheadPropertyKey) {
  	            kind = 'set';
  	            computed = this.match('[');
  	            key = this.parseObjectPropertyKey();
  	            value = this.parseSetterMethod();
  	        }
  	        else if (token.type === 7 /* Punctuator */ && token.value === '*' && lookaheadPropertyKey) {
  	            kind = 'init';
  	            computed = this.match('[');
  	            key = this.parseObjectPropertyKey();
  	            value = this.parseGeneratorMethod();
  	            method = true;
  	        }
  	        else {
  	            if (!key) {
  	                this.throwUnexpectedToken(this.lookahead);
  	            }
  	            kind = 'init';
  	            if (this.match(':') && !isAsync) {
  	                if (!computed && this.isPropertyKey(key, '__proto__')) {
  	                    if (hasProto.value) {
  	                        this.tolerateError(messages_1.Messages.DuplicateProtoProperty);
  	                    }
  	                    hasProto.value = true;
  	                }
  	                this.nextToken();
  	                value = this.inheritCoverGrammar(this.parseAssignmentExpression);
  	            }
  	            else if (this.match('(')) {
  	                value = isAsync ? this.parsePropertyMethodAsyncFunction() : this.parsePropertyMethodFunction();
  	                method = true;
  	            }
  	            else if (token.type === 3 /* Identifier */) {
  	                var id = this.finalize(node, new Node.Identifier(token.value));
  	                if (this.match('=')) {
  	                    this.context.firstCoverInitializedNameError = this.lookahead;
  	                    this.nextToken();
  	                    shorthand = true;
  	                    var init = this.isolateCoverGrammar(this.parseAssignmentExpression);
  	                    value = this.finalize(node, new Node.AssignmentPattern(id, init));
  	                }
  	                else {
  	                    shorthand = true;
  	                    value = id;
  	                }
  	            }
  	            else {
  	                this.throwUnexpectedToken(this.nextToken());
  	            }
  	        }
  	        return this.finalize(node, new Node.Property(kind, key, computed, value, method, shorthand));
  	    };
  	    Parser.prototype.parseObjectInitializer = function () {
  	        var node = this.createNode();
  	        this.expect('{');
  	        var properties = [];
  	        var hasProto = { value: false };
  	        while (!this.match('}')) {
  	            properties.push(this.parseObjectProperty(hasProto));
  	            if (!this.match('}')) {
  	                this.expectCommaSeparator();
  	            }
  	        }
  	        this.expect('}');
  	        return this.finalize(node, new Node.ObjectExpression(properties));
  	    };
  	    // https://tc39.github.io/ecma262/#sec-template-literals
  	    Parser.prototype.parseTemplateHead = function () {
  	        assert_1.assert(this.lookahead.head, 'Template literal must start with a template head');
  	        var node = this.createNode();
  	        var token = this.nextToken();
  	        var raw = token.value;
  	        var cooked = token.cooked;
  	        return this.finalize(node, new Node.TemplateElement({ raw: raw, cooked: cooked }, token.tail));
  	    };
  	    Parser.prototype.parseTemplateElement = function () {
  	        if (this.lookahead.type !== 10 /* Template */) {
  	            this.throwUnexpectedToken();
  	        }
  	        var node = this.createNode();
  	        var token = this.nextToken();
  	        var raw = token.value;
  	        var cooked = token.cooked;
  	        return this.finalize(node, new Node.TemplateElement({ raw: raw, cooked: cooked }, token.tail));
  	    };
  	    Parser.prototype.parseTemplateLiteral = function () {
  	        var node = this.createNode();
  	        var expressions = [];
  	        var quasis = [];
  	        var quasi = this.parseTemplateHead();
  	        quasis.push(quasi);
  	        while (!quasi.tail) {
  	            expressions.push(this.parseExpression());
  	            quasi = this.parseTemplateElement();
  	            quasis.push(quasi);
  	        }
  	        return this.finalize(node, new Node.TemplateLiteral(quasis, expressions));
  	    };
  	    // https://tc39.github.io/ecma262/#sec-grouping-operator
  	    Parser.prototype.reinterpretExpressionAsPattern = function (expr) {
  	        switch (expr.type) {
  	            case syntax_1.Syntax.Identifier:
  	            case syntax_1.Syntax.MemberExpression:
  	            case syntax_1.Syntax.RestElement:
  	            case syntax_1.Syntax.AssignmentPattern:
  	                break;
  	            case syntax_1.Syntax.SpreadElement:
  	                expr.type = syntax_1.Syntax.RestElement;
  	                this.reinterpretExpressionAsPattern(expr.argument);
  	                break;
  	            case syntax_1.Syntax.ArrayExpression:
  	                expr.type = syntax_1.Syntax.ArrayPattern;
  	                for (var i = 0; i < expr.elements.length; i++) {
  	                    if (expr.elements[i] !== null) {
  	                        this.reinterpretExpressionAsPattern(expr.elements[i]);
  	                    }
  	                }
  	                break;
  	            case syntax_1.Syntax.ObjectExpression:
  	                expr.type = syntax_1.Syntax.ObjectPattern;
  	                for (var i = 0; i < expr.properties.length; i++) {
  	                    this.reinterpretExpressionAsPattern(expr.properties[i].value);
  	                }
  	                break;
  	            case syntax_1.Syntax.AssignmentExpression:
  	                expr.type = syntax_1.Syntax.AssignmentPattern;
  	                delete expr.operator;
  	                this.reinterpretExpressionAsPattern(expr.left);
  	                break;
  	            default:
  	                // Allow other node type for tolerant parsing.
  	                break;
  	        }
  	    };
  	    Parser.prototype.parseGroupExpression = function () {
  	        var expr;
  	        this.expect('(');
  	        if (this.match(')')) {
  	            this.nextToken();
  	            if (!this.match('=>')) {
  	                this.expect('=>');
  	            }
  	            expr = {
  	                type: ArrowParameterPlaceHolder,
  	                params: [],
  	                async: false
  	            };
  	        }
  	        else {
  	            var startToken = this.lookahead;
  	            var params = [];
  	            if (this.match('...')) {
  	                expr = this.parseRestElement(params);
  	                this.expect(')');
  	                if (!this.match('=>')) {
  	                    this.expect('=>');
  	                }
  	                expr = {
  	                    type: ArrowParameterPlaceHolder,
  	                    params: [expr],
  	                    async: false
  	                };
  	            }
  	            else {
  	                var arrow = false;
  	                this.context.isBindingElement = true;
  	                expr = this.inheritCoverGrammar(this.parseAssignmentExpression);
  	                if (this.match(',')) {
  	                    var expressions = [];
  	                    this.context.isAssignmentTarget = false;
  	                    expressions.push(expr);
  	                    while (this.lookahead.type !== 2 /* EOF */) {
  	                        if (!this.match(',')) {
  	                            break;
  	                        }
  	                        this.nextToken();
  	                        if (this.match(')')) {
  	                            this.nextToken();
  	                            for (var i = 0; i < expressions.length; i++) {
  	                                this.reinterpretExpressionAsPattern(expressions[i]);
  	                            }
  	                            arrow = true;
  	                            expr = {
  	                                type: ArrowParameterPlaceHolder,
  	                                params: expressions,
  	                                async: false
  	                            };
  	                        }
  	                        else if (this.match('...')) {
  	                            if (!this.context.isBindingElement) {
  	                                this.throwUnexpectedToken(this.lookahead);
  	                            }
  	                            expressions.push(this.parseRestElement(params));
  	                            this.expect(')');
  	                            if (!this.match('=>')) {
  	                                this.expect('=>');
  	                            }
  	                            this.context.isBindingElement = false;
  	                            for (var i = 0; i < expressions.length; i++) {
  	                                this.reinterpretExpressionAsPattern(expressions[i]);
  	                            }
  	                            arrow = true;
  	                            expr = {
  	                                type: ArrowParameterPlaceHolder,
  	                                params: expressions,
  	                                async: false
  	                            };
  	                        }
  	                        else {
  	                            expressions.push(this.inheritCoverGrammar(this.parseAssignmentExpression));
  	                        }
  	                        if (arrow) {
  	                            break;
  	                        }
  	                    }
  	                    if (!arrow) {
  	                        expr = this.finalize(this.startNode(startToken), new Node.SequenceExpression(expressions));
  	                    }
  	                }
  	                if (!arrow) {
  	                    this.expect(')');
  	                    if (this.match('=>')) {
  	                        if (expr.type === syntax_1.Syntax.Identifier && expr.name === 'yield') {
  	                            arrow = true;
  	                            expr = {
  	                                type: ArrowParameterPlaceHolder,
  	                                params: [expr],
  	                                async: false
  	                            };
  	                        }
  	                        if (!arrow) {
  	                            if (!this.context.isBindingElement) {
  	                                this.throwUnexpectedToken(this.lookahead);
  	                            }
  	                            if (expr.type === syntax_1.Syntax.SequenceExpression) {
  	                                for (var i = 0; i < expr.expressions.length; i++) {
  	                                    this.reinterpretExpressionAsPattern(expr.expressions[i]);
  	                                }
  	                            }
  	                            else {
  	                                this.reinterpretExpressionAsPattern(expr);
  	                            }
  	                            var parameters = (expr.type === syntax_1.Syntax.SequenceExpression ? expr.expressions : [expr]);
  	                            expr = {
  	                                type: ArrowParameterPlaceHolder,
  	                                params: parameters,
  	                                async: false
  	                            };
  	                        }
  	                    }
  	                    this.context.isBindingElement = false;
  	                }
  	            }
  	        }
  	        return expr;
  	    };
  	    // https://tc39.github.io/ecma262/#sec-left-hand-side-expressions
  	    Parser.prototype.parseArguments = function () {
  	        this.expect('(');
  	        var args = [];
  	        if (!this.match(')')) {
  	            while (true) {
  	                var expr = this.match('...') ? this.parseSpreadElement() :
  	                    this.isolateCoverGrammar(this.parseAssignmentExpression);
  	                args.push(expr);
  	                if (this.match(')')) {
  	                    break;
  	                }
  	                this.expectCommaSeparator();
  	                if (this.match(')')) {
  	                    break;
  	                }
  	            }
  	        }
  	        this.expect(')');
  	        return args;
  	    };
  	    Parser.prototype.isIdentifierName = function (token) {
  	        return token.type === 3 /* Identifier */ ||
  	            token.type === 4 /* Keyword */ ||
  	            token.type === 1 /* BooleanLiteral */ ||
  	            token.type === 5 /* NullLiteral */;
  	    };
  	    Parser.prototype.parseIdentifierName = function () {
  	        var node = this.createNode();
  	        var token = this.nextToken();
  	        if (!this.isIdentifierName(token)) {
  	            this.throwUnexpectedToken(token);
  	        }
  	        return this.finalize(node, new Node.Identifier(token.value));
  	    };
  	    Parser.prototype.parseNewExpression = function () {
  	        var node = this.createNode();
  	        var id = this.parseIdentifierName();
  	        assert_1.assert(id.name === 'new', 'New expression must start with `new`');
  	        var expr;
  	        if (this.match('.')) {
  	            this.nextToken();
  	            if (this.lookahead.type === 3 /* Identifier */ && this.context.inFunctionBody && this.lookahead.value === 'target') {
  	                var property = this.parseIdentifierName();
  	                expr = new Node.MetaProperty(id, property);
  	            }
  	            else {
  	                this.throwUnexpectedToken(this.lookahead);
  	            }
  	        }
  	        else {
  	            var callee = this.isolateCoverGrammar(this.parseLeftHandSideExpression);
  	            var args = this.match('(') ? this.parseArguments() : [];
  	            expr = new Node.NewExpression(callee, args);
  	            this.context.isAssignmentTarget = false;
  	            this.context.isBindingElement = false;
  	        }
  	        return this.finalize(node, expr);
  	    };
  	    Parser.prototype.parseAsyncArgument = function () {
  	        var arg = this.parseAssignmentExpression();
  	        this.context.firstCoverInitializedNameError = null;
  	        return arg;
  	    };
  	    Parser.prototype.parseAsyncArguments = function () {
  	        this.expect('(');
  	        var args = [];
  	        if (!this.match(')')) {
  	            while (true) {
  	                var expr = this.match('...') ? this.parseSpreadElement() :
  	                    this.isolateCoverGrammar(this.parseAsyncArgument);
  	                args.push(expr);
  	                if (this.match(')')) {
  	                    break;
  	                }
  	                this.expectCommaSeparator();
  	                if (this.match(')')) {
  	                    break;
  	                }
  	            }
  	        }
  	        this.expect(')');
  	        return args;
  	    };
  	    Parser.prototype.parseLeftHandSideExpressionAllowCall = function () {
  	        var startToken = this.lookahead;
  	        var maybeAsync = this.matchContextualKeyword('async');
  	        var previousAllowIn = this.context.allowIn;
  	        this.context.allowIn = true;
  	        var expr;
  	        if (this.matchKeyword('super') && this.context.inFunctionBody) {
  	            expr = this.createNode();
  	            this.nextToken();
  	            expr = this.finalize(expr, new Node.Super());
  	            if (!this.match('(') && !this.match('.') && !this.match('[')) {
  	                this.throwUnexpectedToken(this.lookahead);
  	            }
  	        }
  	        else {
  	            expr = this.inheritCoverGrammar(this.matchKeyword('new') ? this.parseNewExpression : this.parsePrimaryExpression);
  	        }
  	        while (true) {
  	            if (this.match('.')) {
  	                this.context.isBindingElement = false;
  	                this.context.isAssignmentTarget = true;
  	                this.expect('.');
  	                var property = this.parseIdentifierName();
  	                expr = this.finalize(this.startNode(startToken), new Node.StaticMemberExpression(expr, property));
  	            }
  	            else if (this.match('(')) {
  	                var asyncArrow = maybeAsync && (startToken.lineNumber === this.lookahead.lineNumber);
  	                this.context.isBindingElement = false;
  	                this.context.isAssignmentTarget = false;
  	                var args = asyncArrow ? this.parseAsyncArguments() : this.parseArguments();
  	                expr = this.finalize(this.startNode(startToken), new Node.CallExpression(expr, args));
  	                if (asyncArrow && this.match('=>')) {
  	                    for (var i = 0; i < args.length; ++i) {
  	                        this.reinterpretExpressionAsPattern(args[i]);
  	                    }
  	                    expr = {
  	                        type: ArrowParameterPlaceHolder,
  	                        params: args,
  	                        async: true
  	                    };
  	                }
  	            }
  	            else if (this.match('[')) {
  	                this.context.isBindingElement = false;
  	                this.context.isAssignmentTarget = true;
  	                this.expect('[');
  	                var property = this.isolateCoverGrammar(this.parseExpression);
  	                this.expect(']');
  	                expr = this.finalize(this.startNode(startToken), new Node.ComputedMemberExpression(expr, property));
  	            }
  	            else if (this.lookahead.type === 10 /* Template */ && this.lookahead.head) {
  	                var quasi = this.parseTemplateLiteral();
  	                expr = this.finalize(this.startNode(startToken), new Node.TaggedTemplateExpression(expr, quasi));
  	            }
  	            else {
  	                break;
  	            }
  	        }
  	        this.context.allowIn = previousAllowIn;
  	        return expr;
  	    };
  	    Parser.prototype.parseSuper = function () {
  	        var node = this.createNode();
  	        this.expectKeyword('super');
  	        if (!this.match('[') && !this.match('.')) {
  	            this.throwUnexpectedToken(this.lookahead);
  	        }
  	        return this.finalize(node, new Node.Super());
  	    };
  	    Parser.prototype.parseLeftHandSideExpression = function () {
  	        assert_1.assert(this.context.allowIn, 'callee of new expression always allow in keyword.');
  	        var node = this.startNode(this.lookahead);
  	        var expr = (this.matchKeyword('super') && this.context.inFunctionBody) ? this.parseSuper() :
  	            this.inheritCoverGrammar(this.matchKeyword('new') ? this.parseNewExpression : this.parsePrimaryExpression);
  	        while (true) {
  	            if (this.match('[')) {
  	                this.context.isBindingElement = false;
  	                this.context.isAssignmentTarget = true;
  	                this.expect('[');
  	                var property = this.isolateCoverGrammar(this.parseExpression);
  	                this.expect(']');
  	                expr = this.finalize(node, new Node.ComputedMemberExpression(expr, property));
  	            }
  	            else if (this.match('.')) {
  	                this.context.isBindingElement = false;
  	                this.context.isAssignmentTarget = true;
  	                this.expect('.');
  	                var property = this.parseIdentifierName();
  	                expr = this.finalize(node, new Node.StaticMemberExpression(expr, property));
  	            }
  	            else if (this.lookahead.type === 10 /* Template */ && this.lookahead.head) {
  	                var quasi = this.parseTemplateLiteral();
  	                expr = this.finalize(node, new Node.TaggedTemplateExpression(expr, quasi));
  	            }
  	            else {
  	                break;
  	            }
  	        }
  	        return expr;
  	    };
  	    // https://tc39.github.io/ecma262/#sec-update-expressions
  	    Parser.prototype.parseUpdateExpression = function () {
  	        var expr;
  	        var startToken = this.lookahead;
  	        if (this.match('++') || this.match('--')) {
  	            var node = this.startNode(startToken);
  	            var token = this.nextToken();
  	            expr = this.inheritCoverGrammar(this.parseUnaryExpression);
  	            if (this.context.strict && expr.type === syntax_1.Syntax.Identifier && this.scanner.isRestrictedWord(expr.name)) {
  	                this.tolerateError(messages_1.Messages.StrictLHSPrefix);
  	            }
  	            if (!this.context.isAssignmentTarget) {
  	                this.tolerateError(messages_1.Messages.InvalidLHSInAssignment);
  	            }
  	            var prefix = true;
  	            expr = this.finalize(node, new Node.UpdateExpression(token.value, expr, prefix));
  	            this.context.isAssignmentTarget = false;
  	            this.context.isBindingElement = false;
  	        }
  	        else {
  	            expr = this.inheritCoverGrammar(this.parseLeftHandSideExpressionAllowCall);
  	            if (!this.hasLineTerminator && this.lookahead.type === 7 /* Punctuator */) {
  	                if (this.match('++') || this.match('--')) {
  	                    if (this.context.strict && expr.type === syntax_1.Syntax.Identifier && this.scanner.isRestrictedWord(expr.name)) {
  	                        this.tolerateError(messages_1.Messages.StrictLHSPostfix);
  	                    }
  	                    if (!this.context.isAssignmentTarget) {
  	                        this.tolerateError(messages_1.Messages.InvalidLHSInAssignment);
  	                    }
  	                    this.context.isAssignmentTarget = false;
  	                    this.context.isBindingElement = false;
  	                    var operator = this.nextToken().value;
  	                    var prefix = false;
  	                    expr = this.finalize(this.startNode(startToken), new Node.UpdateExpression(operator, expr, prefix));
  	                }
  	            }
  	        }
  	        return expr;
  	    };
  	    // https://tc39.github.io/ecma262/#sec-unary-operators
  	    Parser.prototype.parseAwaitExpression = function () {
  	        var node = this.createNode();
  	        this.nextToken();
  	        var argument = this.parseUnaryExpression();
  	        return this.finalize(node, new Node.AwaitExpression(argument));
  	    };
  	    Parser.prototype.parseUnaryExpression = function () {
  	        var expr;
  	        if (this.match('+') || this.match('-') || this.match('~') || this.match('!') ||
  	            this.matchKeyword('delete') || this.matchKeyword('void') || this.matchKeyword('typeof')) {
  	            var node = this.startNode(this.lookahead);
  	            var token = this.nextToken();
  	            expr = this.inheritCoverGrammar(this.parseUnaryExpression);
  	            expr = this.finalize(node, new Node.UnaryExpression(token.value, expr));
  	            if (this.context.strict && expr.operator === 'delete' && expr.argument.type === syntax_1.Syntax.Identifier) {
  	                this.tolerateError(messages_1.Messages.StrictDelete);
  	            }
  	            this.context.isAssignmentTarget = false;
  	            this.context.isBindingElement = false;
  	        }
  	        else if (this.context.await && this.matchContextualKeyword('await')) {
  	            expr = this.parseAwaitExpression();
  	        }
  	        else {
  	            expr = this.parseUpdateExpression();
  	        }
  	        return expr;
  	    };
  	    Parser.prototype.parseExponentiationExpression = function () {
  	        var startToken = this.lookahead;
  	        var expr = this.inheritCoverGrammar(this.parseUnaryExpression);
  	        if (expr.type !== syntax_1.Syntax.UnaryExpression && this.match('**')) {
  	            this.nextToken();
  	            this.context.isAssignmentTarget = false;
  	            this.context.isBindingElement = false;
  	            var left = expr;
  	            var right = this.isolateCoverGrammar(this.parseExponentiationExpression);
  	            expr = this.finalize(this.startNode(startToken), new Node.BinaryExpression('**', left, right));
  	        }
  	        return expr;
  	    };
  	    // https://tc39.github.io/ecma262/#sec-exp-operator
  	    // https://tc39.github.io/ecma262/#sec-multiplicative-operators
  	    // https://tc39.github.io/ecma262/#sec-additive-operators
  	    // https://tc39.github.io/ecma262/#sec-bitwise-shift-operators
  	    // https://tc39.github.io/ecma262/#sec-relational-operators
  	    // https://tc39.github.io/ecma262/#sec-equality-operators
  	    // https://tc39.github.io/ecma262/#sec-binary-bitwise-operators
  	    // https://tc39.github.io/ecma262/#sec-binary-logical-operators
  	    Parser.prototype.binaryPrecedence = function (token) {
  	        var op = token.value;
  	        var precedence;
  	        if (token.type === 7 /* Punctuator */) {
  	            precedence = this.operatorPrecedence[op] || 0;
  	        }
  	        else if (token.type === 4 /* Keyword */) {
  	            precedence = (op === 'instanceof' || (this.context.allowIn && op === 'in')) ? 7 : 0;
  	        }
  	        else {
  	            precedence = 0;
  	        }
  	        return precedence;
  	    };
  	    Parser.prototype.parseBinaryExpression = function () {
  	        var startToken = this.lookahead;
  	        var expr = this.inheritCoverGrammar(this.parseExponentiationExpression);
  	        var token = this.lookahead;
  	        var prec = this.binaryPrecedence(token);
  	        if (prec > 0) {
  	            this.nextToken();
  	            this.context.isAssignmentTarget = false;
  	            this.context.isBindingElement = false;
  	            var markers = [startToken, this.lookahead];
  	            var left = expr;
  	            var right = this.isolateCoverGrammar(this.parseExponentiationExpression);
  	            var stack = [left, token.value, right];
  	            var precedences = [prec];
  	            while (true) {
  	                prec = this.binaryPrecedence(this.lookahead);
  	                if (prec <= 0) {
  	                    break;
  	                }
  	                // Reduce: make a binary expression from the three topmost entries.
  	                while ((stack.length > 2) && (prec <= precedences[precedences.length - 1])) {
  	                    right = stack.pop();
  	                    var operator = stack.pop();
  	                    precedences.pop();
  	                    left = stack.pop();
  	                    markers.pop();
  	                    var node = this.startNode(markers[markers.length - 1]);
  	                    stack.push(this.finalize(node, new Node.BinaryExpression(operator, left, right)));
  	                }
  	                // Shift.
  	                stack.push(this.nextToken().value);
  	                precedences.push(prec);
  	                markers.push(this.lookahead);
  	                stack.push(this.isolateCoverGrammar(this.parseExponentiationExpression));
  	            }
  	            // Final reduce to clean-up the stack.
  	            var i = stack.length - 1;
  	            expr = stack[i];
  	            markers.pop();
  	            while (i > 1) {
  	                var node = this.startNode(markers.pop());
  	                var operator = stack[i - 1];
  	                expr = this.finalize(node, new Node.BinaryExpression(operator, stack[i - 2], expr));
  	                i -= 2;
  	            }
  	        }
  	        return expr;
  	    };
  	    // https://tc39.github.io/ecma262/#sec-conditional-operator
  	    Parser.prototype.parseConditionalExpression = function () {
  	        var startToken = this.lookahead;
  	        var expr = this.inheritCoverGrammar(this.parseBinaryExpression);
  	        if (this.match('?')) {
  	            this.nextToken();
  	            var previousAllowIn = this.context.allowIn;
  	            this.context.allowIn = true;
  	            var consequent = this.isolateCoverGrammar(this.parseAssignmentExpression);
  	            this.context.allowIn = previousAllowIn;
  	            this.expect(':');
  	            var alternate = this.isolateCoverGrammar(this.parseAssignmentExpression);
  	            expr = this.finalize(this.startNode(startToken), new Node.ConditionalExpression(expr, consequent, alternate));
  	            this.context.isAssignmentTarget = false;
  	            this.context.isBindingElement = false;
  	        }
  	        return expr;
  	    };
  	    // https://tc39.github.io/ecma262/#sec-assignment-operators
  	    Parser.prototype.checkPatternParam = function (options, param) {
  	        switch (param.type) {
  	            case syntax_1.Syntax.Identifier:
  	                this.validateParam(options, param, param.name);
  	                break;
  	            case syntax_1.Syntax.RestElement:
  	                this.checkPatternParam(options, param.argument);
  	                break;
  	            case syntax_1.Syntax.AssignmentPattern:
  	                this.checkPatternParam(options, param.left);
  	                break;
  	            case syntax_1.Syntax.ArrayPattern:
  	                for (var i = 0; i < param.elements.length; i++) {
  	                    if (param.elements[i] !== null) {
  	                        this.checkPatternParam(options, param.elements[i]);
  	                    }
  	                }
  	                break;
  	            case syntax_1.Syntax.ObjectPattern:
  	                for (var i = 0; i < param.properties.length; i++) {
  	                    this.checkPatternParam(options, param.properties[i].value);
  	                }
  	                break;
  	            default:
  	                break;
  	        }
  	        options.simple = options.simple && (param instanceof Node.Identifier);
  	    };
  	    Parser.prototype.reinterpretAsCoverFormalsList = function (expr) {
  	        var params = [expr];
  	        var options;
  	        var asyncArrow = false;
  	        switch (expr.type) {
  	            case syntax_1.Syntax.Identifier:
  	                break;
  	            case ArrowParameterPlaceHolder:
  	                params = expr.params;
  	                asyncArrow = expr.async;
  	                break;
  	            default:
  	                return null;
  	        }
  	        options = {
  	            simple: true,
  	            paramSet: {}
  	        };
  	        for (var i = 0; i < params.length; ++i) {
  	            var param = params[i];
  	            if (param.type === syntax_1.Syntax.AssignmentPattern) {
  	                if (param.right.type === syntax_1.Syntax.YieldExpression) {
  	                    if (param.right.argument) {
  	                        this.throwUnexpectedToken(this.lookahead);
  	                    }
  	                    param.right.type = syntax_1.Syntax.Identifier;
  	                    param.right.name = 'yield';
  	                    delete param.right.argument;
  	                    delete param.right.delegate;
  	                }
  	            }
  	            else if (asyncArrow && param.type === syntax_1.Syntax.Identifier && param.name === 'await') {
  	                this.throwUnexpectedToken(this.lookahead);
  	            }
  	            this.checkPatternParam(options, param);
  	            params[i] = param;
  	        }
  	        if (this.context.strict || !this.context.allowYield) {
  	            for (var i = 0; i < params.length; ++i) {
  	                var param = params[i];
  	                if (param.type === syntax_1.Syntax.YieldExpression) {
  	                    this.throwUnexpectedToken(this.lookahead);
  	                }
  	            }
  	        }
  	        if (options.message === messages_1.Messages.StrictParamDupe) {
  	            var token = this.context.strict ? options.stricted : options.firstRestricted;
  	            this.throwUnexpectedToken(token, options.message);
  	        }
  	        return {
  	            simple: options.simple,
  	            params: params,
  	            stricted: options.stricted,
  	            firstRestricted: options.firstRestricted,
  	            message: options.message
  	        };
  	    };
  	    Parser.prototype.parseAssignmentExpression = function () {
  	        var expr;
  	        if (!this.context.allowYield && this.matchKeyword('yield')) {
  	            expr = this.parseYieldExpression();
  	        }
  	        else {
  	            var startToken = this.lookahead;
  	            var token = startToken;
  	            expr = this.parseConditionalExpression();
  	            if (token.type === 3 /* Identifier */ && (token.lineNumber === this.lookahead.lineNumber) && token.value === 'async') {
  	                if (this.lookahead.type === 3 /* Identifier */ || this.matchKeyword('yield')) {
  	                    var arg = this.parsePrimaryExpression();
  	                    this.reinterpretExpressionAsPattern(arg);
  	                    expr = {
  	                        type: ArrowParameterPlaceHolder,
  	                        params: [arg],
  	                        async: true
  	                    };
  	                }
  	            }
  	            if (expr.type === ArrowParameterPlaceHolder || this.match('=>')) {
  	                // https://tc39.github.io/ecma262/#sec-arrow-function-definitions
  	                this.context.isAssignmentTarget = false;
  	                this.context.isBindingElement = false;
  	                var isAsync = expr.async;
  	                var list = this.reinterpretAsCoverFormalsList(expr);
  	                if (list) {
  	                    if (this.hasLineTerminator) {
  	                        this.tolerateUnexpectedToken(this.lookahead);
  	                    }
  	                    this.context.firstCoverInitializedNameError = null;
  	                    var previousStrict = this.context.strict;
  	                    var previousAllowStrictDirective = this.context.allowStrictDirective;
  	                    this.context.allowStrictDirective = list.simple;
  	                    var previousAllowYield = this.context.allowYield;
  	                    var previousAwait = this.context.await;
  	                    this.context.allowYield = true;
  	                    this.context.await = isAsync;
  	                    var node = this.startNode(startToken);
  	                    this.expect('=>');
  	                    var body = void 0;
  	                    if (this.match('{')) {
  	                        var previousAllowIn = this.context.allowIn;
  	                        this.context.allowIn = true;
  	                        body = this.parseFunctionSourceElements();
  	                        this.context.allowIn = previousAllowIn;
  	                    }
  	                    else {
  	                        body = this.isolateCoverGrammar(this.parseAssignmentExpression);
  	                    }
  	                    var expression = body.type !== syntax_1.Syntax.BlockStatement;
  	                    if (this.context.strict && list.firstRestricted) {
  	                        this.throwUnexpectedToken(list.firstRestricted, list.message);
  	                    }
  	                    if (this.context.strict && list.stricted) {
  	                        this.tolerateUnexpectedToken(list.stricted, list.message);
  	                    }
  	                    expr = isAsync ? this.finalize(node, new Node.AsyncArrowFunctionExpression(list.params, body, expression)) :
  	                        this.finalize(node, new Node.ArrowFunctionExpression(list.params, body, expression));
  	                    this.context.strict = previousStrict;
  	                    this.context.allowStrictDirective = previousAllowStrictDirective;
  	                    this.context.allowYield = previousAllowYield;
  	                    this.context.await = previousAwait;
  	                }
  	            }
  	            else {
  	                if (this.matchAssign()) {
  	                    if (!this.context.isAssignmentTarget) {
  	                        this.tolerateError(messages_1.Messages.InvalidLHSInAssignment);
  	                    }
  	                    if (this.context.strict && expr.type === syntax_1.Syntax.Identifier) {
  	                        var id = expr;
  	                        if (this.scanner.isRestrictedWord(id.name)) {
  	                            this.tolerateUnexpectedToken(token, messages_1.Messages.StrictLHSAssignment);
  	                        }
  	                        if (this.scanner.isStrictModeReservedWord(id.name)) {
  	                            this.tolerateUnexpectedToken(token, messages_1.Messages.StrictReservedWord);
  	                        }
  	                    }
  	                    if (!this.match('=')) {
  	                        this.context.isAssignmentTarget = false;
  	                        this.context.isBindingElement = false;
  	                    }
  	                    else {
  	                        this.reinterpretExpressionAsPattern(expr);
  	                    }
  	                    token = this.nextToken();
  	                    var operator = token.value;
  	                    var right = this.isolateCoverGrammar(this.parseAssignmentExpression);
  	                    expr = this.finalize(this.startNode(startToken), new Node.AssignmentExpression(operator, expr, right));
  	                    this.context.firstCoverInitializedNameError = null;
  	                }
  	            }
  	        }
  	        return expr;
  	    };
  	    // https://tc39.github.io/ecma262/#sec-comma-operator
  	    Parser.prototype.parseExpression = function () {
  	        var startToken = this.lookahead;
  	        var expr = this.isolateCoverGrammar(this.parseAssignmentExpression);
  	        if (this.match(',')) {
  	            var expressions = [];
  	            expressions.push(expr);
  	            while (this.lookahead.type !== 2 /* EOF */) {
  	                if (!this.match(',')) {
  	                    break;
  	                }
  	                this.nextToken();
  	                expressions.push(this.isolateCoverGrammar(this.parseAssignmentExpression));
  	            }
  	            expr = this.finalize(this.startNode(startToken), new Node.SequenceExpression(expressions));
  	        }
  	        return expr;
  	    };
  	    // https://tc39.github.io/ecma262/#sec-block
  	    Parser.prototype.parseStatementListItem = function () {
  	        var statement;
  	        this.context.isAssignmentTarget = true;
  	        this.context.isBindingElement = true;
  	        if (this.lookahead.type === 4 /* Keyword */) {
  	            switch (this.lookahead.value) {
  	                case 'export':
  	                    if (!this.context.isModule) {
  	                        this.tolerateUnexpectedToken(this.lookahead, messages_1.Messages.IllegalExportDeclaration);
  	                    }
  	                    statement = this.parseExportDeclaration();
  	                    break;
  	                case 'import':
  	                    if (!this.context.isModule) {
  	                        this.tolerateUnexpectedToken(this.lookahead, messages_1.Messages.IllegalImportDeclaration);
  	                    }
  	                    statement = this.parseImportDeclaration();
  	                    break;
  	                case 'const':
  	                    statement = this.parseLexicalDeclaration({ inFor: false });
  	                    break;
  	                case 'function':
  	                    statement = this.parseFunctionDeclaration();
  	                    break;
  	                case 'class':
  	                    statement = this.parseClassDeclaration();
  	                    break;
  	                case 'let':
  	                    statement = this.isLexicalDeclaration() ? this.parseLexicalDeclaration({ inFor: false }) : this.parseStatement();
  	                    break;
  	                default:
  	                    statement = this.parseStatement();
  	                    break;
  	            }
  	        }
  	        else {
  	            statement = this.parseStatement();
  	        }
  	        return statement;
  	    };
  	    Parser.prototype.parseBlock = function () {
  	        var node = this.createNode();
  	        this.expect('{');
  	        var block = [];
  	        while (true) {
  	            if (this.match('}')) {
  	                break;
  	            }
  	            block.push(this.parseStatementListItem());
  	        }
  	        this.expect('}');
  	        return this.finalize(node, new Node.BlockStatement(block));
  	    };
  	    // https://tc39.github.io/ecma262/#sec-let-and-const-declarations
  	    Parser.prototype.parseLexicalBinding = function (kind, options) {
  	        var node = this.createNode();
  	        var params = [];
  	        var id = this.parsePattern(params, kind);
  	        if (this.context.strict && id.type === syntax_1.Syntax.Identifier) {
  	            if (this.scanner.isRestrictedWord(id.name)) {
  	                this.tolerateError(messages_1.Messages.StrictVarName);
  	            }
  	        }
  	        var init = null;
  	        if (kind === 'const') {
  	            if (!this.matchKeyword('in') && !this.matchContextualKeyword('of')) {
  	                if (this.match('=')) {
  	                    this.nextToken();
  	                    init = this.isolateCoverGrammar(this.parseAssignmentExpression);
  	                }
  	                else {
  	                    this.throwError(messages_1.Messages.DeclarationMissingInitializer, 'const');
  	                }
  	            }
  	        }
  	        else if ((!options.inFor && id.type !== syntax_1.Syntax.Identifier) || this.match('=')) {
  	            this.expect('=');
  	            init = this.isolateCoverGrammar(this.parseAssignmentExpression);
  	        }
  	        return this.finalize(node, new Node.VariableDeclarator(id, init));
  	    };
  	    Parser.prototype.parseBindingList = function (kind, options) {
  	        var list = [this.parseLexicalBinding(kind, options)];
  	        while (this.match(',')) {
  	            this.nextToken();
  	            list.push(this.parseLexicalBinding(kind, options));
  	        }
  	        return list;
  	    };
  	    Parser.prototype.isLexicalDeclaration = function () {
  	        var state = this.scanner.saveState();
  	        this.scanner.scanComments();
  	        var next = this.scanner.lex();
  	        this.scanner.restoreState(state);
  	        return (next.type === 3 /* Identifier */) ||
  	            (next.type === 7 /* Punctuator */ && next.value === '[') ||
  	            (next.type === 7 /* Punctuator */ && next.value === '{') ||
  	            (next.type === 4 /* Keyword */ && next.value === 'let') ||
  	            (next.type === 4 /* Keyword */ && next.value === 'yield');
  	    };
  	    Parser.prototype.parseLexicalDeclaration = function (options) {
  	        var node = this.createNode();
  	        var kind = this.nextToken().value;
  	        assert_1.assert(kind === 'let' || kind === 'const', 'Lexical declaration must be either let or const');
  	        var declarations = this.parseBindingList(kind, options);
  	        this.consumeSemicolon();
  	        return this.finalize(node, new Node.VariableDeclaration(declarations, kind));
  	    };
  	    // https://tc39.github.io/ecma262/#sec-destructuring-binding-patterns
  	    Parser.prototype.parseBindingRestElement = function (params, kind) {
  	        var node = this.createNode();
  	        this.expect('...');
  	        var arg = this.parsePattern(params, kind);
  	        return this.finalize(node, new Node.RestElement(arg));
  	    };
  	    Parser.prototype.parseArrayPattern = function (params, kind) {
  	        var node = this.createNode();
  	        this.expect('[');
  	        var elements = [];
  	        while (!this.match(']')) {
  	            if (this.match(',')) {
  	                this.nextToken();
  	                elements.push(null);
  	            }
  	            else {
  	                if (this.match('...')) {
  	                    elements.push(this.parseBindingRestElement(params, kind));
  	                    break;
  	                }
  	                else {
  	                    elements.push(this.parsePatternWithDefault(params, kind));
  	                }
  	                if (!this.match(']')) {
  	                    this.expect(',');
  	                }
  	            }
  	        }
  	        this.expect(']');
  	        return this.finalize(node, new Node.ArrayPattern(elements));
  	    };
  	    Parser.prototype.parsePropertyPattern = function (params, kind) {
  	        var node = this.createNode();
  	        var computed = false;
  	        var shorthand = false;
  	        var method = false;
  	        var key;
  	        var value;
  	        if (this.lookahead.type === 3 /* Identifier */) {
  	            var keyToken = this.lookahead;
  	            key = this.parseVariableIdentifier();
  	            var init = this.finalize(node, new Node.Identifier(keyToken.value));
  	            if (this.match('=')) {
  	                params.push(keyToken);
  	                shorthand = true;
  	                this.nextToken();
  	                var expr = this.parseAssignmentExpression();
  	                value = this.finalize(this.startNode(keyToken), new Node.AssignmentPattern(init, expr));
  	            }
  	            else if (!this.match(':')) {
  	                params.push(keyToken);
  	                shorthand = true;
  	                value = init;
  	            }
  	            else {
  	                this.expect(':');
  	                value = this.parsePatternWithDefault(params, kind);
  	            }
  	        }
  	        else {
  	            computed = this.match('[');
  	            key = this.parseObjectPropertyKey();
  	            this.expect(':');
  	            value = this.parsePatternWithDefault(params, kind);
  	        }
  	        return this.finalize(node, new Node.Property('init', key, computed, value, method, shorthand));
  	    };
  	    Parser.prototype.parseObjectPattern = function (params, kind) {
  	        var node = this.createNode();
  	        var properties = [];
  	        this.expect('{');
  	        while (!this.match('}')) {
  	            properties.push(this.parsePropertyPattern(params, kind));
  	            if (!this.match('}')) {
  	                this.expect(',');
  	            }
  	        }
  	        this.expect('}');
  	        return this.finalize(node, new Node.ObjectPattern(properties));
  	    };
  	    Parser.prototype.parsePattern = function (params, kind) {
  	        var pattern;
  	        if (this.match('[')) {
  	            pattern = this.parseArrayPattern(params, kind);
  	        }
  	        else if (this.match('{')) {
  	            pattern = this.parseObjectPattern(params, kind);
  	        }
  	        else {
  	            if (this.matchKeyword('let') && (kind === 'const' || kind === 'let')) {
  	                this.tolerateUnexpectedToken(this.lookahead, messages_1.Messages.LetInLexicalBinding);
  	            }
  	            params.push(this.lookahead);
  	            pattern = this.parseVariableIdentifier(kind);
  	        }
  	        return pattern;
  	    };
  	    Parser.prototype.parsePatternWithDefault = function (params, kind) {
  	        var startToken = this.lookahead;
  	        var pattern = this.parsePattern(params, kind);
  	        if (this.match('=')) {
  	            this.nextToken();
  	            var previousAllowYield = this.context.allowYield;
  	            this.context.allowYield = true;
  	            var right = this.isolateCoverGrammar(this.parseAssignmentExpression);
  	            this.context.allowYield = previousAllowYield;
  	            pattern = this.finalize(this.startNode(startToken), new Node.AssignmentPattern(pattern, right));
  	        }
  	        return pattern;
  	    };
  	    // https://tc39.github.io/ecma262/#sec-variable-statement
  	    Parser.prototype.parseVariableIdentifier = function (kind) {
  	        var node = this.createNode();
  	        var token = this.nextToken();
  	        if (token.type === 4 /* Keyword */ && token.value === 'yield') {
  	            if (this.context.strict) {
  	                this.tolerateUnexpectedToken(token, messages_1.Messages.StrictReservedWord);
  	            }
  	            else if (!this.context.allowYield) {
  	                this.throwUnexpectedToken(token);
  	            }
  	        }
  	        else if (token.type !== 3 /* Identifier */) {
  	            if (this.context.strict && token.type === 4 /* Keyword */ && this.scanner.isStrictModeReservedWord(token.value)) {
  	                this.tolerateUnexpectedToken(token, messages_1.Messages.StrictReservedWord);
  	            }
  	            else {
  	                if (this.context.strict || token.value !== 'let' || kind !== 'var') {
  	                    this.throwUnexpectedToken(token);
  	                }
  	            }
  	        }
  	        else if ((this.context.isModule || this.context.await) && token.type === 3 /* Identifier */ && token.value === 'await') {
  	            this.tolerateUnexpectedToken(token);
  	        }
  	        return this.finalize(node, new Node.Identifier(token.value));
  	    };
  	    Parser.prototype.parseVariableDeclaration = function (options) {
  	        var node = this.createNode();
  	        var params = [];
  	        var id = this.parsePattern(params, 'var');
  	        if (this.context.strict && id.type === syntax_1.Syntax.Identifier) {
  	            if (this.scanner.isRestrictedWord(id.name)) {
  	                this.tolerateError(messages_1.Messages.StrictVarName);
  	            }
  	        }
  	        var init = null;
  	        if (this.match('=')) {
  	            this.nextToken();
  	            init = this.isolateCoverGrammar(this.parseAssignmentExpression);
  	        }
  	        else if (id.type !== syntax_1.Syntax.Identifier && !options.inFor) {
  	            this.expect('=');
  	        }
  	        return this.finalize(node, new Node.VariableDeclarator(id, init));
  	    };
  	    Parser.prototype.parseVariableDeclarationList = function (options) {
  	        var opt = { inFor: options.inFor };
  	        var list = [];
  	        list.push(this.parseVariableDeclaration(opt));
  	        while (this.match(',')) {
  	            this.nextToken();
  	            list.push(this.parseVariableDeclaration(opt));
  	        }
  	        return list;
  	    };
  	    Parser.prototype.parseVariableStatement = function () {
  	        var node = this.createNode();
  	        this.expectKeyword('var');
  	        var declarations = this.parseVariableDeclarationList({ inFor: false });
  	        this.consumeSemicolon();
  	        return this.finalize(node, new Node.VariableDeclaration(declarations, 'var'));
  	    };
  	    // https://tc39.github.io/ecma262/#sec-empty-statement
  	    Parser.prototype.parseEmptyStatement = function () {
  	        var node = this.createNode();
  	        this.expect(';');
  	        return this.finalize(node, new Node.EmptyStatement());
  	    };
  	    // https://tc39.github.io/ecma262/#sec-expression-statement
  	    Parser.prototype.parseExpressionStatement = function () {
  	        var node = this.createNode();
  	        var expr = this.parseExpression();
  	        this.consumeSemicolon();
  	        return this.finalize(node, new Node.ExpressionStatement(expr));
  	    };
  	    // https://tc39.github.io/ecma262/#sec-if-statement
  	    Parser.prototype.parseIfClause = function () {
  	        if (this.context.strict && this.matchKeyword('function')) {
  	            this.tolerateError(messages_1.Messages.StrictFunction);
  	        }
  	        return this.parseStatement();
  	    };
  	    Parser.prototype.parseIfStatement = function () {
  	        var node = this.createNode();
  	        var consequent;
  	        var alternate = null;
  	        this.expectKeyword('if');
  	        this.expect('(');
  	        var test = this.parseExpression();
  	        if (!this.match(')') && this.config.tolerant) {
  	            this.tolerateUnexpectedToken(this.nextToken());
  	            consequent = this.finalize(this.createNode(), new Node.EmptyStatement());
  	        }
  	        else {
  	            this.expect(')');
  	            consequent = this.parseIfClause();
  	            if (this.matchKeyword('else')) {
  	                this.nextToken();
  	                alternate = this.parseIfClause();
  	            }
  	        }
  	        return this.finalize(node, new Node.IfStatement(test, consequent, alternate));
  	    };
  	    // https://tc39.github.io/ecma262/#sec-do-while-statement
  	    Parser.prototype.parseDoWhileStatement = function () {
  	        var node = this.createNode();
  	        this.expectKeyword('do');
  	        var previousInIteration = this.context.inIteration;
  	        this.context.inIteration = true;
  	        var body = this.parseStatement();
  	        this.context.inIteration = previousInIteration;
  	        this.expectKeyword('while');
  	        this.expect('(');
  	        var test = this.parseExpression();
  	        if (!this.match(')') && this.config.tolerant) {
  	            this.tolerateUnexpectedToken(this.nextToken());
  	        }
  	        else {
  	            this.expect(')');
  	            if (this.match(';')) {
  	                this.nextToken();
  	            }
  	        }
  	        return this.finalize(node, new Node.DoWhileStatement(body, test));
  	    };
  	    // https://tc39.github.io/ecma262/#sec-while-statement
  	    Parser.prototype.parseWhileStatement = function () {
  	        var node = this.createNode();
  	        var body;
  	        this.expectKeyword('while');
  	        this.expect('(');
  	        var test = this.parseExpression();
  	        if (!this.match(')') && this.config.tolerant) {
  	            this.tolerateUnexpectedToken(this.nextToken());
  	            body = this.finalize(this.createNode(), new Node.EmptyStatement());
  	        }
  	        else {
  	            this.expect(')');
  	            var previousInIteration = this.context.inIteration;
  	            this.context.inIteration = true;
  	            body = this.parseStatement();
  	            this.context.inIteration = previousInIteration;
  	        }
  	        return this.finalize(node, new Node.WhileStatement(test, body));
  	    };
  	    // https://tc39.github.io/ecma262/#sec-for-statement
  	    // https://tc39.github.io/ecma262/#sec-for-in-and-for-of-statements
  	    Parser.prototype.parseForStatement = function () {
  	        var init = null;
  	        var test = null;
  	        var update = null;
  	        var forIn = true;
  	        var left, right;
  	        var node = this.createNode();
  	        this.expectKeyword('for');
  	        this.expect('(');
  	        if (this.match(';')) {
  	            this.nextToken();
  	        }
  	        else {
  	            if (this.matchKeyword('var')) {
  	                init = this.createNode();
  	                this.nextToken();
  	                var previousAllowIn = this.context.allowIn;
  	                this.context.allowIn = false;
  	                var declarations = this.parseVariableDeclarationList({ inFor: true });
  	                this.context.allowIn = previousAllowIn;
  	                if (declarations.length === 1 && this.matchKeyword('in')) {
  	                    var decl = declarations[0];
  	                    if (decl.init && (decl.id.type === syntax_1.Syntax.ArrayPattern || decl.id.type === syntax_1.Syntax.ObjectPattern || this.context.strict)) {
  	                        this.tolerateError(messages_1.Messages.ForInOfLoopInitializer, 'for-in');
  	                    }
  	                    init = this.finalize(init, new Node.VariableDeclaration(declarations, 'var'));
  	                    this.nextToken();
  	                    left = init;
  	                    right = this.parseExpression();
  	                    init = null;
  	                }
  	                else if (declarations.length === 1 && declarations[0].init === null && this.matchContextualKeyword('of')) {
  	                    init = this.finalize(init, new Node.VariableDeclaration(declarations, 'var'));
  	                    this.nextToken();
  	                    left = init;
  	                    right = this.parseAssignmentExpression();
  	                    init = null;
  	                    forIn = false;
  	                }
  	                else {
  	                    init = this.finalize(init, new Node.VariableDeclaration(declarations, 'var'));
  	                    this.expect(';');
  	                }
  	            }
  	            else if (this.matchKeyword('const') || this.matchKeyword('let')) {
  	                init = this.createNode();
  	                var kind = this.nextToken().value;
  	                if (!this.context.strict && this.lookahead.value === 'in') {
  	                    init = this.finalize(init, new Node.Identifier(kind));
  	                    this.nextToken();
  	                    left = init;
  	                    right = this.parseExpression();
  	                    init = null;
  	                }
  	                else {
  	                    var previousAllowIn = this.context.allowIn;
  	                    this.context.allowIn = false;
  	                    var declarations = this.parseBindingList(kind, { inFor: true });
  	                    this.context.allowIn = previousAllowIn;
  	                    if (declarations.length === 1 && declarations[0].init === null && this.matchKeyword('in')) {
  	                        init = this.finalize(init, new Node.VariableDeclaration(declarations, kind));
  	                        this.nextToken();
  	                        left = init;
  	                        right = this.parseExpression();
  	                        init = null;
  	                    }
  	                    else if (declarations.length === 1 && declarations[0].init === null && this.matchContextualKeyword('of')) {
  	                        init = this.finalize(init, new Node.VariableDeclaration(declarations, kind));
  	                        this.nextToken();
  	                        left = init;
  	                        right = this.parseAssignmentExpression();
  	                        init = null;
  	                        forIn = false;
  	                    }
  	                    else {
  	                        this.consumeSemicolon();
  	                        init = this.finalize(init, new Node.VariableDeclaration(declarations, kind));
  	                    }
  	                }
  	            }
  	            else {
  	                var initStartToken = this.lookahead;
  	                var previousAllowIn = this.context.allowIn;
  	                this.context.allowIn = false;
  	                init = this.inheritCoverGrammar(this.parseAssignmentExpression);
  	                this.context.allowIn = previousAllowIn;
  	                if (this.matchKeyword('in')) {
  	                    if (!this.context.isAssignmentTarget || init.type === syntax_1.Syntax.AssignmentExpression) {
  	                        this.tolerateError(messages_1.Messages.InvalidLHSInForIn);
  	                    }
  	                    this.nextToken();
  	                    this.reinterpretExpressionAsPattern(init);
  	                    left = init;
  	                    right = this.parseExpression();
  	                    init = null;
  	                }
  	                else if (this.matchContextualKeyword('of')) {
  	                    if (!this.context.isAssignmentTarget || init.type === syntax_1.Syntax.AssignmentExpression) {
  	                        this.tolerateError(messages_1.Messages.InvalidLHSInForLoop);
  	                    }
  	                    this.nextToken();
  	                    this.reinterpretExpressionAsPattern(init);
  	                    left = init;
  	                    right = this.parseAssignmentExpression();
  	                    init = null;
  	                    forIn = false;
  	                }
  	                else {
  	                    if (this.match(',')) {
  	                        var initSeq = [init];
  	                        while (this.match(',')) {
  	                            this.nextToken();
  	                            initSeq.push(this.isolateCoverGrammar(this.parseAssignmentExpression));
  	                        }
  	                        init = this.finalize(this.startNode(initStartToken), new Node.SequenceExpression(initSeq));
  	                    }
  	                    this.expect(';');
  	                }
  	            }
  	        }
  	        if (typeof left === 'undefined') {
  	            if (!this.match(';')) {
  	                test = this.parseExpression();
  	            }
  	            this.expect(';');
  	            if (!this.match(')')) {
  	                update = this.parseExpression();
  	            }
  	        }
  	        var body;
  	        if (!this.match(')') && this.config.tolerant) {
  	            this.tolerateUnexpectedToken(this.nextToken());
  	            body = this.finalize(this.createNode(), new Node.EmptyStatement());
  	        }
  	        else {
  	            this.expect(')');
  	            var previousInIteration = this.context.inIteration;
  	            this.context.inIteration = true;
  	            body = this.isolateCoverGrammar(this.parseStatement);
  	            this.context.inIteration = previousInIteration;
  	        }
  	        return (typeof left === 'undefined') ?
  	            this.finalize(node, new Node.ForStatement(init, test, update, body)) :
  	            forIn ? this.finalize(node, new Node.ForInStatement(left, right, body)) :
  	                this.finalize(node, new Node.ForOfStatement(left, right, body));
  	    };
  	    // https://tc39.github.io/ecma262/#sec-continue-statement
  	    Parser.prototype.parseContinueStatement = function () {
  	        var node = this.createNode();
  	        this.expectKeyword('continue');
  	        var label = null;
  	        if (this.lookahead.type === 3 /* Identifier */ && !this.hasLineTerminator) {
  	            var id = this.parseVariableIdentifier();
  	            label = id;
  	            var key = '$' + id.name;
  	            if (!Object.prototype.hasOwnProperty.call(this.context.labelSet, key)) {
  	                this.throwError(messages_1.Messages.UnknownLabel, id.name);
  	            }
  	        }
  	        this.consumeSemicolon();
  	        if (label === null && !this.context.inIteration) {
  	            this.throwError(messages_1.Messages.IllegalContinue);
  	        }
  	        return this.finalize(node, new Node.ContinueStatement(label));
  	    };
  	    // https://tc39.github.io/ecma262/#sec-break-statement
  	    Parser.prototype.parseBreakStatement = function () {
  	        var node = this.createNode();
  	        this.expectKeyword('break');
  	        var label = null;
  	        if (this.lookahead.type === 3 /* Identifier */ && !this.hasLineTerminator) {
  	            var id = this.parseVariableIdentifier();
  	            var key = '$' + id.name;
  	            if (!Object.prototype.hasOwnProperty.call(this.context.labelSet, key)) {
  	                this.throwError(messages_1.Messages.UnknownLabel, id.name);
  	            }
  	            label = id;
  	        }
  	        this.consumeSemicolon();
  	        if (label === null && !this.context.inIteration && !this.context.inSwitch) {
  	            this.throwError(messages_1.Messages.IllegalBreak);
  	        }
  	        return this.finalize(node, new Node.BreakStatement(label));
  	    };
  	    // https://tc39.github.io/ecma262/#sec-return-statement
  	    Parser.prototype.parseReturnStatement = function () {
  	        if (!this.context.inFunctionBody) {
  	            this.tolerateError(messages_1.Messages.IllegalReturn);
  	        }
  	        var node = this.createNode();
  	        this.expectKeyword('return');
  	        var hasArgument = !this.match(';') && !this.match('}') &&
  	            !this.hasLineTerminator && this.lookahead.type !== 2 /* EOF */;
  	        var argument = hasArgument ? this.parseExpression() : null;
  	        this.consumeSemicolon();
  	        return this.finalize(node, new Node.ReturnStatement(argument));
  	    };
  	    // https://tc39.github.io/ecma262/#sec-with-statement
  	    Parser.prototype.parseWithStatement = function () {
  	        if (this.context.strict) {
  	            this.tolerateError(messages_1.Messages.StrictModeWith);
  	        }
  	        var node = this.createNode();
  	        var body;
  	        this.expectKeyword('with');
  	        this.expect('(');
  	        var object = this.parseExpression();
  	        if (!this.match(')') && this.config.tolerant) {
  	            this.tolerateUnexpectedToken(this.nextToken());
  	            body = this.finalize(this.createNode(), new Node.EmptyStatement());
  	        }
  	        else {
  	            this.expect(')');
  	            body = this.parseStatement();
  	        }
  	        return this.finalize(node, new Node.WithStatement(object, body));
  	    };
  	    // https://tc39.github.io/ecma262/#sec-switch-statement
  	    Parser.prototype.parseSwitchCase = function () {
  	        var node = this.createNode();
  	        var test;
  	        if (this.matchKeyword('default')) {
  	            this.nextToken();
  	            test = null;
  	        }
  	        else {
  	            this.expectKeyword('case');
  	            test = this.parseExpression();
  	        }
  	        this.expect(':');
  	        var consequent = [];
  	        while (true) {
  	            if (this.match('}') || this.matchKeyword('default') || this.matchKeyword('case')) {
  	                break;
  	            }
  	            consequent.push(this.parseStatementListItem());
  	        }
  	        return this.finalize(node, new Node.SwitchCase(test, consequent));
  	    };
  	    Parser.prototype.parseSwitchStatement = function () {
  	        var node = this.createNode();
  	        this.expectKeyword('switch');
  	        this.expect('(');
  	        var discriminant = this.parseExpression();
  	        this.expect(')');
  	        var previousInSwitch = this.context.inSwitch;
  	        this.context.inSwitch = true;
  	        var cases = [];
  	        var defaultFound = false;
  	        this.expect('{');
  	        while (true) {
  	            if (this.match('}')) {
  	                break;
  	            }
  	            var clause = this.parseSwitchCase();
  	            if (clause.test === null) {
  	                if (defaultFound) {
  	                    this.throwError(messages_1.Messages.MultipleDefaultsInSwitch);
  	                }
  	                defaultFound = true;
  	            }
  	            cases.push(clause);
  	        }
  	        this.expect('}');
  	        this.context.inSwitch = previousInSwitch;
  	        return this.finalize(node, new Node.SwitchStatement(discriminant, cases));
  	    };
  	    // https://tc39.github.io/ecma262/#sec-labelled-statements
  	    Parser.prototype.parseLabelledStatement = function () {
  	        var node = this.createNode();
  	        var expr = this.parseExpression();
  	        var statement;
  	        if ((expr.type === syntax_1.Syntax.Identifier) && this.match(':')) {
  	            this.nextToken();
  	            var id = expr;
  	            var key = '$' + id.name;
  	            if (Object.prototype.hasOwnProperty.call(this.context.labelSet, key)) {
  	                this.throwError(messages_1.Messages.Redeclaration, 'Label', id.name);
  	            }
  	            this.context.labelSet[key] = true;
  	            var body = void 0;
  	            if (this.matchKeyword('class')) {
  	                this.tolerateUnexpectedToken(this.lookahead);
  	                body = this.parseClassDeclaration();
  	            }
  	            else if (this.matchKeyword('function')) {
  	                var token = this.lookahead;
  	                var declaration = this.parseFunctionDeclaration();
  	                if (this.context.strict) {
  	                    this.tolerateUnexpectedToken(token, messages_1.Messages.StrictFunction);
  	                }
  	                else if (declaration.generator) {
  	                    this.tolerateUnexpectedToken(token, messages_1.Messages.GeneratorInLegacyContext);
  	                }
  	                body = declaration;
  	            }
  	            else {
  	                body = this.parseStatement();
  	            }
  	            delete this.context.labelSet[key];
  	            statement = new Node.LabeledStatement(id, body);
  	        }
  	        else {
  	            this.consumeSemicolon();
  	            statement = new Node.ExpressionStatement(expr);
  	        }
  	        return this.finalize(node, statement);
  	    };
  	    // https://tc39.github.io/ecma262/#sec-throw-statement
  	    Parser.prototype.parseThrowStatement = function () {
  	        var node = this.createNode();
  	        this.expectKeyword('throw');
  	        if (this.hasLineTerminator) {
  	            this.throwError(messages_1.Messages.NewlineAfterThrow);
  	        }
  	        var argument = this.parseExpression();
  	        this.consumeSemicolon();
  	        return this.finalize(node, new Node.ThrowStatement(argument));
  	    };
  	    // https://tc39.github.io/ecma262/#sec-try-statement
  	    Parser.prototype.parseCatchClause = function () {
  	        var node = this.createNode();
  	        this.expectKeyword('catch');
  	        this.expect('(');
  	        if (this.match(')')) {
  	            this.throwUnexpectedToken(this.lookahead);
  	        }
  	        var params = [];
  	        var param = this.parsePattern(params);
  	        var paramMap = {};
  	        for (var i = 0; i < params.length; i++) {
  	            var key = '$' + params[i].value;
  	            if (Object.prototype.hasOwnProperty.call(paramMap, key)) {
  	                this.tolerateError(messages_1.Messages.DuplicateBinding, params[i].value);
  	            }
  	            paramMap[key] = true;
  	        }
  	        if (this.context.strict && param.type === syntax_1.Syntax.Identifier) {
  	            if (this.scanner.isRestrictedWord(param.name)) {
  	                this.tolerateError(messages_1.Messages.StrictCatchVariable);
  	            }
  	        }
  	        this.expect(')');
  	        var body = this.parseBlock();
  	        return this.finalize(node, new Node.CatchClause(param, body));
  	    };
  	    Parser.prototype.parseFinallyClause = function () {
  	        this.expectKeyword('finally');
  	        return this.parseBlock();
  	    };
  	    Parser.prototype.parseTryStatement = function () {
  	        var node = this.createNode();
  	        this.expectKeyword('try');
  	        var block = this.parseBlock();
  	        var handler = this.matchKeyword('catch') ? this.parseCatchClause() : null;
  	        var finalizer = this.matchKeyword('finally') ? this.parseFinallyClause() : null;
  	        if (!handler && !finalizer) {
  	            this.throwError(messages_1.Messages.NoCatchOrFinally);
  	        }
  	        return this.finalize(node, new Node.TryStatement(block, handler, finalizer));
  	    };
  	    // https://tc39.github.io/ecma262/#sec-debugger-statement
  	    Parser.prototype.parseDebuggerStatement = function () {
  	        var node = this.createNode();
  	        this.expectKeyword('debugger');
  	        this.consumeSemicolon();
  	        return this.finalize(node, new Node.DebuggerStatement());
  	    };
  	    // https://tc39.github.io/ecma262/#sec-ecmascript-language-statements-and-declarations
  	    Parser.prototype.parseStatement = function () {
  	        var statement;
  	        switch (this.lookahead.type) {
  	            case 1 /* BooleanLiteral */:
  	            case 5 /* NullLiteral */:
  	            case 6 /* NumericLiteral */:
  	            case 8 /* StringLiteral */:
  	            case 10 /* Template */:
  	            case 9 /* RegularExpression */:
  	                statement = this.parseExpressionStatement();
  	                break;
  	            case 7 /* Punctuator */:
  	                var value = this.lookahead.value;
  	                if (value === '{') {
  	                    statement = this.parseBlock();
  	                }
  	                else if (value === '(') {
  	                    statement = this.parseExpressionStatement();
  	                }
  	                else if (value === ';') {
  	                    statement = this.parseEmptyStatement();
  	                }
  	                else {
  	                    statement = this.parseExpressionStatement();
  	                }
  	                break;
  	            case 3 /* Identifier */:
  	                statement = this.matchAsyncFunction() ? this.parseFunctionDeclaration() : this.parseLabelledStatement();
  	                break;
  	            case 4 /* Keyword */:
  	                switch (this.lookahead.value) {
  	                    case 'break':
  	                        statement = this.parseBreakStatement();
  	                        break;
  	                    case 'continue':
  	                        statement = this.parseContinueStatement();
  	                        break;
  	                    case 'debugger':
  	                        statement = this.parseDebuggerStatement();
  	                        break;
  	                    case 'do':
  	                        statement = this.parseDoWhileStatement();
  	                        break;
  	                    case 'for':
  	                        statement = this.parseForStatement();
  	                        break;
  	                    case 'function':
  	                        statement = this.parseFunctionDeclaration();
  	                        break;
  	                    case 'if':
  	                        statement = this.parseIfStatement();
  	                        break;
  	                    case 'return':
  	                        statement = this.parseReturnStatement();
  	                        break;
  	                    case 'switch':
  	                        statement = this.parseSwitchStatement();
  	                        break;
  	                    case 'throw':
  	                        statement = this.parseThrowStatement();
  	                        break;
  	                    case 'try':
  	                        statement = this.parseTryStatement();
  	                        break;
  	                    case 'var':
  	                        statement = this.parseVariableStatement();
  	                        break;
  	                    case 'while':
  	                        statement = this.parseWhileStatement();
  	                        break;
  	                    case 'with':
  	                        statement = this.parseWithStatement();
  	                        break;
  	                    default:
  	                        statement = this.parseExpressionStatement();
  	                        break;
  	                }
  	                break;
  	            default:
  	                statement = this.throwUnexpectedToken(this.lookahead);
  	        }
  	        return statement;
  	    };
  	    // https://tc39.github.io/ecma262/#sec-function-definitions
  	    Parser.prototype.parseFunctionSourceElements = function () {
  	        var node = this.createNode();
  	        this.expect('{');
  	        var body = this.parseDirectivePrologues();
  	        var previousLabelSet = this.context.labelSet;
  	        var previousInIteration = this.context.inIteration;
  	        var previousInSwitch = this.context.inSwitch;
  	        var previousInFunctionBody = this.context.inFunctionBody;
  	        this.context.labelSet = {};
  	        this.context.inIteration = false;
  	        this.context.inSwitch = false;
  	        this.context.inFunctionBody = true;
  	        while (this.lookahead.type !== 2 /* EOF */) {
  	            if (this.match('}')) {
  	                break;
  	            }
  	            body.push(this.parseStatementListItem());
  	        }
  	        this.expect('}');
  	        this.context.labelSet = previousLabelSet;
  	        this.context.inIteration = previousInIteration;
  	        this.context.inSwitch = previousInSwitch;
  	        this.context.inFunctionBody = previousInFunctionBody;
  	        return this.finalize(node, new Node.BlockStatement(body));
  	    };
  	    Parser.prototype.validateParam = function (options, param, name) {
  	        var key = '$' + name;
  	        if (this.context.strict) {
  	            if (this.scanner.isRestrictedWord(name)) {
  	                options.stricted = param;
  	                options.message = messages_1.Messages.StrictParamName;
  	            }
  	            if (Object.prototype.hasOwnProperty.call(options.paramSet, key)) {
  	                options.stricted = param;
  	                options.message = messages_1.Messages.StrictParamDupe;
  	            }
  	        }
  	        else if (!options.firstRestricted) {
  	            if (this.scanner.isRestrictedWord(name)) {
  	                options.firstRestricted = param;
  	                options.message = messages_1.Messages.StrictParamName;
  	            }
  	            else if (this.scanner.isStrictModeReservedWord(name)) {
  	                options.firstRestricted = param;
  	                options.message = messages_1.Messages.StrictReservedWord;
  	            }
  	            else if (Object.prototype.hasOwnProperty.call(options.paramSet, key)) {
  	                options.stricted = param;
  	                options.message = messages_1.Messages.StrictParamDupe;
  	            }
  	        }
  	        /* istanbul ignore next */
  	        if (typeof Object.defineProperty === 'function') {
  	            Object.defineProperty(options.paramSet, key, { value: true, enumerable: true, writable: true, configurable: true });
  	        }
  	        else {
  	            options.paramSet[key] = true;
  	        }
  	    };
  	    Parser.prototype.parseRestElement = function (params) {
  	        var node = this.createNode();
  	        this.expect('...');
  	        var arg = this.parsePattern(params);
  	        if (this.match('=')) {
  	            this.throwError(messages_1.Messages.DefaultRestParameter);
  	        }
  	        if (!this.match(')')) {
  	            this.throwError(messages_1.Messages.ParameterAfterRestParameter);
  	        }
  	        return this.finalize(node, new Node.RestElement(arg));
  	    };
  	    Parser.prototype.parseFormalParameter = function (options) {
  	        var params = [];
  	        var param = this.match('...') ? this.parseRestElement(params) : this.parsePatternWithDefault(params);
  	        for (var i = 0; i < params.length; i++) {
  	            this.validateParam(options, params[i], params[i].value);
  	        }
  	        options.simple = options.simple && (param instanceof Node.Identifier);
  	        options.params.push(param);
  	    };
  	    Parser.prototype.parseFormalParameters = function (firstRestricted) {
  	        var options;
  	        options = {
  	            simple: true,
  	            params: [],
  	            firstRestricted: firstRestricted
  	        };
  	        this.expect('(');
  	        if (!this.match(')')) {
  	            options.paramSet = {};
  	            while (this.lookahead.type !== 2 /* EOF */) {
  	                this.parseFormalParameter(options);
  	                if (this.match(')')) {
  	                    break;
  	                }
  	                this.expect(',');
  	                if (this.match(')')) {
  	                    break;
  	                }
  	            }
  	        }
  	        this.expect(')');
  	        return {
  	            simple: options.simple,
  	            params: options.params,
  	            stricted: options.stricted,
  	            firstRestricted: options.firstRestricted,
  	            message: options.message
  	        };
  	    };
  	    Parser.prototype.matchAsyncFunction = function () {
  	        var match = this.matchContextualKeyword('async');
  	        if (match) {
  	            var state = this.scanner.saveState();
  	            this.scanner.scanComments();
  	            var next = this.scanner.lex();
  	            this.scanner.restoreState(state);
  	            match = (state.lineNumber === next.lineNumber) && (next.type === 4 /* Keyword */) && (next.value === 'function');
  	        }
  	        return match;
  	    };
  	    Parser.prototype.parseFunctionDeclaration = function (identifierIsOptional) {
  	        var node = this.createNode();
  	        var isAsync = this.matchContextualKeyword('async');
  	        if (isAsync) {
  	            this.nextToken();
  	        }
  	        this.expectKeyword('function');
  	        var isGenerator = isAsync ? false : this.match('*');
  	        if (isGenerator) {
  	            this.nextToken();
  	        }
  	        var message;
  	        var id = null;
  	        var firstRestricted = null;
  	        if (!identifierIsOptional || !this.match('(')) {
  	            var token = this.lookahead;
  	            id = this.parseVariableIdentifier();
  	            if (this.context.strict) {
  	                if (this.scanner.isRestrictedWord(token.value)) {
  	                    this.tolerateUnexpectedToken(token, messages_1.Messages.StrictFunctionName);
  	                }
  	            }
  	            else {
  	                if (this.scanner.isRestrictedWord(token.value)) {
  	                    firstRestricted = token;
  	                    message = messages_1.Messages.StrictFunctionName;
  	                }
  	                else if (this.scanner.isStrictModeReservedWord(token.value)) {
  	                    firstRestricted = token;
  	                    message = messages_1.Messages.StrictReservedWord;
  	                }
  	            }
  	        }
  	        var previousAllowAwait = this.context.await;
  	        var previousAllowYield = this.context.allowYield;
  	        this.context.await = isAsync;
  	        this.context.allowYield = !isGenerator;
  	        var formalParameters = this.parseFormalParameters(firstRestricted);
  	        var params = formalParameters.params;
  	        var stricted = formalParameters.stricted;
  	        firstRestricted = formalParameters.firstRestricted;
  	        if (formalParameters.message) {
  	            message = formalParameters.message;
  	        }
  	        var previousStrict = this.context.strict;
  	        var previousAllowStrictDirective = this.context.allowStrictDirective;
  	        this.context.allowStrictDirective = formalParameters.simple;
  	        var body = this.parseFunctionSourceElements();
  	        if (this.context.strict && firstRestricted) {
  	            this.throwUnexpectedToken(firstRestricted, message);
  	        }
  	        if (this.context.strict && stricted) {
  	            this.tolerateUnexpectedToken(stricted, message);
  	        }
  	        this.context.strict = previousStrict;
  	        this.context.allowStrictDirective = previousAllowStrictDirective;
  	        this.context.await = previousAllowAwait;
  	        this.context.allowYield = previousAllowYield;
  	        return isAsync ? this.finalize(node, new Node.AsyncFunctionDeclaration(id, params, body)) :
  	            this.finalize(node, new Node.FunctionDeclaration(id, params, body, isGenerator));
  	    };
  	    Parser.prototype.parseFunctionExpression = function () {
  	        var node = this.createNode();
  	        var isAsync = this.matchContextualKeyword('async');
  	        if (isAsync) {
  	            this.nextToken();
  	        }
  	        this.expectKeyword('function');
  	        var isGenerator = isAsync ? false : this.match('*');
  	        if (isGenerator) {
  	            this.nextToken();
  	        }
  	        var message;
  	        var id = null;
  	        var firstRestricted;
  	        var previousAllowAwait = this.context.await;
  	        var previousAllowYield = this.context.allowYield;
  	        this.context.await = isAsync;
  	        this.context.allowYield = !isGenerator;
  	        if (!this.match('(')) {
  	            var token = this.lookahead;
  	            id = (!this.context.strict && !isGenerator && this.matchKeyword('yield')) ? this.parseIdentifierName() : this.parseVariableIdentifier();
  	            if (this.context.strict) {
  	                if (this.scanner.isRestrictedWord(token.value)) {
  	                    this.tolerateUnexpectedToken(token, messages_1.Messages.StrictFunctionName);
  	                }
  	            }
  	            else {
  	                if (this.scanner.isRestrictedWord(token.value)) {
  	                    firstRestricted = token;
  	                    message = messages_1.Messages.StrictFunctionName;
  	                }
  	                else if (this.scanner.isStrictModeReservedWord(token.value)) {
  	                    firstRestricted = token;
  	                    message = messages_1.Messages.StrictReservedWord;
  	                }
  	            }
  	        }
  	        var formalParameters = this.parseFormalParameters(firstRestricted);
  	        var params = formalParameters.params;
  	        var stricted = formalParameters.stricted;
  	        firstRestricted = formalParameters.firstRestricted;
  	        if (formalParameters.message) {
  	            message = formalParameters.message;
  	        }
  	        var previousStrict = this.context.strict;
  	        var previousAllowStrictDirective = this.context.allowStrictDirective;
  	        this.context.allowStrictDirective = formalParameters.simple;
  	        var body = this.parseFunctionSourceElements();
  	        if (this.context.strict && firstRestricted) {
  	            this.throwUnexpectedToken(firstRestricted, message);
  	        }
  	        if (this.context.strict && stricted) {
  	            this.tolerateUnexpectedToken(stricted, message);
  	        }
  	        this.context.strict = previousStrict;
  	        this.context.allowStrictDirective = previousAllowStrictDirective;
  	        this.context.await = previousAllowAwait;
  	        this.context.allowYield = previousAllowYield;
  	        return isAsync ? this.finalize(node, new Node.AsyncFunctionExpression(id, params, body)) :
  	            this.finalize(node, new Node.FunctionExpression(id, params, body, isGenerator));
  	    };
  	    // https://tc39.github.io/ecma262/#sec-directive-prologues-and-the-use-strict-directive
  	    Parser.prototype.parseDirective = function () {
  	        var token = this.lookahead;
  	        var node = this.createNode();
  	        var expr = this.parseExpression();
  	        var directive = (expr.type === syntax_1.Syntax.Literal) ? this.getTokenRaw(token).slice(1, -1) : null;
  	        this.consumeSemicolon();
  	        return this.finalize(node, directive ? new Node.Directive(expr, directive) : new Node.ExpressionStatement(expr));
  	    };
  	    Parser.prototype.parseDirectivePrologues = function () {
  	        var firstRestricted = null;
  	        var body = [];
  	        while (true) {
  	            var token = this.lookahead;
  	            if (token.type !== 8 /* StringLiteral */) {
  	                break;
  	            }
  	            var statement = this.parseDirective();
  	            body.push(statement);
  	            var directive = statement.directive;
  	            if (typeof directive !== 'string') {
  	                break;
  	            }
  	            if (directive === 'use strict') {
  	                this.context.strict = true;
  	                if (firstRestricted) {
  	                    this.tolerateUnexpectedToken(firstRestricted, messages_1.Messages.StrictOctalLiteral);
  	                }
  	                if (!this.context.allowStrictDirective) {
  	                    this.tolerateUnexpectedToken(token, messages_1.Messages.IllegalLanguageModeDirective);
  	                }
  	            }
  	            else {
  	                if (!firstRestricted && token.octal) {
  	                    firstRestricted = token;
  	                }
  	            }
  	        }
  	        return body;
  	    };
  	    // https://tc39.github.io/ecma262/#sec-method-definitions
  	    Parser.prototype.qualifiedPropertyName = function (token) {
  	        switch (token.type) {
  	            case 3 /* Identifier */:
  	            case 8 /* StringLiteral */:
  	            case 1 /* BooleanLiteral */:
  	            case 5 /* NullLiteral */:
  	            case 6 /* NumericLiteral */:
  	            case 4 /* Keyword */:
  	                return true;
  	            case 7 /* Punctuator */:
  	                return token.value === '[';
  	            default:
  	                break;
  	        }
  	        return false;
  	    };
  	    Parser.prototype.parseGetterMethod = function () {
  	        var node = this.createNode();
  	        var isGenerator = false;
  	        var previousAllowYield = this.context.allowYield;
  	        this.context.allowYield = false;
  	        var formalParameters = this.parseFormalParameters();
  	        if (formalParameters.params.length > 0) {
  	            this.tolerateError(messages_1.Messages.BadGetterArity);
  	        }
  	        var method = this.parsePropertyMethod(formalParameters);
  	        this.context.allowYield = previousAllowYield;
  	        return this.finalize(node, new Node.FunctionExpression(null, formalParameters.params, method, isGenerator));
  	    };
  	    Parser.prototype.parseSetterMethod = function () {
  	        var node = this.createNode();
  	        var isGenerator = false;
  	        var previousAllowYield = this.context.allowYield;
  	        this.context.allowYield = false;
  	        var formalParameters = this.parseFormalParameters();
  	        if (formalParameters.params.length !== 1) {
  	            this.tolerateError(messages_1.Messages.BadSetterArity);
  	        }
  	        else if (formalParameters.params[0] instanceof Node.RestElement) {
  	            this.tolerateError(messages_1.Messages.BadSetterRestParameter);
  	        }
  	        var method = this.parsePropertyMethod(formalParameters);
  	        this.context.allowYield = previousAllowYield;
  	        return this.finalize(node, new Node.FunctionExpression(null, formalParameters.params, method, isGenerator));
  	    };
  	    Parser.prototype.parseGeneratorMethod = function () {
  	        var node = this.createNode();
  	        var isGenerator = true;
  	        var previousAllowYield = this.context.allowYield;
  	        this.context.allowYield = true;
  	        var params = this.parseFormalParameters();
  	        this.context.allowYield = false;
  	        var method = this.parsePropertyMethod(params);
  	        this.context.allowYield = previousAllowYield;
  	        return this.finalize(node, new Node.FunctionExpression(null, params.params, method, isGenerator));
  	    };
  	    // https://tc39.github.io/ecma262/#sec-generator-function-definitions
  	    Parser.prototype.isStartOfExpression = function () {
  	        var start = true;
  	        var value = this.lookahead.value;
  	        switch (this.lookahead.type) {
  	            case 7 /* Punctuator */:
  	                start = (value === '[') || (value === '(') || (value === '{') ||
  	                    (value === '+') || (value === '-') ||
  	                    (value === '!') || (value === '~') ||
  	                    (value === '++') || (value === '--') ||
  	                    (value === '/') || (value === '/='); // regular expression literal
  	                break;
  	            case 4 /* Keyword */:
  	                start = (value === 'class') || (value === 'delete') ||
  	                    (value === 'function') || (value === 'let') || (value === 'new') ||
  	                    (value === 'super') || (value === 'this') || (value === 'typeof') ||
  	                    (value === 'void') || (value === 'yield');
  	                break;
  	            default:
  	                break;
  	        }
  	        return start;
  	    };
  	    Parser.prototype.parseYieldExpression = function () {
  	        var node = this.createNode();
  	        this.expectKeyword('yield');
  	        var argument = null;
  	        var delegate = false;
  	        if (!this.hasLineTerminator) {
  	            var previousAllowYield = this.context.allowYield;
  	            this.context.allowYield = false;
  	            delegate = this.match('*');
  	            if (delegate) {
  	                this.nextToken();
  	                argument = this.parseAssignmentExpression();
  	            }
  	            else if (this.isStartOfExpression()) {
  	                argument = this.parseAssignmentExpression();
  	            }
  	            this.context.allowYield = previousAllowYield;
  	        }
  	        return this.finalize(node, new Node.YieldExpression(argument, delegate));
  	    };
  	    // https://tc39.github.io/ecma262/#sec-class-definitions
  	    Parser.prototype.parseClassElement = function (hasConstructor) {
  	        var token = this.lookahead;
  	        var node = this.createNode();
  	        var kind = '';
  	        var key = null;
  	        var value = null;
  	        var computed = false;
  	        var method = false;
  	        var isStatic = false;
  	        var isAsync = false;
  	        if (this.match('*')) {
  	            this.nextToken();
  	        }
  	        else {
  	            computed = this.match('[');
  	            key = this.parseObjectPropertyKey();
  	            var id = key;
  	            if (id.name === 'static' && (this.qualifiedPropertyName(this.lookahead) || this.match('*'))) {
  	                token = this.lookahead;
  	                isStatic = true;
  	                computed = this.match('[');
  	                if (this.match('*')) {
  	                    this.nextToken();
  	                }
  	                else {
  	                    key = this.parseObjectPropertyKey();
  	                }
  	            }
  	            if ((token.type === 3 /* Identifier */) && !this.hasLineTerminator && (token.value === 'async')) {
  	                var punctuator = this.lookahead.value;
  	                if (punctuator !== ':' && punctuator !== '(' && punctuator !== '*') {
  	                    isAsync = true;
  	                    token = this.lookahead;
  	                    key = this.parseObjectPropertyKey();
  	                    if (token.type === 3 /* Identifier */) {
  	                        if (token.value === 'get' || token.value === 'set') {
  	                            this.tolerateUnexpectedToken(token);
  	                        }
  	                        else if (token.value === 'constructor') {
  	                            this.tolerateUnexpectedToken(token, messages_1.Messages.ConstructorIsAsync);
  	                        }
  	                    }
  	                }
  	            }
  	        }
  	        var lookaheadPropertyKey = this.qualifiedPropertyName(this.lookahead);
  	        if (token.type === 3 /* Identifier */) {
  	            if (token.value === 'get' && lookaheadPropertyKey) {
  	                kind = 'get';
  	                computed = this.match('[');
  	                key = this.parseObjectPropertyKey();
  	                this.context.allowYield = false;
  	                value = this.parseGetterMethod();
  	            }
  	            else if (token.value === 'set' && lookaheadPropertyKey) {
  	                kind = 'set';
  	                computed = this.match('[');
  	                key = this.parseObjectPropertyKey();
  	                value = this.parseSetterMethod();
  	            }
  	        }
  	        else if (token.type === 7 /* Punctuator */ && token.value === '*' && lookaheadPropertyKey) {
  	            kind = 'init';
  	            computed = this.match('[');
  	            key = this.parseObjectPropertyKey();
  	            value = this.parseGeneratorMethod();
  	            method = true;
  	        }
  	        if (!kind && key && this.match('(')) {
  	            kind = 'init';
  	            value = isAsync ? this.parsePropertyMethodAsyncFunction() : this.parsePropertyMethodFunction();
  	            method = true;
  	        }
  	        if (!kind) {
  	            this.throwUnexpectedToken(this.lookahead);
  	        }
  	        if (kind === 'init') {
  	            kind = 'method';
  	        }
  	        if (!computed) {
  	            if (isStatic && this.isPropertyKey(key, 'prototype')) {
  	                this.throwUnexpectedToken(token, messages_1.Messages.StaticPrototype);
  	            }
  	            if (!isStatic && this.isPropertyKey(key, 'constructor')) {
  	                if (kind !== 'method' || !method || (value && value.generator)) {
  	                    this.throwUnexpectedToken(token, messages_1.Messages.ConstructorSpecialMethod);
  	                }
  	                if (hasConstructor.value) {
  	                    this.throwUnexpectedToken(token, messages_1.Messages.DuplicateConstructor);
  	                }
  	                else {
  	                    hasConstructor.value = true;
  	                }
  	                kind = 'constructor';
  	            }
  	        }
  	        return this.finalize(node, new Node.MethodDefinition(key, computed, value, kind, isStatic));
  	    };
  	    Parser.prototype.parseClassElementList = function () {
  	        var body = [];
  	        var hasConstructor = { value: false };
  	        this.expect('{');
  	        while (!this.match('}')) {
  	            if (this.match(';')) {
  	                this.nextToken();
  	            }
  	            else {
  	                body.push(this.parseClassElement(hasConstructor));
  	            }
  	        }
  	        this.expect('}');
  	        return body;
  	    };
  	    Parser.prototype.parseClassBody = function () {
  	        var node = this.createNode();
  	        var elementList = this.parseClassElementList();
  	        return this.finalize(node, new Node.ClassBody(elementList));
  	    };
  	    Parser.prototype.parseClassDeclaration = function (identifierIsOptional) {
  	        var node = this.createNode();
  	        var previousStrict = this.context.strict;
  	        this.context.strict = true;
  	        this.expectKeyword('class');
  	        var id = (identifierIsOptional && (this.lookahead.type !== 3 /* Identifier */)) ? null : this.parseVariableIdentifier();
  	        var superClass = null;
  	        if (this.matchKeyword('extends')) {
  	            this.nextToken();
  	            superClass = this.isolateCoverGrammar(this.parseLeftHandSideExpressionAllowCall);
  	        }
  	        var classBody = this.parseClassBody();
  	        this.context.strict = previousStrict;
  	        return this.finalize(node, new Node.ClassDeclaration(id, superClass, classBody));
  	    };
  	    Parser.prototype.parseClassExpression = function () {
  	        var node = this.createNode();
  	        var previousStrict = this.context.strict;
  	        this.context.strict = true;
  	        this.expectKeyword('class');
  	        var id = (this.lookahead.type === 3 /* Identifier */) ? this.parseVariableIdentifier() : null;
  	        var superClass = null;
  	        if (this.matchKeyword('extends')) {
  	            this.nextToken();
  	            superClass = this.isolateCoverGrammar(this.parseLeftHandSideExpressionAllowCall);
  	        }
  	        var classBody = this.parseClassBody();
  	        this.context.strict = previousStrict;
  	        return this.finalize(node, new Node.ClassExpression(id, superClass, classBody));
  	    };
  	    // https://tc39.github.io/ecma262/#sec-scripts
  	    // https://tc39.github.io/ecma262/#sec-modules
  	    Parser.prototype.parseModule = function () {
  	        this.context.strict = true;
  	        this.context.isModule = true;
  	        var node = this.createNode();
  	        var body = this.parseDirectivePrologues();
  	        while (this.lookahead.type !== 2 /* EOF */) {
  	            body.push(this.parseStatementListItem());
  	        }
  	        return this.finalize(node, new Node.Module(body));
  	    };
  	    Parser.prototype.parseScript = function () {
  	        var node = this.createNode();
  	        var body = this.parseDirectivePrologues();
  	        while (this.lookahead.type !== 2 /* EOF */) {
  	            body.push(this.parseStatementListItem());
  	        }
  	        return this.finalize(node, new Node.Script(body));
  	    };
  	    // https://tc39.github.io/ecma262/#sec-imports
  	    Parser.prototype.parseModuleSpecifier = function () {
  	        var node = this.createNode();
  	        if (this.lookahead.type !== 8 /* StringLiteral */) {
  	            this.throwError(messages_1.Messages.InvalidModuleSpecifier);
  	        }
  	        var token = this.nextToken();
  	        var raw = this.getTokenRaw(token);
  	        return this.finalize(node, new Node.Literal(token.value, raw));
  	    };
  	    // import {<foo as bar>} ...;
  	    Parser.prototype.parseImportSpecifier = function () {
  	        var node = this.createNode();
  	        var imported;
  	        var local;
  	        if (this.lookahead.type === 3 /* Identifier */) {
  	            imported = this.parseVariableIdentifier();
  	            local = imported;
  	            if (this.matchContextualKeyword('as')) {
  	                this.nextToken();
  	                local = this.parseVariableIdentifier();
  	            }
  	        }
  	        else {
  	            imported = this.parseIdentifierName();
  	            local = imported;
  	            if (this.matchContextualKeyword('as')) {
  	                this.nextToken();
  	                local = this.parseVariableIdentifier();
  	            }
  	            else {
  	                this.throwUnexpectedToken(this.nextToken());
  	            }
  	        }
  	        return this.finalize(node, new Node.ImportSpecifier(local, imported));
  	    };
  	    // {foo, bar as bas}
  	    Parser.prototype.parseNamedImports = function () {
  	        this.expect('{');
  	        var specifiers = [];
  	        while (!this.match('}')) {
  	            specifiers.push(this.parseImportSpecifier());
  	            if (!this.match('}')) {
  	                this.expect(',');
  	            }
  	        }
  	        this.expect('}');
  	        return specifiers;
  	    };
  	    // import <foo> ...;
  	    Parser.prototype.parseImportDefaultSpecifier = function () {
  	        var node = this.createNode();
  	        var local = this.parseIdentifierName();
  	        return this.finalize(node, new Node.ImportDefaultSpecifier(local));
  	    };
  	    // import <* as foo> ...;
  	    Parser.prototype.parseImportNamespaceSpecifier = function () {
  	        var node = this.createNode();
  	        this.expect('*');
  	        if (!this.matchContextualKeyword('as')) {
  	            this.throwError(messages_1.Messages.NoAsAfterImportNamespace);
  	        }
  	        this.nextToken();
  	        var local = this.parseIdentifierName();
  	        return this.finalize(node, new Node.ImportNamespaceSpecifier(local));
  	    };
  	    Parser.prototype.parseImportDeclaration = function () {
  	        if (this.context.inFunctionBody) {
  	            this.throwError(messages_1.Messages.IllegalImportDeclaration);
  	        }
  	        var node = this.createNode();
  	        this.expectKeyword('import');
  	        var src;
  	        var specifiers = [];
  	        if (this.lookahead.type === 8 /* StringLiteral */) {
  	            // import 'foo';
  	            src = this.parseModuleSpecifier();
  	        }
  	        else {
  	            if (this.match('{')) {
  	                // import {bar}
  	                specifiers = specifiers.concat(this.parseNamedImports());
  	            }
  	            else if (this.match('*')) {
  	                // import * as foo
  	                specifiers.push(this.parseImportNamespaceSpecifier());
  	            }
  	            else if (this.isIdentifierName(this.lookahead) && !this.matchKeyword('default')) {
  	                // import foo
  	                specifiers.push(this.parseImportDefaultSpecifier());
  	                if (this.match(',')) {
  	                    this.nextToken();
  	                    if (this.match('*')) {
  	                        // import foo, * as foo
  	                        specifiers.push(this.parseImportNamespaceSpecifier());
  	                    }
  	                    else if (this.match('{')) {
  	                        // import foo, {bar}
  	                        specifiers = specifiers.concat(this.parseNamedImports());
  	                    }
  	                    else {
  	                        this.throwUnexpectedToken(this.lookahead);
  	                    }
  	                }
  	            }
  	            else {
  	                this.throwUnexpectedToken(this.nextToken());
  	            }
  	            if (!this.matchContextualKeyword('from')) {
  	                var message = this.lookahead.value ? messages_1.Messages.UnexpectedToken : messages_1.Messages.MissingFromClause;
  	                this.throwError(message, this.lookahead.value);
  	            }
  	            this.nextToken();
  	            src = this.parseModuleSpecifier();
  	        }
  	        this.consumeSemicolon();
  	        return this.finalize(node, new Node.ImportDeclaration(specifiers, src));
  	    };
  	    // https://tc39.github.io/ecma262/#sec-exports
  	    Parser.prototype.parseExportSpecifier = function () {
  	        var node = this.createNode();
  	        var local = this.parseIdentifierName();
  	        var exported = local;
  	        if (this.matchContextualKeyword('as')) {
  	            this.nextToken();
  	            exported = this.parseIdentifierName();
  	        }
  	        return this.finalize(node, new Node.ExportSpecifier(local, exported));
  	    };
  	    Parser.prototype.parseExportDeclaration = function () {
  	        if (this.context.inFunctionBody) {
  	            this.throwError(messages_1.Messages.IllegalExportDeclaration);
  	        }
  	        var node = this.createNode();
  	        this.expectKeyword('export');
  	        var exportDeclaration;
  	        if (this.matchKeyword('default')) {
  	            // export default ...
  	            this.nextToken();
  	            if (this.matchKeyword('function')) {
  	                // export default function foo () {}
  	                // export default function () {}
  	                var declaration = this.parseFunctionDeclaration(true);
  	                exportDeclaration = this.finalize(node, new Node.ExportDefaultDeclaration(declaration));
  	            }
  	            else if (this.matchKeyword('class')) {
  	                // export default class foo {}
  	                var declaration = this.parseClassDeclaration(true);
  	                exportDeclaration = this.finalize(node, new Node.ExportDefaultDeclaration(declaration));
  	            }
  	            else if (this.matchContextualKeyword('async')) {
  	                // export default async function f () {}
  	                // export default async function () {}
  	                // export default async x => x
  	                var declaration = this.matchAsyncFunction() ? this.parseFunctionDeclaration(true) : this.parseAssignmentExpression();
  	                exportDeclaration = this.finalize(node, new Node.ExportDefaultDeclaration(declaration));
  	            }
  	            else {
  	                if (this.matchContextualKeyword('from')) {
  	                    this.throwError(messages_1.Messages.UnexpectedToken, this.lookahead.value);
  	                }
  	                // export default {};
  	                // export default [];
  	                // export default (1 + 2);
  	                var declaration = this.match('{') ? this.parseObjectInitializer() :
  	                    this.match('[') ? this.parseArrayInitializer() : this.parseAssignmentExpression();
  	                this.consumeSemicolon();
  	                exportDeclaration = this.finalize(node, new Node.ExportDefaultDeclaration(declaration));
  	            }
  	        }
  	        else if (this.match('*')) {
  	            // export * from 'foo';
  	            this.nextToken();
  	            if (!this.matchContextualKeyword('from')) {
  	                var message = this.lookahead.value ? messages_1.Messages.UnexpectedToken : messages_1.Messages.MissingFromClause;
  	                this.throwError(message, this.lookahead.value);
  	            }
  	            this.nextToken();
  	            var src = this.parseModuleSpecifier();
  	            this.consumeSemicolon();
  	            exportDeclaration = this.finalize(node, new Node.ExportAllDeclaration(src));
  	        }
  	        else if (this.lookahead.type === 4 /* Keyword */) {
  	            // export var f = 1;
  	            var declaration = void 0;
  	            switch (this.lookahead.value) {
  	                case 'let':
  	                case 'const':
  	                    declaration = this.parseLexicalDeclaration({ inFor: false });
  	                    break;
  	                case 'var':
  	                case 'class':
  	                case 'function':
  	                    declaration = this.parseStatementListItem();
  	                    break;
  	                default:
  	                    this.throwUnexpectedToken(this.lookahead);
  	            }
  	            exportDeclaration = this.finalize(node, new Node.ExportNamedDeclaration(declaration, [], null));
  	        }
  	        else if (this.matchAsyncFunction()) {
  	            var declaration = this.parseFunctionDeclaration();
  	            exportDeclaration = this.finalize(node, new Node.ExportNamedDeclaration(declaration, [], null));
  	        }
  	        else {
  	            var specifiers = [];
  	            var source = null;
  	            var isExportFromIdentifier = false;
  	            this.expect('{');
  	            while (!this.match('}')) {
  	                isExportFromIdentifier = isExportFromIdentifier || this.matchKeyword('default');
  	                specifiers.push(this.parseExportSpecifier());
  	                if (!this.match('}')) {
  	                    this.expect(',');
  	                }
  	            }
  	            this.expect('}');
  	            if (this.matchContextualKeyword('from')) {
  	                // export {default} from 'foo';
  	                // export {foo} from 'foo';
  	                this.nextToken();
  	                source = this.parseModuleSpecifier();
  	                this.consumeSemicolon();
  	            }
  	            else if (isExportFromIdentifier) {
  	                // export {default}; // missing fromClause
  	                var message = this.lookahead.value ? messages_1.Messages.UnexpectedToken : messages_1.Messages.MissingFromClause;
  	                this.throwError(message, this.lookahead.value);
  	            }
  	            else {
  	                // export {foo};
  	                this.consumeSemicolon();
  	            }
  	            exportDeclaration = this.finalize(node, new Node.ExportNamedDeclaration(null, specifiers, source));
  	        }
  	        return exportDeclaration;
  	    };
  	    return Parser;
  	}());
  	exports.Parser = Parser;


  /***/ },
  /* 9 */
  /***/ function(module, exports) {
  	// Ensure the condition is true, otherwise throw an error.
  	// This is only to have a better contract semantic, i.e. another safety net
  	// to catch a logic error. The condition shall be fulfilled in normal case.
  	// Do NOT use this to enforce a certain condition on any user input.
  	Object.defineProperty(exports, "__esModule", { value: true });
  	function assert(condition, message) {
  	    /* istanbul ignore if */
  	    if (!condition) {
  	        throw new Error('ASSERT: ' + message);
  	    }
  	}
  	exports.assert = assert;


  /***/ },
  /* 10 */
  /***/ function(module, exports) {
  	/* tslint:disable:max-classes-per-file */
  	Object.defineProperty(exports, "__esModule", { value: true });
  	var ErrorHandler = (function () {
  	    function ErrorHandler() {
  	        this.errors = [];
  	        this.tolerant = false;
  	    }
  	    ErrorHandler.prototype.recordError = function (error) {
  	        this.errors.push(error);
  	    };
  	    ErrorHandler.prototype.tolerate = function (error) {
  	        if (this.tolerant) {
  	            this.recordError(error);
  	        }
  	        else {
  	            throw error;
  	        }
  	    };
  	    ErrorHandler.prototype.constructError = function (msg, column) {
  	        var error = new Error(msg);
  	        try {
  	            throw error;
  	        }
  	        catch (base) {
  	            /* istanbul ignore else */
  	            if (Object.create && Object.defineProperty) {
  	                error = Object.create(base);
  	                Object.defineProperty(error, 'column', { value: column });
  	            }
  	        }
  	        /* istanbul ignore next */
  	        return error;
  	    };
  	    ErrorHandler.prototype.createError = function (index, line, col, description) {
  	        var msg = 'Line ' + line + ': ' + description;
  	        var error = this.constructError(msg, col);
  	        error.index = index;
  	        error.lineNumber = line;
  	        error.description = description;
  	        return error;
  	    };
  	    ErrorHandler.prototype.throwError = function (index, line, col, description) {
  	        throw this.createError(index, line, col, description);
  	    };
  	    ErrorHandler.prototype.tolerateError = function (index, line, col, description) {
  	        var error = this.createError(index, line, col, description);
  	        if (this.tolerant) {
  	            this.recordError(error);
  	        }
  	        else {
  	            throw error;
  	        }
  	    };
  	    return ErrorHandler;
  	}());
  	exports.ErrorHandler = ErrorHandler;


  /***/ },
  /* 11 */
  /***/ function(module, exports) {
  	Object.defineProperty(exports, "__esModule", { value: true });
  	// Error messages should be identical to V8.
  	exports.Messages = {
  	    BadGetterArity: 'Getter must not have any formal parameters',
  	    BadSetterArity: 'Setter must have exactly one formal parameter',
  	    BadSetterRestParameter: 'Setter function argument must not be a rest parameter',
  	    ConstructorIsAsync: 'Class constructor may not be an async method',
  	    ConstructorSpecialMethod: 'Class constructor may not be an accessor',
  	    DeclarationMissingInitializer: 'Missing initializer in %0 declaration',
  	    DefaultRestParameter: 'Unexpected token =',
  	    DuplicateBinding: 'Duplicate binding %0',
  	    DuplicateConstructor: 'A class may only have one constructor',
  	    DuplicateProtoProperty: 'Duplicate __proto__ fields are not allowed in object literals',
  	    ForInOfLoopInitializer: '%0 loop variable declaration may not have an initializer',
  	    GeneratorInLegacyContext: 'Generator declarations are not allowed in legacy contexts',
  	    IllegalBreak: 'Illegal break statement',
  	    IllegalContinue: 'Illegal continue statement',
  	    IllegalExportDeclaration: 'Unexpected token',
  	    IllegalImportDeclaration: 'Unexpected token',
  	    IllegalLanguageModeDirective: 'Illegal \'use strict\' directive in function with non-simple parameter list',
  	    IllegalReturn: 'Illegal return statement',
  	    InvalidEscapedReservedWord: 'Keyword must not contain escaped characters',
  	    InvalidHexEscapeSequence: 'Invalid hexadecimal escape sequence',
  	    InvalidLHSInAssignment: 'Invalid left-hand side in assignment',
  	    InvalidLHSInForIn: 'Invalid left-hand side in for-in',
  	    InvalidLHSInForLoop: 'Invalid left-hand side in for-loop',
  	    InvalidModuleSpecifier: 'Unexpected token',
  	    InvalidRegExp: 'Invalid regular expression',
  	    LetInLexicalBinding: 'let is disallowed as a lexically bound name',
  	    MissingFromClause: 'Unexpected token',
  	    MultipleDefaultsInSwitch: 'More than one default clause in switch statement',
  	    NewlineAfterThrow: 'Illegal newline after throw',
  	    NoAsAfterImportNamespace: 'Unexpected token',
  	    NoCatchOrFinally: 'Missing catch or finally after try',
  	    ParameterAfterRestParameter: 'Rest parameter must be last formal parameter',
  	    Redeclaration: '%0 \'%1\' has already been declared',
  	    StaticPrototype: 'Classes may not have static property named prototype',
  	    StrictCatchVariable: 'Catch variable may not be eval or arguments in strict mode',
  	    StrictDelete: 'Delete of an unqualified identifier in strict mode.',
  	    StrictFunction: 'In strict mode code, functions can only be declared at top level or inside a block',
  	    StrictFunctionName: 'Function name may not be eval or arguments in strict mode',
  	    StrictLHSAssignment: 'Assignment to eval or arguments is not allowed in strict mode',
  	    StrictLHSPostfix: 'Postfix increment/decrement may not have eval or arguments operand in strict mode',
  	    StrictLHSPrefix: 'Prefix increment/decrement may not have eval or arguments operand in strict mode',
  	    StrictModeWith: 'Strict mode code may not include a with statement',
  	    StrictOctalLiteral: 'Octal literals are not allowed in strict mode.',
  	    StrictParamDupe: 'Strict mode function may not have duplicate parameter names',
  	    StrictParamName: 'Parameter name eval or arguments is not allowed in strict mode',
  	    StrictReservedWord: 'Use of future reserved word in strict mode',
  	    StrictVarName: 'Variable name may not be eval or arguments in strict mode',
  	    TemplateOctalLiteral: 'Octal literals are not allowed in template strings.',
  	    UnexpectedEOS: 'Unexpected end of input',
  	    UnexpectedIdentifier: 'Unexpected identifier',
  	    UnexpectedNumber: 'Unexpected number',
  	    UnexpectedReserved: 'Unexpected reserved word',
  	    UnexpectedString: 'Unexpected string',
  	    UnexpectedTemplate: 'Unexpected quasi %0',
  	    UnexpectedToken: 'Unexpected token %0',
  	    UnexpectedTokenIllegal: 'Unexpected token ILLEGAL',
  	    UnknownLabel: 'Undefined label \'%0\'',
  	    UnterminatedRegExp: 'Invalid regular expression: missing /'
  	};


  /***/ },
  /* 12 */
  /***/ function(module, exports, __webpack_require__) {
  	Object.defineProperty(exports, "__esModule", { value: true });
  	var assert_1 = __webpack_require__(9);
  	var character_1 = __webpack_require__(4);
  	var messages_1 = __webpack_require__(11);
  	function hexValue(ch) {
  	    return '0123456789abcdef'.indexOf(ch.toLowerCase());
  	}
  	function octalValue(ch) {
  	    return '01234567'.indexOf(ch);
  	}
  	var Scanner = (function () {
  	    function Scanner(code, handler) {
  	        this.source = code;
  	        this.errorHandler = handler;
  	        this.trackComment = false;
  	        this.length = code.length;
  	        this.index = 0;
  	        this.lineNumber = (code.length > 0) ? 1 : 0;
  	        this.lineStart = 0;
  	        this.curlyStack = [];
  	    }
  	    Scanner.prototype.saveState = function () {
  	        return {
  	            index: this.index,
  	            lineNumber: this.lineNumber,
  	            lineStart: this.lineStart
  	        };
  	    };
  	    Scanner.prototype.restoreState = function (state) {
  	        this.index = state.index;
  	        this.lineNumber = state.lineNumber;
  	        this.lineStart = state.lineStart;
  	    };
  	    Scanner.prototype.eof = function () {
  	        return this.index >= this.length;
  	    };
  	    Scanner.prototype.throwUnexpectedToken = function (message) {
  	        if (message === void 0) { message = messages_1.Messages.UnexpectedTokenIllegal; }
  	        return this.errorHandler.throwError(this.index, this.lineNumber, this.index - this.lineStart + 1, message);
  	    };
  	    Scanner.prototype.tolerateUnexpectedToken = function (message) {
  	        if (message === void 0) { message = messages_1.Messages.UnexpectedTokenIllegal; }
  	        this.errorHandler.tolerateError(this.index, this.lineNumber, this.index - this.lineStart + 1, message);
  	    };
  	    // https://tc39.github.io/ecma262/#sec-comments
  	    Scanner.prototype.skipSingleLineComment = function (offset) {
  	        var comments = [];
  	        var start, loc;
  	        if (this.trackComment) {
  	            comments = [];
  	            start = this.index - offset;
  	            loc = {
  	                start: {
  	                    line: this.lineNumber,
  	                    column: this.index - this.lineStart - offset
  	                },
  	                end: {}
  	            };
  	        }
  	        while (!this.eof()) {
  	            var ch = this.source.charCodeAt(this.index);
  	            ++this.index;
  	            if (character_1.Character.isLineTerminator(ch)) {
  	                if (this.trackComment) {
  	                    loc.end = {
  	                        line: this.lineNumber,
  	                        column: this.index - this.lineStart - 1
  	                    };
  	                    var entry = {
  	                        multiLine: false,
  	                        slice: [start + offset, this.index - 1],
  	                        range: [start, this.index - 1],
  	                        loc: loc
  	                    };
  	                    comments.push(entry);
  	                }
  	                if (ch === 13 && this.source.charCodeAt(this.index) === 10) {
  	                    ++this.index;
  	                }
  	                ++this.lineNumber;
  	                this.lineStart = this.index;
  	                return comments;
  	            }
  	        }
  	        if (this.trackComment) {
  	            loc.end = {
  	                line: this.lineNumber,
  	                column: this.index - this.lineStart
  	            };
  	            var entry = {
  	                multiLine: false,
  	                slice: [start + offset, this.index],
  	                range: [start, this.index],
  	                loc: loc
  	            };
  	            comments.push(entry);
  	        }
  	        return comments;
  	    };
  	    Scanner.prototype.skipMultiLineComment = function () {
  	        var comments = [];
  	        var start, loc;
  	        if (this.trackComment) {
  	            comments = [];
  	            start = this.index - 2;
  	            loc = {
  	                start: {
  	                    line: this.lineNumber,
  	                    column: this.index - this.lineStart - 2
  	                },
  	                end: {}
  	            };
  	        }
  	        while (!this.eof()) {
  	            var ch = this.source.charCodeAt(this.index);
  	            if (character_1.Character.isLineTerminator(ch)) {
  	                if (ch === 0x0D && this.source.charCodeAt(this.index + 1) === 0x0A) {
  	                    ++this.index;
  	                }
  	                ++this.lineNumber;
  	                ++this.index;
  	                this.lineStart = this.index;
  	            }
  	            else if (ch === 0x2A) {
  	                // Block comment ends with '*/'.
  	                if (this.source.charCodeAt(this.index + 1) === 0x2F) {
  	                    this.index += 2;
  	                    if (this.trackComment) {
  	                        loc.end = {
  	                            line: this.lineNumber,
  	                            column: this.index - this.lineStart
  	                        };
  	                        var entry = {
  	                            multiLine: true,
  	                            slice: [start + 2, this.index - 2],
  	                            range: [start, this.index],
  	                            loc: loc
  	                        };
  	                        comments.push(entry);
  	                    }
  	                    return comments;
  	                }
  	                ++this.index;
  	            }
  	            else {
  	                ++this.index;
  	            }
  	        }
  	        // Ran off the end of the file - the whole thing is a comment
  	        if (this.trackComment) {
  	            loc.end = {
  	                line: this.lineNumber,
  	                column: this.index - this.lineStart
  	            };
  	            var entry = {
  	                multiLine: true,
  	                slice: [start + 2, this.index],
  	                range: [start, this.index],
  	                loc: loc
  	            };
  	            comments.push(entry);
  	        }
  	        this.tolerateUnexpectedToken();
  	        return comments;
  	    };
  	    Scanner.prototype.scanComments = function () {
  	        var comments;
  	        if (this.trackComment) {
  	            comments = [];
  	        }
  	        var start = (this.index === 0);
  	        while (!this.eof()) {
  	            var ch = this.source.charCodeAt(this.index);
  	            if (character_1.Character.isWhiteSpace(ch)) {
  	                ++this.index;
  	            }
  	            else if (character_1.Character.isLineTerminator(ch)) {
  	                ++this.index;
  	                if (ch === 0x0D && this.source.charCodeAt(this.index) === 0x0A) {
  	                    ++this.index;
  	                }
  	                ++this.lineNumber;
  	                this.lineStart = this.index;
  	                start = true;
  	            }
  	            else if (ch === 0x2F) {
  	                ch = this.source.charCodeAt(this.index + 1);
  	                if (ch === 0x2F) {
  	                    this.index += 2;
  	                    var comment = this.skipSingleLineComment(2);
  	                    if (this.trackComment) {
  	                        comments = comments.concat(comment);
  	                    }
  	                    start = true;
  	                }
  	                else if (ch === 0x2A) {
  	                    this.index += 2;
  	                    var comment = this.skipMultiLineComment();
  	                    if (this.trackComment) {
  	                        comments = comments.concat(comment);
  	                    }
  	                }
  	                else {
  	                    break;
  	                }
  	            }
  	            else if (start && ch === 0x2D) {
  	                // U+003E is '>'
  	                if ((this.source.charCodeAt(this.index + 1) === 0x2D) && (this.source.charCodeAt(this.index + 2) === 0x3E)) {
  	                    // '-->' is a single-line comment
  	                    this.index += 3;
  	                    var comment = this.skipSingleLineComment(3);
  	                    if (this.trackComment) {
  	                        comments = comments.concat(comment);
  	                    }
  	                }
  	                else {
  	                    break;
  	                }
  	            }
  	            else if (ch === 0x3C) {
  	                if (this.source.slice(this.index + 1, this.index + 4) === '!--') {
  	                    this.index += 4; // `<!--`
  	                    var comment = this.skipSingleLineComment(4);
  	                    if (this.trackComment) {
  	                        comments = comments.concat(comment);
  	                    }
  	                }
  	                else {
  	                    break;
  	                }
  	            }
  	            else {
  	                break;
  	            }
  	        }
  	        return comments;
  	    };
  	    // https://tc39.github.io/ecma262/#sec-future-reserved-words
  	    Scanner.prototype.isFutureReservedWord = function (id) {
  	        switch (id) {
  	            case 'enum':
  	            case 'export':
  	            case 'import':
  	            case 'super':
  	                return true;
  	            default:
  	                return false;
  	        }
  	    };
  	    Scanner.prototype.isStrictModeReservedWord = function (id) {
  	        switch (id) {
  	            case 'implements':
  	            case 'interface':
  	            case 'package':
  	            case 'private':
  	            case 'protected':
  	            case 'public':
  	            case 'static':
  	            case 'yield':
  	            case 'let':
  	                return true;
  	            default:
  	                return false;
  	        }
  	    };
  	    Scanner.prototype.isRestrictedWord = function (id) {
  	        return id === 'eval' || id === 'arguments';
  	    };
  	    // https://tc39.github.io/ecma262/#sec-keywords
  	    Scanner.prototype.isKeyword = function (id) {
  	        switch (id.length) {
  	            case 2:
  	                return (id === 'if') || (id === 'in') || (id === 'do');
  	            case 3:
  	                return (id === 'var') || (id === 'for') || (id === 'new') ||
  	                    (id === 'try') || (id === 'let');
  	            case 4:
  	                return (id === 'this') || (id === 'else') || (id === 'case') ||
  	                    (id === 'void') || (id === 'with') || (id === 'enum');
  	            case 5:
  	                return (id === 'while') || (id === 'break') || (id === 'catch') ||
  	                    (id === 'throw') || (id === 'const') || (id === 'yield') ||
  	                    (id === 'class') || (id === 'super');
  	            case 6:
  	                return (id === 'return') || (id === 'typeof') || (id === 'delete') ||
  	                    (id === 'switch') || (id === 'export') || (id === 'import');
  	            case 7:
  	                return (id === 'default') || (id === 'finally') || (id === 'extends');
  	            case 8:
  	                return (id === 'function') || (id === 'continue') || (id === 'debugger');
  	            case 10:
  	                return (id === 'instanceof');
  	            default:
  	                return false;
  	        }
  	    };
  	    Scanner.prototype.codePointAt = function (i) {
  	        var cp = this.source.charCodeAt(i);
  	        if (cp >= 0xD800 && cp <= 0xDBFF) {
  	            var second = this.source.charCodeAt(i + 1);
  	            if (second >= 0xDC00 && second <= 0xDFFF) {
  	                var first = cp;
  	                cp = (first - 0xD800) * 0x400 + second - 0xDC00 + 0x10000;
  	            }
  	        }
  	        return cp;
  	    };
  	    Scanner.prototype.scanHexEscape = function (prefix) {
  	        var len = (prefix === 'u') ? 4 : 2;
  	        var code = 0;
  	        for (var i = 0; i < len; ++i) {
  	            if (!this.eof() && character_1.Character.isHexDigit(this.source.charCodeAt(this.index))) {
  	                code = code * 16 + hexValue(this.source[this.index++]);
  	            }
  	            else {
  	                return null;
  	            }
  	        }
  	        return String.fromCharCode(code);
  	    };
  	    Scanner.prototype.scanUnicodeCodePointEscape = function () {
  	        var ch = this.source[this.index];
  	        var code = 0;
  	        // At least, one hex digit is required.
  	        if (ch === '}') {
  	            this.throwUnexpectedToken();
  	        }
  	        while (!this.eof()) {
  	            ch = this.source[this.index++];
  	            if (!character_1.Character.isHexDigit(ch.charCodeAt(0))) {
  	                break;
  	            }
  	            code = code * 16 + hexValue(ch);
  	        }
  	        if (code > 0x10FFFF || ch !== '}') {
  	            this.throwUnexpectedToken();
  	        }
  	        return character_1.Character.fromCodePoint(code);
  	    };
  	    Scanner.prototype.getIdentifier = function () {
  	        var start = this.index++;
  	        while (!this.eof()) {
  	            var ch = this.source.charCodeAt(this.index);
  	            if (ch === 0x5C) {
  	                // Blackslash (U+005C) marks Unicode escape sequence.
  	                this.index = start;
  	                return this.getComplexIdentifier();
  	            }
  	            else if (ch >= 0xD800 && ch < 0xDFFF) {
  	                // Need to handle surrogate pairs.
  	                this.index = start;
  	                return this.getComplexIdentifier();
  	            }
  	            if (character_1.Character.isIdentifierPart(ch)) {
  	                ++this.index;
  	            }
  	            else {
  	                break;
  	            }
  	        }
  	        return this.source.slice(start, this.index);
  	    };
  	    Scanner.prototype.getComplexIdentifier = function () {
  	        var cp = this.codePointAt(this.index);
  	        var id = character_1.Character.fromCodePoint(cp);
  	        this.index += id.length;
  	        // '\u' (U+005C, U+0075) denotes an escaped character.
  	        var ch;
  	        if (cp === 0x5C) {
  	            if (this.source.charCodeAt(this.index) !== 0x75) {
  	                this.throwUnexpectedToken();
  	            }
  	            ++this.index;
  	            if (this.source[this.index] === '{') {
  	                ++this.index;
  	                ch = this.scanUnicodeCodePointEscape();
  	            }
  	            else {
  	                ch = this.scanHexEscape('u');
  	                if (ch === null || ch === '\\' || !character_1.Character.isIdentifierStart(ch.charCodeAt(0))) {
  	                    this.throwUnexpectedToken();
  	                }
  	            }
  	            id = ch;
  	        }
  	        while (!this.eof()) {
  	            cp = this.codePointAt(this.index);
  	            if (!character_1.Character.isIdentifierPart(cp)) {
  	                break;
  	            }
  	            ch = character_1.Character.fromCodePoint(cp);
  	            id += ch;
  	            this.index += ch.length;
  	            // '\u' (U+005C, U+0075) denotes an escaped character.
  	            if (cp === 0x5C) {
  	                id = id.substr(0, id.length - 1);
  	                if (this.source.charCodeAt(this.index) !== 0x75) {
  	                    this.throwUnexpectedToken();
  	                }
  	                ++this.index;
  	                if (this.source[this.index] === '{') {
  	                    ++this.index;
  	                    ch = this.scanUnicodeCodePointEscape();
  	                }
  	                else {
  	                    ch = this.scanHexEscape('u');
  	                    if (ch === null || ch === '\\' || !character_1.Character.isIdentifierPart(ch.charCodeAt(0))) {
  	                        this.throwUnexpectedToken();
  	                    }
  	                }
  	                id += ch;
  	            }
  	        }
  	        return id;
  	    };
  	    Scanner.prototype.octalToDecimal = function (ch) {
  	        // \0 is not octal escape sequence
  	        var octal = (ch !== '0');
  	        var code = octalValue(ch);
  	        if (!this.eof() && character_1.Character.isOctalDigit(this.source.charCodeAt(this.index))) {
  	            octal = true;
  	            code = code * 8 + octalValue(this.source[this.index++]);
  	            // 3 digits are only allowed when string starts
  	            // with 0, 1, 2, 3
  	            if ('0123'.indexOf(ch) >= 0 && !this.eof() && character_1.Character.isOctalDigit(this.source.charCodeAt(this.index))) {
  	                code = code * 8 + octalValue(this.source[this.index++]);
  	            }
  	        }
  	        return {
  	            code: code,
  	            octal: octal
  	        };
  	    };
  	    // https://tc39.github.io/ecma262/#sec-names-and-keywords
  	    Scanner.prototype.scanIdentifier = function () {
  	        var type;
  	        var start = this.index;
  	        // Backslash (U+005C) starts an escaped character.
  	        var id = (this.source.charCodeAt(start) === 0x5C) ? this.getComplexIdentifier() : this.getIdentifier();
  	        // There is no keyword or literal with only one character.
  	        // Thus, it must be an identifier.
  	        if (id.length === 1) {
  	            type = 3 /* Identifier */;
  	        }
  	        else if (this.isKeyword(id)) {
  	            type = 4 /* Keyword */;
  	        }
  	        else if (id === 'null') {
  	            type = 5 /* NullLiteral */;
  	        }
  	        else if (id === 'true' || id === 'false') {
  	            type = 1 /* BooleanLiteral */;
  	        }
  	        else {
  	            type = 3 /* Identifier */;
  	        }
  	        if (type !== 3 /* Identifier */ && (start + id.length !== this.index)) {
  	            var restore = this.index;
  	            this.index = start;
  	            this.tolerateUnexpectedToken(messages_1.Messages.InvalidEscapedReservedWord);
  	            this.index = restore;
  	        }
  	        return {
  	            type: type,
  	            value: id,
  	            lineNumber: this.lineNumber,
  	            lineStart: this.lineStart,
  	            start: start,
  	            end: this.index
  	        };
  	    };
  	    // https://tc39.github.io/ecma262/#sec-punctuators
  	    Scanner.prototype.scanPunctuator = function () {
  	        var start = this.index;
  	        // Check for most common single-character punctuators.
  	        var str = this.source[this.index];
  	        switch (str) {
  	            case '(':
  	            case '{':
  	                if (str === '{') {
  	                    this.curlyStack.push('{');
  	                }
  	                ++this.index;
  	                break;
  	            case '.':
  	                ++this.index;
  	                if (this.source[this.index] === '.' && this.source[this.index + 1] === '.') {
  	                    // Spread operator: ...
  	                    this.index += 2;
  	                    str = '...';
  	                }
  	                break;
  	            case '}':
  	                ++this.index;
  	                this.curlyStack.pop();
  	                break;
  	            case ')':
  	            case ';':
  	            case ',':
  	            case '[':
  	            case ']':
  	            case ':':
  	            case '?':
  	            case '~':
  	                ++this.index;
  	                break;
  	            default:
  	                // 4-character punctuator.
  	                str = this.source.substr(this.index, 4);
  	                if (str === '>>>=') {
  	                    this.index += 4;
  	                }
  	                else {
  	                    // 3-character punctuators.
  	                    str = str.substr(0, 3);
  	                    if (str === '===' || str === '!==' || str === '>>>' ||
  	                        str === '<<=' || str === '>>=' || str === '**=') {
  	                        this.index += 3;
  	                    }
  	                    else {
  	                        // 2-character punctuators.
  	                        str = str.substr(0, 2);
  	                        if (str === '&&' || str === '||' || str === '==' || str === '!=' ||
  	                            str === '+=' || str === '-=' || str === '*=' || str === '/=' ||
  	                            str === '++' || str === '--' || str === '<<' || str === '>>' ||
  	                            str === '&=' || str === '|=' || str === '^=' || str === '%=' ||
  	                            str === '<=' || str === '>=' || str === '=>' || str === '**') {
  	                            this.index += 2;
  	                        }
  	                        else {
  	                            // 1-character punctuators.
  	                            str = this.source[this.index];
  	                            if ('<>=!+-*%&|^/'.indexOf(str) >= 0) {
  	                                ++this.index;
  	                            }
  	                        }
  	                    }
  	                }
  	        }
  	        if (this.index === start) {
  	            this.throwUnexpectedToken();
  	        }
  	        return {
  	            type: 7 /* Punctuator */,
  	            value: str,
  	            lineNumber: this.lineNumber,
  	            lineStart: this.lineStart,
  	            start: start,
  	            end: this.index
  	        };
  	    };
  	    // https://tc39.github.io/ecma262/#sec-literals-numeric-literals
  	    Scanner.prototype.scanHexLiteral = function (start) {
  	        var num = '';
  	        while (!this.eof()) {
  	            if (!character_1.Character.isHexDigit(this.source.charCodeAt(this.index))) {
  	                break;
  	            }
  	            num += this.source[this.index++];
  	        }
  	        if (num.length === 0) {
  	            this.throwUnexpectedToken();
  	        }
  	        if (character_1.Character.isIdentifierStart(this.source.charCodeAt(this.index))) {
  	            this.throwUnexpectedToken();
  	        }
  	        return {
  	            type: 6 /* NumericLiteral */,
  	            value: parseInt('0x' + num, 16),
  	            lineNumber: this.lineNumber,
  	            lineStart: this.lineStart,
  	            start: start,
  	            end: this.index
  	        };
  	    };
  	    Scanner.prototype.scanBinaryLiteral = function (start) {
  	        var num = '';
  	        var ch;
  	        while (!this.eof()) {
  	            ch = this.source[this.index];
  	            if (ch !== '0' && ch !== '1') {
  	                break;
  	            }
  	            num += this.source[this.index++];
  	        }
  	        if (num.length === 0) {
  	            // only 0b or 0B
  	            this.throwUnexpectedToken();
  	        }
  	        if (!this.eof()) {
  	            ch = this.source.charCodeAt(this.index);
  	            /* istanbul ignore else */
  	            if (character_1.Character.isIdentifierStart(ch) || character_1.Character.isDecimalDigit(ch)) {
  	                this.throwUnexpectedToken();
  	            }
  	        }
  	        return {
  	            type: 6 /* NumericLiteral */,
  	            value: parseInt(num, 2),
  	            lineNumber: this.lineNumber,
  	            lineStart: this.lineStart,
  	            start: start,
  	            end: this.index
  	        };
  	    };
  	    Scanner.prototype.scanOctalLiteral = function (prefix, start) {
  	        var num = '';
  	        var octal = false;
  	        if (character_1.Character.isOctalDigit(prefix.charCodeAt(0))) {
  	            octal = true;
  	            num = '0' + this.source[this.index++];
  	        }
  	        else {
  	            ++this.index;
  	        }
  	        while (!this.eof()) {
  	            if (!character_1.Character.isOctalDigit(this.source.charCodeAt(this.index))) {
  	                break;
  	            }
  	            num += this.source[this.index++];
  	        }
  	        if (!octal && num.length === 0) {
  	            // only 0o or 0O
  	            this.throwUnexpectedToken();
  	        }
  	        if (character_1.Character.isIdentifierStart(this.source.charCodeAt(this.index)) || character_1.Character.isDecimalDigit(this.source.charCodeAt(this.index))) {
  	            this.throwUnexpectedToken();
  	        }
  	        return {
  	            type: 6 /* NumericLiteral */,
  	            value: parseInt(num, 8),
  	            octal: octal,
  	            lineNumber: this.lineNumber,
  	            lineStart: this.lineStart,
  	            start: start,
  	            end: this.index
  	        };
  	    };
  	    Scanner.prototype.isImplicitOctalLiteral = function () {
  	        // Implicit octal, unless there is a non-octal digit.
  	        // (Annex B.1.1 on Numeric Literals)
  	        for (var i = this.index + 1; i < this.length; ++i) {
  	            var ch = this.source[i];
  	            if (ch === '8' || ch === '9') {
  	                return false;
  	            }
  	            if (!character_1.Character.isOctalDigit(ch.charCodeAt(0))) {
  	                return true;
  	            }
  	        }
  	        return true;
  	    };
  	    Scanner.prototype.scanNumericLiteral = function () {
  	        var start = this.index;
  	        var ch = this.source[start];
  	        assert_1.assert(character_1.Character.isDecimalDigit(ch.charCodeAt(0)) || (ch === '.'), 'Numeric literal must start with a decimal digit or a decimal point');
  	        var num = '';
  	        if (ch !== '.') {
  	            num = this.source[this.index++];
  	            ch = this.source[this.index];
  	            // Hex number starts with '0x'.
  	            // Octal number starts with '0'.
  	            // Octal number in ES6 starts with '0o'.
  	            // Binary number in ES6 starts with '0b'.
  	            if (num === '0') {
  	                if (ch === 'x' || ch === 'X') {
  	                    ++this.index;
  	                    return this.scanHexLiteral(start);
  	                }
  	                if (ch === 'b' || ch === 'B') {
  	                    ++this.index;
  	                    return this.scanBinaryLiteral(start);
  	                }
  	                if (ch === 'o' || ch === 'O') {
  	                    return this.scanOctalLiteral(ch, start);
  	                }
  	                if (ch && character_1.Character.isOctalDigit(ch.charCodeAt(0))) {
  	                    if (this.isImplicitOctalLiteral()) {
  	                        return this.scanOctalLiteral(ch, start);
  	                    }
  	                }
  	            }
  	            while (character_1.Character.isDecimalDigit(this.source.charCodeAt(this.index))) {
  	                num += this.source[this.index++];
  	            }
  	            ch = this.source[this.index];
  	        }
  	        if (ch === '.') {
  	            num += this.source[this.index++];
  	            while (character_1.Character.isDecimalDigit(this.source.charCodeAt(this.index))) {
  	                num += this.source[this.index++];
  	            }
  	            ch = this.source[this.index];
  	        }
  	        if (ch === 'e' || ch === 'E') {
  	            num += this.source[this.index++];
  	            ch = this.source[this.index];
  	            if (ch === '+' || ch === '-') {
  	                num += this.source[this.index++];
  	            }
  	            if (character_1.Character.isDecimalDigit(this.source.charCodeAt(this.index))) {
  	                while (character_1.Character.isDecimalDigit(this.source.charCodeAt(this.index))) {
  	                    num += this.source[this.index++];
  	                }
  	            }
  	            else {
  	                this.throwUnexpectedToken();
  	            }
  	        }
  	        if (character_1.Character.isIdentifierStart(this.source.charCodeAt(this.index))) {
  	            this.throwUnexpectedToken();
  	        }
  	        return {
  	            type: 6 /* NumericLiteral */,
  	            value: parseFloat(num),
  	            lineNumber: this.lineNumber,
  	            lineStart: this.lineStart,
  	            start: start,
  	            end: this.index
  	        };
  	    };
  	    // https://tc39.github.io/ecma262/#sec-literals-string-literals
  	    Scanner.prototype.scanStringLiteral = function () {
  	        var start = this.index;
  	        var quote = this.source[start];
  	        assert_1.assert((quote === '\'' || quote === '"'), 'String literal must starts with a quote');
  	        ++this.index;
  	        var octal = false;
  	        var str = '';
  	        while (!this.eof()) {
  	            var ch = this.source[this.index++];
  	            if (ch === quote) {
  	                quote = '';
  	                break;
  	            }
  	            else if (ch === '\\') {
  	                ch = this.source[this.index++];
  	                if (!ch || !character_1.Character.isLineTerminator(ch.charCodeAt(0))) {
  	                    switch (ch) {
  	                        case 'u':
  	                            if (this.source[this.index] === '{') {
  	                                ++this.index;
  	                                str += this.scanUnicodeCodePointEscape();
  	                            }
  	                            else {
  	                                var unescaped_1 = this.scanHexEscape(ch);
  	                                if (unescaped_1 === null) {
  	                                    this.throwUnexpectedToken();
  	                                }
  	                                str += unescaped_1;
  	                            }
  	                            break;
  	                        case 'x':
  	                            var unescaped = this.scanHexEscape(ch);
  	                            if (unescaped === null) {
  	                                this.throwUnexpectedToken(messages_1.Messages.InvalidHexEscapeSequence);
  	                            }
  	                            str += unescaped;
  	                            break;
  	                        case 'n':
  	                            str += '\n';
  	                            break;
  	                        case 'r':
  	                            str += '\r';
  	                            break;
  	                        case 't':
  	                            str += '\t';
  	                            break;
  	                        case 'b':
  	                            str += '\b';
  	                            break;
  	                        case 'f':
  	                            str += '\f';
  	                            break;
  	                        case 'v':
  	                            str += '\x0B';
  	                            break;
  	                        case '8':
  	                        case '9':
  	                            str += ch;
  	                            this.tolerateUnexpectedToken();
  	                            break;
  	                        default:
  	                            if (ch && character_1.Character.isOctalDigit(ch.charCodeAt(0))) {
  	                                var octToDec = this.octalToDecimal(ch);
  	                                octal = octToDec.octal || octal;
  	                                str += String.fromCharCode(octToDec.code);
  	                            }
  	                            else {
  	                                str += ch;
  	                            }
  	                            break;
  	                    }
  	                }
  	                else {
  	                    ++this.lineNumber;
  	                    if (ch === '\r' && this.source[this.index] === '\n') {
  	                        ++this.index;
  	                    }
  	                    this.lineStart = this.index;
  	                }
  	            }
  	            else if (character_1.Character.isLineTerminator(ch.charCodeAt(0))) {
  	                break;
  	            }
  	            else {
  	                str += ch;
  	            }
  	        }
  	        if (quote !== '') {
  	            this.index = start;
  	            this.throwUnexpectedToken();
  	        }
  	        return {
  	            type: 8 /* StringLiteral */,
  	            value: str,
  	            octal: octal,
  	            lineNumber: this.lineNumber,
  	            lineStart: this.lineStart,
  	            start: start,
  	            end: this.index
  	        };
  	    };
  	    // https://tc39.github.io/ecma262/#sec-template-literal-lexical-components
  	    Scanner.prototype.scanTemplate = function () {
  	        var cooked = '';
  	        var terminated = false;
  	        var start = this.index;
  	        var head = (this.source[start] === '`');
  	        var tail = false;
  	        var rawOffset = 2;
  	        ++this.index;
  	        while (!this.eof()) {
  	            var ch = this.source[this.index++];
  	            if (ch === '`') {
  	                rawOffset = 1;
  	                tail = true;
  	                terminated = true;
  	                break;
  	            }
  	            else if (ch === '$') {
  	                if (this.source[this.index] === '{') {
  	                    this.curlyStack.push('${');
  	                    ++this.index;
  	                    terminated = true;
  	                    break;
  	                }
  	                cooked += ch;
  	            }
  	            else if (ch === '\\') {
  	                ch = this.source[this.index++];
  	                if (!character_1.Character.isLineTerminator(ch.charCodeAt(0))) {
  	                    switch (ch) {
  	                        case 'n':
  	                            cooked += '\n';
  	                            break;
  	                        case 'r':
  	                            cooked += '\r';
  	                            break;
  	                        case 't':
  	                            cooked += '\t';
  	                            break;
  	                        case 'u':
  	                            if (this.source[this.index] === '{') {
  	                                ++this.index;
  	                                cooked += this.scanUnicodeCodePointEscape();
  	                            }
  	                            else {
  	                                var restore = this.index;
  	                                var unescaped_2 = this.scanHexEscape(ch);
  	                                if (unescaped_2 !== null) {
  	                                    cooked += unescaped_2;
  	                                }
  	                                else {
  	                                    this.index = restore;
  	                                    cooked += ch;
  	                                }
  	                            }
  	                            break;
  	                        case 'x':
  	                            var unescaped = this.scanHexEscape(ch);
  	                            if (unescaped === null) {
  	                                this.throwUnexpectedToken(messages_1.Messages.InvalidHexEscapeSequence);
  	                            }
  	                            cooked += unescaped;
  	                            break;
  	                        case 'b':
  	                            cooked += '\b';
  	                            break;
  	                        case 'f':
  	                            cooked += '\f';
  	                            break;
  	                        case 'v':
  	                            cooked += '\v';
  	                            break;
  	                        default:
  	                            if (ch === '0') {
  	                                if (character_1.Character.isDecimalDigit(this.source.charCodeAt(this.index))) {
  	                                    // Illegal: \01 \02 and so on
  	                                    this.throwUnexpectedToken(messages_1.Messages.TemplateOctalLiteral);
  	                                }
  	                                cooked += '\0';
  	                            }
  	                            else if (character_1.Character.isOctalDigit(ch.charCodeAt(0))) {
  	                                // Illegal: \1 \2
  	                                this.throwUnexpectedToken(messages_1.Messages.TemplateOctalLiteral);
  	                            }
  	                            else {
  	                                cooked += ch;
  	                            }
  	                            break;
  	                    }
  	                }
  	                else {
  	                    ++this.lineNumber;
  	                    if (ch === '\r' && this.source[this.index] === '\n') {
  	                        ++this.index;
  	                    }
  	                    this.lineStart = this.index;
  	                }
  	            }
  	            else if (character_1.Character.isLineTerminator(ch.charCodeAt(0))) {
  	                ++this.lineNumber;
  	                if (ch === '\r' && this.source[this.index] === '\n') {
  	                    ++this.index;
  	                }
  	                this.lineStart = this.index;
  	                cooked += '\n';
  	            }
  	            else {
  	                cooked += ch;
  	            }
  	        }
  	        if (!terminated) {
  	            this.throwUnexpectedToken();
  	        }
  	        if (!head) {
  	            this.curlyStack.pop();
  	        }
  	        return {
  	            type: 10 /* Template */,
  	            value: this.source.slice(start + 1, this.index - rawOffset),
  	            cooked: cooked,
  	            head: head,
  	            tail: tail,
  	            lineNumber: this.lineNumber,
  	            lineStart: this.lineStart,
  	            start: start,
  	            end: this.index
  	        };
  	    };
  	    // https://tc39.github.io/ecma262/#sec-literals-regular-expression-literals
  	    Scanner.prototype.testRegExp = function (pattern, flags) {
  	        // The BMP character to use as a replacement for astral symbols when
  	        // translating an ES6 "u"-flagged pattern to an ES5-compatible
  	        // approximation.
  	        // Note: replacing with '\uFFFF' enables false positives in unlikely
  	        // scenarios. For example, `[\u{1044f}-\u{10440}]` is an invalid
  	        // pattern that would not be detected by this substitution.
  	        var astralSubstitute = '\uFFFF';
  	        var tmp = pattern;
  	        var self = this;
  	        if (flags.indexOf('u') >= 0) {
  	            tmp = tmp
  	                .replace(/\\u\{([0-9a-fA-F]+)\}|\\u([a-fA-F0-9]{4})/g, function ($0, $1, $2) {
  	                var codePoint = parseInt($1 || $2, 16);
  	                if (codePoint > 0x10FFFF) {
  	                    self.throwUnexpectedToken(messages_1.Messages.InvalidRegExp);
  	                }
  	                if (codePoint <= 0xFFFF) {
  	                    return String.fromCharCode(codePoint);
  	                }
  	                return astralSubstitute;
  	            })
  	                .replace(/[\uD800-\uDBFF][\uDC00-\uDFFF]/g, astralSubstitute);
  	        }
  	        // First, detect invalid regular expressions.
  	        try {
  	        }
  	        catch (e) {
  	            this.throwUnexpectedToken(messages_1.Messages.InvalidRegExp);
  	        }
  	        // Return a regular expression object for this pattern-flag pair, or
  	        // `null` in case the current environment doesn't support the flags it
  	        // uses.
  	        try {
  	            return new RegExp(pattern, flags);
  	        }
  	        catch (exception) {
  	            /* istanbul ignore next */
  	            return null;
  	        }
  	    };
  	    Scanner.prototype.scanRegExpBody = function () {
  	        var ch = this.source[this.index];
  	        assert_1.assert(ch === '/', 'Regular expression literal must start with a slash');
  	        var str = this.source[this.index++];
  	        var classMarker = false;
  	        var terminated = false;
  	        while (!this.eof()) {
  	            ch = this.source[this.index++];
  	            str += ch;
  	            if (ch === '\\') {
  	                ch = this.source[this.index++];
  	                // https://tc39.github.io/ecma262/#sec-literals-regular-expression-literals
  	                if (character_1.Character.isLineTerminator(ch.charCodeAt(0))) {
  	                    this.throwUnexpectedToken(messages_1.Messages.UnterminatedRegExp);
  	                }
  	                str += ch;
  	            }
  	            else if (character_1.Character.isLineTerminator(ch.charCodeAt(0))) {
  	                this.throwUnexpectedToken(messages_1.Messages.UnterminatedRegExp);
  	            }
  	            else if (classMarker) {
  	                if (ch === ']') {
  	                    classMarker = false;
  	                }
  	            }
  	            else {
  	                if (ch === '/') {
  	                    terminated = true;
  	                    break;
  	                }
  	                else if (ch === '[') {
  	                    classMarker = true;
  	                }
  	            }
  	        }
  	        if (!terminated) {
  	            this.throwUnexpectedToken(messages_1.Messages.UnterminatedRegExp);
  	        }
  	        // Exclude leading and trailing slash.
  	        return str.substr(1, str.length - 2);
  	    };
  	    Scanner.prototype.scanRegExpFlags = function () {
  	        var str = '';
  	        var flags = '';
  	        while (!this.eof()) {
  	            var ch = this.source[this.index];
  	            if (!character_1.Character.isIdentifierPart(ch.charCodeAt(0))) {
  	                break;
  	            }
  	            ++this.index;
  	            if (ch === '\\' && !this.eof()) {
  	                ch = this.source[this.index];
  	                if (ch === 'u') {
  	                    ++this.index;
  	                    var restore = this.index;
  	                    var char = this.scanHexEscape('u');
  	                    if (char !== null) {
  	                        flags += char;
  	                        for (str += '\\u'; restore < this.index; ++restore) {
  	                            str += this.source[restore];
  	                        }
  	                    }
  	                    else {
  	                        this.index = restore;
  	                        flags += 'u';
  	                        str += '\\u';
  	                    }
  	                    this.tolerateUnexpectedToken();
  	                }
  	                else {
  	                    str += '\\';
  	                    this.tolerateUnexpectedToken();
  	                }
  	            }
  	            else {
  	                flags += ch;
  	                str += ch;
  	            }
  	        }
  	        return flags;
  	    };
  	    Scanner.prototype.scanRegExp = function () {
  	        var start = this.index;
  	        var pattern = this.scanRegExpBody();
  	        var flags = this.scanRegExpFlags();
  	        var value = this.testRegExp(pattern, flags);
  	        return {
  	            type: 9 /* RegularExpression */,
  	            value: '',
  	            pattern: pattern,
  	            flags: flags,
  	            regex: value,
  	            lineNumber: this.lineNumber,
  	            lineStart: this.lineStart,
  	            start: start,
  	            end: this.index
  	        };
  	    };
  	    Scanner.prototype.lex = function () {
  	        if (this.eof()) {
  	            return {
  	                type: 2 /* EOF */,
  	                value: '',
  	                lineNumber: this.lineNumber,
  	                lineStart: this.lineStart,
  	                start: this.index,
  	                end: this.index
  	            };
  	        }
  	        var cp = this.source.charCodeAt(this.index);
  	        if (character_1.Character.isIdentifierStart(cp)) {
  	            return this.scanIdentifier();
  	        }
  	        // Very common: ( and ) and ;
  	        if (cp === 0x28 || cp === 0x29 || cp === 0x3B) {
  	            return this.scanPunctuator();
  	        }
  	        // String literal starts with single quote (U+0027) or double quote (U+0022).
  	        if (cp === 0x27 || cp === 0x22) {
  	            return this.scanStringLiteral();
  	        }
  	        // Dot (.) U+002E can also start a floating-point number, hence the need
  	        // to check the next character.
  	        if (cp === 0x2E) {
  	            if (character_1.Character.isDecimalDigit(this.source.charCodeAt(this.index + 1))) {
  	                return this.scanNumericLiteral();
  	            }
  	            return this.scanPunctuator();
  	        }
  	        if (character_1.Character.isDecimalDigit(cp)) {
  	            return this.scanNumericLiteral();
  	        }
  	        // Template literals start with ` (U+0060) for template head
  	        // or } (U+007D) for template middle or template tail.
  	        if (cp === 0x60 || (cp === 0x7D && this.curlyStack[this.curlyStack.length - 1] === '${')) {
  	            return this.scanTemplate();
  	        }
  	        // Possible identifier start in a surrogate pair.
  	        if (cp >= 0xD800 && cp < 0xDFFF) {
  	            if (character_1.Character.isIdentifierStart(this.codePointAt(this.index))) {
  	                return this.scanIdentifier();
  	            }
  	        }
  	        return this.scanPunctuator();
  	    };
  	    return Scanner;
  	}());
  	exports.Scanner = Scanner;


  /***/ },
  /* 13 */
  /***/ function(module, exports) {
  	Object.defineProperty(exports, "__esModule", { value: true });
  	exports.TokenName = {};
  	exports.TokenName[1 /* BooleanLiteral */] = 'Boolean';
  	exports.TokenName[2 /* EOF */] = '<end>';
  	exports.TokenName[3 /* Identifier */] = 'Identifier';
  	exports.TokenName[4 /* Keyword */] = 'Keyword';
  	exports.TokenName[5 /* NullLiteral */] = 'Null';
  	exports.TokenName[6 /* NumericLiteral */] = 'Numeric';
  	exports.TokenName[7 /* Punctuator */] = 'Punctuator';
  	exports.TokenName[8 /* StringLiteral */] = 'String';
  	exports.TokenName[9 /* RegularExpression */] = 'RegularExpression';
  	exports.TokenName[10 /* Template */] = 'Template';


  /***/ },
  /* 14 */
  /***/ function(module, exports) {
  	// Generated by generate-xhtml-entities.js. DO NOT MODIFY!
  	Object.defineProperty(exports, "__esModule", { value: true });
  	exports.XHTMLEntities = {
  	    quot: '\u0022',
  	    amp: '\u0026',
  	    apos: '\u0027',
  	    gt: '\u003E',
  	    nbsp: '\u00A0',
  	    iexcl: '\u00A1',
  	    cent: '\u00A2',
  	    pound: '\u00A3',
  	    curren: '\u00A4',
  	    yen: '\u00A5',
  	    brvbar: '\u00A6',
  	    sect: '\u00A7',
  	    uml: '\u00A8',
  	    copy: '\u00A9',
  	    ordf: '\u00AA',
  	    laquo: '\u00AB',
  	    not: '\u00AC',
  	    shy: '\u00AD',
  	    reg: '\u00AE',
  	    macr: '\u00AF',
  	    deg: '\u00B0',
  	    plusmn: '\u00B1',
  	    sup2: '\u00B2',
  	    sup3: '\u00B3',
  	    acute: '\u00B4',
  	    micro: '\u00B5',
  	    para: '\u00B6',
  	    middot: '\u00B7',
  	    cedil: '\u00B8',
  	    sup1: '\u00B9',
  	    ordm: '\u00BA',
  	    raquo: '\u00BB',
  	    frac14: '\u00BC',
  	    frac12: '\u00BD',
  	    frac34: '\u00BE',
  	    iquest: '\u00BF',
  	    Agrave: '\u00C0',
  	    Aacute: '\u00C1',
  	    Acirc: '\u00C2',
  	    Atilde: '\u00C3',
  	    Auml: '\u00C4',
  	    Aring: '\u00C5',
  	    AElig: '\u00C6',
  	    Ccedil: '\u00C7',
  	    Egrave: '\u00C8',
  	    Eacute: '\u00C9',
  	    Ecirc: '\u00CA',
  	    Euml: '\u00CB',
  	    Igrave: '\u00CC',
  	    Iacute: '\u00CD',
  	    Icirc: '\u00CE',
  	    Iuml: '\u00CF',
  	    ETH: '\u00D0',
  	    Ntilde: '\u00D1',
  	    Ograve: '\u00D2',
  	    Oacute: '\u00D3',
  	    Ocirc: '\u00D4',
  	    Otilde: '\u00D5',
  	    Ouml: '\u00D6',
  	    times: '\u00D7',
  	    Oslash: '\u00D8',
  	    Ugrave: '\u00D9',
  	    Uacute: '\u00DA',
  	    Ucirc: '\u00DB',
  	    Uuml: '\u00DC',
  	    Yacute: '\u00DD',
  	    THORN: '\u00DE',
  	    szlig: '\u00DF',
  	    agrave: '\u00E0',
  	    aacute: '\u00E1',
  	    acirc: '\u00E2',
  	    atilde: '\u00E3',
  	    auml: '\u00E4',
  	    aring: '\u00E5',
  	    aelig: '\u00E6',
  	    ccedil: '\u00E7',
  	    egrave: '\u00E8',
  	    eacute: '\u00E9',
  	    ecirc: '\u00EA',
  	    euml: '\u00EB',
  	    igrave: '\u00EC',
  	    iacute: '\u00ED',
  	    icirc: '\u00EE',
  	    iuml: '\u00EF',
  	    eth: '\u00F0',
  	    ntilde: '\u00F1',
  	    ograve: '\u00F2',
  	    oacute: '\u00F3',
  	    ocirc: '\u00F4',
  	    otilde: '\u00F5',
  	    ouml: '\u00F6',
  	    divide: '\u00F7',
  	    oslash: '\u00F8',
  	    ugrave: '\u00F9',
  	    uacute: '\u00FA',
  	    ucirc: '\u00FB',
  	    uuml: '\u00FC',
  	    yacute: '\u00FD',
  	    thorn: '\u00FE',
  	    yuml: '\u00FF',
  	    OElig: '\u0152',
  	    oelig: '\u0153',
  	    Scaron: '\u0160',
  	    scaron: '\u0161',
  	    Yuml: '\u0178',
  	    fnof: '\u0192',
  	    circ: '\u02C6',
  	    tilde: '\u02DC',
  	    Alpha: '\u0391',
  	    Beta: '\u0392',
  	    Gamma: '\u0393',
  	    Delta: '\u0394',
  	    Epsilon: '\u0395',
  	    Zeta: '\u0396',
  	    Eta: '\u0397',
  	    Theta: '\u0398',
  	    Iota: '\u0399',
  	    Kappa: '\u039A',
  	    Lambda: '\u039B',
  	    Mu: '\u039C',
  	    Nu: '\u039D',
  	    Xi: '\u039E',
  	    Omicron: '\u039F',
  	    Pi: '\u03A0',
  	    Rho: '\u03A1',
  	    Sigma: '\u03A3',
  	    Tau: '\u03A4',
  	    Upsilon: '\u03A5',
  	    Phi: '\u03A6',
  	    Chi: '\u03A7',
  	    Psi: '\u03A8',
  	    Omega: '\u03A9',
  	    alpha: '\u03B1',
  	    beta: '\u03B2',
  	    gamma: '\u03B3',
  	    delta: '\u03B4',
  	    epsilon: '\u03B5',
  	    zeta: '\u03B6',
  	    eta: '\u03B7',
  	    theta: '\u03B8',
  	    iota: '\u03B9',
  	    kappa: '\u03BA',
  	    lambda: '\u03BB',
  	    mu: '\u03BC',
  	    nu: '\u03BD',
  	    xi: '\u03BE',
  	    omicron: '\u03BF',
  	    pi: '\u03C0',
  	    rho: '\u03C1',
  	    sigmaf: '\u03C2',
  	    sigma: '\u03C3',
  	    tau: '\u03C4',
  	    upsilon: '\u03C5',
  	    phi: '\u03C6',
  	    chi: '\u03C7',
  	    psi: '\u03C8',
  	    omega: '\u03C9',
  	    thetasym: '\u03D1',
  	    upsih: '\u03D2',
  	    piv: '\u03D6',
  	    ensp: '\u2002',
  	    emsp: '\u2003',
  	    thinsp: '\u2009',
  	    zwnj: '\u200C',
  	    zwj: '\u200D',
  	    lrm: '\u200E',
  	    rlm: '\u200F',
  	    ndash: '\u2013',
  	    mdash: '\u2014',
  	    lsquo: '\u2018',
  	    rsquo: '\u2019',
  	    sbquo: '\u201A',
  	    ldquo: '\u201C',
  	    rdquo: '\u201D',
  	    bdquo: '\u201E',
  	    dagger: '\u2020',
  	    Dagger: '\u2021',
  	    bull: '\u2022',
  	    hellip: '\u2026',
  	    permil: '\u2030',
  	    prime: '\u2032',
  	    Prime: '\u2033',
  	    lsaquo: '\u2039',
  	    rsaquo: '\u203A',
  	    oline: '\u203E',
  	    frasl: '\u2044',
  	    euro: '\u20AC',
  	    image: '\u2111',
  	    weierp: '\u2118',
  	    real: '\u211C',
  	    trade: '\u2122',
  	    alefsym: '\u2135',
  	    larr: '\u2190',
  	    uarr: '\u2191',
  	    rarr: '\u2192',
  	    darr: '\u2193',
  	    harr: '\u2194',
  	    crarr: '\u21B5',
  	    lArr: '\u21D0',
  	    uArr: '\u21D1',
  	    rArr: '\u21D2',
  	    dArr: '\u21D3',
  	    hArr: '\u21D4',
  	    forall: '\u2200',
  	    part: '\u2202',
  	    exist: '\u2203',
  	    empty: '\u2205',
  	    nabla: '\u2207',
  	    isin: '\u2208',
  	    notin: '\u2209',
  	    ni: '\u220B',
  	    prod: '\u220F',
  	    sum: '\u2211',
  	    minus: '\u2212',
  	    lowast: '\u2217',
  	    radic: '\u221A',
  	    prop: '\u221D',
  	    infin: '\u221E',
  	    ang: '\u2220',
  	    and: '\u2227',
  	    or: '\u2228',
  	    cap: '\u2229',
  	    cup: '\u222A',
  	    int: '\u222B',
  	    there4: '\u2234',
  	    sim: '\u223C',
  	    cong: '\u2245',
  	    asymp: '\u2248',
  	    ne: '\u2260',
  	    equiv: '\u2261',
  	    le: '\u2264',
  	    ge: '\u2265',
  	    sub: '\u2282',
  	    sup: '\u2283',
  	    nsub: '\u2284',
  	    sube: '\u2286',
  	    supe: '\u2287',
  	    oplus: '\u2295',
  	    otimes: '\u2297',
  	    perp: '\u22A5',
  	    sdot: '\u22C5',
  	    lceil: '\u2308',
  	    rceil: '\u2309',
  	    lfloor: '\u230A',
  	    rfloor: '\u230B',
  	    loz: '\u25CA',
  	    spades: '\u2660',
  	    clubs: '\u2663',
  	    hearts: '\u2665',
  	    diams: '\u2666',
  	    lang: '\u27E8',
  	    rang: '\u27E9'
  	};


  /***/ },
  /* 15 */
  /***/ function(module, exports, __webpack_require__) {
  	Object.defineProperty(exports, "__esModule", { value: true });
  	var error_handler_1 = __webpack_require__(10);
  	var scanner_1 = __webpack_require__(12);
  	var token_1 = __webpack_require__(13);
  	var Reader = (function () {
  	    function Reader() {
  	        this.values = [];
  	        this.curly = this.paren = -1;
  	    }
  	    // A function following one of those tokens is an expression.
  	    Reader.prototype.beforeFunctionExpression = function (t) {
  	        return ['(', '{', '[', 'in', 'typeof', 'instanceof', 'new',
  	            'return', 'case', 'delete', 'throw', 'void',
  	            // assignment operators
  	            '=', '+=', '-=', '*=', '**=', '/=', '%=', '<<=', '>>=', '>>>=',
  	            '&=', '|=', '^=', ',',
  	            // binary/unary operators
  	            '+', '-', '*', '**', '/', '%', '++', '--', '<<', '>>', '>>>', '&',
  	            '|', '^', '!', '~', '&&', '||', '?', ':', '===', '==', '>=',
  	            '<=', '<', '>', '!=', '!=='].indexOf(t) >= 0;
  	    };
  	    // Determine if forward slash (/) is an operator or part of a regular expression
  	    // https://github.com/mozilla/sweet.js/wiki/design
  	    Reader.prototype.isRegexStart = function () {
  	        var previous = this.values[this.values.length - 1];
  	        var regex = (previous !== null);
  	        switch (previous) {
  	            case 'this':
  	            case ']':
  	                regex = false;
  	                break;
  	            case ')':
  	                var keyword = this.values[this.paren - 1];
  	                regex = (keyword === 'if' || keyword === 'while' || keyword === 'for' || keyword === 'with');
  	                break;
  	            case '}':
  	                // Dividing a function by anything makes little sense,
  	                // but we have to check for that.
  	                regex = false;
  	                if (this.values[this.curly - 3] === 'function') {
  	                    // Anonymous function, e.g. function(){} /42
  	                    var check = this.values[this.curly - 4];
  	                    regex = check ? !this.beforeFunctionExpression(check) : false;
  	                }
  	                else if (this.values[this.curly - 4] === 'function') {
  	                    // Named function, e.g. function f(){} /42/
  	                    var check = this.values[this.curly - 5];
  	                    regex = check ? !this.beforeFunctionExpression(check) : true;
  	                }
  	                break;
  	            default:
  	                break;
  	        }
  	        return regex;
  	    };
  	    Reader.prototype.push = function (token) {
  	        if (token.type === 7 /* Punctuator */ || token.type === 4 /* Keyword */) {
  	            if (token.value === '{') {
  	                this.curly = this.values.length;
  	            }
  	            else if (token.value === '(') {
  	                this.paren = this.values.length;
  	            }
  	            this.values.push(token.value);
  	        }
  	        else {
  	            this.values.push(null);
  	        }
  	    };
  	    return Reader;
  	}());
  	var Tokenizer = (function () {
  	    function Tokenizer(code, config) {
  	        this.errorHandler = new error_handler_1.ErrorHandler();
  	        this.errorHandler.tolerant = config ? (typeof config.tolerant === 'boolean' && config.tolerant) : false;
  	        this.scanner = new scanner_1.Scanner(code, this.errorHandler);
  	        this.scanner.trackComment = config ? (typeof config.comment === 'boolean' && config.comment) : false;
  	        this.trackRange = config ? (typeof config.range === 'boolean' && config.range) : false;
  	        this.trackLoc = config ? (typeof config.loc === 'boolean' && config.loc) : false;
  	        this.buffer = [];
  	        this.reader = new Reader();
  	    }
  	    Tokenizer.prototype.errors = function () {
  	        return this.errorHandler.errors;
  	    };
  	    Tokenizer.prototype.getNextToken = function () {
  	        if (this.buffer.length === 0) {
  	            var comments = this.scanner.scanComments();
  	            if (this.scanner.trackComment) {
  	                for (var i = 0; i < comments.length; ++i) {
  	                    var e = comments[i];
  	                    var value = this.scanner.source.slice(e.slice[0], e.slice[1]);
  	                    var comment = {
  	                        type: e.multiLine ? 'BlockComment' : 'LineComment',
  	                        value: value
  	                    };
  	                    if (this.trackRange) {
  	                        comment.range = e.range;
  	                    }
  	                    if (this.trackLoc) {
  	                        comment.loc = e.loc;
  	                    }
  	                    this.buffer.push(comment);
  	                }
  	            }
  	            if (!this.scanner.eof()) {
  	                var loc = void 0;
  	                if (this.trackLoc) {
  	                    loc = {
  	                        start: {
  	                            line: this.scanner.lineNumber,
  	                            column: this.scanner.index - this.scanner.lineStart
  	                        },
  	                        end: {}
  	                    };
  	                }
  	                var startRegex = (this.scanner.source[this.scanner.index] === '/') && this.reader.isRegexStart();
  	                var token = startRegex ? this.scanner.scanRegExp() : this.scanner.lex();
  	                this.reader.push(token);
  	                var entry = {
  	                    type: token_1.TokenName[token.type],
  	                    value: this.scanner.source.slice(token.start, token.end)
  	                };
  	                if (this.trackRange) {
  	                    entry.range = [token.start, token.end];
  	                }
  	                if (this.trackLoc) {
  	                    loc.end = {
  	                        line: this.scanner.lineNumber,
  	                        column: this.scanner.index - this.scanner.lineStart
  	                    };
  	                    entry.loc = loc;
  	                }
  	                if (token.type === 9 /* RegularExpression */) {
  	                    var pattern = token.pattern;
  	                    var flags = token.flags;
  	                    entry.regex = { pattern: pattern, flags: flags };
  	                }
  	                this.buffer.push(entry);
  	            }
  	        }
  	        return this.buffer.shift();
  	    };
  	    return Tokenizer;
  	}());
  	exports.Tokenizer = Tokenizer;


  /***/ }
  /******/ ])
  });
  });

  unwrapExports(esprima$1);

  // This module is suitable for passing as options.parser when calling
  // recast.parse to process ECMAScript code with Esprima:
  //
  //   const ast = recast.parse(source, {
  //     parser: require("recast/parsers/esprima")
  //   });
  //
  const getOption = util_1$1.getOption;

  var parse = function (source, options) {
    const comments = [];
    const ast = esprima$1.parse(source, {
      loc: true,
      locations: true,
      comment: true,
      onComment: comments,
      range: getOption(options, "range", false),
      tolerant: getOption(options, "tolerant", true),
      tokens: true
    });

    if (! Array.isArray(ast.comments)) {
      ast.comments = comments;
    }

    return ast;
  };

  var esprima$3 = {
  	parse: parse
  };

  var defaults = {
    // If you want to use a different branch of esprima, or any other module
    // that supports a .parse function, pass that module object to
    // recast.parse as options.parser (legacy synonym: options.esprima).
    parser: esprima$3,

    // Number of spaces the pretty-printer should use per tab for
    // indentation. If you do not pass this option explicitly, it will be
    // (quite reliably!) inferred from the original code.
    tabWidth: 4,

    // If you really want the pretty-printer to use tabs instead of spaces,
    // make this option true.
    useTabs: false,

    // The reprinting code leaves leading whitespace untouched unless it has
    // to reindent a line, or you pass false for this option.
    reuseWhitespace: true,

    // Override this option to use a different line terminator, e.g. \r\n.
    lineTerminator: require$$1.EOL || "\n",

    // Some of the pretty-printer code (such as that for printing function
    // parameter lists) makes a valiant attempt to prevent really long
    // lines. You can adjust the limit by changing this option; however,
    // there is no guarantee that line length will fit inside this limit.
    wrapColumn: 74, // Aspirational for now.

    // Pass a string as options.sourceFileName to recast.parse to tell the
    // reprinter to keep track of reused code so that it can construct a
    // source map automatically.
    sourceFileName: null,

    // Pass a string as options.sourceMapName to recast.print, and (provided
    // you passed options.sourceFileName earlier) the PrintResult of
    // recast.print will have a .map property for the generated source map.
    sourceMapName: null,

    // If provided, this option will be passed along to the source map
    // generator as a root directory for relative source file paths.
    sourceRoot: null,

    // If you provide a source map that was generated from a previous call
    // to recast.print as options.inputSourceMap, the old source map will be
    // composed with the new source map.
    inputSourceMap: null,

    // If you want esprima to generate .range information (recast only uses
    // .loc internally), pass true for this option.
    range: false,

    // If you want esprima not to throw exceptions when it encounters
    // non-fatal errors, keep this option true.
    tolerant: true,

    // If you want to override the quotes used in string literals, specify
    // either "single", "double", or "auto" here ("auto" will select the one
    // which results in the shorter literal) Otherwise, use double quotes.
    quote: null,

    // Controls the printing of trailing commas in object literals, array
    // expressions and function parameters.
    //
    // This option could either be:
    // * Boolean - enable/disable in all contexts (objects, arrays and function params).
    // * Object - enable/disable per context.
    //
    // Example:
    // trailingComma: {
    //   objects: true,
    //   arrays: true,
    //   parameters: false,
    // }
    trailingComma: false,

    // Controls the printing of spaces inside array brackets.
    // See: http://eslint.org/docs/rules/array-bracket-spacing
    arrayBracketSpacing: false,

    // Controls the printing of spaces inside object literals,
    // destructuring assignments, and import/export specifiers.
    // See: http://eslint.org/docs/rules/object-curly-spacing
    objectCurlySpacing: true,

    // If you want parenthesis to wrap single-argument arrow function
    // parameter lists, pass true for this option.
    arrowParensAlways: false,

    // There are 2 supported syntaxes (`,` and `;`) in Flow Object Types;
    // The use of commas is in line with the more popular style and matches
    // how objects are defined in JS, making it a bit more natural to write.
    flowObjectCommas: true,

    // Whether to return an array of .tokens on the root AST node.
    tokens: true

  }, hasOwn$4 = defaults.hasOwnProperty;

  // Copy options and fill in default values.
  var normalize = function(options) {
    options = options || defaults;

    function get(key) {
      return hasOwn$4.call(options, key)
        ? options[key]
        : defaults[key];
    }

    return {
      tabWidth: +get("tabWidth"),
      useTabs: !!get("useTabs"),
      reuseWhitespace: !!get("reuseWhitespace"),
      lineTerminator: get("lineTerminator"),
      wrapColumn: Math.max(get("wrapColumn"), 0),
      sourceFileName: get("sourceFileName"),
      sourceMapName: get("sourceMapName"),
      sourceRoot: get("sourceRoot"),
      inputSourceMap: get("inputSourceMap"),
      parser: get("esprima") || get("parser"),
      range: get("range"),
      tolerant: get("tolerant"),
      quote: get("quote"),
      trailingComma: get("trailingComma"),
      arrayBracketSpacing: get("arrayBracketSpacing"),
      objectCurlySpacing: get("objectCurlySpacing"),
      arrowParensAlways: get("arrowParensAlways"),
      flowObjectCommas: get("flowObjectCommas"),
      tokens: !!get("tokens")
    };
  };

  var options = {
  	normalize: normalize
  };

  var originalObject = Object;
  var originalDefProp = Object.defineProperty;
  var originalCreate = Object.create;

  function defProp(obj, name, value) {
    if (originalDefProp) try {
      originalDefProp.call(originalObject, obj, name, { value: value });
    } catch (definePropertyIsBrokenInIE8) {
      obj[name] = value;
    } else {
      obj[name] = value;
    }
  }

  // For functions that will be invoked using .call or .apply, we need to
  // define those methods on the function objects themselves, rather than
  // inheriting them from Function.prototype, so that a malicious or clumsy
  // third party cannot interfere with the functionality of this module by
  // redefining Function.prototype.call or .apply.
  function makeSafeToCall(fun) {
    if (fun) {
      defProp(fun, "call", fun.call);
      defProp(fun, "apply", fun.apply);
    }
    return fun;
  }

  makeSafeToCall(originalDefProp);
  makeSafeToCall(originalCreate);

  var hasOwn$5 = makeSafeToCall(Object.prototype.hasOwnProperty);
  var numToStr = makeSafeToCall(Number.prototype.toString);
  var strSlice = makeSafeToCall(String.prototype.slice);

  var cloner = function(){};
  function create$7(prototype) {
    if (originalCreate) {
      return originalCreate.call(originalObject, prototype);
    }
    cloner.prototype = prototype || null;
    return new cloner;
  }

  var rand = Math.random;
  var uniqueKeys = create$7(null);

  function makeUniqueKey() {
    // Collisions are highly unlikely, but this module is in the business of
    // making guarantees rather than safe bets.
    do var uniqueKey = internString(strSlice.call(numToStr.call(rand(), 36), 2));
    while (hasOwn$5.call(uniqueKeys, uniqueKey));
    return uniqueKeys[uniqueKey] = uniqueKey;
  }

  function internString(str) {
    var obj = {};
    obj[str] = true;
    return Object.keys(obj)[0];
  }

  // External users might find this function useful, but it is not necessary
  // for the typical use of this module.
  var makeUniqueKey_1 = makeUniqueKey;

  // Object.getOwnPropertyNames is the only way to enumerate non-enumerable
  // properties, so if we wrap it to ignore our secret keys, there should be
  // no way (except guessing) to access those properties.
  var originalGetOPNs = Object.getOwnPropertyNames;
  Object.getOwnPropertyNames = function getOwnPropertyNames(object) {
    for (var names = originalGetOPNs(object),
             src = 0,
             dst = 0,
             len = names.length;
         src < len;
         ++src) {
      if (!hasOwn$5.call(uniqueKeys, names[src])) {
        if (src > dst) {
          names[dst] = names[src];
        }
        ++dst;
      }
    }
    names.length = dst;
    return names;
  };

  function defaultCreatorFn(object) {
    return create$7(null);
  }

  function makeAccessor(secretCreatorFn) {
    var brand = makeUniqueKey();
    var passkey = create$7(null);

    secretCreatorFn = secretCreatorFn || defaultCreatorFn;

    function register(object) {
      var secret; // Created lazily.

      function vault(key, forget) {
        // Only code that has access to the passkey can retrieve (or forget)
        // the secret object.
        if (key === passkey) {
          return forget
            ? secret = null
            : secret || (secret = secretCreatorFn(object));
        }
      }

      defProp(object, brand, vault);
    }

    function accessor(object) {
      if (!hasOwn$5.call(object, brand))
        register(object);
      return object[brand](passkey);
    }

    accessor.forget = function(object) {
      if (hasOwn$5.call(object, brand))
        object[brand](passkey, true);
    };

    return accessor;
  }

  var makeAccessor_1 = makeAccessor;

  var _private = {
  	makeUniqueKey: makeUniqueKey_1,
  	makeAccessor: makeAccessor_1
  };

  var isString$1 = types$1.builtInTypes.string;
  var isNumber = types$1.builtInTypes.number;
  var SourceLocation = types$1.namedTypes.SourceLocation;
  var Position = types$1.namedTypes.Position;

  var comparePos = util_1$1.comparePos;

  function Mapping$2(sourceLines, sourceLoc, targetLoc) {
      SourceLocation.assert(sourceLoc);

      if (targetLoc) ; else {
          // Assume identity mapping if no targetLoc specified.
          targetLoc = sourceLoc;
      }

      Object.defineProperties(this, {
          sourceLines: { value: sourceLines },
          sourceLoc: { value: sourceLoc },
          targetLoc: { value: targetLoc }
      });
  }

  var Mp = Mapping$2.prototype;
  var mapping = Mapping$2;

  Mp.slice = function(lines$$1, start, end) {
      Position.assert(start);

      if (end) {
          Position.assert(end);
      } else {
          end = lines$$1.lastPos();
      }

      var sourceLines = this.sourceLines;
      var sourceLoc = this.sourceLoc;
      var targetLoc = this.targetLoc;

      function skip(name) {
          var sourceFromPos = sourceLoc[name];
          var targetFromPos = targetLoc[name];
          var targetToPos = start;

          if (name === "end") {
              targetToPos = end;
          }

          return skipChars(
              sourceLines, sourceFromPos,
              lines$$1, targetFromPos, targetToPos
          );
      }

      if (comparePos(start, targetLoc.start) <= 0) {
          if (comparePos(targetLoc.end, end) <= 0) {
              targetLoc = {
                  start: subtractPos(targetLoc.start, start.line, start.column),
                  end: subtractPos(targetLoc.end, start.line, start.column)
              };

              // The sourceLoc can stay the same because the contents of the
              // targetLoc have not changed.

          } else if (comparePos(end, targetLoc.start) <= 0) {
              return null;

          } else {
              sourceLoc = {
                  start: sourceLoc.start,
                  end: skip("end")
              };

              targetLoc = {
                  start: subtractPos(targetLoc.start, start.line, start.column),
                  end: subtractPos(end, start.line, start.column)
              };
          }

      } else {
          if (comparePos(targetLoc.end, start) <= 0) {
              return null;
          }

          if (comparePos(targetLoc.end, end) <= 0) {
              sourceLoc = {
                  start: skip("start"),
                  end: sourceLoc.end
              };

              targetLoc = {
                  // Same as subtractPos(start, start.line, start.column):
                  start: { line: 1, column: 0 },
                  end: subtractPos(targetLoc.end, start.line, start.column)
              };

          } else {
              sourceLoc = {
                  start: skip("start"),
                  end: skip("end")
              };

              targetLoc = {
                  // Same as subtractPos(start, start.line, start.column):
                  start: { line: 1, column: 0 },
                  end: subtractPos(end, start.line, start.column)
              };
          }
      }

      return new Mapping$2(this.sourceLines, sourceLoc, targetLoc);
  };

  Mp.add = function(line, column) {
      return new Mapping$2(this.sourceLines, this.sourceLoc, {
          start: addPos(this.targetLoc.start, line, column),
          end: addPos(this.targetLoc.end, line, column)
      });
  };

  function addPos(toPos, line, column) {
      return {
          line: toPos.line + line - 1,
          column: (toPos.line === 1)
              ? toPos.column + column
              : toPos.column
      };
  }

  Mp.subtract = function(line, column) {
      return new Mapping$2(this.sourceLines, this.sourceLoc, {
          start: subtractPos(this.targetLoc.start, line, column),
          end: subtractPos(this.targetLoc.end, line, column)
      });
  };

  function subtractPos(fromPos, line, column) {
      return {
          line: fromPos.line - line + 1,
          column: (fromPos.line === line)
              ? fromPos.column - column
              : fromPos.column
      };
  }

  Mp.indent = function(by, skipFirstLine, noNegativeColumns) {
      if (by === 0) {
          return this;
      }

      var targetLoc = this.targetLoc;
      var startLine = targetLoc.start.line;
      var endLine = targetLoc.end.line;

      if (skipFirstLine && startLine === 1 && endLine === 1) {
          return this;
      }

      targetLoc = {
          start: targetLoc.start,
          end: targetLoc.end
      };

      if (!skipFirstLine || startLine > 1) {
          var startColumn = targetLoc.start.column + by;
          targetLoc.start = {
              line: startLine,
              column: noNegativeColumns
                  ? Math.max(0, startColumn)
                  : startColumn
          };
      }

      if (!skipFirstLine || endLine > 1) {
          var endColumn = targetLoc.end.column + by;
          targetLoc.end = {
              line: endLine,
              column: noNegativeColumns
                  ? Math.max(0, endColumn)
                  : endColumn
          };
      }

      return new Mapping$2(this.sourceLines, this.sourceLoc, targetLoc);
  };

  function skipChars(
      sourceLines, sourceFromPos,
      targetLines, targetFromPos, targetToPos
  ) {
      Position.assert(sourceFromPos);
      Position.assert(targetFromPos);
      Position.assert(targetToPos);

      var targetComparison = comparePos(targetFromPos, targetToPos);
      if (targetComparison === 0) {
          // Trivial case: no characters to skip.
          return sourceFromPos;
      }

      if (targetComparison < 0) {
          // Skipping forward.

          var sourceCursor = sourceLines.skipSpaces(sourceFromPos);
          var targetCursor = targetLines.skipSpaces(targetFromPos);

          var lineDiff = targetToPos.line - targetCursor.line;
          sourceCursor.line += lineDiff;
          targetCursor.line += lineDiff;

          if (lineDiff > 0) {
              // If jumping to later lines, reset columns to the beginnings
              // of those lines.
              sourceCursor.column = 0;
              targetCursor.column = 0;
          }

          while (comparePos(targetCursor, targetToPos) < 0 &&
                 targetLines.nextPos(targetCursor, true)) {
          }

      } else {
          // Skipping backward.

          var sourceCursor = sourceLines.skipSpaces(sourceFromPos, true);
          var targetCursor = targetLines.skipSpaces(targetFromPos, true);

          var lineDiff = targetToPos.line - targetCursor.line;
          sourceCursor.line += lineDiff;
          targetCursor.line += lineDiff;

          if (lineDiff < 0) {
              // If jumping to earlier lines, reset columns to the ends of
              // those lines.
              sourceCursor.column = sourceLines.getLineLength(sourceCursor.line);
              targetCursor.column = targetLines.getLineLength(targetCursor.line);
          }

          while (comparePos(targetToPos, targetCursor) < 0 &&
                 targetLines.prevPos(targetCursor, true)) {
          }
      }

      return sourceCursor;
  }

  var normalizeOptions = options.normalize;
  var secretKey = _private.makeUniqueKey();

  var isString$2 = types$1.builtInTypes.string;
  var comparePos$1 = util_1$1.comparePos;


  // Goals:
  // 1. Minimize new string creation.
  // 2. Keep (de)identation O(lines) time.
  // 3. Permit negative indentations.
  // 4. Enforce immutability.
  // 5. No newline characters.

  var useSymbol = typeof Symbol === "function";
  var secretKey = "recastLinesSecret";
  if (useSymbol) {
    secretKey = Symbol.for(secretKey);
  }

  function getSecret(lines) {
    return lines[secretKey];
  }

  function Lines(infos, sourceFileName) {

    if (sourceFileName) {
      isString$2.assert(sourceFileName);
    } else {
      sourceFileName = null;
    }

    setSymbolOrKey(this, secretKey, {
      infos: infos,
      mappings: [],
      name: sourceFileName,
      cachedSourceMap: null
    });

    this.length = infos.length;
    this.name = sourceFileName;

    if (sourceFileName) {
      getSecret(this).mappings.push(new mapping(this, {
        start: this.firstPos(),
        end: this.lastPos()
      }));
    }
  }

  function setSymbolOrKey(obj, key, value) {
    if (useSymbol) {
      return obj[key] = value;
    }

    Object.defineProperty(obj, key, {
      value: value,
      enumerable: false,
      writable: false,
      configurable: true
    });

    return value;
  }

  // Exposed for instanceof checks. The fromString function should be used
  // to create new Lines objects.
  var Lines_1 = Lines;
  var Lp = Lines.prototype;

  function copyLineInfo(info) {
    return {
      line: info.line,
      indent: info.indent,
      locked: info.locked,
      sliceStart: info.sliceStart,
      sliceEnd: info.sliceEnd
    };
  }

  var fromStringCache = {};
  var hasOwn$6 = fromStringCache.hasOwnProperty;
  var maxCacheKeyLen = 10;

  function countSpaces(spaces, tabWidth) {
    var count = 0;
    var len = spaces.length;

    for (var i = 0; i < len; ++i) {
      switch (spaces.charCodeAt(i)) {
      case 9: // '\t'

        var next = Math.ceil(count / tabWidth) * tabWidth;
        if (next === count) {
          count += tabWidth;
        } else {
          count = next;
        }

        break;

      case 11: // '\v'
      case 12: // '\f'
      case 13: // '\r'
      case 0xfeff: // zero-width non-breaking space
        // These characters contribute nothing to indentation.
        break;

      case 32: // ' '
      default: // Treat all other whitespace like ' '.
        count += 1;
        break;
      }
    }

    return count;
  }
  var countSpaces_1 = countSpaces;

  var leadingSpaceExp = /^\s*/;

  // As specified here: http://www.ecma-international.org/ecma-262/6.0/#sec-line-terminators
  var lineTerminatorSeqExp =
    /\u000D\u000A|\u000D(?!\u000A)|\u000A|\u2028|\u2029/;

  /**
   * @param {Object} options - Options object that configures printing.
   */
  function fromString(string, options$$1) {
    if (string instanceof Lines)
      return string;

    string += "";

    var tabWidth = options$$1 && options$$1.tabWidth;
    var tabless = string.indexOf("\t") < 0;
    var locked = !! (options$$1 && options$$1.locked);
    var cacheable = !options$$1 && tabless && (string.length <= maxCacheKeyLen);

    if (cacheable && hasOwn$6.call(fromStringCache, string))
      return fromStringCache[string];

    var lines = new Lines(string.split(lineTerminatorSeqExp).map(function(line) {
      var spaces = leadingSpaceExp.exec(line)[0];
      return {
        line: line,
        indent: countSpaces(spaces, tabWidth),
        // Boolean indicating whether this line can be reindented.
        locked: locked,
        sliceStart: spaces.length,
        sliceEnd: line.length
      };
    }), normalizeOptions(options$$1).sourceFileName);

    if (cacheable)
      fromStringCache[string] = lines;

    return lines;
  }
  var fromString_1 = fromString;

  function isOnlyWhitespace(string) {
    return !/\S/.test(string);
  }

  Lp.toString = function(options$$1) {
    return this.sliceString(this.firstPos(), this.lastPos(), options$$1);
  };

  Lp.getSourceMap = function(sourceMapName, sourceRoot) {
    if (!sourceMapName) {
      // Although we could make up a name or generate an anonymous
      // source map, instead we assume that any consumer who does not
      // provide a name does not actually want a source map.
      return null;
    }

    var targetLines = this;

    function updateJSON(json) {
      json = json || {};

      isString$2.assert(sourceMapName);
      json.file = sourceMapName;

      if (sourceRoot) {
        isString$2.assert(sourceRoot);
        json.sourceRoot = sourceRoot;
      }

      return json;
    }

    var secret = getSecret(targetLines);
    if (secret.cachedSourceMap) {
      // Since Lines objects are immutable, we can reuse any source map
      // that was previously generated. Nevertheless, we return a new
      // JSON object here to protect the cached source map from outside
      // modification.
      return updateJSON(secret.cachedSourceMap.toJSON());
    }

    var smg = new sourceMap.SourceMapGenerator(updateJSON());
    var sourcesToContents = {};

    secret.mappings.forEach(function(mapping$$1) {
      var sourceCursor = mapping$$1.sourceLines.skipSpaces(
        mapping$$1.sourceLoc.start
      ) || mapping$$1.sourceLines.lastPos();

      var targetCursor = targetLines.skipSpaces(
        mapping$$1.targetLoc.start
      ) || targetLines.lastPos();

      while (comparePos$1(sourceCursor, mapping$$1.sourceLoc.end) < 0 &&
             comparePos$1(targetCursor, mapping$$1.targetLoc.end) < 0) {

        var sourceChar = mapping$$1.sourceLines.charAt(sourceCursor);
        var targetChar = targetLines.charAt(targetCursor);

        var sourceName = mapping$$1.sourceLines.name;

        // Add mappings one character at a time for maximum resolution.
        smg.addMapping({
          source: sourceName,
          original: { line: sourceCursor.line,
                      column: sourceCursor.column },
          generated: { line: targetCursor.line,
                       column: targetCursor.column }
        });

        if (!hasOwn$6.call(sourcesToContents, sourceName)) {
          var sourceContent = mapping$$1.sourceLines.toString();
          smg.setSourceContent(sourceName, sourceContent);
          sourcesToContents[sourceName] = sourceContent;
        }

        targetLines.nextPos(targetCursor, true);
        mapping$$1.sourceLines.nextPos(sourceCursor, true);
      }
    });

    secret.cachedSourceMap = smg;

    return smg.toJSON();
  };

  Lp.bootstrapCharAt = function(pos) {

    var line = pos.line,
    column = pos.column,
    strings = this.toString().split(lineTerminatorSeqExp),
    string = strings[line - 1];

    if (typeof string === "undefined")
      return "";

    if (column === string.length &&
        line < strings.length)
      return "\n";

    if (column >= string.length)
      return "";

    return string.charAt(column);
  };

  Lp.charAt = function(pos) {

    var line = pos.line,
    column = pos.column,
    secret = getSecret(this),
    infos = secret.infos,
    info = infos[line - 1],
    c = column;

    if (typeof info === "undefined" || c < 0)
      return "";

    var indent = this.getIndentAt(line);
    if (c < indent)
      return " ";

    c += info.sliceStart - indent;

    if (c === info.sliceEnd &&
        line < this.length)
      return "\n";

    if (c >= info.sliceEnd)
      return "";

    return info.line.charAt(c);
  };

  Lp.stripMargin = function(width, skipFirstLine) {
    if (width === 0)
      return this;

    if (skipFirstLine && this.length === 1)
      return this;

    var secret = getSecret(this);

    var lines = new Lines(secret.infos.map(function(info, i) {
      if (info.line && (i > 0 || !skipFirstLine)) {
        info = copyLineInfo(info);
        info.indent = Math.max(0, info.indent - width);
      }
      return info;
    }));

    if (secret.mappings.length > 0) {
      var newMappings = getSecret(lines).mappings;
      secret.mappings.forEach(function(mapping$$1) {
        newMappings.push(mapping$$1.indent(width, skipFirstLine, true));
      });
    }

    return lines;
  };

  Lp.indent = function(by) {
    if (by === 0)
      return this;

    var secret = getSecret(this);

    var lines = new Lines(secret.infos.map(function(info) {
      if (info.line && ! info.locked) {
        info = copyLineInfo(info);
        info.indent += by;
      }
      return info
    }));

    if (secret.mappings.length > 0) {
      var newMappings = getSecret(lines).mappings;
      secret.mappings.forEach(function(mapping$$1) {
        newMappings.push(mapping$$1.indent(by));
      });
    }

    return lines;
  };

  Lp.indentTail = function(by) {
    if (by === 0)
      return this;

    if (this.length < 2)
      return this;

    var secret = getSecret(this);

    var lines = new Lines(secret.infos.map(function(info, i) {
      if (i > 0 && info.line && ! info.locked) {
        info = copyLineInfo(info);
        info.indent += by;
      }

      return info;
    }));

    if (secret.mappings.length > 0) {
      var newMappings = getSecret(lines).mappings;
      secret.mappings.forEach(function(mapping$$1) {
        newMappings.push(mapping$$1.indent(by, true));
      });
    }

    return lines;
  };

  Lp.lockIndentTail = function () {
    if (this.length < 2) {
      return this;
    }

    var infos = getSecret(this).infos;

    return new Lines(infos.map(function (info, i) {
      info = copyLineInfo(info);
      info.locked = i > 0;
      return info;
    }));
  };

  Lp.getIndentAt = function(line) {
    var secret = getSecret(this),
    info = secret.infos[line - 1];
    return Math.max(info.indent, 0);
  };

  Lp.guessTabWidth = function() {
    var secret = getSecret(this);
    if (hasOwn$6.call(secret, "cachedTabWidth")) {
      return secret.cachedTabWidth;
    }

    var counts = []; // Sparse array.
    var lastIndent = 0;

    for (var line = 1, last = this.length; line <= last; ++line) {
      var info = secret.infos[line - 1];
      var sliced = info.line.slice(info.sliceStart, info.sliceEnd);

      // Whitespace-only lines don't tell us much about the likely tab
      // width of this code.
      if (isOnlyWhitespace(sliced)) {
        continue;
      }

      var diff = Math.abs(info.indent - lastIndent);
      counts[diff] = ~~counts[diff] + 1;
      lastIndent = info.indent;
    }

    var maxCount = -1;
    var result = 2;

    for (var tabWidth = 1;
         tabWidth < counts.length;
         tabWidth += 1) {
      if (hasOwn$6.call(counts, tabWidth) &&
          counts[tabWidth] > maxCount) {
        maxCount = counts[tabWidth];
        result = tabWidth;
      }
    }

    return secret.cachedTabWidth = result;
  };

  // Determine if the list of lines has a first line that starts with a //
  // or /* comment. If this is the case, the code may need to be wrapped in
  // parens to avoid ASI issues.
  Lp.startsWithComment = function () {
    var secret = getSecret(this);
    if (secret.infos.length === 0) {
      return false;
    }
    var firstLineInfo = secret.infos[0],
    sliceStart = firstLineInfo.sliceStart,
    sliceEnd = firstLineInfo.sliceEnd,
    firstLine = firstLineInfo.line.slice(sliceStart, sliceEnd).trim();
    return firstLine.length === 0 ||
      firstLine.slice(0, 2) === "//" ||
      firstLine.slice(0, 2) === "/*";
  };

  Lp.isOnlyWhitespace = function() {
    return isOnlyWhitespace(this.toString());
  };

  Lp.isPrecededOnlyByWhitespace = function(pos) {
    var secret = getSecret(this);
    var info = secret.infos[pos.line - 1];
    var indent = Math.max(info.indent, 0);

    var diff = pos.column - indent;
    if (diff <= 0) {
      // If pos.column does not exceed the indentation amount, then
      // there must be only whitespace before it.
      return true;
    }

    var start = info.sliceStart;
    var end = Math.min(start + diff, info.sliceEnd);
    var prefix = info.line.slice(start, end);

    return isOnlyWhitespace(prefix);
  };

  Lp.getLineLength = function(line) {
    var secret = getSecret(this),
    info = secret.infos[line - 1];
    return this.getIndentAt(line) + info.sliceEnd - info.sliceStart;
  };

  Lp.nextPos = function(pos, skipSpaces) {
    var l = Math.max(pos.line, 0),
    c = Math.max(pos.column, 0);

    if (c < this.getLineLength(l)) {
      pos.column += 1;

      return skipSpaces
        ? !!this.skipSpaces(pos, false, true)
        : true;
    }

    if (l < this.length) {
      pos.line += 1;
      pos.column = 0;

      return skipSpaces
        ? !!this.skipSpaces(pos, false, true)
        : true;
    }

    return false;
  };

  Lp.prevPos = function(pos, skipSpaces) {
    var l = pos.line,
    c = pos.column;

    if (c < 1) {
      l -= 1;

      if (l < 1)
        return false;

      c = this.getLineLength(l);

    } else {
      c = Math.min(c - 1, this.getLineLength(l));
    }

    pos.line = l;
    pos.column = c;

    return skipSpaces
      ? !!this.skipSpaces(pos, true, true)
      : true;
  };

  Lp.firstPos = function() {
    // Trivial, but provided for completeness.
    return { line: 1, column: 0 };
  };

  Lp.lastPos = function() {
    return {
      line: this.length,
      column: this.getLineLength(this.length)
    };
  };

  Lp.skipSpaces = function(pos, backward, modifyInPlace) {
    if (pos) {
      pos = modifyInPlace ? pos : {
        line: pos.line,
        column: pos.column
      };
    } else if (backward) {
      pos = this.lastPos();
    } else {
      pos = this.firstPos();
    }

    if (backward) {
      while (this.prevPos(pos)) {
        if (!isOnlyWhitespace(this.charAt(pos)) &&
            this.nextPos(pos)) {
          return pos;
        }
      }

      return null;

    } else {
      while (isOnlyWhitespace(this.charAt(pos))) {
        if (!this.nextPos(pos)) {
          return null;
        }
      }

      return pos;
    }
  };

  Lp.trimLeft = function() {
    var pos = this.skipSpaces(this.firstPos(), false, true);
    return pos ? this.slice(pos) : emptyLines;
  };

  Lp.trimRight = function() {
    var pos = this.skipSpaces(this.lastPos(), true, true);
    return pos ? this.slice(this.firstPos(), pos) : emptyLines;
  };

  Lp.trim = function() {
    var start = this.skipSpaces(this.firstPos(), false, true);
    if (start === null)
      return emptyLines;

    var end = this.skipSpaces(this.lastPos(), true, true);

    return this.slice(start, end);
  };

  Lp.eachPos = function(callback, startPos, skipSpaces) {
    var pos = this.firstPos();

    if (startPos) {
      pos.line = startPos.line,
      pos.column = startPos.column;
    }

    if (skipSpaces && !this.skipSpaces(pos, false, true)) {
      return; // Encountered nothing but spaces.
    }

    do callback.call(this, pos);
    while (this.nextPos(pos, skipSpaces));
  };

  Lp.bootstrapSlice = function(start, end) {
    var strings = this.toString().split(
      lineTerminatorSeqExp
    ).slice(
      start.line - 1,
      end.line
    );

    strings.push(strings.pop().slice(0, end.column));
    strings[0] = strings[0].slice(start.column);

    return fromString(strings.join("\n"));
  };

  Lp.slice = function(start, end) {
    if (!end) {
      if (!start) {
        // The client seems to want a copy of this Lines object, but
        // Lines objects are immutable, so it's perfectly adequate to
        // return the same object.
        return this;
      }

      // Slice to the end if no end position was provided.
      end = this.lastPos();
    }

    var secret = getSecret(this);
    var sliced = secret.infos.slice(start.line - 1, end.line);

    if (start.line === end.line) {
      sliced[0] = sliceInfo(sliced[0], start.column, end.column);
    } else {
      sliced[0] = sliceInfo(sliced[0], start.column);
      sliced.push(sliceInfo(sliced.pop(), 0, end.column));
    }

    var lines = new Lines(sliced);

    if (secret.mappings.length > 0) {
      var newMappings = getSecret(lines).mappings;
      secret.mappings.forEach(function(mapping$$1) {
        var sliced = mapping$$1.slice(this, start, end);
        if (sliced) {
          newMappings.push(sliced);
        }
      }, this);
    }

    return lines;
  };

  function sliceInfo(info, startCol, endCol) {
    var sliceStart = info.sliceStart;
    var sliceEnd = info.sliceEnd;
    var indent = Math.max(info.indent, 0);
    var lineLength = indent + sliceEnd - sliceStart;

    if (typeof endCol === "undefined") {
      endCol = lineLength;
    }

    startCol = Math.max(startCol, 0);
    endCol = Math.min(endCol, lineLength);
    endCol = Math.max(endCol, startCol);

    if (endCol < indent) {
      indent = endCol;
      sliceEnd = sliceStart;
    } else {
      sliceEnd -= lineLength - endCol;
    }

    lineLength = endCol;
    lineLength -= startCol;

    if (startCol < indent) {
      indent -= startCol;
    } else {
      startCol -= indent;
      indent = 0;
      sliceStart += startCol;
    }

    if (info.indent === indent &&
        info.sliceStart === sliceStart &&
        info.sliceEnd === sliceEnd) {
      return info;
    }

    return {
      line: info.line,
      indent: indent,
      // A destructive slice always unlocks indentation.
      locked: false,
      sliceStart: sliceStart,
      sliceEnd: sliceEnd
    };
  }

  Lp.bootstrapSliceString = function(start, end, options$$1) {
    return this.slice(start, end).toString(options$$1);
  };

  Lp.sliceString = function(start, end, options$$1) {
    if (!end) {
      if (!start) {
        // The client seems to want a copy of this Lines object, but
        // Lines objects are immutable, so it's perfectly adequate to
        // return the same object.
        return this;
      }

      // Slice to the end if no end position was provided.
      end = this.lastPos();
    }

    options$$1 = normalizeOptions(options$$1);

    var infos = getSecret(this).infos;
    var parts = [];
    var tabWidth = options$$1.tabWidth;

    for (var line = start.line; line <= end.line; ++line) {
      var info = infos[line - 1];

      if (line === start.line) {
        if (line === end.line) {
          info = sliceInfo(info, start.column, end.column);
        } else {
          info = sliceInfo(info, start.column);
        }
      } else if (line === end.line) {
        info = sliceInfo(info, 0, end.column);
      }

      var indent = Math.max(info.indent, 0);

      var before = info.line.slice(0, info.sliceStart);
      if (options$$1.reuseWhitespace &&
          isOnlyWhitespace(before) &&
          countSpaces(before, options$$1.tabWidth) === indent) {
        // Reuse original spaces if the indentation is correct.
        parts.push(info.line.slice(0, info.sliceEnd));
        continue;
      }

      var tabs = 0;
      var spaces = indent;

      if (options$$1.useTabs) {
        tabs = Math.floor(indent / tabWidth);
        spaces -= tabs * tabWidth;
      }

      var result = "";

      if (tabs > 0) {
        result += new Array(tabs + 1).join("\t");
      }

      if (spaces > 0) {
        result += new Array(spaces + 1).join(" ");
      }

      result += info.line.slice(info.sliceStart, info.sliceEnd);

      parts.push(result);
    }

    return parts.join(options$$1.lineTerminator);
  };

  Lp.isEmpty = function() {
    return this.length < 2 && this.getLineLength(1) < 1;
  };

  Lp.join = function(elements) {
    var separator = this;
    var separatorSecret = getSecret(separator);
    var infos = [];
    var mappings = [];
    var prevInfo;

    function appendSecret(secret) {
      if (secret === null)
        return;

      if (prevInfo) {
        var info = secret.infos[0];
        var indent = new Array(info.indent + 1).join(" ");
        var prevLine = infos.length;
        var prevColumn = Math.max(prevInfo.indent, 0) +
          prevInfo.sliceEnd - prevInfo.sliceStart;

        prevInfo.line = prevInfo.line.slice(
          0, prevInfo.sliceEnd) + indent + info.line.slice(
            info.sliceStart, info.sliceEnd);

        // If any part of a line is indentation-locked, the whole line
        // will be indentation-locked.
        prevInfo.locked = prevInfo.locked || info.locked;

        prevInfo.sliceEnd = prevInfo.line.length;

        if (secret.mappings.length > 0) {
          secret.mappings.forEach(function(mapping$$1) {
            mappings.push(mapping$$1.add(prevLine, prevColumn));
          });
        }

      } else if (secret.mappings.length > 0) {
        mappings.push.apply(mappings, secret.mappings);
      }

      secret.infos.forEach(function(info, i) {
        if (!prevInfo || i > 0) {
          prevInfo = copyLineInfo(info);
          infos.push(prevInfo);
        }
      });
    }

    function appendWithSeparator(secret, i) {
      if (i > 0)
        appendSecret(separatorSecret);
      appendSecret(secret);
    }

    elements.map(function(elem) {
      var lines = fromString(elem);
      if (lines.isEmpty())
        return null;
      return getSecret(lines);
    }).forEach(separator.isEmpty()
               ? appendSecret
               : appendWithSeparator);

    if (infos.length < 1)
      return emptyLines;

    var lines = new Lines(infos);

    getSecret(lines).mappings = mappings;

    return lines;
  };

  var concat = function(elements) {
    return emptyLines.join(elements);
  };

  Lp.concat = function(other) {
    var args = arguments,
    list = [this];
    list.push.apply(list, args);
    return emptyLines.join(list);
  };

  // The emptyLines object needs to be created all the way down here so that
  // Lines.prototype will be fully populated.
  var emptyLines = fromString("");

  var lines = {
  	Lines: Lines_1,
  	countSpaces: countSpaces_1,
  	fromString: fromString_1,
  	concat: concat
  };

  var n = types$1.namedTypes;
  var Node$1 = n.Node;
  var isArray = types$1.builtInTypes.array;
  var isNumber$1 = types$1.builtInTypes.number;


  function FastPath(value) {
    this.stack = [value];
  }

  var FPp = FastPath.prototype;
  var fastPath = FastPath;

  // Static convenience function for coercing a value to a FastPath.
  FastPath.from = function(obj) {
    if (obj instanceof FastPath) {
      // Return a defensive copy of any existing FastPath instances.
      return obj.copy();
    }

    if (obj instanceof types$1.NodePath) {
      // For backwards compatibility, unroll NodePath instances into
      // lightweight FastPath [..., name, value] stacks.
      var copy = Object.create(FastPath.prototype);
      var stack = [obj.value];
      for (var pp; (pp = obj.parentPath); obj = pp)
        stack.push(obj.name, pp.value);
      copy.stack = stack.reverse();
      return copy;
    }

    // Otherwise use obj as the value of the new FastPath instance.
    return new FastPath(obj);
  };

  FPp.copy = function copy() {
    var copy = Object.create(FastPath.prototype);
    copy.stack = this.stack.slice(0);
    return copy;
  };

  // The name of the current property is always the penultimate element of
  // this.stack, and always a String.
  FPp.getName = function getName() {
    var s = this.stack;
    var len = s.length;
    if (len > 1) {
      return s[len - 2];
    }
    // Since the name is always a string, null is a safe sentinel value to
    // return if we do not know the name of the (root) value.
    return null;
  };

  // The value of the current property is always the final element of
  // this.stack.
  FPp.getValue = function getValue() {
    var s = this.stack;
    return s[s.length - 1];
  };

  FPp.valueIsDuplicate = function () {
    var s = this.stack;
    var valueIndex = s.length - 1;
    return s.lastIndexOf(s[valueIndex], valueIndex - 1) >= 0;
  };

  function getNodeHelper(path, count) {
    var s = path.stack;

    for (var i = s.length - 1; i >= 0; i -= 2) {
      var value = s[i];
      if (n.Node.check(value) && --count < 0) {
        return value;
      }
    }

    return null;
  }

  FPp.getNode = function getNode(count) {
    return getNodeHelper(this, ~~count);
  };

  FPp.getParentNode = function getParentNode(count) {
    return getNodeHelper(this, ~~count + 1);
  };

  // The length of the stack can be either even or odd, depending on whether
  // or not we have a name for the root value. The difference between the
  // index of the root value and the index of the final value is always
  // even, though, which allows us to return the root value in constant time
  // (i.e. without iterating backwards through the stack).
  FPp.getRootValue = function getRootValue() {
    var s = this.stack;
    if (s.length % 2 === 0) {
      return s[1];
    }
    return s[0];
  };

  // Temporarily push properties named by string arguments given after the
  // callback function onto this.stack, then call the callback with a
  // reference to this (modified) FastPath object. Note that the stack will
  // be restored to its original state after the callback is finished, so it
  // is probably a mistake to retain a reference to the path.
  FPp.call = function call(callback/*, name1, name2, ... */) {
    var s = this.stack;
    var origLen = s.length;
    var value = s[origLen - 1];
    var argc = arguments.length;
    for (var i = 1; i < argc; ++i) {
      var name = arguments[i];
      value = value[name];
      s.push(name, value);
    }
    var result = callback(this);
    s.length = origLen;
    return result;
  };

  // Similar to FastPath.prototype.call, except that the value obtained by
  // accessing this.getValue()[name1][name2]... should be array-like. The
  // callback will be called with a reference to this path object for each
  // element of the array.
  FPp.each = function each(callback/*, name1, name2, ... */) {
    var s = this.stack;
    var origLen = s.length;
    var value = s[origLen - 1];
    var argc = arguments.length;

    for (var i = 1; i < argc; ++i) {
      var name = arguments[i];
      value = value[name];
      s.push(name, value);
    }

    for (var i = 0; i < value.length; ++i) {
      if (i in value) {
        s.push(i, value[i]);
        // If the callback needs to know the value of i, call
        // path.getName(), assuming path is the parameter name.
        callback(this);
        s.length -= 2;
      }
    }

    s.length = origLen;
  };

  // Similar to FastPath.prototype.each, except that the results of the
  // callback function invocations are stored in an array and returned at
  // the end of the iteration.
  FPp.map = function map(callback/*, name1, name2, ... */) {
    var s = this.stack;
    var origLen = s.length;
    var value = s[origLen - 1];
    var argc = arguments.length;

    for (var i = 1; i < argc; ++i) {
      var name = arguments[i];
      value = value[name];
      s.push(name, value);
    }

    var result = new Array(value.length);

    for (var i = 0; i < value.length; ++i) {
      if (i in value) {
        s.push(i, value[i]);
        result[i] = callback(this, i);
        s.length -= 2;
      }
    }

    s.length = origLen;

    return result;
  };

  // Returns true if the node at the tip of the path is wrapped with
  // parentheses, OR if the only reason the node needed parentheses was that
  // it couldn't be the first expression in the enclosing statement (see
  // FastPath#canBeFirstInStatement), and it has an opening `(` character.
  // For example, the FunctionExpression in `(function(){}())` appears to
  // need parentheses only because it's the first expression in the AST, but
  // since it happens to be preceded by a `(` (which is not apparent from
  // the AST but can be determined using FastPath#getPrevToken), there is no
  // ambiguity about how to parse it, so it counts as having parentheses,
  // even though it is not immediately followed by a `)`.
  FPp.hasParens = function () {
    const node = this.getNode();

    const prevToken = this.getPrevToken(node);
    if (! prevToken) {
      return false;
    }

    const nextToken = this.getNextToken(node);
    if (! nextToken) {
      return false;
    }

    if (prevToken.value === "(") {
      if (nextToken.value === ")") {
        // If the node preceded by a `(` token and followed by a `)` token,
        // then of course it has parentheses.
        return true;
      }

      // If this is one of the few Expression types that can't come first in
      // the enclosing statement because of parsing ambiguities (namely,
      // FunctionExpression, ObjectExpression, and ClassExpression) and
      // this.firstInStatement() returns true, and the node would not need
      // parentheses in an expression context because this.needsParens(true)
      // returns false, then it just needs an opening parenthesis to resolve
      // the parsing ambiguity that made it appear to need parentheses.
      const justNeedsOpeningParen =
        ! this.canBeFirstInStatement() &&
        this.firstInStatement() &&
        ! this.needsParens(true);

      if (justNeedsOpeningParen) {
        return true;
      }
    }

    return false;
  };

  FPp.getPrevToken = function (node) {
    node = node || this.getNode();
    const loc = node && node.loc;
    const tokens = loc && loc.tokens;
    if (tokens && loc.start.token > 0) {
      const token = tokens[loc.start.token - 1];
      if (token) {
        // Do not return tokens that fall outside the root subtree.
        const rootLoc = this.getRootValue().loc;
        if (util_1$1.comparePos(rootLoc.start, token.loc.start) <= 0) {
          return token;
        }
      }
    }
    return null;
  };

  FPp.getNextToken = function (node) {
    node = node || this.getNode();
    const loc = node && node.loc;
    const tokens = loc && loc.tokens;
    if (tokens && loc.end.token < tokens.length) {
      const token = tokens[loc.end.token];
      if (token) {
        // Do not return tokens that fall outside the root subtree.
        const rootLoc = this.getRootValue().loc;
        if (util_1$1.comparePos(token.loc.end, rootLoc.end) <= 0) {
          return token;
        }
      }
    }
    return null;
  };

  // Inspired by require("ast-types").NodePath.prototype.needsParens, but
  // more efficient because we're iterating backwards through a stack.
  FPp.needsParens = function(assumeExpressionContext) {
    var node = this.getNode();

    // This needs to come before `if (!parent) { return false }` because
    // an object destructuring assignment requires parens for
    // correctness even when it's the topmost expression.
    if (node.type === "AssignmentExpression" && node.left.type === 'ObjectPattern') {
      return true;
    }

    var parent = this.getParentNode();
    if (!parent) {
      return false;
    }

    var name = this.getName();

    // If the value of this path is some child of a Node and not a Node
    // itself, then it doesn't need parentheses. Only Node objects (in fact,
    // only Expression nodes) need parentheses.
    if (this.getValue() !== node) {
      return false;
    }

    // Only statements don't need parentheses.
    if (n.Statement.check(node)) {
      return false;
    }

    // Identifiers never need parentheses.
    if (node.type === "Identifier") {
      return false;
    }

    if (parent.type === "ParenthesizedExpression") {
      return false;
    }

    switch (node.type) {
    case "UnaryExpression":
    case "SpreadElement":
    case "SpreadProperty":
      return parent.type === "MemberExpression"
        && name === "object"
        && parent.object === node;

    case "BinaryExpression":
    case "LogicalExpression":
      switch (parent.type) {
      case "CallExpression":
        return name === "callee"
          && parent.callee === node;

      case "UnaryExpression":
      case "SpreadElement":
      case "SpreadProperty":
        return true;

      case "MemberExpression":
        return name === "object"
          && parent.object === node;

      case "BinaryExpression":
      case "LogicalExpression":
        var po = parent.operator;
        var pp = PRECEDENCE[po];
        var no = node.operator;
        var np = PRECEDENCE[no];

        if (pp > np) {
          return true;
        }

        if (pp === np && name === "right") {
          return true;
        }

      default:
        return false;
      }

    case "SequenceExpression":
      switch (parent.type) {
      case "ReturnStatement":
        return false;

      case "ForStatement":
        // Although parentheses wouldn't hurt around sequence expressions in
        // the head of for loops, traditional style dictates that e.g. i++,
        // j++ should not be wrapped with parentheses.
        return false;

      case "ExpressionStatement":
        return name !== "expression";

      default:
        // Otherwise err on the side of overparenthesization, adding
        // explicit exceptions above if this proves overzealous.
        return true;
      }

    case "YieldExpression":
      switch (parent.type) {
      case "BinaryExpression":
      case "LogicalExpression":
      case "UnaryExpression":
      case "SpreadElement":
      case "SpreadProperty":
      case "CallExpression":
      case "MemberExpression":
      case "NewExpression":
      case "ConditionalExpression":
      case "YieldExpression":
        return true;

      default:
        return false;
      }

    case "IntersectionTypeAnnotation":
    case "UnionTypeAnnotation":
      return parent.type === "NullableTypeAnnotation";

    case "Literal":
      return parent.type === "MemberExpression"
        && isNumber$1.check(node.value)
        && name === "object"
        && parent.object === node;

    // Babel 6 Literal split
    case "NumericLiteral":
      return parent.type === "MemberExpression"
        && name === "object"
        && parent.object === node;

    case "AssignmentExpression":
    case "ConditionalExpression":
      switch (parent.type) {
      case "UnaryExpression":
      case "SpreadElement":
      case "SpreadProperty":
      case "BinaryExpression":
      case "LogicalExpression":
        return true;

      case "CallExpression":
      case "NewExpression":
        return name === "callee"
          && parent.callee === node;

      case "ConditionalExpression":
        return name === "test"
          && parent.test === node;

      case "MemberExpression":
        return name === "object"
          && parent.object === node;

      default:
        return false;
      }

    case "ArrowFunctionExpression":
      if (n.CallExpression.check(parent) &&
          name === 'callee') {
        return true;
      }

      if (n.MemberExpression.check(parent) &&
          name === 'object') {
        return true;
      }

      return isBinary(parent);

    case "ObjectExpression":
      if (parent.type === "ArrowFunctionExpression" &&
          name === "body") {
        return true;
      }

      break;

    case "CallExpression":
      if (name === "declaration" &&
          n.ExportDefaultDeclaration.check(parent) &&
          n.FunctionExpression.check(node.callee)) {
        return true;
      }
    }

    if (parent.type === "NewExpression" &&
        name === "callee" &&
        parent.callee === node) {
      return containsCallExpression(node);
    }

    if (assumeExpressionContext !== true &&
        !this.canBeFirstInStatement() &&
        this.firstInStatement()) {
      return true;
    }

    return false;
  };

  function isBinary(node) {
    return n.BinaryExpression.check(node)
      || n.LogicalExpression.check(node);
  }

  var PRECEDENCE = {};
  [["||"],
   ["&&"],
   ["|"],
   ["^"],
   ["&"],
   ["==", "===", "!=", "!=="],
   ["<", ">", "<=", ">=", "in", "instanceof"],
   [">>", "<<", ">>>"],
   ["+", "-"],
   ["*", "/", "%", "**"]
  ].forEach(function(tier, i) {
    tier.forEach(function(op) {
      PRECEDENCE[op] = i;
    });
  });

  function containsCallExpression(node) {
    if (n.CallExpression.check(node)) {
      return true;
    }

    if (isArray.check(node)) {
      return node.some(containsCallExpression);
    }

    if (n.Node.check(node)) {
      return types$1.someField(node, function(name, child) {
        return containsCallExpression(child);
      });
    }

    return false;
  }

  FPp.canBeFirstInStatement = function() {
    var node = this.getNode();

    if (n.FunctionExpression.check(node)) {
      return false;
    }

    if (n.ObjectExpression.check(node)) {
      return false;
    }

    if (n.ClassExpression.check(node)) {
      return false;
    }

    return true;
  };

  FPp.firstInStatement = function() {
    var s = this.stack;
    var parentName, parent;
    var childName, child;

    for (var i = s.length - 1; i >= 0; i -= 2) {
      if (n.Node.check(s[i])) {
        childName = parentName;
        child = parent;
        parentName = s[i - 1];
        parent = s[i];
      }

      if (!parent || !child) {
        continue;
      }

      if (n.BlockStatement.check(parent) &&
          parentName === "body" &&
          childName === 0) {
        return true;
      }

      if (n.ExpressionStatement.check(parent) &&
          childName === "expression") {
        return true;
      }

      if (n.AssignmentExpression.check(parent) &&
          childName === "left") {
        return true;
      }

      if (n.ArrowFunctionExpression.check(parent) &&
          childName === "body") {
        return true;
      }

      if (n.SequenceExpression.check(parent) &&
          parentName === "expressions" &&
          childName === 0) {
        continue;
      }

      if (n.CallExpression.check(parent) &&
          childName === "callee") {
        continue;
      }

      if (n.MemberExpression.check(parent) &&
          childName === "object") {
        continue;
      }

      if (n.ConditionalExpression.check(parent) &&
          childName === "test") {
        continue;
      }

      if (isBinary(parent) &&
          childName === "left") {
        continue;
      }

      if (n.UnaryExpression.check(parent) &&
          !parent.prefix &&
          childName === "argument") {
        continue;
      }

      return false;
    }

    return true;
  };

  var getFieldValue = types$1.getFieldValue;
  var Node$2 = types$1.namedTypes.Node;
  var Printable = types$1.namedTypes.Printable;
  var Expression$1 = types$1.namedTypes.Expression;
  var ReturnStatement = types$1.namedTypes.ReturnStatement;
  var SourceLocation$1 = types$1.namedTypes.SourceLocation;

  var comparePos$2 = util_1$1.comparePos;

  var isObject = types$1.builtInTypes.object;
  var isArray$1 = types$1.builtInTypes.array;
  var isString$3 = types$1.builtInTypes.string;
  var riskyAdjoiningCharExp = /[0-9a-z_$]/i;

  function Patcher(lines$$1) {

    var self = this,
    replacements = [];

    self.replace = function(loc, lines$$1) {
      if (isString$3.check(lines$$1))
        lines$$1 = lines.fromString(lines$$1);

      replacements.push({
        lines: lines$$1,
        start: loc.start,
        end: loc.end
      });
    };

    self.get = function(loc) {
      // If no location is provided, return the complete Lines object.
      loc = loc || {
        start: { line: 1, column: 0 },
        end: { line: lines$$1.length,
               column: lines$$1.getLineLength(lines$$1.length) }
      };

      var sliceFrom = loc.start,
      toConcat = [];

      function pushSlice(from, to) {
        toConcat.push(lines$$1.slice(from, to));
      }

      replacements.sort(function(a, b) {
        return comparePos$2(a.start, b.start);
      }).forEach(function(rep) {
        if (comparePos$2(sliceFrom, rep.start) > 0) ; else {
          pushSlice(sliceFrom, rep.start);
          toConcat.push(rep.lines);
          sliceFrom = rep.end;
        }
      });

      pushSlice(sliceFrom, loc.end);

      return lines.concat(toConcat);
    };
  }
  var Patcher_1 = Patcher;

  var Pp = Patcher.prototype;

  Pp.tryToReprintComments = function(newNode, oldNode, print) {
    var patcher = this;

    if (!newNode.comments &&
        !oldNode.comments) {
      // We were (vacuously) able to reprint all the comments!
      return true;
    }

    var newPath = fastPath.from(newNode);
    var oldPath = fastPath.from(oldNode);

    newPath.stack.push("comments", getSurroundingComments(newNode));
    oldPath.stack.push("comments", getSurroundingComments(oldNode));

    var reprints = [];
    var ableToReprintComments =
      findArrayReprints(newPath, oldPath, reprints);

    // No need to pop anything from newPath.stack or oldPath.stack, since
    // newPath and oldPath are fresh local variables.

    if (ableToReprintComments && reprints.length > 0) {
      reprints.forEach(function(reprint) {
        var oldComment = reprint.oldPath.getValue();
        patcher.replace(
          oldComment.loc,
          // Comments can't have .comments, so it doesn't matter whether we
          // print with comments or without.
          print(reprint.newPath).indentTail(oldComment.loc.indent)
        );
      });
    }

    return ableToReprintComments;
  };

  // Get all comments that are either leading or trailing, ignoring any
  // comments that occur inside node.loc. Returns an empty array for nodes
  // with no leading or trailing comments.
  function getSurroundingComments(node) {
    var result = [];
    if (node.comments &&
        node.comments.length > 0) {
      node.comments.forEach(function(comment) {
        if (comment.leading || comment.trailing) {
          result.push(comment);
        }
      });
    }
    return result;
  }

  Pp.deleteComments = function(node) {
    if (!node.comments) {
      return;
    }

    var patcher = this;

    node.comments.forEach(function(comment) {
      if (comment.leading) {
        // Delete leading comments along with any trailing whitespace they
        // might have.
        patcher.replace({
          start: comment.loc.start,
          end: node.loc.lines.skipSpaces(
            comment.loc.end, false, false)
        }, "");

      } else if (comment.trailing) {
        // Delete trailing comments along with any leading whitespace they
        // might have.
        patcher.replace({
          start: node.loc.lines.skipSpaces(
            comment.loc.start, true, false),
          end: comment.loc.end
        }, "");
      }
    });
  };

  var getReprinter = function(path) {

    // Make sure that this path refers specifically to a Node, rather than
    // some non-Node subproperty of a Node.
    var node = path.getValue();
    if (!Printable.check(node))
      return;

    var orig = node.original;
    var origLoc = orig && orig.loc;
    var lines$$1 = origLoc && origLoc.lines;
    var reprints = [];

    if (!lines$$1 || !findReprints(path, reprints))
      return;

    return function(print) {
      var patcher = new Patcher(lines$$1);

      reprints.forEach(function(reprint) {
        var newNode = reprint.newPath.getValue();
        var oldNode = reprint.oldPath.getValue();

        SourceLocation$1.assert(oldNode.loc, true);

        var needToPrintNewPathWithComments =
          !patcher.tryToReprintComments(newNode, oldNode, print);

        if (needToPrintNewPathWithComments) {
          // Since we were not able to preserve all leading/trailing
          // comments, we delete oldNode's comments, print newPath with
          // comments, and then patch the resulting lines where oldNode used
          // to be.
          patcher.deleteComments(oldNode);
        }

        var newLines = print(reprint.newPath, {
          includeComments: needToPrintNewPathWithComments,
          // If the oldNode we're replacing already had parentheses, we may
          // not need to print the new node with any extra parentheses,
          // because the existing parentheses will suffice. However, if the
          // newNode has a different type than the oldNode, let the printer
          // decide if reprint.newPath needs parentheses, as usual.
          avoidRootParens: (oldNode.type === newNode.type &&
                            reprint.oldPath.hasParens())
        }).indentTail(oldNode.loc.indent);

        var nls = needsLeadingSpace(lines$$1, oldNode.loc, newLines);
        var nts = needsTrailingSpace(lines$$1, oldNode.loc, newLines);

        // If we try to replace the argument of a ReturnStatement like
        // return"asdf" with e.g. a literal null expression, we run the risk
        // of ending up with returnnull, so we need to add an extra leading
        // space in situations where that might happen. Likewise for
        // "asdf"in obj. See #170.
        if (nls || nts) {
          var newParts = [];
          nls && newParts.push(" ");
          newParts.push(newLines);
          nts && newParts.push(" ");
          newLines = lines.concat(newParts);
        }

        patcher.replace(oldNode.loc, newLines);
      });

      // Recall that origLoc is the .loc of an ancestor node that is
      // guaranteed to contain all the reprinted nodes and comments.
      const patchedLines = patcher.get(origLoc).indentTail(-orig.loc.indent);

      if (path.needsParens()) {
        return lines.concat(["(", patchedLines, ")"]);
      }

      return patchedLines;
    };
  };

  // If the last character before oldLoc and the first character of newLines
  // are both identifier characters, they must be separated by a space,
  // otherwise they will most likely get fused together into a single token.
  function needsLeadingSpace(oldLines, oldLoc, newLines) {
    var posBeforeOldLoc = util_1$1.copyPos(oldLoc.start);

    // The character just before the location occupied by oldNode.
    var charBeforeOldLoc =
      oldLines.prevPos(posBeforeOldLoc) &&
      oldLines.charAt(posBeforeOldLoc);

    // First character of the reprinted node.
    var newFirstChar = newLines.charAt(newLines.firstPos());

    return charBeforeOldLoc &&
      riskyAdjoiningCharExp.test(charBeforeOldLoc) &&
      newFirstChar &&
      riskyAdjoiningCharExp.test(newFirstChar);
  }

  // If the last character of newLines and the first character after oldLoc
  // are both identifier characters, they must be separated by a space,
  // otherwise they will most likely get fused together into a single token.
  function needsTrailingSpace(oldLines, oldLoc, newLines) {
    // The character just after the location occupied by oldNode.
    var charAfterOldLoc = oldLines.charAt(oldLoc.end);

    var newLastPos = newLines.lastPos();

    // Last character of the reprinted node.
    var newLastChar = newLines.prevPos(newLastPos) &&
      newLines.charAt(newLastPos);

    return newLastChar &&
      riskyAdjoiningCharExp.test(newLastChar) &&
      charAfterOldLoc &&
      riskyAdjoiningCharExp.test(charAfterOldLoc);
  }

  function findReprints(newPath, reprints) {
    var newNode = newPath.getValue();
    Printable.assert(newNode);

    var oldNode = newNode.original;
    Printable.assert(oldNode);

    if (newNode.type !== oldNode.type) {
      return false;
    }

    var oldPath = new fastPath(oldNode);
    var canReprint = findChildReprints(newPath, oldPath, reprints);

    if (!canReprint) {
      // Make absolutely sure the calling code does not attempt to reprint
      // any nodes.
      reprints.length = 0;
    }

    return canReprint;
  }

  function findAnyReprints(newPath, oldPath, reprints) {
    var newNode = newPath.getValue();
    var oldNode = oldPath.getValue();

    if (newNode === oldNode)
      return true;

    if (isArray$1.check(newNode))
      return findArrayReprints(newPath, oldPath, reprints);

    if (isObject.check(newNode))
      return findObjectReprints(newPath, oldPath, reprints);

    return false;
  }

  function findArrayReprints(newPath, oldPath, reprints) {
    var newNode = newPath.getValue();
    var oldNode = oldPath.getValue();

    if (newNode === oldNode ||
        newPath.valueIsDuplicate() ||
        oldPath.valueIsDuplicate()) {
      return true;
    }

    isArray$1.assert(newNode);
    var len = newNode.length;

    if (!(isArray$1.check(oldNode) &&
          oldNode.length === len))
      return false;

    for (var i = 0; i < len; ++i) {
      newPath.stack.push(i, newNode[i]);
      oldPath.stack.push(i, oldNode[i]);
      var canReprint = findAnyReprints(newPath, oldPath, reprints);
      newPath.stack.length -= 2;
      oldPath.stack.length -= 2;
      if (!canReprint) {
        return false;
      }
    }

    return true;
  }

  function findObjectReprints(newPath, oldPath, reprints) {
    var newNode = newPath.getValue();
    isObject.assert(newNode);

    if (newNode.original === null) {
      // If newNode.original node was set to null, reprint the node.
      return false;
    }

    var oldNode = oldPath.getValue();
    if (!isObject.check(oldNode))
      return false;

    if (newNode === oldNode ||
        newPath.valueIsDuplicate() ||
        oldPath.valueIsDuplicate()) {
      return true;
    }

    if (Printable.check(newNode)) {
      if (!Printable.check(oldNode)) {
        return false;
      }

      // Here we need to decide whether the reprinted code for newNode is
      // appropriate for patching into the location of oldNode.

      if (newNode.type === oldNode.type) {
        var childReprints = [];

        if (findChildReprints(newPath, oldPath, childReprints)) {
          reprints.push.apply(reprints, childReprints);
        } else if (oldNode.loc) {
          // If we have no .loc information for oldNode, then we won't be
          // able to reprint it.
          reprints.push({
            oldPath: oldPath.copy(),
            newPath: newPath.copy()
          });
        } else {
          return false;
        }

        return true;
      }

      if (Expression$1.check(newNode) &&
          Expression$1.check(oldNode) &&
          // If we have no .loc information for oldNode, then we won't be
          // able to reprint it.
          oldNode.loc) {

        // If both nodes are subtypes of Expression, then we should be able
        // to fill the location occupied by the old node with code printed
        // for the new node with no ill consequences.
        reprints.push({
          oldPath: oldPath.copy(),
          newPath: newPath.copy()
        });

        return true;
      }

      // The nodes have different types, and at least one of the types is
      // not a subtype of the Expression type, so we cannot safely assume
      // the nodes are syntactically interchangeable.
      return false;
    }

    return findChildReprints(newPath, oldPath, reprints);
  }

  function findChildReprints(newPath, oldPath, reprints) {
    var newNode = newPath.getValue();
    var oldNode = oldPath.getValue();

    isObject.assert(newNode);
    isObject.assert(oldNode);

    if (newNode.original === null) {
      // If newNode.original node was set to null, reprint the node.
      return false;
    }

    // If this node needs parentheses and will not be wrapped with
    // parentheses when reprinted, then return false to skip reprinting and
    // let it be printed generically.
    if (newPath.needsParens() &&
        ! oldPath.hasParens()) {
      return false;
    }

    var keys = util_1$1.getUnionOfKeys(oldNode, newNode);

    if (oldNode.type === "File" ||
        newNode.type === "File") {
      // Don't bother traversing file.tokens, an often very large array
      // returned by Babylon, and useless for our purposes.
      delete keys.tokens;
    }

    // Don't bother traversing .loc objects looking for reprintable nodes.
    delete keys.loc;

    var originalReprintCount = reprints.length;

    for (var k in keys) {
      if (k.charAt(0) === "_") {
        // Ignore "private" AST properties added by e.g. Babel plugins and
        // parsers like Babylon.
        continue;
      }

      newPath.stack.push(k, types$1.getFieldValue(newNode, k));
      oldPath.stack.push(k, types$1.getFieldValue(oldNode, k));
      var canReprint = findAnyReprints(newPath, oldPath, reprints);
      newPath.stack.length -= 2;
      oldPath.stack.length -= 2;

      if (!canReprint) {
        return false;
      }
    }

    // Return statements might end up running into ASI issues due to
    // comments inserted deep within the tree, so reprint them if anything
    // changed within them.
    if (ReturnStatement.check(newPath.getNode()) &&
        reprints.length > originalReprintCount) {
      return false;
    }

    return true;
  }

  var patcher = {
  	Patcher: Patcher_1,
  	getReprinter: getReprinter
  };

  var n$1 = types$1.namedTypes;
  var isArray$2 = types$1.builtInTypes.array;
  var isObject$1 = types$1.builtInTypes.object;
  var Lines$1 = lines.Lines;
  var concat$1 = lines.concat;

  var comparePos$3 = util_1$1.comparePos;
  var childNodesCacheKey = _private.makeUniqueKey();

  // TODO Move a non-caching implementation of this function into ast-types,
  // and implement a caching wrapper function here.
  function getSortedChildNodes(node, lines$$1, resultArray) {
      if (!node) {
          return;
      }

      // The .loc checks below are sensitive to some of the problems that
      // are fixed by this utility function. Specifically, if it decides to
      // set node.loc to null, indicating that the node's .loc information
      // is unreliable, then we don't want to add node to the resultArray.
      util_1$1.fixFaultyLocations(node, lines$$1);

      if (resultArray) {
          if (n$1.Node.check(node) &&
              n$1.SourceLocation.check(node.loc)) {
              // This reverse insertion sort almost always takes constant
              // time because we almost always (maybe always?) append the
              // nodes in order anyway.
              for (var i = resultArray.length - 1; i >= 0; --i) {
                  if (comparePos$3(resultArray[i].loc.end,
                                 node.loc.start) <= 0) {
                      break;
                  }
              }
              resultArray.splice(i + 1, 0, node);
              return;
          }
      } else if (node[childNodesCacheKey]) {
          return node[childNodesCacheKey];
      }

      var names;
      if (isArray$2.check(node)) {
          names = Object.keys(node);
      } else if (isObject$1.check(node)) {
          names = types$1.getFieldNames(node);
      } else {
          return;
      }

      if (!resultArray) {
          Object.defineProperty(node, childNodesCacheKey, {
              value: resultArray = [],
              enumerable: false
          });
      }

      for (var i = 0, nameCount = names.length; i < nameCount; ++i) {
          getSortedChildNodes(node[names[i]], lines$$1, resultArray);
      }

      return resultArray;
  }

  // As efficiently as possible, decorate the comment object with
  // .precedingNode, .enclosingNode, and/or .followingNode properties, at
  // least one of which is guaranteed to be defined.
  function decorateComment(node, comment, lines$$1) {
      var childNodes = getSortedChildNodes(node, lines$$1);

      // Time to dust off the old binary search robes and wizard hat.
      var left = 0, right = childNodes.length;
      while (left < right) {
          var middle = (left + right) >> 1;
          var child = childNodes[middle];

          if (comparePos$3(child.loc.start, comment.loc.start) <= 0 &&
              comparePos$3(comment.loc.end, child.loc.end) <= 0) {
              // The comment is completely contained by this child node.
              decorateComment(comment.enclosingNode = child, comment, lines$$1);
              return; // Abandon the binary search at this level.
          }

          if (comparePos$3(child.loc.end, comment.loc.start) <= 0) {
              // This child node falls completely before the comment.
              // Because we will never consider this node or any nodes
              // before it again, this node must be the closest preceding
              // node we have encountered so far.
              var precedingNode = child;
              left = middle + 1;
              continue;
          }

          if (comparePos$3(comment.loc.end, child.loc.start) <= 0) {
              // This child node falls completely after the comment.
              // Because we will never consider this node or any nodes after
              // it again, this node must be the closest following node we
              // have encountered so far.
              var followingNode = child;
              right = middle;
              continue;
          }

          throw new Error("Comment location overlaps with node location");
      }

      if (precedingNode) {
          comment.precedingNode = precedingNode;
      }

      if (followingNode) {
          comment.followingNode = followingNode;
      }
  }

  var attach = function(comments, ast, lines$$1) {
      if (!isArray$2.check(comments)) {
          return;
      }

      var tiesToBreak = [];

      comments.forEach(function(comment) {
          comment.loc.lines = lines$$1;
          decorateComment(ast, comment, lines$$1);

          var pn = comment.precedingNode;
          var en = comment.enclosingNode;
          var fn = comment.followingNode;

          if (pn && fn) {
              var tieCount = tiesToBreak.length;
              if (tieCount > 0) {
                  var lastTie = tiesToBreak[tieCount - 1];

                  if (lastTie.followingNode !== comment.followingNode) {
                      breakTies(tiesToBreak, lines$$1);
                  }
              }

              tiesToBreak.push(comment);

          } else if (pn) {
              // No contest: we have a trailing comment.
              breakTies(tiesToBreak, lines$$1);
              addTrailingComment(pn, comment);

          } else if (fn) {
              // No contest: we have a leading comment.
              breakTies(tiesToBreak, lines$$1);
              addLeadingComment(fn, comment);

          } else if (en) {
              // The enclosing node has no child nodes at all, so what we
              // have here is a dangling comment, e.g. [/* crickets */].
              breakTies(tiesToBreak, lines$$1);
              addDanglingComment(en, comment);

          } else {
              throw new Error("AST contains no nodes at all?");
          }
      });

      breakTies(tiesToBreak, lines$$1);

      comments.forEach(function(comment) {
          // These node references were useful for breaking ties, but we
          // don't need them anymore, and they create cycles in the AST that
          // may lead to infinite recursion if we don't delete them here.
          delete comment.precedingNode;
          delete comment.enclosingNode;
          delete comment.followingNode;
      });
  };

  function breakTies(tiesToBreak, lines$$1) {
      var tieCount = tiesToBreak.length;
      if (tieCount === 0) {
          return;
      }

      var pn = tiesToBreak[0].precedingNode;
      var fn = tiesToBreak[0].followingNode;
      var gapEndPos = fn.loc.start;

      // Iterate backwards through tiesToBreak, examining the gaps
      // between the tied comments. In order to qualify as leading, a
      // comment must be separated from fn by an unbroken series of
      // whitespace-only gaps (or other comments).
      for (var indexOfFirstLeadingComment = tieCount;
           indexOfFirstLeadingComment > 0;
           --indexOfFirstLeadingComment) {
          var comment = tiesToBreak[indexOfFirstLeadingComment - 1];

          var gap = lines$$1.sliceString(comment.loc.end, gapEndPos);
          if (/\S/.test(gap)) {
              // The gap string contained something other than whitespace.
              break;
          }

          gapEndPos = comment.loc.start;
      }

      while (indexOfFirstLeadingComment <= tieCount &&
             (comment = tiesToBreak[indexOfFirstLeadingComment]) &&
             // If the comment is a //-style comment and indented more
             // deeply than the node itself, reconsider it as trailing.
             (comment.type === "Line" || comment.type === "CommentLine") &&
             comment.loc.start.column > fn.loc.start.column) {
          ++indexOfFirstLeadingComment;
      }

      tiesToBreak.forEach(function(comment, i) {
          if (i < indexOfFirstLeadingComment) {
              addTrailingComment(pn, comment);
          } else {
              addLeadingComment(fn, comment);
          }
      });

      tiesToBreak.length = 0;
  }

  function addCommentHelper(node, comment) {
      var comments = node.comments || (node.comments = []);
      comments.push(comment);
  }

  function addLeadingComment(node, comment) {
      comment.leading = true;
      comment.trailing = false;
      addCommentHelper(node, comment);
  }

  function addDanglingComment(node, comment) {
      comment.leading = false;
      comment.trailing = false;
      addCommentHelper(node, comment);
  }

  function addTrailingComment(node, comment) {
      comment.leading = false;
      comment.trailing = true;
      addCommentHelper(node, comment);
  }

  function printLeadingComment(commentPath, print) {
      var comment = commentPath.getValue();
      n$1.Comment.assert(comment);

      var loc = comment.loc;
      var lines$$1 = loc && loc.lines;
      var parts = [print(commentPath)];

      if (comment.trailing) {
          // When we print trailing comments as leading comments, we don't
          // want to bring any trailing spaces along.
          parts.push("\n");

      } else if (lines$$1 instanceof Lines$1) {
          var trailingSpace = lines$$1.slice(
              loc.end,
              lines$$1.skipSpaces(loc.end)
          );

          if (trailingSpace.length === 1) {
              // If the trailing space contains no newlines, then we want to
              // preserve it exactly as we found it.
              parts.push(trailingSpace);
          } else {
              // If the trailing space contains newlines, then replace it
              // with just that many newlines, with all other spaces removed.
              parts.push(new Array(trailingSpace.length).join("\n"));
          }

      } else {
          parts.push("\n");
      }

      return concat$1(parts);
  }

  function printTrailingComment(commentPath, print) {
      var comment = commentPath.getValue(commentPath);
      n$1.Comment.assert(comment);

      var loc = comment.loc;
      var lines$$1 = loc && loc.lines;
      var parts = [];

      if (lines$$1 instanceof Lines$1) {
          var fromPos = lines$$1.skipSpaces(loc.start, true) || lines$$1.firstPos();
          var leadingSpace = lines$$1.slice(fromPos, loc.start);

          if (leadingSpace.length === 1) {
              // If the leading space contains no newlines, then we want to
              // preserve it exactly as we found it.
              parts.push(leadingSpace);
          } else {
              // If the leading space contains newlines, then replace it
              // with just that many newlines, sans all other spaces.
              parts.push(new Array(leadingSpace.length).join("\n"));
          }
      }

      parts.push(print(commentPath));

      return concat$1(parts);
  }

  var printComments = function(path, print) {
      var value = path.getValue();
      var innerLines = print(path);
      var comments = n$1.Node.check(value) &&
          types$1.getFieldValue(value, "comments");

      if (!comments || comments.length === 0) {
          return innerLines;
      }

      var leadingParts = [];
      var trailingParts = [innerLines];

      path.each(function(commentPath) {
          var comment = commentPath.getValue();
          var leading = types$1.getFieldValue(comment, "leading");
          var trailing = types$1.getFieldValue(comment, "trailing");

          if (leading || (trailing && !(n$1.Statement.check(value) ||
                                        comment.type === "Block" ||
                                        comment.type === "CommentBlock"))) {
              leadingParts.push(printLeadingComment(commentPath, print));
          } else if (trailing) {
              trailingParts.push(printTrailingComment(commentPath, print));
          }
      }, "comments");

      leadingParts.push.apply(leadingParts, trailingParts);
      return concat$1(leadingParts);
  };

  var comments = {
  	attach: attach,
  	printComments: printComments
  };

  var n$2 = types$1.namedTypes;
  var b = types$1.builders;
  var isObject$2 = types$1.builtInTypes.object;
  var isArray$3 = types$1.builtInTypes.array;
  var isFunction$1 = types$1.builtInTypes.function;
  var normalizeOptions$1 = options.normalize;
  var fromString$2 = lines.fromString;
  var attachComments = comments.attach;


  var parse$1 = function parse(source, options$$1) {
    options$$1 = normalizeOptions$1(options$$1);

    const lines$$1 = fromString$2(source, options$$1);

    const sourceWithoutTabs = lines$$1.toString({
      tabWidth: options$$1.tabWidth,
      reuseWhitespace: false,
      useTabs: false
    });

    let comments$$1 = [];
    const ast = options$$1.parser.parse(sourceWithoutTabs, {
      jsx: true,
      loc: true,
      locations: true,
      range: options$$1.range,
      comment: true,
      onComment: comments$$1,
      tolerant: util_1$1.getOption(options$$1, "tolerant", true),
      ecmaVersion: 6,
      sourceType: util_1$1.getOption(options$$1, "sourceType", "module")
    });

    // Use ast.tokens if possible, and otherwise fall back to the Esprima
    // tokenizer. All the preconfigured ../parsers/* expose ast.tokens
    // automatically, but custom parsers might need additional configuration
    // to avoid this fallback.
    const tokens = Array.isArray(ast.tokens)
      ? ast.tokens
      : esprima$1.tokenize(sourceWithoutTabs, {
          loc: true
        });

    // We will reattach the tokens array to the file object below.
    delete ast.tokens;

    // Make sure every token has a token.value string.
    tokens.forEach(function (token) {
      if (typeof token.value !== "string") {
        token.value = lines$$1.sliceString(token.loc.start, token.loc.end);
      }
    });

    if (Array.isArray(ast.comments)) {
      comments$$1 = ast.comments;
      delete ast.comments;
    }

    if (ast.loc) {
      // If the source was empty, some parsers give loc.{start,end}.line
      // values of 0, instead of the minimum of 1.
      util_1$1.fixFaultyLocations(ast, lines$$1);
    } else {
      ast.loc = {
        start: lines$$1.firstPos(),
        end: lines$$1.lastPos()
      };
    }

    ast.loc.lines = lines$$1;
    ast.loc.indent = 0;

    let file;
    let program;
    if (ast.type === "Program") {
      program = ast;
      // In order to ensure we reprint leading and trailing program
      // comments, wrap the original Program node with a File node. Only
      // ESTree parsers (Acorn and Esprima) return a Program as the root AST
      // node. Most other (Babylon-like) parsers return a File.
      file = b.file(ast, options$$1.sourceFileName || null);
      file.loc = {
        start: lines$$1.firstPos(),
        end: lines$$1.lastPos(),
        lines: lines$$1,
        indent: 0
      };
    } else if (ast.type === "File") {
      file = ast;
      program = file.program;
    }

    // Expose file.tokens unless the caller passed false for options.tokens.
    if (options$$1.tokens) {
      file.tokens = tokens;
    }

    // Expand the Program's .loc to include all comments (not just those
    // attached to the Program node, as its children may have comments as
    // well), since sometimes program.loc.{start,end} will coincide with the
    // .loc.{start,end} of the first and last *statements*, mistakenly
    // excluding comments that fall outside that region.
    var trueProgramLoc = util_1$1.getTrueLoc({
      type: program.type,
      loc: program.loc,
      body: [],
      comments: comments$$1
    }, lines$$1);
    program.loc.start = trueProgramLoc.start;
    program.loc.end = trueProgramLoc.end;

    // Passing file.program here instead of just file means that initial
    // comments will be attached to program.body[0] instead of program.
    attachComments(
      comments$$1,
      program.body.length ? file.program : file,
      lines$$1
    );

    // Return a copy of the original AST so that any changes made may be
    // compared to the original.
    return new TreeCopier(lines$$1, tokens).copy(file);
  };

  function TreeCopier(lines$$1, tokens) {
    this.lines = lines$$1;
    this.tokens = tokens;
    this.startTokenIndex = 0;
    this.endTokenIndex = tokens.length;
    this.indent = 0;
    this.seen = new Map;
  }

  var TCp = TreeCopier.prototype;

  TCp.copy = function(node) {
    if (this.seen.has(node)) {
      return this.seen.get(node);
    }

    if (isArray$3.check(node)) {
      var copy = new Array(node.length);
      this.seen.set(node, copy);
      node.forEach(function (item, i) {
        copy[i] = this.copy(item);
      }, this);
      return copy;
    }

    if (!isObject$2.check(node)) {
      return node;
    }

    util_1$1.fixFaultyLocations(node, this.lines);

    var copy = Object.create(Object.getPrototypeOf(node), {
      original: { // Provide a link from the copy to the original.
        value: node,
        configurable: false,
        enumerable: false,
        writable: true
      }
    });

    this.seen.set(node, copy);

    var loc = node.loc;
    var oldIndent = this.indent;
    var newIndent = oldIndent;

    const oldStartTokenIndex = this.startTokenIndex;
    const oldEndTokenIndex = this.endTokenIndex;

    if (loc) {
      // When node is a comment, we set node.loc.indent to
      // node.loc.start.column so that, when/if we print the comment by
      // itself, we can strip that much whitespace from the left margin of
      // the comment. This only really matters for multiline Block comments,
      // but it doesn't hurt for Line comments.
      if (node.type === "Block" || node.type === "Line" ||
          node.type === "CommentBlock" || node.type === "CommentLine" ||
          this.lines.isPrecededOnlyByWhitespace(loc.start)) {
        newIndent = this.indent = loc.start.column;
      }

      // Every node.loc has a reference to the original source lines as well
      // as a complete list of source tokens.
      loc.lines = this.lines;
      loc.tokens = this.tokens;
      loc.indent = newIndent;

      // Set loc.start.token and loc.end.token such that
      // loc.tokens.slice(loc.start.token, loc.end.token) returns a list of
      // all the tokens that make up this node.
      this.findTokenRange(loc);
    }

    var keys = Object.keys(node);
    var keyCount = keys.length;
    for (var i = 0; i < keyCount; ++i) {
      var key = keys[i];
      if (key === "loc") {
        copy[key] = node[key];
      } else if (key === "tokens" &&
                 node.type === "File") {
        // Preserve file.tokens (uncopied) in case client code cares about
        // it, even though Recast ignores it when reprinting.
        copy[key] = node[key];
      } else {
        copy[key] = this.copy(node[key]);
      }
    }

    this.indent = oldIndent;
    this.startTokenIndex = oldStartTokenIndex;
    this.endTokenIndex = oldEndTokenIndex;

    return copy;
  };

  // If we didn't have any idea where in loc.tokens to look for tokens
  // contained by this loc, a binary search would be appropriate, but
  // because we maintain this.startTokenIndex and this.endTokenIndex as we
  // traverse the AST, we only need to make small (linear) adjustments to
  // those indexes with each recursive iteration.
  TCp.findTokenRange = function (loc) {
    // In the unlikely event that loc.tokens[this.startTokenIndex] starts
    // *after* loc.start, we need to rewind this.startTokenIndex first.
    while (this.startTokenIndex > 0) {
      const token = loc.tokens[this.startTokenIndex];
      if (util_1$1.comparePos(loc.start, token.loc.start) < 0) {
        --this.startTokenIndex;
      } else break;
    }

    // In the unlikely event that loc.tokens[this.endTokenIndex - 1] ends
    // *before* loc.end, we need to fast-forward this.endTokenIndex first.
    while (this.endTokenIndex < loc.tokens.length) {
      const token = loc.tokens[this.endTokenIndex];
      if (util_1$1.comparePos(token.loc.end, loc.end) < 0) {
        ++this.endTokenIndex;
      } else break;
    }

    // Increment this.startTokenIndex until we've found the first token
    // contained by this node.
    while (this.startTokenIndex < this.endTokenIndex) {
      const token = loc.tokens[this.startTokenIndex];
      if (util_1$1.comparePos(token.loc.start, loc.start) < 0) {
        ++this.startTokenIndex;
      } else break;
    }

    // Index into loc.tokens of the first token within this node.
    loc.start.token = this.startTokenIndex;

    // Decrement this.endTokenIndex until we've found the first token after
    // this node (not contained by the node).
    while (this.endTokenIndex > this.startTokenIndex) {
      const token = loc.tokens[this.endTokenIndex - 1];
      if (util_1$1.comparePos(loc.end, token.loc.end) < 0) {
        --this.endTokenIndex;
      } else break;
    }

    // Index into loc.tokens of the first token *after* this node.
    // If loc.start.token === loc.end.token, the node contains no tokens,
    // and the index is that of the next token following this node.
    loc.end.token = this.endTokenIndex;
  };

  var parser = {
  	parse: parse$1
  };

  var printComments$1 = comments.printComments;

  var fromString$3 = lines.fromString;
  var concat$2 = lines.concat;
  var normalizeOptions$2 = options.normalize;
  var getReprinter$1 = patcher.getReprinter;

  var namedTypes = types$1.namedTypes;
  var isString$4 = types$1.builtInTypes.string;
  var isObject$3 = types$1.builtInTypes.object;



  function PrintResult(code, sourceMap$$1) {

      isString$4.assert(code);
      this.code = code;

      if (sourceMap$$1) {
          isObject$3.assert(sourceMap$$1);
          this.map = sourceMap$$1;
      }
  }

  var PRp = PrintResult.prototype;
  var warnedAboutToString = false;

  PRp.toString = function() {
      if (!warnedAboutToString) {
          console.warn(
              "Deprecation warning: recast.print now returns an object with " +
              "a .code property. You appear to be treating the object as a " +
              "string, which might still work but is strongly discouraged."
          );

          warnedAboutToString = true;
      }

      return this.code;
  };

  var emptyPrintResult = new PrintResult("");

  function Printer(config) {

      const explicitTabWidth = config && config.tabWidth;
      config = normalizeOptions$2(config);

      // It's common for client code to pass the same options into both
      // recast.parse and recast.print, but the Printer doesn't need (and
      // can be confused by) config.sourceFileName, so we null it out.
      config.sourceFileName = null;

      // Non-destructively modifies options with overrides, and returns a
      // new print function that uses the modified options.
      function makePrintFunctionWith(options$$1, overrides) {
          options$$1 = Object.assign({}, options$$1, overrides);
          return function (path) {
              return print(path, options$$1);
          };
      }

      function print(path, options$$1) {
          options$$1 = options$$1 || {};

          if (options$$1.includeComments) {
              return printComments$1(path, makePrintFunctionWith(options$$1, {
                  includeComments: false
              }));
          }

          const oldTabWidth = config.tabWidth;

          if (! explicitTabWidth) {
              const loc = path.getNode().loc;
              if (loc && loc.lines && loc.lines.guessTabWidth) {
                  config.tabWidth = loc.lines.guessTabWidth();
              }
          }

          const reprinter = getReprinter$1(path);
          const lines$$1 = reprinter
              // Since the print function that we pass to the reprinter will
              // be used to print "new" nodes, it's tempting to think we
              // should pass printRootGenerically instead of print, to avoid
              // calling maybeReprint again, but that would be a mistake
              // because the new nodes might not be entirely new, but merely
              // moved from elsewhere in the AST. The print function is the
              // right choice because it gives us the opportunity to reprint
              // such nodes using their original source.
              ? reprinter(print)
              : genericPrint(
                  path,
                  config,
                  options$$1,
                  makePrintFunctionWith(options$$1, {
                      includeComments: true,
                      avoidRootParens: false
                  })
              );

          config.tabWidth = oldTabWidth;

          return lines$$1;
      }

      this.print = function(ast) {
          if (!ast) {
              return emptyPrintResult;
          }

          const lines$$1 = print(fastPath.from(ast), {
              includeComments: true,
              avoidRootParens: false
          });

          return new PrintResult(
              lines$$1.toString(config),
              util_1$1.composeSourceMaps(
                  config.inputSourceMap,
                  lines$$1.getSourceMap(
                      config.sourceMapName,
                      config.sourceRoot
                  )
              )
          );
      };

      this.printGenerically = function(ast) {
          if (!ast) {
              return emptyPrintResult;
          }

          // Print the entire AST generically.
          function printGenerically(path) {
              return printComments$1(path, function (path) {
                  return genericPrint(path, config, {
                      includeComments: true,
                      avoidRootParens: false
                  }, printGenerically);
              });
          }

          const path = fastPath.from(ast);
          const oldReuseWhitespace = config.reuseWhitespace;

          // Do not reuse whitespace (or anything else, for that matter)
          // when printing generically.
          config.reuseWhitespace = false;

          // TODO Allow printing of comments?
          const pr = new PrintResult(printGenerically(path).toString(config));
          config.reuseWhitespace = oldReuseWhitespace;
          return pr;
      };
  }

  var Printer_1 = Printer;

  function genericPrint(path, config, options$$1, printPath) {

      const node = path.getValue();
      const parts = [];
      const linesWithoutParens =
          genericPrintNoParens(path, config, printPath);

      if (! node || linesWithoutParens.isEmpty()) {
          return linesWithoutParens;
      }

      let shouldAddParens = false;
      const decoratorsLines = printDecorators(path, printPath);

      if (decoratorsLines.isEmpty()) {
          // Nodes with decorators can't have parentheses, so we can avoid
          // computing path.needsParens() except in this case.
          if (! options$$1.avoidRootParens) {
              shouldAddParens = path.needsParens();
          }
      } else {
          parts.push(decoratorsLines);
      }

      if (shouldAddParens) {
          parts.unshift("(");
      }

      parts.push(linesWithoutParens);

      if (shouldAddParens) {
          parts.push(")");
      }

      return concat$2(parts);
  }

  // Note that the `options` parameter of this function is what other
  // functions in this file call the `config` object (that is, the
  // configuration object originally passed into the Printer constructor).
  // Its properties are documented in lib/options.js.
  function genericPrintNoParens(path, options$$1, print) {
      var n = path.getValue();

      if (!n) {
          return fromString$3("");
      }

      if (typeof n === "string") {
          return fromString$3(n, options$$1);
      }

      namedTypes.Printable.assert(n);

      var parts = [];

      switch (n.type) {
      case "File":
          return path.call(print, "program");

      case "Program":
          // Babel 6
          if (n.directives) {
              path.each(function(childPath) {
                  parts.push(print(childPath), ";\n");
              }, "directives");
          }

          parts.push(path.call(function(bodyPath) {
              return printStatementSequence(bodyPath, options$$1, print);
          }, "body"));

          return concat$2(parts);

      case "Noop": // Babel extension.
      case "EmptyStatement":
          return fromString$3("");

      case "ExpressionStatement":
          return concat$2([path.call(print, "expression"), ";"]);

      case "ParenthesizedExpression": // Babel extension.
          return concat$2(["(", path.call(print, "expression"), ")"]);

      case "BinaryExpression":
      case "LogicalExpression":
      case "AssignmentExpression":
          return fromString$3(" ").join([
              path.call(print, "left"),
              n.operator,
              path.call(print, "right")
          ]);

      case "AssignmentPattern":
          return concat$2([
              path.call(print, "left"),
              " = ",
              path.call(print, "right")
          ]);

      case "MemberExpression":
      case "OptionalMemberExpression":
          parts.push(path.call(print, "object"));

          var property = path.call(print, "property");
          var optional = n.type === "OptionalMemberExpression";

          if (n.computed) {
              parts.push(optional ? "?.[" : "[", property, "]");
          } else {
              parts.push(optional ? "?." : ".", property);
          }

          return concat$2(parts);

      case "MetaProperty":
          return concat$2([
              path.call(print, "meta"),
              ".",
              path.call(print, "property")
          ]);

      case "BindExpression":
          if (n.object) {
              parts.push(path.call(print, "object"));
          }

          parts.push("::", path.call(print, "callee"));

          return concat$2(parts);

      case "Path":
          return fromString$3(".").join(n.body);

      case "Identifier":
          return concat$2([
              fromString$3(n.name, options$$1),
              n.optional ? "?" : "",
              path.call(print, "typeAnnotation")
          ]);

      case "SpreadElement":
      case "SpreadElementPattern":
      case "RestProperty": // Babel 6 for ObjectPattern
      case "SpreadProperty":
      case "SpreadPropertyPattern":
      case "ObjectTypeSpreadProperty":
      case "RestElement":
          return concat$2([
              "...",
              path.call(print, "argument"),
              path.call(print, "typeAnnotation")
          ]);

      case "FunctionDeclaration":
      case "FunctionExpression":
      case "TSDeclareFunction":
          if (n.declare) {
              parts.push("declare ");
          }

          if (n.async) {
              parts.push("async ");
          }

          parts.push("function");

          if (n.generator)
              parts.push("*");

          if (n.id) {
              parts.push(
                  " ",
                  path.call(print, "id"),
                  path.call(print, "typeParameters")
              );
          }

          parts.push(
              "(",
              printFunctionParams(path, options$$1, print),
              ")",
              path.call(print, "returnType")
          );

          if (n.body) {
              parts.push(" ", path.call(print, "body"));
          }

          return concat$2(parts);

      case "ArrowFunctionExpression":
          if (n.async) {
              parts.push("async ");
          }

          if (n.typeParameters) {
              parts.push(path.call(print, "typeParameters"));
          }

          if (! options$$1.arrowParensAlways &&
              n.params.length === 1 &&
              ! n.rest &&
              n.params[0].type === 'Identifier' &&
              ! n.params[0].typeAnnotation &&
              ! n.returnType) {
              parts.push(path.call(print, "params", 0));
          } else {
              parts.push(
                  "(",
                  printFunctionParams(path, options$$1, print),
                  ")",
                  path.call(print, "returnType")
              );
          }

          parts.push(" => ", path.call(print, "body"));

          return concat$2(parts);

      case "MethodDefinition":
          return printMethod(path, options$$1, print);

      case "YieldExpression":
          parts.push("yield");

          if (n.delegate)
              parts.push("*");

          if (n.argument)
              parts.push(" ", path.call(print, "argument"));

          return concat$2(parts);

      case "AwaitExpression":
          parts.push("await");

          if (n.all)
              parts.push("*");

          if (n.argument)
              parts.push(" ", path.call(print, "argument"));

          return concat$2(parts);

      case "ModuleDeclaration":
          parts.push("module", path.call(print, "id"));

          if (n.source) {
              parts.push("from", path.call(print, "source"));
          } else {
              parts.push(path.call(print, "body"));
          }

          return fromString$3(" ").join(parts);

      case "ImportSpecifier":
          if (n.importKind && n.importKind !== "value") {
              parts.push(n.importKind + " ");
          }
          if (n.imported) {
              parts.push(path.call(print, "imported"));
              if (n.local &&
                  n.local.name !== n.imported.name) {
                  parts.push(" as ", path.call(print, "local"));
              }
          } else if (n.id) {
              parts.push(path.call(print, "id"));
              if (n.name) {
                  parts.push(" as ", path.call(print, "name"));
              }
          }

          return concat$2(parts);

      case "ExportSpecifier":
          if (n.local) {
              parts.push(path.call(print, "local"));
              if (n.exported &&
                  n.exported.name !== n.local.name) {
                  parts.push(" as ", path.call(print, "exported"));
              }
          } else if (n.id) {
              parts.push(path.call(print, "id"));
              if (n.name) {
                  parts.push(" as ", path.call(print, "name"));
              }
          }

          return concat$2(parts);

      case "ExportBatchSpecifier":
          return fromString$3("*");

      case "ImportNamespaceSpecifier":
          parts.push("* as ");
          if (n.local) {
              parts.push(path.call(print, "local"));
          } else if (n.id) {
              parts.push(path.call(print, "id"));
          }
          return concat$2(parts);

      case "ImportDefaultSpecifier":
          if (n.local) {
              return path.call(print, "local");
          }
          return path.call(print, "id");

      case "TSExportAssignment":
          return concat$2(["export = ", path.call(print, "expression")]);

      case "ExportDeclaration":
      case "ExportDefaultDeclaration":
      case "ExportNamedDeclaration":
          return printExportDeclaration(path, options$$1, print);

      case "ExportAllDeclaration":
          parts.push("export *");

          if (n.exported) {
              parts.push(" as ", path.call(print, "exported"));
          }

          parts.push(
              " from ",
              path.call(print, "source")
          );

          return concat$2(parts);

      case "TSNamespaceExportDeclaration":
          parts.push("export as namespace ", path.call(print, "id"));
          return maybeAddSemicolon(concat$2(parts));

      case "ExportNamespaceSpecifier":
          return concat$2(["* as ", path.call(print, "exported")]);

      case "ExportDefaultSpecifier":
          return path.call(print, "exported");

      case "Import":
          return fromString$3("import", options$$1);

      case "ImportDeclaration": {
          parts.push("import ");

          if (n.importKind && n.importKind !== "value") {
              parts.push(n.importKind + " ");
          }

          if (n.specifiers &&
              n.specifiers.length > 0) {

              const unbracedSpecifiers = [];
              const bracedSpecifiers = [];

              path.each(function (specifierPath) {
                  const spec = specifierPath.getValue();
                  if (spec.type === "ImportSpecifier") {
                      bracedSpecifiers.push(print(specifierPath));
                  } else if (spec.type === "ImportDefaultSpecifier" ||
                             spec.type === "ImportNamespaceSpecifier") {
                      unbracedSpecifiers.push(print(specifierPath));
                  }
              }, "specifiers");

              unbracedSpecifiers.forEach((lines$$1, i) => {
                  if (i > 0) {
                      parts.push(", ");
                  }
                  parts.push(lines$$1);
              });

              if (bracedSpecifiers.length > 0) {
                  let lines$$1 = fromString$3(", ").join(bracedSpecifiers);
                  if (lines$$1.getLineLength(1) > options$$1.wrapColumn) {
                      lines$$1 = concat$2([
                          fromString$3(",\n").join(
                              bracedSpecifiers
                          ).indent(options$$1.tabWidth),
                          ","
                      ]);
                  }

                  if (unbracedSpecifiers.length > 0) {
                      parts.push(", ");
                  }

                  if (lines$$1.length > 1) {
                      parts.push("{\n", lines$$1, "\n}");
                  } else if (options$$1.objectCurlySpacing) {
                      parts.push("{ ", lines$$1, " }");
                  } else {
                      parts.push("{", lines$$1, "}");
                  }
              }

              parts.push(" from ");
          }

          parts.push(path.call(print, "source"), ";");

          return concat$2(parts);
      }

      case "BlockStatement":
          var naked = path.call(function(bodyPath) {
              return printStatementSequence(bodyPath, options$$1, print);
          }, "body");


          if (naked.isEmpty()) {
              if (!n.directives || n.directives.length === 0) {
                  return fromString$3("{}");
              }
          }

          parts.push("{\n");
          // Babel 6
          if (n.directives) {
              path.each(function(childPath) {
                  parts.push(
                      print(childPath).indent(options$$1.tabWidth),
                      ";",
                      n.directives.length > 1 || !naked.isEmpty() ? "\n" : ""
                  );
              }, "directives");
          }
          parts.push(naked.indent(options$$1.tabWidth));
          parts.push("\n}");

          return concat$2(parts);

      case "ReturnStatement":
          parts.push("return");

          if (n.argument) {
              var argLines = path.call(print, "argument");
              if (argLines.startsWithComment() ||
                  (argLines.length > 1 &&
                      namedTypes.JSXElement &&
                      namedTypes.JSXElement.check(n.argument)
                  )) {
                  parts.push(
                      " (\n",
                      argLines.indent(options$$1.tabWidth),
                      "\n)"
                  );
              } else {
                  parts.push(" ", argLines);
              }
          }

          parts.push(";");

          return concat$2(parts);

      case "CallExpression":
      case "OptionalCallExpression":
          var parts = [path.call(print, "callee")];

          if (n.type === "OptionalCallExpression" &&
              n.callee.type !== "OptionalMemberExpression") {
              parts.push("?.");
          }

          parts.push(printArgumentsList(path, options$$1, print));

          return concat$2(parts);

      case "ObjectExpression":
      case "ObjectPattern":
      case "ObjectTypeAnnotation":
          var allowBreak = false;
          var isTypeAnnotation = n.type === "ObjectTypeAnnotation";
          var separator = options$$1.flowObjectCommas ? "," : (isTypeAnnotation ? ";" : ",");
          var fields = [];

          if (isTypeAnnotation) {
              fields.push("indexers", "callProperties");
          }

          fields.push("properties");

          var len = 0;
          fields.forEach(function(field) {
              len += n[field].length;
          });

          var oneLine = (isTypeAnnotation && len === 1) || len === 0;
          var leftBrace = n.exact ? "{|" : "{";
          var rightBrace = n.exact ? "|}" : "}";
          parts.push(oneLine ? leftBrace : leftBrace + "\n");
          var leftBraceIndex = parts.length - 1;

          var i = 0;
          fields.forEach(function(field) {
              path.each(function(childPath) {
                  var lines$$1 = print(childPath);

                  if (!oneLine) {
                      lines$$1 = lines$$1.indent(options$$1.tabWidth);
                  }

                  var multiLine = !isTypeAnnotation && lines$$1.length > 1;
                  if (multiLine && allowBreak) {
                      // Similar to the logic for BlockStatement.
                      parts.push("\n");
                  }

                  parts.push(lines$$1);

                  if (i < len - 1) {
                      // Add an extra line break if the previous object property
                      // had a multi-line value.
                      parts.push(separator + (multiLine ? "\n\n" : "\n"));
                      allowBreak = !multiLine;
                  } else if (len !== 1 && isTypeAnnotation) {
                      parts.push(separator);
                  } else if (!oneLine && util_1$1.isTrailingCommaEnabled(options$$1, "objects")) {
                      parts.push(separator);
                  }
                  i++;
              }, field);
          });

          parts.push(oneLine ? rightBrace : "\n" + rightBrace);

          if (i !== 0 && oneLine && options$$1.objectCurlySpacing) {
              parts[leftBraceIndex] = leftBrace + " ";
              parts[parts.length - 1] = " " + rightBrace;
          }

          if (n.typeAnnotation) {
              parts.push(path.call(print, "typeAnnotation"));
          }

          return concat$2(parts);

      case "PropertyPattern":
          return concat$2([
              path.call(print, "key"),
              ": ",
              path.call(print, "pattern")
          ]);

      case "ObjectProperty": // Babel 6
      case "Property": // Non-standard AST node type.
          if (n.method || n.kind === "get" || n.kind === "set") {
              return printMethod(path, options$$1, print);
          }

          var key = path.call(print, "key");
          if (n.computed) {
              parts.push("[", key, "]");
          } else {
              parts.push(key);
          }

          if (! n.shorthand) {
              parts.push(": ", path.call(print, "value"));
          }

          return concat$2(parts);

      case "ClassMethod": // Babel 6
      case "ObjectMethod": // Babel 6
      case "TSDeclareMethod":
          return printMethod(path, options$$1, print);

      case "Decorator":
          return concat$2(["@", path.call(print, "expression")]);

      case "ArrayExpression":
      case "ArrayPattern":
          var elems = n.elements,
              len = elems.length;

          var printed = path.map(print, "elements");
          var joined = fromString$3(", ").join(printed);
          var oneLine = joined.getLineLength(1) <= options$$1.wrapColumn;
          if (oneLine) {
            if (options$$1.arrayBracketSpacing) {
              parts.push("[ ");
            } else {
              parts.push("[");
            }
          } else {
            parts.push("[\n");
          }

          path.each(function(elemPath) {
              var i = elemPath.getName();
              var elem = elemPath.getValue();
              if (!elem) {
                  // If the array expression ends with a hole, that hole
                  // will be ignored by the interpreter, but if it ends with
                  // two (or more) holes, we need to write out two (or more)
                  // commas so that the resulting code is interpreted with
                  // both (all) of the holes.
                  parts.push(",");
              } else {
                  var lines$$1 = printed[i];
                  if (oneLine) {
                      if (i > 0)
                          parts.push(" ");
                  } else {
                      lines$$1 = lines$$1.indent(options$$1.tabWidth);
                  }
                  parts.push(lines$$1);
                  if (i < len - 1 || (!oneLine && util_1$1.isTrailingCommaEnabled(options$$1, "arrays")))
                      parts.push(",");
                  if (!oneLine)
                      parts.push("\n");
              }
          }, "elements");

          if (oneLine && options$$1.arrayBracketSpacing) {
            parts.push(" ]");
          } else {
            parts.push("]");
          }

          return concat$2(parts);

      case "SequenceExpression":
          return fromString$3(", ").join(path.map(print, "expressions"));

      case "ThisExpression":
          return fromString$3("this");

      case "Super":
          return fromString$3("super");

      case "NullLiteral": // Babel 6 Literal split
          return fromString$3("null");

      case "RegExpLiteral": // Babel 6 Literal split
          return fromString$3(n.extra.raw);

      case "BigIntLiteral": // Babel 7 Literal split
          return fromString$3(n.value + "n");

      case "NumericLiteral": // Babel 6 Literal Split
          // Keep original representation for numeric values not in base 10.
          if (n.extra &&
              typeof n.extra.raw === "string" &&
              Number(n.extra.raw) === n.value) {
              return fromString$3(n.extra.raw, options$$1);
          }

          return fromString$3(n.value, options$$1);

      case "BooleanLiteral": // Babel 6 Literal split

      case "StringLiteral": // Babel 6 Literal split
      case "Literal":
          // Numeric values may be in bases other than 10. Use their raw
          // representation if equivalent.
          if (typeof n.value === "number" &&
              typeof n.raw === "string" &&
              Number(n.raw) === n.value) {
              return fromString$3(n.raw, options$$1);
          }

          if (typeof n.value !== "string") {
              return fromString$3(n.value, options$$1);
          }

          return fromString$3(nodeStr(n.value, options$$1), options$$1);

      case "Directive": // Babel 6
          return path.call(print, "value");

      case "DirectiveLiteral": // Babel 6
          return fromString$3(nodeStr(n.value, options$$1));

      case "ModuleSpecifier":
          if (n.local) {
              throw new Error(
                  "The ESTree ModuleSpecifier type should be abstract"
              );
          }

          // The Esprima ModuleSpecifier type is just a string-valued
          // Literal identifying the imported-from module.
          return fromString$3(nodeStr(n.value, options$$1), options$$1);

      case "UnaryExpression":
          parts.push(n.operator);
          if (/[a-z]$/.test(n.operator))
              parts.push(" ");
          parts.push(path.call(print, "argument"));
          return concat$2(parts);

      case "UpdateExpression":
          parts.push(
              path.call(print, "argument"),
              n.operator
          );

          if (n.prefix)
              parts.reverse();

          return concat$2(parts);

      case "ConditionalExpression":
          return concat$2([
              path.call(print, "test"),
              " ? ", path.call(print, "consequent"),
              " : ", path.call(print, "alternate")
          ]);

      case "NewExpression":
          parts.push("new ", path.call(print, "callee"));
          var args = n.arguments;
          if (args) {
              parts.push(printArgumentsList(path, options$$1, print));
          }

          return concat$2(parts);

      case "VariableDeclaration":
          if (n.declare) {
              parts.push("declare ");
          }

          parts.push(n.kind, " ");

          var maxLen = 0;
          var printed = path.map(function(childPath) {
              var lines$$1 = print(childPath);
              maxLen = Math.max(lines$$1.length, maxLen);
              return lines$$1;
          }, "declarations");

          if (maxLen === 1) {
              parts.push(fromString$3(", ").join(printed));
          } else if (printed.length > 1 ) {
              parts.push(
                  fromString$3(",\n").join(printed)
                      .indentTail(n.kind.length + 1)
              );
          } else {
              parts.push(printed[0]);
          }

          // We generally want to terminate all variable declarations with a
          // semicolon, except when they are children of for loops.
          var parentNode = path.getParentNode();
          if (!namedTypes.ForStatement.check(parentNode) &&
              !namedTypes.ForInStatement.check(parentNode) &&
              !(namedTypes.ForOfStatement &&
                namedTypes.ForOfStatement.check(parentNode)) &&
              !(namedTypes.ForAwaitStatement &&
                namedTypes.ForAwaitStatement.check(parentNode))) {
              parts.push(";");
          }

          return concat$2(parts);

      case "VariableDeclarator":
          return n.init ? fromString$3(" = ").join([
              path.call(print, "id"),
              path.call(print, "init")
          ]) : path.call(print, "id");

      case "WithStatement":
          return concat$2([
              "with (",
              path.call(print, "object"),
              ") ",
              path.call(print, "body")
          ]);

      case "IfStatement":
          var con = adjustClause(path.call(print, "consequent"), options$$1),
              parts = ["if (", path.call(print, "test"), ")", con];

          if (n.alternate)
              parts.push(
                  endsWithBrace(con) ? " else" : "\nelse",
                  adjustClause(path.call(print, "alternate"), options$$1));

          return concat$2(parts);

      case "ForStatement":
          // TODO Get the for (;;) case right.
          var init = path.call(print, "init"),
              sep = init.length > 1 ? ";\n" : "; ",
              forParen = "for (",
              indented = fromString$3(sep).join([
                  init,
                  path.call(print, "test"),
                  path.call(print, "update")
              ]).indentTail(forParen.length),
              head = concat$2([forParen, indented, ")"]),
              clause = adjustClause(path.call(print, "body"), options$$1),
              parts = [head];

          if (head.length > 1) {
              parts.push("\n");
              clause = clause.trimLeft();
          }

          parts.push(clause);

          return concat$2(parts);

      case "WhileStatement":
          return concat$2([
              "while (",
              path.call(print, "test"),
              ")",
              adjustClause(path.call(print, "body"), options$$1)
          ]);

      case "ForInStatement":
          // Note: esprima can't actually parse "for each (".
          return concat$2([
              n.each ? "for each (" : "for (",
              path.call(print, "left"),
              " in ",
              path.call(print, "right"),
              ")",
              adjustClause(path.call(print, "body"), options$$1)
          ]);

      case "ForOfStatement":
      case "ForAwaitStatement":
          parts.push("for ");

          if (n.await || n.type === "ForAwaitStatement") {
              parts.push("await ");
          }

          parts.push(
              "(",
              path.call(print, "left"),
              " of ",
              path.call(print, "right"),
              ")",
              adjustClause(path.call(print, "body"), options$$1)
          );

          return concat$2(parts);

      case "DoWhileStatement":
          var doBody = concat$2([
              "do",
              adjustClause(path.call(print, "body"), options$$1)
          ]), parts = [doBody];

          if (endsWithBrace(doBody))
              parts.push(" while");
          else
              parts.push("\nwhile");

          parts.push(" (", path.call(print, "test"), ");");

          return concat$2(parts);

      case "DoExpression":
          var statements = path.call(function(bodyPath) {
              return printStatementSequence(bodyPath, options$$1, print);
          }, "body");

          return concat$2([
              "do {\n",
              statements.indent(options$$1.tabWidth),
              "\n}"
          ]);

      case "BreakStatement":
          parts.push("break");
          if (n.label)
              parts.push(" ", path.call(print, "label"));
          parts.push(";");
          return concat$2(parts);

      case "ContinueStatement":
          parts.push("continue");
          if (n.label)
              parts.push(" ", path.call(print, "label"));
          parts.push(";");
          return concat$2(parts);

      case "LabeledStatement":
          return concat$2([
              path.call(print, "label"),
              ":\n",
              path.call(print, "body")
          ]);

      case "TryStatement":
          parts.push(
              "try ",
              path.call(print, "block")
          );

          if (n.handler) {
              parts.push(" ", path.call(print, "handler"));
          } else if (n.handlers) {
              path.each(function(handlerPath) {
                  parts.push(" ", print(handlerPath));
              }, "handlers");
          }

          if (n.finalizer) {
              parts.push(" finally ", path.call(print, "finalizer"));
          }

          return concat$2(parts);

      case "CatchClause":
          parts.push("catch ");

          if (n.param) {
              parts.push("(", path.call(print, "param"));
          }

          if (n.guard) {
              // Note: esprima does not recognize conditional catch clauses.
              parts.push(" if ", path.call(print, "guard"));
          }

          if (n.param) {
              parts.push(") ");
          }

          parts.push(path.call(print, "body"));

          return concat$2(parts);

      case "ThrowStatement":
          return concat$2(["throw ", path.call(print, "argument"), ";"]);

      case "SwitchStatement":
          return concat$2([
              "switch (",
              path.call(print, "discriminant"),
              ") {\n",
              fromString$3("\n").join(path.map(print, "cases")),
              "\n}"
          ]);

          // Note: ignoring n.lexical because it has no printing consequences.

      case "SwitchCase":
          if (n.test)
              parts.push("case ", path.call(print, "test"), ":");
          else
              parts.push("default:");

          if (n.consequent.length > 0) {
              parts.push("\n", path.call(function(consequentPath) {
                  return printStatementSequence(consequentPath, options$$1, print);
              }, "consequent").indent(options$$1.tabWidth));
          }

          return concat$2(parts);

      case "DebuggerStatement":
          return fromString$3("debugger;");

      // JSX extensions below.

      case "JSXAttribute":
          parts.push(path.call(print, "name"));
          if (n.value)
              parts.push("=", path.call(print, "value"));
          return concat$2(parts);

      case "JSXIdentifier":
          return fromString$3(n.name, options$$1);

      case "JSXNamespacedName":
          return fromString$3(":").join([
              path.call(print, "namespace"),
              path.call(print, "name")
          ]);

      case "JSXMemberExpression":
          return fromString$3(".").join([
              path.call(print, "object"),
              path.call(print, "property")
          ]);

      case "JSXSpreadAttribute":
          return concat$2(["{...", path.call(print, "argument"), "}"]);

      case "JSXSpreadChild":
          return concat$2(["{...", path.call(print, "expression"), "}"]);

      case "JSXExpressionContainer":
          return concat$2(["{", path.call(print, "expression"), "}"]);

      case "JSXElement":
      case "JSXFragment":
          var openingPropName = "opening" + (
              n.type === "JSXElement" ? "Element" : "Fragment");

          var closingPropName = "closing" + (
              n.type === "JSXElement" ? "Element" : "Fragment");

          var openingLines = path.call(print, openingPropName);

          if (n[openingPropName].selfClosing) {
              return openingLines;
          }

          var childLines = concat$2(
              path.map(function(childPath) {
                  var child = childPath.getValue();

                  if (namedTypes.Literal.check(child) &&
                      typeof child.value === "string") {
                      if (/\S/.test(child.value)) {
                          return child.value.replace(/^\s+|\s+$/g, "");
                      } else if (/\n/.test(child.value)) {
                          return "\n";
                      }
                  }

                  return print(childPath);
              }, "children")
          ).indentTail(options$$1.tabWidth);

          var closingLines = path.call(print, closingPropName);

          return concat$2([
              openingLines,
              childLines,
              closingLines
          ]);

      case "JSXOpeningElement":
          parts.push("<", path.call(print, "name"));
          var attrParts = [];

          path.each(function(attrPath) {
              attrParts.push(" ", print(attrPath));
          }, "attributes");

          var attrLines = concat$2(attrParts);

          var needLineWrap = (
              attrLines.length > 1 ||
              attrLines.getLineLength(1) > options$$1.wrapColumn
          );

          if (needLineWrap) {
              attrParts.forEach(function(part, i) {
                  if (part === " ") {
                      attrParts[i] = "\n";
                  }
              });

              attrLines = concat$2(attrParts).indentTail(options$$1.tabWidth);
          }

          parts.push(attrLines, n.selfClosing ? " />" : ">");

          return concat$2(parts);

      case "JSXClosingElement":
          return concat$2(["</", path.call(print, "name"), ">"]);

      case "JSXOpeningFragment":
          return fromString$3("<>");

      case "JSXClosingFragment":
          return fromString$3("</>")

      case "JSXText":
          return fromString$3(n.value, options$$1);

      case "JSXEmptyExpression":
          return fromString$3("");

      case "TypeAnnotatedIdentifier":
          return concat$2([
              path.call(print, "annotation"),
              " ",
              path.call(print, "identifier")
          ]);

      case "ClassBody":
          if (n.body.length === 0) {
              return fromString$3("{}");
          }

          return concat$2([
              "{\n",
              path.call(function(bodyPath) {
                  return printStatementSequence(bodyPath, options$$1, print);
              }, "body").indent(options$$1.tabWidth),
              "\n}"
          ]);

      case "ClassPropertyDefinition":
          parts.push("static ", path.call(print, "definition"));
          if (!namedTypes.MethodDefinition.check(n.definition))
              parts.push(";");
          return concat$2(parts);

      case "ClassProperty":
          if (typeof n.accessibility === "string") {
              parts.push(n.accessibility, " ");
          }

          if (n.static) {
              parts.push("static ");
          }

          if (n.abstract) {
              parts.push("abstract ");
          }

          if (n.readonly) {
              parts.push("readonly ");
          }

          var key = path.call(print, "key");

          if (n.computed) {
              key = concat$2(["[", key, "]"]);
          }

          if (n.variance) {
              key = concat$2([printVariance(path, print), key]);
          }

          parts.push(key);

          if (n.optional) {
              parts.push("?");
          }

          if (n.typeAnnotation) {
              parts.push(path.call(print, "typeAnnotation"));
          }

          if (n.value) {
              parts.push(" = ", path.call(print, "value"));
          }

          parts.push(";");
          return concat$2(parts);

      case "ClassDeclaration":
      case "ClassExpression":
          if (n.declare) {
              parts.push("declare ");
          }

          if (n.abstract) {
              parts.push("abstract ");
          }

          parts.push("class");

          if (n.id) {
              parts.push(
                  " ",
                  path.call(print, "id")
              );
          }

          if (n.typeParameters) {
              parts.push(path.call(print, "typeParameters"));
          }

          if (n.superClass) {
              parts.push(
                  " extends ",
                  path.call(print, "superClass"),
                  path.call(print, "superTypeParameters")
              );
          }

          if (n["implements"] && n['implements'].length > 0) {
              parts.push(
                  " implements ",
                  fromString$3(", ").join(path.map(print, "implements"))
              );
          }

          parts.push(" ", path.call(print, "body"));

          return concat$2(parts);

      case "TemplateElement":
          return fromString$3(n.value.raw, options$$1).lockIndentTail();

      case "TemplateLiteral":
          var expressions = path.map(print, "expressions");
          parts.push("`");

          path.each(function(childPath) {
              var i = childPath.getName();
              parts.push(print(childPath));
              if (i < expressions.length) {
                  parts.push("${", expressions[i], "}");
              }
          }, "quasis");

          parts.push("`");

          return concat$2(parts).lockIndentTail();

      case "TaggedTemplateExpression":
          return concat$2([
              path.call(print, "tag"),
              path.call(print, "quasi")
          ]);

      // These types are unprintable because they serve as abstract
      // supertypes for other (printable) types.
      case "Node":
      case "Printable":
      case "SourceLocation":
      case "Position":
      case "Statement":
      case "Function":
      case "Pattern":
      case "Expression":
      case "Declaration":
      case "Specifier":
      case "NamedSpecifier":
      case "Comment": // Supertype of Block and Line
      case "Flow": // Supertype of all Flow AST node types
      case "FlowType": // Supertype of all Flow types
      case "FlowPredicate": // Supertype of InferredPredicate and DeclaredPredicate
      case "MemberTypeAnnotation": // Flow
      case "Type": // Flow
      case "TSHasOptionalTypeParameters":
      case "TSHasOptionalTypeAnnotation":
          throw new Error("unprintable type: " + JSON.stringify(n.type));

      case "CommentBlock": // Babel block comment.
      case "Block": // Esprima block comment.
          return concat$2(["/*", fromString$3(n.value, options$$1), "*/"]);

      case "CommentLine": // Babel line comment.
      case "Line": // Esprima line comment.
          return concat$2(["//", fromString$3(n.value, options$$1)]);

      // Type Annotations for Facebook Flow, typically stripped out or
      // transformed away before printing.
      case "TypeAnnotation":
          if (n.typeAnnotation) {
              if (n.typeAnnotation.type !== "FunctionTypeAnnotation") {
                  parts.push(": ");
              }
              parts.push(path.call(print, "typeAnnotation"));
              return concat$2(parts);
          }

          return fromString$3("");

      case "ExistentialTypeParam":
      case "ExistsTypeAnnotation":
          return fromString$3("*", options$$1);

      case "EmptyTypeAnnotation":
          return fromString$3("empty", options$$1);

      case "AnyTypeAnnotation":
          return fromString$3("any", options$$1);

      case "MixedTypeAnnotation":
          return fromString$3("mixed", options$$1);

      case "ArrayTypeAnnotation":
          return concat$2([
              path.call(print, "elementType"),
              "[]"
          ]);

      case "TupleTypeAnnotation":
          var printed = path.map(print, "types");
          var joined = fromString$3(", ").join(printed);
          var oneLine = joined.getLineLength(1) <= options$$1.wrapColumn;
          if (oneLine) {
            if (options$$1.arrayBracketSpacing) {
              parts.push("[ ");
            } else {
              parts.push("[");
            }
          } else {
            parts.push("[\n");
          }

          path.each(function(elemPath) {
              var i = elemPath.getName();
              var elem = elemPath.getValue();
              if (!elem) {
                  // If the array expression ends with a hole, that hole
                  // will be ignored by the interpreter, but if it ends with
                  // two (or more) holes, we need to write out two (or more)
                  // commas so that the resulting code is interpreted with
                  // both (all) of the holes.
                  parts.push(",");
              } else {
                  var lines$$1 = printed[i];
                  if (oneLine) {
                      if (i > 0)
                          parts.push(" ");
                  } else {
                      lines$$1 = lines$$1.indent(options$$1.tabWidth);
                  }
                  parts.push(lines$$1);
                  if (i < n.types.length - 1 || (!oneLine && util_1$1.isTrailingCommaEnabled(options$$1, "arrays")))
                      parts.push(",");
                  if (!oneLine)
                      parts.push("\n");
              }
          }, "types");

          if (oneLine && options$$1.arrayBracketSpacing) {
            parts.push(" ]");
          } else {
            parts.push("]");
          }

          return concat$2(parts);

      case "BooleanTypeAnnotation":
          return fromString$3("boolean", options$$1);

      case "BooleanLiteralTypeAnnotation":
          return fromString$3("" + n.value, options$$1);

      case "DeclareClass":
          return printFlowDeclaration(path, [
              "class ",
              path.call(print, "id"),
              " ",
              path.call(print, "body"),
          ]);

      case "DeclareFunction":
          return printFlowDeclaration(path, [
              "function ",
              path.call(print, "id"),
              ";"
          ]);

      case "DeclareModule":
          return printFlowDeclaration(path, [
              "module ",
              path.call(print, "id"),
              " ",
              path.call(print, "body"),
          ]);

      case "DeclareModuleExports":
          return printFlowDeclaration(path, [
              "module.exports",
              path.call(print, "typeAnnotation"),
          ]);

      case "DeclareVariable":
          return printFlowDeclaration(path, [
              "var ",
              path.call(print, "id"),
              ";"
          ]);

      case "DeclareExportDeclaration":
      case "DeclareExportAllDeclaration":
          return concat$2([
              "declare ",
              printExportDeclaration(path, options$$1, print)
          ]);

      case "InferredPredicate":
          return fromString$3("%checks", options$$1);

      case "DeclaredPredicate":
          return concat$2([
              "%checks(",
              path.call(print, "value"),
              ")"
          ]);

      case "FunctionTypeAnnotation":
          // FunctionTypeAnnotation is ambiguous:
          // declare function(a: B): void; OR
          // var A: (a: B) => void;
          var parent = path.getParentNode(0);
          var isArrowFunctionTypeAnnotation = !(
              namedTypes.ObjectTypeCallProperty.check(parent) ||
              namedTypes.DeclareFunction.check(path.getParentNode(2))
          );

          var needsColon =
              isArrowFunctionTypeAnnotation &&
              !namedTypes.FunctionTypeParam.check(parent);

          if (needsColon) {
              parts.push(": ");
          }

          parts.push(
              "(",
              fromString$3(", ").join(path.map(print, "params")),
              ")"
          );

          // The returnType is not wrapped in a TypeAnnotation, so the colon
          // needs to be added separately.
          if (n.returnType) {
              parts.push(
                  isArrowFunctionTypeAnnotation ? " => " : ": ",
                  path.call(print, "returnType")
              );
          }

          return concat$2(parts);

      case "FunctionTypeParam":
          return concat$2([
              path.call(print, "name"),
              n.optional ? '?' : '',
              ": ",
              path.call(print, "typeAnnotation"),
          ]);

      case "GenericTypeAnnotation":
          return concat$2([
              path.call(print, "id"),
              path.call(print, "typeParameters")
          ]);

      case "DeclareInterface":
          parts.push("declare ");
          // Fall through to InterfaceDeclaration...

      case "InterfaceDeclaration":
      case "TSInterfaceDeclaration":
          if (n.declare) {
              parts.push("declare ");
          }

          parts.push(
              "interface ",
              path.call(print, "id"),
              path.call(print, "typeParameters"),
              " "
          );

          if (n["extends"] && n["extends"].length > 0) {
              parts.push(
                  "extends ",
                  fromString$3(", ").join(path.map(print, "extends")),
                  " "
              );
          }

          if (n.body) {
              parts.push(path.call(print, "body"));
          }

          return concat$2(parts);

      case "ClassImplements":
      case "InterfaceExtends":
          return concat$2([
              path.call(print, "id"),
              path.call(print, "typeParameters")
          ]);

      case "IntersectionTypeAnnotation":
          return fromString$3(" & ").join(path.map(print, "types"));

      case "NullableTypeAnnotation":
          return concat$2([
              "?",
              path.call(print, "typeAnnotation")
          ]);

      case "NullLiteralTypeAnnotation":
          return fromString$3("null", options$$1);

      case "ThisTypeAnnotation":
          return fromString$3("this", options$$1);

      case "NumberTypeAnnotation":
          return fromString$3("number", options$$1);

      case "ObjectTypeCallProperty":
          return path.call(print, "value");

      case "ObjectTypeIndexer":
          return concat$2([
              printVariance(path, print),
              "[",
              path.call(print, "id"),
              ": ",
              path.call(print, "key"),
              "]: ",
              path.call(print, "value")
          ]);

      case "ObjectTypeProperty":
          return concat$2([
              printVariance(path, print),
              path.call(print, "key"),
              n.optional ? "?" : "",
              ": ",
              path.call(print, "value")
          ]);

      case "QualifiedTypeIdentifier":
          return concat$2([
              path.call(print, "qualification"),
              ".",
              path.call(print, "id")
          ]);

      case "StringLiteralTypeAnnotation":
          return fromString$3(nodeStr(n.value, options$$1), options$$1);

      case "NumberLiteralTypeAnnotation":
      case "NumericLiteralTypeAnnotation":
          return fromString$3(JSON.stringify(n.value), options$$1);

      case "StringTypeAnnotation":
          return fromString$3("string", options$$1);

      case "DeclareTypeAlias":
          parts.push("declare ");
          // Fall through to TypeAlias...

      case "TypeAlias":
          return concat$2([
              "type ",
              path.call(print, "id"),
              path.call(print, "typeParameters"),
              " = ",
              path.call(print, "right"),
              ";"
          ]);

      case "DeclareOpaqueType":
          parts.push("declare ");
          // Fall through to OpaqueType...

      case "OpaqueType":
          parts.push(
              "opaque type ",
              path.call(print, "id"),
              path.call(print, "typeParameters")
          );

          if (n["supertype"]) {
              parts.push(": ", path.call(print, "supertype"));
          }

          if (n["impltype"]) {
              parts.push(" = ", path.call(print, "impltype"));
          }

          parts.push(";");

          return concat$2(parts);

      case "TypeCastExpression":
          return concat$2([
              "(",
              path.call(print, "expression"),
              path.call(print, "typeAnnotation"),
              ")"
          ]);

      case "TypeParameterDeclaration":
      case "TypeParameterInstantiation":
          return concat$2([
              "<",
              fromString$3(", ").join(path.map(print, "params")),
              ">"
          ]);

      case "Variance":
          if (n.kind === "plus") {
              return fromString$3("+");
          }

          if (n.kind === "minus") {
              return fromString$3("-");
          }

          return fromString$3("");

      case "TypeParameter":
          if (n.variance) {
              parts.push(printVariance(path, print));
          }

          parts.push(path.call(print, 'name'));

          if (n.bound) {
              parts.push(path.call(print, 'bound'));
          }

          if (n['default']) {
              parts.push('=', path.call(print, 'default'));
          }

          return concat$2(parts);

      case "TypeofTypeAnnotation":
          return concat$2([
              fromString$3("typeof ", options$$1),
              path.call(print, "argument")
          ]);

      case "UnionTypeAnnotation":
          return fromString$3(" | ").join(path.map(print, "types"));

      case "VoidTypeAnnotation":
          return fromString$3("void", options$$1);

      case "NullTypeAnnotation":
          return fromString$3("null", options$$1);

      // Type Annotations for TypeScript (when using Babylon as parser)
      case "TSType":
          throw new Error("unprintable type: " + JSON.stringify(n.type));

      case "TSNumberKeyword":
          return fromString$3("number", options$$1);

      case "TSObjectKeyword":
          return fromString$3("object", options$$1);

      case "TSBooleanKeyword":
          return fromString$3("boolean", options$$1);

      case "TSStringKeyword":
          return fromString$3("string", options$$1);

      case "TSSymbolKeyword":
          return fromString$3("symbol", options$$1);

      case "TSAnyKeyword":
          return fromString$3("any", options$$1);

      case "TSVoidKeyword":
          return fromString$3("void", options$$1);

      case "TSThisType":
          return fromString$3("this", options$$1);

      case "TSNullKeyword":
          return fromString$3("null", options$$1);

      case "TSUndefinedKeyword":
          return fromString$3("undefined", options$$1);

      case "TSUnknownKeyword":
          return fromString$3("unknown", options$$1);

      case "TSNeverKeyword":
          return fromString$3("never", options$$1);

      case "TSArrayType":
          return concat$2([
              path.call(print, "elementType"),
              "[]"
          ]);

      case "TSLiteralType":
          return path.call(print, "literal")

      case "TSUnionType":
          return fromString$3(" | ").join(path.map(print, "types"));

      case "TSIntersectionType":
          return fromString$3(" & ").join(path.map(print, "types"));

      case "TSConditionalType":
          parts.push(
              path.call(print, "checkType"),
              " extends ",
              path.call(print, "extendsType"),
              " ? ",
              path.call(print, "trueType"),
              " : ",
              path.call(print, "falseType")
          );

          return concat$2(parts);

      case "TSInferType":
          parts.push(
              "infer ",
              path.call(print, "typeParameter")
          );

          return concat$2(parts);

      case "TSParenthesizedType":
          return concat$2([
              "(",
              path.call(print, "typeAnnotation"),
              ")"
          ]);

      case "TSFunctionType":
      case "TSConstructorType":
          return concat$2([
              path.call(print, "typeParameters"),
              "(",
              printFunctionParams(path, options$$1, print),
              ")",
              path.call(print, "typeAnnotation")
          ]);

      case "TSMappedType": {
          parts.push(
              n.readonly ? "readonly " : "",
              "[",
              path.call(print, "typeParameter"),
              "]",
              n.optional ? "?" : ""
          );

          if (n.typeAnnotation) {
              parts.push(": ", path.call(print, "typeAnnotation"), ";");
          }

          return concat$2([
              "{\n",
              concat$2(parts).indent(options$$1.tabWidth),
              "\n}",
          ]);
      }

      case "TSTupleType":
          return concat$2([
              "[",
              fromString$3(", ").join(path.map(print, "elementTypes")),
              "]"
          ]);

      case "TSRestType":
          return concat$2([
              "...",
              path.call(print, "typeAnnotation"),
              "[]"
          ]);

      case "TSOptionalType":
          return concat$2([
              path.call(print, "typeAnnotation"),
              "?"
          ]);

      case "TSIndexedAccessType":
          return concat$2([
              path.call(print, "objectType"),
              "[",
              path.call(print, "indexType"),
              "]"
          ]);

      case "TSTypeOperator":
          return concat$2([
              path.call(print, "operator"),
              " ",
              path.call(print, "typeAnnotation")
          ]);

      case "TSTypeLiteral": {
          const memberLines =
              fromString$3(",\n").join(path.map(print, "members"));

          if (memberLines.isEmpty()) {
              return fromString$3("{}", options$$1);
          }

          parts.push(
              "{\n",
              memberLines.indent(options$$1.tabWidth),
              "\n}"
          );

          return concat$2(parts);
      }

      case "TSEnumMember":
          parts.push(path.call(print, "id"));
          if (n.initializer) {
              parts.push(
                  " = ",
                  path.call(print, "initializer")
              );
          }
          return concat$2(parts);

      case "TSTypeQuery":
          return concat$2([
              "typeof ",
              path.call(print, "exprName"),
          ]);

      case "TSParameterProperty":
          if (n.accessibility) {
              parts.push(n.accessibility, " ");
          }

          if (n.export) {
              parts.push("export ");
          }

          if (n.static) {
              parts.push("static ");
          }

          if (n.readonly) {
              parts.push("readonly ");
          }

          parts.push(path.call(print, "parameter"));

          return concat$2(parts);

      case "TSTypeReference":
          return concat$2([
              path.call(print, "typeName"),
              path.call(print, "typeParameters")
          ]);

      case "TSQualifiedName":
          return concat$2([
              path.call(print, "left"),
              ".",
              path.call(print, "right")
          ]);

      case "TSAsExpression": {
          var withParens = n.extra && n.extra.parenthesized === true;
          parts = [];
          if (withParens) parts.push("(");
          parts.push(
              path.call(print, "expression"),
              fromString$3(" as "),
              path.call(print, "typeAnnotation")
          );
          if (withParens) parts.push(")");

          return concat$2(parts);
      }

      case "TSNonNullExpression":
          return concat$2([
              path.call(print, "expression"),
              "!"
          ]);

      case "TSTypeAnnotation": {
          // similar to flow's FunctionTypeAnnotation, this can be
          // ambiguous: it can be prefixed by => or :
          // in a type predicate, it takes the for u is U
          var parent = path.getParentNode(0);
          var prefix = ": ";
          var isFunctionType = namedTypes.TSFunctionType.check(parent);
          if (namedTypes.TSFunctionType.check(parent)) {
              prefix = " => ";
          }

          if (namedTypes.TSTypePredicate.check(parent)) {
              prefix = " is ";
          }

          return concat$2([
              prefix,
              path.call(print, "typeAnnotation")
          ]);
      }

      case "TSIndexSignature":
          return concat$2([
              n.readonly ? "readonly " : "",
              "[",
              path.map(print, "parameters"),
              "]",
              path.call(print, "typeAnnotation")
          ]);

      case "TSPropertySignature":
          parts.push(
              printVariance(path, print),
              n.readonly ? "readonly " : ""
          );

          if (n.computed) {
              parts.push(
                  "[",
                  path.call(print, "key"),
                  "]"
              );
          } else {
              parts.push(path.call(print, "key"));
          }

          parts.push(
              n.optional ? "?" : "",
              path.call(print, "typeAnnotation")
          );

          return concat$2(parts);

      case "TSMethodSignature":
          if (n.computed) {
              parts.push(
                  "[",
                  path.call(print, "key"),
                  "]"
              );
          } else {
              parts.push(path.call(print, "key"));
          }

          if (n.optional) {
              parts.push("?");
          }

          parts.push(
              path.call(print, "typeParameters"),
              "(",
              printFunctionParams(path, options$$1, print),
              ")",
              path.call(print, "typeAnnotation")
          );

          return concat$2(parts);

      case "TSTypePredicate":
          return concat$2([
              path.call(print, "parameterName"),
              path.call(print, "typeAnnotation")
          ]);

      case "TSCallSignatureDeclaration":
          return concat$2([
              path.call(print, "typeParameters"),
              "(",
              printFunctionParams(path, options$$1, print),
              ")",
              path.call(print, "typeAnnotation")
          ]);

      case "TSConstructSignatureDeclaration":
          if (n.typeParameters) {
              parts.push(
                  "new",
                  path.call(print, "typeParameters")
              );
          } else {
              parts.push("new ");
          }

          parts.push(
              "(",
              printFunctionParams(path, options$$1, print),
              ")",
              path.call(print, "typeAnnotation")
          );

          return concat$2(parts);

      case "TSTypeAliasDeclaration":
          return concat$2([
              n.declare ? "declare " : "",
              "type ",
              path.call(print, "id"),
              path.call(print, "typeParameters"),
              " = ",
              path.call(print, "typeAnnotation"),
              ";"
          ]);

      case "TSTypeParameter":
          parts.push(path.call(print, "name"));

          // ambiguous because of TSMappedType
          var parent = path.getParentNode(0);
          var isInMappedType = namedTypes.TSMappedType.check(parent);

          if (n.constraint) {
              parts.push(
                  isInMappedType ? " in " : " extends ",
                  path.call(print, "constraint")
              );
          }

          if (n["default"]) {
              parts.push(" = ", path.call(print, "default"));
          }

          return concat$2(parts);

      case "TSTypeAssertion":
          var withParens = n.extra && n.extra.parenthesized === true;
          if (withParens) {
              parts.push("(");
          }

          parts.push(
              "<",
              path.call(print, "typeAnnotation"),
              "> ",
              path.call(print, "expression")
          );

          if (withParens) {
              parts.push(")");
          }

          return concat$2(parts);

      case "TSTypeParameterDeclaration":
      case "TSTypeParameterInstantiation":
          return concat$2([
              "<",
              fromString$3(", ").join(path.map(print, "params")),
              ">"
          ]);

      case "TSEnumDeclaration":
          parts.push(
              n.declare ? "declare " : "",
              n.const ? "const " : "",
              "enum ",
              path.call(print, "id")
          );

          const memberLines =
              fromString$3(",\n").join(path.map(print, "members"));

          if (memberLines.isEmpty()) {
              parts.push(" {}");
          } else {
              parts.push(
                  " {\n",
                  memberLines.indent(options$$1.tabWidth),
                  "\n}"
              );
          }

          return concat$2(parts);

      case "TSExpressionWithTypeArguments":
          return concat$2([
              path.call(print, "expression"),
              path.call(print, "typeParameters")
          ]);

      case "TSInterfaceBody":
          var lines$$1 = fromString$3(";\n").join(path.map(print, "body"));
          if (lines$$1.isEmpty()) {
              return fromString$3("{}", options$$1);
          }

          return concat$2([
              "{\n",
              lines$$1.indent(options$$1.tabWidth), ";",
              "\n}",
          ]);

      case "TSImportEqualsDeclaration":
          if (n.isExport) {
              parts.push("export ");
          }

          parts.push(
              "import ",
              path.call(print, "id"),
              " = ",
              path.call(print, "moduleReference")
          );

          return maybeAddSemicolon(concat$2(parts));

      case "TSExternalModuleReference":
        return concat$2(["require(", path.call(print, "expression"), ")"]);

      case "TSModuleDeclaration": {
          const parent = path.getParentNode();

          if (parent.type === "TSModuleDeclaration") {
              parts.push(".");
          } else {
              if (n.declare) {
                  parts.push("declare ");
              }

              if (! n.global) {
                  const isExternal = n.id.type === "StringLiteral" ||
                      (n.id.type === "Literal" &&
                       typeof n.id.value === "string");

                  if (isExternal) {
                      parts.push("module ");

                  } else if (n.loc &&
                             n.loc.lines &&
                             n.id.loc) {
                      const prefix = n.loc.lines.sliceString(
                          n.loc.start,
                          n.id.loc.start
                      );

                      // These keywords are fundamentally ambiguous in the
                      // Babylon parser, and not reflected in the AST, so
                      // the best we can do is to match the original code,
                      // when possible.
                      if (prefix.indexOf("module") >= 0) {
                          parts.push("module ");
                      } else {
                          parts.push("namespace ");
                      }

                  } else {
                      parts.push("namespace ");
                  }
              }
          }

          parts.push(path.call(print, "id"));

          if (n.body && n.body.type === "TSModuleDeclaration") {
              parts.push(path.call(print, "body"));
          } else if (n.body) {
              const bodyLines = path.call(print, "body");
              if (bodyLines.isEmpty()) {
                  parts.push(" {}");
              } else {
                  parts.push(
                      " {\n",
                      bodyLines.indent(options$$1.tabWidth),
                      "\n}"
                  );
              }
          }

          return concat$2(parts);
      }

      case "TSModuleBlock":
          return path.call(function (bodyPath) {
              return printStatementSequence(bodyPath, options$$1, print);
          }, "body");

      // Unhandled types below. If encountered, nodes of these types should
      // be either left alone or desugared into AST types that are fully
      // supported by the pretty-printer.
      case "ClassHeritage": // TODO
      case "ComprehensionBlock": // TODO
      case "ComprehensionExpression": // TODO
      case "Glob": // TODO
      case "GeneratorExpression": // TODO
      case "LetStatement": // TODO
      case "LetExpression": // TODO
      case "GraphExpression": // TODO
      case "GraphIndexExpression": // TODO

      // XML types that nobody cares about or needs to print.
      case "XMLDefaultDeclaration":
      case "XMLAnyName":
      case "XMLQualifiedIdentifier":
      case "XMLFunctionQualifiedIdentifier":
      case "XMLAttributeSelector":
      case "XMLFilterExpression":
      case "XML":
      case "XMLElement":
      case "XMLList":
      case "XMLEscape":
      case "XMLText":
      case "XMLStartTag":
      case "XMLEndTag":
      case "XMLPointTag":
      case "XMLName":
      case "XMLAttribute":
      case "XMLCdata":
      case "XMLComment":
      case "XMLProcessingInstruction":
      default:
          throw new Error("unknown type: " + JSON.stringify(n.type));
      }

      return p;
  }

  function printDecorators(path, printPath) {
      const parts = [];
      const node = path.getValue();

      if (node.decorators &&
          node.decorators.length > 0 &&
          // If the parent node is an export declaration, it will be
          // responsible for printing node.decorators.
          ! util_1$1.getParentExportDeclaration(path)) {

          path.each(function(decoratorPath) {
              parts.push(printPath(decoratorPath), "\n");
          }, "decorators");

      } else if (util_1$1.isExportDeclaration(node) &&
                 node.declaration &&
                 node.declaration.decorators) {
          // Export declarations are responsible for printing any decorators
          // that logically apply to node.declaration.
          path.each(function(decoratorPath) {
              parts.push(printPath(decoratorPath), "\n");
          }, "declaration", "decorators");
      }

      return concat$2(parts);
  }

  function printStatementSequence(path, options$$1, print) {
      var inClassBody =
          namedTypes.ClassBody &&
          namedTypes.ClassBody.check(path.getParentNode());

      var filtered = [];

      path.each(function(stmtPath) {
          var i = stmtPath.getName();
          var stmt = stmtPath.getValue();

          // Just in case the AST has been modified to contain falsy
          // "statements," it's safer simply to skip them.
          if (!stmt) {
              return;
          }

          // Skip printing EmptyStatement nodes to avoid leaving stray
          // semicolons lying around.
          if (stmt.type === "EmptyStatement" &&
              ! (stmt.comments && stmt.comments.length > 0)) {
              return;
          }

          if (namedTypes.Comment.check(stmt)) ; else if (namedTypes.Statement.check(stmt)) ; else {
              // When the pretty printer encounters a string instead of an
              // AST node, it just prints the string. This behavior can be
              // useful for fine-grained formatting decisions like inserting
              // blank lines.
              isString$4.assert(stmt);
          }

          // We can't hang onto stmtPath outside of this function, because
          // it's just a reference to a mutable FastPath object, so we have
          // to go ahead and print it here.
          filtered.push({
              node: stmt,
              printed: print(stmtPath)
          });
      });

      var prevTrailingSpace = null;
      var len = filtered.length;
      var parts = [];

      filtered.forEach(function(info, i) {
          var printed = info.printed;
          var stmt = info.node;
          var multiLine = printed.length > 1;
          var notFirst = i > 0;
          var notLast = i < len - 1;
          var leadingSpace;
          var trailingSpace;
          var lines$$1 = stmt && stmt.loc && stmt.loc.lines;
          var trueLoc = lines$$1 && options$$1.reuseWhitespace &&
              util_1$1.getTrueLoc(stmt, lines$$1);

          if (notFirst) {
              if (trueLoc) {
                  var beforeStart = lines$$1.skipSpaces(trueLoc.start, true);
                  var beforeStartLine = beforeStart ? beforeStart.line : 1;
                  var leadingGap = trueLoc.start.line - beforeStartLine;
                  leadingSpace = Array(leadingGap + 1).join("\n");
              } else {
                  leadingSpace = multiLine ? "\n\n" : "\n";
              }
          } else {
              leadingSpace = "";
          }

          if (notLast) {
              if (trueLoc) {
                  var afterEnd = lines$$1.skipSpaces(trueLoc.end);
                  var afterEndLine = afterEnd ? afterEnd.line : lines$$1.length;
                  var trailingGap = afterEndLine - trueLoc.end.line;
                  trailingSpace = Array(trailingGap + 1).join("\n");
              } else {
                  trailingSpace = multiLine ? "\n\n" : "\n";
              }
          } else {
              trailingSpace = "";
          }

          parts.push(
              maxSpace(prevTrailingSpace, leadingSpace),
              printed
          );

          if (notLast) {
              prevTrailingSpace = trailingSpace;
          } else if (trailingSpace) {
              parts.push(trailingSpace);
          }
      });

      return concat$2(parts);
  }

  function maxSpace(s1, s2) {
      if (!s1 && !s2) {
          return fromString$3("");
      }

      if (!s1) {
          return fromString$3(s2);
      }

      if (!s2) {
          return fromString$3(s1);
      }

      var spaceLines1 = fromString$3(s1);
      var spaceLines2 = fromString$3(s2);

      if (spaceLines2.length > spaceLines1.length) {
          return spaceLines2;
      }

      return spaceLines1;
  }

  function printMethod(path, options$$1, print) {
      var node = path.getNode();
      var kind = node.kind;
      var parts = [];

      var nodeValue = node.value;
      if (! namedTypes.FunctionExpression.check(nodeValue)) {
          nodeValue = node;
      }

      var access = node.accessibility || node.access;
      if (typeof access === "string") {
          parts.push(access, " ");
      }

      if (node.static) {
          parts.push("static ");
      }

      if (node.abstract) {
          parts.push("abstract ");
      }

      if (node.readonly) {
          parts.push("readonly ");
      }

      if (nodeValue.async) {
          parts.push("async ");
      }

      if (nodeValue.generator) {
          parts.push("*");
      }

      if (kind === "get" || kind === "set") {
          parts.push(kind, " ");
      }

      var key = path.call(print, "key");
      if (node.computed) {
          key = concat$2(["[", key, "]"]);
      }

      parts.push(key);

      if (node.optional) {
          parts.push("?");
      }

      if (node === nodeValue) {
          parts.push(
              path.call(print, "typeParameters"),
              "(",
              printFunctionParams(path, options$$1, print),
              ")",
              path.call(print, "returnType")
          );

          if (node.body) {
              parts.push(" ", path.call(print, "body"));
          } else {
              parts.push(";");
          }

      } else {
          parts.push(
              path.call(print, "value", "typeParameters"),
              "(",
              path.call(function(valuePath) {
                  return printFunctionParams(valuePath, options$$1, print);
              }, "value"),
              ")",
              path.call(print, "value", "returnType")
          );

          if (nodeValue.body) {
              parts.push(" ", path.call(print, "value", "body"));
          } else {
              parts.push(";");
          }
      }

      return concat$2(parts);
  }

  function printArgumentsList(path, options$$1, print) {
      var printed = path.map(print, "arguments");
      var trailingComma = util_1$1.isTrailingCommaEnabled(options$$1, "parameters");

      var joined = fromString$3(", ").join(printed);
      if (joined.getLineLength(1) > options$$1.wrapColumn) {
          joined = fromString$3(",\n").join(printed);
          return concat$2([
              "(\n",
              joined.indent(options$$1.tabWidth),
              trailingComma ? ",\n)" : "\n)"
          ]);
      }

      return concat$2(["(", joined, ")"]);
  }

  function printFunctionParams(path, options$$1, print) {
      var fun = path.getValue();

      if (fun.params) {
          var params = fun.params;
          var printed = path.map(print, "params");
      } else if (fun.parameters) {
          params = fun.parameters;
          printed = path.map(print, "parameters");
      }

      if (fun.defaults) {
          path.each(function(defExprPath) {
              var i = defExprPath.getName();
              var p = printed[i];
              if (p && defExprPath.getValue()) {
                  printed[i] = concat$2([p, " = ", print(defExprPath)]);
              }
          }, "defaults");
      }

      if (fun.rest) {
          printed.push(concat$2(["...", path.call(print, "rest")]));
      }

      var joined = fromString$3(", ").join(printed);
      if (joined.length > 1 ||
          joined.getLineLength(1) > options$$1.wrapColumn) {
          joined = fromString$3(",\n").join(printed);
          if (util_1$1.isTrailingCommaEnabled(options$$1, "parameters") &&
              !fun.rest &&
              params[params.length - 1].type !== 'RestElement') {
              joined = concat$2([joined, ",\n"]);
          } else {
              joined = concat$2([joined, "\n"]);
          }
          return concat$2(["\n", joined.indent(options$$1.tabWidth)]);
      }

      return joined;
  }

  function printExportDeclaration(path, options$$1, print) {
      var decl = path.getValue();
      var parts = ["export "];
      if (decl.exportKind && decl.exportKind !== "value") {
          parts.push(decl.exportKind + " ");
      }
      var shouldPrintSpaces = options$$1.objectCurlySpacing;

      namedTypes.Declaration.assert(decl);

      if (decl["default"] ||
          decl.type === "ExportDefaultDeclaration") {
          parts.push("default ");
      }

      if (decl.declaration) {
          parts.push(path.call(print, "declaration"));

      } else if (decl.specifiers) {

          if (decl.specifiers.length === 1 &&
              decl.specifiers[0].type === "ExportBatchSpecifier") {
              parts.push("*");

          } else if (decl.specifiers.length === 0) {
              parts.push("{}");

          } else if (decl.specifiers[0].type === 'ExportDefaultSpecifier') {
              const unbracedSpecifiers = [];
              const bracedSpecifiers = [];

              path.each(function (specifierPath) {
                  const spec = specifierPath.getValue();
                  if (spec.type === "ExportDefaultSpecifier") {
                      unbracedSpecifiers.push(print(specifierPath));
                  } else {
                      bracedSpecifiers.push(print(specifierPath));
                  }
              }, "specifiers");

              unbracedSpecifiers.forEach((lines$$1, i) => {
                  if (i > 0) {
                      parts.push(", ");
                  }
                  parts.push(lines$$1);
              });

              if (bracedSpecifiers.length > 0) {
                  let lines$$1 = fromString$3(", ").join(bracedSpecifiers);
                  if (lines$$1.getLineLength(1) > options$$1.wrapColumn) {
                      lines$$1 = concat$2([
                          fromString$3(",\n").join(
                              bracedSpecifiers
                          ).indent(options$$1.tabWidth),
                          ","
                      ]);
                  }

                  if (unbracedSpecifiers.length > 0) {
                      parts.push(", ");
                  }

                  if (lines$$1.length > 1) {
                      parts.push("{\n", lines$$1, "\n}");
                  } else if (options$$1.objectCurlySpacing) {
                      parts.push("{ ", lines$$1, " }");
                  } else {
                      parts.push("{", lines$$1, "}");
                  }
              }

          } else {
              parts.push(
                  shouldPrintSpaces ? "{ " : "{",
                  fromString$3(", ").join(path.map(print, "specifiers")),
                  shouldPrintSpaces ? " }" : "}"
              );
          }

          if (decl.source) {
              parts.push(" from ", path.call(print, "source"));
          }
      }

      var lines$$1 = concat$2(parts);
      if (lastNonSpaceCharacter(lines$$1) !== ";" &&
          ! (decl.declaration &&
             (decl.declaration.type === "FunctionDeclaration" ||
              decl.declaration.type === "ClassDeclaration" ||
              decl.declaration.type === "TSModuleDeclaration" ||
              decl.declaration.type === "TSInterfaceDeclaration" ||
              decl.declaration.type === "TSEnumDeclaration"))) {
          lines$$1 = concat$2([lines$$1, ";"]);
      }
      return lines$$1;
  }

  function printFlowDeclaration(path, parts) {
      var parentExportDecl = util_1$1.getParentExportDeclaration(path);

      if (parentExportDecl) ; else {
          // If the parent node has type DeclareExportDeclaration, then it
          // will be responsible for printing the "declare" token. Otherwise
          // it needs to be printed with this non-exported declaration node.
          parts.unshift("declare ");
      }

      return concat$2(parts);
  }

  function printVariance(path, print) {
      return path.call(function (variancePath) {
          var value = variancePath.getValue();

          if (value) {
              if (value === "plus") {
                  return fromString$3("+");
              }

              if (value === "minus") {
                  return fromString$3("-");
              }

              return print(variancePath);
          }

          return fromString$3("");
      }, "variance");
  }

  function adjustClause(clause, options$$1) {
      if (clause.length > 1)
          return concat$2([" ", clause]);

      return concat$2([
          "\n",
          maybeAddSemicolon(clause).indent(options$$1.tabWidth)
      ]);
  }

  function lastNonSpaceCharacter(lines$$1) {
      var pos = lines$$1.lastPos();
      do {
          var ch = lines$$1.charAt(pos);
          if (/\S/.test(ch))
              return ch;
      } while (lines$$1.prevPos(pos));
  }

  function endsWithBrace(lines$$1) {
      return lastNonSpaceCharacter(lines$$1) === "}";
  }

  function swapQuotes(str) {
      return str.replace(/['"]/g, function(m) {
          return m === '"' ? '\'' : '"';
      });
  }

  function nodeStr(str, options$$1) {
      isString$4.assert(str);
      switch (options$$1.quote) {
      case "auto":
          var double = JSON.stringify(str);
          var single = swapQuotes(JSON.stringify(swapQuotes(str)));
          return double.length > single.length ? single : double;
      case "single":
          return swapQuotes(JSON.stringify(swapQuotes(str)));
      case "double":
      default:
          return JSON.stringify(str);
      }
  }

  function maybeAddSemicolon(lines$$1) {
      var eoc = lastNonSpaceCharacter(lines$$1);
      if (!eoc || "\n};".indexOf(eoc) < 0)
          return concat$2([lines$$1, ";"]);
      return lines$$1;
  }

  var printer = {
  	Printer: Printer_1
  };

  var main$1 = createCommonjsModule(function (module, exports) {
  var parse = parser.parse;
  var Printer = printer.Printer;

  function print(node, options) {
      return new Printer(options).print(node);
  }

  function prettyPrint(node, options) {
      return new Printer(options).printGenerically(node);
  }

  function run(transformer, options) {
      return runFile(process.argv[2], transformer, options);
  }

  function runFile(path, transformer, options) {
      require$$1.readFile(path, "utf-8", function(err, code) {
          if (err) {
              console.error(err);
              return;
          }

          runString(code, transformer, options);
      });
  }

  function defaultWriteback(output) {
      process.stdout.write(output);
  }

  function runString(code, transformer, options) {
      var writeback = options && options.writeback || defaultWriteback;
      transformer(parse(code, options), function(node) {
          writeback(print(node, options).code);
      });
  }

  Object.defineProperties(exports, {
      /**
       * Parse a string of code into an augmented syntax tree suitable for
       * arbitrary modification and reprinting.
       */
      parse: {
          enumerable: true,
          value: parse
      },

      /**
       * Traverse and potentially modify an abstract syntax tree using a
       * convenient visitor syntax:
       *
       *   recast.visit(ast, {
       *     names: [],
       *     visitIdentifier: function(path) {
       *       var node = path.value;
       *       this.visitor.names.push(node.name);
       *       this.traverse(path);
       *     }
       *   });
       */
      visit: {
          enumerable: true,
          value: types$1.visit
      },

      /**
       * Reprint a modified syntax tree using as much of the original source
       * code as possible.
       */
      print: {
          enumerable: true,
          value: print
      },

      /**
       * Print without attempting to reuse any original source code.
       */
      prettyPrint: {
          enumerable: false,
          value: prettyPrint
      },

      /**
       * Customized version of require("ast-types").
       */
      types: {
          enumerable: false,
          value: types$1
      },

      /**
       * Convenient command-line interface (see e.g. example/add-braces).
       */
      run: {
          enumerable: false,
          value: run
      }
  });
  });

  /**
   * Not all the types are handled in this module.
   *
   * @enum {number}
   * @readonly
   */
  const TAG$1 = 1; /* TAG */
  const ATTR = 2; /* ATTR */
  const TEXT$1 = 3; /* TEXT */
  const CDATA = 4; /* CDATA */
  const COMMENT = 8; /* COMMENT */
  const DOCUMENT = 9; /* DOCUMENT */
  const DOCTYPE = 10; /* DOCTYPE */
  const DOCUMENT_FRAGMENT = 11; /* DOCUMENT_FRAGMENT */

  var types$2 = /*#__PURE__*/Object.freeze({
    TAG: TAG$1,
    ATTR: ATTR,
    TEXT: TEXT$1,
    CDATA: CDATA,
    COMMENT: COMMENT,
    DOCUMENT: DOCUMENT,
    DOCTYPE: DOCTYPE,
    DOCUMENT_FRAGMENT: DOCUMENT_FRAGMENT
  });

  const rootTagNotFound = 'Root tag not found.';
  const unclosedTemplateLiteral = 'Unclosed ES6 template literal.';
  const unexpectedEndOfFile = 'Unexpected end of file.';
  const unclosedComment = 'Unclosed comment.';
  const unclosedNamedBlock = 'Unclosed "%1" block.';
  const duplicatedNamedTag = 'Duplicate tag "<%1>".';
  const unexpectedCharInExpression = 'Unexpected character %1.';
  const unclosedExpression = 'Unclosed expression.';

  /**
   * Matches the start of valid tags names; used with the first 2 chars after the `'<'`.
   * @const
   * @private
   */
  const TAG_2C = /^(?:\/[a-zA-Z]|[a-zA-Z][^\s>/]?)/;
  /**
   * Matches valid tags names AFTER the validation with `TAG_2C`.
   * $1: tag name including any `'/'`, $2: non self-closing brace (`>`) w/o attributes.
   * @const
   * @private
   */
  const TAG_NAME = /(\/?[^\s>/]+)\s*(>)?/g;
  /**
   * Matches an attribute name-value pair (both can be empty).
   * $1: attribute name, $2: value including any quotes.
   * @const
   * @private
   */
  const ATTR_START = /(\S[^>/=\s]*)(?:\s*=\s*([^>/])?)?/g;

  /**
   * Matches the spread operator
   * it will be used for the spread attributes
   * @type {RegExp}
   */
  const SPREAD_OPERATOR = /\.\.\./;
  /**
   * Matches the closing tag of a `script` and `style` block.
   * Used by parseText fo find the end of the block.
   * @const
   * @private
   */
  const RE_SCRYLE = {
    script: /<\/script\s*>/gi,
    style: /<\/style\s*>/gi,
    textarea: /<\/textarea\s*>/gi
  };

  // Do not touch text content inside this tags
  const RAW_TAGS = /^\/?(?:pre|textarea)$/;

  const JAVASCRIPT_OUTPUT_NAME = 'javascript';
  const CSS_OUTPUT_NAME = 'css';
  const TEMPLATE_OUTPUT_NAME = 'template';

  // Tag names
  const JAVASCRIPT_TAG = 'script';
  const STYLE_TAG = 'style';
  const TEXTAREA_TAG = 'textarea';

  // Boolean attributes
  const IS_RAW = 'isRaw';
  const IS_SELF_CLOSING = 'isSelfClosing';
  const IS_VOID = 'isVoid';
  const IS_BOOLEAN = 'isBoolean';
  const IS_CUSTOM = 'isCustom';
  const IS_SPREAD = 'isSpread';

  /**
   * Add an item into a collection, if the collection is not an array
   * we create one and add the item to it
   * @param   {Array} collection - target collection
   * @param   {*} item - item to add to the collection
   * @returns {Array} array containing the new item added to it
   */
  function addToCollection(collection = [], item) {
    collection.push(item);
    return collection
  }

  /**
   * Run RegExp.exec starting from a specific position
   * @param   {RegExp} re - regex
   * @param   {number} pos - last index position
   * @param   {string} string - regex target
   * @returns {Array} regex result
   */
  function execFromPos(re, pos, string) {
    re.lastIndex = pos;
    return re.exec(string)
  }

  /**
   * Escape special characters in a given string, in preparation to create a regex.
   *
   * @param   {string} str - Raw string
   * @returns {string} Escaped string.
   */
  var escapeStr = (str) => str.replace(/(?=[-[\](){^*+?.$|\\])/g, '\\');

  function formatError(data, message, pos) {
    if (!pos) {
      pos = data.length;
    }
    // count unix/mac/win eols
    const line = (data.slice(0, pos).match(/\r\n?|\n/g) || '').length + 1;
    let col = 0;
    while (--pos >= 0 && !/[\r\n]/.test(data[pos])) {
      ++col;
    }
    return `[${line},${col}]: ${message}`
  }

  const $_ES6_BQ = '`';

  /**
   * Searches the next backquote that signals the end of the ES6 Template Literal
   * or the "${" sequence that starts a JS expression, skipping any escaped
   * character.
   *
   * @param   {string}    code  - Whole code
   * @param   {number}    pos   - The start position of the template
   * @param   {string[]}  stack - To save nested ES6 TL count
   * @returns {number}    The end of the string (-1 if not found)
   */
  function skipES6TL(code, pos, stack) {
    // we are in the char following the backquote (`),
    // find the next unescaped backquote or the sequence "${"
    const re = /[`$\\]/g;
    let c;
    while (re.lastIndex = pos, re.exec(code)) {
      pos = re.lastIndex;
      c = code[pos - 1];
      if (c === '`') {
        return pos
      }
      if (c === '$' && code[pos++] === '{') {
        stack.push($_ES6_BQ, '}');
        return pos
      }
      // else this is an escaped char
    }
    throw formatError(code, unclosedTemplateLiteral, pos)
  }

  /**
   * Custom error handler can be implemented replacing this method.
   * The `state` object includes the buffer (`data`)
   * The error position (`loc`) contains line (base 1) and col (base 0).
   * @param {string} data - string containing the error
   * @param {string} msg - Error message
   * @param {number} pos - Position of the error
   * @returns {undefined} throw an exception error
   */
  function panic$1(data, msg, pos) {
    const message = formatError(data, msg, pos);
    throw new Error(message)
  }

  // forked from https://github.com/aMarCruz/skip-regex

  // safe characters to precced a regex (including `=>`, `**`, and `...`)
  const beforeReChars = '[{(,;:?=|&!^~>%*/';
  const beforeReSign = `${beforeReChars}+-`;

  // keyword that can preceed a regex (`in` is handled as special case)
  const beforeReWords = [
    'case',
    'default',
    'do',
    'else',
    'in',
    'instanceof',
    'prefix',
    'return',
    'typeof',
    'void',
    'yield'
  ];

  // Last chars of all the beforeReWords elements to speed up the process.
  const wordsEndChar = beforeReWords.reduce((s, w) => s + w.slice(-1), '');

  // Matches literal regex from the start of the buffer.
  // The buffer to search must not include line-endings.
  const RE_LIT_REGEX = /^\/(?=[^*>/])[^[/\\]*(?:(?:\\.|\[(?:\\.|[^\]\\]*)*\])[^[\\/]*)*?\/[gimuy]*/;

  // Valid characters for JavaScript variable names and literal numbers.
  const RE_JS_VCHAR = /[$\w]/;

  // Match dot characters that could be part of tricky regex
  const RE_DOT_CHAR = /.*/g;

  /**
   * Searches the position of the previous non-blank character inside `code`,
   * starting with `pos - 1`.
   *
   * @param   {string} code - Buffer to search
   * @param   {number} pos  - Starting position
   * @returns {number} Position of the first non-blank character to the left.
   * @private
   */
  function _prev(code, pos) {
    while (--pos >= 0 && /\s/.test(code[pos]));
    return pos
  }



  /**
   * Check if the character in the `start` position within `code` can be a regex
   * and returns the position following this regex or `start+1` if this is not
   * one.
   *
   * NOTE: Ensure `start` points to a slash (this is not checked).
   *
   * @function skipRegex
   * @param   {string} code  - Buffer to test in
   * @param   {number} start - Position the first slash inside `code`
   * @returns {number} Position of the char following the regex.
   *
   */
  /* istanbul ignore next */
  function skipRegex(code, start) {
    let pos = RE_DOT_CHAR.lastIndex = start++;

    // `exec()` will extract from the slash to the end of the line
    //   and the chained `match()` will match the possible regex.
    const match = (RE_DOT_CHAR.exec(code) || ' ')[0].match(RE_LIT_REGEX);

    if (match) {
      const next = pos + match[0].length;      // result comes from `re.match`

      pos = _prev(code, pos);
      let c = code[pos];

      // start of buffer or safe prefix?
      if (pos < 0 || beforeReChars.includes(c)) {
        return next
      }

      // from here, `pos` is >= 0 and `c` is code[pos]
      if (c === '.') {
        // can be `...` or something silly like 5./2
        if (code[pos - 1] === '.') {
          start = next;
        }

      } else {

        if (c === '+' || c === '-') {
          // tricky case
          if (code[--pos] !== c ||            // if have a single operator or
               (pos = _prev(code, pos)) < 0 ||  // ...have `++` and no previous token
               beforeReSign.includes(c = code[pos])) {
            return next                       // ...this is a regex
          }
        }

        if (wordsEndChar.includes(c)) {  // looks like a keyword?
          const end = pos + 1;

          // get the complete (previous) keyword
          while (--pos >= 0 && RE_JS_VCHAR.test(code[pos]));

          // it is in the allowed keywords list?
          if (beforeReWords.includes(code.slice(pos + 1, end))) {
            start = next;
          }
        }
      }
    }

    return start
  }

  /*
   * Mini-parser for expressions.
   * The main pourpose of this module is to find the end of an expression
   * and return its text without the enclosing brackets.
   * Does not works with comments, but supports ES6 template strings.
   */
  /**
   * @exports exprExtr
   */
  const S_SQ_STR = /'[^'\n\r\\]*(?:\\(?:\r\n?|[\S\s])[^'\n\r\\]*)*'/.source;
  /**
   * Matches double quoted JS strings taking care about nested quotes
   * and EOLs (escaped EOLs are Ok).
   *
   * @const
   * @private
   */
  const S_STRING = `${S_SQ_STR}|${S_SQ_STR.replace(/'/g, '"')}`;
  /**
   * Regex cache
   *
   * @type {Object.<string, RegExp>}
   * @const
   * @private
   */
  const reBr = {};
  /**
   * Makes an optimal regex that matches quoted strings, brackets, backquotes
   * and the closing brackets of an expression.
   *
   * @param   {string} b - Closing brackets
   * @returns {RegExp} - optimized regex
   */
  function _regex(b) {
    let re = reBr[b];
    if (!re) {
      let s = escapeStr(b);
      if (b.length > 1) {
        s = `${s}|[`;
      } else {
        s = /[{}[\]()]/.test(b) ? '[' : `[${s}`;
      }
      reBr[b] = re = new RegExp(`${S_STRING}|${s}\`/\\{}[\\]()]`, 'g');
    }
    return re
  }

  /**
   * Update the scopes stack removing or adding closures to it
   * @param   {Array} stack - array stacking the expression closures
   * @param   {string} char - current char to add or remove from the stack
   * @param   {string} idx  - matching index
   * @param   {string} code - expression code
   * @returns {Object} result
   * @returns {Object} result.char - either the char received or the closing braces
   * @returns {Object} result.index - either a new index to skip part of the source code,
   *                                  or 0 to keep from parsing from the old position
   */
  function updateStack(stack, char, idx, code) {
    let index = 0;

    switch (char) {
    case '[':
    case '(':
    case '{':
      stack.push(char === '[' ? ']' : char === '(' ? ')' : '}');
      break
    case ')':
    case ']':
    case '}':
      if (char !== stack.pop()) {
        panic$1(code, unexpectedCharInExpression.replace('%1', char), index);
      }

      if (char === '}' && stack[stack.length - 1] === $_ES6_BQ) {
        char = stack.pop();
      }

      index = idx + 1;
      break
    case '/':
      index = skipRegex(code, idx);
    }

    return { char, index }
  }

  /**
     * Parses the code string searching the end of the expression.
     * It skips braces, quoted strings, regexes, and ES6 template literals.
     *
     * @function exprExtr
     * @param   {string}  code  - Buffer to parse
     * @param   {number}  start - Position of the opening brace
     * @param   {[string,string]} bp - Brackets pair
     * @returns {Object} Expression's end (after the closing brace) or -1
     *                            if it is not an expr.
     */
  function exprExtr(code, start, bp) {
    const [openingBraces, closingBraces] = bp;
    const offset = start + openingBraces.length; // skips the opening brace
    const stack = []; // expected closing braces ('`' for ES6 TL)
    const re = _regex(closingBraces);

    re.lastIndex = offset; // begining of the expression

    let end;
    let match;

    while (match = re.exec(code)) { // eslint-disable-line
      const idx = match.index;
      const str = match[0];
      end = re.lastIndex;

      // end the iteration
      if (str === closingBraces && !stack.length) {
        return {
          text: code.slice(offset, idx),
          start,
          end
        }
      }

      const { char, index } = updateStack(stack, str[0], idx, code);
      // update the end value depending on the new index received
      end = index || end;
      // update the regex last index
      re.lastIndex = char === $_ES6_BQ ? skipES6TL(code, end, stack) : end;
    }

    if (stack.length) {
      panic$1(code, unclosedExpression, end);
    }
  }

  /**
   * Outputs the last parsed node. Can be used with a builder too.
   *
   * @param   {ParserStore} store - Parsing store
   * @returns {undefined} void function
   * @private
   */
  function flush(store) {
    const last = store.last;
    store.last = null;
    if (last && store.root) {
      store.builder.push(last);
    }
  }

  /**
   * Get the code chunks from start and end range
   * @param   {string}  source  - source code
   * @param   {number}  start   - Start position of the chunk we want to extract
   * @param   {number}  end     - Ending position of the chunk we need
   * @returns {string}  chunk of code extracted from the source code received
   * @private
   */
  function getChunk(source, start, end) {
    return source.slice(start, end)
  }

  /**
   * states text in the last text node, or creates a new one if needed.
   *
   * @param {ParserState}   state   - Current parser state
   * @param {number}  start   - Start position of the tag
   * @param {number}  end     - Ending position (last char of the tag)
   * @param {Object}  extra   - extra properties to add to the text node
   * @param {RawExpr[]} extra.expressions  - Found expressions
   * @param {string}    extra.unescape     - Brackets to unescape
   * @returns {undefined} - void function
   * @private
   */
  function pushText(state, start, end, extra = {}) {
    const text = getChunk(state.data, start, end);
    const expressions = extra.expressions;
    const unescape = extra.unescape;

    let q = state.last;
    state.pos = end;

    if (q && q.type === TEXT$1) {
      q.text += text;
      q.end = end;
    } else {
      flush(state);
      state.last = q = { type: TEXT$1, text, start, end };
    }

    if (expressions && expressions.length) {
      q.expressions = (q.expressions || []).concat(expressions);
    }

    if (unescape) {
      q.unescape = unescape;
    }

    return TEXT$1
  }

  /**
   * Find the end of the attribute value or text node
   * Extract expressions.
   * Detect if value have escaped brackets.
   *
   * @param   {ParserState} state  - Parser state
   * @param   {HasExpr} node       - Node if attr, info if text
   * @param   {string} endingChars - Ends the value or text
   * @param   {number} start       - Starting position
   * @returns {number} Ending position
   * @private
   */
  function expr(state, node, endingChars, start) {
    const re = b0re(state, endingChars);

    re.lastIndex = start; // reset re position

    const { unescape, expressions, end } = parseExpressions(state, re);

    if (node) {
      if (unescape) {
        node.unescape = unescape;
      }
      if (expressions.length) {
        node.expressions = expressions;
      }
    } else {
      pushText(state, start, end, {expressions, unescape});
    }

    return end
  }

  /**
   * Parse a text chunk finding all the expressions in it
   * @param   {ParserState} state  - Parser state
   * @param   {RegExp} re - regex to match the expressions contents
   * @returns {Object} result containing the expression found, the string to unescape and the end position
   */
  function parseExpressions(state, re) {
    const { data, options } = state;
    const { brackets } = options;
    const expressions = [];
    let unescape, pos, match;

    // Anything captured in $1 (closing quote or character) ends the loop...
    while ((match = re.exec(data)) && !match[1]) {
      // ...else, we have an opening bracket and maybe an expression.
      pos = match.index;
      if (data[pos - 1] === '\\') {
        unescape = match[0]; // it is an escaped opening brace
      } else {
        const tmpExpr = exprExtr(data, pos, brackets);
        if (tmpExpr) {
          expressions.push(tmpExpr);
          re.lastIndex = tmpExpr.end;
        }
      }
    }

    // Even for text, the parser needs match a closing char
    if (!match) {
      panic$1(data, unexpectedEndOfFile, pos);
    }

    return {
      unescape,
      expressions,
      end: match.index
    }
  }



  /**
   * Creates a regex for the given string and the left bracket.
   * The string is captured in $1.
   *
   * @param   {ParserState} state  - Parser state
   * @param   {string} str - String to search
   * @returns {RegExp} Resulting regex.
   * @private
   */
  function b0re(state, str) {
    const { brackets } = state.options;
    const re = state.regexCache[str];

    if (re) return re

    const b0 = escapeStr(brackets[0]);
    // cache the regex extending the regexCache object
    Object.assign(state.regexCache, { [str]: new RegExp(`(${str})|${b0}`, 'g') });

    return state.regexCache[str]
  }

  /**
   * SVG void elements that cannot be auto-closed and shouldn't contain child nodes.
   * @const {Array}
   */
  const VOID_SVG_TAGS_LIST = [
    'circle',
    'ellipse',
    'line',
    'path',
    'polygon',
    'polyline',
    'rect',
    'stop',
    'use'
  ];

  /**
   * List of all the available svg tags
   * @const {Array}
   * @see {@link https://github.com/wooorm/svg-tag-names}
   */
  const SVG_TAGS_LIST = [
    'a',
    'altGlyph',
    'altGlyphDef',
    'altGlyphItem',
    'animate',
    'animateColor',
    'animateMotion',
    'animateTransform',
    'animation',
    'audio',
    'canvas',
    'clipPath',
    'color-profile',
    'cursor',
    'defs',
    'desc',
    'discard',
    'feBlend',
    'feColorMatrix',
    'feComponentTransfer',
    'feComposite',
    'feConvolveMatrix',
    'feDiffuseLighting',
    'feDisplacementMap',
    'feDistantLight',
    'feDropShadow',
    'feFlood',
    'feFuncA',
    'feFuncB',
    'feFuncG',
    'feFuncR',
    'feGaussianBlur',
    'feImage',
    'feMerge',
    'feMergeNode',
    'feMorphology',
    'feOffset',
    'fePointLight',
    'feSpecularLighting',
    'feSpotLight',
    'feTile',
    'feTurbulence',
    'filter',
    'font',
    'font-face',
    'font-face-format',
    'font-face-name',
    'font-face-src',
    'font-face-uri',
    'foreignObject',
    'g',
    'glyph',
    'glyphRef',
    'handler',
    'hatch',
    'hatchpath',
    'hkern',
    'iframe',
    'image',
    'linearGradient',
    'listener',
    'marker',
    'mask',
    'mesh',
    'meshgradient',
    'meshpatch',
    'meshrow',
    'metadata',
    'missing-glyph',
    'mpath',
    'pattern',
    'prefetch',
    'radialGradient',
    'script',
    'set',
    'solidColor',
    'solidcolor',
    'style',
    'svg',
    'switch',
    'symbol',
    'tbreak',
    'text',
    'textArea',
    'textPath',
    'title',
    'tref',
    'tspan',
    'unknown',
    'video',
    'view',
    'vkern'
  ].concat(VOID_SVG_TAGS_LIST).sort();

  /**
   * HTML void elements that cannot be auto-closed and shouldn't contain child nodes.
   * @type {Array}
   * @see   {@link http://www.w3.org/TR/html-markup/syntax.html#syntax-elements}
   * @see   {@link http://www.w3.org/TR/html5/syntax.html#void-elements}
   */
  const VOID_HTML_TAGS_LIST = [
    'area',
    'base',
    'br',
    'col',
    'embed',
    'hr',
    'img',
    'input',
    'keygen',
    'link',
    'menuitem',
    'meta',
    'param',
    'source',
    'track',
    'wbr'
  ];

  /**
   * List of all the html tags
   * @const {Array}
   * @see {@link https://github.com/sindresorhus/html-tags}
   */
  const HTML_TAGS_LIST = [
    'a',
    'abbr',
    'address',
    'article',
    'aside',
    'audio',
    'b',
    'bdi',
    'bdo',
    'blockquote',
    'body',
    'button',
    'canvas',
    'caption',
    'cite',
    'code',
    'colgroup',
    'data',
    'datalist',
    'dd',
    'del',
    'details',
    'dfn',
    'dialog',
    'div',
    'dl',
    'dt',
    'em',
    'fieldset',
    'figcaption',
    'figure',
    'footer',
    'form',
    'h1',
    'h2',
    'h3',
    'h4',
    'h5',
    'h6',
    'head',
    'header',
    'hgroup',
    'html',
    'i',
    'iframe',
    'ins',
    'kbd',
    'label',
    'legend',
    'li',
    'main',
    'map',
    'mark',
    'math',
    'menu',
    'meter',
    'nav',
    'noscript',
    'object',
    'ol',
    'optgroup',
    'option',
    'output',
    'p',
    'picture',
    'pre',
    'progress',
    'q',
    'rb',
    'rp',
    'rt',
    'rtc',
    'ruby',
    's',
    'samp',
    'script',
    'section',
    'select',
    'slot',
    'small',
    'span',
    'strong',
    'style',
    'sub',
    'summary',
    'sup',
    'svg',
    'table',
    'tbody',
    'td',
    'template',
    'textarea',
    'tfoot',
    'th',
    'thead',
    'time',
    'title',
    'tr',
    'u',
    'ul',
    'var',
    'video'
  ].concat(VOID_HTML_TAGS_LIST).sort();

  /**
   * Matches boolean HTML attributes in the riot tag definition.
   * With a long list like this, a regex is faster than `[].indexOf` in most browsers.
   * @const {RegExp}
   * @see [attributes.md](https://github.com/riot/compiler/blob/dev/doc/attributes.md)
   */
  const BOOLEAN_ATTRIBUTES_LIST = [
    'disabled',
    'visible',
    'checked',
    'readonly',
    'required',
    'allowfullscreen',
    'autofocus',
    'autoplay',
    'compact',
    'controls',
    'default',
    'formnovalidate',
    'hidden',
    'ismap',
    'itemscope',
    'loop',
    'multiple',
    'muted',
    'noresize',
    'noshade',
    'novalidate',
    'nowrap',
    'open',
    'reversed',
    'seamless',
    'selected',
    'sortable',
    'truespeed',
    'typemustmatch'
  ];

  /**
   * Join a list of items with the pipe symbol (usefull for regex list concatenation)
   * @private
   * @param   {Array} list - list of strings
   * @returns {String} the list received joined with pipes
   */
  function joinWithPipe(list) {
    return list.join('|')
  }

  /**
   * Convert list of strings to regex in order to test against it ignoring the cases
   * @private
   * @param   {...Array} lists - array of strings
   * @returns {RegExp} regex that will match all the strings in the array received ignoring the cases
   */
  function listsToRegex(...lists) {
    return new RegExp(`^/?(?:${joinWithPipe(lists.map(joinWithPipe))})$`, 'i')
  }

  /**
   * Regex matching all the html tags ignoring the cases
   * @const {RegExp}
   */
  const HTML_TAGS_RE = listsToRegex(HTML_TAGS_LIST);

  /**
   * Regex matching all the svg tags ignoring the cases
   * @const {RegExp}
   */
  const SVG_TAGS_RE = listsToRegex(SVG_TAGS_LIST);

  /**
   * Regex matching all the void html tags ignoring the cases
   * @const {RegExp}
   */
  const VOID_HTML_TAGS_RE =  listsToRegex(VOID_HTML_TAGS_LIST);

  /**
   * Regex matching all the void svg tags ignoring the cases
   * @const {RegExp}
   */
  const VOID_SVG_TAGS_RE =  listsToRegex(VOID_SVG_TAGS_LIST);

  /**
   * Regex matching all the boolean attributes
   * @const {RegExp}
   */
  const BOOLEAN_ATTRIBUTES_RE =  listsToRegex(BOOLEAN_ATTRIBUTES_LIST);

  /**
   * True if it's a self closing tag
   * @param   {String}  tag - test tag
   * @returns {Boolean}
   * @example
   * isVoid('meta') // true
   * isVoid('circle') // true
   * isVoid('IMG') // true
   * isVoid('div') // false
   * isVoid('mask') // false
   */
  function isVoid(tag) {
    return [
      VOID_HTML_TAGS_RE,
      VOID_SVG_TAGS_RE
    ].some(r => r.test(tag))
  }

  /**
   * True if it's not SVG nor a HTML known tag
   * @param   {String}  tag - test tag
   * @returns {Boolean}
   * @example
   * isCustom('my-component') // true
   * isCustom('div') // false
   */
  function isCustom(tag) {
    return [
      HTML_TAGS_RE,
      SVG_TAGS_RE
    ].every(l => !l.test(tag))
  }

  /**
   * True if it's a boolean attribute
   * @param   {String} attribute - test attribute
   * @returns {Boolean}
   * @example
   * isBoolAttribute('selected') // true
   * isBoolAttribute('class') // false
   */
  function isBoolAttribute(attribute) {
    return BOOLEAN_ATTRIBUTES_RE.test(attribute)
  }

  /**
   * The more complex parsing is for attributes as it can contain quoted or
   * unquoted values or expressions.
   *
   * @param   {ParserStore} state  - Parser state
   * @returns {number} New parser mode.
   * @private
   */
  function attr(state) {
    const { data, last, pos, root } = state;
    const tag = last; // the last (current) tag in the output
    const _CH = /\S/g; // matches the first non-space char
    const ch = execFromPos(_CH, pos, data);

    switch (true) {
    case !ch:
      state.pos = data.length; // reaching the end of the buffer with
      // NodeTypes.ATTR will generate error
      break
    case ch[0] === '>':
      // closing char found. If this is a self-closing tag with the name of the
      // Root tag, we need decrement the counter as we are changing mode.
      state.pos = tag.end = _CH.lastIndex;
      if (tag[IS_SELF_CLOSING]) {
        state.scryle = null; // allow selfClosing script/style tags
        if (root && root.name === tag.name) {
          state.count--; // "pop" root tag
        }
      }
      return TEXT$1
    case ch[0] === '/':
      state.pos = _CH.lastIndex; // maybe. delegate the validation
      tag[IS_SELF_CLOSING] = true; // the next loop
      break
    default:
      delete tag[IS_SELF_CLOSING]; // ensure unmark as selfclosing tag
      setAttribute$1(state, ch.index, tag);
    }

    return ATTR
  }

  /**
   * Parses an attribute and its expressions.
   *
   * @param   {ParserStore}  state  - Parser state
   * @param   {number} pos    - Starting position of the attribute
   * @param   {Object} tag    - Current parent tag
   * @returns {undefined} void function
   * @private
   */
  function setAttribute$1(state, pos, tag) {
    const { data } = state;
    const re = ATTR_START; // (\S[^>/=\s]*)(?:\s*=\s*([^>/])?)? g
    const start = re.lastIndex = pos; // first non-whitespace
    const match = re.exec(data);

    if (match) {
      const end = re.lastIndex;
      const attr = parseAttribute(state, match, start, end);

      //assert(q && q.type === Mode.TAG, 'no previous tag for the attr!')
      // Pushes the attribute and shifts the `end` position of the tag (`last`).
      state.pos = tag.end = attr.end;
      tag.attributes = addToCollection(tag.attributes, attr);
    }
  }

  function parseNomalAttribute(state, attr, quote) {
    const { data } = state;
    let { end } = attr;

    if (isBoolAttribute(attr.name)) {
      attr[IS_BOOLEAN] = true;
    }

    // parse the whole value (if any) and get any expressions on it
    if (quote) {
      // Usually, the value's first char (`quote`) is a quote and the lastIndex
      // (`end`) is the start of the value.
      let valueStart = end;
      // If it not, this is an unquoted value and we need adjust the start.
      if (quote !== '"' && quote !== '\'') {
        quote = ''; // first char of value is not a quote
        valueStart--; // adjust the starting position
      }

      end = expr(state, attr, quote || '[>/\\s]', valueStart);

      // adjust the bounds of the value and save its content
      return Object.assign(attr, {
        value: getChunk(data, valueStart, end),
        valueStart,
        end: quote ? ++end : end
      })
    }

    return attr
  }

  function parseSpreadAttribute(state, attr, quote) {
    let end = expr(state, attr, quote || '[>/\\s]', attr.start);

    return {
      [IS_SPREAD]: true,
      start: attr.start,
      expressions: attr.expressions.map(expr$$1 => Object.assign(expr$$1, {
        text: expr$$1.text.replace(SPREAD_OPERATOR, '')
      })),
      end: quote ? ++end : end
    }
  }

  /**
   * Parse the attribute values normalising the quotes
   * @param   {ParserStore}  state  - Parser state
   * @param   {Array} match - results of the attributes regex
   * @param   {number} start - attribute start position
   * @param   {number} end - attribute end position
   * @returns {Object} attribute object
   */
  function parseAttribute(state, match, start, end) {
    const attr = {
      name: match[1],
      value: '',
      start,
      end
    };
    const quote = match[2]; // first letter of value or nothing

    if (SPREAD_OPERATOR.test(attr.name)) {
      return parseSpreadAttribute(state, attr, quote)
    }

    return parseNomalAttribute(state, attr, quote)
  }

  /**
   * Parses comments in long or short form
   * (any DOCTYPE & CDATA blocks are parsed as comments).
   *
   * @param   {ParserState} state  - Parser state
   * @param   {string} data       - Buffer to parse
   * @param   {number} start      - Position of the '<!' sequence
   * @returns {number} node type id
   * @private
   */
  function comment(state, data, start) {
    const pos = start + 2; // skip '<!'
    const str = data.substr(pos, 2) === '--' ? '-->' : '>';
    const end = data.indexOf(str, pos);
    if (end < 0) {
      panic$1(data, unclosedComment, start);
    }
    pushComment(state, start, end + str.length);

    return TEXT$1
  }

  /**
   * Parse a comment.
   *
   * @param   {ParserState}  state - Current parser state
   * @param   {number}  start - Start position of the tag
   * @param   {number}  end   - Ending position (last char of the tag)
   * @returns {undefined} void function
   * @private
   */
  function pushComment(state, start, end) {
    flush(state);
    state.pos = end;
    if (state.options.comments === true) {
      state.last = { type: COMMENT, start, end };
    }
  }

  /**
   * Pushes a new *tag* and set `last` to this, so any attributes
   * will be included on this and shifts the `end`.
   *
   * @param   {ParserState} state  - Current parser state
   * @param   {string}  name      - Name of the node including any slash
   * @param   {number}  start     - Start position of the tag
   * @param   {number}  end       - Ending position (last char of the tag + 1)
   * @returns {undefined} - void function
   * @private
   */
  function pushTag(state, name, start, end) {
    const root = state.root;
    const last = { type: TAG$1, name, start, end };

    if (isCustom(name)) {
      last[IS_CUSTOM] = true;
    }

    if (isVoid(name)) {
      last[IS_VOID] = true;
    }

    state.pos = end;

    if (root) {
      if (name === root.name) {
        state.count++;
      } else if (name === root.close) {
        state.count--;
      }
      flush(state);
    } else {
      // start with root (keep ref to output)
      state.root = { name: last.name, close: `/${name}` };
      state.count = 1;
    }

    state.last = last;
  }

  /**
   * Parse the tag following a '<' character, or delegate to other parser
   * if an invalid tag name is found.
   *
   * @param   {ParserState} state  - Parser state
   * @returns {number} New parser mode
   * @private
   */
  function tag(state) {
    const { pos, data } = state; // pos of the char following '<'
    const start = pos - 1; // pos of '<'
    const str = data.substr(pos, 2); // first two chars following '<'

    switch (true) {
    case str[0] === '!':
      return comment(state, data, start)
    case TAG_2C.test(str):
      return parseTag(state, start)
    default:
      return pushText(state, start, pos) // pushes the '<' as text
    }
  }

  function parseTag(state, start) {
    const { data, pos } = state;
    const re = TAG_NAME; // (\/?[^\s>/]+)\s*(>)? g
    const match = execFromPos(re, pos, data);
    const end = re.lastIndex;
    const name = match[1].toLowerCase(); // $1: tag name including any '/'
    // script/style block is parsed as another tag to extract attributes
    if (name in RE_SCRYLE) {
      state.scryle = name; // used by parseText
    }

    pushTag(state, name, start, end);
    // only '>' can ends the tag here, the '/' is handled in parseAttribute
    if (!match[2]) {
      return ATTR
    }

    return TEXT$1
  }

  /**
   * Parses regular text and script/style blocks ...scryle for short :-)
   * (the content of script and style is text as well)
   *
   * @param   {ParserState} state - Parser state
   * @returns {number} New parser mode.
   * @private
   */
  function text(state) {
    const { pos, data, scryle } = state;

    switch (true) {
    case typeof scryle === 'string': {
      const name = scryle;
      const re = RE_SCRYLE[name];
      const match = execFromPos(re, pos, data);

      if (!match) {
        panic$1(data, unclosedNamedBlock.replace('%1', name), pos - 1);
      }

      const start = match.index;
      const end = re.lastIndex;
      state.scryle = null; // reset the script/style flag now
      // write the tag content, if any
      if (start > pos) {
        parseSpecialTagsContent(state, name, match);
      }
      // now the closing tag, either </script> or </style>
      pushTag(state, `/${name}`, start, end);
      break
    }
    case data[pos] === '<':
      state.pos++;
      return TAG$1
    default:
      expr(state, null, '<', pos);
    }

    return TEXT$1
  }

  /**
   * Parse the text content depending on the name
   * @param   {ParserState} state - Parser state
   * @param   {string} name  - one of the tags matched by the RE_SCRYLE regex
   * @param   {Array}  match - result of the regex matching the content of the parsed tag
   * @returns {undefined} void function
   */
  function parseSpecialTagsContent(state, name, match) {
    const { pos } = state;
    const start = match.index;

    if (name === TEXTAREA_TAG) {
      expr(state, null, match[0], pos);
    } else {
      pushText(state, pos, start);
    }
  }

  /*---------------------------------------------------------------------
   * Tree builder for the riot tag parser.
   *
   * The output has a root property and separate arrays for `html`, `css`,
   * and `js` tags.
   *
   * The root tag is included as first element in the `html` array.
   * Script tags marked with "defer" are included in `html` instead `js`.
   *
   * - Mark SVG tags
   * - Mark raw tags
   * - Mark void tags
   * - Split prefixes from expressions
   * - Unescape escaped brackets and escape EOLs and backslashes
   * - Compact whitespace (option `compact`) for non-raw tags
   * - Create an array `parts` for text nodes and attributes
   *
   * Throws on unclosed tags or closing tags without start tag.
   * Selfclosing and void tags has no nodes[] property.
   */

  /**
   * Escape the carriage return and the line feed from a string
   * @param   {string} string - input string
   * @returns {string} output string escaped
   */
  function escapeReturn(string) {
    return string
      .replace(/\r/g, '\\r')
      .replace(/\n/g, '\\n')
  }

  /**
   * Escape double slashes in a string
   * @param   {string} string - input string
   * @returns {string} output string escaped
   */
  function escapeSlashes(string) {
    return string.replace(/\\/g, '\\\\')
  }

  /**
   * Replace the multiple spaces with only one
   * @param   {string} string - input string
   * @returns {string} string without trailing spaces
   */
  function cleanSpaces(string) {
    return string.replace(/\s+/g, ' ')
  }

  const TREE_BUILDER_STRUCT = Object.seal({
    get() {
      const store = this.store;
      // The real root tag is in store.root.nodes[0]
      return {
        [TEMPLATE_OUTPUT_NAME]: store.root.nodes[0],
        [CSS_OUTPUT_NAME]: store[STYLE_TAG],
        [JAVASCRIPT_OUTPUT_NAME]: store[JAVASCRIPT_TAG]
      }
    },

    /**
    * Process the current tag or text.
    * @param {Object} node - Raw pseudo-node from the parser
    * @returns {undefined} void function
    */
    push(node) {
      const store = this.store;

      switch (node.type) {
      case TEXT$1:
        this.pushText(store, node);
        break
      case TAG$1: {
        const name = node.name;
        if (name[0] === '/') {
          this.closeTag(store, node, name);
        } else {
          this.openTag(store, node);
        }
        break
      }
      }
    },
    closeTag(store, node) {
      const last = store.scryle || store.last;

      last.end = node.end;

      if (store.scryle) {
        store.scryle = null;
      } else {
        store.last = store.stack.pop();
      }
    },

    openTag(store, node) {
      const name = node.name;
      const attrs = node.attributes;

      if ([JAVASCRIPT_TAG, STYLE_TAG].includes(name)) {
        // Only accept one of each
        if (store[name]) {
          panic$1(this.store.data, duplicatedNamedTag.replace('%1', name), node.start);
        }

        store[name] = node;
        store.scryle = store[name];

      } else {
        // store.last holds the last tag pushed in the stack and this are
        // non-void, non-empty tags, so we are sure the `lastTag` here
        // have a `nodes` property.
        const lastTag = store.last;
        const newNode = node;

        lastTag.nodes.push(newNode);

        if (lastTag[IS_RAW] || RAW_TAGS.test(name)) {
          node[IS_RAW] = true;
        }

        if (!node[IS_SELF_CLOSING] && !node[IS_VOID]) {
          store.stack.push(lastTag);
          newNode.nodes = [];
          store.last = newNode;
        }
      }

      if (attrs) {
        this.attrs(attrs);
      }
    },
    attrs(attributes) {
      attributes.forEach(attr => {
        if (attr.value) {
          this.split(attr, attr.value, attr.valueStart, true);
        }
      });
    },
    pushText(store, node) {
      const text = node.text;
      const empty = !/\S/.test(text);
      const scryle = store.scryle;
      if (!scryle) {
        // store.last always have a nodes property
        const parent = store.last;
        const pack = this.compact && !parent[IS_RAW];
        if (pack && empty) {
          return
        }
        this.split(node, text, node.start, pack);
        parent.nodes.push(node);
      } else if (!empty) {
        scryle.text = node;
      }
    },
    split(node, source, start, pack) {
      const expressions = node.expressions;
      const parts = [];

      if (expressions) {
        let pos = 0;

        expressions.forEach(expr => {
          const text = source.slice(pos, expr.start - start);
          const code = expr.text;
          parts.push(this.sanitise(node, text, pack), escapeReturn(escapeSlashes(code).trim()));
          pos = expr.end - start;
        });

        if ((pos += start) < node.end) {
          parts.push(this.sanitise(node, source.slice(pos), pack));
        }
      } else {
        parts[0] = this.sanitise(node, source, pack);
      }

      node.parts = parts.filter(p => p); // remove the empty strings
    },
    // unescape escaped brackets and split prefixes of expressions
    sanitise(node, text, pack) {
      let rep = node.unescape;
      if (rep) {
        let idx = 0;
        rep = `\\${rep}`;
        while ((idx = text.indexOf(rep, idx)) !== -1) {
          text = text.substr(0, idx) + text.substr(idx + 1);
          idx++;
        }
      }

      text = escapeSlashes(text);

      return pack ? cleanSpaces(text) : escapeReturn(text)
    }
  });

  function createTreeBuilder(data, options) {
    const root = {
      type: TAG$1,
      name: '',
      start: 0,
      end: 0,
      nodes: []
    };

    return Object.assign(Object.create(TREE_BUILDER_STRUCT), {
      compact: options.compact !== false,
      store: {
        last: root,
        stack: [],
        scryle: null,
        root,
        style: null,
        script: null,
        data
      }
    })
  }

  /**
   * Factory for the Parser class, exposing only the `parse` method.
   * The export adds the Parser class as property.
   *
   * @param   {Object}   options - User Options
   * @param   {Function} customBuilder - Tree builder factory
   * @returns {Function} Public Parser implementation.
   */
  function parser$1(options, customBuilder) {
    const state = curry(createParserState)(options, customBuilder || createTreeBuilder);
    return {
      parse: (data) => parse$2(state(data))
    }
  }

  /**
   * Create a new state object
   * @param   {Object} userOptions - parser options
   * @param   {Function} builder - Tree builder factory
   * @param   {string} data - data to parse
   * @returns {ParserState} it represents the current parser state
   */
  function createParserState(userOptions, builder, data) {
    const options = Object.assign({
      brackets: ['{', '}']
    }, userOptions);

    return {
      options,
      regexCache: {},
      pos: 0,
      count: -1,
      root: null,
      last: null,
      scryle: null,
      builder: builder(data, options),
      data
    }
  }

  /**
   * It creates a raw output of pseudo-nodes with one of three different types,
   * all of them having a start/end position:
   *
   * - TAG     -- Opening or closing tags
   * - TEXT    -- Raw text
   * - COMMENT -- Comments
   *
   * @param   {ParserState}  state - Current parser state
   * @returns {ParserResult} Result, contains data and output properties.
   */
  function parse$2(state) {
    const { data } = state;

    walk(state);
    flush(state);

    if (state.count) {
      panic$1(data, state.count > 0 ? unexpectedEndOfFile : rootTagNotFound, state.pos);
    }

    return {
      data,
      output: state.builder.get()
    }
  }

  /**
   * Parser walking recursive function
   * @param {ParserState}  state - Current parser state
   * @param {string} type - current parsing context
   * @returns {undefined} void function
   */
  function walk(state, type) {
    const { data } = state;
    // extend the state adding the tree builder instance and the initial data
    const length = data.length;

    // The "count" property is set to 1 when the first tag is found.
    // This becomes the root and precedent text or comments are discarded.
    // So, at the end of the parsing count must be zero.
    if (state.pos < length && state.count) {
      walk(state, eat(state, type));
    }
  }

  /**
   * Function to help iterating on the current parser state
   * @param {ParserState}  state - Current parser state
   * @param   {string} type - current parsing context
   * @returns {string} parsing context
   */
  function eat(state, type) {
    switch (type) {
    case TAG$1:
      return tag(state)
    case ATTR:
      return attr(state)
    default:
      return text(state)
    }
  }

  /**
   * The nodeTypes definition
   */
  const nodeTypes = types$2;

  /**
   * Cancel token
   * @private
   * @type { Symbol }
   */
  const CANCEL = Symbol();

  /**
   * Helper that can be returned by ruit function to cancel the tasks chain
   * @returns { Symbol } internal private constant
   * @example
   *
   * ruit(
   *   100,
   *   num => Math.random() * num
   *   num => num > 50 ? ruit.cancel() : num
   *   num => num - 2
   * ).then(result => {
   *   console.log(result) // here we will get only number lower than 50
   * })
   *
   */
  ruit.cancel = () => CANCEL;

  /**
   * The same as ruit() but with the arguments inverted from right to left
   * @param   { * } tasks - list of tasks to process sequentially
   * @returns { Promise } a promise containing the result of the whole chain
   * @example
   *
   * const curry = f => a => b => f(a, b)
   * const add = (a, b) => a + b
   *
   * const addOne = curry(add)(1)
   *
   * const squareAsync = (num) => {
   *   return new Promise(r => {
   *     setTimeout(r, 500, num * 2)
   *   })
   * }
   *
   * // a -> a + a -> a * 2
   * // basically from right to left: 1 => 1 + 1 => 2 * 2
   * ruit.compose(squareAsync, addOne, 1).then(result => console.log(result)) // 4
   */
  ruit.compose = (...tasks) => ruit(...tasks.reverse());

  /**
   * Serialize a list of sync and async tasks from left to right
   * @param   { * } tasks - list of tasks to process sequentially
   * @returns { Promise } a promise containing the result of the whole chain
   * @example
   *
   * const curry = f => a => b => f(a, b)
   * const add = (a, b) => a + b
   *
   * const addOne = curry(add)(1)
   *
   * const squareAsync = (num) => {
   *   return new Promise(r => {
   *     setTimeout(r, 500, num * 2)
   *   })
   * }
   *
   * // a -> a + a -> a * 2
   * // basically from left to right: 1 => 1 + 1 => 2 * 2
   * ruit(1, addOne, squareAsync).then(result => console.log(result)) // 4
   */
  function ruit(...tasks) {
    return new Promise((resolve, reject) => {
      return (function run(queue, result) {
        if (!queue.length) return resolve(result)

        const [task, ...rest] = queue;
        const value = typeof task === 'function' ? task(result) : task;
        const done = v => run(rest, v);

        // check against nil values
        if (value != null) {
          if (value === CANCEL) return
          if (value.then) return value.then(done, reject)
        }

        return Promise.resolve(done(value))
      })(tasks)
    })
  }

  var builtin = {
  	"Array": false,
  	"ArrayBuffer": false,
  	Atomics: false,
  	BigInt: false,
  	BigInt64Array: false,
  	BigUint64Array: false,
  	"Boolean": false,
  	constructor: false,
  	"DataView": false,
  	"Date": false,
  	"decodeURI": false,
  	"decodeURIComponent": false,
  	"encodeURI": false,
  	"encodeURIComponent": false,
  	"Error": false,
  	"escape": false,
  	"eval": false,
  	"EvalError": false,
  	"Float32Array": false,
  	"Float64Array": false,
  	"Function": false,
  	hasOwnProperty: false,
  	"Infinity": false,
  	"Int16Array": false,
  	"Int32Array": false,
  	"Int8Array": false,
  	"isFinite": false,
  	"isNaN": false,
  	isPrototypeOf: false,
  	"JSON": false,
  	"Map": false,
  	"Math": false,
  	"NaN": false,
  	"Number": false,
  	"Object": false,
  	"parseFloat": false,
  	"parseInt": false,
  	"Promise": false,
  	propertyIsEnumerable: false,
  	"Proxy": false,
  	"RangeError": false,
  	"ReferenceError": false,
  	"Reflect": false,
  	"RegExp": false,
  	"Set": false,
  	SharedArrayBuffer: false,
  	"String": false,
  	"Symbol": false,
  	"SyntaxError": false,
  	toLocaleString: false,
  	toString: false,
  	"TypeError": false,
  	"Uint16Array": false,
  	"Uint32Array": false,
  	"Uint8Array": false,
  	"Uint8ClampedArray": false,
  	"undefined": false,
  	"unescape": false,
  	"URIError": false,
  	valueOf: false,
  	"WeakMap": false,
  	"WeakSet": false
  };
  var es5 = {
  	"Array": false,
  	"Boolean": false,
  	constructor: false,
  	"Date": false,
  	"decodeURI": false,
  	"decodeURIComponent": false,
  	"encodeURI": false,
  	"encodeURIComponent": false,
  	"Error": false,
  	"escape": false,
  	"eval": false,
  	"EvalError": false,
  	"Function": false,
  	hasOwnProperty: false,
  	"Infinity": false,
  	"isFinite": false,
  	"isNaN": false,
  	isPrototypeOf: false,
  	"JSON": false,
  	"Math": false,
  	"NaN": false,
  	"Number": false,
  	"Object": false,
  	"parseFloat": false,
  	"parseInt": false,
  	propertyIsEnumerable: false,
  	"RangeError": false,
  	"ReferenceError": false,
  	"RegExp": false,
  	"String": false,
  	"SyntaxError": false,
  	toLocaleString: false,
  	toString: false,
  	"TypeError": false,
  	"undefined": false,
  	"unescape": false,
  	"URIError": false,
  	valueOf: false
  };
  var es2015 = {
  	"Array": false,
  	"ArrayBuffer": false,
  	"Boolean": false,
  	constructor: false,
  	"DataView": false,
  	"Date": false,
  	"decodeURI": false,
  	"decodeURIComponent": false,
  	"encodeURI": false,
  	"encodeURIComponent": false,
  	"Error": false,
  	"escape": false,
  	"eval": false,
  	"EvalError": false,
  	"Float32Array": false,
  	"Float64Array": false,
  	"Function": false,
  	hasOwnProperty: false,
  	"Infinity": false,
  	"Int16Array": false,
  	"Int32Array": false,
  	"Int8Array": false,
  	"isFinite": false,
  	"isNaN": false,
  	isPrototypeOf: false,
  	"JSON": false,
  	"Map": false,
  	"Math": false,
  	"NaN": false,
  	"Number": false,
  	"Object": false,
  	"parseFloat": false,
  	"parseInt": false,
  	"Promise": false,
  	propertyIsEnumerable: false,
  	"Proxy": false,
  	"RangeError": false,
  	"ReferenceError": false,
  	"Reflect": false,
  	"RegExp": false,
  	"Set": false,
  	"String": false,
  	"Symbol": false,
  	"SyntaxError": false,
  	toLocaleString: false,
  	toString: false,
  	"TypeError": false,
  	"Uint16Array": false,
  	"Uint32Array": false,
  	"Uint8Array": false,
  	"Uint8ClampedArray": false,
  	"undefined": false,
  	"unescape": false,
  	"URIError": false,
  	valueOf: false,
  	"WeakMap": false,
  	"WeakSet": false
  };
  var es2017 = {
  	"Array": false,
  	"ArrayBuffer": false,
  	Atomics: false,
  	"Boolean": false,
  	constructor: false,
  	"DataView": false,
  	"Date": false,
  	"decodeURI": false,
  	"decodeURIComponent": false,
  	"encodeURI": false,
  	"encodeURIComponent": false,
  	"Error": false,
  	"escape": false,
  	"eval": false,
  	"EvalError": false,
  	"Float32Array": false,
  	"Float64Array": false,
  	"Function": false,
  	hasOwnProperty: false,
  	"Infinity": false,
  	"Int16Array": false,
  	"Int32Array": false,
  	"Int8Array": false,
  	"isFinite": false,
  	"isNaN": false,
  	isPrototypeOf: false,
  	"JSON": false,
  	"Map": false,
  	"Math": false,
  	"NaN": false,
  	"Number": false,
  	"Object": false,
  	"parseFloat": false,
  	"parseInt": false,
  	"Promise": false,
  	propertyIsEnumerable: false,
  	"Proxy": false,
  	"RangeError": false,
  	"ReferenceError": false,
  	"Reflect": false,
  	"RegExp": false,
  	"Set": false,
  	SharedArrayBuffer: false,
  	"String": false,
  	"Symbol": false,
  	"SyntaxError": false,
  	toLocaleString: false,
  	toString: false,
  	"TypeError": false,
  	"Uint16Array": false,
  	"Uint32Array": false,
  	"Uint8Array": false,
  	"Uint8ClampedArray": false,
  	"undefined": false,
  	"unescape": false,
  	"URIError": false,
  	valueOf: false,
  	"WeakMap": false,
  	"WeakSet": false
  };
  var browser = {
  	AbortController: false,
  	AbortSignal: false,
  	addEventListener: false,
  	alert: false,
  	AnalyserNode: false,
  	Animation: false,
  	AnimationEffectReadOnly: false,
  	AnimationEffectTiming: false,
  	AnimationEffectTimingReadOnly: false,
  	AnimationEvent: false,
  	AnimationPlaybackEvent: false,
  	AnimationTimeline: false,
  	applicationCache: false,
  	ApplicationCache: false,
  	ApplicationCacheErrorEvent: false,
  	atob: false,
  	Attr: false,
  	Audio: false,
  	AudioBuffer: false,
  	AudioBufferSourceNode: false,
  	AudioContext: false,
  	AudioDestinationNode: false,
  	AudioListener: false,
  	AudioNode: false,
  	AudioParam: false,
  	AudioProcessingEvent: false,
  	AudioScheduledSourceNode: false,
  	"AudioWorkletGlobalScope ": false,
  	AudioWorkletNode: false,
  	AudioWorkletProcessor: false,
  	BarProp: false,
  	BaseAudioContext: false,
  	BatteryManager: false,
  	BeforeUnloadEvent: false,
  	BiquadFilterNode: false,
  	Blob: false,
  	BlobEvent: false,
  	blur: false,
  	BroadcastChannel: false,
  	btoa: false,
  	BudgetService: false,
  	ByteLengthQueuingStrategy: false,
  	Cache: false,
  	caches: false,
  	CacheStorage: false,
  	cancelAnimationFrame: false,
  	cancelIdleCallback: false,
  	CanvasCaptureMediaStreamTrack: false,
  	CanvasGradient: false,
  	CanvasPattern: false,
  	CanvasRenderingContext2D: false,
  	ChannelMergerNode: false,
  	ChannelSplitterNode: false,
  	CharacterData: false,
  	clearInterval: false,
  	clearTimeout: false,
  	clientInformation: false,
  	ClipboardEvent: false,
  	close: false,
  	closed: false,
  	CloseEvent: false,
  	Comment: false,
  	CompositionEvent: false,
  	confirm: false,
  	console: false,
  	ConstantSourceNode: false,
  	ConvolverNode: false,
  	CountQueuingStrategy: false,
  	createImageBitmap: false,
  	Credential: false,
  	CredentialsContainer: false,
  	crypto: false,
  	Crypto: false,
  	CryptoKey: false,
  	CSS: false,
  	CSSConditionRule: false,
  	CSSFontFaceRule: false,
  	CSSGroupingRule: false,
  	CSSImportRule: false,
  	CSSKeyframeRule: false,
  	CSSKeyframesRule: false,
  	CSSMediaRule: false,
  	CSSNamespaceRule: false,
  	CSSPageRule: false,
  	CSSRule: false,
  	CSSRuleList: false,
  	CSSStyleDeclaration: false,
  	CSSStyleRule: false,
  	CSSStyleSheet: false,
  	CSSSupportsRule: false,
  	CustomElementRegistry: false,
  	customElements: false,
  	CustomEvent: false,
  	DataTransfer: false,
  	DataTransferItem: false,
  	DataTransferItemList: false,
  	defaultstatus: false,
  	defaultStatus: false,
  	DelayNode: false,
  	DeviceMotionEvent: false,
  	DeviceOrientationEvent: false,
  	devicePixelRatio: false,
  	dispatchEvent: false,
  	document: false,
  	Document: false,
  	DocumentFragment: false,
  	DocumentType: false,
  	DOMError: false,
  	DOMException: false,
  	DOMImplementation: false,
  	DOMMatrix: false,
  	DOMMatrixReadOnly: false,
  	DOMParser: false,
  	DOMPoint: false,
  	DOMPointReadOnly: false,
  	DOMQuad: false,
  	DOMRect: false,
  	DOMRectReadOnly: false,
  	DOMStringList: false,
  	DOMStringMap: false,
  	DOMTokenList: false,
  	DragEvent: false,
  	DynamicsCompressorNode: false,
  	Element: false,
  	ErrorEvent: false,
  	event: false,
  	Event: false,
  	EventSource: false,
  	EventTarget: false,
  	external: false,
  	fetch: false,
  	File: false,
  	FileList: false,
  	FileReader: false,
  	find: false,
  	focus: false,
  	FocusEvent: false,
  	FontFace: false,
  	FontFaceSetLoadEvent: false,
  	FormData: false,
  	frameElement: false,
  	frames: false,
  	GainNode: false,
  	Gamepad: false,
  	GamepadButton: false,
  	GamepadEvent: false,
  	getComputedStyle: false,
  	getSelection: false,
  	HashChangeEvent: false,
  	Headers: false,
  	history: false,
  	History: false,
  	HTMLAllCollection: false,
  	HTMLAnchorElement: false,
  	HTMLAreaElement: false,
  	HTMLAudioElement: false,
  	HTMLBaseElement: false,
  	HTMLBodyElement: false,
  	HTMLBRElement: false,
  	HTMLButtonElement: false,
  	HTMLCanvasElement: false,
  	HTMLCollection: false,
  	HTMLContentElement: false,
  	HTMLDataElement: false,
  	HTMLDataListElement: false,
  	HTMLDetailsElement: false,
  	HTMLDialogElement: false,
  	HTMLDirectoryElement: false,
  	HTMLDivElement: false,
  	HTMLDListElement: false,
  	HTMLDocument: false,
  	HTMLElement: false,
  	HTMLEmbedElement: false,
  	HTMLFieldSetElement: false,
  	HTMLFontElement: false,
  	HTMLFormControlsCollection: false,
  	HTMLFormElement: false,
  	HTMLFrameElement: false,
  	HTMLFrameSetElement: false,
  	HTMLHeadElement: false,
  	HTMLHeadingElement: false,
  	HTMLHRElement: false,
  	HTMLHtmlElement: false,
  	HTMLIFrameElement: false,
  	HTMLImageElement: false,
  	HTMLInputElement: false,
  	HTMLLabelElement: false,
  	HTMLLegendElement: false,
  	HTMLLIElement: false,
  	HTMLLinkElement: false,
  	HTMLMapElement: false,
  	HTMLMarqueeElement: false,
  	HTMLMediaElement: false,
  	HTMLMenuElement: false,
  	HTMLMetaElement: false,
  	HTMLMeterElement: false,
  	HTMLModElement: false,
  	HTMLObjectElement: false,
  	HTMLOListElement: false,
  	HTMLOptGroupElement: false,
  	HTMLOptionElement: false,
  	HTMLOptionsCollection: false,
  	HTMLOutputElement: false,
  	HTMLParagraphElement: false,
  	HTMLParamElement: false,
  	HTMLPictureElement: false,
  	HTMLPreElement: false,
  	HTMLProgressElement: false,
  	HTMLQuoteElement: false,
  	HTMLScriptElement: false,
  	HTMLSelectElement: false,
  	HTMLShadowElement: false,
  	HTMLSlotElement: false,
  	HTMLSourceElement: false,
  	HTMLSpanElement: false,
  	HTMLStyleElement: false,
  	HTMLTableCaptionElement: false,
  	HTMLTableCellElement: false,
  	HTMLTableColElement: false,
  	HTMLTableElement: false,
  	HTMLTableRowElement: false,
  	HTMLTableSectionElement: false,
  	HTMLTemplateElement: false,
  	HTMLTextAreaElement: false,
  	HTMLTimeElement: false,
  	HTMLTitleElement: false,
  	HTMLTrackElement: false,
  	HTMLUListElement: false,
  	HTMLUnknownElement: false,
  	HTMLVideoElement: false,
  	IDBCursor: false,
  	IDBCursorWithValue: false,
  	IDBDatabase: false,
  	IDBFactory: false,
  	IDBIndex: false,
  	IDBKeyRange: false,
  	IDBObjectStore: false,
  	IDBOpenDBRequest: false,
  	IDBRequest: false,
  	IDBTransaction: false,
  	IDBVersionChangeEvent: false,
  	IdleDeadline: false,
  	IIRFilterNode: false,
  	Image: false,
  	ImageBitmap: false,
  	ImageBitmapRenderingContext: false,
  	ImageCapture: false,
  	ImageData: false,
  	indexedDB: false,
  	innerHeight: false,
  	innerWidth: false,
  	InputEvent: false,
  	IntersectionObserver: false,
  	IntersectionObserverEntry: false,
  	"Intl": false,
  	isSecureContext: false,
  	KeyboardEvent: false,
  	KeyframeEffect: false,
  	KeyframeEffectReadOnly: false,
  	length: false,
  	localStorage: false,
  	location: false,
  	Location: false,
  	locationbar: false,
  	matchMedia: false,
  	MediaDeviceInfo: false,
  	MediaDevices: false,
  	MediaElementAudioSourceNode: false,
  	MediaEncryptedEvent: false,
  	MediaError: false,
  	MediaKeyMessageEvent: false,
  	MediaKeySession: false,
  	MediaKeyStatusMap: false,
  	MediaKeySystemAccess: false,
  	MediaList: false,
  	MediaQueryList: false,
  	MediaQueryListEvent: false,
  	MediaRecorder: false,
  	MediaSettingsRange: false,
  	MediaSource: false,
  	MediaStream: false,
  	MediaStreamAudioDestinationNode: false,
  	MediaStreamAudioSourceNode: false,
  	MediaStreamEvent: false,
  	MediaStreamTrack: false,
  	MediaStreamTrackEvent: false,
  	menubar: false,
  	MessageChannel: false,
  	MessageEvent: false,
  	MessagePort: false,
  	MIDIAccess: false,
  	MIDIConnectionEvent: false,
  	MIDIInput: false,
  	MIDIInputMap: false,
  	MIDIMessageEvent: false,
  	MIDIOutput: false,
  	MIDIOutputMap: false,
  	MIDIPort: false,
  	MimeType: false,
  	MimeTypeArray: false,
  	MouseEvent: false,
  	moveBy: false,
  	moveTo: false,
  	MutationEvent: false,
  	MutationObserver: false,
  	MutationRecord: false,
  	name: false,
  	NamedNodeMap: false,
  	NavigationPreloadManager: false,
  	navigator: false,
  	Navigator: false,
  	NetworkInformation: false,
  	Node: false,
  	NodeFilter: false,
  	NodeIterator: false,
  	NodeList: false,
  	Notification: false,
  	OfflineAudioCompletionEvent: false,
  	OfflineAudioContext: false,
  	offscreenBuffering: false,
  	OffscreenCanvas: true,
  	onabort: true,
  	onafterprint: true,
  	onanimationend: true,
  	onanimationiteration: true,
  	onanimationstart: true,
  	onappinstalled: true,
  	onauxclick: true,
  	onbeforeinstallprompt: true,
  	onbeforeprint: true,
  	onbeforeunload: true,
  	onblur: true,
  	oncancel: true,
  	oncanplay: true,
  	oncanplaythrough: true,
  	onchange: true,
  	onclick: true,
  	onclose: true,
  	oncontextmenu: true,
  	oncuechange: true,
  	ondblclick: true,
  	ondevicemotion: true,
  	ondeviceorientation: true,
  	ondeviceorientationabsolute: true,
  	ondrag: true,
  	ondragend: true,
  	ondragenter: true,
  	ondragleave: true,
  	ondragover: true,
  	ondragstart: true,
  	ondrop: true,
  	ondurationchange: true,
  	onemptied: true,
  	onended: true,
  	onerror: true,
  	onfocus: true,
  	ongotpointercapture: true,
  	onhashchange: true,
  	oninput: true,
  	oninvalid: true,
  	onkeydown: true,
  	onkeypress: true,
  	onkeyup: true,
  	onlanguagechange: true,
  	onload: true,
  	onloadeddata: true,
  	onloadedmetadata: true,
  	onloadstart: true,
  	onlostpointercapture: true,
  	onmessage: true,
  	onmessageerror: true,
  	onmousedown: true,
  	onmouseenter: true,
  	onmouseleave: true,
  	onmousemove: true,
  	onmouseout: true,
  	onmouseover: true,
  	onmouseup: true,
  	onmousewheel: true,
  	onoffline: true,
  	ononline: true,
  	onpagehide: true,
  	onpageshow: true,
  	onpause: true,
  	onplay: true,
  	onplaying: true,
  	onpointercancel: true,
  	onpointerdown: true,
  	onpointerenter: true,
  	onpointerleave: true,
  	onpointermove: true,
  	onpointerout: true,
  	onpointerover: true,
  	onpointerup: true,
  	onpopstate: true,
  	onprogress: true,
  	onratechange: true,
  	onrejectionhandled: true,
  	onreset: true,
  	onresize: true,
  	onscroll: true,
  	onsearch: true,
  	onseeked: true,
  	onseeking: true,
  	onselect: true,
  	onstalled: true,
  	onstorage: true,
  	onsubmit: true,
  	onsuspend: true,
  	ontimeupdate: true,
  	ontoggle: true,
  	ontransitionend: true,
  	onunhandledrejection: true,
  	onunload: true,
  	onvolumechange: true,
  	onwaiting: true,
  	onwheel: true,
  	open: false,
  	openDatabase: false,
  	opener: false,
  	Option: false,
  	origin: false,
  	OscillatorNode: false,
  	outerHeight: false,
  	outerWidth: false,
  	PageTransitionEvent: false,
  	pageXOffset: false,
  	pageYOffset: false,
  	PannerNode: false,
  	parent: false,
  	Path2D: false,
  	PaymentAddress: false,
  	PaymentRequest: false,
  	PaymentRequestUpdateEvent: false,
  	PaymentResponse: false,
  	performance: false,
  	Performance: false,
  	PerformanceEntry: false,
  	PerformanceLongTaskTiming: false,
  	PerformanceMark: false,
  	PerformanceMeasure: false,
  	PerformanceNavigation: false,
  	PerformanceNavigationTiming: false,
  	PerformanceObserver: false,
  	PerformanceObserverEntryList: false,
  	PerformancePaintTiming: false,
  	PerformanceResourceTiming: false,
  	PerformanceTiming: false,
  	PeriodicWave: false,
  	Permissions: false,
  	PermissionStatus: false,
  	personalbar: false,
  	PhotoCapabilities: false,
  	Plugin: false,
  	PluginArray: false,
  	PointerEvent: false,
  	PopStateEvent: false,
  	postMessage: false,
  	Presentation: false,
  	PresentationAvailability: false,
  	PresentationConnection: false,
  	PresentationConnectionAvailableEvent: false,
  	PresentationConnectionCloseEvent: false,
  	PresentationConnectionList: false,
  	PresentationReceiver: false,
  	PresentationRequest: false,
  	print: false,
  	ProcessingInstruction: false,
  	ProgressEvent: false,
  	PromiseRejectionEvent: false,
  	prompt: false,
  	PushManager: false,
  	PushSubscription: false,
  	PushSubscriptionOptions: false,
  	queueMicrotask: false,
  	RadioNodeList: false,
  	Range: false,
  	ReadableStream: false,
  	registerProcessor: false,
  	RemotePlayback: false,
  	removeEventListener: false,
  	Request: false,
  	requestAnimationFrame: false,
  	requestIdleCallback: false,
  	resizeBy: false,
  	ResizeObserver: false,
  	ResizeObserverEntry: false,
  	resizeTo: false,
  	Response: false,
  	RTCCertificate: false,
  	RTCDataChannel: false,
  	RTCDataChannelEvent: false,
  	RTCDtlsTransport: false,
  	RTCIceCandidate: false,
  	RTCIceGatherer: false,
  	RTCIceTransport: false,
  	RTCPeerConnection: false,
  	RTCPeerConnectionIceEvent: false,
  	RTCRtpContributingSource: false,
  	RTCRtpReceiver: false,
  	RTCRtpSender: false,
  	RTCSctpTransport: false,
  	RTCSessionDescription: false,
  	RTCStatsReport: false,
  	RTCTrackEvent: false,
  	screen: false,
  	Screen: false,
  	screenLeft: false,
  	ScreenOrientation: false,
  	screenTop: false,
  	screenX: false,
  	screenY: false,
  	ScriptProcessorNode: false,
  	scroll: false,
  	scrollbars: false,
  	scrollBy: false,
  	scrollTo: false,
  	scrollX: false,
  	scrollY: false,
  	SecurityPolicyViolationEvent: false,
  	Selection: false,
  	self: false,
  	ServiceWorker: false,
  	ServiceWorkerContainer: false,
  	ServiceWorkerRegistration: false,
  	sessionStorage: false,
  	setInterval: false,
  	setTimeout: false,
  	ShadowRoot: false,
  	SharedWorker: false,
  	SourceBuffer: false,
  	SourceBufferList: false,
  	speechSynthesis: false,
  	SpeechSynthesisEvent: false,
  	SpeechSynthesisUtterance: false,
  	StaticRange: false,
  	status: false,
  	statusbar: false,
  	StereoPannerNode: false,
  	stop: false,
  	Storage: false,
  	StorageEvent: false,
  	StorageManager: false,
  	styleMedia: false,
  	StyleSheet: false,
  	StyleSheetList: false,
  	SubtleCrypto: false,
  	SVGAElement: false,
  	SVGAngle: false,
  	SVGAnimatedAngle: false,
  	SVGAnimatedBoolean: false,
  	SVGAnimatedEnumeration: false,
  	SVGAnimatedInteger: false,
  	SVGAnimatedLength: false,
  	SVGAnimatedLengthList: false,
  	SVGAnimatedNumber: false,
  	SVGAnimatedNumberList: false,
  	SVGAnimatedPreserveAspectRatio: false,
  	SVGAnimatedRect: false,
  	SVGAnimatedString: false,
  	SVGAnimatedTransformList: false,
  	SVGAnimateElement: false,
  	SVGAnimateMotionElement: false,
  	SVGAnimateTransformElement: false,
  	SVGAnimationElement: false,
  	SVGCircleElement: false,
  	SVGClipPathElement: false,
  	SVGComponentTransferFunctionElement: false,
  	SVGDefsElement: false,
  	SVGDescElement: false,
  	SVGDiscardElement: false,
  	SVGElement: false,
  	SVGEllipseElement: false,
  	SVGFEBlendElement: false,
  	SVGFEColorMatrixElement: false,
  	SVGFEComponentTransferElement: false,
  	SVGFECompositeElement: false,
  	SVGFEConvolveMatrixElement: false,
  	SVGFEDiffuseLightingElement: false,
  	SVGFEDisplacementMapElement: false,
  	SVGFEDistantLightElement: false,
  	SVGFEDropShadowElement: false,
  	SVGFEFloodElement: false,
  	SVGFEFuncAElement: false,
  	SVGFEFuncBElement: false,
  	SVGFEFuncGElement: false,
  	SVGFEFuncRElement: false,
  	SVGFEGaussianBlurElement: false,
  	SVGFEImageElement: false,
  	SVGFEMergeElement: false,
  	SVGFEMergeNodeElement: false,
  	SVGFEMorphologyElement: false,
  	SVGFEOffsetElement: false,
  	SVGFEPointLightElement: false,
  	SVGFESpecularLightingElement: false,
  	SVGFESpotLightElement: false,
  	SVGFETileElement: false,
  	SVGFETurbulenceElement: false,
  	SVGFilterElement: false,
  	SVGForeignObjectElement: false,
  	SVGGElement: false,
  	SVGGeometryElement: false,
  	SVGGradientElement: false,
  	SVGGraphicsElement: false,
  	SVGImageElement: false,
  	SVGLength: false,
  	SVGLengthList: false,
  	SVGLinearGradientElement: false,
  	SVGLineElement: false,
  	SVGMarkerElement: false,
  	SVGMaskElement: false,
  	SVGMatrix: false,
  	SVGMetadataElement: false,
  	SVGMPathElement: false,
  	SVGNumber: false,
  	SVGNumberList: false,
  	SVGPathElement: false,
  	SVGPatternElement: false,
  	SVGPoint: false,
  	SVGPointList: false,
  	SVGPolygonElement: false,
  	SVGPolylineElement: false,
  	SVGPreserveAspectRatio: false,
  	SVGRadialGradientElement: false,
  	SVGRect: false,
  	SVGRectElement: false,
  	SVGScriptElement: false,
  	SVGSetElement: false,
  	SVGStopElement: false,
  	SVGStringList: false,
  	SVGStyleElement: false,
  	SVGSVGElement: false,
  	SVGSwitchElement: false,
  	SVGSymbolElement: false,
  	SVGTextContentElement: false,
  	SVGTextElement: false,
  	SVGTextPathElement: false,
  	SVGTextPositioningElement: false,
  	SVGTitleElement: false,
  	SVGTransform: false,
  	SVGTransformList: false,
  	SVGTSpanElement: false,
  	SVGUnitTypes: false,
  	SVGUseElement: false,
  	SVGViewElement: false,
  	TaskAttributionTiming: false,
  	Text: false,
  	TextDecoder: false,
  	TextEncoder: false,
  	TextEvent: false,
  	TextMetrics: false,
  	TextTrack: false,
  	TextTrackCue: false,
  	TextTrackCueList: false,
  	TextTrackList: false,
  	TimeRanges: false,
  	toolbar: false,
  	top: false,
  	Touch: false,
  	TouchEvent: false,
  	TouchList: false,
  	TrackEvent: false,
  	TransitionEvent: false,
  	TreeWalker: false,
  	UIEvent: false,
  	URL: false,
  	URLSearchParams: false,
  	ValidityState: false,
  	visualViewport: false,
  	VisualViewport: false,
  	VTTCue: false,
  	WaveShaperNode: false,
  	WebAssembly: false,
  	WebGL2RenderingContext: false,
  	WebGLActiveInfo: false,
  	WebGLBuffer: false,
  	WebGLContextEvent: false,
  	WebGLFramebuffer: false,
  	WebGLProgram: false,
  	WebGLQuery: false,
  	WebGLRenderbuffer: false,
  	WebGLRenderingContext: false,
  	WebGLSampler: false,
  	WebGLShader: false,
  	WebGLShaderPrecisionFormat: false,
  	WebGLSync: false,
  	WebGLTexture: false,
  	WebGLTransformFeedback: false,
  	WebGLUniformLocation: false,
  	WebGLVertexArrayObject: false,
  	WebSocket: false,
  	WheelEvent: false,
  	window: false,
  	Window: false,
  	Worker: false,
  	WritableStream: false,
  	XMLDocument: false,
  	XMLHttpRequest: false,
  	XMLHttpRequestEventTarget: false,
  	XMLHttpRequestUpload: false,
  	XMLSerializer: false,
  	XPathEvaluator: false,
  	XPathExpression: false,
  	XPathResult: false,
  	XSLTProcessor: false
  };
  var worker = {
  	applicationCache: false,
  	atob: false,
  	Blob: false,
  	BroadcastChannel: false,
  	btoa: false,
  	Cache: false,
  	caches: false,
  	clearInterval: false,
  	clearTimeout: false,
  	close: true,
  	console: false,
  	fetch: false,
  	FileReaderSync: false,
  	FormData: false,
  	Headers: false,
  	IDBCursor: false,
  	IDBCursorWithValue: false,
  	IDBDatabase: false,
  	IDBFactory: false,
  	IDBIndex: false,
  	IDBKeyRange: false,
  	IDBObjectStore: false,
  	IDBOpenDBRequest: false,
  	IDBRequest: false,
  	IDBTransaction: false,
  	IDBVersionChangeEvent: false,
  	ImageData: false,
  	importScripts: true,
  	indexedDB: false,
  	location: false,
  	MessageChannel: false,
  	MessagePort: false,
  	name: false,
  	navigator: false,
  	Notification: false,
  	onclose: true,
  	onconnect: true,
  	onerror: true,
  	onlanguagechange: true,
  	onmessage: true,
  	onoffline: true,
  	ononline: true,
  	onrejectionhandled: true,
  	onunhandledrejection: true,
  	performance: false,
  	Performance: false,
  	PerformanceEntry: false,
  	PerformanceMark: false,
  	PerformanceMeasure: false,
  	PerformanceNavigation: false,
  	PerformanceResourceTiming: false,
  	PerformanceTiming: false,
  	postMessage: true,
  	"Promise": false,
  	queueMicrotask: false,
  	Request: false,
  	Response: false,
  	self: true,
  	ServiceWorkerRegistration: false,
  	setInterval: false,
  	setTimeout: false,
  	TextDecoder: false,
  	TextEncoder: false,
  	URL: false,
  	URLSearchParams: false,
  	WebSocket: false,
  	Worker: false,
  	WorkerGlobalScope: false,
  	XMLHttpRequest: false
  };
  var node = {
  	__dirname: false,
  	__filename: false,
  	Buffer: false,
  	clearImmediate: false,
  	clearInterval: false,
  	clearTimeout: false,
  	console: false,
  	exports: true,
  	global: false,
  	"Intl": false,
  	module: false,
  	process: false,
  	queueMicrotask: false,
  	require: false,
  	setImmediate: false,
  	setInterval: false,
  	setTimeout: false,
  	TextDecoder: false,
  	TextEncoder: false,
  	URL: false,
  	URLSearchParams: false
  };
  var commonjs = {
  	exports: true,
  	global: false,
  	module: false,
  	require: false
  };
  var amd = {
  	define: false,
  	require: false
  };
  var mocha = {
  	after: false,
  	afterEach: false,
  	before: false,
  	beforeEach: false,
  	context: false,
  	describe: false,
  	it: false,
  	mocha: false,
  	run: false,
  	setup: false,
  	specify: false,
  	suite: false,
  	suiteSetup: false,
  	suiteTeardown: false,
  	teardown: false,
  	test: false,
  	xcontext: false,
  	xdescribe: false,
  	xit: false,
  	xspecify: false
  };
  var jasmine = {
  	afterAll: false,
  	afterEach: false,
  	beforeAll: false,
  	beforeEach: false,
  	describe: false,
  	expect: false,
  	fail: false,
  	fdescribe: false,
  	fit: false,
  	it: false,
  	jasmine: false,
  	pending: false,
  	runs: false,
  	spyOn: false,
  	spyOnProperty: false,
  	waits: false,
  	waitsFor: false,
  	xdescribe: false,
  	xit: false
  };
  var jest = {
  	afterAll: false,
  	afterEach: false,
  	beforeAll: false,
  	beforeEach: false,
  	describe: false,
  	expect: false,
  	fdescribe: false,
  	fit: false,
  	it: false,
  	jest: false,
  	pit: false,
  	require: false,
  	test: false,
  	xdescribe: false,
  	xit: false,
  	xtest: false
  };
  var qunit = {
  	asyncTest: false,
  	deepEqual: false,
  	equal: false,
  	expect: false,
  	module: false,
  	notDeepEqual: false,
  	notEqual: false,
  	notOk: false,
  	notPropEqual: false,
  	notStrictEqual: false,
  	ok: false,
  	propEqual: false,
  	QUnit: false,
  	raises: false,
  	start: false,
  	stop: false,
  	strictEqual: false,
  	test: false,
  	throws: false
  };
  var phantomjs = {
  	console: true,
  	exports: true,
  	phantom: true,
  	require: true,
  	WebPage: true
  };
  var couch = {
  	emit: false,
  	exports: false,
  	getRow: false,
  	log: false,
  	module: false,
  	provides: false,
  	require: false,
  	respond: false,
  	send: false,
  	start: false,
  	sum: false
  };
  var rhino = {
  	defineClass: false,
  	deserialize: false,
  	gc: false,
  	help: false,
  	importClass: false,
  	importPackage: false,
  	java: false,
  	load: false,
  	loadClass: false,
  	Packages: false,
  	print: false,
  	quit: false,
  	readFile: false,
  	readUrl: false,
  	runCommand: false,
  	seal: false,
  	serialize: false,
  	spawn: false,
  	sync: false,
  	toint32: false,
  	version: false
  };
  var nashorn = {
  	__DIR__: false,
  	__FILE__: false,
  	__LINE__: false,
  	com: false,
  	edu: false,
  	exit: false,
  	java: false,
  	Java: false,
  	javafx: false,
  	JavaImporter: false,
  	javax: false,
  	JSAdapter: false,
  	load: false,
  	loadWithNewGlobal: false,
  	org: false,
  	Packages: false,
  	print: false,
  	quit: false
  };
  var wsh = {
  	ActiveXObject: true,
  	Enumerator: true,
  	GetObject: true,
  	ScriptEngine: true,
  	ScriptEngineBuildVersion: true,
  	ScriptEngineMajorVersion: true,
  	ScriptEngineMinorVersion: true,
  	VBArray: true,
  	WScript: true,
  	WSH: true,
  	XDomainRequest: true
  };
  var jquery = {
  	$: false,
  	jQuery: false
  };
  var yui = {
  	YAHOO: false,
  	YAHOO_config: false,
  	YUI: false,
  	YUI_config: false
  };
  var shelljs = {
  	cat: false,
  	cd: false,
  	chmod: false,
  	config: false,
  	cp: false,
  	dirs: false,
  	echo: false,
  	env: false,
  	error: false,
  	exec: false,
  	exit: false,
  	find: false,
  	grep: false,
  	ln: false,
  	ls: false,
  	mkdir: false,
  	mv: false,
  	popd: false,
  	pushd: false,
  	pwd: false,
  	rm: false,
  	sed: false,
  	set: false,
  	target: false,
  	tempdir: false,
  	test: false,
  	touch: false,
  	which: false
  };
  var prototypejs = {
  	$: false,
  	$$: false,
  	$A: false,
  	$break: false,
  	$continue: false,
  	$F: false,
  	$H: false,
  	$R: false,
  	$w: false,
  	Abstract: false,
  	Ajax: false,
  	Autocompleter: false,
  	Builder: false,
  	Class: false,
  	Control: false,
  	Draggable: false,
  	Draggables: false,
  	Droppables: false,
  	Effect: false,
  	Element: false,
  	Enumerable: false,
  	Event: false,
  	Field: false,
  	Form: false,
  	Hash: false,
  	Insertion: false,
  	ObjectRange: false,
  	PeriodicalExecuter: false,
  	Position: false,
  	Prototype: false,
  	Scriptaculous: false,
  	Selector: false,
  	Sortable: false,
  	SortableObserver: false,
  	Sound: false,
  	Template: false,
  	Toggle: false,
  	Try: false
  };
  var meteor = {
  	_: false,
  	$: false,
  	Accounts: false,
  	AccountsClient: false,
  	AccountsCommon: false,
  	AccountsServer: false,
  	App: false,
  	Assets: false,
  	Blaze: false,
  	check: false,
  	Cordova: false,
  	DDP: false,
  	DDPRateLimiter: false,
  	DDPServer: false,
  	Deps: false,
  	EJSON: false,
  	Email: false,
  	HTTP: false,
  	Log: false,
  	Match: false,
  	Meteor: false,
  	Mongo: false,
  	MongoInternals: false,
  	Npm: false,
  	Package: false,
  	Plugin: false,
  	process: false,
  	Random: false,
  	ReactiveDict: false,
  	ReactiveVar: false,
  	Router: false,
  	ServiceConfiguration: false,
  	Session: false,
  	share: false,
  	Spacebars: false,
  	Template: false,
  	Tinytest: false,
  	Tracker: false,
  	UI: false,
  	Utils: false,
  	WebApp: false,
  	WebAppInternals: false
  };
  var mongo = {
  	_isWindows: false,
  	_rand: false,
  	BulkWriteResult: false,
  	cat: false,
  	cd: false,
  	connect: false,
  	db: false,
  	getHostName: false,
  	getMemInfo: false,
  	hostname: false,
  	ISODate: false,
  	listFiles: false,
  	load: false,
  	ls: false,
  	md5sumFile: false,
  	mkdir: false,
  	Mongo: false,
  	NumberInt: false,
  	NumberLong: false,
  	ObjectId: false,
  	PlanCache: false,
  	print: false,
  	printjson: false,
  	pwd: false,
  	quit: false,
  	removeFile: false,
  	rs: false,
  	sh: false,
  	UUID: false,
  	version: false,
  	WriteResult: false
  };
  var applescript = {
  	$: false,
  	Application: false,
  	Automation: false,
  	console: false,
  	delay: false,
  	Library: false,
  	ObjC: false,
  	ObjectSpecifier: false,
  	Path: false,
  	Progress: false,
  	Ref: false
  };
  var serviceworker = {
  	Cache: false,
  	caches: false,
  	CacheStorage: false,
  	Client: false,
  	clients: false,
  	Clients: false,
  	ExtendableEvent: false,
  	ExtendableMessageEvent: false,
  	FetchEvent: false,
  	importScripts: false,
  	registration: false,
  	self: false,
  	ServiceWorker: false,
  	ServiceWorkerContainer: false,
  	ServiceWorkerGlobalScope: false,
  	ServiceWorkerMessageEvent: false,
  	ServiceWorkerRegistration: false,
  	skipWaiting: false,
  	WindowClient: false
  };
  var atomtest = {
  	advanceClock: false,
  	fakeClearInterval: false,
  	fakeClearTimeout: false,
  	fakeSetInterval: false,
  	fakeSetTimeout: false,
  	resetTimeouts: false,
  	waitsForPromise: false
  };
  var embertest = {
  	andThen: false,
  	click: false,
  	currentPath: false,
  	currentRouteName: false,
  	currentURL: false,
  	fillIn: false,
  	find: false,
  	findAll: false,
  	findWithAssert: false,
  	keyEvent: false,
  	pauseTest: false,
  	resumeTest: false,
  	triggerEvent: false,
  	visit: false,
  	wait: false
  };
  var protractor = {
  	$: false,
  	$$: false,
  	browser: false,
  	by: false,
  	By: false,
  	DartObject: false,
  	element: false,
  	protractor: false
  };
  var webextensions = {
  	browser: false,
  	chrome: false,
  	opr: false
  };
  var greasemonkey = {
  	GM: false,
  	GM_addStyle: false,
  	GM_deleteValue: false,
  	GM_getResourceText: false,
  	GM_getResourceURL: false,
  	GM_getValue: false,
  	GM_info: false,
  	GM_listValues: false,
  	GM_log: false,
  	GM_openInTab: false,
  	GM_registerMenuCommand: false,
  	GM_setClipboard: false,
  	GM_setValue: false,
  	GM_xmlhttpRequest: false,
  	unsafeWindow: false
  };
  var devtools = {
  	$: false,
  	$_: false,
  	$$: false,
  	$0: false,
  	$1: false,
  	$2: false,
  	$3: false,
  	$4: false,
  	$x: false,
  	chrome: false,
  	clear: false,
  	copy: false,
  	debug: false,
  	dir: false,
  	dirxml: false,
  	getEventListeners: false,
  	inspect: false,
  	keys: false,
  	monitor: false,
  	monitorEvents: false,
  	profile: false,
  	profileEnd: false,
  	queryObjects: false,
  	table: false,
  	undebug: false,
  	unmonitor: false,
  	unmonitorEvents: false,
  	values: false
  };
  var globals$1 = {
  	builtin: builtin,
  	es5: es5,
  	es2015: es2015,
  	es2017: es2017,
  	browser: browser,
  	worker: worker,
  	node: node,
  	commonjs: commonjs,
  	amd: amd,
  	mocha: mocha,
  	jasmine: jasmine,
  	jest: jest,
  	qunit: qunit,
  	phantomjs: phantomjs,
  	couch: couch,
  	rhino: rhino,
  	nashorn: nashorn,
  	wsh: wsh,
  	jquery: jquery,
  	yui: yui,
  	shelljs: shelljs,
  	prototypejs: prototypejs,
  	meteor: meteor,
  	mongo: mongo,
  	applescript: applescript,
  	serviceworker: serviceworker,
  	atomtest: atomtest,
  	embertest: embertest,
  	protractor: protractor,
  	"shared-node-browser": {
  	clearInterval: false,
  	clearTimeout: false,
  	console: false,
  	setInterval: false,
  	setTimeout: false,
  	URL: false,
  	URLSearchParams: false
  },
  	webextensions: webextensions,
  	greasemonkey: greasemonkey,
  	devtools: devtools
  };

  var globals$2 = /*#__PURE__*/Object.freeze({
    builtin: builtin,
    es5: es5,
    es2015: es2015,
    es2017: es2017,
    browser: browser,
    worker: worker,
    node: node,
    commonjs: commonjs,
    amd: amd,
    mocha: mocha,
    jasmine: jasmine,
    jest: jest,
    qunit: qunit,
    phantomjs: phantomjs,
    couch: couch,
    rhino: rhino,
    nashorn: nashorn,
    wsh: wsh,
    jquery: jquery,
    yui: yui,
    shelljs: shelljs,
    prototypejs: prototypejs,
    meteor: meteor,
    mongo: mongo,
    applescript: applescript,
    serviceworker: serviceworker,
    atomtest: atomtest,
    embertest: embertest,
    protractor: protractor,
    webextensions: webextensions,
    greasemonkey: greasemonkey,
    devtools: devtools,
    default: globals$1
  });

  var require$$0 = getCjsExportFromNamespace(globals$2);

  var globals$3 = require$$0;

  /* Riot Compiler v4.0.0-alpha.0, @license MIT */

  const TAG_LOGIC_PROPERTY = 'tag';
  const TAG_CSS_PROPERTY = 'css';
  const TAG_TEMPLATE_PROPERTY = 'template';

  /**
   * Detect node js environements
   * @returns { boolean } true if the runtime is node
   */
  function isNode() {
    return typeof process !== 'undefined'
  }

  /**
   * Return a source map as JSON, it it has not the toJSON method it means it can
   * be used right the way
   * @param   { SourceMapGenerator|Object } map - a sourcemap generator or simply an json object
   * @returns { Object } the source map as JSON
   */
  function asJSON(map) {
    if (map.toJSON) return map.toJSON()
    return map
  }
  /**
   * Compose two sourcemaps
   * @param   { SourceMapGenerator } formerMap - original sourcemap
   * @param   { SourceMapGenerator } latterMap - target sourcemap
   * @returns { Object } sourcemap json
   */
  function composeSourcemaps(formerMap, latterMap) {
    if (
      isNode() &&
      formerMap && latterMap && latterMap.mappings
    ) {
      return util_1$1.composeSourceMaps(asJSON(formerMap), asJSON(latterMap))
    } else if (isNode() && formerMap) {
      return asJSON(formerMap)
    }

    return {}
  }

  /**
   * Create a new sourcemap generator
   * @param   { Object } options - sourcemap options
   * @returns { SourceMapGenerator } SourceMapGenerator instance
   */
  function createSourcemap(options) {
    return new SourceMapGenerator$5(options)
  }

  const Output = Object.freeze({
    code: '',
    ast: [],
    map: null
  });

  /**
   * Create the right output data result of a parsing
   * @param   { Object } data - output data
   * @param   { string } data.code - code generated
   * @param   { AST } data.ast - ast representing the code
   * @param   { SourceMapGenerator } data.map - source map generated along with the code
   * @param   { Object } options - user options, probably containing the path to the source file
   * @returns { Output } output container object
   */
  function createOutput(data, options) {
    const output = Object.seal({
      ...Output,
      ...data
    });

    if (!output.map && options && options.file) Object.assign(output, {
      map: createSourcemap({ file: options.file })
    });

    return output
  }

  /**
   * Transform the source code received via a compiler function
   * @param   { Function } compiler - function needed to generate the output code
   * @param   { Object } options - options to pass to the compilert
   * @param   { string } source - source code
   * @returns { Promise<Output> } output - the result of the compiler
   */
  async function transform(compiler, options, source) {
    const result = await (compiler ? compiler(source, options) : { code: source });
    return createOutput(result, options)
  }

  /**
   * Throw an error with a descriptive message
   * @param   { string } message - error message
   * @returns { undefined } hoppla.. at this point the program should stop working
   */
  function panic$2(message) {
    throw new Error(message)
  }

  const postprocessors = new Set();

  /**
   * Register a postprocessor that will be used after the parsing and compilation of the riot tags
   * @param { Function } postprocessor - transformer that will receive the output code ans sourcemap
   * @returns { Set } the postprocessors collection
   */
  function register$1(postprocessor) {
    if (postprocessors.has(postprocessor)) {
      panic$2(`This postprocessor "${postprocessor.name || postprocessor.toString()}" was already registered`);
    }

    postprocessors.add(postprocessor);

    return postprocessors
  }

  /**
   * Exec all the postprocessors in sequence combining the sourcemaps generated
   * @param   { Output } compilerOutput - output generated by the compiler
   * @param   { Object } options - user options received by the compiler
   * @returns { Promise<Output> } object containing output code and source map
   */
  async function execute(compilerOutput, options) {
    return Array.from(postprocessors).reduce(async function(acc, postprocessor) {
      const { code, map } = await acc;
      const output = await postprocessor(code, options);

      return {
        code: output.code,
        map: composeSourcemaps(output.map, map)
      }
    }, Promise.resolve(createOutput(compilerOutput, options)))
  }

  /**
   * Parsers that can be registered by users to preparse components fragments
   * @type { Object }
   */
  const preprocessors = Object.freeze({
    javascript: new Map(),
    css: new Map(),
    template: new Map()
      .set('default', function(code) { return { code } })
  });

  // throw a processor type error
  function preprocessorTypeError(type) {
    panic$2(`No preprocessor of type "${type}" was found, please make sure to use one of these: 'javascript', 'css' or 'template'`);
  }

  // throw an error if the preprocessor was not registered
  function preprocessorNameNotFoundError(name) {
    panic$2(`No preprocessor named "${name}" was found, are you sure you have registered it?'`);
  }

  /**
   * Exec the compilation of a preprocessor
   * @param   { string } type - preprocessor type either 'js', 'css' or 'template'
   * @param   { string } name - unique preprocessor id
   * @param   { Object } options - preprocessor options
   * @param   { string } source - source code
   * @returns { Promise<Output> } object containing a sourcemap and a code string
   */
  async function execute$1(type, name, options, source) {
    if (!preprocessors[type]) preprocessorTypeError(type);
    if (!preprocessors[type].has(name)) preprocessorNameNotFoundError(name);

    return await transform(preprocessors[type].get(name), options, source)
  }

  const ATTRIBUTE_TYPE_NAME = 'type';

  /**
   * Get the type attribute from a node generated by the riot parser
   * @param   { Object} sourceNode - riot parser node
   * @returns { string|null } a valid type to identify the preprocessor to use or nothing
   */
  function getPreprocessorTypeByAttribute(sourceNode) {
    const typeAttribute = sourceNode.attributes ?
      sourceNode.attributes.find(attribute => attribute.name === ATTRIBUTE_TYPE_NAME) :
      null;

    return typeAttribute ? normalize$1(typeAttribute.value) : null
  }


  /**
   * Remove the noise in case a user has defined the preprocessor type='text/scss'
   * @param   { string } value - input string
   * @returns { string } normalized string
   */
  function normalize$1(value) {
    return value.replace('text/', '')
  }

  /**
   * Count spaces before first character of a string
   * @param   { string } string - target string
   * @returns { number } amount of spaces before the first char
   */
  function getColum(string) {
    const spacesAmount = string.search(/\S/);
    return spacesAmount > 0 ? spacesAmount : 0
  }

  const LINES_RE = /\r\n?|\n/g;

  /**
   * Split a string into a rows array generated from its EOL matches
   * @param   { string } string [description]
   * @returns { Array } array containing all the string rows
   */
  function splitStringByEOL(string) {
    return string.split(LINES_RE)
  }

  /**
   * Get the line and the column of a source text based on its position in the string
   * @param   { string } string - target string
   * @param   { number } position - target position
   * @returns { Object } object containing the source text line and column
   */
  function getLineAndColumnByPosition(string, position) {
    const lines = splitStringByEOL(string.slice(0, position));

    return {
      line: lines.length,
      column: getColum(lines[lines.length - 1])
    }
  }

  /**
   * Preprocess a riot parser node
   * @param   { string } preprocessorType - either css, js
   * @param   { string } preprocessorName - preprocessor id
   * @param   { Object } options - options that will be passed to the compiler
   * @param   { string } source - tag source code
   * @param   { RiotParser.nodeTypes } node - css node detected by the parser
   * @returns { Output } code and sourcemap generated by the preprocessor
   */
  async function preprocess(preprocessorType, preprocessorName, options, source, node) {
    const { column } = getLineAndColumnByPosition(source, node.start);
    const offsetTop = '\n'.repeat(column);
    const code = `${offsetTop}\n${node.text}`;

    return await (preprocessorName ?
      execute$1(preprocessorType, preprocessorName, options, code) :
      { code }
    )
  }

  const types$3 = main$1.types;
  const builders = types$3.builders;
  const namedTypes$1 = types$3.namedTypes;

  /**
   * Source for creating regexes matching valid quoted, single-line JavaScript strings.
   * It recognizes escape characters, including nested quotes and line continuation.
   * @const {string}
   */
  const S_LINESTR = /"[^"\n\\]*(?:\\[\S\s][^"\n\\]*)*"|'[^'\n\\]*(?:\\[\S\s][^'\n\\]*)*'/.source;

  /**
   * Matches CSS selectors, excluding those beginning with '@' and quoted strings.
   * @const {RegExp}
   */

  const CSS_SELECTOR = RegExp(`([{}]|^)[; ]*((?:[^@ ;{}][^{}]*)?[^@ ;{}:] ?)(?={)|${S_LINESTR}`, 'g');

  /**
   * Parses styles enclosed in a "scoped" tag
   * The "css" string is received without comments or surrounding spaces.
   *
   * @param   {string} tag - Tag name of the root element
   * @param   {string} css - The CSS code
   * @returns {string} CSS with the styles scoped to the root element
   */
  function scopedCSS(tag, css) {
    const host = ':host';
    const selectorsBlacklist = ['from', 'to'];

    return css.replace(CSS_SELECTOR, function(m, p1, p2) {
      // skip quoted strings
      if (!p2) return m

      // we have a selector list, parse each individually
      p2 = p2.replace(/[^,]+/g, function(sel) {
        const s = sel.trim();

        // skip selectors already using the tag name
        if (s.indexOf(tag) === 0) {
          return sel
        }

        // skips the keywords and percents of css animations
        if (!s || selectorsBlacklist.indexOf(s) > -1 || s.slice(-1) === '%') {
          return sel
        }

        // replace the `:host` pseudo-selector, where it is, with the root tag name;
        // if `:host` was not included, add the tag name as prefix, and mirror all
        // `[data-is]`
        if (s.indexOf(host) < 0) {
          return `${tag} ${s},[is="${tag}"] ${s}`
        } else {
          return `${s.replace(host, tag)},${
          s.replace(host, `[is="${tag}"]`)}`
        }
      });

      // add the danling bracket char and return the processed selector list
      return p1 ? `${p1} ${p2}` : p2
    })
  }

  /**
   * Generate the component css
   * @param   { Object } sourceNode - node generated by the riot compiler
   * @param   { string } source - original component source code
   * @param   { Object } options - user options
   * @param   { Object } output - current compiler output
   * @returns { Promise<Output> } - the current ast program and the original sourcemap
   */
  async function css(sourceNode, source, options, { ast, map }) {
    const preprocessorName = getPreprocessorTypeByAttribute(sourceNode);
    const cssNode = sourceNode.text;
    const preprocessorOutput = await preprocess('css', preprocessorName, options, source, cssNode);
    const cssCode = (options.scopedCss ?
      scopedCSS(options.tagName, preprocessorOutput.code) :
      preprocessorOutput.code
    ).trim();

    const generatedCss = main$1.parse(`\`${cssCode}\``, {
      sourceFileName: options.file,
      inputSourceMap: composeSourcemaps(map, preprocessorOutput.map)
    });

    types$3.visit(ast, {
      visitProperty(path) {
        if (path.value.key.name === TAG_CSS_PROPERTY) {
          path.value.value = generatedCss.program.body[0].expression;
          return false
        }

        this.traverse(path);
      }
    });

    return { ast, map, code: main$1.print(ast).code }
  }

  const isExportDefaultStatement = namedTypes$1.ExportDefaultDeclaration.check;

  /**
   * Find the export default statement
   * @param   { Array } body - tree structure containing the program code
   * @returns { Object } node containing only the code of the export default statement
   */
  function findExportDefaultStatement(body) {
    return body.find(isExportDefaultStatement)
  }

  /**
   * Find all the code in an ast program except for the export default statements
   * @param   { Array } body - tree structure containing the program code
   * @returns { Array } array containing all the program code except the export default expressions
   */
  function filterNonExportDefaultStatements(body) {
    return body.filter(node => !isExportDefaultStatement(node))
  }

  /**
   * Get the body of the AST structure
   * @param   { Object } ast - ast object generated by recast
   * @returns { Array } array containing the program code
   */
  function getProgramBody(ast) {
    return ast.program.body
  }

  /**
   * Extend the AST adding the new tag method containing our tag sourcecode
   * @param   { Object } ast - current output ast
   * @param   { Object } exportDefaultNode - tag export default node
   * @returns { Object } the output ast having the "tag" key extended with the content of the export default
   */
  function extendTagProperty(ast, exportDefaultNode) {
    types$3.visit(ast, {
      visitProperty(path) {
        if (path.value.key.name === TAG_LOGIC_PROPERTY) {
          path.value.value = exportDefaultNode.declaration;
          return false
        }

        this.traverse(path);
      }
    });

    return ast
  }

  /**
   * Generate the component javascript logic
   * @param   { Object } sourceNode - node generated by the riot compiler
   * @param   { string } source - original component source code
   * @param   { Object } options - user options
   * @param   { Output } output - current compiler output
   * @returns { Promise<Output> } - enhanced output with the result of the current generator
   */
  async function javascript(sourceNode, source, options, { ast, map }) {
    const preprocessorName = getPreprocessorTypeByAttribute(sourceNode);
    const javascriptNode = sourceNode.text;
    const preprocessorOutput = await preprocess('js', preprocessorName, options, source, javascriptNode);
    const generatedAst = main$1.parse(preprocessorOutput.code, {
      sourceFileName: options.file,
      inputSourceMap: composeSourcemaps(map, preprocessorOutput.map)
    });
    const generatedAstBody = getProgramBody(generatedAst);
    const bodyWithoutExportDefault = filterNonExportDefaultStatements(generatedAstBody);
    const exportDefaultNode = findExportDefaultStatement(generatedAstBody);
    const outputBody = getProgramBody(ast);

    // add to the ast the "private" javascript content of our tag script node
    outputBody.unshift(...bodyWithoutExportDefault);

    // convert the export default adding its content to the "tag" property exported
    if (exportDefaultNode) extendTagProperty(ast, exportDefaultNode);

    return {
      ast,
      map,
      code: main$1.print(ast).code
    }
  }

  // import {IS_BOOLEAN,IS_CUSTOM,IS_RAW,IS_SPREAD,IS_VOID} from '@riotjs/parser/src/constants'

  const BINDING_TYPES = 'bindingTypes';
  const EACH_BINDING_TYPE = 'EACH';
  const IF_BINDING_TYPE = 'IF';
  const TAG_BINDING_TYPE = 'TAG';


  const EXPRESSION_TYPES = 'expressionTypes';
  const ATTRIBUTE_EXPRESSION_TYPE = 'ATTRIBUTE';
  const VALUE_EXPRESSION_TYPE = 'VALUE';
  const TEXT_EXPRESSION_TYPE = 'TEXT';
  const EVENT_EXPRESSION_TYPE = 'EVENT';

  const TEMPLATE_FN = 'template';
  const SCOPE = 'scope';
  const GET_COMPONENT_FN = 'getComponent';

  // keys needed to create the DOM bindings
  const BINDING_SELECTOR_KEY = 'selector';
  const BINDING_GET_COMPONENT_KEY = 'getComponent';
  const BINDING_TEMPLATE_KEY = 'template';
  const BINDING_TYPE_KEY = 'type';
  const BINDING_REDUNDANT_ATTRIBUTE_KEY = 'redundantAttribute';
  const BINDING_CONDITION_KEY = 'condition';
  const BINDING_ITEM_NAME_KEY = 'itemName';
  const BINDING_GET_KEY_KEY = 'getKey';
  const BINDING_INDEX_NAME_KEY = 'indexName';
  const BINDING_EVALUATE_KEY = 'evaluate';
  const BINDING_NAME_KEY = 'name';
  const BINDING_SLOTS_KEY = 'slots';
  const BINDING_EXPRESSIONS_KEY = 'expressions';
  const BINDING_CHILD_NODE_INDEX_KEY = 'childNodeIndex';
  // slots keys
  const BINDING_BINDINGS_KEY = 'bindings';
  const BINDING_ID_KEY = 'id';
  const BINDING_HTML_KEY = 'html';
  const BINDING_ATTRIBUTES_KEY = 'attributes';

  // DOM directives
  const IF_DIRECTIVE = 'if';
  const EACH_DIRECTIVE = 'each';
  const KEY_ATTRIBUTE = 'key';
  const SLOT_ATTRIBUTE = 'slot';
  const IS_DIRECTIVE$1 = 'is';

  const TEXT_NODE_EXPRESSION_PLACEHOLDER = '<!---->';
  const BINDING_SELECTOR_PREFIX = 'expr';
  const IS_VOID_NODE = 'isVoid';
  const IS_CUSTOM_NODE = 'isCustom';
  const IS_BOOLEAN_ATTRIBUTE = 'isBoolean';
  const IS_SPREAD_ATTRIBUTE = 'isSpread';

  const browserAPIs = Object.keys(globals$3.browser);
  const builtinAPIs = Object.keys(globals$3.builtin);

  const isIdentifier = namedTypes$1.Identifier.check;
  const isLiteral = namedTypes$1.Literal.check;
  const isExpressionStatement = namedTypes$1.ExpressionStatement.check;
  const isObjectExpression = namedTypes$1.ObjectExpression.check;
  const isThisExpression = namedTypes$1.ThisExpression.check;
  const isSequenceExpression = namedTypes$1.SequenceExpression.check;
  const isBinaryExpression = namedTypes$1.BinaryExpression.check;

  const isBrowserAPI = ({name}) => browserAPIs.includes(name);
  const isBuiltinAPI = ({name}) => builtinAPIs.includes(name);
  const isRaw = (node) => node && node.raw;

  function nullNode() {
    return builders.literal(null)
  }

  function simplePropertyNode(key, value) {
    return builders.property('init', builders.literal(key), value, false)
  }

  /* eslint-disable */
  // source: https://30secondsofcode.org/function#compose
  var compose = (...fns) => fns.reduce((f, g) => (...args) => f(g(...args)));

  const scope$1 = builders.identifier(SCOPE);
  const getName$1 = node => node && node.name ? node.name : node;

  /**
   * Find the attribute node
   * @param   { string } name -  name of the attribute we want to find
   * @param   { riotParser.nodeTypes.TAG } node - a tag node
   * @returns { riotParser.nodeTypes.ATTR } attribute node
   */
  function findAttribute(name, node) {
    return node.attributes && node.attributes.find(attr => getName$1(attr) === name)
  }

  const findIfAttribute = curry(findAttribute)(IF_DIRECTIVE);
  const findEachAttribute = curry(findAttribute)(EACH_DIRECTIVE);
  const findKeyAttribute = curry(findAttribute)(KEY_ATTRIBUTE);
  const findIsAttribute = curry(findAttribute)(IS_DIRECTIVE$1);
  const hasIfAttribute = compose(Boolean, findIfAttribute);
  const hasEachAttribute = compose(Boolean, findEachAttribute);
  const hasKeyAttribute = compose(Boolean, findKeyAttribute);
  const hasIsAttribute = compose(Boolean, findIsAttribute);

  function createExpressionSourcemap(expression, sourceFile, sourceCode) {
    const sourcemap = createSourcemap({ file: sourceFile })

    ;[expression.start, expression.end].forEach(position => {
      const location = getLineAndColumnByPosition(sourceCode, position);

      sourcemap.addMapping({
        source: sourceFile,
        generated: location,
        original: location
      });
    });
  }

  /**
   * Check if a node name is part of the browser or builtin javascript api or it belongs to the current scope
   * @param   { types.NodePath } path - containing the current node visited
   * @returns {boolean} true if it's a global api variable
   */
  function isGlobal({ scope, node }) {
    return isRaw(node) || isBuiltinAPI(node) || isBrowserAPI(node) || scope.lookup(getName$1(node))
  }

  /**
   * Replace the path scope with a member Expression
   * @param   { types.NodePath } path - containing the current node visited
   * @param   { types.Node } property - node we want to prefix with the scope identifier
   * @returns {undefined} this is a void function
   */
  function replacePathScope(path, property) {
    path.replace(builders.memberExpression(
      scope$1,
      property,
      false
    ));
  }

  /**
   * Change the nodes scope adding the `scope` prefix
   * @param   { types.NodePath } path - containing the current node visited
   * @returns { boolean } return false if we want to stop the tree traversal
   * @context { types.visit }
   */
  function updateNodeScope(path) {
    if (!isGlobal(path)) {
      replacePathScope(path, path.node);

      return false
    }

    this.traverse(path);
  }

  /**
   * Change the scope of the member expressions
   * @param   { types.NodePath } path - containing the current node visited
   * @returns { boolean } return always false because we want to check only the first node object
   */
  function visitMemberExpression(path) {
    if (!isGlobal({ node: path.node.object, scope: path.scope })) {
      replacePathScope(path, isThisExpression(path.node.object) ? path.node.property : path.node);
    }

    return false
  }


  /**
   * Objects properties should be handled a bit differently from the Identifier
   * @param   { types.NodePath } path - containing the current node visited
   * @returns { boolean } return false if we want to stop the tree traversal
   */
  function visitProperty(path) {
    const value = path.node.value;

    if (isIdentifier(value)) {
      updateNodeScope(path.get('value'));
    } else {
      this.traverse(path.get('value'));
    }

    return false
  }

  /**
   * The this expressions should be replaced with the scope
   * @param   { types.NodePath } path - containing the current node visited
   * @returns { boolean } return false if we want to stop the tree traversal
   */
  function visitThisExpression(path) {
    path.replace(scope$1);
    this.traverse(path);
  }

  /**
   * Update the scope of the global nodes
   * @param   { Object } ast - ast program
   * @returns { Object } the ast program with all the global nodes updated
   */
  function updateNodesScope(ast) {
    const ignorePath = () => false;

    types$3.visit(ast, {
      visitIdentifier: updateNodeScope,
      visitMemberExpression,
      visitProperty,
      visitThisExpression,
      visitClassExpression: ignorePath
    });

    return ast
  }

  /**
   * Convert any expression to an AST tree
   * @param   { Object } expression - expression parsed by the riot parser
   * @param   { string } sourceFile - original tag file
   * @param   { string } sourceCode - original tag source code
   * @returns { Object } the ast generated
   */
  function createASTFromExpression(expression, sourceFile, sourceCode) {
    return main$1.parse(`(${expression.text})`, {
      sourceFileName: sourceFile,
      inputSourceMap: sourceFile && createExpressionSourcemap(expression, sourceFile, sourceCode)
    })
  }

  const getEachItemName = expression => isSequenceExpression(expression.left) ? expression.left.expressions[0] : expression.left;
  const getEachIndexName = expression => isSequenceExpression(expression.left) ? expression.left.expressions[1] : null;
  const getEachValue = expression => expression.right;
  const nameToliteral = compose(builders.literal, getName$1);

  /**
   * Get the each expression properties to create properly the template binding
   * @param   { DomBinding.Expression } eachExpression - original each expression data
   * @param   { string } sourceFile - original tag file
   * @param   { string } sourceCode - original tag source code
   * @returns { Array } AST nodes that are needed to build an each binding
   */
  function getEachExpressionProperties(eachExpression, sourceFile, sourceCode) {
    const ast = createASTFromExpression(eachExpression, sourceFile, sourceCode);
    const body = ast.program.body;
    const firstNode = body[0];


    if (!isExpressionStatement(firstNode)) {
      panic$2(`The each directives supported should be of type "ExpressionStatement",you have provided a "${firstNode.type}"`);
    }

    const { expression } = firstNode;

    return [
      simplePropertyNode(
        BINDING_ITEM_NAME_KEY,
        compose(nameToliteral, getEachItemName)(expression)
      ),
      simplePropertyNode(
        BINDING_INDEX_NAME_KEY,
        compose(nameToliteral, getEachIndexName)(expression)
      ),
      simplePropertyNode(
        BINDING_EVALUATE_KEY,
        compose(
          e => toScopedFunction(e, sourceFile, sourceCode),
          e => ({
            ...eachExpression,
            text: main$1.print(e).code
          }),
          getEachValue
        )(expression)
      )
    ]
  }

  /**
   * Create the bindings template property
   * @param   {Array} args - arguments to pass to the template function
   * @returns {ASTNode} a binding template key
   */
  function createTemplateProperty(args) {
    return simplePropertyNode(
      BINDING_TEMPLATE_KEY,
      args ? callTemplateFunction(...args) : nullNode()
    )
  }

  /**
   * Try to get the expression of an attribute node
   * @param   { RiotParser.Node.Attribute } attribute - riot parser attribute node
   * @returns { RiotParser.Node.Expression } attribute expression value
   */
  function getAttributeExpression(attribute) {
    return attribute.expressions ? attribute.expressions[0] : {
      // if no expression was found try to typecast the attribute value
      ...attribute,
      text: attribute.value
    }
  }

  /**
   * Convert any parser option to a valid template one
   * @param   { RiotParser.Node.Expression } expression - expression parsed by the riot parser
   * @param   { string } sourceFile - original tag file
   * @param   { string } sourceCode - original tag source code
   * @returns { Object } a FunctionExpression object
   *
   * @example
   *  toScopedFunction('foo + bar') // scope.foo + scope.bar
   *
   * @example
   *  toScopedFunction('foo.baz + bar') // scope.foo.baz + scope.bar
   */
  function toScopedFunction(expression, sourceFile, sourceCode) {
    const ast = createASTFromExpression(expression, sourceFile, sourceCode);
    const generatedAST = updateNodesScope(ast);
    const astBody = generatedAST.program.body;
    const expressionAST = astBody[0] ? astBody[0].expression : astBody;

    return builders.functionExpression(
      null,
      [scope$1],
      builders.blockStatement([builders.returnStatement(
        expressionAST
      )])
    )
  }

  /**
   * Wrap a string in a template literal expression
   * @param   {string} string - target string
   * @returns {string} a template literal
   */
  function wrapInBacktick(string) {
    return `\`${string}\``
  }

  /**
   * Simple bindings might contain multiple expressions like for example: "{foo} and {bar}"
   * This helper aims to merge them in a template literal if it's necessary
   * @param   {RiotParser.Node} node - riot parser node
   * @returns {string} either a string representing a template literal or simply the first expression matched
   */
  function mergeNodeExpressions(node) {
    if (node.expressions.length === 1) {
      return node.expressions[0].text
    }

    const charAddedProIteration = 3; // ${}

    // a tricky way to merge nodes containing multiple expressions in the same string
    const values = node.expressions.reduce((string, expression, index) => {
      const bracketsLength = expression.end - expression.start - expression.text.length;
      const offset = index * (charAddedProIteration - bracketsLength);
      const expressionStart = expression.start - node.start + offset;
      const expressionEnd = expression.end - node.start + offset;

      return `${string.substring(0, expressionStart)}$\{${expression.text}}${string.substring(expressionEnd)}`
    }, node.text);

    return wrapInBacktick(values)
  }

  /**
   * Create the template call function
   * @param   {Array|string|Node.Literal} template - template string
   * @param   {Array<AST.Nodes>} bindings - template bindings provided as AST nodes
   * @returns {Node.CallExpression} template call expression
   */
  function callTemplateFunction(template, bindings) {
    return builders.callExpression(builders.identifier(TEMPLATE_FN), [
      template ? builders.literal(template) : nullNode(),
      bindings ? builders.arrayExpression(bindings) : nullNode()
    ])
  }

  /**
   * Convert any DOM attribute into a valid DOM selector useful for the querySelector API
   * @param   { string } attributeName - name of the attribute to query
   * @returns { string } the attribute transformed to a query selector
   */
  const attributeNameToDOMQuerySelector = attributeName => `[${attributeName}]`;

  /**
   * Create the properties to query a DOM node
   * @param   { string } attributeName - attribute name needed to identify a DOM node
   * @returns { Array<AST.Node> } array containing the selector properties needed for the binding
   */
  function createSelectorProperties(attributeName) {
    return attributeName ? [
      simplePropertyNode(BINDING_REDUNDANT_ATTRIBUTE_KEY, builders.literal(attributeName)),
      simplePropertyNode(BINDING_SELECTOR_KEY,
        compose(builders.literal, attributeNameToDOMQuerySelector)(attributeName)
      )
    ] : []
  }

  /**
   * Clean binding or custom attributes
   * @param   {RiotParser.Node} node - riot parser node
   * @returns {Array<RiotParser.Node.Attr>} only the attributes that are not bindings or directives
   */
  function cleanAttributes(node) {
    return getNodeAttributes(node).filter(attribute => ![
      IF_DIRECTIVE,
      EACH_DIRECTIVE,
      KEY_ATTRIBUTE,
      SLOT_ATTRIBUTE,
      IS_DIRECTIVE$1
    ].includes(attribute.name))
  }

  /**
   * Clone the node filtering out the selector attribute from the attributes list
   * @param   {RiotParser.Node} node - riot parser node
   * @param   {string} selectorAttribute - name of the selector attribute to filter out
   * @returns {RiotParser.Node} the node with the attribute cleaned up
   */
  function cloneNodeWithoutSelectorAttribute(node, selectorAttribute) {
    return {
      ...node,
      attributes: getNodeAttributes(node).filter(attribute => attribute.name !== selectorAttribute)
    }
  }

  /**
   * Create a root node proxing only its nodes and attributes
   * @param   {RiotParser.Node} node - riot parser node
   * @returns {RiotParser.Node} root node
   */
  function createRootNode(node) {
    return {
      nodes: getChildrenNodes(node),
      isRoot: true,
      // root nodes shuold't have directives
      attributes: cleanAttributes(node)
    }
  }

  /**
   * Get all the child nodes of a RiotParser.Node
   * @param   {RiotParser.Node} node - riot parser node
   * @returns {Array<RiotParser.Node>} all the child nodes found
   */
  function getChildrenNodes(node) {
    return node && node.nodes ? node.nodes : []
  }

  /**
   * Get all the attributes of a riot parser node
   * @param   {RiotParser.Node} node - riot parser node
   * @returns {Array<RiotParser.Node.Attribute>} all the attributes find
   */
  function getNodeAttributes(node) {
    return node.attributes ? node.attributes : []
  }
  /**
   * Get the name of a custom node transforming it into an expression node
   * @param   {RiotParser.Node} node - riot parser node
   * @returns {RiotParser.Node.Attr} the node name as expression attribute
   */
  function getCustomNodeNameAsExpression(node) {
    const isAttribute = findIsAttribute(node);
    const toRawString = val => `'${val}'`;

    if (isAttribute) {
      return isAttribute.expressions ? isAttribute.expressions[0] : {
        ...isAttribute,
        text: toRawString(isAttribute.value)
      }
    }

    return { ...node, text: toRawString(getName$1(node)) }
  }

  /**
   * Find all the node attributes that are not expressions
   * @param   {RiotParser.Node} node - riot parser node
   * @returns {Array} list of all the static attributes
   */
  function findStaticAttributes(node) {
    return getNodeAttributes(node).filter(attribute => !hasExpressions(attribute))
  }

  /**
   * Find all the node attributes that have expressions
   * @param   {RiotParser.Node} node - riot parser node
   * @returns {Array} list of all the dynamic attributes
   */
  function findDynamicAttributes(node) {
    return getNodeAttributes(node).filter(hasExpressions)
  }

  /**
   * True if the node has the isCustom attribute set
   * @param   {RiotParser.Node} node - riot parser node
   * @returns {boolean} true if either it's a riot component or a custom element
   */
  function isCustomNode(node) {
    return !!(node[IS_CUSTOM_NODE] || hasIsAttribute(node))
  }


  /**
   * True if the node has the isVoid attribute set
   * @param   {RiotParser.Node} node - riot parser node
   * @returns {boolean} true if the node is self closing
   */
  function isVoidNode(node) {
    return !!node[IS_VOID_NODE]
  }

  /**
   * True if the riot parser did find a tag node
   * @param   {RiotParser.Node} node - riot parser node
   * @returns {boolean} true only for the tag nodes
   */
  function isTagNode(node) {
    return node.type === nodeTypes.TAG
  }

  /**
   * True if the riot parser did find a text node
   * @param   {RiotParser.Node} node - riot parser node
   * @returns {boolean} true only for the text nodes
   */
  function isTextNode(node) {
    return node.type === nodeTypes.TEXT
  }

  /**
   * True if the node parsed is the root one
   * @param   {RiotParser.Node} node - riot parser node
   * @returns {boolean} true only for the root nodes
   */
  function isRootNode(node) {
    return node.isRoot
  }

  /**
   * True if the attribute parsed is of type spread one
   * @param   {RiotParser.Node} node - riot parser node
   * @returns {boolean} true if the attribute node is of type spread
   */
  function isSpreadAttribute(node) {
    return node[IS_SPREAD_ATTRIBUTE]
  }

  /**
   * True if the node is an attribute and its name is "value"
   * @param   {RiotParser.Node} node - riot parser node
   * @returns {boolean} true only for value attribute nodes
   */
  function isValueAttribute(node) {
    return node.name === 'value'
  }

  /**
   * True if the node is an attribute and a DOM handler
   * @param   {RiotParser.Node} node - riot parser node
   * @returns {boolean} true only for dom listener attribute nodes
   */
  const isEventAttribute = (() => {
    const EVENT_ATTR_RE = /^on/;
    return node => EVENT_ATTR_RE.test(node.name)
  })();

  /**
   * True if the node has expressions or expression attributes
   * @param   {RiotParser.Node} node - riot parser node
   * @returns {boolean} ditto
   */
  function hasExpressions(node) {
    return !!(
      node.expressions ||
      // has expression attributes
      (getNodeAttributes(node).some(attribute => hasExpressions(attribute))) ||
      // has child text nodes with expressions
      (node.nodes && node.nodes.some(node => isTextNode(node) && hasExpressions(node)))
    )
  }

  /**
   * Convert all the node static attributes to strings
   * @param   {RiotParser.Node} node - riot parser node
   * @returns {string} all the node static concatenated as string
   */
  function staticAttributesToString(node) {
    return findStaticAttributes(node)
      .map(attribute => attribute[IS_BOOLEAN_ATTRIBUTE] || !attribute.value ?
        attribute.name :
        `${attribute.name}="${attribute.value}"`
      ).join(' ')
  }


  /**
   * Convert a riot parser opening node into a string
   * @param   {RiotParser.Node} node - riot parser node
   * @returns {string} the node as string
   */
  function nodeToString(node) {
    const attributes = staticAttributesToString(node);

    switch(true) {
    case isTagNode(node):
      return `<${node.name}${attributes ? ` ${attributes}` : ''}${isVoidNode(node) ? '/' : ''}>`
    case isTextNode(node):
      return hasExpressions(node) ? TEXT_NODE_EXPRESSION_PLACEHOLDER : node.text
    default:
      return ''
    }
  }

  /**
   * Close an html node
   * @param   {RiotParser.Node} node - riot parser node
   * @returns {string} the closing tag of the html tag node passed to this function
   */
  function closeTag(node) {
    return node.name ? `</${node.name}>` : ''
  }

  /**
   * True if the node has not expression set nor bindings directives
   * @param   {RiotParser.Node} node - riot parser node
   * @returns {boolean} true only if it's a static node that doesn't need bindings or expressions
   */
  function isStaticNode(node) {
    return [
      hasExpressions,
      findEachAttribute,
      findIfAttribute,
      isCustomNode
    ].every(test => !test(node))
  }

  /**
   * True if the node is a directive having its own template
   * @param   {RiotParser.Node} node - riot parser node
   * @returns {boolean} true only for the IF EACH and TAG bindings
   */
  function hasItsOwnTemplate(node) {
    return [
      findEachAttribute,
      findIfAttribute,
      isCustomNode
    ].some(test => test(node))
  }


  /**
   * Create a selector that will be used to find the node via dom-bindings
   * @param   {number} id - temporary variable that will be increased anytime this function will be called
   * @returns {string} selector attribute needed to bind a riot expression
   */
  const createBindingSelector = (function createSelector(id = 0) {
    return () => `${BINDING_SELECTOR_PREFIX}${id++}`
  }());

  /**
   * Simple clone deep function, do not use it for classes or recursive objects!
   * @param   {*} source - possibily an object to clone
   * @returns {*} the object we wanted to clone
   */
  function cloneDeep(source) {
    return JSON.parse(JSON.stringify(source))
  }

  /**
   * Create a simple attribute expression
   * @param   {RiotParser.Node.Attr} sourceNode - the custom tag
   * @param   {stiring} sourceFile - source file path
   * @param   {string} sourceCode - original source
   * @returns {AST.Node} object containing the expression binding keys
   */
  function createAttributeExpression(sourceNode, sourceFile, sourceCode) {
    return builders.objectExpression([
      simplePropertyNode(BINDING_TYPE_KEY,
        builders.memberExpression(
          builders.identifier(EXPRESSION_TYPES),
          builders.identifier(ATTRIBUTE_EXPRESSION_TYPE),
          false
        ),
      ),
      simplePropertyNode(BINDING_NAME_KEY, isSpreadAttribute(sourceNode) ? nullNode() : builders.literal(sourceNode.name)),
      simplePropertyNode(
        BINDING_EVALUATE_KEY,
        hasExpressions(sourceNode) ?
          // dynamic attribute
          toScopedFunction(sourceNode.expressions[0], sourceFile, sourceCode) :
          // static attribute
          builders.functionExpression(
            null,
            [],
            builders.blockStatement([
              builders.returnStatement(builders.literal(sourceNode.value || true))
            ])
          )
      )
    ])
  }

  /**
   * Find the slots in the current component and group them under the same id
   * @param   {RiotParser.Node.Tag} sourceNode - the custom tag
   * @returns {Object} object containing all the slots grouped by name
   */
  function groupSlots(sourceNode) {
    return sourceNode.nodes.reduce((acc, node) => {
      const slotAttribute = findSlotAttribute(node);

      if (slotAttribute) {
        acc[slotAttribute.value] = node;
      } else {
        acc.default = {
          nodes: [...getChildrenNodes(acc.default), node]
        };
      }

      return acc
    }, {
      default: null
    })
  }

  /**
   * Create the slot entity to pass to the riot-dom bindings
   * @param   {string} id - slot id
   * @param   {RiotParser.Node.Tag} sourceNode - slot root node
   * @param   {stiring} sourceFile - source file path
   * @param   {string} sourceCode - original source
   * @returns {AST.Node} ast node containing the slot object properties
   */
  function buildSlot(id, sourceNode, sourceFile, sourceCode) {
    const cloneNode = {
      ...sourceNode,
      // avoid to render the slot attribute
      attributes: getNodeAttributes(sourceNode).filter(attribute => attribute.name !== SLOT_ATTRIBUTE)
    };
    const [html, bindings] = build(cloneNode, sourceFile, sourceCode);

    return builders.objectExpression([
      simplePropertyNode(BINDING_ID_KEY, builders.literal(id)),
      simplePropertyNode(BINDING_HTML_KEY, builders.literal(html)),
      simplePropertyNode(BINDING_BINDINGS_KEY, builders.arrayExpression(bindings))
    ])
  }

  /**
   * Find the slot attribute if it exists
   * @param   {RiotParser.Node.Tag} sourceNode - the custom tag
   * @returns {RiotParser.Node.Attr|undefined} the slot attribute found
   */
  function findSlotAttribute(sourceNode) {
    return getNodeAttributes(sourceNode).find(attribute => attribute.name === SLOT_ATTRIBUTE)
  }

  /**
   * Transform a RiotParser.Node.Tag into a tag binding
   * @param   { RiotParser.Node.Tag } sourceNode - the custom tag
   * @param   { string } selectorAttribute - attribute needed to select the target node
   * @param   { stiring } sourceFile - source file path
   * @param   { string } sourceCode - original source
   * @returns { AST.Node } an each binding node
   */
  function createTagBinding(sourceNode, selectorAttribute, sourceFile, sourceCode) {
    return builders.objectExpression([
      simplePropertyNode(BINDING_TYPE_KEY,
        builders.memberExpression(
          builders.identifier(BINDING_TYPES),
          builders.identifier(TAG_BINDING_TYPE),
          false
        ),
      ),
      simplePropertyNode(BINDING_GET_COMPONENT_KEY, builders.identifier(GET_COMPONENT_FN)),
      simplePropertyNode(
        BINDING_EVALUATE_KEY,
        toScopedFunction(getCustomNodeNameAsExpression(sourceNode), sourceFile, sourceCode)
      ),
      simplePropertyNode(BINDING_SLOTS_KEY, builders.arrayExpression([
        ...Object.entries(groupSlots(sourceNode))
          .filter(([,value]) => value)
          .map(([key, value]) => buildSlot(key, value, sourceFile, sourceCode))
      ])),
      simplePropertyNode(BINDING_ATTRIBUTES_KEY, builders.arrayExpression([
        ...cleanAttributes(sourceNode)
          .filter(attribute => attribute.name !== selectorAttribute)
          .map(attribute => createAttributeExpression(attribute, sourceFile, sourceCode))
      ])),
      ...createSelectorProperties(selectorAttribute)
    ])
  }

  /**
   * Transform a RiotParser.Node.Tag into an each binding
   * @param   { RiotParser.Node.Tag } sourceNode - tag containing the each attribute
   * @param   { string } selectorAttribute - attribute needed to select the target node
   * @param   { stiring } sourceFile - source file path
   * @param   { string } sourceCode - original source
   * @returns { AST.Node } an each binding node
   */
  function createEachBinding(sourceNode, selectorAttribute, sourceFile, sourceCode) {
    const [ifAttribute, eachAttribute, keyAttribute] = [
      findIfAttribute,
      findEachAttribute,
      findKeyAttribute
    ].map(f => f(sourceNode));
    const mightBeARiotComponent = isCustomNode(sourceNode);
    const attributeOrNull = attribute => attribute ? toScopedFunction(getAttributeExpression(attribute), sourceFile, sourceCode) : nullNode();

    return builders.objectExpression([
      simplePropertyNode(BINDING_TYPE_KEY,
        builders.memberExpression(
          builders.identifier(BINDING_TYPES),
          builders.identifier(EACH_BINDING_TYPE),
          false
        ),
      ),
      simplePropertyNode(BINDING_GET_KEY_KEY, attributeOrNull(keyAttribute)),
      simplePropertyNode(BINDING_CONDITION_KEY, attributeOrNull(ifAttribute)),
      createTemplateProperty(mightBeARiotComponent ?
        [null, [
          createTagBinding(
            cloneNodeWithoutSelectorAttribute(sourceNode),
            null,
            sourceCode,
            sourceCode
          )]
        ] :
        build(createRootNode(sourceNode), sourceCode, sourceCode)
      ),
      ...createSelectorProperties(selectorAttribute),
      ...compose(getEachExpressionProperties, getAttributeExpression)(eachAttribute)
    ])
  }

  /**
   * Transform a RiotParser.Node.Tag into an if binding
   * @param   { RiotParser.Node.Tag } sourceNode - tag containing the if attribute
   * @param   { string } selectorAttribute - attribute needed to select the target node
   * @param   { stiring } sourceFile - source file path
   * @param   { string } sourceCode - original source
   * @returns { AST.Node } an each binding node
   */
  function createIfBinding(sourceNode, selectorAttribute, sourceFile, sourceCode) {
    const ifAttribute = findIfAttribute(sourceNode);
    const mightBeARiotComponent = isCustomNode(sourceNode);

    return builders.objectExpression([
      simplePropertyNode(BINDING_TYPE_KEY,
        builders.memberExpression(
          builders.identifier(BINDING_TYPES),
          builders.identifier(IF_BINDING_TYPE),
          false
        ),
      ),
      simplePropertyNode(
        BINDING_EVALUATE_KEY,
        toScopedFunction(ifAttribute.expressions[0], sourceFile, sourceCode)
      ),
      ...createSelectorProperties(selectorAttribute),
      createTemplateProperty(mightBeARiotComponent ?
        [null, [
          createTagBinding(
            cloneNodeWithoutSelectorAttribute(sourceNode),
            null,
            sourceCode,
            sourceCode
          )]
        ] :
        build(createRootNode(sourceNode), sourceCode, sourceCode)
      )
    ])
  }

  /**
   * Create a simple event expression
   * @param   {RiotParser.Node.Attr} sourceNode - attribute containing the event handlers
   * @param   {stiring} sourceFile - source file path
   * @param   {string} sourceCode - original source
   * @returns {AST.Node} object containing the expression binding keys
   */
  function createEventExpression(sourceNode, sourceFile, sourceCode) {
    return builders.objectExpression([
      simplePropertyNode(BINDING_TYPE_KEY,
        builders.memberExpression(
          builders.identifier(EXPRESSION_TYPES),
          builders.identifier(EVENT_EXPRESSION_TYPE),
          false
        ),
      ),
      simplePropertyNode(BINDING_NAME_KEY, builders.literal(sourceNode.name)),
      simplePropertyNode(
        BINDING_EVALUATE_KEY,
        toScopedFunction(sourceNode.expressions[0], sourceFile, sourceCode)
      )
    ])
  }

  /**
   * Create a text expression
   * @param   {RiotParser.Node.Text} sourceNode - text node to parse
   * @param   {stiring} sourceFile - source file path
   * @param   {string} sourceCode - original source
   * @param   {number} childNodeIndex - position of the child text node in its parent children nodes
   * @returns {AST.Node} object containing the expression binding keys
   */
  function createTextExpression(sourceNode, sourceFile, sourceCode, childNodeIndex) {
    return builders.objectExpression([
      simplePropertyNode(BINDING_TYPE_KEY,
        builders.memberExpression(
          builders.identifier(EXPRESSION_TYPES),
          builders.identifier(TEXT_EXPRESSION_TYPE),
          false
        ),
      ),
      simplePropertyNode(
        BINDING_CHILD_NODE_INDEX_KEY,
        builders.literal(childNodeIndex)
      ),
      simplePropertyNode(
        BINDING_EVALUATE_KEY,
        toScopedFunction({
          start: sourceNode.start,
          end: sourceNode.end,
          text: mergeNodeExpressions(sourceNode)
        }, sourceFile, sourceCode)
      )
    ])
  }

  function createValueExpression(sourceNode, sourceFile, sourceCode) {
    return builders.objectExpression([
      simplePropertyNode(BINDING_TYPE_KEY,
        builders.memberExpression(
          builders.identifier(EXPRESSION_TYPES),
          builders.identifier(VALUE_EXPRESSION_TYPE),
          false
        ),
      ),
      simplePropertyNode(
        BINDING_EVALUATE_KEY,
        toScopedFunction(sourceNode.expressions[0], sourceFile, sourceCode)
      )
    ])
  }

  function createExpression(sourceNode, sourceFile, sourceCode, childNodeIndex) {
    switch (true) {
    case isTextNode(sourceNode):
      return createTextExpression(sourceNode, sourceFile, sourceCode, childNodeIndex)
    case isValueAttribute(sourceNode):
      return createValueExpression(sourceNode, sourceFile, sourceCode)
    case isEventAttribute(sourceNode):
      return createEventExpression(sourceNode, sourceFile, sourceCode)
    default:
      return createAttributeExpression(sourceNode, sourceFile, sourceCode)
    }
  }

  /**
   * Create the attribute expressions
   * @param   {RiotParser.Node} sourceNode - any kind of node parsed via riot parser
   * @param   {stiring} sourceFile - source file path
   * @param   {string} sourceCode - original source
   * @returns {Array} array containing all the attribute expressions
   */
  function createAttributeExpressions(sourceNode, sourceFile, sourceCode) {
    return findDynamicAttributes(sourceNode)
      .map(attribute => createExpression(attribute, sourceFile, sourceCode))
  }

  /**
   * Create the text node expressions
   * @param   {RiotParser.Node} sourceNode - any kind of node parsed via riot parser
   * @param   {stiring} sourceFile - source file path
   * @param   {string} sourceCode - original source
   * @returns {Array} array containing all the text node expressions
   */
  function createTextNodeExpressions(sourceNode, sourceFile, sourceCode) {
    const childrenNodes = getChildrenNodes(sourceNode);

    return childrenNodes
      .filter(isTextNode)
      .filter(hasExpressions)
      .map(node => createExpression(
        node,
        sourceFile,
        sourceCode,
        childrenNodes.indexOf(node)
      ))
  }

  /**
   * Add a simple binding to a riot parser node
   * @param   { RiotParser.Node.Tag } sourceNode - tag containing the if attribute
   * @param   { string } selectorAttribute - attribute needed to select the target node
   * @param   { stiring } sourceFile - source file path
   * @param   { string } sourceCode - original source
   * @returns { AST.Node } an each binding node
   */
  function createSimpleBinding(sourceNode, selectorAttribute, sourceFile, sourceCode) {
    return builders.objectExpression([
      ...createSelectorProperties(selectorAttribute),
      simplePropertyNode(
        BINDING_EXPRESSIONS_KEY,
        builders.arrayExpression([
          ...createTextNodeExpressions(sourceNode, sourceFile, sourceCode),
          ...createAttributeExpressions(sourceNode, sourceFile, sourceCode)
        ])
      )
    ])
  }

  const BuildingState = Object.freeze({
    html: [],
    bindings: [],
    parent: null
  });

  /**
   * Nodes having bindings should be cloned and new selector properties should be added to them
   * @param   {RiotParser.Node} sourceNode - any kind of node parsed via riot parser
   * @param   {string} bindingsSelector - temporary string to identify the current node
   * @returns {RiotParser.Node} the original node parsed having the new binding selector attribute
   */
  function createBindingsTag(sourceNode, bindingsSelector) {
    if (!bindingsSelector) return sourceNode

    return {
      ...sourceNode,
      // inject the selector bindings into the node attributes
      attributes: [{
        name: bindingsSelector
      }, ...getNodeAttributes(sourceNode)]
    }
  }

  /**
   * Create a generic dynamic node (text or tag) and generate its bindings
   * @param   {RiotParser.Node} sourceNode - any kind of node parsed via riot parser
   * @param   {stiring} sourceFile - source file path
   * @param   {string} sourceCode - original source
   * @param   {BuildingState} state - state representing the current building tree state during the recursion
   * @returns {Array} array containing the html output and bindings for the current node
   */
  function createDynamicNode(sourceNode, sourceFile, sourceCode, state) {
    switch (true) {
    case isTextNode(sourceNode):
      // text nodes will not have any bindings
      return [nodeToString(sourceNode), []]
    default:
      return createTagWithBindings(sourceNode, sourceFile, sourceCode, state)
    }
  }

  /**
   * Create only a dynamic tag node with generating a custom selector and its bindings
   * @param   {RiotParser.Node} sourceNode - any kind of node parsed via riot parser
   * @param   {stiring} sourceFile - source file path
   * @param   {string} sourceCode - original source
   * @param   {BuildingState} state - state representing the current building tree state during the recursion
   * @returns {Array} array containing the html output and bindings for the current node
   */
  function createTagWithBindings(sourceNode, sourceFile, sourceCode) {
    const bindingsSelector = isRootNode(sourceNode) ? null : createBindingSelector();
    const cloneNode = createBindingsTag(sourceNode, bindingsSelector);
    const tagOpeningHTML = nodeToString(cloneNode);

    switch(true) {
    // EACH bindings have prio 1
    case hasEachAttribute(cloneNode):
      return [tagOpeningHTML, [createEachBinding(cloneNode, bindingsSelector, sourceFile, sourceCode)]]
    // IF bindings have prio 2
    case hasIfAttribute(cloneNode):
      return [tagOpeningHTML, [createIfBinding(cloneNode, bindingsSelector, sourceFile, sourceCode)]]
    // TAG bindings have prio 3
    case isCustomNode(cloneNode):
      return [tagOpeningHTML, [createTagBinding(cloneNode, bindingsSelector, sourceFile, sourceCode)]]
    // this node has expressions bound to it
    default:
      return [tagOpeningHTML, [createSimpleBinding(cloneNode, bindingsSelector, sourceFile, sourceCode)]]
    }
  }

  /**
   * Parse a node trying to extract its template and bindings
   * @param   {RiotParser.Node} sourceNode - any kind of node parsed via riot parser
   * @param   {stiring} sourceFile - source file path
   * @param   {string} sourceCode - original source
   * @param   {BuildingState} state - state representing the current building tree state during the recursion
   * @returns {Array} array containing the html output and bindings for the current node
   */
  function parseNode(sourceNode, sourceFile, sourceCode, state) {
    // static nodes have no bindings
    if (isStaticNode(sourceNode)) return [nodeToString(sourceNode), []]
    return createDynamicNode(sourceNode, sourceFile, sourceCode, state)
  }

  /**
   * Build the template and the bindings
   * @param   {RiotParser.Node} sourceNode - any kind of node parsed via riot parser
   * @param   {stiring} sourceFile - source file path
   * @param   {string} sourceCode - original source
   * @param   {BuildingState} state - state representing the current building tree state during the recursion
   * @returns {Array} array containing the html output and the dom bindings
   */
  function build(
    sourceNode,
    sourceFile,
    sourceCode,
    state
  ) {
    if (!sourceNode) panic$2('Something went wrong with your tag DOM parsing, your tag template can\'t be created');

    const [nodeHTML, nodeBindings] = parseNode(sourceNode, sourceFile, sourceCode, state);
    const childrenNodes = getChildrenNodes(sourceNode);
    const currentState = { ...cloneDeep(BuildingState), ...state };

    // mutate the original arrays
    currentState.html.push(...nodeHTML);
    currentState.bindings.push(...nodeBindings);

    // do recursion if
    // this tag has children and it has no special directives bound to it
    if (childrenNodes.length && !hasItsOwnTemplate(sourceNode)) {
      childrenNodes.forEach(node => build(node, sourceFile, sourceCode, { parent: sourceNode, ...currentState }));
    }

    // close the tag if it's not a void one
    if (isTagNode(sourceNode) && !isVoidNode(sourceNode)) {
      currentState.html.push(closeTag(sourceNode));
    }

    return [
      currentState.html.join(''),
      currentState.bindings
    ]
  }

  /**
   * Extend the AST adding the new template property containing our template call to render the component
   * @param   { Object } ast - current output ast
  * @param   { stiring } sourceFile - source file path
   * @param   { string } sourceCode - original source
   * @param   { Object } sourceNode - node generated by the riot compiler
   * @returns { Object } the output ast having the "template" key
   */
  function extendTemplateProperty(ast, sourceFile, sourceCode, sourceNode) {
    types$3.visit(ast, {
      visitProperty(path) {
        if (path.value.key.name === TAG_TEMPLATE_PROPERTY) {
          path.value.value = builders.functionExpression(
            null,
            [
              TEMPLATE_FN,
              EXPRESSION_TYPES,
              BINDING_TYPES,
              GET_COMPONENT_FN
            ].map(builders.identifier),
            builders.blockStatement([
              builders.returnStatement(
                callTemplateFunction(
                  ...build(
                    createRootNode(sourceNode),
                    sourceFile,
                    sourceCode
                  )
                )
              )
            ])
          );

          return false
        }

        this.traverse(path);
      }
    });

    return ast
  }

  /**
   * Generate the component template logic
   * @param   { Object } sourceNode - node generated by the riot compiler
   * @param   { string } source - original component source code
   * @param   { Object } options - user options
   * @param   { Output } output - current compiler output
   * @returns { Promise<Output> } - enhanced output with the result of the current generator
   */
  async function template(sourceNode, source, options, { ast, map }) {
    const output = extendTemplateProperty(ast, options.file, source, sourceNode);

    return { ast: output, map, code: main$1.print(output).code }
  }

  const DEFAULT_OPTIONS = {
    template: 'default',
    file: '[unknown-source-file]',
    scopedCss: true
  };

  /**
   * Create the initial output
   * @param { Sourcemap } map - initial sourcemap
   * @returns { Output } containing the initial code and  AST
   *
   * @example
   * // the output represents the following string in AST
   */
  function createInitialInput(map) {
    const code = `export default { ${TAG_CSS_PROPERTY}: null, ${TAG_LOGIC_PROPERTY}: null, ${TAG_TEMPLATE_PROPERTY}: null }`;
    return {
      ast: main$1.parse(code),
      code,
      map
    }
  }

  /**
   * Generate the output code source together with the sourcemap
   * @param { string } source - source code of the tag we will need to compile
   * @param { string } options - compiling options
   * @returns { Promise<Output> } object containing output code and source map
   */
  async function compile(source, options = {}) {
    const opts = {
      ...DEFAULT_OPTIONS,
      ...options
    };

    const { code, map } = await execute$1('template', opts.template, opts, source);
    const { template: template$$1, css: css$$1, javascript: javascript$$1 } = parser$1(opts).parse(code).output;

    // generate the tag name in runtime
    Object.assign(opts, {
      tagName: template$$1.name
    });

    return ruit(createInitialInput(map),
      hookGenerator(css, css$$1, code, opts),
      hookGenerator(javascript, javascript$$1, code, opts),
      hookGenerator(template, template$$1, code, opts),
      ({ ast }) => main$1.prettyPrint(ast),
      (result) => execute(result, opts),
    )
  }

  /**
   * Prepare the riot parser node transformers
   * @param   { Function } transformer - transformer function
   * @param   { Object } sourceNode - riot parser node
   * @param   { string } source - component source code
   * @param   { Object } options - compiling options
   * @returns { Promise<Output> } object containing output code and source map
   */
  function hookGenerator(transformer, sourceNode, source, options) {
    if (!sourceNode || (sourceNode.nodes && !sourceNode.nodes.length)) {
      return result => result
    }

    return curry(transformer)(sourceNode, source, options)
  }

  // This function can allow you to register postprocessors that will parse the output code
  // here we can run prettifiers, eslint fixes...
  const registerPostprocessor = register$1;

  /** Detect free variable `global` from Node.js. */
  var freeGlobal = typeof commonjsGlobal == 'object' && commonjsGlobal && commonjsGlobal.Object === Object && commonjsGlobal;

  var _freeGlobal = freeGlobal;

  /** Detect free variable `self`. */
  var freeSelf = typeof self == 'object' && self && self.Object === Object && self;

  /** Used as a reference to the global object. */
  var root = _freeGlobal || freeSelf || Function('return this')();

  var _root = root;

  /** Built-in value references. */
  var Symbol$1 = _root.Symbol;

  var _Symbol = Symbol$1;

  /**
   * A specialized version of `_.map` for arrays without support for iteratee
   * shorthands.
   *
   * @private
   * @param {Array} [array] The array to iterate over.
   * @param {Function} iteratee The function invoked per iteration.
   * @returns {Array} Returns the new mapped array.
   */
  function arrayMap(array, iteratee) {
    var index = -1,
        length = array == null ? 0 : array.length,
        result = Array(length);

    while (++index < length) {
      result[index] = iteratee(array[index], index, array);
    }
    return result;
  }

  var _arrayMap = arrayMap;

  /**
   * Checks if `value` is classified as an `Array` object.
   *
   * @static
   * @memberOf _
   * @since 0.1.0
   * @category Lang
   * @param {*} value The value to check.
   * @returns {boolean} Returns `true` if `value` is an array, else `false`.
   * @example
   *
   * _.isArray([1, 2, 3]);
   * // => true
   *
   * _.isArray(document.body.children);
   * // => false
   *
   * _.isArray('abc');
   * // => false
   *
   * _.isArray(_.noop);
   * // => false
   */
  var isArray$4 = Array.isArray;

  var isArray_1 = isArray$4;

  /** Used for built-in method references. */
  var objectProto = Object.prototype;

  /** Used to check objects for own properties. */
  var hasOwnProperty = objectProto.hasOwnProperty;

  /**
   * Used to resolve the
   * [`toStringTag`](http://ecma-international.org/ecma-262/7.0/#sec-object.prototype.tostring)
   * of values.
   */
  var nativeObjectToString = objectProto.toString;

  /** Built-in value references. */
  var symToStringTag = _Symbol ? _Symbol.toStringTag : undefined;

  /**
   * A specialized version of `baseGetTag` which ignores `Symbol.toStringTag` values.
   *
   * @private
   * @param {*} value The value to query.
   * @returns {string} Returns the raw `toStringTag`.
   */
  function getRawTag(value) {
    var isOwn = hasOwnProperty.call(value, symToStringTag),
        tag = value[symToStringTag];

    try {
      value[symToStringTag] = undefined;
      var unmasked = true;
    } catch (e) {}

    var result = nativeObjectToString.call(value);
    if (unmasked) {
      if (isOwn) {
        value[symToStringTag] = tag;
      } else {
        delete value[symToStringTag];
      }
    }
    return result;
  }

  var _getRawTag = getRawTag;

  /** Used for built-in method references. */
  var objectProto$1 = Object.prototype;

  /**
   * Used to resolve the
   * [`toStringTag`](http://ecma-international.org/ecma-262/7.0/#sec-object.prototype.tostring)
   * of values.
   */
  var nativeObjectToString$1 = objectProto$1.toString;

  /**
   * Converts `value` to a string using `Object.prototype.toString`.
   *
   * @private
   * @param {*} value The value to convert.
   * @returns {string} Returns the converted string.
   */
  function objectToString(value) {
    return nativeObjectToString$1.call(value);
  }

  var _objectToString = objectToString;

  /** `Object#toString` result references. */
  var nullTag = '[object Null]',
      undefinedTag = '[object Undefined]';

  /** Built-in value references. */
  var symToStringTag$1 = _Symbol ? _Symbol.toStringTag : undefined;

  /**
   * The base implementation of `getTag` without fallbacks for buggy environments.
   *
   * @private
   * @param {*} value The value to query.
   * @returns {string} Returns the `toStringTag`.
   */
  function baseGetTag(value) {
    if (value == null) {
      return value === undefined ? undefinedTag : nullTag;
    }
    return (symToStringTag$1 && symToStringTag$1 in Object(value))
      ? _getRawTag(value)
      : _objectToString(value);
  }

  var _baseGetTag = baseGetTag;

  /**
   * Checks if `value` is object-like. A value is object-like if it's not `null`
   * and has a `typeof` result of "object".
   *
   * @static
   * @memberOf _
   * @since 4.0.0
   * @category Lang
   * @param {*} value The value to check.
   * @returns {boolean} Returns `true` if `value` is object-like, else `false`.
   * @example
   *
   * _.isObjectLike({});
   * // => true
   *
   * _.isObjectLike([1, 2, 3]);
   * // => true
   *
   * _.isObjectLike(_.noop);
   * // => false
   *
   * _.isObjectLike(null);
   * // => false
   */
  function isObjectLike(value) {
    return value != null && typeof value == 'object';
  }

  var isObjectLike_1 = isObjectLike;

  /** `Object#toString` result references. */
  var symbolTag = '[object Symbol]';

  /**
   * Checks if `value` is classified as a `Symbol` primitive or object.
   *
   * @static
   * @memberOf _
   * @since 4.0.0
   * @category Lang
   * @param {*} value The value to check.
   * @returns {boolean} Returns `true` if `value` is a symbol, else `false`.
   * @example
   *
   * _.isSymbol(Symbol.iterator);
   * // => true
   *
   * _.isSymbol('abc');
   * // => false
   */
  function isSymbol(value) {
    return typeof value == 'symbol' ||
      (isObjectLike_1(value) && _baseGetTag(value) == symbolTag);
  }

  var isSymbol_1 = isSymbol;

  /** Used as references for various `Number` constants. */
  var INFINITY = 1 / 0;

  /** Used to convert symbols to primitives and strings. */
  var symbolProto = _Symbol ? _Symbol.prototype : undefined,
      symbolToString = symbolProto ? symbolProto.toString : undefined;

  /**
   * The base implementation of `_.toString` which doesn't convert nullish
   * values to empty strings.
   *
   * @private
   * @param {*} value The value to process.
   * @returns {string} Returns the string.
   */
  function baseToString(value) {
    // Exit early for strings to avoid a performance hit in some environments.
    if (typeof value == 'string') {
      return value;
    }
    if (isArray_1(value)) {
      // Recursively convert values (susceptible to call stack limits).
      return _arrayMap(value, baseToString) + '';
    }
    if (isSymbol_1(value)) {
      return symbolToString ? symbolToString.call(value) : '';
    }
    var result = (value + '');
    return (result == '0' && (1 / value) == -INFINITY) ? '-0' : result;
  }

  var _baseToString = baseToString;

  /**
   * Converts `value` to a string. An empty string is returned for `null`
   * and `undefined` values. The sign of `-0` is preserved.
   *
   * @static
   * @memberOf _
   * @since 4.0.0
   * @category Lang
   * @param {*} value The value to convert.
   * @returns {string} Returns the converted string.
   * @example
   *
   * _.toString(null);
   * // => ''
   *
   * _.toString(-0);
   * // => '-0'
   *
   * _.toString([1, 2, 3]);
   * // => '1,2,3'
   */
  function toString(value) {
    return value == null ? '' : _baseToString(value);
  }

  var toString_1 = toString;

  /**
   * The base implementation of `_.slice` without an iteratee call guard.
   *
   * @private
   * @param {Array} array The array to slice.
   * @param {number} [start=0] The start position.
   * @param {number} [end=array.length] The end position.
   * @returns {Array} Returns the slice of `array`.
   */
  function baseSlice(array, start, end) {
    var index = -1,
        length = array.length;

    if (start < 0) {
      start = -start > length ? 0 : (length + start);
    }
    end = end > length ? length : end;
    if (end < 0) {
      end += length;
    }
    length = start > end ? 0 : ((end - start) >>> 0);
    start >>>= 0;

    var result = Array(length);
    while (++index < length) {
      result[index] = array[index + start];
    }
    return result;
  }

  var _baseSlice = baseSlice;

  /**
   * Casts `array` to a slice if it's needed.
   *
   * @private
   * @param {Array} array The array to inspect.
   * @param {number} start The start position.
   * @param {number} [end=array.length] The end position.
   * @returns {Array} Returns the cast slice.
   */
  function castSlice(array, start, end) {
    var length = array.length;
    end = end === undefined ? length : end;
    return (!start && end >= length) ? array : _baseSlice(array, start, end);
  }

  var _castSlice = castSlice;

  /** Used to compose unicode character classes. */
  var rsAstralRange = '\\ud800-\\udfff',
      rsComboMarksRange = '\\u0300-\\u036f',
      reComboHalfMarksRange = '\\ufe20-\\ufe2f',
      rsComboSymbolsRange = '\\u20d0-\\u20ff',
      rsComboRange = rsComboMarksRange + reComboHalfMarksRange + rsComboSymbolsRange,
      rsVarRange = '\\ufe0e\\ufe0f';

  /** Used to compose unicode capture groups. */
  var rsZWJ = '\\u200d';

  /** Used to detect strings with [zero-width joiners or code points from the astral planes](http://eev.ee/blog/2015/09/12/dark-corners-of-unicode/). */
  var reHasUnicode = RegExp('[' + rsZWJ + rsAstralRange  + rsComboRange + rsVarRange + ']');

  /**
   * Checks if `string` contains Unicode symbols.
   *
   * @private
   * @param {string} string The string to inspect.
   * @returns {boolean} Returns `true` if a symbol is found, else `false`.
   */
  function hasUnicode(string) {
    return reHasUnicode.test(string);
  }

  var _hasUnicode = hasUnicode;

  /**
   * Converts an ASCII `string` to an array.
   *
   * @private
   * @param {string} string The string to convert.
   * @returns {Array} Returns the converted array.
   */
  function asciiToArray(string) {
    return string.split('');
  }

  var _asciiToArray = asciiToArray;

  /** Used to compose unicode character classes. */
  var rsAstralRange$1 = '\\ud800-\\udfff',
      rsComboMarksRange$1 = '\\u0300-\\u036f',
      reComboHalfMarksRange$1 = '\\ufe20-\\ufe2f',
      rsComboSymbolsRange$1 = '\\u20d0-\\u20ff',
      rsComboRange$1 = rsComboMarksRange$1 + reComboHalfMarksRange$1 + rsComboSymbolsRange$1,
      rsVarRange$1 = '\\ufe0e\\ufe0f';

  /** Used to compose unicode capture groups. */
  var rsAstral = '[' + rsAstralRange$1 + ']',
      rsCombo = '[' + rsComboRange$1 + ']',
      rsFitz = '\\ud83c[\\udffb-\\udfff]',
      rsModifier = '(?:' + rsCombo + '|' + rsFitz + ')',
      rsNonAstral = '[^' + rsAstralRange$1 + ']',
      rsRegional = '(?:\\ud83c[\\udde6-\\uddff]){2}',
      rsSurrPair = '[\\ud800-\\udbff][\\udc00-\\udfff]',
      rsZWJ$1 = '\\u200d';

  /** Used to compose unicode regexes. */
  var reOptMod = rsModifier + '?',
      rsOptVar = '[' + rsVarRange$1 + ']?',
      rsOptJoin = '(?:' + rsZWJ$1 + '(?:' + [rsNonAstral, rsRegional, rsSurrPair].join('|') + ')' + rsOptVar + reOptMod + ')*',
      rsSeq = rsOptVar + reOptMod + rsOptJoin,
      rsSymbol = '(?:' + [rsNonAstral + rsCombo + '?', rsCombo, rsRegional, rsSurrPair, rsAstral].join('|') + ')';

  /** Used to match [string symbols](https://mathiasbynens.be/notes/javascript-unicode). */
  var reUnicode = RegExp(rsFitz + '(?=' + rsFitz + ')|' + rsSymbol + rsSeq, 'g');

  /**
   * Converts a Unicode `string` to an array.
   *
   * @private
   * @param {string} string The string to convert.
   * @returns {Array} Returns the converted array.
   */
  function unicodeToArray(string) {
    return string.match(reUnicode) || [];
  }

  var _unicodeToArray = unicodeToArray;

  /**
   * Converts `string` to an array.
   *
   * @private
   * @param {string} string The string to convert.
   * @returns {Array} Returns the converted array.
   */
  function stringToArray(string) {
    return _hasUnicode(string)
      ? _unicodeToArray(string)
      : _asciiToArray(string);
  }

  var _stringToArray = stringToArray;

  /**
   * Creates a function like `_.lowerFirst`.
   *
   * @private
   * @param {string} methodName The name of the `String` case method to use.
   * @returns {Function} Returns the new case function.
   */
  function createCaseFirst(methodName) {
    return function(string) {
      string = toString_1(string);

      var strSymbols = _hasUnicode(string)
        ? _stringToArray(string)
        : undefined;

      var chr = strSymbols
        ? strSymbols[0]
        : string.charAt(0);

      var trailing = strSymbols
        ? _castSlice(strSymbols, 1).join('')
        : string.slice(1);

      return chr[methodName]() + trailing;
    };
  }

  var _createCaseFirst = createCaseFirst;

  /**
   * Converts the first character of `string` to upper case.
   *
   * @static
   * @memberOf _
   * @since 4.0.0
   * @category String
   * @param {string} [string=''] The string to convert.
   * @returns {string} Returns the converted string.
   * @example
   *
   * _.upperFirst('fred');
   * // => 'Fred'
   *
   * _.upperFirst('FRED');
   * // => 'FRED'
   */
  var upperFirst = _createCaseFirst('toUpperCase');

  var upperFirst_1 = upperFirst;

  /**
   * Converts the first character of `string` to upper case and the remaining
   * to lower case.
   *
   * @static
   * @memberOf _
   * @since 3.0.0
   * @category String
   * @param {string} [string=''] The string to capitalize.
   * @returns {string} Returns the capitalized string.
   * @example
   *
   * _.capitalize('FRED');
   * // => 'Fred'
   */
  function capitalize(string) {
    return upperFirst_1(toString_1(string).toLowerCase());
  }

  var capitalize_1 = capitalize;

  /**
   * A specialized version of `_.reduce` for arrays without support for
   * iteratee shorthands.
   *
   * @private
   * @param {Array} [array] The array to iterate over.
   * @param {Function} iteratee The function invoked per iteration.
   * @param {*} [accumulator] The initial value.
   * @param {boolean} [initAccum] Specify using the first element of `array` as
   *  the initial value.
   * @returns {*} Returns the accumulated value.
   */
  function arrayReduce(array, iteratee, accumulator, initAccum) {
    var index = -1,
        length = array == null ? 0 : array.length;

    if (initAccum && length) {
      accumulator = array[++index];
    }
    while (++index < length) {
      accumulator = iteratee(accumulator, array[index], index, array);
    }
    return accumulator;
  }

  var _arrayReduce = arrayReduce;

  /**
   * The base implementation of `_.propertyOf` without support for deep paths.
   *
   * @private
   * @param {Object} object The object to query.
   * @returns {Function} Returns the new accessor function.
   */
  function basePropertyOf(object) {
    return function(key) {
      return object == null ? undefined : object[key];
    };
  }

  var _basePropertyOf = basePropertyOf;

  /** Used to map Latin Unicode letters to basic Latin letters. */
  var deburredLetters = {
    // Latin-1 Supplement block.
    '\xc0': 'A',  '\xc1': 'A', '\xc2': 'A', '\xc3': 'A', '\xc4': 'A', '\xc5': 'A',
    '\xe0': 'a',  '\xe1': 'a', '\xe2': 'a', '\xe3': 'a', '\xe4': 'a', '\xe5': 'a',
    '\xc7': 'C',  '\xe7': 'c',
    '\xd0': 'D',  '\xf0': 'd',
    '\xc8': 'E',  '\xc9': 'E', '\xca': 'E', '\xcb': 'E',
    '\xe8': 'e',  '\xe9': 'e', '\xea': 'e', '\xeb': 'e',
    '\xcc': 'I',  '\xcd': 'I', '\xce': 'I', '\xcf': 'I',
    '\xec': 'i',  '\xed': 'i', '\xee': 'i', '\xef': 'i',
    '\xd1': 'N',  '\xf1': 'n',
    '\xd2': 'O',  '\xd3': 'O', '\xd4': 'O', '\xd5': 'O', '\xd6': 'O', '\xd8': 'O',
    '\xf2': 'o',  '\xf3': 'o', '\xf4': 'o', '\xf5': 'o', '\xf6': 'o', '\xf8': 'o',
    '\xd9': 'U',  '\xda': 'U', '\xdb': 'U', '\xdc': 'U',
    '\xf9': 'u',  '\xfa': 'u', '\xfb': 'u', '\xfc': 'u',
    '\xdd': 'Y',  '\xfd': 'y', '\xff': 'y',
    '\xc6': 'Ae', '\xe6': 'ae',
    '\xde': 'Th', '\xfe': 'th',
    '\xdf': 'ss',
    // Latin Extended-A block.
    '\u0100': 'A',  '\u0102': 'A', '\u0104': 'A',
    '\u0101': 'a',  '\u0103': 'a', '\u0105': 'a',
    '\u0106': 'C',  '\u0108': 'C', '\u010a': 'C', '\u010c': 'C',
    '\u0107': 'c',  '\u0109': 'c', '\u010b': 'c', '\u010d': 'c',
    '\u010e': 'D',  '\u0110': 'D', '\u010f': 'd', '\u0111': 'd',
    '\u0112': 'E',  '\u0114': 'E', '\u0116': 'E', '\u0118': 'E', '\u011a': 'E',
    '\u0113': 'e',  '\u0115': 'e', '\u0117': 'e', '\u0119': 'e', '\u011b': 'e',
    '\u011c': 'G',  '\u011e': 'G', '\u0120': 'G', '\u0122': 'G',
    '\u011d': 'g',  '\u011f': 'g', '\u0121': 'g', '\u0123': 'g',
    '\u0124': 'H',  '\u0126': 'H', '\u0125': 'h', '\u0127': 'h',
    '\u0128': 'I',  '\u012a': 'I', '\u012c': 'I', '\u012e': 'I', '\u0130': 'I',
    '\u0129': 'i',  '\u012b': 'i', '\u012d': 'i', '\u012f': 'i', '\u0131': 'i',
    '\u0134': 'J',  '\u0135': 'j',
    '\u0136': 'K',  '\u0137': 'k', '\u0138': 'k',
    '\u0139': 'L',  '\u013b': 'L', '\u013d': 'L', '\u013f': 'L', '\u0141': 'L',
    '\u013a': 'l',  '\u013c': 'l', '\u013e': 'l', '\u0140': 'l', '\u0142': 'l',
    '\u0143': 'N',  '\u0145': 'N', '\u0147': 'N', '\u014a': 'N',
    '\u0144': 'n',  '\u0146': 'n', '\u0148': 'n', '\u014b': 'n',
    '\u014c': 'O',  '\u014e': 'O', '\u0150': 'O',
    '\u014d': 'o',  '\u014f': 'o', '\u0151': 'o',
    '\u0154': 'R',  '\u0156': 'R', '\u0158': 'R',
    '\u0155': 'r',  '\u0157': 'r', '\u0159': 'r',
    '\u015a': 'S',  '\u015c': 'S', '\u015e': 'S', '\u0160': 'S',
    '\u015b': 's',  '\u015d': 's', '\u015f': 's', '\u0161': 's',
    '\u0162': 'T',  '\u0164': 'T', '\u0166': 'T',
    '\u0163': 't',  '\u0165': 't', '\u0167': 't',
    '\u0168': 'U',  '\u016a': 'U', '\u016c': 'U', '\u016e': 'U', '\u0170': 'U', '\u0172': 'U',
    '\u0169': 'u',  '\u016b': 'u', '\u016d': 'u', '\u016f': 'u', '\u0171': 'u', '\u0173': 'u',
    '\u0174': 'W',  '\u0175': 'w',
    '\u0176': 'Y',  '\u0177': 'y', '\u0178': 'Y',
    '\u0179': 'Z',  '\u017b': 'Z', '\u017d': 'Z',
    '\u017a': 'z',  '\u017c': 'z', '\u017e': 'z',
    '\u0132': 'IJ', '\u0133': 'ij',
    '\u0152': 'Oe', '\u0153': 'oe',
    '\u0149': "'n", '\u017f': 's'
  };

  /**
   * Used by `_.deburr` to convert Latin-1 Supplement and Latin Extended-A
   * letters to basic Latin letters.
   *
   * @private
   * @param {string} letter The matched letter to deburr.
   * @returns {string} Returns the deburred letter.
   */
  var deburrLetter = _basePropertyOf(deburredLetters);

  var _deburrLetter = deburrLetter;

  /** Used to match Latin Unicode letters (excluding mathematical operators). */
  var reLatin = /[\xc0-\xd6\xd8-\xf6\xf8-\xff\u0100-\u017f]/g;

  /** Used to compose unicode character classes. */
  var rsComboMarksRange$2 = '\\u0300-\\u036f',
      reComboHalfMarksRange$2 = '\\ufe20-\\ufe2f',
      rsComboSymbolsRange$2 = '\\u20d0-\\u20ff',
      rsComboRange$2 = rsComboMarksRange$2 + reComboHalfMarksRange$2 + rsComboSymbolsRange$2;

  /** Used to compose unicode capture groups. */
  var rsCombo$1 = '[' + rsComboRange$2 + ']';

  /**
   * Used to match [combining diacritical marks](https://en.wikipedia.org/wiki/Combining_Diacritical_Marks) and
   * [combining diacritical marks for symbols](https://en.wikipedia.org/wiki/Combining_Diacritical_Marks_for_Symbols).
   */
  var reComboMark = RegExp(rsCombo$1, 'g');

  /**
   * Deburrs `string` by converting
   * [Latin-1 Supplement](https://en.wikipedia.org/wiki/Latin-1_Supplement_(Unicode_block)#Character_table)
   * and [Latin Extended-A](https://en.wikipedia.org/wiki/Latin_Extended-A)
   * letters to basic Latin letters and removing
   * [combining diacritical marks](https://en.wikipedia.org/wiki/Combining_Diacritical_Marks).
   *
   * @static
   * @memberOf _
   * @since 3.0.0
   * @category String
   * @param {string} [string=''] The string to deburr.
   * @returns {string} Returns the deburred string.
   * @example
   *
   * _.deburr('déjà vu');
   * // => 'deja vu'
   */
  function deburr(string) {
    string = toString_1(string);
    return string && string.replace(reLatin, _deburrLetter).replace(reComboMark, '');
  }

  var deburr_1 = deburr;

  /** Used to match words composed of alphanumeric characters. */
  var reAsciiWord = /[^\x00-\x2f\x3a-\x40\x5b-\x60\x7b-\x7f]+/g;

  /**
   * Splits an ASCII `string` into an array of its words.
   *
   * @private
   * @param {string} The string to inspect.
   * @returns {Array} Returns the words of `string`.
   */
  function asciiWords(string) {
    return string.match(reAsciiWord) || [];
  }

  var _asciiWords = asciiWords;

  /** Used to detect strings that need a more robust regexp to match words. */
  var reHasUnicodeWord = /[a-z][A-Z]|[A-Z]{2}[a-z]|[0-9][a-zA-Z]|[a-zA-Z][0-9]|[^a-zA-Z0-9 ]/;

  /**
   * Checks if `string` contains a word composed of Unicode symbols.
   *
   * @private
   * @param {string} string The string to inspect.
   * @returns {boolean} Returns `true` if a word is found, else `false`.
   */
  function hasUnicodeWord(string) {
    return reHasUnicodeWord.test(string);
  }

  var _hasUnicodeWord = hasUnicodeWord;

  /** Used to compose unicode character classes. */
  var rsAstralRange$2 = '\\ud800-\\udfff',
      rsComboMarksRange$3 = '\\u0300-\\u036f',
      reComboHalfMarksRange$3 = '\\ufe20-\\ufe2f',
      rsComboSymbolsRange$3 = '\\u20d0-\\u20ff',
      rsComboRange$3 = rsComboMarksRange$3 + reComboHalfMarksRange$3 + rsComboSymbolsRange$3,
      rsDingbatRange = '\\u2700-\\u27bf',
      rsLowerRange = 'a-z\\xdf-\\xf6\\xf8-\\xff',
      rsMathOpRange = '\\xac\\xb1\\xd7\\xf7',
      rsNonCharRange = '\\x00-\\x2f\\x3a-\\x40\\x5b-\\x60\\x7b-\\xbf',
      rsPunctuationRange = '\\u2000-\\u206f',
      rsSpaceRange = ' \\t\\x0b\\f\\xa0\\ufeff\\n\\r\\u2028\\u2029\\u1680\\u180e\\u2000\\u2001\\u2002\\u2003\\u2004\\u2005\\u2006\\u2007\\u2008\\u2009\\u200a\\u202f\\u205f\\u3000',
      rsUpperRange = 'A-Z\\xc0-\\xd6\\xd8-\\xde',
      rsVarRange$2 = '\\ufe0e\\ufe0f',
      rsBreakRange = rsMathOpRange + rsNonCharRange + rsPunctuationRange + rsSpaceRange;

  /** Used to compose unicode capture groups. */
  var rsApos = "['\u2019]",
      rsBreak = '[' + rsBreakRange + ']',
      rsCombo$2 = '[' + rsComboRange$3 + ']',
      rsDigits = '\\d+',
      rsDingbat = '[' + rsDingbatRange + ']',
      rsLower = '[' + rsLowerRange + ']',
      rsMisc = '[^' + rsAstralRange$2 + rsBreakRange + rsDigits + rsDingbatRange + rsLowerRange + rsUpperRange + ']',
      rsFitz$1 = '\\ud83c[\\udffb-\\udfff]',
      rsModifier$1 = '(?:' + rsCombo$2 + '|' + rsFitz$1 + ')',
      rsNonAstral$1 = '[^' + rsAstralRange$2 + ']',
      rsRegional$1 = '(?:\\ud83c[\\udde6-\\uddff]){2}',
      rsSurrPair$1 = '[\\ud800-\\udbff][\\udc00-\\udfff]',
      rsUpper = '[' + rsUpperRange + ']',
      rsZWJ$2 = '\\u200d';

  /** Used to compose unicode regexes. */
  var rsMiscLower = '(?:' + rsLower + '|' + rsMisc + ')',
      rsMiscUpper = '(?:' + rsUpper + '|' + rsMisc + ')',
      rsOptContrLower = '(?:' + rsApos + '(?:d|ll|m|re|s|t|ve))?',
      rsOptContrUpper = '(?:' + rsApos + '(?:D|LL|M|RE|S|T|VE))?',
      reOptMod$1 = rsModifier$1 + '?',
      rsOptVar$1 = '[' + rsVarRange$2 + ']?',
      rsOptJoin$1 = '(?:' + rsZWJ$2 + '(?:' + [rsNonAstral$1, rsRegional$1, rsSurrPair$1].join('|') + ')' + rsOptVar$1 + reOptMod$1 + ')*',
      rsOrdLower = '\\d*(?:1st|2nd|3rd|(?![123])\\dth)(?=\\b|[A-Z_])',
      rsOrdUpper = '\\d*(?:1ST|2ND|3RD|(?![123])\\dTH)(?=\\b|[a-z_])',
      rsSeq$1 = rsOptVar$1 + reOptMod$1 + rsOptJoin$1,
      rsEmoji = '(?:' + [rsDingbat, rsRegional$1, rsSurrPair$1].join('|') + ')' + rsSeq$1;

  /** Used to match complex or compound words. */
  var reUnicodeWord = RegExp([
    rsUpper + '?' + rsLower + '+' + rsOptContrLower + '(?=' + [rsBreak, rsUpper, '$'].join('|') + ')',
    rsMiscUpper + '+' + rsOptContrUpper + '(?=' + [rsBreak, rsUpper + rsMiscLower, '$'].join('|') + ')',
    rsUpper + '?' + rsMiscLower + '+' + rsOptContrLower,
    rsUpper + '+' + rsOptContrUpper,
    rsOrdUpper,
    rsOrdLower,
    rsDigits,
    rsEmoji
  ].join('|'), 'g');

  /**
   * Splits a Unicode `string` into an array of its words.
   *
   * @private
   * @param {string} The string to inspect.
   * @returns {Array} Returns the words of `string`.
   */
  function unicodeWords(string) {
    return string.match(reUnicodeWord) || [];
  }

  var _unicodeWords = unicodeWords;

  /**
   * Splits `string` into an array of its words.
   *
   * @static
   * @memberOf _
   * @since 3.0.0
   * @category String
   * @param {string} [string=''] The string to inspect.
   * @param {RegExp|string} [pattern] The pattern to match words.
   * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
   * @returns {Array} Returns the words of `string`.
   * @example
   *
   * _.words('fred, barney, & pebbles');
   * // => ['fred', 'barney', 'pebbles']
   *
   * _.words('fred, barney, & pebbles', /[^, ]+/g);
   * // => ['fred', 'barney', '&', 'pebbles']
   */
  function words(string, pattern, guard) {
    string = toString_1(string);
    pattern = guard ? undefined : pattern;

    if (pattern === undefined) {
      return _hasUnicodeWord(string) ? _unicodeWords(string) : _asciiWords(string);
    }
    return string.match(pattern) || [];
  }

  var words_1 = words;

  /** Used to compose unicode capture groups. */
  var rsApos$1 = "['\u2019]";

  /** Used to match apostrophes. */
  var reApos = RegExp(rsApos$1, 'g');

  /**
   * Creates a function like `_.camelCase`.
   *
   * @private
   * @param {Function} callback The function to combine each word.
   * @returns {Function} Returns the new compounder function.
   */
  function createCompounder(callback) {
    return function(string) {
      return _arrayReduce(words_1(deburr_1(string).replace(reApos, '')), callback, '');
    };
  }

  var _createCompounder = createCompounder;

  /**
   * Converts `string` to [camel case](https://en.wikipedia.org/wiki/CamelCase).
   *
   * @static
   * @memberOf _
   * @since 3.0.0
   * @category String
   * @param {string} [string=''] The string to convert.
   * @returns {string} Returns the camel cased string.
   * @example
   *
   * _.camelCase('Foo Bar');
   * // => 'fooBar'
   *
   * _.camelCase('--foo-bar--');
   * // => 'fooBar'
   *
   * _.camelCase('__FOO_BAR__');
   * // => 'fooBar'
   */
  var camelCase = _createCompounder(function(result, word, index) {
    word = word.toLowerCase();
    return result + (index ? capitalize_1(word) : word);
  });

  var camelcase = camelCase;

  /**
   * Converts `string` to
   * [kebab case](https://en.wikipedia.org/wiki/Letter_case#Special_case_styles).
   *
   * @static
   * @memberOf _
   * @since 3.0.0
   * @category String
   * @param {string} [string=''] The string to convert.
   * @returns {string} Returns the kebab cased string.
   * @example
   *
   * _.kebabCase('Foo Bar');
   * // => 'foo-bar'
   *
   * _.kebabCase('fooBar');
   * // => 'foo-bar'
   *
   * _.kebabCase('__FOO_BAR__');
   * // => 'foo-bar'
   */
  var kebabCase = _createCompounder(function(result, word, index) {
    return result + (index ? '-' : '') + word.toLowerCase();
  });

  var kebabcase = kebabCase;

  const GLOBAL_REGISTRY = '__riot_registry__';
  const TMP_TAG_NAME_VARIABLE = '__CURRENT_RIOT_TAG_NAME__';
  window[GLOBAL_REGISTRY] = {};

  // evaluates a compiled tag within the global context
  function globalEval(js, url) {
    const node = document.createElement('script');
    const root = document.documentElement;

    // make the source available in the "(no domain)" tab
    // of Chrome DevTools, with a .js extension
    if (url) js += `\n//# sourceURL=${url}.js`;

    node.text = js;
    root.appendChild(node);
    root.removeChild(node);
  }

  registerPostprocessor(async function(code, { tagName }){
    // cheap transpilation
    return {
      code: `${TMP_TAG_NAME_VARIABLE}=${tagName};(function (global){${code}})(window)`
        .replace('export default',
          `global['${GLOBAL_REGISTRY}']['${camelcase(tagName)}'] =`
        ),
      map: {}
    }
  });

  async function compileFromUrl(url) {
    const response = await fetch(url);
    const code = await response.text();

    return await compile(code, { file: url })
  }

  async function compileFromString(string, options) {
    return await compile(string, options)
  }

  async function compile$1() {
    const scripts = $$('script[type="riot"]');
    const urls = scripts.map(s => getAttribute(s, 'src') || getAttribute(s, 'data-src'));
    const tags = await Promise.all(urls.map(compileFromUrl));

    tags.forEach(({code}, i) => {
      const url = urls[i];
      const tagNameRe = new RegExp(`${TMP_TAG_NAME_VARIABLE}=((.*?);)`);
      const tagName = tagNameRe.exec(code)[1];

      globalEval(code.replace(tagNameRe, ''), url);
      register(kebabcase(tagName), window[GLOBAL_REGISTRY][camelcase(tagName)]);
    });
  }

  var riot_compiler = {
    ...riot,
    compile: compile$1,
    compileFromString,
    compileFromUrl
  };

  return riot_compiler;

}));
